Package: gseries
Title: Improve the Coherence of Your Time Series Data
Version: 3.0.3
Authors@R: c(
    person("Michel", "Ferland", email = "michel.ferland@statcan.gc.ca", 
    role = c("aut", "cre")),
    person("Statistics Canada", role = c("cph", "fnd"))
    )
Description: 
    'R' version of 'G-Series', Statistics Canada's generalized system devoted
    to the benchmarking and reconciliation of time series data. The methods
    used in 'G-Series' essentially come from Dagum, E. B., and P. Cholette 
    (2006) <doi:10.1007/0-387-35439-5>.
License: GPL (>= 3)
URL: https://StatCan.github.io/gensol-gseries/en/,
        https://StatCan.github.io/gensol-gseries/fr/
BugReports: https://github.com/StatCan/gensol-gseries/issues/
Depends: R (>= 4.0)
Imports: ggplot2, ggtext, graphics, grDevices, gridExtra, lifecycle,
        osqp (>= 1.0), rlang (>= 1.1.0), stats, utils, xmpdf
Suggests: knitr, rmarkdown, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
LazyData: true
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-09-01 15:09:00 UTC; ferlmic
Author: Michel Ferland [aut, cre],
  Statistics Canada [cph, fnd]
Maintainer: Michel Ferland <michel.ferland@statcan.gc.ca>
Repository: CRAN
Date/Publication: 2026-09-01 16:30:02 UTC

# Generated by roxygen2: do not edit by hand

export(GR_plot)
export(GR_table)
export(adj_plot)
export(benchmarking)
export(build_balancing_problem)
export(build_raking_problem)
export(gs.build_proc_grps)
export(gs.gInv_MP)
export(gs.time2per)
export(gs.time2str)
export(gs.time2year)
export(macro_gseriestsbalancing)
export(ori_plot)
export(plot_benchAdj)
export(plot_graphTable)
export(proc_benchmarking)
export(proc_tsraking)
export(rkMeta_to_blSpecs)
export(stack_bmkDF)
export(stack_tsDF)
export(stock_benchmarking)
export(tsDF_to_ts)
export(ts_to_bmkDF)
export(ts_to_tsDF)
export(tsbalancing)
export(tsraking)
export(tsraking_driver)
export(unstack_tsDF)
importFrom(lifecycle,deprecated)
importFrom(rlang,.data)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# G-Series in R (package gseries)

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/gseries)](https://cran.r-project.org/package=gseries)
[![Lifecycle:
stable](man/figures/lifecycle-stable.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![R-CMD-check
status](https://github.com/StatCan/gensol-gseries/actions/workflows/R-CMD-check.yaml/badge.svg?branch=main)](https://github.com/StatCan/gensol-gseries/actions/workflows/R-CMD-check.yaml?query=branch%3Amain)
[![codecov](https://codecov.io/gh/StatCan/gensol-gseries/branch/main/graph/badge.svg?token=ZUL7LPM7EV)](https://app.codecov.io/gh/StatCan/gensol-gseries?branch=main)
[![Mentioned in Awesome Official
Statistics](https://awesome.re/mentioned-badge.svg)](https://github.com/SNStatComp/awesome-official-statistics-software)

<!-- badges: end -->

([Français](#g-s%C3%A9ries-en-r-librairie-gseries))

## Description

R version of Statistics Canada’s (StatCan) generalized system
**G‑Series** initially developed in SAS<sup>®</sup>. This project is
devoted to G‑Series in R (package gseries). Email us at
<g-series@statcan.gc.ca> for information about the SAS<sup>®</sup>
versions.

> ***Note - StatCan intranet***  
> StatCan employees can also visit the G‑Series Confluence page on the
> agency’s intranet (search for “G-Series \| G-Séries” in Confluence) as
> well as the G‑Series GitLab development project also hosted on the
> agency’s intranet (search for “G-Series in R - G-Séries en R” in
> GitLab). The latter includes information and instructions specific to
> the StatCan IT infrastructure (e.g., Artifactory and GitLab); see file
> `index_StatCan.md` in the GitLab project root folder.

G‑Series 3.0 (package gseries 3.0.3) is the initial open-source version
of the software. It includes the rewriting in R of all SAS<sup>®</sup>
G‑Series 2.0 functionalities, that is PROC BENCHMARKING, PROC TSRAKING
and macro ***GSeriesTSBalancing*** along with a function for
benchmarking *stocks* using a spline interpolation approach where the
spline knots correspond to the benchmark-to-indicator ratios or
differences.

<https://StatCan.github.io/gensol-gseries/en/> (the gseries package
website) is your *go-to* place for everything there is to know about
this package:

- installation instructions (home page);
- complete function documentation with examples (**Reference** page);
- package vignettes containing complementary documentation (**Articles**
  menu);
- G‑Series version *changelog* (**News** \> **Changelog** page);
- how to download local copies of the G‑Series HTML and PDF
  documentation (home page);
- etc.

## Project Structure

With the exception of the following directories, the files in this
project follow the common structure of a R package with a pkgdown
website:

- `docs/` (\*): bilingual pkgdown website files (usually built in a
  GitHub Actions workflow or a GitLab CI/CD pipeline rather than being
  saved with the project files).
- `fr/`: files required to generate the French version of the complete
  HTML and PDF package documentation (function reference, vignettes,
  etc.). The contents of the `fr/` directory resemble those of the root
  project directory.
- `misc/`: various files related to the package maintenance.
- `pdf/` (\*): bilingual package documentation and vignettes in PDF
  format.

(\*) The purpose of directories `docs/` and `pdf/` is to provide access
to the complete library of available G‑Series documentation in HTML and
PDF formats, as opposed to only the CRAN version for the package website
or the installed version through R and RStudio. So whether one is
interested in the documentation for a previous version of G‑Series
(*tag* \< CRAN version) or for the development version (when *main*
branch version \> CRAN version), the contents of these folders can be
downloaded for offline consultation.

## Contact - Support

User support for G‑Series is provided by the Time Series Research and
Analysis Centre (TSRAC) in the Economic Statistics Methods Division
(ESMD) and the Digital Processing Solutions Division (DPSD). Email us at
<g-series@statcan.gc.ca> for information or help using G‑Series. GitHub
account holders can also request information, ask questions or report
problems through the G‑Series GitHub project
[Issues](https://github.com/StatCan/gensol-gseries/issues) page. StatCan
employees can do the same through the **Issues** page of the G‑Series
GitLab development project hosted on the StatCan intranet (search for
“G-Series in R - G-Séries en R” in GitLab).

<br>

------------------------------------------------------------------------

<br>

# G-Séries en R (librairie gseries)

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/gseries)](https://cran.r-project.org/package=gseries)
[![Lifecycle:
stable](man/figures/lifecycle-stable.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![R-CMD-check
status](https://github.com/StatCan/gensol-gseries/actions/workflows/R-CMD-check.yaml/badge.svg?branch=main)](https://github.com/StatCan/gensol-gseries/actions/workflows/R-CMD-check.yaml?query=branch%3Amain)
[![codecov](https://codecov.io/gh/StatCan/gensol-gseries/branch/main/graph/badge.svg?token=ZUL7LPM7EV)](https://app.codecov.io/gh/StatCan/gensol-gseries?branch=main)
[![Mentioned in Awesome Official
Statistics](https://awesome.re/mentioned-badge.svg)](https://github.com/SNStatComp/awesome-official-statistics-software)

<!-- badges: end -->

([English](#g-series-in-r-package-gseries))

## Description

Version R du système généralisé de Statistique Canada (StatCan)
**G‑Séries** initialement développé en SAS<sup>®</sup>. Ce projet est
consacré à la version R de G‑Séries (librairie gseries). Écrivez-nous à
<g-series@statcan.gc.ca> pour obtenir des informations sur les versions
SAS<sup>®</sup>.

> ***Note - intranet de StatCan***  
> Les employés de StatCan peuvent également visiter la page Confluence
> de G‑Séries dans l’intranet de l’agence (cherchez « G-Series \|
> G-Séries » dans Confluence) ainsi que le projet de développement
> GitLab de G‑Séries également hébergé dans l’intranet de l’agence
> (cherchez « G-Series in R - G-Séries en R » dans GitLab). Ce dernier
> inclut des informations et instructions qui sont spécifiques à
> l’infrastructure TI de StatCan (ex., Artifactory et GitLab) ; voir le
> fichier `index_StatCan.md` dans le répertoire racine du projet GitLab.

G‑Séries 3.0 (librairie gseries 3.0.3) est la première version du
logiciel offerte en libre accès (logiciel libre). Elle inclut le
recodage en R de toutes les fonctionalités SAS<sup>®</sup> de G‑Séries
2.0, soit PROC BENCHMARKING, PROC TSRAKING et la macro
***GSeriesTSBalancing***, ainsi qu’une fonction pour l’étalonnage de
*séries de stocks* par l’entremise d’interpolations par spline cubique
où les noeuds de la spline correspondent aux ratios ou différences entre
les valeurs des étalons et de la série indicatrice.

<https://StatCan.github.io/gensol-gseries/fr/> (le site web de la
librairie gseries) est l’endroit de prédilection pour tout ce qu’il y a
à savoir sur la librairie :

- instructions d’installation (page d’accueil) ;
- documentation complète des fonctions avec des exemples (page
  **Référence**) ;
- vignettes de la librairie contenant de la documentation complémentaire
  (menu **Articles**) ;
- historique des versions de G‑Séries (page **Nouveautés** \>
  **Changements**) ;
- comment télécharger une copie locale de la documentation HTML et PDF
  de G‑Séries (page d’accueil) ;
- etc.

## Structure du projet

A l’exception des répertoires suivants, les fichiers de ce projet
suivent la structure commune (habituelle) d’une librairie R avec un site
web pkgdown :

- `docs/` (\*) : fichiers du site web pkgdown (habituellement construits
  dans un *GitHub Actions workflow* ou un *GitLab CI/CD pipeline* plutôt
  que d’être sauvegardés avec les fichiers du projet).
- `fr/` : fichiers nécessaires pour générer la version française de la
  documentation complète de la librairie gseries en format HTML et PDF
  (référence des fonctions, vignettes, etc.). Le contenu du répertoire
  `fr/` ressemble à celui du répertoire racine du projet.
- `misc/` : divers fichiers liés à la maintenance de la librairie.
- `pdf/` (\*) : documentation et vignettes de la librairie gseries en
  format PDF.

(\*) La raison d’être des répertoires `docs/` et `pdf/` est de fournir
un accès à la bibliothèque complète de la documentation disponible de
G‑Séries en format HTML et PDF, par opposition à uniquement la version
CRAN pour le site web de la librairie ou la version installée via R et
RStudio. Ainsi, que l’on soit intéressé par la documentation d’une
version précédente de G‑Séries (*tag* \< version CRAN) ou de la version
de développement (lorsque version branche *main* \> version CRAN), le
contenu de ces répertoires peut être téléchargé pour consultation hors
ligne.

## Contact - Assistance

L’assistance aux utilisateurs de G‑Séries assurée par le Centre de
recherche et d’analyse en séries chronologiques (CRASC) de la Division
des méthodes de la statistique économique (DMSE) et par la Division des
solutions de traitement numérique (DSTN). Écrivez-nous à
<g-series@statcan.gc.ca> pour obtenir des informations ou de l’aide sur
l’utilisation de G‑Séries. Les détenteurs d’un compte GitHub peuvent
également demander des informations, poser des questions ou signaler des
problèmes via la page
[*Issues*](https://github.com/StatCan/gensol-gseries/issues) du projet
GitHub de G‑Séries. Les employés de StatCan peuvent faire de même via la
page **Tickets** du projet de développement GitLab de G‑Séries hébergé
dans l’intranet de StatCan (recherchez « G-Series in R - G-Séries en R »
dans GitLab).

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, include = FALSE---------------------------------------------------
library(gseries)

## ----eval = FALSE-------------------------------------------------------------
# # Load G-Series in R (package gseries)
# library(gseries)
# 
# # Set the working directory (for the PDF graph files)
# iniwd <- getwd()
# setwd("C:/Temp")
# 
# 
# 
# 
# #### Example 1 (single quarterly series) ####
# # Simple case with a single quarterly series to benchmark to annual values
# 
# # Quarterly indicator series
# my_ts_qtr <- ts(c(1.9, 2.4, 3.1, 2.2, 2.0, 2.6, 3.4, 2.4, 2.3),
#                 start = c(2015, 1),
#                 frequency = 4)
# my_series1 <- ts_to_tsDF(my_ts_qtr)
# 
# # Annual benchmarks for quarterly data
# my_ts_ann <- ts(c(10.3, 10.2),
#                 start = 2015,
#                 frequency = 1)
# my_benchmarks1 <- ts_to_bmkDF(my_ts_ann, ind_frequency = 4)
# 
# # Benchmarking using...
# #   - recommended `rho` value for quarterly series (`rho = 0.729`)
# #   - proportional model (`lambda = 1`)
# #   - bias-corrected indicator series with the estimated bias (`biasOption = 3`)
# out_bench1 <- benchmarking(my_series1,
#                            my_benchmarks1,
#                            rho = 0.729,
#                            lambda = 1,
#                            biasOption = 3)
# 
# # Outputs
# View(out_bench1$series)
# View(out_bench1$benchmarks)
# View(out_bench1$graphTable)
# 
# # Graphics (default set of plots)
# plot_graphTable(out_bench1$graphTable,
#                 "Ex1_graphs.pdf")
# # With the G. R. Table
# plot_graphTable(out_bench1$graphTable,
#                 "Ex1_graphs_with_GRTable.pdf",
#                 GR_table_flag = TRUE)
# 
# 
# # Compare with package tempdisagg that allows Denton benchmarking
# # (modified version by Cholette)
# 
# #install.packages("tempdisagg")
# library(tempdisagg)
# #install.packages("dplyr")
# library(dplyr)
# 
# 
# # Proportional Denton
# mult_denton <- benchmarking(my_series1,
#                             my_benchmarks1,
#                             rho = 1,
#                             lambda = 1)
# 
# td_mult_denton <- td(formula = my_ts_ann ~ 0 + my_ts_qtr,
#                      method = "denton-cholette",
#                      criterion = "proportional")
# 
# compare_mult <- ts.union(gseries = tsDF_to_ts(mult_denton$series, frequency = 4),
#                          tempdisagg = predict(td_mult_denton),
#                          dframe = TRUE) %>%
#   cbind(., diff = .[, 1] - .[, 2])
# 
# # Same results
# View(compare_mult)
# 
# 
# # Additive Denton
# add_denton <- benchmarking(my_series1,
#                            my_benchmarks1,
#                            rho = 1,
#                            lambda = 0)
# 
# td_add_denton <- td(formula = my_ts_ann ~ 0 + my_ts_qtr,
#                     method = "denton-cholette",
#                     criterion = "additive")
# 
# compare_add <- ts.union(gseries = tsDF_to_ts(add_denton$series, frequency = 4),
#                         tempdisagg = predict(td_add_denton),
#                         dframe = TRUE) %>%
#   cbind(., diff = .[, 1] - .[, 2])
# 
# # Same results
# View(compare_add)
# 
# 
# 
# 
# #### Example 2 (multiple quarterly series) ####
# # Two quarterly series to benchmark to annual values,
# # with BY-groups and user-defined alterability coefficients
# 
# # Sales data (same sales for groups A and B; only alter coefs for van sales differ)
# qtr_sales <- ts(matrix(c(# Car sales
#   1851, 2436, 3115, 2205, 1987, 2635, 3435, 2361, 2183, 2822,
#   3664, 2550, 2342, 3001, 3779, 2538, 2363, 3090, 3807, 2631,
#   2601, 3063, 3961, 2774, 2476, 3083, 3864, 2773, 2489, 3082,
#   # Van sales
#   1900, 2200, 3000, 2000, 1900, 2500, 3800, 2500, 2100, 3100,
#   3650, 2950, 3300, 4000, 3290, 2600, 2010, 3600, 3500, 2100,
#   2050, 3500, 4290, 2800, 2770, 3080, 3100, 2800, 3100, 2860),
#   ncol = 2),
#   start = c(2011, 1),
#   frequency = 4,
#   names = c("car_sales", "van_sales"))
# 
# class(qtr_sales)
# 
# ann_sales <- ts(matrix(c(# Car sales
#   10324, 10200, 10582, 11097, 11582, 11092,
#   # Van sales
#   12000, 10400, 11550, 11400, 14500, 16000),
#   ncol = 2),
#   start = 2011,
#   frequency = 1,
#   names = c("car_sales", "van_sales"))
# 
# # Quarterly indicator series (with default alter coefs for now)
# my_series2 <- rbind(cbind(data.frame(group = rep("A", nrow(qtr_sales)),
#                                      alt_van = rep(1, nrow(qtr_sales))),
#                           ts_to_tsDF(qtr_sales)),
#                     cbind(data.frame(group = rep("B", nrow(qtr_sales)),
#                                      alt_van = rep(1, nrow(qtr_sales))),
#                           ts_to_tsDF(qtr_sales)))
# 
# # Set binding van sales (alter coef = 0) for 2012 Q1 and Q2 in group A (rows 5 and 6)
# my_series2$alt_van[c(5, 6)] <- 0
# 
# # Annual benchmarks for quarterly data (without alter coefs)
# my_benchmarks2 <- rbind(cbind(data.frame(group = rep("A", nrow(ann_sales))),
#                               ts_to_bmkDF(ann_sales, ind_frequency = 4)),
#                         cbind(data.frame(group = rep("B", nrow(ann_sales))),
#                               ts_to_bmkDF(ann_sales, ind_frequency = 4)))
# 
# # Benchmarking using...
# #   - recommended RHO value for quarterly series (rho = 0.729)
# #   - proportional model (lambda = 1)
# #   - without bias correction (biasOption = 1 and bias not specified)
# out_bench2 <- benchmarking(my_series2,
#                            my_benchmarks2,
#                            rho = 0.729,
#                            lambda = 1,
#                            biasOption = 1,
#                            var = c("car_sales", "van_sales / alt_van"),
#                            with = c("car_sales", "van_sales"),
#                            by = "group")
# 
# # Outputs
# View(out_bench2$series)
# View(out_bench2$benchmarks)
# View(out_bench2$graphTable)
# 
# # Graphics
# plot_graphTable(out_bench2$graphTable,
#                 "Ex2_graphs.pdf")
# 
# 
# 
# 
# #### Example 3 (multiple quarterly series) ####
# # Same as example 2, but benchmarking all 4 series as BY-groups
# # (4 BY-groups of 1 series instead of 2 BY-groups of 2 series)
# 
# my_series3 <- stack_tsDF(ts_to_tsDF(ts.union(A = qtr_sales, B = qtr_sales)))
# my_series3$alter <- 1
# my_series3$alter[my_series3$series == "A.van_sales" &
#                    my_series3$year == 2012 & my_series3$period <= 2] <- 0
# my_benchmarks3 <- stack_bmkDF(ts_to_bmkDF(ts.union(A = ann_sales, B = ann_sales),
#                                           ind_frequency = 4))
# out_bench3 <- benchmarking(my_series3,
#                            my_benchmarks3,
#                            rho = 0.729,
#                            lambda = 1,
#                            biasOption = 1,
#                            var = "value / alter",
#                            with = "value",
#                            by = "series")
# 
# # Outputs
# View(out_bench3$series)
# View(out_bench3$benchmarks)
# View(out_bench3$graphTable)
# 
# # Graphics
# plot_graphTable(out_bench3$graphTable,
#                 "Ex3_graphs.pdf")
# 
# # Convert the benchmarked series as a "mts" object
# my_out_series3 <- tsDF_to_ts(unstack_tsDF(out_bench3$series), frequency = 4)
# class(my_out_series3)
# plot(my_out_series3)
# 
# 
# 
# 
# #### Monthly data (Box & Jenkins airline series) ####
# 
# data(AirPassengers)
# my_AP_ind <- ts_to_tsDF(AirPassengers)
# 
# # Create annual benchmarks by changing the level (5 times larger), adding some random
# # noise and dropping the last 2 benchmarks
# set.seed(as.Date("2003-03-25"))  # for results reproducibility (select any date you want)
# #set.seed(NULL)
# ann_AP <- round(jitter(aggregate.ts(AirPassengers, nfrequency = 1, FUN = sum) * 5,
#                        amount = 2500))
# my_AP_bmk <- (ts_to_bmkDF(ann_AP, ind_frequency = 12))[1:10, ]
# 
# 
# # With bias correction (estimated bias with `biasOption = 3`)
# #   => everything looks good
# out_bench_AP <- benchmarking(my_AP_ind,
#                              my_AP_bmk,
#                              rho = 0.9,
#                              lambda = 1,
#                              biasOption = 3)
# View(out_bench_AP$series)
# View(out_bench_AP$benchmarks)
# View(out_bench_AP$graphTable)
# 
# plot_graphTable(out_bench_AP$graphTable,
#                 "AP_graphs.pdf")
# 
# 
# # Without bias correction (`biasOption = 1`)
# #   => issues with the projected adjustments at the end of the series for periods
# #      not covered by a benchmark
# out_bench_AP_noBias <- benchmarking(my_AP_ind,
#                                     my_AP_bmk,
#                                     rho = 0.9,
#                                     lambda = 1,
#                                     biasOption = 1)
# View(out_bench_AP_noBias$series)
# View(out_bench_AP_noBias$benchmarks)
# View(out_bench_AP_noBias$graphTable)
# 
# plot_graphTable(out_bench_AP_noBias$graphTable,
#                 "AP_graphs_noBias.pdf")
# 
# 
# # Denton benchmarking (`rho = 1`, bias correction is irrelevant)
# #   => last adjustment repeated (forever) at the end of the series
# #      (strong assumption, but not necessarily problematic)
# out_bench_AP_denton <- benchmarking(my_AP_ind,
#                                     my_AP_bmk,
#                                     rho = 1,
#                                     lambda = 1)
# View(out_bench_AP_denton$series)
# View(out_bench_AP_denton$benchmarks)
# View(out_bench_AP_denton$graphTable)
# 
# plot_graphTable(out_bench_AP_denton$graphTable,
#                 "AP_graphs_Denton.pdf")
# 
# 
# # Denton benchmarking approximation (`rho = 0.999`, `biasOption = 3`)
# #   => regression benchmarking model approximation of the "pure" Denton method
# #   => last adjustment repeated at the end of the series (mild convergence to the bias
# #      compared to the "pure" Denton method)
# out_bench_AP_dentonApprox <- benchmarking(my_AP_ind,
#                                           my_AP_bmk,
#                                           rho = 0.999,
#                                           lambda = 1,
#                                           biasOption = 3)
# View(out_bench_AP_dentonApprox$series)
# View(out_bench_AP_dentonApprox$benchmarks)
# View(out_bench_AP_dentonApprox$graphTable)
# 
# plot_graphTable(out_bench_AP_dentonApprox$graphTable,
#                 "AP_graphs_DentonApprox.pdf")
# 
# 
# # Pro-rating (`lambda = 0.5` and `rho = 0`)
# #   => no movement preservation: all adjustments are lumped into January every year
# #      (with a disastrous impact on some of the initial December to January movements)
# #   => immediate convergence to the bias (estimated with `biasOption = 3`) for the
# #      projected adjustments at the end of the series
# out_bench_AP_proRate <- benchmarking(my_AP_ind,
#                                      my_AP_bmk,
#                                      rho = 0,
#                                      lambda = 0.5,
#                                      biasOption = 3)
# View(out_bench_AP_proRate$series)
# View(out_bench_AP_proRate$benchmarks)
# View(out_bench_AP_proRate$graphTable)
# 
# plot_graphTable(out_bench_AP_proRate$graphTable,
#                 "AP_graphs_proRating.pdf")
# 
# 
# 
# 
# #### End of year stocks ("kinks" in the adjustments with `benchmarking()`) ####
# 
# # Quarterly indicator stock series (same pattern repeated every year)
# my_stock_ind <- ts_to_tsDF(ts(rep(c(85, 95, 125, 95), 7),
#                               start = c(2013, 1),
#                               frequency = 4))
# 
# # Annual benchmarks (end-of-year stocks)
# my_stock_bmk <- ts_to_bmkDF(ts(c(135, 125, 155, 145, 165),
#                                start = 2013,
#                                frequency = 1),
#                             discrete_flag = TRUE,
#                             alignment = "e",
#                             ind_frequency = 4)
# 
# # With `benchmarking()` ("Proc Benchmarking" approach)
# out_stock_PB <- benchmarking(my_stock_ind,
#                              my_stock_bmk,
#                              rho = 0.729,
#                              lambda = 1,
#                              biasOption = 3)
# 
# # With `stock_benchmarking()` ("Stock Benchmarking" approach)
# out_stock_SB <- stock_benchmarking(my_stock_ind,
#                                    my_stock_bmk,
#                                    rho = 0.729,
#                                    lambda = 1,
#                                    biasOption = 3)
# 
# # Benchmarking adjustments of both approaches
# plot_benchAdj(PB_graphTable = out_stock_PB$graphTable,
#               SB_graphTable = out_stock_SB$graphTable)
# 
# # Have you noticed how smoother the `stock_benchmarking()` adjustments are compared
# # to the `benchmarking()` ones?
# 
# # The gain in the quality of the resulting benchmarked stocks might not necessarily
# # be obvious in this example.
# plot(out_stock_SB$graphTable$t, out_stock_SB$graphTable$benchmarked,
#      type = "b", col = "red", xlab = "t", ylab = "Benchmarked Stock")
# lines(out_stock_PB$graphTable$t, out_stock_PB$graphTable$benchmarked,
#       type = "b", col = "blue")
# legend(x = "topleft", bty = "n", inset = 0.05, lty = 1, pch = 1,
#        col = c("red", "blue"), legend = c("out_stock_SB", "out_stock_PB"))
# title("Benchmarked Stock")
# 
# # PDF graphics
# plot_graphTable(out_stock_PB$graphTable, "Stock_graphs_PB.pdf")
# plot_graphTable(out_stock_SB$graphTable, "Stock_graphs_SB.pdf")
# 
# # What about cases where a flat indicator is used, which may happen in practice
# # in absence of a good indicator of the quarterly (sub-annual) movement?
# my_flat_ind <- my_stock_ind
# my_flat_ind$value <- 1
# out_stock_PB2 <- benchmarking(my_flat_ind,
#                               my_stock_bmk,
#                               rho = 0.729,
#                               lambda = 1,
#                               biasOption = 3)
# out_stock_SB2 <- stock_benchmarking(my_flat_ind,
#                                     my_stock_bmk,
#                                     rho = 0.729,
#                                     lambda = 1,
#                                     biasOption = 3)
# 
# plot(out_stock_SB2$graphTable$t, out_stock_SB2$graphTable$benchmarked,
#      type = "b", col = "red", xlab = "t", ylab = "Benchmarked Stock")
# lines(out_stock_PB2$graphTable$t, out_stock_PB2$graphTable$benchmarked,
#       type = "b", col = "blue")
# legend(x = "bottomright", bty = "n", inset = 0.05, lty = 1, pch = 1,
#        col = c("red", "blue"), legend = c("out_stock_SB2", "out_stock_PB2"))
# title("Benchmarked Stock - Flat Indicator")
# 
# # The awkwardness of the benchmarked stocks produced by `benchmarking()` suddenly
# # becomes obvious. That's because the benchmarked series corresponds to the
# # benchmarking adjustments when using a flat indicator (e.g., a series of 1's
# # with proportional benchmarking):
# plot_benchAdj(PB_graphTable = out_stock_PB2$graphTable,
#               SB_graphTable = out_stock_SB2$graphTable)
# 
# # The shortcomings of the "Proc Benchmarking" approach (function `benchmarking()`)
# # with stocks is also quite noticeable in this case when looking at the resulting
# # quarterly growth rates, which are conveniently produced by `plot_graphTable()`.
# # Pay particular attention to the transition in the growth rates from Q4 to Q1
# # every year in the generated PDF graphs.
# plot_graphTable(out_stock_PB2$graphTable, "Stock_graphs_PB_flat_indicator.pdf")
# plot_graphTable(out_stock_SB2$graphTable, "Stock_graphs_SB_flat_indicator.pdf")
# 
# 
# # Reset the working directory to its initial location
# setwd(iniwd)

######
# Objects and functions in this "benchmarking utility script" are prefixed with "bk."
# to make them easy to spot in the (regular) benchmarking scripts.


# The functions defined in this script are internal functions shared by `benchmarking()`
# and `stock_benchmarking()` that would otherwise need to be defined inside (duplicated in)
# both function scripts. Environment `bk.e` created below contains objects that are not (cannot be)
# passed as arguments to the shared internal benchmarking functions. Notes:
#   - Environment `bk.e` must be compiled before the benchmarking function scripts, hence 'aaa_'
#     in the current script name.
#   - Although environment `bk.e` is primarily used for the shared internal benchmarking functions,
#     it can be used, if needed, by any package function to store objects.
#   - Large objects created in environment `bk.e` that are not needed after the function execution
#     should be removed at the end of the function execution with `rlang::env_unbind()`.
bk.e <- new.env(parent = emptyenv())


# Bias adjustment functions according to (main function) arguments 'lambda' and 'biasOption':
#   - add/mult bias estimation functions
#   - add/mult bias application (correction) functions
#   - bias option implementation functions
bk.calc_add_bias <- function(s_lowFreq, a, denom, ...) {
  sum(a - s_lowFreq) / denom
}
bk.calc_mult_bias <- function(s_lowFreq, a, tol = gs.tolerance, ...) {
  numer <- sum(a)
  denom <- sum(s_lowFreq)
  if (abs(denom) <= tol) {
    if (abs(numer) > tol) {
      warning("Bias calculation involves division by zero. Assuming no bias (1.0).\n",
              call. = FALSE, immediate. = TRUE)
      bk.e$warning_flag <- TRUE
    }
    1
  } else {
    numer / denom
  }
}
bk.apply_add_bias <- function(s, b) {
  s + b
}
bk.apply_mult_bias <- function(s, b) {
  s * b
}
bk.apply_biasOption1 <- function(msg_f, calc_f, s_lowFreq, a, denom, tol, apply_f, s, bias, str) {
  msg_f("BIAS = ", format(bias), " (", str, ")\n")
  bk.e$actual_bias <- bias  # package `bk.e` environment assignment
  apply_f(s, bias)  # `apply_f` is either `bk.apply_add_bias` or `bk.apply_mult_bias`
}
bk.apply_biasOption2 <- function(msg_f, calc_f, s_lowFreq, a, denom, tol, apply_f, s, bias, str) {
  b <- calc_f(s_lowFreq, a, denom = denom, tol = tol)  # `calc_f` is either `bk.calc_add_bias` or `bk.calc_mult_bias`
  # `denom` is `sum(J)` for PB and `length(a)` for SB
  msg_f("BIAS = ", format(bias), " (", str, ")")
  msg_f("BIAS = ", format(b), " (calculated, but NOT used)\n")
  bk.e$actual_bias <- bias  # package `bk.e` environment assignment
  apply_f(s, bias)  # `apply_f` is either `bk.apply_add_bias` or `bk.apply_mult_bias`
}
bk.apply_biasOption3 <- function(msg_f, calc_f, s_lowFreq, a, denom, tol, apply_f, s, ...) {
  b <- calc_f(s_lowFreq, a, denom = denom, tol = tol)  # `calc_f` is either `bk.calc_add_bias` or `bk.calc_mult_bias`
  # `denom` is `sum(J)` for PB and `length(a)` for SB
  msg_f("BIAS = ", format(b), " (calculated)\n")
  bk.e$actual_bias <- b  # package `bk.e` environment assignment
  apply_f(s, b)  # `apply_f` is either `bk.apply_add_bias` or `bk.apply_mult_bias`
}


# Binding benchmarks validation functions according to (main function) arguments 'tolV' and 'tolP':
#   - bk.binding_bmk_diff_valid   : tolV is specified (check differences against tolV)
#   - bk.binding_bmk_relDiff_valid: tolP is specified (check relative differences against tolP)
bk.binding_bmk_diff_valid <- function(a, theta_lowFreq, c_a, bmk_start, bmk_end, tol, ...) {
  discr <- abs(theta_lowFreq - a) * (c_a == 0)
  prob_id <- which(discr > tol)
  if (length(prob_id) > 0) {
    warning("The following binding benchmarks were not met (tolerance = ", format(tol), "):",
            paste0("\n  [", bmk_start[prob_id], ", ", bmk_end[prob_id], "]: difference = ",
                   format(discr[prob_id]), collapse = ""),
            "\n", call. = FALSE, immediate. = TRUE)
    bk.e$warning_flag <- TRUE
  }
  invisible(NULL)
}
bk.binding_bmk_relDiff_valid <- function(a, theta_lowFreq, c_a, bmk_start, bmk_end, tol, zero_tol = gs.tolerance) {
  a[abs(a) <= zero_tol] <- NA
  discr <- abs(theta_lowFreq / a - 1) * 100 * (c_a == 0)
  tol <- tol * 100
  prob_id <- which(discr > tol)
  if (length(prob_id) > 0) {
    warning("The following binding benchmarks were not met (tolerance = ", format(tol), "%):",
            paste0("\n  [", bmk_start[prob_id], ", ", bmk_end[prob_id], "]: relative difference = ",
                   format(discr[prob_id]), "%", collapse = ""),
            "\n", call. = FALSE, immediate. = TRUE)
    bk.e$warning_flag <- TRUE
  }
  invisible(NULL)
}


# By-group initialization functions:
#   - bk.noByGrp_ini: argument 'by' (main function) is empty
#   - bk.byGrp_ini  : otherwise
# => They return nothing
bk.noByGrp_ini <- function(series_df, benchmarks_df, ...) {
  bk.e$ser_df_byGrp <- series_df[order(series_df$year, series_df$period), ]
  bk.e$bmk_df_byGrp <- benchmarks_df[order(benchmarks_df$startYear, benchmarks_df$startPeriod,
                                           benchmarks_df$endYear, benchmarks_df$endPeriod), ]
  invisible(NULL)
}
bk.byGrp_ini <- function(series_df, benchmarks_df, by_grps, ii, by) {
  subset <- by_grps[ii, by, drop = FALSE]
  bk.e$ser_df_byGrp <- merge(subset, series_df, by = by)
  bk.e$ser_df_byGrp <- bk.e$ser_df_byGrp[order(bk.e$ser_df_byGrp$year, bk.e$ser_df_byGrp$period), ]
  bk.e$bmk_df_byGrp <- merge(subset, benchmarks_df, by = by)
  bk.e$bmk_df_byGrp <- bk.e$bmk_df_byGrp[order(bk.e$bmk_df_byGrp$startYear, bk.e$bmk_df_byGrp$startPeriod,
                                               bk.e$bmk_df_byGrp$endYear, bk.e$bmk_df_byGrp$endPeriod), ]
  invisible(NULL)
}


# Negative input data validation functions
# => return `TRUE` if the situation corresponds to an "error" and return `FALSE` otherwise
bk.check_neg_err <- function(x, tol = gs.tolerance, data_str = "input data") {
  if (gs.check_neg(x, tol = tol)) {

    # Negative values are NOT allowed and some were found (this is an "error")
    TRUE

  } else {

    # Negative values not found (this is NOT an "error")
    FALSE
  }
}
bk.check_neg_warn <- function(x, tol = gs.tolerance, data_str = "input data") {
  if (gs.check_neg(x, tol = tol)) {
    warning("Negative values found in the ", data_str, ". Suspicious use of proportional benchmarking.\n",
            call. = FALSE, immediate. = TRUE)
    bk.e$warning_flag <- TRUE
  }

  # Negative values are allowed (whether negative values are found or not, this is NOT an "error")
  FALSE
}

#' Helper function for [tsraking()]
#'
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/tsraking_driver.html}})}
#' 
#' Helper function for the [tsraking()] function that conveniently determines the required set of raking
#' problems to be solved and internally generates the individual calls to [tsraking()]. It is especially
#' useful in the context of temporal total (e.g., annual total) preservation where each individual raking
#' problem either involves a single period for incomplete temporal groups (e.g., incomplete years) or several
#' periods for complete temporal groups (e.g., the set of periods of a complete year).
#'
#'
#' @usage
#' tsraking_driver(
#'   in_ts,
#'   ...,  # `tsraking()` arguments excluding `data_df`
#'   temporal_grp_periodicity = 1,
#'   temporal_grp_start = 1
#' )
#'
#' 
#' @param in_ts (mandatory)
#'
#' Time series (object of class "ts" or "mts") that contains the time series data to be reconciled. 
#' They are the raking problems' input data (initial solutions).
#'
#' @inheritDotParams tsraking -data_df
#' 
#' @inheritParams tsbalancing
#' 
#' 
#' @details
#' This function solves one raking problem with [tsraking()] per processing group (see section **Processing groups** for 
#' details). The mathematical expression of these raking problem can be found in the **Details** section of the [tsraking()] 
#' documentation.
#' 
#' The alterability coefficients data frame (argument `alterability_df`) specified with [tsraking_driver()] can either 
#' contain:
#' * A single observation: the specified coefficients will be used for all periods of input time series object (argument 
#' `in_ts`).
#' * A number of observations equal to `frequency(in_ts)`: the specified coefficients will be used for the 
#' corresponding *cycle* of the input time series object (argument `in_ts`) periods. Monthly data example: 1<sup>st</sup> 
#' observation for January, 2<sup>nd</sup> observation for February, etc.).
#' * A number of observations equal to `nrow(in_ts)`: the specified coefficients will be used for the corresponding 
#' periods of the input time series object (argument `in_ts`), i.e., 1<sup>st</sup> observation for the 1<sup>st</sup> 
#' period, 2<sup>nd</sup> observation for the 2<sup>nd</sup> period, etc.).
#'
#' Specifying `quiet = TRUE` will suppress the [tsraking()] messages (e.g., function header) and only
#' display essential information such as warnings, errors and the period (or set of periods) being reconciled.
#' We advise against _wrapping_ your [tsraking_driver()] function call with [suppressMessages()] to further
#' suppress the display of the _raking period(s)_ information as this would make troubleshooting difficult
#' in case of issues with individual raking problems.
#'
#' Although [tsraking()] could be called with `*apply()` to successively reconcile all the periods of the input time 
#' series (`in_ts`), using [tsraking_driver()] has a few advantages, namely:
#' - temporal total preservation (only period-by-period processing, without temporal total preservation,
#' would be possible with `*apply()`);
#' - more flexibility in the specification of user-defined alterability coefficients (e.g., period-specific values);
#' - display of the period being processed (reconciled) in the console, which is useful for troubleshooting individual
#' raking problems;
#' - improved error handling, i.e., better management of warnings or errors if they were to occur only for some
#' raking problems (periods);
#' - readily returns a "ts" ("mts") object.
#'
#' 
#' @inheritSection tsbalancing Processing groups
#'
#'
#' @returns
#' The function returns a time series object (class "ts" or "mts") containing the reconciled component series, 
#' reconciled cross-sectional control totals and other series specified with [tsraking()] argument `id`. It can be 
#' explicitly coerced to another type of object with the appropriate `as*()` function (e.g., `tsibble::as_tsibble()` 
#' would coerce it to a tsibble). 
#' 
#' Note that a `NULL` object is returned if an error occurs before data processing could start. Otherwise, if execution 
#' gets far enough so that data processing could start, then an incomplete object (with `NA` values) would be returned 
#' in case of errors. 
#'
#'
#' @references Statistics Canada (2018). "Chapter 6: Advanced topics", **Theory and Application of Reconciliation
#' (Course code 0437)**, Statistics Canada, Ottawa, Canada.
#'
#'
#' @seealso [tsraking()] [tsbalancing()] [rkMeta_to_blSpecs()] [gs.build_proc_grps()] 
#'
#'
#' @example misc/function_examples/tsraking_driver-ex.R
#'
#'
#' @export
tsraking_driver <- function(in_ts,                         # input data as a "ts" ("mts") object
                            ...,                           # `tsraking()` arguments excluding `data_df`
                            temporal_grp_periodicity = 1,  # positive integer (number of periods for temporal group preservation)
                            temporal_grp_start = 1) {      # integer in the [1..`temporal_grp_periodicity`] interval




  ### Internal functions ###


  # Generate the `alterability_df` argument data frame
  gen_alter_arg <- function(per_vec) {
    alter_df[per_vec, , drop = FALSE]
  }




  ### Main function ###
  
  # Define a pointer to the main function environment and create a NULL object
  main.e <- environment()
  main.e$dummy <- NULL
  
  # Initialize the object to be returned by the function via `on.exit()`
  out_ts <- NULL
  on.exit(return(out_ts))
  try_error <- FALSE
  try_error_msg <- ""
  warning_flag <- FALSE
  
  # Enforce the default R "error" option (`options(error = NULL)`). E.g. this Turns off traceback
  # generated by calls to the stop() function inside internal functions in R Studio.
  ini_error_opt <- getOption("error")
  on.exit(options(error = ini_error_opt), add = TRUE)
  options(error = NULL)


  # Argument validation

  # Mandatory arguments (without default values)
  if (nchar(deparse1(substitute(in_ts))) == 0) {
    stop("Argument 'in_ts' is mandatory (it must be specified).\n\n", call. = FALSE)
  }
  in_ts <- in_ts
  if (!stats::is.ts(in_ts)) {
    stop("Argument 'in_ts' is not a 'ts' object.\n\n", call. = FALSE)
  }
  in_ts <- stats::as.ts(in_ts)
  ts_freq <- stats::frequency(in_ts)
  periods <- gs.time2str(in_ts)
  
  # Optional arguments (with default values)
  tmp <- (unlist(temporal_grp_periodicity))[1]
  if (!identical(temporal_grp_periodicity, tmp) || is.null(tmp) || !is.finite(tmp) ||
      is.finite(tmp) && (tmp <= 0 || tmp != as.integer(tmp))) {
    stop("Argument 'temporal_grp_periodicity' must be a positive integer.\n\n", call. = FALSE)
  }
  temporal_grp_periodicity <- as.integer(temporal_grp_periodicity)
  if (temporal_grp_periodicity > 1) {
    tmp <- (unlist(temporal_grp_start))[1]
    if (!identical(temporal_grp_start, tmp) || is.null(tmp) || !is.finite(tmp) ||
        is.finite(tmp) && (tmp <= 0 || tmp > temporal_grp_periodicity || tmp != as.integer(tmp))) {
      stop("Argument 'temporal_grp_start' must be an integer in the [1..", temporal_grp_periodicity, "] interval.\n\n",
           call. = FALSE)
    }
    temporal_grp_start <- as.integer(temporal_grp_start)
  }


  # Build the initial version of the input data frame
  n_per <- nrow(in_ts)
  in_df <- data.frame(RK._t_ = 1:n_per,
                      RK._year_ = gs.time2year(in_ts),
                      RK._per_ = gs.time2per(in_ts),
                      as.data.frame(in_ts))
  row.names(in_df) <- NULL


  # `tsraking()` arguments, i.e. the dot (...) arguments
  n_dot_args <- ...length()
  if (n_dot_args == 0) {
    stop("`tsraking()` argument 'metadata_df' must be specified.\n\n", call. = FALSE)
  }
  dot_args_list <- list(...)
  dot_args <- ...names()
  if (length(dot_args) == 0) {
    dot_args <- rep("", n_dot_args)
    names(dot_args_list) <- dot_args
  }
  tsraking_args <- names(formals("tsraking"))

  # "named" arguments
  named_args <- intersect(dot_args, tsraking_args)
  args_list <- dot_args_list[named_args]
  if ("data_df" %in% names(args_list)) {
    stop("`tsraking()` argument 'data_df' must NOT be specified with `tsraking_driver()`. ",
         "Use argument 'in_ts' instead.\n\n", call. = FALSE)
  }

  # "unnamed" arguments
  n_unnamed <- n_dot_args - length(args_list)
  if (n_unnamed > 0) {
    temp <- setdiff(dot_args, named_args)
    invalid_args <- temp[temp != ""]
    if (length(invalid_args) > 0) {
      stop("The following arguments are not defined for `tsraking()`:",
           paste0("\n  ", invalid_args, collapse = ""), "\n\n", call. = FALSE)
    }
    remaining_tsraking_args <- setdiff(tsraking_args, c("data_df", named_args))
    if (n_unnamed > length(remaining_tsraking_args)) {
      stop("Unable to match some of the specified dot (...) arguments with `tsraking()` arguments. ",
           "Remember not to specify argument 'data_df' with `tsraking_driver()`. Use argument 'in_ts' instead.\n\n",
           call. = FALSE)
    }

    # Assign names to the unnamed `tsraking()` arguments and add them to list `args_list`
    unnamed_args_list <- dot_args_list[which(dot_args == "")]
    names(unnamed_args_list) <- remaining_tsraking_args[1:n_unnamed]
    args_list <- c(args_list, unnamed_args_list)
  }
  specified_args <- names(args_list)


  # Quick validation of argument `metadata_df`
  if (!("metadata_df" %in% specified_args)) {
    stop("`tsraking()` argument 'metadata_df' must be specified.\n\n", call. = FALSE)
  }
  if (!is.data.frame(args_list$metadata_df)) {
    stop("`tsraking()` argument 'metadata_df' must be a 'data.frame' object.\n\n", call. = FALSE)
  }
  args_list$metadata_df <- as.data.frame(args_list$metadata_df)
  meta_cols <- toupper(names(args_list$metadata_df))
  if (length(intersect(c("SERIES", "TOTAL1"), meta_cols)) < 2) {
    stop("`tsraking()` argument 'metadata_df' contains invalid data. Suspecting argument ",
         "'data_df' to be inadvertently specified for argument 'metadata_df'.\n\n", call. = FALSE)
  }
  meta_df <- args_list$metadata_df
  names(meta_df) <- meta_cols


  # Process the alterability file (argument 'alterability_df')
  alter_arg_id <- which(specified_args == "alterability_df")
  if (length(alter_arg_id) == 1) {
    if (!is.null(args_list$alterability_df)) {

      if (!is.data.frame(args_list$alterability_df)) {
        stop("`tsraking()` argument 'alterability_df' must be a 'data.frame' object.\n\n", call. = FALSE)
      }
      args_list$alterability_df <- as.data.frame(args_list$alterability_df)


      # Assign the relevant alter coefs to each individual period
      n_alter_obs <- nrow(args_list$alterability_df)

      # Single obs: propagate (replicate) the alter coefs for all periods
      if (n_alter_obs == 1) {
        alter_df <- args_list$alterability_df[rep(1, n_per), , drop = FALSE]
        row.names(alter_df) <- NULL

      # Number of obs matches the 'ts' frequency: propagate the alter coefs to the corresponding
      # periods (matching cycles) in the 'ts'
      } else if (n_alter_obs == ts_freq) {
        alter_df <- args_list$alterability_df
        row.names(alter_df) <- NULL
        alter_df$RK._per_ <- as.integer(row.names(alter_df))
        temp <- c("RK._per_", "RK._t_")
        alter_df <- merge(alter_df, in_df[temp], by = "RK._per_")
        alter_df <- alter_df[order(alter_df$RK._t_), setdiff(names(alter_df), temp), drop = FALSE]
        row.names(alter_df) <- NULL

      # Number of obs matches the number of periods in the 'ts': nothing to do
      } else if (n_alter_obs == n_per)  {
        alter_df <- args_list$alterability_df
        row.names(alter_df) <- NULL

      # Incompatible number of obs in the `alterability_df` data frame
      } else {
        stop("`tsraking()` argument 'alterability_df' contains a number of observations ",
             "that is not compatible with the specified time series (argument 'in_ts').\n\n", call. = FALSE)
      }
      alter_arg_func <- gen_alter_arg

    # No alterability file
    } else {
      alter_arg_func <- gs.NULL_func
    }

    # Temporarily remove `alterability_df` from the list of arguments (will be (re)added for each
    # `tsraking()` call)
    args_list <- args_list[-alter_arg_id]

    # No `alterability_df` argument
  } else {
    alter_arg_func <- gs.NULL_func
  }


  # Set the output time series columns and problem variables (`NA` in case of errors)
  all_cols <- colnames(in_ts)
  prob_cols <- intersect(all_cols, unlist(apply(meta_df[intersect(meta_cols, c("SERIES", "TOTAL1", "TOTAL2"))],
                                                2, unique)))
  names(prob_cols) <- NULL
  if ("id" %in% specified_args) {
    xtra_cols <- intersect(all_cols, gs.cleanup_col_list(args_list$id))
  } else {
    xtra_cols <- character(0)
  }
  out_cols <- intersect(all_cols, unique(c(prob_cols, xtra_cols)))


  # Define the processing groups (set of raking problems)
  grp_df <- gs.build_proc_grps(in_df$RK._year_,
                               in_df$RK._per_,
                               n_per,
                               ts_freq,
                               temporal_grp_periodicity,
                               temporal_grp_start)

  # Activate message display
  n_grps <- nrow(grp_df)
  if (n_grps > 1) {
    msg_func <- message
    final_msg_flag <- TRUE
    try_stop_func <- gs.try_stop
  } else {
    msg_func <- gs.NULL_func
    final_msg_flag <- FALSE
    try_stop_func <- gs.NULL_func
  }


  # Rake each processing group (generate the `tsraking()` calls)
  out_df <- NULL
  for (grp in 1:n_grps) {

    # Print the processing group header
    if (grp_df$complete_grp[grp]) {
      msg_str <- paste0("Raking periods [", periods[grp_df$beg_per[grp]], " - ", periods[grp_df$end_per[grp]], "]")
    } else {
      msg_str <- paste0("Raking period [", periods[grp_df$beg_per[grp]], "]")
    }
    msg_func("\n\n", msg_str)
    msg_func(strrep("=", nchar(msg_str)), "\n")

    per_vec <- grp_df$beg_per[grp]:grp_df$end_per[grp]
    grp_data_df <- in_df[per_vec, out_cols, drop = FALSE]

    # The warning and error handler functions defined below (inside `withCallingHandlers()` and `tryCatch()` 
    # respectively) use environment pointer `main.e` to update objects in the current (main) function environment 
    # (instead of using operator `<<-`)
    out_df <- rbind(out_df,
                    tryCatch(
                      withCallingHandlers(

                        do.call("tsraking",
                                c(list(data_df = grp_data_df,
                                       alterability_df = alter_arg_func(per_vec)),
                                  args_list)
                        ),

                        warning = function(wCnd) {
                          main.e$warning_flag <- TRUE
                        }
                      ),

                      error = function(eCnd) {
                        main.e$try_error_msg <- conditionMessage(eCnd)
                        try_stop_func(try_error_msg)
                        main.e$try_error <- TRUE
                        df <- grp_data_df
                        df[prob_cols] <- NA
                        df
                      }
                    ))
  }


  # Create the output ts object
  out_ser <- intersect(all_cols, names(out_df))
  out_ts <- stats::ts(data = out_df[out_ser],
                      start = stats::start(in_ts),
                      frequency = ts_freq)


  # Display a final warning/error message for multiple processing groups
  if (final_msg_flag) {
    if (warning_flag) {
      warning("Warnings were generated during processing. See relevant message(s) for details.\n",
              call. = FALSE, immediate. = TRUE)
    }
    # Non-muted error message (for proper condition catching by users of the function)
    if (try_error) {
      stop("Problems were encontered during processing. See preceeding error message(s) for details.\n\n",
           call. = FALSE)
    }
    
  # Display the error message for single processing groups
  } else if (try_error) {
    stop(try_error_msg, call. = FALSE)
  }
  
  # Output object returned via function `on.exit()`
}

#' Convert a "ts" object to a benchmarks data frame
#'
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/ts_to_bmkDF.html}})}
#' 
#' Convert a "ts" (or "mts") object into a benchmarks data frame for the benchmarking functions with five or more variables
#' (columns):
#' * four (4) for the benchmark coverage
#' * one (1) for each benchmark time series
#'
#' For discrete benchmarks (anchor points covering a single period of the indicator series, e.g., end of year
#' stocks), specify `discrete_flag = TRUE` and `alignment = "b"`, `"e"` or `"m"`.
#'
#'
#' @param in_ts (mandatory)
#'
#' Time series (object of class "ts" or "mts") to be converted.
#'
#' @param ind_frequency (mandatory)
#'
#' Integer specifying the frequency of the indicator (high frequency) series for which the benchmarks (low
#' frequency series) are related to. The frequency of a time series corresponds to the maximum number of periods
#' in a year (e.g., 12 for a monthly data, 4 for a quarterly data, 1 for annual data).
#'
#' @param discrete_flag (optional)
#'
#' Logical argument specifying whether the benchmarks correspond to discrete values (anchor points covering a single
#' period of the indicator series, e.g., end of year stocks) or not. `discrete_flag = FALSE` defines non-discrete
#' benchmarks, i.e., benchmarks that cover several periods of the indicator series (e.g. annual benchmarks cover
#' 4 quarters or 12 months, quarterly benchmarks cover 3 months, etc.).
#'
#' **Default value** is `discrete_flag = FALSE`.
#'
#' @param alignment (optional)
#'
#' Character identifying the alignment of discrete benchmarks (argument `discrete_flag = TRUE`) in the benchmark
#' (low frequency series) interval coverage window:
#' * `alignment = "b"`: beginning of the benchmark interval window (first period)
#' * `alignment = "e"`: end of the benchmark interval window (last period)
#' * `alignment = "m"`: middle of the benchmark interval window (middle period)
#'
#' This argument has no effect for non-discrete benchmarks (`discrete_flag = FALSE`).
#'
#' **Default value** is `alignment = "b"`.
#'
#' @param bmk_interval_start (optional)
#'
#' Integer in the \[1 .. `ind_frequency`\] interval specifying the period (cycle) of the indicator (high frequency) 
#' series at which the benchmark (low frequency series) interval window starts. E.g., annual benchmarks corresponding to 
#' fiscal years defined from April to March of the following year would be specified with `bmk_interval_start = 4` 
#' for a monthly indicator series (`ind_frequency = 12`) and `bmk_interval_start = 2` for a quarterly indicator 
#' series (`ind_frequency = 4`).
#'
#' **Default value** is `bmk_interval_start = 1`.
#'
#' @param startYr_cName,startPer_cName,endYr_cName,endPer_cName (optional)
#'
#' Strings specifying the name of the numeric variables (columns) in the output data frame that will define the
#' benchmarks coverage, i.e., the starting and ending year and period (cycle) identifiers.
#'
#' **Default values** are `startYr_cName = "startYear"`, `startPer_cName = "startPeriod"`
#' `endYr_cName = "endYear"` and `endPer_cName   = "endPeriod"`.
#'
#' @param val_cName (optional)
#'
#' String specifying the name of the numeric variable (column) in the output data frame that will contain the
#' benchmark values. This argument has no effect for "mts" objects (benchmark variable names are automatically
#' inherited from the "mts" object).
#'
#' **Default value** is `val_cName = "value"`.
#'
#'
#' @returns
#' The function returns a data frame with five or more variables:
#' * Benchmark coverage starting year, type numeric (see argument `startYr_cName`)
#' * Benchmark coverage starting period (cycle), type numeric (see argument `startPer_cName`)
#' * Benchmark coverage ending year, type numeric (see argument `endtYr_cName`)
#' * Benchmark coverage ending period (cycle), type numeric (see argument `endPer_cName`)
#' * One ("ts" object) or many ("mts" object) benchmark data variable(s), type numeric (see argument `val_cName`)
#'
#' Note: the function returns a "data.frame" object than can be explicitly coerced to another type of object 
#' with the appropriate `as*()` function (e.g., `tibble::as_tibble()` would coerce it to a tibble).
#'
#'
#' @seealso [ts_to_tsDF()] [stack_bmkDF()] [benchmarking()] [stock_benchmarking()] [time_values_conv]
#'
#'
#' @example misc/function_examples/ts_to_bmkDF-ex.R
#'
#'
#' @export
ts_to_bmkDF <- function(in_ts,
                        ind_frequency,
                        discrete_flag = FALSE,
                        alignment = "b",
                        bmk_interval_start = 1,
                        startYr_cName = "startYear",
                        startPer_cName = "startPeriod",
                        endYr_cName = "endYear",
                        endPer_cName = "endPeriod",
                        val_cName = "value") {

  
  # Enforce the default R "error" option (`options(error = NULL)`). E.g. this Turns off traceback
  # generated by calls to the stop() function inside internal functions in R Studio.
  ini_error_opt <- getOption("error")
  on.exit(options(error = ini_error_opt))
  options(error = NULL)
  
  
  # validate object
  if (!stats::is.ts(in_ts)) {
    stop("Argument 'in_ts' is not a 'ts' object.\n\n", call. = FALSE)
  }
  in_ts <- stats::as.ts(in_ts)


  # (Re)align the `in_ts` time values at the start of the benchmark interval (to ensure proper year and indicator period 
  # determination) and then shift them to reflect `bmk_interval_start`
  #
  # => Note that since the `in-ts` frequency corresponds to the benchmark series (and not the indicator series), function 
  #    `gs.time2per()` (`stats::cycle()`) cannot be used to determine the periods of the indicator series.
  #    Function 'gs.time2year()` could be used to determine the years of the indicator series, but it's not necessary since
  #    the time values are (re)aligned at the start of the intervals (simply "truncating" the time values works fine)
  #
  bmk_freq <- stats::frequency(in_ts)
  beg_time <- stats::time(stats::ts(in_ts,
                                    start = c(gs.time2year(in_ts)[1], gs.time2per(in_ts)[1]),
                                    frequency = stats::frequency(in_ts))) +
    ((bmk_interval_start - 1) %% (ind_frequency / bmk_freq)) / ind_frequency


  # Create the initial data frame with generic names for the benchmark window columns
  # (first 4 columns)

  # Non-discrete benchmarks (e.g. flow series benchmarking)
  if (!gs.validate_arg_logi(discrete_flag)) {

    end_time <- beg_time + (ind_frequency / bmk_freq - 1) / ind_frequency
    out_df <- data.frame(
      col1 = as.integer(signif(beg_time, gs.signif_digits)),
      col2 = as.integer(round((beg_time - as.integer(beg_time)) * ind_frequency)) + 1L,
      col3 = as.integer(signif(end_time, gs.signif_digits)),
      col4 = as.integer(round((end_time - as.integer(end_time)) * ind_frequency)) + 1L
    )

  # Discrete benchmarks (e.g. stock series benchmarking)
  } else {

    tmp <- toupper(substr(alignment, 1, 1))
    if (tmp == "E") {
      shift <- (ind_frequency / bmk_freq - 1) / ind_frequency
    } else if (tmp == "M") {
      # An obvious formula (based on the above one) would probably be `round((ind_frequency / bmk_freq - 1) / 2) / ind_frequency`.
      # However, `base::round()` implements 'go to the even digit' for rounding off a 5 (0.5 -> 0, 1.5 = 2.5 -> 2, etc.) while we
      # want them (half intervals) always rounded up (half years are closer to July 1st than June 1st, i.e., a shift of 6 months).
      # For example, we want...
      #   - monthly indicators with annual benchmarks     : shift = round((12 / 1 - 1) / 2) / 12 = round(5.5) / 12 = 6 / 12 = 0.5)
      #                                                     (base::round(5.5) = 6 -> all good)
      #   - monthly indicators with semi-annual benchmarks: shift = round((12 / 2 - 1) / 2) / 12 = round(2.5) / 12 = 3 / 12 = 0.25)
      #                                                     (base::round(2.5) = 2 -> not what we want!)
      # The following formula corresponds to always rounding up value `(ind_frequency / bmk_freq - 1) / 2`
      shift <- as.integer(ind_frequency / bmk_freq / 2) / ind_frequency
    } else {
      shift <- 0
    }
    time <- beg_time + shift
    yr <- as.integer(signif(time, gs.signif_digits))
    per <- as.integer(round((time - as.integer(time)) * ind_frequency)) + 1L
    out_df <- data.frame(
      col1 = yr,
      col2 = per,
      col3 = yr,
      col4 = per
    )
  }

  # Set the date column names
  names(out_df) <- c(startYr_cName, startPer_cName, endYr_cName, endPer_cName)


  # Multivariate series data columns for a mts object
  if (stats::is.mts(in_ts)) {
    out_df <- cbind(out_df, in_ts)

  # Single series data column for a ts object
  } else {
    temp_df <- data.frame(col1 = as.numeric(in_ts))
    names(temp_df) <- val_cName[1]
    out_df <- cbind(out_df, temp_df)
  }

  # Reset the now names (numbers)
  row.names(out_df) <- NULL

  out_df
}

#' Stack time series data
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/stack_tsDF.html}})}
#' 
#' Convert a multivariate time series data frame (see [ts_to_tsDF()]) for the benchmarking functions
#' ([benchmarking()] and [stock_benchmarking()]) into a stacked (tall) data frame with four variables (columns):
#' * one (1) for the series name
#' * two (2) for the data point identification (year and period)
#' * one (1) for the data point value
#'
#' Missing (`NA`) series values are not included in the output stacked data frame by default. Specify argument 
#' `keep_NA = TRUE` in order to keep them.
#'
#' This function is useful when intending to use the `by` argument (*BY-group* processing mode) of the benchmarking
#' functions in order to benchmark multiple series in a single function call.
#'
#'
#' @param ts_df (mandatory)
#'
#' Data frame (object of class "data.frame") that contains the multivariate time series data to be stacked.
#'
#' @param ser_cName (optional)
#'
#' String specifying the name of the character variable (column) in the output stacked data frame that will
#' contain the series names (name of the time series variables in the input multivariate time series data
#' frame). This variable can then be used as the BY-group variable (argument `by`) with the benchmarking functions.
#'
#' **Default value** is `ser_cName = "series"`.
#'
#' @param yr_cName,per_cName (optional)
#'
#' Strings specifying the name of the numeric variables (columns) in the input multivariate time series data frame
#' that contain the data point year and period (cycle) identifiers. These variables are *transferred* to the output 
#' stacked data frame with the same variable names.
#'
#' **Default values** are `yr_cName = "year"` and `per_cName   = "period"`.
#'
#' @param val_cName (optional)
#'
#' String specifying the name of the numeric variable (column) in the output stacked data frame that will
#' contain the data point values.
#'
#' **Default value** is `val_cName = "value"`.
#'
#' @param keep_NA (optional)
#'
#' Logical argument specifying whether `NA` time series values in the input multivariate time series data frame
#' should be kept in the output stacked data frame.
#'
#' **Default value** is `keep_NA = FALSE`.
#'
#'
#' @returns
#' The function returns a data frame with four variables:
#' * Series name, type character (see argument `ser_cName`)
#' * Data point year, type numeric (see argument `yr_cName`)
#' * Data point period, type numeric (see argument `per_cName`)
#' * Data point value, type numeric (see argument `val_cName`)
#'
#' Note: the function returns a "data.frame" object than can be explicitly coerced to another type of object 
#' with the appropriate `as*()` function (e.g., `tibble::as_tibble()` would coerce it to a tibble).
#'
#'
#' @seealso [unstack_tsDF()] [stack_bmkDF()] [ts_to_tsDF()] [benchmarking()] [stock_benchmarking()]
#'
#'
#' @example misc/function_examples/stack_tsDF-ex.R
#'
#'
#' @export
stack_tsDF <- function(ts_df,
                       ser_cName = "series",
                       yr_cName = "year",
                       per_cName = "period",
                       val_cName = "value",
                       keep_NA = FALSE) {
  
  # Enforce the default R "error" option (`options(error = NULL)`). E.g. this Turns off traceback
  # generated by calls to the stop() function inside internal functions in R Studio.
  ini_error_opt <- getOption("error")
  on.exit(options(error = ini_error_opt))
  options(error = NULL)
  
  # validate object
  if (!is.data.frame(ts_df)) {
    stop("Argument 'ts_df' is not a 'data.frame' object.\n\n", call. = FALSE)
  }
  ts_df <- as.data.frame(ts_df)
  df_cols <- names(ts_df)
  date_cols <- c(yr_cName, per_cName)
  date_args <- c("yr_cName", "per_cName")
  for (ii in seq_along(date_cols)) {
    if (!(date_cols[ii] %in% df_cols)) {
      stop("The input data frame does not contain column \"", date_cols[ii], "\" (argument '",
           date_args[ii], "').\n\n", call. = FALSE)
    }
  }

  ser_list <- setdiff(names(ts_df), date_cols)
  out_df <- data.frame(col1 = character(),
                       col2 = integer(),
                       col3 = integer(),
                       col4 = double(),
                       stringsAsFactors = FALSE)
  for (ser in ser_list) {
    if (gs.validate_arg_logi(keep_NA)) {
      tmp_df <- ts_df[c(date_cols, ser)]
    } else {
      tmp_df <- ts_df[!is.na(ts_df[ser]), c(date_cols, ser)]
    }
    out_df <- rbind(out_df, data.frame(col1 = ser,
                                       col2 = as.integer(tmp_df[[yr_cName]]),
                                       col3 = as.integer(tmp_df[[per_cName]]),
                                       col4 = as.numeric(tmp_df[[ser]]),
                                       stringsAsFactors = FALSE))
  }

  # Set the column names and reset the now names (numbers)
  names(out_df) <- c(ser_cName, date_cols, val_cName)
  row.names(out_df) <- NULL
  out_df
}

#' Reciprocal function of [ts_to_tsDF()]
#'
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/tsDF_to_ts.html}})}
#' 
#' Convert a (non-stacked) time series data frame ([benchmarking()] and [stock_benchmarking()] data format)
#' into a "ts" (or "mts") object.
#'
#' This function is useful to convert the benchmarked data frame returned by a call to [benchmarking()] or 
#' [stock_benchmarking()] into a "ts" object, where one or several series were benchmarked in *non BY-group*
#' processing mode. Stacked time series data frames associated to executions in *BY-group* mode must first be
#' *unstacked* with [unstack_tsDF()].
#'
#'
#' @usage
#' tsDF_to_ts(
#'   ts_df,
#'   frequency,
#'   yr_cName = "year",
#'   per_cName = "period"
#' )
#'
#'
#' @param ts_df (mandatory)
#'
#' Data frame (object of class "data.frame") to be converted.
#'
#' @param frequency (mandatory)
#' 
#' Integer specifying the frequency of the time series to be converted. The frequency of a time series corresponds
#' to the maximum number of periods in a year (12 for a monthly data, 4 for a quarterly data, 1 for annual data).
#'
#' @param yr_cName,per_cName (optional)
#'
#' Strings specifying the name of the numeric variables (columns) in the input data frame that contain the data
#' point year and period identifiers.
#'
#' **Default values** are `yr_cName = "year"` and `per_cName   = "period"`.
#'
#'
#' @returns
#' The function returns a time series object (class "ts" or "mts"), which can be explicitly coerced to another type 
#' of object with the appropriate `as*()` function (e.g., `tsibble::as_tsibble()` would coerce it to a tsibble).
#'
#'
#' @seealso [ts_to_tsDF()] [unstack_tsDF()] [benchmarking()] [stock_benchmarking()]
#'
#'
#' @example misc/function_examples/tsDF_to_ts-ex.R
#'
#'
#' @export
tsDF_to_ts <- function(ts_df,
                       frequency,
                       yr_cName = "year",
                       per_cName = "period") {

  # validate object
  if (!is.data.frame(ts_df)) {
    stop("Argument 'ts_df' is not a 'data.frame' object.\n\n", call. = FALSE)
  }
  if (nrow(ts_df) == 0) {
    stop("The input data frame must contain at leat one observation (row).\n\n", call. = FALSE)
  }  
  df_cols <- names(ts_df)
  date_cols <- c(yr_cName, per_cName)
  date_args <- c("yr_cName", "per_cName")
  for (ii in seq_along(date_cols)) {
    if (!(date_cols[ii] %in% df_cols)) {
      stop("The input data frame does not contain column \"", date_cols[ii], "\" (argument '",
           date_args[ii], "').\n\n", call. = FALSE)
    }
  }

  # Sort the data frame by date (safety measure)
  sorted_df <- ts_df[order(ts_df[[yr_cName]], ts_df[[per_cName]]), ]

  # Extract the numeric columns (excluding the year and period id columns)
  if (length(df_cols) > 2) {
    tmp_df <- sorted_df[setdiff(df_cols, date_cols)]
    # `sapply()` is safe: it always returns a logical vector of length minimum 1
    series_obj <- tmp_df[sapply(tmp_df, is.numeric)]
    nTS <- ncol(series_obj)
  } else {
    nTS <- 0
  }
  if (nTS == 0) {
    stop("The input data frame must contain at leat one (numeric) time series.\n\n", call. = FALSE)
  } else if (nTS == 1) {
    # Univariate ts object: transform the data frame into a vector
    series_obj <- series_obj[[1]]
  }

  # Create the ts object
  stats::ts(series_obj,
            start = as.integer(c(sorted_df[1, yr_cName], sorted_df[1, per_cName])),
            frequency = as.integer(frequency))
}

#' Convert a "ts" object to a time series data frame
#'
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/ts_to_tsDF.html}})}
#' 
#' Convert a "ts" (or "mts") object into a time series data frame for the benchmarking functions with three or more 
#' variables (columns):
#' * two (2) for the data point identification (year and period)
#' * one (1) for each time series
#'
#'
#' @usage
#' ts_to_tsDF(
#'   in_ts,
#'   yr_cName = "year",
#'   per_cName = "period",
#'   val_cName = "value"
#' )
#'
#'
#' @param in_ts (mandatory)
#'
#' Time series (object of class "ts" or "mts") to be converted.
#'
#' @param yr_cName,per_cName (optional)
#'
#' Strings specifying the name of the numeric variables (columns) in the output data frame that will contain
#' the data point year and period identifiers.
#'
#' **Default values** are `yr_cName = "year"` and `per_cName   = "period"`.
#'
#' @param val_cName (optional)
#'
#' String specifying the name of the numeric variable (column) in the output data frame that will contain the
#' data point value. This argument has no effect for "mts" objects (time series data variable names are
#' automatically inherited from the "mts" object).
#'
#' **Default value** is `val_cName = "value"`.
#'
#'
#' @returns
#' The function returns a data frame with three or more variables:
#' * Data point year, type numeric (see argument `startYr_cName`)
#' * Data point period, type numeric (see argument `startPer_cName`)
#' * One ("ts" object) or many ("mts" object) time series data variable(s), type numeric (see argument `val_cName`)
#'
#' Note: the function returns a "data.frame" object than can be explicitly coerced to another type of object 
#' with the appropriate `as*()` function (e.g., `tibble::as_tibble()` would coerce it to a tibble).
#'
#'
#' @seealso [tsDF_to_ts()] [ts_to_bmkDF()] [stack_tsDF()] [benchmarking()] [stock_benchmarking()] [time_values_conv]
#'
#'
#' @example misc/function_examples/ts_to_tsDF-ex.R
#'
#'
#' @export
ts_to_tsDF <- function(in_ts,
                       yr_cName = "year",
                       per_cName = "period",
                       val_cName = "value") {


  # validate object
  if (!stats::is.ts(in_ts)) {
    stop("Argument 'in_ts' is not a 'ts' object.\n\n", call. = FALSE)
  }
  in_ts <- stats::as.ts(in_ts)

  if (stats::is.mts(in_ts)) {

    # Create the initial data frame with only the date columns (first 2 columns)
    out_df <- data.frame(
      col1 = gs.time2year(in_ts),
      col2 = gs.time2per(in_ts)
    )

    # Set date the column names
    names(out_df) <- cbind(yr_cName, per_cName)

    # Add the series data
    out_df <- cbind(out_df, in_ts)

  } else {

    # Create the initial data frame with generic column names
    out_df <- data.frame(
      col1 = gs.time2year(in_ts),
      col2 = gs.time2per(in_ts),
      col3 = as.numeric(in_ts)
    )

    # Set the column names
    names(out_df) <- c(yr_cName, per_cName, val_cName)
  }

  # Reset the now names (numbers)
  row.names(out_df) <- NULL

  out_df
}

#' Plot benchmarking adjustments
#'
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/plot_benchAdj.html}})}
#' 
#' Plot benchmarking adjustments for a single series in the current (active) graphics device. Up to three types of
#' adjustments can be overlayed in the same plot:
#' - Adjustments generated by function [benchmarking()]
#' - Adjustments generated by function [stock_benchmarking()]
#' - Cubic spline associated to adjustments generated by function [stock_benchmarking()]
#'
#' These plots can be useful to assess the quality of the benchmarking results and compare the adjustments
#' generated by both benchmarking functions ([benchmarking()] and [stock_benchmarking()]) for stock series.
#'
#'
#' @param PB_graphTable (optional)
#'
#' Data frame (object of class "data.frame") corresponding to the [benchmarking()] (PB for "Proc Benchmarking"
#' approach) function output `graphTable` data frame. Specify `NULL` not to include the [benchmarking()] adjustments
#' in the plot.
#'
#' **Default value** is `PB_graphTable = NULL`.
#'
#' @param SB_graphTable (optional)
#'
#' Data frame (object of class "data.frame") corresponding to the [stock_benchmarking()] (SB) function output
#' `graphTable` data frame. Specify `NULL` not to include the [stock_benchmarking()] adjustments in the plot.
#'
#' **Default value** is `SB_graphTable = NULL`.
#'
#' @param SB_splineKnots (optional)
#'
#' Data frame (object of class "data.frame") corresponding to the [stock_benchmarking()] (SB) function output
#' `splineKnots` data frame. Specify `NULL` not to include the [stock_benchmarking()] cubic spline in the plot.
#'
#' **Default value** is `SB_splineKnots = NULL`.
#'
#' @param legendPos (optional)
#'
#' String (keyword) specifying the location of the legend in the plot. See the description of argument `x` in
#' the documentation of `graphics::legend()` for the list of valid keywords. Specify `NULL` not to include a
#' legend in the plot.
#'
#' **Default value** is `legendPos = "bottomright"`.
#'
#'
#' @details
#' `graphTable` data frame (arguments `PB_graphTable` and `SB_graphTable`) variables used in the plot:
#' - `t` for the x-axis values (*t*)
#' - `benchmarkedSubAnnualRatio` for the *Stock Bench. (SB)* and *Proc Bench. (PB)* lines
#' - `bias` for the *Bias* line (when \eqn{\rho < 1})
#' 
#' `splineKnots` data frame (argument `SB_splineKnots`) variables used in the plot:
#' - `x` for the x-axis values (*t*)
#' - `y` for the *Cubic spline* line and the *Extra knot* and *Original knot* points
#' - `extraKnot` for the type of knot (*Extra knot* vs. *Original knot*)
#' 
#' See section **Value** of [benchmarking()] and [stock_benchmarking()] for more details on these data frames.
#' 
#'
#' @returns
#' This function returns nothing (`invisible(NULL)`).
#'
#'
#' @seealso [plot_graphTable()] [bench_graphs] [benchmarking()] [stock_benchmarking()]
#'
#'
#' @example misc/function_examples/plot_benchAdj-ex.R
#'
#'
#' @export
plot_benchAdj <- function(PB_graphTable = NULL,
                          SB_graphTable = NULL,
                          SB_splineKnots = NULL,
                          legendPos = "bottomright") {

  # Initialize info
  lab_list <- NULL
  col_list <- NULL
  lty_list <- NULL
  lwd_list <- NULL
  pch_list <- NULL
  pt.cex_list <- NULL
  min_x <- .Machine$double.xmax
  max_x <- .Machine$double.xmin
  min_y <- .Machine$double.xmax
  max_y <- .Machine$double.xmin
  bias <- NULL
  BI_string <- ""

  # Process `stock_benchmarking()` `splineKnots` data frame info
  if (!is.null(SB_splineKnots)) {
    
    # Accept `SB_splineKnots = NA` as `SB_splineKnots = NULL`
    tmp <- (unlist(SB_splineKnots))[1]
    if (identical(SB_splineKnots, tmp) && is.na(tmp)) {
      SB_splineKnots <- NULL
    } else {
      
      if (!is.data.frame(SB_splineKnots)) {
        stop("Argument 'SB_splineKnots' is not a 'data.frame' object.\n\n", call. = FALSE)
      }
      SB_splineKnots <- as.data.frame(SB_splineKnots)
      var_list <- names(SB_splineKnots)
      if ("extraKnot" %in% var_list) {
        extraKnot_flag <- TRUE
        lab_list <- c(lab_list, "Cubic spline", "Extra knot", "Original knot")
        col_list <- c(col_list, "black", "black", "green3")
        lty_list <- c(lty_list, 1, NA, NA)
        lwd_list <- c(lwd_list, 1, NA, NA)
        pch_list <- c(pch_list, NA, 1, 19)
        pt.cex_list <- c(pt.cex_list, NA, 1.5, 1.5)
      } else {
        extraKnot_flag <- FALSE
        lab_list <- c(lab_list, "Cubic spline")
        col_list <- c(col_list, "black")
        lty_list <- c(lty_list, 1)
        lwd_list <- c(lwd_list, 1)
        pch_list <- c(pch_list, 1)
        pt.cex_list <- c(pt.cex_list, 1.5)
      }
      interpol <- stats::spline(SB_splineKnots$x, SB_splineKnots$y, method = "natural")
      min_x <- min(min_x, interpol$x, na.rm = TRUE)
      max_x <- max(max_x, interpol$x, na.rm = TRUE)
      min_y <- min(min_y, interpol$y, na.rm = TRUE)
      max_y <- max(max_y, interpol$y, na.rm = TRUE)
    }
  }

  # Process `stock_benchmarking()` `graphTable` data frame info
  if (!is.null(SB_graphTable)) {
    
    # Accept `SB_graphTable = NA` as `SB_graphTable = NULL`
    tmp <- (unlist(SB_graphTable))[1]
    if (identical(SB_graphTable, tmp) && is.na(tmp)) {
      SB_graphTable <- NULL
    } else {
      
      if (!is.data.frame(SB_graphTable)) {
        stop("Argument 'SB_graphTable' is not a 'data.frame' object.\n\n", call. = FALSE)
      }
      SB_graphTable <- as.data.frame(SB_graphTable)
      lab_list <- c(lab_list, "Stock Bench. (SB)")
      col_list <- c(col_list, "red")
      lty_list <- c(lty_list, 1)
      lwd_list <- c(lwd_list, 1)
      pch_list <- c(pch_list, 1)
      pt.cex_list <- c(pt.cex_list, 1)
      min_x <- min(min_x, SB_graphTable$t, na.rm = TRUE)
      max_x <- max(max_x, SB_graphTable$t, na.rm = TRUE)
      min_y <- min(min_y, SB_graphTable$benchmarkedSubAnnualRatio, na.rm = TRUE)
      max_y <- max(max_y, SB_graphTable$benchmarkedSubAnnualRatio, na.rm = TRUE)
      var_list <- names(SB_graphTable)
      if ("bias" %in% var_list) {
        
        # Display the bias line only for non-Denton benchmarking (`rho < 1`)
        if ("rho" %in% var_list) {
          if (abs(SB_graphTable$rho[1] - 1) > gs.tolerance) {
            bias <- SB_graphTable$bias[1]
          }
        } else {
          bias <- SB_graphTable$bias[1]
        }
      }
      if ("lambda" %in% var_list) {
        if (SB_graphTable$lambda[1] == 0) {
          BI_string <- "(BI differences)"
        } else {
          BI_string <- "(BI ratios)"
        }
      }
    }
    
  }

  # Process `benchmarking()` `graphTable` data frame info
  if (!is.null(PB_graphTable)) {
    
    # Accept `PB_graphTable = NA` as `PB_graphTable = NULL`
    tmp <- (unlist(PB_graphTable))[1]
    if (identical(PB_graphTable, tmp) && is.na(tmp)) {
      PB_graphTable <- NULL
    } else {
      
      if (!is.data.frame(PB_graphTable)) {
        stop("Argument 'PB_graphTable' is not a 'data.frame' object.\n\n", call. = FALSE)
      }
      PB_graphTable <- as.data.frame(PB_graphTable)
      PB_graphTable <- PB_graphTable[!duplicated(PB_graphTable[c("t", "benchmarkedSubAnnualRatio")]), ]
      lab_list <- c(lab_list, "Proc Bench. (PB)")
      col_list <- c(col_list, "blue")
      lty_list <- c(lty_list, 1)
      lwd_list <- c(lwd_list, 1)
      pch_list <- c(pch_list, 1)
      pt.cex_list <- c(pt.cex_list, 1)
      min_x <- min(min_x, PB_graphTable$t, na.rm = TRUE)
      max_x <- max(max_x, PB_graphTable$t, na.rm = TRUE)
      min_y <- min(min_y, PB_graphTable$benchmarkedSubAnnualRatio, na.rm = TRUE)
      max_y <- max(max_y, PB_graphTable$benchmarkedSubAnnualRatio, na.rm = TRUE)
      var_list <- names(PB_graphTable)
      if ("bias" %in% var_list) {
        
        # Display the bias line only for non-Denton benchmarking (`rho < 1`)
        if ("rho" %in% var_list) {
          if (abs(PB_graphTable$rho[1] - 1) > gs.tolerance) {
            bias <- PB_graphTable$bias[1]
          }
        } else {
          bias <- PB_graphTable$bias[1]
        }
      }
      if ("lambda" %in% var_list) {
        if (PB_graphTable$lambda[1] == 0) {
          BI_string <- "(BI differences)"
        } else {
          BI_string <- "(BI ratios)"
        }
      }
    }
  }

  if (!is.null(lab_list)) {

    # Impose a minimum range for the 'x' and y' axis values
    # (to avoid "new" warning messages introduced after R-4.2.3)
    if (max_x - min_x < gs.tolerance) {
      # E.g., single adjustment value
      min_x <- min_x - gs.tolerance / 2
      max_x <- max_x + gs.tolerance / 2
    }
    if (max_y - min_y < gs.tolerance) {
      # E.g., identical/repeated adjustment values
      min_y <- min_y - gs.tolerance / 2
      max_y <- max_y + gs.tolerance / 2
    }

    # Initialize the plot
    plot(min_x, min_y, type = "n", xlim = c(min_x, max_x), ylim = c(min_y, max_y),
         xlab = "t", ylab = paste("Adjustments", BI_string))
    graphics::title(paste("Benchmarking Adjustments", BI_string))

    # Plot the cubic spline with knots
    if (!is.null(SB_splineKnots)) {
      graphics::lines(interpol, type = "l")
      graphics::points(SB_splineKnots$x, SB_splineKnots$y, cex = 1.5)
      if (extraKnot_flag) {
        graphics::points(SB_splineKnots$x[SB_splineKnots$extraKnot == FALSE],
                         SB_splineKnots$y[SB_splineKnots$extraKnot == FALSE],
                         pch = 19, cex = 1.5, col = "green3")
      }
    }

    # Plot the stock_benchmarking() adjustments
    if (!is.null(SB_graphTable)) {
      graphics::lines(SB_graphTable$t, SB_graphTable$benchmarkedSubAnnualRatio,
                      type = "l", col = "red", lwd = 2)
      graphics::points(SB_graphTable$t, SB_graphTable$benchmarkedSubAnnualRatio,
                       col = "red")
    }

    # Plot the benchmarking() adjustments
    if (!is.null(PB_graphTable)) {
      graphics::lines(PB_graphTable$t, PB_graphTable$benchmarkedSubAnnualRatio,
                      type = "l", col = "blue", lwd = 2)
      graphics::points(PB_graphTable$t, PB_graphTable$benchmarkedSubAnnualRatio,
                       col = "blue")
    }

    # Plot the bias line
    if (!is.null(bias)) {
      bias_line <- data.frame(x = seq(min_x, max_x),
                              y = rep(bias, max_x - min_x + 1))
      graphics::lines(bias_line, col = "green4", lty = 2)
      lab_list <- c(lab_list, "Bias")
      col_list <- c(col_list, "green4")
      lty_list <- c(lty_list, 2)
      lwd_list <- c(lwd_list, 1)
      pch_list <- c(pch_list, NA)
      pt.cex_list <- c(pt.cex_list, NA)
    }

    # Add the legend
    if (!is.null(legendPos)) {
      graphics::legend(x = legendPos,
                       bty = "n",
                       inset = c(0.025, 0.025),
                       cex = 0.8,
                       legend = lab_list,
                       col = col_list,
                       lty = lty_list,
                       lwd = lwd_list,
                       pch = pch_list,
                       pt.cex = pt.cex_list)
    }
  }

  invisible(NULL)
}

#' Restore cross-sectional (contemporaneous) linear constraints
#'
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/tsbalancing.html}})}
#' 
#' \emph{Replication of the G-Series 2.0 SAS\eqn{^\circledR}{®} \strong{GSeriesTSBalancing} macro. See the
#' G-Series 2.0 documentation for details (Statistics Canada 2016).}
#'
#' This function balances (reconciles) a system of time series according to a set of linear constraints. The balancing 
#' solution is obtained by solving one or several quadratic minimization problems (see section **Details**) with the OSQP 
#' solver (Stellato et al. 2020). Given the feasibility of the balancing problem(s), the resulting time series data respect 
#' the specified constraints for every time period. Linear equality and inequality constraints are allowed. Optionally, the 
#' preservation of temporal totals may also be specified.
#' 
#' 
#' @usage
#' tsbalancing(
#'   in_ts,
#'   problem_specs_df,
#'   temporal_grp_periodicity = 1,
#'   temporal_grp_start = 1,
#'   osqp_settings_df = default_osqp_sequence,
#'   display_level = 1,
#'   alter_pos = 1,
#'   alter_neg = 1,
#'   alter_mix = 1,
#'   alter_temporal = 0,
#'   lower_bound = -Inf,
#'   upper_bound = Inf,
#'   tolV = 0,
#'   tolV_temporal = 0,
#'   tolP_temporal = NA,
#'
#'   # New in G-Series 3.0
#'   validation_tol = 0.001,
#'   trunc_to_zero_tol = validation_tol,
#'   full_sequence = FALSE,
#'   validation_only = FALSE,
#'   quiet = FALSE
#' )
#' 
#'
#' @param in_ts (mandatory) 
#' 
#' Time series (object of class "ts" or "mts") that contains the time series data to be reconciled. 
#' They are the balancing problems' input data (initial solutions).
#' 
#' @param problem_specs_df (mandatory) 
#' 
#' Balancing problem specifications data frame (object of class "data.frame"). Using a sparse format inspired from the 
#' SAS/OR\eqn{^\circledR}{®} LP procedure’s *sparse data input format* (SAS Institute 2015), it contains only the relevant 
#' information such as the nonzero coefficients of the balancing constraints as well as the non-default alterability 
#' coefficients and lower/upper bounds (i.e., values that would take precedence over those defined with arguments `alter_pos`, 
#' `alter_neg`, `alter_mix`, `alter_temporal`, `lower_bound` and `upper_bound`).
#' 
#' The information is provided using four mandatory variables (`type`, `col`, `row` and `coef`) and one optional variable 
#' (`timeVal`). An observation (a row) in the problem specs data frame either defines a label for one of the seven types of the 
#' balancing problem elements with columns `type` and `row` (see *Label definition records* below) or specifies coefficients 
#' (numerical values) for those balancing problem elements with variables `col`, `row`, `coef` and `timeVal` (see *Information 
#' specification records* below).
#' 
#' * **Label definition records** (`type` is not missing (is not `NA`))
#' 
#'   * `type` (chr): reserved keyword identifying the type of problem element being defined: 
#'     * `EQ`: equality (\eqn{=}) balancing constraint
#'     * `LE`: lower or equal (\eqn{\le}{<=}) balancing constraint
#'     * `GE`: greater or equal (\eqn{\ge}{>=}) balancing constraint 
#'     * `lowerBd`: period value lower bound
#'     * `upperBd`: period value upper bound
#'     * `alter`: period values alterability coefficient
#'     * `alterTmp`: temporal total alterability coefficient
#'   * `row` (chr): label to be associated to the problem element (_`type` keyword_)
#'   * _all other variables are irrelevant and should contain missing data (`NA` values)_ \cr \cr
#' 
#' * **Information specification records** (`type` is missing (is `NA`))
#' 
#'   * `type` (chr): not applicable (`NA`)
#'   * `col` (chr): series name or reserved word `_rhs_` to specify a balancing constraint right-hand side (RHS) value.
#'   * `row` (chr): problem element label.
#'   * `coef` (num): problem element value:
#'     * balancing constraint series coefficient or RHS value
#'     * series period value lower or upper bound
#'     * series period value or temporal total alterability coefficient
#'   * `timeVal` (num): optional time value to restrict the application of series bounds or alterability coefficients 
#'   to a specific time period (or temporal group). It corresponds to the time value, as returned by `stats::time()`, of a given 
#'   input time series (argument `in_ts`) period (observation) and is conceptually equivalent to \eqn{year + (period - 1) / 
#'   frequency}.
#'
#' Note that empty strings (`""` or `''`) for character variables are interpreted as missing (`NA`) by the function. Variable 
#' `row` identifies the elements of the balancing problem and is the key variable that makes the link between both types of 
#' records. The same label (`row`) cannot be associated with more than one type of problem element (`type`) and multiple labels 
#' (`row`) cannot be defined for the same given type of problem element (`type`), except for balancing constraints (values 
#' `"EQ"`, `"LE"` and `"GE"` of column `type`). User-friendly features of the problem specs data frame include:
#' * The order of the observations (rows) is not important.
#' * Character values (variables `type`, `row` and `col`) are not case sensitive (e.g., strings `"Constraint 1"` and 
#' `"CONSTRAINT 1"` for `row` would be considered as the same problem element label), except when `col` is used to specify a 
#' series name (a column of the input time series object) where **case sensitivity is enforced**.
#' * The variable names of the problem specs data frame are also not case sensitive (e.g., `type`, `Type` or `TYPE` are all 
#' valid) and `time_val` is an accepted variable name (instead of `timeVal`).
#' 
#' Finally, the following table lists valid aliases for the *`type` keywords* (type of problem element):
#' | **Keyword** | **Aliases** |
#' |:-----------:|:------------|
#' | `EQ`        | `==`, `=` |
#' | `LE`        | `<=`, `<` |
#' | `GE`        | `>=`, `>` |
#' | `lowerBd`   | `lowerBound`, `lowerBnd`, + *same terms with '_', '.' or ' ' between words*|
#' | `upperBd`   | `upperBound`, `upperBnd`, + *same terms with '_', '.' or ' ' between words*|
#' | `alterTmp`  | `alterTemporal`, `alterTemp`, + *same terms with '_', '.' or ' ' between words*|
#' 
#' Reviewing the **Examples** should help conceptualize the balancing problem specifications data frame.
#' 
#' @param temporal_grp_periodicity (optional)
#'
#' Positive integer defining the number of periods in temporal groups for which the totals should be preserved.
#' E.g., specify `temporal_grp_periodicity = 3` with a monthly time series for quarterly total preservation and
#' `temporal_grp_periodicity = 12` (or `temporal_grp_periodicity = frequency(in_ts)`) for annual total preservation.
#' Specifying `temporal_grp_periodicity = 1` (*default*) corresponds to period-by-period processing without
#' temporal total preservation.
#'
#' **Default value** is `temporal_grp_periodicity = 1` (period-by-period processing without temporal total preservation).
#'
#' @param temporal_grp_start (optional)
#'
#' Integer in the \[1 .. `temporal_grp_periodicity`\] interval specifying the starting period (cycle) for temporal
#' total preservation. E.g., annual totals corresponding to fiscal years defined from April to March of the
#' following year would be specified with `temporal_grp_start = 4` for a monthly time series (`frequency(in_ts) = 12`)
#' and `temporal_grp_start = 2` for a quarterly time series (`frequency(in_ts) = 4`). This argument has no effect
#' for period-by-period processing without temporal total preservation (`temporal_grp_periodicity = 1`).
#'
#' **Default value** is `temporal_grp_start = 1`.
#' 
#' @param osqp_settings_df (optional)
#' 
#' Data frame (object of class "data.frame") containing a sequence of OSQP settings for solving the balancing problems. 
#' The package includes two predefined OSQP settings sequence data frames:
#' * [default_osqp_sequence]: fast and effective (default);
#' * [alternate_osqp_sequence]: geared towards precision at the expense of execution time.
#' 
#' See `vignette("osqp-settings-sequence-dataframe")` for more details on this topic and to see the actual contents 
#' of these two data frames. Note that the concept of a *solving sequence* with different sets of solver settings is 
#' new in G-Series 3.0 (a single solving attempt was made in G-Series 2.0).
#'
#' **Default value** is `osqp_settings_df = default_osqp_sequence`.
#' 
#' @param display_level (optional)
#' 
#' Integer in the \[0 .. 3\] interval specifying the level of information to display in the console (`stdout()`). 
#' Note that specifying argument `quiet = TRUE` would *nullify* argument `display_level` (none of the following information
#' would be displayed).
#' 
#' | **Displayed information**                           |       **`0`**       |       **`1`**       |       **`2`**       |       **`3`**       |
#' |:----------------------------------------------------|:-------------------:|:-------------------:|:-------------------:|:-------------------:|
#' | Function header                                     | \eqn{\checkmark}{Y} | \eqn{\checkmark}{Y} | \eqn{\checkmark}{Y} | \eqn{\checkmark}{Y} |
#' | Balancing problem elements                          |                     | \eqn{\checkmark}{Y} | \eqn{\checkmark}{Y} | \eqn{\checkmark}{Y} |
#' | Individual problem solving details                  |                     |                     | \eqn{\checkmark}{Y} | \eqn{\checkmark}{Y} |
#' | Individual problem results (values and constraints) |                     |                     |                     | \eqn{\checkmark}{Y} |
#'
#' **Default value** is `display_level = 1`.
#'  
#' @param alter_pos (optional)
#' 
#' Nonnegative real number specifying the default alterability coefficient associated to the values of time series with **positive** 
#' coefficients in all balancing constraints in which they are involved (e.g., component series in aggregation table raking problems). 
#' Alterability coefficients provided in the problem specification data frame (argument `problem_specs_df`) override this value.
#'
#' **Default value** is `alter_pos = 1.0` (nonbinding values).
#' 
#' @param alter_neg (optional)
#' 
#' Nonnegative real number specifying the default alterability coefficient associated to the values of time series with **negative** 
#' coefficients in all balancing constraints in which they are involved (e.g., marginal totals in aggregation table raking problems). 
#' Alterability coefficients provided in the problem specification data frame (argument `problem_specs_df`) override this value.
#'
#' **Default value** is `alter_neg = 1.0` (nonbinding values).
#' 
#' @param alter_mix (optional)
#' 
#' Nonnegative real number specifying the default alterability coefficient associated to the values of time series with a mix of 
#' **positive and negative** coefficients in the balancing constraints in which they are involved. Alterability coefficients provided 
#' in the problem specification data frame (argument `problem_specs_df`) override this value.
#'
#' **Default value** is `alter_mix = 1.0` (nonbinding values).
#' 
#' @param alter_temporal (optional)
#' 
#' Nonnegative real number specifying the default alterability coefficient associated to the time series temporal totals. 
#' Alterability coefficients provided in the problem specification data frame (argument `problem_specs_df`) override this value.
#'
#' **Default value** is `alter_temporal = 0.0` (binding values).
#' 
#' @param lower_bound (optional)
#' 
#' Real number specifying the default lower bound for the time series values. Lower bounds provided in the problem specification 
#' data frame (argument \ifelse{latex}{\code{problem _specs_df}}{\code{problem_specs_df}}) override this value.
#'
#' **Default value** is `lower_bound = -Inf` (unbounded).
#' 
#' @param upper_bound (optional)
#' 
#' Real number specifying the default upper bound for the time series values. Upper bounds provided in the problem specification 
#' data frame (argument \ifelse{latex}{\code{problem _specs_df}}{\code{problem_specs_df}}) override this value.
#'
#' **Default value** is `upper_bound = Inf` (unbounded).
#' 
#' @param tolV (optional)
#' 
#' Nonnegative real number specifying the tolerance, in absolute value, for the balancing constraints right-hand side (RHS) 
#' values:
#' * `EQ` constraints: \eqn{\quad A\mathbf{x} = \mathbf{b} \quad}{Ax = b} become \eqn{\quad \mathbf{b} - \epsilon \le A\mathbf{x} 
#' \le \mathbf{b} + \epsilon}{b - eps <= Ax <= b + eps}
#' * `LE` constraints: \eqn{\quad A\mathbf{x} \le \mathbf{b} \quad}{Ax <= b} become \eqn{\quad A\mathbf{x} \le \mathbf{b} + 
#' \epsilon}{Ax <= b + eps}
#' * `GE` constraints: \eqn{\quad A\mathbf{x} \ge \mathbf{b} \quad}{Ax >= b} become \eqn{\quad A\mathbf{x} \ge \mathbf{b} - 
#' \epsilon}{Ax >= b - eps}
#' 
#' where \eqn{\epsilon}{eps} is the tolerance specified with `tolV`. This argument does not apply to the *period value (lower 
#' and upper) bounds* specified with arguments `lower_bound` and `upper_bound` or in the problem specs data frame (argument 
#' `prob_specs_df`). I.e., `tolV` does not affect the time series values lower and upper bounds, unless they are specified as 
#' *balancing constraints* instead (with `GE` and `LE` constraints in the problem specs data frame).
#' 
#' **Default value** is `tolV = 0.0` (no tolerance).
#' 
#' @param tolV_temporal,tolP_temporal (optional)
#' 
#' Nonnegative real number, or `NA`, specifying the tolerance, in percentage (`tolP_temporal`) or absolute value 
#' (`tolV_temporal`), for the implicit temporal aggregation constraints associated to **binding temporal totals** 
#' \eqn{\left( \sum_t{x_{i,t}} = \sum_t{y_{i,t}} \right)}{(sum_t{x_{i,t}} = sum_t{y_{i,t}})}, which become:
#' \deqn{\sum_t{y_{i,t}} - \epsilon_\text{abs} \le \sum_t{x_{i,t}} \le \sum_t{y_{i,t}} + \epsilon_\text{abs}}{sum_t{y_{i,t}} - 
#' eps_abs <= sum_t{x_{i,t}} <= sum_t{y_{i,t}} + eps_abs}
#' or 
#' \deqn{\sum_t{y_{i,t}} \left( 1 - \epsilon_\text{rel} \right) \le \sum_t{x_{i,t}} \le \sum_t{y_{i,t}} \left( 1 + 
#' \epsilon_\text{rel} \right)}{sum_t{y_{i,t}} (1 - eps_rel) <= sum_t{x_{i,t}} <= sum_t{y_{i,t}} (1 + eps_rel)}
#' 
#' where \eqn{\epsilon_\text{abs}}{eps_abs} and \eqn{\epsilon_\text{rel}}{eps_rel} are the absolute and percentage tolerances 
#' specified respectively with `tolV_temporal` and  `tolP_temporal`. Both arguments cannot be specified together (one must be 
#' specified while the other must be `NA`).
#'
#' **Example:** to set a tolerance of 10 *units*, specify \ifelse{latex}{\code{tolV_temporal = 10, tolP_temporal} \code{= NA}}{
#' \code{tolV_temporal = 10, tolP_temporal = NA}}; to set a tolerance of 1%, specify`tolV_temporal = NA, tolP_temporal = 0.01`. 
#' 
#' **Default values** are `tolV_temporal = 0.0` and `tolP_temporal = NA` (no tolerance).
#' 
#' @param validation_tol (optional)
#' 
#' Nonnegative real number specifying the tolerance for the validation of the balancing results. The function verifies if 
#' the final (reconciled) time series values meet the constraints, allowing for discrepancies up to the value specified with 
#' this argument. A warning is issued as soon as one constraint is not met (discrepancy greater than `validation_tol`).
#' 
#' With constraints defined as \eqn{\mathbf{l} \le A\mathbf{x} \le \mathbf{u}}{l <= Ax <= u}, where \eqn{\mathbf{l = u}}{l = u} 
#' for `EQ` constraints, \eqn{\mathbf{l} = -\infty}{l = -Inf} for `LE` constraints and \eqn{\mathbf{u} = \infty}{u = Inf} for 
#' `GE` constraints, **constraint discrepancies** correspond to \eqn{\max \left( 0, \mathbf{l} - A\mathbf{x}, A\mathbf{x} - 
#' \mathbf{u} \right)}{max(0, l - Ax, Ax - u)}, where constraint bounds \eqn{\mathbf{l}}{l} and \eqn{\mathbf{u}}{u} include 
#' the tolerances, when applicable, specified with arguments `tolV`, `tolV_temporal` and `tolP_temporal`.  
#' 
#' **Default value** is `validation_tol = 0.001`.
#' 
#' @param trunc_to_zero_tol (optional)
#' 
#' Nonnegative real number specifying the tolerance, in absolute value, for replacing by zero (small) values in the output 
#' (reconciled) time series data (output object `out_ts`). Specify `trunc_to_zero_tol = 0` to disable this *truncation to zero* 
#' process on the reconciled data. Otherwise, specify `trunc_to_zero_tol > 0` to replace with \eqn{0.0} any value in the 
#' \eqn{\left[ -\epsilon, \epsilon \right]}{[-eps, eps]} interval, where \eqn{\epsilon}{eps} is the tolerance specified with 
#' `trunc_to_zero_tol`. 
#' 
#' Note that the final constraint discrepancies (see argument `validation_tol`) are calculated on the *zero truncated* 
#' reconciled time series values, therefore ensuring accurate validation of the actual reconciled data returned by the 
#' function.
#' 
#' **Default value** is `trunc_to_zero_tol = validation_tol`.
#' 
#' @param full_sequence (optional)
#' 
#' Logical argument specifying whether all the steps of the *OSQP settings sequence data frame* should be performed or not. 
#' See argument `osqp_settings_df` and `vignette("osqp-settings-sequence-dataframe")` for more details on this topic.
#' 
#' **Default value** is `full_sequence = FALSE`.
#' 
#' @param validation_only (optional)
#' 
#' Logical argument specifying whether the function should only perform input data validation or not. When 
#' `validation_only = TRUE`, the specified *balancing constraints* and *period value (lower and upper) bounds* constraints 
#' are validated against the input time series data, allowing for discrepancies up to the value specified with argument 
#' `validation_tol`. Otherwise, when `validation_only = FALSE` (default), the input data are first reconciled and the 
#' resulting (output) data are then validated.
#' 
#' **Default value** is `validation_only = FALSE`.
#' 
#' @param quiet (optional)
#' 
#' Logical argument specifying whether or not to display only essential information such as warnings, errors and the period 
#' (or set of periods) being reconciled. You could further suppress, if desired, the display of the _balancing period(s)_ 
#' information by _wrapping_ your [tsbalancing()] call with [suppressMessages()]. In that case, the `proc_grp_df` output 
#' data frame can be used to identify (unsuccessful) balancing problems associated with warning messages (if any). Note that 
#' specifying `quiet = TRUE` would also *nullify* argument `display_level`.
#'
#' **Default value** is `quiet = FALSE`.
#'   
#'
#' @details 
#' This function solves one balancing problem per processing group (see section **Processing groups** for details). Each of these balancing 
#' problems is a quadratic minimization problem of the following form:
#' \deqn{\displaystyle 
#' \begin{aligned}
#' & \underset{\mathbf{x}}{\text{minimize}} 
#' & & \mathbf{\left( y - x \right)}^{\mathrm{T}} W \mathbf{\left( y - x \right)} \\
#' & \text{subject to} 
#' & & \mathbf{l} \le A \mathbf{x} \le \mathbf{u}
#' \end{aligned}
#' }{min(x) (y - x)' W (y - x), subject to l <= Ax <= u}
#' where
#' - \eqn{\mathbf{y}}{y} is the vector of the initial problem values, i.e., the initial time series period values and, when applicable, 
#' temporal totals;
#' - \eqn{\mathbf{x}}{x} is the final (reconciled) version of vector \eqn{\mathbf{y}}{y};
#' - matrix \eqn{W = \mathrm{diag} \left( \mathbf{w} \right)}{W = diag(w)} with vector \eqn{\mathbf{w}}{w} elements \eqn{w_i = \left\{
#'     \begin{array}{cl}
#'       0 & \text{if } |c_i y_i| = 0 \\
#'       \frac{1}{|c_i y_i|} & \text{otherwise}
#'     \end{array} \right.
#'     }{w_i = 0 if |c_i y_i| = 0, w_i = 1/|c_i y_i| otherwise}, 
#'     where \eqn{c_i} is the alterability coefficient of problem value \eqn{y_i} and cases corresponding to \eqn{|c_i y_i| 
#'     = 0} are fixed problem values (binding period values or temporal totals);
#' - matrix \eqn{A} and vectors \eqn{\mathbf{l}}{l} and \eqn{\mathbf{u}}{u} specify the _balancing constraints_, the _implicit 
#' temporal total aggregation constraints_ (when applicable), the _period value (upper and lower) bounds_ as well as _\eqn{x_i 
#' = y_i} constraints for fixed \eqn{y_i} values_ \eqn{\left( \left| c_i y_i \right| = 0 \right)}{(|c_i y_i| = 0)}.
#'
#' In practice, the objective function of the problem solved by OSQP excludes constant term \eqn{\mathbf{y}^{\mathrm{T}} W 
#' \mathbf{y}}{y' W y}, therefore corresponding to \eqn{\mathbf{x}^{\mathrm{T}} W \mathbf{x} - 2 \left( \mathbf{w} \mathbf{y} 
#' \right)^{\mathrm{T}} \mathbf{x}}{x' W x - 2 (w y)' x}, and the fixed \eqn{y_i} values \eqn{\left( \left| c_i y_i \right| = 0 
#' \right)}{(|c_i y_i| = 0)} are removed from the problem, adjusting the constraints accordingly, i.e.:
#' - rows corresponding to the *\eqn{x_i = y_i} constraints for fixed \eqn{y_i} values* are removed from \eqn{A}, \eqn{
#' \mathbf{l}}{l} and \eqn{\mathbf{u}}{u};
#' - columns corresponding to fixed \eqn{y_i} values are removed from \eqn{A} while appropriately adjusting \eqn{
#' \mathbf{l}}{l} and \eqn{\mathbf{u}}{u}.
#'     
#'     
######
# This subsection differs slightly between `tsraking()` and `tsbalancing` and is therefore maintained for both functions 
# (in both sets of roxygen2 comments) as opposed to being shared with `roxygen2 tag `@inheritSection`.
# => the "Temporal total preservation* paragraph is the SAME, however: keep them "in sync"!.
######
#' ## Alterability Coefficients
#' Alterability coefficients are nonnegative numbers that change the relative cost of modifying an initial problem value. 
#' By changing the actual objective function to minimize, they allow the generation of a wide range of solutions. Since they 
#' appear in the denominator of the objective function (matrix \eqn{W}), the larger the alterability coefficient the less costly 
#' it is to modify a problem value (period value or temporal total) and, conversely, the smaller the alterability coefficient 
#' the more costly it becomes. This results in problem values with larger alterability coefficients proportionally changing more 
#' than the ones with smaller alterability coefficients. Alterability coefficients of \eqn{0.0} define fixed (binding) problem 
#' values while alterability coefficients greater than \eqn{0.0} define free (nonbinding) values. The default alterability 
#' coefficients are \eqn{0.0} for temporal totals (argument `alter_temporal`) and \eqn{1.0} for period values (arguments 
#' `alter_pos`, `alter_neg`, `alter_mix`). In the common case of aggregation table raking problems, the period values of the 
#' marginal totals (time series with a coefficient of \eqn{-1} in the balancing constraints) are usually binding (specified 
#' with `alter_neg = 0`) while the period values of the component series (time series with a coefficient of \eqn{1} in the 
#' balancing constraints) are usually nonbinding (specified with `alter_pos > 0`, e.g., `alter_pos = 1`). *Almost binding* 
#' problem values (e.g., marginal totals or temporal totals) can be obtained in practice by specifying very small (almost 
#' \eqn{0.0}) alterability coefficients relative to those of the other (nonbinding) problem values.
#' 
#' **Temporal total preservation** refers to the fact that temporal totals, when applicable, are usually kept “as close as 
#' possible” to their initial value. *Pure preservation* is achieved by default with binding temporal totals while the change 
#' is minimized with nonbinding temporal totals (in accordance with the set of alterability coefficients).
#' 
#' 
#' ## Validation and troubleshooting
#' Successful balancing problems (problems with a valid solution) have `sol_status_val > 0` or, equivalently, 
#' `n_unmet_con = 0` or `max_discr <= validation_tol` in the output `proc_grp_df` data frame. Troubleshooting 
#' unsuccessful balancing problems is not necessarily straightforward. Following are some suggestions:
#' 
#' - Investigate the failed constraints (`unmet_flag = TRUE` or, equivalently, `discr_out > validation_tol` in the 
#' output `prob_con_df` data frame) to make sure that they do not cause an empty solution space (infeasible problem).
#' 
#' - Change the OSQP solving sequence. E.g., try:
#'   1. argument `full_sequence = TRUE`
#'   2. argument `osqp_settings_df = alternate_osqp_sequence`
#'   3. arguments `osqp_settings_df = alternate_osqp_sequence` and `full_sequence = TRUE`
#'   
#'   See `vignette("osqp-settings-sequence-dataframe")` for more details on this topic.
#'   
#' - Increase (review) the `validation_tol` value. Although this may sound like *cheating*, the default `validation_tol` 
#' value (\eqn{1 \times 10^{-3}}) may actually be too small for balancing problems that involve very large values (e.g., 
#' in billions) or, conversely, too large with very small problem values (e.g, \eqn{< 1.0}). Multiplying the average scale 
#' of the problem data by the *machine tolerance* (`.Machine$double.eps`) gives an approximation of the average size of the 
#' discrepancies that [tsbalancing()] should be able to handle (distinguish from \eqn{0}) and should probably constitute an 
#' **absolute lower bound** for argument `validation_tol`. In practice, a reasonable `validation_tol` value would likely be 
#' \eqn{1 \times 10^3} to \eqn{1 \times 10^6} times larger than this *lower bound*.
#'   
#' - Address constraints redundancy. Multi-dimensional aggregation table raking problems are over-specified (involve 
#' redundant constraints) when all totals of all dimensions of the *data cube* are binding (fixed) and a constraint is 
#' defined for all of them. Redundancy also occurs for the implicit temporal aggregation constraints in single- or 
#' multi-dimensional aggregation table raking problems with binding (fixed) temporal totals. Over-specification is generally 
#' not an issue for [tsbalancing()] if the input data are not contradictory with regards to the redundant constraints, i.e., 
#' if there are no inconsistencies (discrepancies) associated to the redundant constraints in the input data or if they are 
#' *negligible* (reasonably small relative to the scale of the problem data). Otherwise, this may lead to unsuccessful 
#' balancing problems with [tsbalancing()]. Possible solutions would then include:
#'   1. Resolve (or reduce) the discrepancies associated to the redundant constraints in the input data.
#'   2. Select one marginal total in every dimension, but one, of the data cube and remove the corresponding balancing 
#'   constraints from the problem. *This cannot be done for the implicit temporal aggregation constraints*.
#'   3. Select one marginal total in every dimension, but one, of the data cube and make them nonbinding (alterability 
#'   coefficient of, say, \eqn{1.0}). 
#'   4. Do the same as (3) for the temporal totals of one of the inner-cube component series (make them nonbinding).
#'   5. Make all marginal totals of every dimension, but one, of the data cube *amlost binding*, i.e., specify very small 
#'   alterability coefficients (say \eqn{1 \times 10^{-6}}) compared to those of the inner-cube component series. 
#'   6. Do the same as (5) for the temporal totals of all inner-cube component series (very small alterability 
#'   coefficients, e.g., with argument `alter_temporal`).
#'   7. Use [tsraking()] (if applicable), which handles these inconsistencies by using the Moore-Penrose inverse  
#'   (uniform distribution among all binding totals).
#'        
#'   Solutions (2) to (7) above should only be considered if the discrepancies associated to the redundant constraints 
#'   in the input data are *reasonably small* as they would be distributed among the omitted or nonbinding totals with 
#'   [tsbalancing()] and all binding totals with [tsraking()]. Otherwise, one should first investigate solution (1) above.
#' 
#' - Relax the bounds of the problem constraints, e.g.:
#'   - argument `tolV` for the balancing constraints;
#'   - arguments `tolV_temporal` and `tolP_temporal` for the implicit temporal aggregation constraints;
#'   - arguments `lower_bound` and `upper_bound`.
#' 
#' 
#' # Processing groups
#' The set of periods of a given reconciliation (raking or balancing) problem is called a *processing group* 
#' and either corresponds to:
#' - a **single period** with period-by-period processing or, when preserving temporal totals, for the individual 
#' periods of an incomplete temporal group (e.g., an incomplete year)
#' - or the **set of periods of a complete temporal group** (e.g., a complete year) when preserving temporal 
#' totals. 
#' 
#' The total number of processing groups (total number of reconciliation problems) depends on the set of 
#' periods in the input time series object (argument `in_ts`) and on the value of arguments 
#' `temporal_grp_periodicity` and `temporal_grp_start`. 
#' 
#' Common scenarios include `temporal_grp_periodicity = 1` (default) for period-by period processing without 
#' temporal total preservation and `temporal_grp_periodicity = frequency(in_ts)` for the preservation of annual 
#' totals (calendar years by default). Argument `temporal_grp_start` allows the specification of other types of 
#' (_non-calendar_) years. E.g., fiscal years starting on April correspond to `temporal_grp_start = 4` with monthly 
#' data and `temporal_grp_start = 2` with quarterly data. Preserving quarterly totals with monthly data would 
#' correspond to \ifelse{latex}{\code{temporal_grp _periodicity = 3}}{\code{temporal_grp_periodicity = 3}}. 
#' 
#' By default, temporal groups covering more than a year (i.e., corresponding to \ifelse{latex}{\code{
#' temporal_grp _periodicity > frequency(in_ts)}}{\code{temporal_grp_periodicity > frequency(in_ts)}} start on a 
#' year that is a multiple of \ifelse{latex}{\code{ceiling(temporal_grp _periodicity / frequency(in_ts))}}{\code{
#' ceiling(temporal_grp_periodicity / frequency(in_ts))}}. E.g., biennial groups corresponding to \ifelse{latex}{
#' \code{temporal_grp_per iodicity = 2 * frequency(in_ts)}}{\code{temporal_grp_periodicity = 2 * frequency(in_ts)}}
#' start on an _even year_ by default. This behaviour can be changed with argument `temporal_grp_start`. E.g., the 
#' preservation of biennial totals starting on an _odd year_ instead of an _even year_ (default) corresponds to 
#' `temporal_grp_start = frequency(in_ts) + 1` (along with `temporal_grp_periodicity = 2 * frequency(in_ts)`).
#' 
#' See the [gs.build_proc_grps()] **Examples** for common processing group scenarios.
#' 
#' 
#' # Comparing [tsraking()] and [tsbalancing()]
#' - [tsraking()] is limited to one- and two-dimensional aggregation table raking problems (with temporal 
#' total preservation if required) while [tsbalancing()] handles more general balancing problems (e.g., 
#' higher dimensional raking problems, nonnegative solutions, general linear equality and inequality 
#' constraints as opposed to aggregation rules only, etc.).
#' - [tsraking()] returns the generalized least squared solution of the Dagum and Cholette regression-based 
#' raking model (Dagum and Cholette 2006) while [tsbalancing()] solves the corresponding quadratic 
#' minimization problem using a numerical solver. In most cases, _convergence to the minimum_ is achieved 
#' and the [tsbalancing()] solution matches the (exact) [tsraking()] least square solution. It may not be 
#' the case, however, if convergence could not be achieved after a reasonable number of iterations. 
#' Having said that, only in very rare occasions will the [tsbalancing()] solution _significantly_ differ 
#' from the [tsraking()] solution.
#' - [tsbalancing()] is usually faster than [tsraking()], especially for large raking problems, but is 
#' generally more sensitive to the presence of (small) inconsistencies in the input data associated to 
#' the redundant constraints of fully specified (over-specified) raking problems. [tsraking()] handles 
#' these inconsistencies by using the Moore-Penrose inverse (uniform distribution among all binding totals). 
#' - [tsbalancing()] accommodates the specification of sparse problems in their reduced form. This is not 
#' true in the case of [tsraking()] where aggregation rules must always be fully specified since a 
#' _complete data cube_ without missing data is expected as input (every single _inner-cube_ component 
#' series must contribute to all dimensions of the cube, i.e., to every single _outer-cube_ marginal 
#' total series).
#' - Both tools handle negative values in the input data differently by default. While the solutions of 
#' raking problems obtained from [tsbalancing()] and [tsraking()] are identical when all input data 
#' points are positive, they will differ if some data points are negative (unless argument 
#' `Vmat_option = 2` is specified with [tsraking()]).
#' - While both [tsbalancing()] and [tsraking()] allow the preservation of temporal totals, time 
#' management is not incorporated in [tsraking()]. For example, the construction of the processing groups 
#' (sets of periods of each raking problem) is left to the user with [tsraking()] and separate calls must 
#' be submitted for each processing group (each raking problem). That's where helper function 
#' [tsraking_driver()] comes in handy with [tsraking()].
#' - The marginal totals (_outer-cube_ marginal total series) returned by [tsraking()] are always 
#' recalculated as the sum of the relevant component series (_inner-cube_ component series). The output 
#' values of **binding** marginal totals may therefore slightly differ from the input values. The 
#' magnitude of the differences depends on the initial inconsistencies (that end up being uniformly 
#' distributed by the Moore-Penrose inverse). This is not the case for [tsbalancing()] where the values 
#' of **binding** marginal totals always remain perfectly unchanged, potentially resulting in (small) 
#' discrepancies between the sum of the returned component series and the corresponding marginal totals.
#' - [tsbalancing()] returns the same set of series as the input time series object while [tsraking()] 
#' returns the set of series involved in the raking problem plus those specified with argument `id` 
#' (which could correspond to a subset of the input series).
#'
#'
#' @returns
#' The function returns is a list of seven objects:
#' 
#' - `out_ts`: modified version of the input time series object (class "ts" or "mts"; see argument `in_ts`) with the resulting 
#' reconciled time series values (primary function output). It can be explicitly coerced to another type of object with the 
#' appropriate `as*()` function (e.g., `tsibble::as_tsibble()` would coerce it to a tsibble).
#' 
#' - `proc_grp_df`: processing group summary data frame, useful to identify problems that have succeeded or failed. 
#' It contains one observation (row) for each balancing problem with the following columns:
#'   - `proc_grp` (num): processing group id.
#'   - `proc_grp_type` (chr): processing group type. Possible values are:
#'     - `"period"`;
#'     - `"temporal group"`.
#'   - `proc_grp_label` (chr): string describing the processing group in the following format:
#'     - `"<year>-<period>"` (single periods)
#'     - `"<start year>-<start period> - <end year>-<end period>"` (temporal groups)
#'   - `sol_status_val`, `sol_status` (num, chr): solution status numerical (integer) value and description string:
#'     - ` 1`: `"valid initial solution"`;
#'     - `-1`: `"invalid initial solution"`;
#'     - ` 2`: `"valid polished osqp solution"`;
#'     - `-2`: `"invalid polished osqp solution"`;
#'     - ` 3`: `"valid unpolished osqp solution"`;
#'     - `-3`: `"invalid unpolished osqp solution"`;
#'     - `-4`: `"unsolvable fixed problem"` (invalid initial solution).
#'   - `n_unmet_con` (num): number of unmet constraints (`sum(prob_conf_df$unmet_flag)`).
#'   - `max_discr` (num): maximum constraint discrepancy (`max(prob_conf_df$discr_out)`).
#'   - `validation_tol` (num): specified tolerance for validation purposes (argument `validation_tol`).
#'   - `sol_type` (chr): returned solution type. Possible values are:
#'     - `"initial"` (initial solution, i.e., input data values);
#'     - `"osqp"` (OSQP solution).
#'   - `osqp_attempts` (num): number of attempts made with OSQP (depth achieved in the solving sequence).
#'   - `osqp_seqno` (num): step # of the solving sequence corresponding to the returned solution. `NA` when 
#'   `sol_type = "initial"`.
#'   - `osqp_status` (chr): OSQP status description string (`osqp_sol_info_df$status`). `NA` when `sol_type = "initial"`.
#'   - `osqp_polished` (logi): `TRUE` if the returned OSQP solution is polished (\ifelse{latex}{\code{osqp_sol_info_df 
#'   $status_polish = 1}}{\code{osqp_sol_info_df$status_polish = 1}}), `FALSE` otherwise. `NA` when `sol_type = "initial"`.
#'   - `total_solve_time` (num): total time, in seconds, of the solving sequence.
#'   
#'   Column `proc_grp` constitutes a *unique key* (distinct rows) for the data frame. Successful balancing problems 
#'   (problems with a valid solution) correspond to rows with `sol_status_val > 0` or, equivalently, to `n_unmet_con = 0` 
#'   or to `max_discr <= validation_tol`. The *initial solution* (`sol_type = "initial"`) is returned only if **a)** 
#'   there are no initial constraint discrepancies, **b)** the problem is fixed (all values are binding) or **c)** it beats 
#'   the OSQP solution (smaller total constraint discrepancies). The OSQP solving sequence is described in \ifelse{latex}{
#'   \code{vignette("osqp-settings -sequence-dataframe")}}{\code{vignette("osqp-settings-sequence-dataframe")}}.
#'     
#' - `periods_df`: time periods data frame, useful to match periods to processing groups. It contains one observation 
#' (row) for each period of the input time series object (argument `in_ts`) with the following columns:
#'   - `proc_grp` (num): processing group id.
#'   - `t` (num): time id (`1:nrow(in_ts)`).
#'   - `time_val` (num): time value (`stats::time(in_ts)`). It conceptually corresponds to \eqn{year + (period - 1) / frequency}.
#'   
#'   Columns `t` and `time_val` both constitute a *unique key* (distinct rows) for the data frame.
#' 
#' - `prob_val_df`: problem values data frame, useful to analyze change diagnostics, i.e., initial vs final (reconciled) 
#' values. It contains one observation (row) for each value involved in each balancing problem, with the following columns:
#'   - `proc_grp` (num): processing group id.
#'   - `val_type` (chr): problem value type. Possible values are:
#'     - `"period value"`;
#'     - `"temporal total"`.
#'   - `name` (chr): time series (variable) name.
#'   - `t` (num): time id (`1:nrow(in_ts)`); id of the first period of the temporal group for a *temporal total*.
#'   - `time_val` (num): time value (`stats::time(in_ts)`); value of the first period of the temporal group for a 
#'   *temporal total*. It conceptually corresponds to \eqn{year + (period - 1) / frequency}.
#'   - `lower_bd`, `upper_bd` (num): period value bounds; always `-Inf` and `Inf` for a *temporal total*.
#'   - `alter` (num): alterability coefficient.
#'   - `value_in`, `value_out` (num): initial and final (reconciled) values.
#'   - `dif` (num): `value_out - value_in`.
#'   - `rdif` (num): `dif / value_in`; `NA` if `value_in = 0`.
#'   
#'   Columns `val_type + name + t` and `val_type + name + time_val` both constitute a *unique key* (distinct rows) for the 
#'   data frame. Binding (fixed) problem values correspond to rows with `alter = 0` or `value_in = 0`. Conversely, nonbinding 
#'   (free) problem values correspond to rows with `alter != 0` and `value_in != 0`.
#'   
#' - `prob_con_df`: problem constraints data frame, useful for troubleshooting problems that failed (identify unmet 
#' constraints). It contains one observation (row) for each constraint involved in each balancing problem, with the following 
#' columns:
#'   - `proc_grp` (num): processing group id.
#'   - `con_type` (chr): problem constraint type. Possible values are:
#'     - `"balancing constraint"`;
#'     - `"temporal aggregation constraint"`;
#'     - `"period value bounds"`.
#'     
#'     While *balancing constraints* are specicied by the user, the other two types of constraints (*temporal aggregation 
#'     constraints* and *period value bounds*) are automatically added to the problem by the function (when applicable).
#'   - `name` (chr): constraint label or time series (variable) name.
#'   - `t` (num): time id (`1:nrow(in_ts)`); id of the first period of the temporal group for a *temporal aggregation constraint*.
#'   - `time_val` (num): time value (`stats::time(in_ts)`); value of the first period of the temporal group for a *temporal 
#'   aggregation constraint*. It conceptually corresponds to \eqn{year + (period - 1) / frequency}.
#'   - `l`, `u`, `Ax_in`, `Ax_out` (num): initial and final constraint elements \eqn{\left( \mathbf{l} \le A \mathbf{x} \le 
#'   \mathbf{u} \right)}{(l <= Ax <= u)}.
#'   - `discr_in`, `discr_out` (num): initial and final constraint discrepancies \eqn{\left( \max \left( 0, \mathbf{l} - A 
#'   \mathbf{x}, A \mathbf{x} - \mathbf{u} \right) \right)}{(max(0, l - Ax, Ax - u))}.
#'   - `validation_tol` (num): specified tolerance for validation purposes (argument `validation_tol`).
#'   - `unmet_flag` (logi): `TRUE` if the constraint is not met (`discr_out > validation_tol`), `FALSE` otherwise.
#'   
#'   Columns `con_type + name + t` and `con_type + name + time_val` both constitute a *unique key* (distinct rows) for the 
#'   data frame. Constraint bounds \eqn{\mathbf{l = u}}{l = u} for `EQ` constraints, \eqn{\mathbf{l} = -\infty}{l = -Inf} for 
#'   `LE` constraints, \eqn{\mathbf{u} = \infty}{u = Inf} for `GE` constraints, and include the tolerances, when applicable, 
#'   specified with arguments `tolV`, `tolV_temporal` and `tolP_temporal`.
#' 
#' - `osqp_settings_df`: OSQP settings data frame. It contains one observation (row) for each problem (processing group) 
#' solved with OSQP (`proc_grp_df$sol_type = "osqp"`), with the following columns:
#'   - `proc_grp` (num): processing group id.
#'   - one column corresponding to each element of the list returned by the `osqp::GetParams()` method applied 
#'   to a *OSQP solver object* (class "OSQP_Model" object as returned by [osqp::osqp()]), e.g.:
#'     - Maximum iterations (`max_iter`);
#'     - Primal and dual infeasibility tolerances (`eps_prim_inf` and `eps_dual_inf`);
#'     - Solution polishing flag (`polishing`);
#'     - Number of scaling iterations (`scaling`);
#'     - etc.
#'   - extra settings specific to [tsbalancing()]:
#'     - `prior_scaling` (logi): `TRUE` if the problem data were scaled (using the average of the free (nonbinding) problem 
#'     values as the scaling factor) prior to solving with OSQP, `FALSE` otherwise.
#'     - `require_polished` (logi): `TRUE` if a polished solution from OSQP (\ifelse{latex}{\code{osqp_sol_info_df 
#'     $status_polish = 1}}{\code{osqp_sol_info_df$status_polish = 1}}) was required for this step in order to end the solving 
#'     sequence, `FALSE` otherwise. See `vignette("osqp-settings-sequence-dataframe")` for more details on the solving sequence 
#'     used by [tsbalancing()].
#'     
#'   Column `proc_grp` constitutes a *unique key* (distinct rows) for the data frame. Visit 
#'   <https://osqp.org/docs/interfaces/solver_settings.html> for all available OSQP settings. Problems (processing groups) for 
#'   which the initial solution was returned (`proc_grp_df$sol_type = "initial"`) are not included in this data frame.
#' 
#' - `osqp_sol_info_df`: OSQP solution information data frame. It contains one observation (row) for each problem 
#' (processing group) solved with OSQP (`proc_grp_df$sol_type = "osqp"`), with the following columns:
#'   - `proc_grp` (num): processing group id.
#'   - one column corresponding to each element of the `info` list of a *OSQP solver object* (class "OSQP_Model" 
#'   object as returned by [osqp::osqp()]) after having been solved with the `osqp::Solve()` method, e.g.:
#'     - Solution status (`status` and `status_val`);
#'     - Polishing status (`status_polish`);
#'     - Number of iterations (`iter`);
#'     - Objective function value (`obj_val`);
#'     - Primal and dual residuals (`pri_res` and `dua_res`);
#'     - Solve time (`solve_time`);
#'     - etc.
#'   - extra information specific to [tsbalancing()]:
#'     - `prior_scaling_factor` (num): value of the scaling factor when \ifelse{latex}{\code{osqp_settings_df $prior_scaling 
#'     = TRUE}}{\code{osqp_settings_df$prior_scaling = TRUE}} (`prior_scaling_factor = 1.0` otherwise).
#'     - `obj_val_ori_prob` (num): original balancing problem's objective function value, which is the OSQP objective function 
#'     value (`obj_val`) on the original scale (when \code{osqp_settings_df$prior_scaling = TRUE}) plus the constant term of 
#'     the original balancing problem's objective function, i.e., \ifelse{latex}{\code{obj_val_ori_prob = obj_val * 
#'     prior _scaling_factor + <constant term>}}{\code{obj_val_ori_prob = obj_val * prior_scaling_factor + <constant term>}}, 
#'     where `<constant term>` corresponds to 
#'     \eqn{\mathbf{y}^{\mathrm{T}} W \mathbf{y}}{y' W y}. See section **Details** for the definition of vector 
#'     \eqn{\mathbf{y}}{y}, matrix \eqn{W} and, more generally speaking, the complete expression of the balancing problem's 
#'     objective function.
#' 
#'   Column `proc_grp` constitutes a *unique key* (distinct rows) for the data frame. Visit <https://osqp.org> for more information on 
#'   OSQP. Problems (processing groups) for which the initial solution was returned (`proc_grp_df$sol_type = "initial"`) are not 
#'   included in this data frame.
#'
#' Note that the "data.frame" objects returned by the function can be explicitly coerced to other types of objects with 
#' the appropriate `as*()` function (e.g., `tibble::as_tibble()` would coerce any of them to a tibble).
#'
#'
#' @references Dagum, E. B. and P. Cholette (2006). **Benchmarking, Temporal Distribution and Reconciliation Methods
#' of Time Series**. Springer-Verlag, New York, Lecture Notes in Statistics, Vol. 186. \doi{10.1007/0-387-35439-5}
#'
#' @references Ferland, M., S. Fortier and J. Bérubé (2016). "A Mathematical Optimization Approach to Balancing Time Series: 
#' Statistics Canada’s GSeriesTSBalancing". In **JSM Proceedings, Business and Economic Statistics Section**. Alexandria, 
#' VA: American Statistical Association. 2292-2306.
#' 
#' @references Ferland, M. (2018). "Time Series Balancing Quadratic Problem — Hessian matrix and vector of linear objective 
#' function coefficients". **Internal document**. Statistics Canada, Ottawa, Canada.
#'
#' @references Quenneville, B. and S. Fortier (2012). "Restoring Accounting Constraints in Time Series – Methods and
#' Software for a Statistical Agency". **Economic Time Series: Modeling and Seasonality**. Chapman & Hall, New York.
#' 
#' @references SAS Institute Inc. (2015). "The LP Procedure Sparse Data Input Format". **SAS/OR\eqn{^\circledR}{®} 14.1 
#' User's Guide: Mathematical Programming Legacy Procedures**. 
#' <https://support.sas.com/documentation/cdl/en/ormplpug/68158/HTML/default/viewer.htm#ormplpug_lp_details03.htm>
#' 
#' @references Statistics Canada (2016). "The ***GSeriesTSBalancing*** Macro". **G-Series 2.0 User Guide**.
#' Statistics Canada, Ottawa, Canada.
#'
#' @references Statistics Canada (2018). **Theory and Application of Reconciliation (Course code 0437)**.
#' Statistics Canada, Ottawa, Canada.
#'
#' @references Stellato, B., G. Banjac, P. Goulart et al. (2020). "OSQP: an operator splitting solver for quadratic programs". 
#' **Math. Prog. Comp. 12**, 637–672 (2020). \doi{10.1007/s12532-020-00179-2}
#'
#'
#' @seealso [tsraking()] [tsraking_driver()] [rkMeta_to_blSpecs()] [gs.build_proc_grps()] [build_balancing_problem()] [aliases]
#'
#'
#' @example misc/function_examples/tsbalancing-ex.R
#'
#'
#' @export
tsbalancing <- function(in_ts,
                        problem_specs_df,
                        temporal_grp_periodicity = 1,
                        temporal_grp_start = 1,
                        osqp_settings_df = default_osqp_sequence,
                        display_level = 1,
                        alter_pos = 1,
                        alter_neg = 1,
                        alter_mix = 1,
                        alter_temporal = 0,
                        lower_bound = -Inf,
                        upper_bound = Inf,
                        tolV = 0,
                        tolV_temporal = 0,
                        tolP_temporal = NA,

                        # New in G-Series 3.0
                        validation_tol = 0.001,
                        trunc_to_zero_tol = validation_tol,
                        full_sequence = FALSE,
                        validation_only = FALSE,
                        quiet = FALSE) {




  ### Internal functions ###

  # Display the Problem Specs data frame info
  #
  # Main (parent) function objects used in this function:
  #   - pb         : balancing problem core elements (building blocks), namely (elements of the list):
  #     - labels_df: problem specs labels
  #     - coefs_df : problem specs coefficients 
  #     - ser_names: vector of all series names involved in the balancing problem
  #     - lb       : lower bounds info (list object)
  #     - ub       : upper bounds info (list object)
  #     - alter    : period value alterability coefficients (list object)
  #     - altertmp : temporal total alterability coefficients (list object)
  #     - pos_ser  : vector of series names with only positive constraint coefficients (across all constraints)
  #     - neg_ser  : vector of series names with only negative constraint coefficients (across all constraints)
  #     - mix_ser  : vector of series names with both positive and negative constraint coefficients (across all constraints)
  #   - n_ser      : number of series (length of vector `pb$ser_names`)
  #   - n_con      : number of constraints (length of vector `pb$labels_df$row.lc[pb$labels_df$con.flag]`)
  #   - arguments `alter_pos`, `alter_neg`, `alter_mix`, `alter_temporal`, `lower_bound`, `upper_bound`, 
  #     `validation_only` and `temporal_grp_periodicity`
  print_specs <- function() {

    # Function to build the balancing constraint string for display
    build_con_str <- function(type_val, row_val, col_vec, coef_vec) {
      rhs_logic <- (col_vec == "_rhs_")
      if (any(rhs_logic)) {
        rhs_str <- format(coef_vec[rhs_logic], big.mark = ",")
      } else {
        rhs_str <- "0"
      }
      series <- col_vec[!rhs_logic]
      coefs <- coef_vec[!rhs_logic]

      # A loop is used as opposed to a single `paste0(, collapse = "")` essentially for optimized display
      # of each single series constraint coefficient (e.g., avoid having "200x - 10.5y + 300z >= 1,000"
      # being displayed as "200.0x - 10.5y + 300.0z >= 1,000", i.e., with a decimal everywhere)
      #   => the loop is not that costly (compared to `paste0(, collapse = "")`)
      tmp <- format(abs(coefs[1]), big.mark = ",")
      con_str <- paste0(ifelse(coefs[1] < 0, "-", ""), ifelse(tmp == "1", "", tmp), series[1])
      for (ii in seq.int(from = 2, length.out = length(coefs) - 1)) {
        tmp <- format(abs(coefs[ii]), big.mark = ",")
        con_str <- paste0(con_str, ifelse(coefs[ii] < 0, " - ", " + "), ifelse(tmp == "1", "", tmp), series[ii])
      }

      if (type_val == "eq") {
        op <- " == "
      } else if (type_val == "le") {
        op <- " <= "
      } else {
        op <- " >= "
      }

      paste0("  ", row_val, ":\n    ", con_str, op, rhs_str)
    }

    header <- "Balancing Problem Elements"
    message("\n\n\n", header, "\n", paste0(rep.int("=", nchar(header)), collapse = ""))

    # Display constraints
    con_labels_df <- pb$labels_df[pb$labels_df$con.flag, , drop = FALSE]
    header <- paste0("Balancing Constraints (", n_con, ")")
    message("\n\n  ", header, "\n  ", paste0(rep.int("-", nchar(header)), collapse = ""))
    for (ii in seq.int(n_con)) {
      logi_vec <- pb$coefs_df$row.lc == con_labels_df$row.lc[ii]
      message("\n", build_con_str(type_val = con_labels_df$type.lc[ii],
                                  row_val = con_labels_df$row[ii],
                                  col_vec = pb$coefs_df$col[logi_vec],
                                  coef_vec = pb$coefs_df$coef[logi_vec]))
    }

    
    # Function to pad the values of a character data frame column (character vector) to the right 
    # so that it is displayed as a left-aligned character column
    pad_right <- function(x) {
      len <- max(nchar(x))
      # `sapply()` is safe: it always returns a character vector (`x` is of length minimum 1)
      paste0(x, sapply(len - nchar(x), 
                       function(x) paste0(rep(" ", x), collapse = "")))
      
    }
    
    # Initialize the series info data frame (lower/upper bounds and alter/alterTmp coef values and source)
    series_info_df <- data.frame(
      name = pb$ser_names,
      lb_val = rep.int(lower_bound, n_ser),
      lb_per = rep.int("", n_ser),
      lb_src = rep.int(lower_bound_label, n_ser),
      ub_val = rep.int(upper_bound, n_ser),
      ub_per = rep.int("", n_ser),
      ub_src = rep.int(upper_bound_label, n_ser),
      alter_val = rep.int(NA_real_, n_ser),
      alter_per = rep.int("", n_ser),
      alter_src = rep.int(NA_character_, n_ser),
      altertmp_val = rep.int(alter_temporal, n_ser),
      altertmp_per = rep.int("", n_ser),
      altertmp_src = rep.int(alter_temporal_label, n_ser)
    )

    # Assign the lower/upper bounds and alter coefs
    dated_info <- FALSE
    col_id_vec <- 1
    series_info_df$lb_val[pb$lb$nondated_id_vec] <- pb$lb$nondated_coefs
    series_info_df$lb_src[pb$lb$nondated_id_vec] <- "(problem specs)"
    series_info_df$lb_src <- pad_right(series_info_df$lb_src)
    # Treat period-specific values
    if (length(pb$lb$dated_id_vec) > 0) {
      dated_info <- TRUE
      series_info_df$lb_per[pb$lb$dated_id_vec] <- "*"
      col_id_vec <- c(col_id_vec, 2, 3, 4)
    } else {
      col_id_vec <- c(col_id_vec, 2, 4)
    }
    series_info_df$ub_val[pb$ub$nondated_id_vec] <- pb$ub$nondated_coefs
    series_info_df$ub_src[pb$ub$nondated_id_vec] <- "(problem specs)"
    series_info_df$ub_src <- pad_right(series_info_df$ub_src)
    if (length(pb$ub$dated_id_vec) > 0) {
      dated_info <- TRUE
      series_info_df$ub_per[pb$ub$dated_id_vec] <- "*"
      col_id_vec <- c(col_id_vec, 5, 6, 7)
    } else {
      col_id_vec <- c(col_id_vec, 5, 7)
    }
    if (!validation_only) {
      id_vec <- match(pb$pos_ser, pb$ser_names)
      series_info_df$alter_val[id_vec] <- alter_pos
      series_info_df$alter_src[id_vec] <- alter_pos_label
      id_vec <- match(pb$neg_ser, pb$ser_names)
      series_info_df$alter_val[id_vec] <- alter_neg
      series_info_df$alter_src[id_vec] <- alter_neg_label
      id_vec <- match(pb$mix_ser, pb$ser_names)
      series_info_df$alter_val[id_vec] <- alter_mix
      series_info_df$alter_src[id_vec] <- alter_mix_label
      series_info_df$alter_val[pb$alter$nondated_id_vec] <- pb$alter$nondated_coefs
      series_info_df$alter_src[pb$alter$nondated_id_vec] <- "(problem specs)"
      series_info_df$alter_src <- pad_right(series_info_df$alter_src)
      if (length(pb$alter$dated_id_vec) > 0) {
        dated_info <- TRUE
        series_info_df$alter_per[pb$alter$dated_id_vec] <- "*"
        col_id_vec <- c(col_id_vec, 8, 9, 10)
      } else {
        col_id_vec <- c(col_id_vec, 8, 10)
      }
      if (temporal_grp_periodicity != 1) {
        series_info_df$altertmp_val[pb$altertmp$nondated_id_vec] <- pb$altertmp$nondated_coefs
        series_info_df$altertmp_src[pb$altertmp$nondated_id_vec] <- "(problem specs)"
        series_info_df$altertmp_src <- pad_right(series_info_df$altertmp_src)
        if (length(pb$altertmp$dated_id_vec) > 0) {
          dated_info <- TRUE
          series_info_df$altertmp_per[pb$altertmp$dated_id_vec] <- "*"
          col_id_vec <- c(col_id_vec, 11, 12, 13)
        } else {
          col_id_vec <- c(col_id_vec, 11, 13)
        }
      }
    }
    
    # Keep relevant columns and assign the (duplicate/empty) column names for display
    series_info_df <- data.frame(series_info_df[col_id_vec])
    cols_names <- c("name",
                    "lowerBd", "", "",
                    "upperBd", "", "",
                    "alter", "", "", 
                    "alterTmp", "", "")
    names(series_info_df) <- cols_names[col_id_vec]

    # Display the time series info
    header <- "Time Series Info"
    message("\n\n  ", header, "\n  ", paste0(rep.int("-", nchar(header)), collapse = ""))
    message("\n", paste0("  ", utils::capture.output(print.data.frame(series_info_df)), collapse = "\n"))
    if (dated_info) {
      message("\n  * indicates cases where period-specific values (`timeVal` is not `NA`) are specified in the problem specs data frame.")
    }
    message("\n")
  }


  # Functions to set up `sink()` in order to store "output" type display (sent to `stdout()`) in a text file,
  # wrap up (close) `sink()` and display the text file contents as a "message" (sent to `stderr()`)
  #
  # Argument:
  #   - txtfile: name and path of the (temporary) text file that will store the output
  #
  # Other main (parent) function objects used in this function:
  #   - osqp_output_file: name and path of the (temporary) text file to be deleted as the function exits
  #                       (updated using environment pointer `main.e` instead of operator `<<-`)
  sink_setup <- function(txtfile) {
    sink(file = txtfile, type = c("output", "message"))
    invisible(NULL)
  }
  sink_wrapup <- function(txtfile) {
    sink()
    main.e$osqp_output_file <- txtfile
    msg <- paste0("  ", readLines(txtfile), collapse = "\n")
    message("\n", msg)
    invisible(NULL)
  }


  # Display the balancing results (problem values and constraints data frames)
  #
  # Main (parent) function objects used in this function:
  #   - prob_val_df: balancing problem values data frame
  #   - prob_con_df: balancing problem constraints data frame
  print_results <- function() {

    
    # Problem values data frame
    
    header <- "Problem Values"
    lines <- paste0(rep.int("-", nchar(header)), collapse = "")
    message("\n  ", lines, "\n  ", header, "\n  ", lines, "\n")
    logi_vec <- (prob_val_df$val_type == "temporal total")
    if (sum(logi_vec) == 0) {
      df <- prob_val_df[, -c(1, 2)]
      message(paste0("  ", utils::capture.output(print.data.frame(df, row.names = FALSE, big_mark = ",")),
                     collapse = "\n"), "\n")
    } else {
      header <- "Period Values"
      lines <- paste0(rep.int("-", nchar(header)), collapse = "")
      message("  ", header, "\n  ", lines)
      df <- prob_val_df[!logi_vec, -c(1, 2)]
      message(paste0("  ", utils::capture.output(print.data.frame(df, row.names = FALSE)), collapse = "\n"), "\n")
      header <- "Temporal Totals"
      lines <- paste0(rep.int("-", nchar(header)), collapse = "")
      message("  ", header, "\n  ", lines)
      df <- prob_val_df[logi_vec, -c(1, 2, 4, 5, 6, 7)]
      message(paste0("  ", utils::capture.output(print.data.frame(df, row.names = FALSE)), collapse = "\n"), "\n")
    }

    
    # Problem constraints data frame
    
    header <- "Problem Constraints (l <= Ax <= u)"
    lines <- paste0(rep.int("-", nchar(header)), collapse = "")
    message("\n  ", lines, "\n  ", header, "\n  ", lines, "\n")

    if (length(unique(prob_con_df$con_type)) == 1) {
      df <- prob_con_df[, -c(1, 2)]
      message(paste0("  ", utils::capture.output(print.data.frame(df, row.names = FALSE)),
                     collapse = "\n"), "\n")
    } else {
      header <- "Balancing Constraints"
      lines <- paste0(rep.int("-", nchar(header)), collapse = "")
      message("  ", header, "\n  ", lines)
      df <- prob_con_df[prob_con_df$con_type == "balancing constraint", -c(1, 2)]
      message(paste0("  ", utils::capture.output(print.data.frame(df, row.names = FALSE)), collapse = "\n"), "\n")
      logi_vec <- (prob_con_df$con_type == "period value bounds")
      if (sum(logi_vec) > 0) {
        header <- "Period Value Bounds"
        lines <- paste0(rep.int("-", nchar(header)), collapse = "")
        message("  ", header, "\n  ", lines)
        df <- prob_con_df[logi_vec, -c(1, 2)]
        message(paste0("  ", utils::capture.output(print.data.frame(df, row.names = FALSE)),
                       collapse = "\n"), "\n")
      }
      logi_vec <- (prob_con_df$con_type == "temporal aggregation constraint")
      if (sum(logi_vec) > 0) {
        header <- "Temporal Aggregation Constraints"
        lines <- paste0(rep.int("-", nchar(header)), collapse = "")
        message("  ", header, "\n  ", lines)
        df <- prob_con_df[logi_vec, -c(1, 2, 4, 5)]
        message(paste0("  ", utils::capture.output(print.data.frame(df, row.names = FALSE)),
                       collapse = "\n"), "\n")
      }
    }
    message("")
  }




  ### Main function ###
  
  # Define a pointer to the main function environment and create a NULL object
  main.e <- environment()
  main.e$dummy <- NULL
  
  # Initialize the object to be returned by the function via `on.exit()`
  out_list <- NULL
  on.exit(return(out_list))
  warning_flag <- FALSE
  
  # Enforce the default R "error" option (`options(error = NULL)`). E.g. this Turns off traceback
  # generated by calls to the stop() function inside internal functions in R Studio.
  ini_error_opt <- getOption("error")
  on.exit(options(error = ini_error_opt), add = TRUE)
  options(error = NULL)

  # Validate argument `quiet` and implement the quiet setting
  quiet <- gs.validate_arg_logi(quiet)
  if (quiet) {
    quiet_msg_func <- gs.NULL_func
    quiet_lab <- ""  # won't be displayed anyway
    displayLevel_lab <- ""  # won't be displayed anyway
    specs_display_func <- gs.NULL_func
    solve_msg_func <- gs.NULL_func
    osqp_verbose <- FALSE
    osqpBeg_display_func <- gs.NULL_func
    osqpEnd_display_func <- gs.NULL_func
    osqp_output_file <- NULL
    results_display_func <- gs.NULL_func
  } else {
    quiet_msg_func <- message
    quiet_lab <- "    (*)quiet                 = FALSE (default)"
    
    tmp <- (unlist(display_level))[1]
    if (!identical(display_level, tmp) || is.null(tmp) || !(tmp %in% 0:3)) {
      stop("Argument 'display_level' must take value 0, 1, 2, or 3.\n\n", call. = FALSE)
    }
    display_level <- as.integer(display_level)
    displayLevel_lab <- paste0("    display_level            = ", format(display_level))
    width_opt <- getOption("width")
    if (width_opt < 128) {
      on.exit(options(width = width_opt), add = TRUE)
      options(width = 128)
    }
    if (display_level == 1) {
      displayLevel_lab <- paste0(displayLevel_lab, " (default)")
      specs_display_func <- print_specs
      solve_msg_func <- gs.NULL_func
      osqp_verbose <- FALSE
      osqpBeg_display_func <- gs.NULL_func
      osqpEnd_display_func <- gs.NULL_func
      osqp_output_file <- NULL
      results_display_func <- gs.NULL_func
    } else if (display_level == 0) {
      specs_display_func <- gs.NULL_func
      solve_msg_func <- gs.NULL_func
      osqp_verbose <- FALSE
      osqpBeg_display_func <- gs.NULL_func
      osqpEnd_display_func <- gs.NULL_func
      osqp_output_file <- NULL
      results_display_func <- gs.NULL_func
    } else {
      specs_display_func <- print_specs
      solve_msg_func <- message
      osqp_verbose <- TRUE
      osqpBeg_display_func <- sink_setup
      osqpEnd_display_func <- sink_wrapup
      osqp_output_file <- ".gseries_osqp_output.txt"
      on.exit(invisible(suppressWarnings(file.remove(osqp_output_file))), add = TRUE)
      if (display_level == 3) {
        results_display_func <- print_results
      } else {
        results_display_func <- gs.NULL_func
      }
    }
  }

  
  # Display the function header
  quiet_msg_func("\n\n", gs.header, "\n\ntsbalancing() function:\n")


  # Initial argument validation

  # Mandatory arguments (without default values)
  in_ts_name <- deparse1(substitute(in_ts))
  tmp <- nchar(in_ts_name)
  if (tmp == 0) {
    stop("Argument 'in_ts' is mandatory (it must be specified).\n\n", call. = FALSE)
  } else if (tmp >= 60) {
    in_ts_name <- paste0(substr(in_ts_name, 1, 55), "<...>")
  }
  if (grepl("structure(", in_ts_name, fixed = TRUE)) {
    in_ts_name <- "<argument 'in_ts'>"
  }
  in_ts <- in_ts
  if (!stats::is.ts(in_ts)) {
    stop("Argument 'in_ts' is not a 'ts' object.\n\n", call. = FALSE)
  }
  in_ts <- stats::as.ts(in_ts)
  ts_freq <- stats::frequency(in_ts)
  time_values <- as.numeric(stats::time(in_ts))
  periods <- gs.time2str(in_ts)
  n_per <- length(periods)
  
  specs_df_name <- deparse1(substitute(problem_specs_df))
  tmp <- nchar(specs_df_name)
  if (tmp == 0) {
    stop("Argument 'problem_specs_df' is mandatory (it must be specified).\n\n", call. = FALSE)
  } else if (tmp >= 60) {
    specs_df_name <- paste0(substr(specs_df_name, 1, 55), "<...>")
  }
  if (grepl("structure(", specs_df_name, fixed = TRUE)) {
    specs_df_name <- "<argument 'problem_specs_df'>"
  }
  problem_specs_df <- problem_specs_df
  if (!is.data.frame(problem_specs_df)) {
    stop("Argument 'problem_specs_df' is not a 'data.frame' object.\n\n", call. = FALSE)
  }
  problem_specs_df <- as.data.frame(problem_specs_df)
  row.names(problem_specs_df) <- NULL

  # Optional arguments (with default values).
  # => Note that the following optional arguments, only relevant when `validation_only == FALSE`,
  #    will be validated later:
  #      - `temporal_grp_periodicity`
  #      - `temporal_grp_start`
  #      - `osqp_settings_df`
  #      - `alter_pos`
  #      - `alter_neg`
  #      - `alter_mix`
  #      - `alter_temporal`
  #      - `tolV_temporal` and `tolP_temporal`
  #      - `trunc_to_zero_tol`
  tmp <- (unlist(lower_bound))[1]
  if (!identical(lower_bound, tmp) || is.null(tmp) || is.na(tmp) || !is.finite(tmp) && tmp != -Inf) {
    stop("Argument 'lower_bound' must be a finite real number or -Inf.\n\n", call. = FALSE)
  }
  tmp <- (unlist(upper_bound))[1]
  if (!identical(upper_bound, tmp) || is.null(tmp) || is.na(tmp) || !is.finite(tmp) && tmp != Inf) {
    stop("Argument 'upper_bound' must be a finite real number or Inf.\n\n", call. = FALSE)
  }
  if (lower_bound > upper_bound) {
    stop("Arguments `lower_bound` and 'upper_bound' are not compatible (`lower_bound = ",
         format(lower_bound, big.mark = ","), " > ", format(upper_bound, big.mark = ","),
         " = upper_bound`).\n\n", call. = FALSE)
  }
  tmp <- (unlist(tolV))[1]
  if (!identical(tolV, tmp) || is.null(tmp) || !is.finite(tmp) || is.finite(tmp) && tmp < 0) {
    stop("Argument 'tolV' must be a nonnegative finite real number.\n\n", call. = FALSE)
  }
  tmp <- (unlist(validation_tol))[1]
  if (!identical(validation_tol, tmp) || is.null(tmp) || !is.finite(tmp) || is.finite(tmp) && tmp < 0) {
    stop("Argument 'validation_tol' must be a nonnegative finite real number.\n\n", call. = FALSE)
  }
  validation_only <- gs.validate_arg_logi(validation_only)

  if (validation_only) {
    valid_only_lab <- "    (*)validation_only       = TRUE"
    solve_func <- return_initial_sol
    solve_str <- "Input Values Validation"
    solve_word <- "Validating"

    # Force period-by-period processing
    temporal_grp_periodicity <- 1
    temp_grp_per_lab <- "    temporal_grp_periodicity (ignored)"
    temporal_grp_start <- NA_integer_
    temp_grp_start_lab <- "    temporal_grp_start       (ignored)"

    # Set other ignored arguments
    settings_df <- osqp_settings_df
    settingsDF_lab <- "    osqp_settings_df         (ignored)"
    alter_pos <- NA_real_
    alterPos_lab <- "    alter_pos                (ignored)"
    alter_pos_label <- NA_character_
    alter_neg <- NA_real_
    alterNeg_lab <- "    alter_neg                (ignored)"
    alter_neg_label <- NA_character_
    alter_mix <- NA_real_
    alterMix_lab <- "    alter_mix                (ignored)"
    alter_mix_label <- NA_character_
    alter_temporal <- NA_real_
    alterTmp_lab <- "    alter_temporal           (ignored)"
    alter_temporal_label <- NA_character_
    tolV_temporal <- NA_real_
    tolP_temporal <- NA_real_
    tol_temporal_lab <- "    tolV_temporal            (ignored)"
    trunc_to_zero_tol <- NA_real_
    zero_trunc_lab <- "    (*)trunc_to_zero_tol     (ignored)"

  } else {
    valid_only_lab <- paste0("    (*)validation_only       = FALSE (default)")
    solve_func <- solve_one_osqp
    solve_str <- "Problem Solving"
    solve_word <- "Balancing"

    # Validate optional argument `temporal_grp_periodicity` (and set the header label)
    tmp <- (unlist(temporal_grp_periodicity))[1]
    if (!identical(temporal_grp_periodicity, tmp) || is.null(tmp) || !is.finite(tmp) ||
        is.finite(tmp) && (tmp <= 0 || tmp != as.integer(tmp))) {
      stop("Argument 'temporal_grp_periodicity' must be a positive integer.\n\n", call. = FALSE)
    }
    temporal_grp_periodicity <- as.integer(temporal_grp_periodicity)
    temp_grp_per_lab <- paste0("    temporal_grp_periodicity = ", format(temporal_grp_periodicity))
    if (temporal_grp_periodicity == 1) {
      temp_grp_per_lab <- paste0(temp_grp_per_lab, " (default)")
      temporal_grp_start <- NA_integer_
      temp_grp_start_lab <- paste0("    temporal_grp_start       (ignored)")
      alter_temporal <- NA_real_
      alterTmp_lab <- "    alter_temporal           (ignored)"
      alter_temporal_label <- NA_character_
      tolV_temporal <- NA_real_
      tolP_temporal <- NA_real_
      tol_temporal_lab <- "    tolV_temporal            (ignored)"
    } else {

      # Validate optional argument `temporal_grp_start` (and set the header label)
      tmp <- (unlist(temporal_grp_start))[1]
      if (!identical(temporal_grp_start, tmp) || is.null(tmp) || !is.finite(tmp) ||
          is.finite(tmp) && (tmp <= 0 || tmp > temporal_grp_periodicity || tmp != as.integer(tmp))) {
        stop("Argument 'temporal_grp_start' must be an integer in the [1..", temporal_grp_periodicity, "] interval.\n\n",
             call. = FALSE)
      }
      temporal_grp_start <- as.integer(temporal_grp_start)
      temp_grp_start_lab <- paste0("    temporal_grp_start       = ", format(temporal_grp_start))
      if (temporal_grp_start == 1) {
        temp_grp_start_lab <- paste0(temp_grp_start_lab, " (default)")
      }
      
      # Validate optional argument `alter_temporal`
      tmp <- (unlist(alter_temporal))[1]
      if (!identical(alter_temporal, tmp) || is.null(tmp) || !is.finite(tmp) || tmp < 0) {
        stop("Argument 'alter_temporal' must be a nonnegative finite real number.\n\n", call. = FALSE)
      }
      alterTmp_lab <- paste0("    alter_temporal           = ", format(alter_temporal, big.mark = ","))
      if (alter_temporal == 0) {
        alterTmp_lab <- paste0(alterTmp_lab, " (default)")
        alter_temporal_label <- "(default)"
      } else {
        alter_temporal_label <- "(arg `alter_temporal`)"
      }
      
      # Validate the binding temporal total tolerances (and set the header label)
      tmp <- (unlist(tolV_temporal))[1]
      if (is.null(tmp)) {
        tolV_temporal <- NA_real_
      } else if (!identical(tolV_temporal, tmp) || !is.finite(tmp) && !is.na(tmp) || is.finite(tmp) && tmp < 0) {
        stop("Argument 'tolV_temporal' must be a nonnegative finite real number or NA.\n\n", call. = FALSE)
      }
      tmp <- (unlist(tolP_temporal))[1]
      if (is.null(tmp)) {
        tolP_temporal <- NA_real_
      } else if (!identical(tolP_temporal, tmp) || !is.finite(tmp) && !is.na(tmp) || is.finite(tmp) && tmp < 0) {
        stop("Argument 'tolP_temporal' must be a nonnegative finite real number or NA.\n\n", call. = FALSE)
      }
      if (!is.na(tolV_temporal)) {
        if (!is.na(tolP_temporal)) {
          stop("Arguments 'tolV_temporal' and 'tolP_temporal' cannot be both specified (one must be NA).\n\n", call. = FALSE)
        }
        tol_temporal_lab <- paste0("    tolV_temporal            = ", format(tolV_temporal))
        if (abs(tolV_temporal) < gs.tolerance) {
          tol_temporal_lab <- paste0(tol_temporal_lab, " (default)")
        }
        tolP_temporal <- 0
      } else {
        if (is.na(tolP_temporal)) {
          stop("Arguments 'tolV_temporal' and 'tolP_temporal' cannot be both NA (one must be specified).\n\n", call. = FALSE)
        }
        tol_temporal_lab <- paste0("    tolP_temporal            = ", format(tolP_temporal))
        tolV_temporal <- 0
      }
    }
    
    # Validate optional argument `osqp_settings_df` (and set the header label)
    settingsDF_lab <- "    osqp_settings_df         = "
    if (is.null(osqp_settings_df)) {
      settingsDF_lab <- paste0(settingsDF_lab, "NULL (default OSQP settings)")
      # Set OSQP setting `verbose` (based on arguments `quiet` and `display_level`), 
      # `require_polished = FALSE` and `prior_scaling = FALSE`
      settings_df <- data.frame(verbose = osqp_verbose, require_polished = FALSE, prior_scaling = FALSE)
    } else {
      settings_df_name <- deparse1(substitute(osqp_settings_df))
      if (nchar(settings_df_name) >= 60) {
        settings_df_name <- paste0(substr(settings_df_name, 1, 55), "<...>")
      }
      if (grepl("structure(", settings_df_name, fixed = TRUE)) {
        settings_df_name <- "<argument 'osqp_settings_df'>"
      }
      settings_df <- osqp_settings_df
      if (!is.data.frame(settings_df)) {
        warning("Argument 'settings_df_name' is not a 'data.frame' object. It will be ignored and the default OSQP ",
                "settings sequence data frame (package data frame `default_osqp_sequence`) will be  used instead.\n",
                call. = FALSE, immediate. = TRUE)
        settings_df <- default_osqp_sequence
        settings_df_name <- "default_osqp_sequence"
      }
      # Set (add/overwrite) OSQP setting `verbose` (based on arguments `quiet` and `display_level`)
      settings_df$verbose <- osqp_verbose
      if (settings_df_name %in% c("default_osqp_sequence", "gseries::default_osqp_sequence")) {
        settingsDF_lab <- paste0(settingsDF_lab, settings_df_name, " (default)")
      } else {
        settingsDF_lab <- paste0(settingsDF_lab, settings_df_name)
        # Remove duplicate rows and row names (just in case)
        settings_df <- unique(settings_df)
        row.names(settings_df) <- NULL
      }
      # Replace old setting `polish` (before OSQP V1.0) with new setting `polishing` (since OSQP v1.0)
      settings_cols <- names(settings_df)
      if ("polish" %in% settings_cols) {
        if ("polishing" %in% settings_cols) {
          # Ignore the old `polish` setting (rely on the new `polishing` setting)
          settings_df$polish <- NULL
        } else {
          # Rename `polish` to `polishing`
          names(settings_df)[names(settings_df) == "polish"] <- "polishing"
        }
        settings_cols <- names(settings_df)
      }
      # Set `require_polished`
      if ("require_polished" %in% settings_cols) {
        # Impose polishing when `require_polished == TRUE`
        if ("polishing" %in% settings_cols) {
          settings_df$polishing <- (settings_df$polishing | settings_df$require_polished)
        } else {
          settings_df$polishing <- settings_df$require_polished
        }
      } else {
        settings_df$require_polished <- FALSE
      }
      # Set `prior_scaling` 
      if (!("prior_scaling" %in% settings_cols)) {
        settings_df$prior_scaling <- FALSE
      }
    }
    
    # Validate the alterability coefficient optional arguments
    tmp <- (unlist(alter_pos))[1]
    if (!identical(alter_pos, tmp) || is.null(tmp) || !is.finite(tmp) || tmp < 0) {
      stop("Argument 'alter_pos' must be a nonnegative finite real number.\n\n", call. = FALSE)
    }
    alterPos_lab <- paste0("    alter_pos                = ", format(alter_pos, big.mark = ","))
    if (alter_pos == 1) {
      alterPos_lab <- paste0(alterPos_lab, " (default)")
      alter_pos_label <- "(default)"
    } else {
      alter_pos_label <- "(arg `alter_pos`)"
    }
    tmp <- (unlist(alter_neg))[1]
    if (!identical(alter_neg, tmp) || is.null(tmp) || !is.finite(tmp) || tmp < 0) {
      stop("Argument 'alter_neg' must be a nonnegative finite real number.\n\n", call. = FALSE)
    }
    alterNeg_lab <- paste0("    alter_neg                = ", format(alter_neg, big.mark = ","))
    if (alter_neg == 1) {
      alterNeg_lab <- paste0(alterNeg_lab, " (default)")
      alter_neg_label <- "(default)"
    } else {
      alter_neg_label <- "(arg `alter_neg`)"
    }
    tmp <- (unlist(alter_mix))[1]
    if (!identical(alter_mix, tmp) || is.null(tmp) || !is.finite(tmp) || tmp < 0) {
      stop("Argument 'alter_mix' must be a nonnegative finite real number.\n\n", call. = FALSE)
    }
    alterMix_lab <- paste0("    alter_mix                = ", format(alter_mix, big.mark = ","))
    if (alter_mix == 1) {
      alterMix_lab <- paste0(alterMix_lab, " (default)")
      alter_mix_label <- "(default)"
    } else {
      alter_mix_label <- "(arg `alter_mix`)"
    }

    # Validate optional argument `trunc_to_zero_tol` (and set the header label)
    tmp <- (unlist(trunc_to_zero_tol))[1]
    if (!identical(trunc_to_zero_tol, tmp) || is.null(tmp) || !is.finite(tmp) || is.finite(tmp) && tmp < 0) {
      stop("Argument 'trunc_to_zero_tol' must be a nonnegative finite real number.\n\n",
           call. = FALSE)
    }
    if (abs(trunc_to_zero_tol - validation_tol) < gs.tolerance) {
      zero_trunc_lab <- "    (*)trunc_to_zero_tol     = validation_tol (default)"
    } else {
      zero_trunc_lab <- paste0("    (*)trunc_to_zero_tol     = ", format(trunc_to_zero_tol))
    }

    # Validate optional argument `full_sequence`
    full_sequence <- gs.validate_arg_logi(full_sequence)
  }

  # Display the function call (argument values)
  quiet_msg_func("    in_ts                    = ", in_ts_name)
  quiet_msg_func("    problem_specs_df         = ", specs_df_name)
  quiet_msg_func(temp_grp_per_lab)
  quiet_msg_func(temp_grp_start_lab)
  quiet_msg_func(settingsDF_lab)
  quiet_msg_func(displayLevel_lab)
  quiet_msg_func(alterPos_lab)
  quiet_msg_func(alterNeg_lab)
  quiet_msg_func(alterMix_lab)
  quiet_msg_func(alterTmp_lab)
  lab <- paste0("    lower_bound              = ", format(lower_bound, big.mark = ","))
  if (lower_bound == -Inf) {
    lab <- paste0(lab, " (default)")
    lower_bound_label <- "(default)"
  } else {
    lower_bound_label <- "(arg `lower_bound`)"
  }
  quiet_msg_func(lab)
  lab <- paste0("    upper_bound              = ", format(upper_bound, big.mark = ","))
  if (upper_bound == Inf) {
    lab <- paste0(lab, " (default)")
    upper_bound_label <- "(default)"
  } else {
    upper_bound_label <- "(arg `upper_bound`)"
  }
  quiet_msg_func(lab)
  lab <- paste0("    tolV                     = ", format(tolV))
  if (tolV < gs.tolerance) {
    lab <- paste0(lab, " (default)")
  }
  quiet_msg_func(lab)
  quiet_msg_func(tol_temporal_lab, "\n")
  lab <- paste0("    (*)validation_tol        = ", format(validation_tol))
  if (abs(validation_tol - 0.001) < gs.tolerance) {
    lab <- paste0(lab, " (default)")
  }
  quiet_msg_func(lab)
  quiet_msg_func(zero_trunc_lab)
  quiet_msg_func(valid_only_lab)
  quiet_msg_func(quiet_lab, "\n")
  quiet_msg_func("    (*) indicates new arguments in G-Series 3.0\n")
  
  
  # Build the core elements (building blocks) for the balancing problems (while validating the specified 
  # info), excluding the temporal totals info (will be added later, inside the processing groups loop):
  #   - labels_df: cleaned-up version of the label definition records from `problem_specs_df` 
  #                (rows where `type` is non-missing); extra columns:
  #                  - type.lc : lowercase version of `type`
  #                  - row.lc  : lowercase version of `row` 
  #                  - con.flag: `TRUE` if `type.lc` is "eq", "le" or "ge"
  #   - coefs_df : cleaned-up version of the information specification records from `problem_specs_df` 
  #                (rows where `type` is missing); extra columns:
  #                  - row.lc  : lowercase version of `row` 
  #                  - con.flag: `labels_df$con.flag` allocated through `row.lc`
  #   - values_ts: reduced version of 'in_ts' with only the relevant series (see vector `ser_names`)
  #   - lb       : lower bound info (`type.lc = "lowerbd"`) for the relevant series; list object with the 
  #                following elements:
  #                  - coefs_ts       : lower bound values for series and period
  #                  - nondated_coefs : vector of nondated lower bounds from `problem_specs_df` (`timeVal` is `NA`)
  #                  - nondated_id_vec: vector of `ser_names` id's associated to vector `nondated_coefs`
  #                  - dated_id_vec   : vector of `ser_names` id's associated to dated lower bounds from 
  #                                     `problem_specs_df` (`timeVal` is not `NA`)
  #   - ub       : same as `lb` but for upper bounds (`type.lc = "upperbd"`)
  #   - alter    : same as `lb` but for period value alterability coefficients (`type.lc = "alter"`)
  #   - altertmp : same as `lb` but for temporal total alterability coefficients (`type.lc = "altertmp"`)
  #   - ser_names: vector of the relevant series names (set of series involved in the balancing constraints)
  #   - pos_ser  : vector of series names that have only positive coefficients across all balancing constraints
  #   - neg_ser  : vector of series names that have only negative coefficients across all balancing constraints
  #   - mix_ser  : vector of series names that have both positive and negative coefficients across all balancing 
  #                constraints
  #   - A1,op1,b1: period level (single-period) balancing constraint elements (`A1 %*% x op1 b1` for 
  #                coherent/reconciled data)
  #   - A2,op2,b2: temporal group level (multi-period) balancing constraint elements (`A2 %*% x op2 b2` for 
  #                coherent/reconciled data)
  #
  # Notes:
  #
  #   - The returned balancing problem elements do not include the implicit temporal totals for multi-period 
  #     processing (i.e., elements `A2`, `op2` and `b2` only contain the balancing constraints info).
  #       
  #   - Multi-period balancing problem elements `A2`, `op2` and `b2` are constructed in "column-major order",
  #     corresponding to the default behaviour of R for converting matrices into vectors. E.g.:
  #       - single-period processing (period `t`):
  #           `A1 %*% as.vector(values_ts[t, ]) op1 b1` for coherent/reconciled data
  #       - multi-period processing (periods `t1:t2` of length `temporal_grp_periodicity`):
  #           `A2 %*% as.vector(values_ts[t1:t2, ]) op2 b2` for coherent/reconciled data
  # 
  #   - Default temporal total alterability coefficients (`altertmp$coefs_ts`) for cases not specified in the 
  #     problem specs data frame (`problem_specs_df`) remain `NA` at this stage (i.e., argument `alter_temporal` 
  #     has not been used yet at this point). When time comes, this will allow for easy identification of the
  #     first specified (non `NA`) temporal total alter coef in the specs inside a complete temporal group.
  # 
  #   - A "wide range" of reserved keywords for column `type` in the specs (for label definition records) 
  #     are accepted in practice (i.e., the list of accepted keywords, in lowercase, include):
  #   - A "wide range" of reserved keywords for column `type` in the specs (for label definition records) 
  #     are accepted in practice. The list of accepted keywords, in lowercase, include:
  #       - "==" and "=" for "eq"
  #       - "<=" and "<" for "le"
  #       - ">=" and ">" for "ge"
  #       - "lower" + ("" or "_" or "-" or "." or " ") + ("bd" or "bnd" or "bound) for "lowerbd"
  #       - "upper" + ("" or "_" or "-" or "." or " ") + ("bd" or "bnd" or "bound) for "upperbd"
  #       - "alter" + ("" or "_" or "-" or "." or " ") + ("tmp" or "temp" or "temporal) for "altertmp"
  #
  pb <- build_balancing_problem(in_ts                    = in_ts,
                                problem_specs_df         = problem_specs_df,
                                in_ts_name               = in_ts_name,
                                ts_freq                  = ts_freq,
                                periods                  = periods,
                                n_per                    = n_per,
                                specs_df_name            = specs_df_name,
                                temporal_grp_periodicity = temporal_grp_periodicity,
                                alter_pos                = alter_pos,
                                alter_neg                = alter_neg,
                                alter_mix                = alter_mix,
                                lower_bound              = lower_bound, 
                                upper_bound              = upper_bound,
                                validation_only          = validation_only)
  
  # Number of series and balancing constraints
  n_ser <- length(pb$ser_names)
  n_con <- length(pb$labels_df$row.lc[pb$labels_df$con.flag])
  
  # Display the problem specs info
  specs_display_func()

    
  # Initial temporal aggregation matrix (for all series)
  #   => will be reduced later to the set of non fully binding series only
  #      (temporal totals of fully binding represent redundant constraints)
  dim_A2 <- dim(pb$A2)  # `dim_A2[1]` = number of rows
  # `dim_A2[2]` = number of columns
  A_a <- matrix(0, nrow = n_ser, ncol = dim_A2[2] + n_ser)
  for (ii in 1:n_ser) {
    A_a[ii, ((ii - 1) * temporal_grp_periodicity + 1):(ii * temporal_grp_periodicity)] <- 1
    A_a[ii, dim_A2[2] + ii] <- -1
  }
  

  # Define the processing groups (set of balancing problems).
  # Columns of the returned data frame:
  #   - grp         : processing group id (1:<number-of-groups>)
  #   - beg_per     : first period id
  #   - end_per     : last period id
  #   - complete_grp: complete group flag (logical, TRUE if end_per != beg_per)
  grp_df <- gs.build_proc_grps(gs.time2year(pb$values_ts),
                               gs.time2per(pb$values_ts),
                               n_per,
                               ts_freq,
                               temporal_grp_periodicity,
                               temporal_grp_start)

  # Activate message display
  n_grps <- nrow(grp_df)
  if (n_grps > 1) {
    solve_header_msg_func <- gs.NULL_func
    grp_msg_func <- message
    final_msg_flag <- TRUE
  } else {
    solve_header_msg_func <- solve_msg_func
    grp_msg_func <- gs.NULL_func
    final_msg_flag <- FALSE
  }

  solve_header_msg_func("\n\n", solve_str)
  solve_header_msg_func(strrep("=", nchar(solve_str)), "\n")


  # Initialize the output objects
  out_mat <- matrix(NA_real_, nrow = n_per, ncol = n_ser,
                    dimnames = list(NULL, pb$ser_names))
  tmp1 <- rep.int(NA_character_, n_grps)
  tmp2 <- rep.int(NA_real_, n_grps)
  tmp3 <- rep.int(NA_integer_, n_grps)
  out_proc_grp_df <- data.frame(
    proc_grp = grp_df$grp,
    proc_grp_type = tmp1,
    proc_grp_label = tmp1,
    sol_status = tmp1,
    sol_status_val = tmp3,
    n_unmet_con = tmp2,
    max_discr = tmp2,
    validation_tol = rep.int(validation_tol, n_grps),
    sol_type = tmp1,
    osqp_attempts = rep.int(0L, n_grps),
    osqp_seqno = tmp3,
    osqp_status = tmp1,
    osqp_polished = rep.int(as.logical(NA), n_grps),
    total_solve_time = tmp3)
  out_prob_val_df <- NULL
  out_prob_con_df <- NULL
  out_osqp_settings_df <- NULL
  out_osqp_sol_info_df <- NULL


  # Balance each processing group (generate the `balancing_process()` calls)
  for (grp in 1:n_grps) {

    # Build the processing group header and balancing problem elements list
    per_id_vec <- grp_df$beg_per[grp]:grp_df$end_per[grp]
    n_per_grp <- length(per_id_vec)
    n_values_grp <- n_ser * n_per_grp

    # Initialize the (non-constraint) balancing problem elements
    bl <- list(
      x = as.vector(pb$values_ts[per_id_vec, , drop = FALSE]),
      c = as.vector(pb$alter$coefs_ts[per_id_vec, , drop = FALSE]),
      lb = as.vector(pb$lb$coefs_ts[per_id_vec, , drop = FALSE]),
      ub = as.vector(pb$ub$coefs_ts[per_id_vec, , drop = FALSE]),
      bd_tol = rep.int(0, n_values_grp)
    )


    # Temporal group processing (multiple periods)
    #   => balancing problem elements must be "augmented" with temporal total info
    #   => initial constraint elements are the "v2" objects
    if (grp_df$complete_grp[grp]) {

      out_proc_grp_df$proc_grp_type[grp] <- "temporal group"
      out_proc_grp_df$proc_grp_label[grp] <- paste0(periods[grp_df$beg_per[grp]], " - ", periods[grp_df$end_per[grp]])

      # Build the processing group header
      msg_str <- paste0(solve_word, " periods [", out_proc_grp_df$proc_grp_label[grp], "]")

      # Reduce matrix `A_a` for non fully binding series only
      # (temporal totals of fully binding series are irrelevant)
      non_fully_binding <- (A_a[, 1:dim_A2[2], drop = FALSE] %*% bl$c) != 0
      n_a <- sum(non_fully_binding)
      A_a2 <- A_a[non_fully_binding, c(rep.int(TRUE, dim_A2[2]), non_fully_binding), drop = FALSE]

      # Add the temporal totals info
      a <- A_a2[, 1:dim_A2[2], drop = FALSE] %*% bl$x
      bl$x <- c(bl$x, a)
      c_a <- apply(pb$altertmp$coefs_ts[per_id_vec, pb$ser_names[non_fully_binding], drop = FALSE], 2,
                   function(x) {
                     # Return the first non missing (non `NA`) temporal total alter. coef. of the temporal group.
                     # Otherwise (all `NA`), return the default alter. coef. (argument `alter_temporal`)
                     y <- x[!is.na(x)][1]
                     if (is.na(y)) {
                       alter_temporal
                     } else {
                       y
                     }
                   })
      bl$c <- c(bl$c, c_a)
      names(bl$c) <- NULL
      bl$lb <- c(bl$lb, rep.int(-Inf, n_a))
      bl$ub <- c(bl$ub, rep.int(Inf, n_a))
      bl$bd_tol <- c(bl$bd_tol, rep.int(0, n_a))
      bl$A <- rbind(cbind(pb$A2, matrix(0, nrow = dim_A2[1], ncol = n_a)),
                    A_a2)
      bl$op <- c(pb$op2, rep.int("==", n_a))
      bl$b <- c(pb$b2, rep.int(0, n_a))

      # Temporal aggregation constraint tolerances are for binding temporal totals only!
      # (no need for "additional slack" in nonbinding temporal total aggregation constraints:
      # "slack" for nonbinding temporal totals is already accounted for in the optimization
      # function where changes to the initial temporal totals are minimized)
      bl$b_tol <- c(rep.int(tolV, dim_A2[1]), (a * tolP_temporal + tolV_temporal) * (c_a == 0))


    # Single-period processing
    #   => constraint elements are the "v1" objects
    } else {

      out_proc_grp_df$proc_grp_type[grp] <- "period"
      out_proc_grp_df$proc_grp_label[grp] <- periods[grp_df$beg_per[grp]]

      # Build the processing group header
      msg_str <- paste0(solve_word, " period [", out_proc_grp_df$proc_grp_label[grp], "]")

      # Define the constraint elements
      bl$A <- pb$A1
      bl$op <- pb$op1
      bl$b <- pb$b1
      bl$b_tol <- rep.int(tolV, n_con)

      non_fully_binding <- integer(0L)
      n_a <- 0
    }


    # Display the processing group header
    grp_msg_func("\n\n", msg_str)
    grp_msg_func(strrep("=", nchar(msg_str)), "\n")


    # Convert the balancing problem into a (reduced) QP problem (for OSQP):
    #   - construct the QP problem's Hessian matrix (`P`) and vector of linear coefficients (`q`)
    #   - rewrite the balancing constraints as `l <= Ax <= u` linear constraints. transferring binding values 
    #     (`c * x = 0`) into the constraints bounds (`l` and `u`)
    #   - add `l <= Ax <= u` linear constraints for and upper/lower bounds
    qp <- build_reduced_qp(bl$x, bl$c, bl$lb, bl$ub, bl$bd_tol,  # problem variables info
                           bl$A, bl$op, bl$b, bl$b_tol)          # problem constraints info
    n_con_qp <- length(qp$l)

    
    # Calculate the initial discrepancies
    Ax_ini <- qp$A %*% qp$x
    discr_ini <- calc_discr(Ax_ini, qp$l, qp$u)
    

    # Solve the balancing problem
    sol <- solve_func(x = qp$x,                                          # initial problem values (initial solution)
                      Ax = Ax_ini,                                       # evaluated constraint body values with the initial solution
                      discr = discr_ini,                                 # initial constraint discrepancies list
                      validation_tol = validation_tol,                   # solution validation tolerance (max. allowed constraint discr.)
                      P = qp$P, q = qp$q, A = qp$A, l = qp$l, u = qp$u,  # OSQP problem elements
                      settings_df = settings_df,                         # OSQP settings sequence data frame
                      full_sequence = full_sequence,                     # (logical) perform all steps (rows) of `settings_df`
                      solve_msg_func = solve_msg_func,                   # (function) message display function
                      osqpBeg_display_func = osqpBeg_display_func,       # (function) OSQP model and iterations display set up function
                      osqpEnd_display_func = osqpEnd_display_func,       # (function) OSQP model and iterations display wrap up function
                      osqp_output_file = osqp_output_file,               # temporary file (name and path) for the OSQP model and iterations contents
                      trunc_to_zero_tol = trunc_to_zero_tol)             # tolerance for changing nonzero solution data points to 0

    # Set the balancing solution:
    #  - initial solution (`bl$x`) for binding values
    #  - returned solution (`sol$x`) for nonbinding values
    x_out <- bl$x
    x_out[qp$id_x] <- sol$x

    # Cumulate the solutions
    out_mat[per_id_vec, ] <- matrix(x_out[1:n_values_grp], ncol = n_ser)


    # Create the problem values data frame
    prob_val_df <- data.frame(
      proc_grp = rep.int(grp, length(bl$x)),
      val_type = rep.int(c("period value", "temporal total"), c(n_values_grp, n_a)),
      name = c(rep(pb$ser_names, each = n_per_grp), pb$ser_names[non_fully_binding]),
      t = c(rep.int(per_id_vec, n_ser), rep.int(per_id_vec[1], n_a)),
      time_val = c(rep.int(time_values[per_id_vec], n_ser), rep.int(time_values[per_id_vec[1]], n_a)),
      lower_bd = bl$lb,
      upper_bd = bl$ub,
      alter = bl$c,
      value_in = bl$x,
      value_out = x_out,
      dif = x_out - bl$x,
      rdif = gs.calc_relDiff(x_out, bl$x)
    )


    # Create the problem constraints info data frame
    prob_con_df <- data.frame(
      proc_grp = rep.int(grp, n_con_qp),
      con_type = rep.int(c("balancing constraint",
                           "temporal aggregation constraint",
                           "period value bounds"),
                         c(n_con * n_per_grp,
                           n_a,
                           length(qp$id_bd_con))),
      name = c(rep(pb$labels_df$row[pb$labels_df$con.flag], each = n_per_grp),
               pb$ser_names[non_fully_binding],
               prob_val_df$name[qp$id_bd_con]),
      t = c(rep.int(per_id_vec, n_con),
            rep.int(per_id_vec[1], n_a),
            prob_val_df$t[qp$id_bd_con]),
      time_val = c(rep.int(time_values[per_id_vec], n_con),
                   rep.int(time_values[per_id_vec[1]], n_a),
                   prob_val_df$time_val[qp$id_bd_con]),
      l = qp$l,
      u = qp$u,
      Ax_in = as.vector(Ax_ini),
      Ax_out = as.vector(sol$Ax),
      discr_in = discr_ini$discr_vec,
      discr_out = sol$discr$discr_vec,
      validation_tol = rep.int(validation_tol, n_con_qp),
      unmet_flag = (sol$discr$discr_vec > validation_tol)
    )


    # Fill in the processing groups data frame and both OSQP info data frames
    out_proc_grp_df$sol_status[grp] <- sol$status
    out_proc_grp_df$sol_status_val[grp] <- sol$status_val
    n_unmet <- sum(prob_con_df$unmet_flag)
    out_proc_grp_df$n_unmet_con[grp] <- n_unmet
    out_proc_grp_df$max_discr[grp] <- sol$discr$max_discr
    out_proc_grp_df$sol_type[grp] <- sol$type
    out_proc_grp_df$total_solve_time[grp] <- sol$seq_time
    if (!is.null(sol$osqp_info)) {
      out_proc_grp_df$osqp_attempts[grp] <- sol$osqp_attempts
      out_proc_grp_df$osqp_seqno[grp] <- sol$osqp_seqno
      out_proc_grp_df$osqp_status[grp] <- sol$osqp_info$status
      out_proc_grp_df$osqp_polished[grp] <- sol$osqp_info$status_polish == 1
      out_osqp_settings_df <- rbind(out_osqp_settings_df, as.data.frame(c(list(proc_grp = grp), sol$osqp_settings)))
      out_osqp_sol_info_df <- rbind(out_osqp_sol_info_df,
                                    as.data.frame(c(
                                      list(proc_grp = grp),
                                      sol$osqp_info,
                                      # Calculate the original balancing problem's objective value
                                      list(obj_val_ori_prob = 
                                             sol$osqp_info$obj_val * sol$osqp_info$prior_scaling_factor  # original scale
                                             + sum(-0.5 * qp$q * qp$x)))))                               # constant terms
    }


    # Sort, display and cumulate the problem values and constraints info data frames
    prob_val_df <- prob_val_df[order(rep.int(1:2, c(n_values_grp, n_a)),
                                     prob_val_df$t,
                                     c(rep(1:n_ser, each = n_per_grp), seq_len(n_a)))
                               , ]
    out_prob_val_df <- rbind(out_prob_val_df, prob_val_df)
    prob_con_df <- prob_con_df[order(
      # 1- constraint type (periods, then temporal totals)
      rep.int(c(1, 2, 1), c(n_con * n_per_grp,
                                  n_a,
                                  length(qp$id_bd_con))),
      # 2- period
      prob_con_df$t,
      # 3- constraint sub-type
      rep.int(c(1, 3, 2), c(n_con * n_per_grp,
                            n_a,
                            length(qp$id_bd_con))),
      # 4- constraint/series name (keeping this initial order)
      c(rep(1:n_con, each = n_per_grp),
        seq_len(n_a),
        seq_along(qp$id_bd_con)))
      , ]
    out_prob_con_df <- rbind(out_prob_con_df, prob_con_df)
    results_display_func()
    
    # Validate the solution
    if (n_unmet > 0) {
      warning("Constraints were not met in ", n_unmet, " occasion(s). Maximum discrepancy: ",
              format(sol$discr$max_discr, big.mark = ","), ". See constraints with `unmet_flag = TRUE` ",
              "for details.\n", call. = FALSE, immediate. = TRUE)
      warning_flag <- TRUE
    }
  }


  # Create the output list
  out_ts <- in_ts
  out_ts[, pb$ser_names] <- out_mat
  out_list$out_ts <- out_ts
  row.names(out_proc_grp_df) <- NULL
  out_list$proc_grp_df <- out_proc_grp_df
  out_list$periods_df <- data.frame(
    # `sapply()` is safe: it always returns a numeric vector (`grp_df$grp` is of length minimum 1)
    proc_grp = rep.int(grp_df$grp, sapply(grp_df$grp, function(ii) length(grp_df$beg_per[ii]:grp_df$end_per[ii]))),
    t = 1:n_per,
    time_val = time_values)
  row.names(out_prob_val_df) <- NULL
  out_list$prob_val_df <- out_prob_val_df
  row.names(out_prob_con_df) <- NULL
  out_list$prob_con_df <- out_prob_con_df
  row.names(out_osqp_settings_df) <- NULL
  out_list$osqp_settings_df <- out_osqp_settings_df
  row.names(out_osqp_sol_info_df) <- NULL
  out_list$osqp_sol_info_df <- out_osqp_sol_info_df


  if (final_msg_flag && warning_flag) {
    warning("Warnings were generated during processing. See relevant message(s) for details.\n",
            call. = FALSE, immediate. = TRUE)
  }

  # Output object returned via function `on.exit()`
}


#' Build the elements of balancing problems.
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/build_balancing_problem.html}})}
#' 
#' This function is used internally by [tsbalancing()] to build the elements of the balancing problems. 
#' It can also be useful to derive the indirect series associated to equality balancing constraints manually 
#' (outside of the [tsbalancing()] context). 
#' 
#' 
#' @inheritParams tsbalancing
#' 
#' @param in_ts_name (optional) 
#' 
#' String containing the value of argument `in_ts`.
#'
#' **Default value** is `in_ts_name = deparse1(substitute(in_ts))`.
#' 
#' @param ts_freq (optional)
#' 
#' Frequency of the time series object (argument `in_ts`).
#' 
#' **Default value** is `ts_freq = stats::frequency(in_ts)`.
#' 
#' @param periods (optional) 
#' 
#' Character vector describing the time series object (argument `in_ts`) periods.
#' 
#' **Default value** is `periods = gs.time2str(in_ts)`.
#' 
#' @param n_per (optional) 
#' 
#' Number of periods of the time series object (argument `in_ts`).
#' 
#' **Default value** is `n_per = nrow(as.matrix(in_ts))`.
#' 
#' @param specs_df_name (optional)
#' 
#' String containing the value of argument `problem_specs_df`.
#' 
#' **Default value** is `specs_df_name = deparse1(substitute(problem_specs_df))`.
#' 
#'
#' @returns
#' A list with the elements of the balancing problems (excluding the temporal totals info):
#' - `labels_df`: cleaned-up version of the _label definition records_ from `problem_specs_df` 
#'                (`type` is not missing (is not `NA`)); extra columns:
#'   - `type.lc` : `tolower(type)`
#'   - `row.lc`  : `tolower(row)` 
#'   - `con.flag`: `type.lc %in% c("eq", "le", "ge")`
#' - `coefs_df` : cleaned-up version of the information specification records from `problem_specs_df` 
#'                (`type` is missing (is `NA`); extra columns:
#'   - `row.lc`  : `tolower(row)` 
#'   - `con.flag`: `labels_df$con.flag` allocated through `row.lc`
#' - `values_ts`: reduced version of `in_ts` with only the relevant series (see vector `ser_names`)
#' - `lb`       : lower bound info (`type.lc  = "lowerbd"`) for the relevant series; list object with the 
#'                following elements:
#'   - `coefs_ts`       : lower bound values for series and period
#'   - `nondated_coefs` : vector of nondated lower bounds from `problem_specs_df` (`timeVal` is `NA`)
#'   - `nondated_id_vec`: vector of `ser_names` id's associated to vector `nondated_coefs`
#'   - `dated_id_vec`   : vector of `ser_names` id's associated to dated lower bounds from 
#'                        `problem_specs_df` (`timeVal` is not `NA`)
#' - `ub`       : `lb` equivalent for upper bounds (`type.lc = "upperbd"`)
#' - `alter`    : `lb` equivalent for period value alterability coefficients (`type.lc = "alter"`)
#' - `altertmp` : `lb` equivalent for temporal total alterability coefficients (`type.lc = "altertmp"`)
#' - `ser_names`: vector of the relevant series names (set of series involved in the balancing constraints)
#' - `pos_ser`  : vector of series names that have only positive nonzero coefficients across all balancing constraints
#' - `neg_ser`  : vector of series names that have only negative nonzero coefficients across all balancing constraints
#' - `mix_ser`  : vector of series names that have both positive and negative nonzero coefficients across all balancing 
#'                constraints
#' - `A1`,`op1`,`b1`: balancing constraint elements for problems involving a single period (e.g., each period of an 
#'                    incomplete temporal group)
#' - `A2`,`op2`,`b2`: balancing constraint elements for problems involving `temporal_grp_periodicity` periods (e.g., the set 
#'                    of periods of a complete temporal group)
#' 
#' 
#' @details
#' See [tsbalancing()] for a detailed description of _time series balancing_ problems.
#' 
#' Any missing (`NA`) value found in the input time series object (argument `in_ts`) would be replaced with 0 in `values_ts` 
#' and trigger a warning message.
#' 
#' The returned elements of the balancing problems do not include the implicit temporal totals (i.e., elements `A2`, `op2` 
#' and `b2` only contain the balancing constraints).
#'       
#' Multi-period balancing problem elements `A2`, `op2` and `b2` (when `temporal_grp_periodicity > 1`) are constructed 
#' _column by column_ (in "column-major order"), corresponding to the default behaviour of R for converting objects of class 
#' "matrix" into vectors. I.e., the balancing constraints conceptually correspond to:
#' - `A1 %*% values_ts[t, ] op1 b1` for problems involving a single period (`t`)
#' - `A2 %*% as.vector(values_ts[t1:t2, ]) op2 b2` for problems involving `temporal_grp_periodicity` periods (`t1:t2`).
#' 
#' Notes:
#' - Argument `alter_temporal` has not been applied yet at this point and `altertmp$coefs_ts` only contains the 
#' coefficients specified in the problem specs data frame (argument `problem_specs_df`). I.e., `altertmp$coefs_ts` contains 
#' missing (`NA`) values except for the temporal total alterability coefficients included in (specified with) `problem_specs_df`. 
#' This is done in order to simplify the identification of the first non missing (non `NA`) temporal total alterability 
#' coefficient of each complete temporal group (to occur later, when applicable, inside [tsbalancing()]).
#' - Argument validation is not performed here; it is (bluntly) assumed that the function is called by [tsbalancing()] 
#' where a thorough validation of the arguments is done.

#'
#'
#' @seealso [tsbalancing()] [build_raking_problem()]
#' 
#' @example misc/function_examples/build_balancing_problem-ex.R
#' 
#' @export
build_balancing_problem <- function(in_ts,
                                    problem_specs_df,
                                    in_ts_name = deparse1(substitute(in_ts)),
                                    ts_freq = stats::frequency(in_ts),
                                    periods = gs.time2str(in_ts),
                                    n_per = nrow(as.matrix(in_ts)),
                                    specs_df_name = deparse1(substitute(problem_specs_df)),
                                    temporal_grp_periodicity = 1,
                                    alter_pos = 1,
                                    alter_neg = 1,
                                    alter_mix = 1,
                                    lower_bound = -Inf, 
                                    upper_bound = Inf,
                                    validation_only = FALSE) {
  
  
  # Assign non-constraint coefficients from the problem specs data frame.
  # Note: coefs with time values (`timeval`) outside the time series span are ignored
  #
  # Arguments:
  #   - coefs_ts      : 'ts' object to be updated
  #   - label         : string (`coef_df$type.lc` value) identifying the type of coefs to assign
  #
  # Other main (parent) function objects used in this function:
  #   - coefs_df : problem specs coefficients data frame (rows for which `type` is `NA`)
  #   - ser_names: vector of the series names involved in the balancing problem
  #   - n_per    : `nrow(coefs_ts)` = number of periods in the input 'ts' object
  #   - ts_freq  : `frequency(coefs_ts)` = frequency of the input 'ts' object
  #   - periods  : `gs.time2str(coefs_ts)` = vector of period string labels ("<year>-<per>")
  # 
  # Value:
  #   Returns a list of 4 objects:
  #     - coefs_ts       : updated 'ts' object
  #     - nondated_coefs : vector of nondated coefficients (`timeval` is `NA`)
  #     - nondated_id_vec: vector of `ser_names` id's associated to vector `nondated_coefs`
  #     - dated_id_vec   : vector of `ser_names` id's associated to dated coefficient (`timeval` is not `NA`)
  assign_coefs <- function(coefs_ts, label) {
    
    # All periods (`timeval` is `NA`)
    logi_vec <- coefs_df$type.lc == label & is.na(coefs_df$timeval)
    coefs_ts[, coefs_df$col[logi_vec]] <- rep(coefs_df$coef[logi_vec], each = n_per)
    
    # Specific periods (`timeval` is not `NA`)
    id_vec <- which(coefs_df$type.lc == label & !is.na(coefs_df$timeval))
    for (ii in seq_along(id_vec)) {
      coefs_ts[match(gs.time2str(stats::ts(NA, start = coefs_df$timeval[id_vec[ii]], frequency = ts_freq)),
                     periods, nomatch = 0),
               coefs_df$col[id_vec[ii]]] <- coefs_df$coef[id_vec[ii]]
    }
    
    list(coefs_ts = coefs_ts,
         nondated_coefs = coefs_df$coef[logi_vec],
         nondated_id_vec = match(coefs_df$col[logi_vec], ser_names),
         dated_id_vec = match(unique(coefs_df$col[id_vec]), ser_names))
  }
  
  
  ### Main function ###

  
  # Initial clean-up/standardization of the problem specs df:
  #   - convert column names to lowercase
  #   - replace empty strings ("") with `NA` for character columns (`type`, `col`, `row`)
  #   - standardize constraint RHS specification (`col = "_rhs_"`)
  #   - add column `timeval` if not present 
  #     => if column `time_val` is present (without column `timeval`), it is recoded to `timeval`
  in_ts_cols <- colnames(in_ts)
  specsDF_cols <- tolower(names(problem_specs_df))
  names(problem_specs_df) <- specsDF_cols
  core_cols <- c("type", "col", "row", "coef")
  gs.validate_cols(tolower(setdiff(core_cols, specsDF_cols)), NULL, specs_df_name)
  # `sapply()` is safe: it always returns a logical vector of length 3
  logi_vec <- sapply(core_cols[1:3], function(col) !is.character(problem_specs_df[[col]]))
  if (any(logi_vec)) {
    stop("The following problem specs data frame column(s) must be character:",
         paste0("\n  ", core_cols[c(logi_vec, FALSE)], collapse = ""), call. = FALSE)
  }
  if (!is.numeric(problem_specs_df$coef)) {
    stop("Problem specs data frame column `coef` must be numeric.", call. = FALSE)
  }
  problem_specs_df$type <- trimws(problem_specs_df$type)
  problem_specs_df$type[problem_specs_df$type == ""] <- NA
  problem_specs_df$col <- trimws(problem_specs_df$col)
  problem_specs_df$col[problem_specs_df$col == ""] <- NA
  problem_specs_df$col[tolower(problem_specs_df$col) == "_rhs_"] <- "_rhs_"
  problem_specs_df$row <- trimws(problem_specs_df$row)
  problem_specs_df$row[problem_specs_df$row == ""] <- NA
  logi_vec <- c("timeval", "time_val") %in% specsDF_cols
  if (!any(logi_vec)) {
    problem_specs_df$timeval <- NA_real_
  } else {
    if (all(logi_vec == c(FALSE, TRUE))) {
      problem_specs_df$timeval <- problem_specs_df$time_val
    }
    logi_vec <- !is.na(problem_specs_df$timeval)
    if (any(logi_vec)) {
      if (!is.numeric(problem_specs_df$timeval[logi_vec])) {
        stop("Problem specs data frame column `timeVal` must be numeric.", call. = FALSE)
      }
      if (any(!is.finite(problem_specs_df$timeval[logi_vec]))) {
        stop("Invalid time values (column `timeVal`) found in the problem specs data frame (time values must be ",
             "finite real numbers or `NA`).", call. = FALSE)
      }
    } else {
      problem_specs_df$timeval <- NA_real_
    }
  }
  
  
  # Split specs data frame into labels and coefs data frames
  labels_df <- problem_specs_df[!is.na(problem_specs_df$type), c("type", "row"), drop = FALSE]
  coefs_df <- problem_specs_df[is.na(problem_specs_df$type), c("col", "row", "coef", "timeval"), drop = FALSE]
  
  # Initial (basic) problems specs validation
  if (any(is.na(labels_df$row))) {
    stop("Invalid label definition record found in the problem specs data frame (missing `row` value).", call. = FALSE)
  }
  if (any(is.na(coefs_df$col) | is.na(coefs_df$row))) {
    stop("Invalid information specification record found in the problem specs data frame ",
         "(missing `col` or `row` value).", call. = FALSE)
  }
  # Reject (ignore) missing (`NA`) coefs
  coefs_df <- coefs_df[!is.na(coefs_df$coef), , drop = FALSE]
  
  # Cleanup the labels (lowercase version of columns `type` and `row`)
  # => Allow the following lowercase `type` values (convert them to the expected lowercase value):
  #    - "==" and "=" for "eq"
  #    - "<=" and "<" for "le"
  #    - ">=" and ">" for "ge"
  #    - "lower" + ("" or "_" or "-" or "." or " ") + ("bd" or "bnd" or "bound) for "lowerbd"
  #    - "upper" + ("" or "_" or "-" or "." or " ") + ("bd" or "bnd" or "bound) for "upperbd"
  #    - "alter" + ("" or "_" or "-" or "." or " ") + ("tmp" or "temp" or "temporal) for "altertmp"
  labels_df$type.lc <- sub(sub(sub(sub(sub(tolower(labels_df$type), 
                                           pattern = "^(==|=)$", 
                                           replacement = "eq", 
                                           perl = TRUE),
                                       pattern = "^(<=|<)$", 
                                       replacement = "le", 
                                       perl = TRUE),
                                   pattern = "^(>=|>)$", 
                                   replacement = "ge", 
                                   perl = TRUE),
                               pattern = "^(lower|upper)[\\.[:blank:]_-]?(bd|bnd|bound)$", 
                               replacement = "\\1bd", 
                               perl = TRUE),
                           pattern = "^alter[\\.[:blank:]_-]?(tmp|temp|temporal)$", 
                           replacement = "altertmp", 
                           perl = TRUE)
  labels_df$row.lc <- tolower(labels_df$row)
  coefs_df$row.lc <- tolower(coefs_df$row)
  labels_df$con.flag <- labels_df$type.lc %in% c("eq", "le", "ge")
  coefs_df$con.flag <- coefs_df$row.lc %in% unique(labels_df$row.lc[labels_df$con.flag])
  
  # Get the list of series involved in balancing constraints
  ser_names <- unique(coefs_df$col[coefs_df$con.flag & coefs_df$col != "_rhs_" & coefs_df$coef != 0])
  n_ser <- length(ser_names)
  if (n_ser == 0) {
    stop("The problem specs data frame must include at least one valid (non-empty) balancing constraint.", call. = FALSE)
  }
  missing_cols <- setdiff(ser_names, in_ts_cols)
  if (length(missing_cols) > 0) {
    stop("The following series, listed in balancing constraints in the problem specs data frame, are missing from ",
         "input 'ts' object \"", in_ts_name, "\": ", paste0("\n  ", missing_cols, collapse = ""), call. = FALSE)
  }
  # Order `ser_names` according to the input time series columns
  ser_names <- intersect(in_ts_cols, ser_names)
  
  # Keep only the relevant info from the problem specification data frame: info related to the
  # "balancing problem series", i.e., the set of series involved in the balancing constraint.
  coefs_df <- coefs_df[
    # coefs of relevant series (involved in balancing constraints)
    coefs_df$col %in% ser_names & (!coefs_df$con.flag | coefs_df$con.flag & coefs_df$coef != 0)
    # RHS values for relevant constraints (involving at least one series)
    | coefs_df$col == "_rhs_" & coefs_df$row.lc %in% unique(coefs_df$row.lc[coefs_df$con.flag & coefs_df$col %in% ser_names])
    , , drop = FALSE]
  row.names(coefs_df) <- NULL
  labels_df <- labels_df[labels_df$row.lc %in% unique(coefs_df$row.lc), , drop = FALSE]
  row.names(labels_df) <- NULL
  
  # Bring the type labels in the coefs data frame
  coefs_df <- merge(labels_df[c("type.lc", "row.lc")], coefs_df, by = "row.lc", all = TRUE, sort = FALSE)
  
  
  # Extra problems specs validation (more in-depth validation for the remaining relevant info)
  
  # Invalid problem elements
  invalid_types <- setdiff(unique(labels_df$type.lc), c("eq", "le", "ge", "lowerbd", "upperbd", "alter", "altertmp"))
  if (length(invalid_types) > 0) {
    stop("The following invalid problem elements ('type' column values in lowercase) were found in the problem specs data frame: ",
         paste0("\n  ", invalid_types, collapse = ""), call. = FALSE)
  }
  
  # Invalid labels
  invalid_labels <- unique(coefs_df$row.lc[is.na(coefs_df$type.lc)])
  if (length(invalid_labels) > 0) {
    stop("The following labels ('row' column values in lowercase) in the problem specs data frame have not been properly ",
         "defined (no coresponding \"label definition record\"): ", paste0("\n  ", invalid_labels, collapse = ""),
         call. = FALSE)
  }
  
  # Duplicate labels
  if (any(table(labels_df$row.lc) > 1)) {
    stop("The problem specs data frame contains duplicate labels, i.e., same 'row' value used for several ",
         "problem elements (`type` values).", call. = FALSE)
  }
  
  # Duplicate label definition records for non-constraint elements
  if (any(table(labels_df$type.lc[!labels_df$con.flag]) > 1)) {
    stop("The problem specs data frame contains duplicate label definition records for non-constraint elements ",
         "(i.e., a given non-constraint `type` value is assigned a `row` value more than once).", call. = FALSE)
  }
  
  # Specified time values (`timeval`) for constraints
  if (any(coefs_df$con.flag & !is.na(coefs_df$timeval))) {
    stop("Time values (column `timeVal` in the problem specs data frame) cannot be specified for constraint ",
         "coefficients or RHS values.", call. = FALSE)
  }
  
  # Invalid numeric `coef` values:
  #   - constraint coefficients  : infinite
  #   - alterability coefficients: negative or infinite
  #   - lower bound              : Inf
  #   - upper bound              : -Inf
  if (any(!is.finite(coefs_df$coef[coefs_df$con.flag]))) {
    stop("Invalid constraint coefficients (column `coef`) found in the problem specs data frame (constraint ",
         "coefficients must be finite real numbers).", call. = FALSE)
  }
  logi_vec <- (coefs_df$type.lc %in% c("alter", "altertmp"))
  if (any(!is.finite(coefs_df$coef[logi_vec]) | coefs_df$coef[logi_vec] < 0)) {
    stop("Invalid alterability coefficients (column `coef`) found in the problem specs data frame (alterability ",
         "coefficients must be nonnegative finite real numbers).", call. = FALSE)
  }
  if (any(coefs_df$coef[coefs_df$type.lc == "lowerbd"] == Inf)) {
    stop("Invalid period value lower bounds (column `coef`) found in the problem specs data frame (lower bounds ",
         "must be a finite real number or -Inf).", call. = FALSE)
  }
  if (any(coefs_df$coef[coefs_df$type.lc == "upperbd"] == -Inf)) {
    stop("Invalid period value upper bounds (column `coef`) found in the problem specs data frame (upper bounds ",
         "must be a finite real number or Inf).", call. = FALSE)
  }
  
  # Duplicate coefficient specification
  if (any(table(coefs_df[c("col", "row.lc", "timeval")], useNA = "always") > 1)) {
    stop("The problem specs data frame contains duplicate information specification records, i.e., multiple 'coef' values ",
         "for a given set of `col` (time series or constraint RHS), `row` (problem element) and 'timeVal' values.",
         call. = FALSE)
  }
  
  # 'Reduce the input 'ts' 'in_ts' version with only the relevant series
  values_ts <- in_ts[, ser_names, drop = FALSE]
  logi_vec <- is.na(values_ts)
  if (sum(logi_vec) > 0) {
    warning("Missing data were found in the input 'ts' object. They were replaced with 0.",
            call. = FALSE, immediate. = TRUE)
    values_ts[logi_vec] <- 0
  }
  
  
  # Assign the lower/upper bounds
  lb <- list(coefs_ts = values_ts)
  lb$coefs_ts[, ] <- lower_bound
  lb <- assign_coefs(lb$coefs_ts, "lowerbd")
  ub <- list(coefs_ts = values_ts)
  ub$coefs_ts[, ] <- upper_bound
  ub <- assign_coefs(ub$coefs_ts, "upperbd")
  if (any(lb$coefs_ts > ub$coefs_ts)) {
    stop("Incompatible period value bounds (column `coef`) specified in the problem specs data frame (lower bounds ",
         "cannot be greater than upper bounds).", call. = FALSE)
  }
  
  # Assign the alterability coefficients
  alter <- list(coefs_ts = values_ts)
  alter$coefs_ts[, ] <- NA_real_
  altertmp <- alter
  if (!validation_only) {
    
    # Default (function arguments `alter_pos`, `alter_neg` and `alter_mix`)
    pos_ser <- unique(coefs_df$col[coefs_df$con.flag & coefs_df$col != "_rhs_" & coefs_df$coef > 0])
    neg_ser <- unique(coefs_df$col[coefs_df$con.flag & coefs_df$col != "_rhs_" & coefs_df$coef < 0])
    mix_ser <- intersect(pos_ser, neg_ser)
    pos_ser <- setdiff(pos_ser, mix_ser)
    neg_ser <- setdiff(neg_ser, mix_ser)
    alter$coefs_ts[, pos_ser] <- alter_pos
    alter$coefs_ts[, neg_ser] <- alter_neg
    alter$coefs_ts[, mix_ser] <- alter_mix
    
    # From the problem specs data frame
    alter <- assign_coefs(alter$coefs_ts, "alter")
    altertmp <- assign_coefs(altertmp$coefs_ts, "altertmp")
    # Note: as opposed to the period value alter coefs (`alter$coefs_ts`), the default temporal total 
    #       alterability coefficients (`altertmp$coefs_ts`) for cases not specified in the problem specs 
    #       data frame remain `NA` for now (i.e., argument `alter_temporal` is not used right away).
    #       This will allow for easy identification of the first specified temporal total alter coef in the specs 
    #       (when applicable) inside a complete temporal group.
    
  } else {
    pos_ser <- character(0L)
    neg_ser <- character(0L)
    mix_ser <- character(0L)
    alter <- c(alter, list(nondated_coefs = numeric(0L),
                           nondated_id_vec = integer(0L),
                           dated_id_vec = integer(0L)))
    altertmp <- c(altertmp, list(nondated_coefs = numeric(0L),
                                 nondated_id_vec = integer(0L),
                                 dated_id_vec = integer(0L)))
  }
  
  
  # Build the initial version of problem constraints elements:
  #   - matrix `A` : balancing constraint coefficients matrix
  #   - vector `op`: balancing constraint operator strings ("==", "<=" or ">=")
  #   - vector `b` : balancing constraint RHS
  #
  # Two versions of these elements are built:
  #   - <obj>1: for single period processing (incomplete temporal groups or period-by-period processing)
  #   - <obj>2: for multiple periods processing (complete temporal groups)
  #
  # Note: these constraint elements will eventually be "augmented" for OSQP to implement:
  #         - (implicit) temporal aggregation constraints (for complete temporal groups)
  #         - period value (lower/upper) bounds
  #
  con_labels <- labels_df$row.lc[labels_df$con.flag]
  n_con <- length(con_labels)
  A1 <- matrix(0, nrow = n_con, ncol = n_ser)
  A2 <- matrix(0, nrow = n_con * temporal_grp_periodicity, ncol = n_ser * temporal_grp_periodicity)
  for (ii in 1:n_con) {
    logi_vec <- coefs_df$row.lc == con_labels[ii] & coefs_df$col != "_rhs_"
    id_vec <- match(coefs_df$col[logi_vec], ser_names)
    A1[ii, id_vec] <- coefs_df$coef[logi_vec]
    for (jj in 1:temporal_grp_periodicity) {
      A2[(ii - 1) * temporal_grp_periodicity + jj, (id_vec - 1) * temporal_grp_periodicity + jj] <- coefs_df$coef[logi_vec]
    }
  }
  b1 <- rep.int(0, n_con)
  b1[match(coefs_df$row.lc[coefs_df$col == "_rhs_"], con_labels)] <- coefs_df$coef[coefs_df$col == "_rhs_"]
  b2 <- b1[rep(1:n_con, each = temporal_grp_periodicity)]
  op1 <- labels_df$type.lc[labels_df$con.flag]
  op1[op1 == "eq"] <- "=="
  op1[op1 == "le"] <- "<="
  op1[op1 == "ge"] <- ">="
  op2 <- op1[rep(1:n_con, each = temporal_grp_periodicity)]
  
  
  # Return the balancing problem core elements (building blocks)
  list(labels_df = labels_df,
       coefs_df = coefs_df,
       values_ts = values_ts,
       lb = lb,
       ub = ub,
       alter = alter,
       altertmp = altertmp,
       ser_names = ser_names,
       pos_ser = pos_ser,
       neg_ser = neg_ser,
       mix_ser = mix_ser,
       A1 = A1,
       op1 = op1,
       b1 = b1,
       A2 = A2,
       op2 = op2,
       b2 = b2)
}


#' Convert a time series balancing problem into a QP problem for OSQP.
#'
#' Balancing problem values arguments (numeric vectors of size `n`) including time series
#' temporal totals (when applicable)
#' @param x vector of initial values
#' @param c vector of alterability coefficients
#' @param lb,ub vectors of lower/upper bounds (`-Inf` or `Inf` correspond to no bounds)
#' @param bd_tol vector of tolerances for lower/upper bounds (0 corresponds to no tolerance)
#'
#' Balancing problem constraints arguments (matrices and vectors of size `m X n` and `m`)
#' including the implicit temporal aggregation constraints (when applicable)
#' @param A left-hand-side matrix (numeric)
#' @param op vector of operators (character: "==", "<=" or ">=")
#' @param b vector of right-hand side values (numeric)
#' @param b_tol vector of right-hand side value tolerances (0 corresponds to no tolerance)
#'
#' @returns
#' Elements of the (reduced) QP problem for OSQP:
#'   - P    : Hessian (quadratic coefficients) matrix
#'   - q    : vector of linear coefficients
#'   - A,l,u: constraints matrix and bounds (`l <= Ax <= u`)
#' Additional information:
#'   - x        : reduced problem initial values (input vector `x` stripped of the binding values)
#'   - id_x     : id's of input vector `x` corresponding to the reduced problem values
#'   - id_bd_con: id's of input vector `x` corresponding to the lower/upper bound constraints
#'
#' @details
#' The returned QP problem corresponds the balancing problem in its "reduced form" where binding values 
#' (`c * x = 0`) are removed from the problem and transferred into the constraints bounds (`l` and `u`) 
#' instead. 
#' 
#' Construct the reduced QP problem's Hessian matrix (`P`) and vector of linear coefficients (`q`), 
#' rewrite the balancing constraints as `l <= Ax <= u` linear constraints, i.e.,
#'   - `l = b - b_tol` and `u = b + b_tol` for "==" constraints,
#'   - `l = -Inf` and `u = b + b_tol` for "<=" constraints,
#'   - `l = b - b_tol` and `u = Inf` for ">=" constraints,
#' and define additional `l <= Ax <= u` linear constraints for lower/upper bounds, i.e.,
#'   - `l = lb - bd_tol` and `u = Inf` for lower bounds (w/o upper bounds),
#'   - `l = -Inf` and `u = ub + bd_tol` for upper bounds (w/o lower bounds),
#'   - `l = lb - bd_tol` and `u = ub + bd_tol` for both lower AND upper bounds.
#'   
#' Fixed QP problems (`validation_only = TRUE` or all values of `x` are binding) have missing (`NA`) values 
#' for `P` and `q` and should not be sent to OSQP (the initial solution should be returned).
#'
#' @noRd
build_reduced_qp <- function(x, c, lb, ub, bd_tol,  # problem variables info
                             A, op, b, b_tol) {     # problem constraints info
  
  # Initial number of constraints (`dim_A[1]`) and variables (`dim_A[2]`)
  dim_A <- dim(A)

  # Identify the set of free (nonbinding) problem values (when in "solving mode", i.e., `validation_only = FALSE`)
  # => alterability coefficients are all `NA` when in "validation mode" (`validation_only = TRUE`)
  logi_vec <- !is.na(c) & (x * c) != 0

  # Reduce the problem for fixed (binding) values
  if (any(logi_vec)) {
    id_x <- which(logi_vec)

    # Update the constraints given the fixed (binding) values:
    #   - adjust the constraint RHS (`b`) values
    #   - remove the corresponding column from constrain coefficients matrix (`A`)
    b <- b - A[, !logi_vec, drop = FALSE] %*% x[!logi_vec]
    A <- A[, id_x, drop = FALSE]
    dim_A[2] <- length(id_x)
    
    # QP problem quadratic coefficients (Hessian) matrix and vector of linear coefficients
    w <- 1 / abs(c[id_x] * x[id_x])
    P <- diag(2 * w, nrow = dim_A[2])
    q <- -2 * w * x[id_x]

  # Fixed problem (`validation_only = TRUE` or all values are binding) => do not reduce the problem
  } else {
    id_x <- seq_along(x)

    # (dummy) QP problem quadratic coefficients (Hessian) matrix and vector of linear coefficients
    # (won't be sent to OSQP for solving anyway)
    P <- as.matrix(NA_real_)
    q <- NA_real_
  }

  # Transform constraint RHS values (`b`) into `l` and `u` vectors (constraints lower/upper bounds)
  l <- rep.int(-Inf, dim_A[1])
  u <- rep.int(Inf, dim_A[1])
  logi_vec <- op != "<="
  l[logi_vec] <- b[logi_vec] - b_tol[logi_vec]
  logi_vec <- op != ">="
  u[logi_vec] <- b[logi_vec] + b_tol[logi_vec]
  
  # Add constraints for period value lower/upper bounds.
  logi_vec <- lb[id_x] != -Inf | ub[id_x] != Inf
  if (any(logi_vec)) {
    # Note regarding both vectors of id's defined here:
    #   - `id_bd`    : id's of the output 'x' vector (reduced problem)
    #   - `id_bd_con`: id's of the input 'x' vector (full scale problem)
    id_bd <- which(logi_vec)
    id_bd_con <- id_x[id_bd]
    n_bd <- length(id_bd)
    
    # Initial transposed matrix (0 values) for the new rows (new columns in this transposed matrix)
    tmp_mat <- matrix(0, nrow = dim_A[2], ncol = n_bd)
    
    # Set the proper matrix elements to 1 (column-major order)
    tmp_mat[id_bd + seq.int(from = 0, by = dim_A[2], length.out = n_bd)] <- 1
    
    # Add the new constraints
    A <- rbind(A, t(tmp_mat))
    l <- c(l, lb[id_bd_con] - bd_tol[id_bd_con])  # `-Inf` minus any value remains `-Inf`
    u <- c(u, ub[id_bd_con] + bd_tol[id_bd_con])  #  `Inf`  plus any value remains  `Inf`
  } else {
    id_bd_con <- integer(0L)
  }

  # Return the reduced QP problem info
  list(P = P,
       q = q,
       A = A,
       l = l,
       u = u,
       x = x[id_x],
       id_x = id_x,
       id_bd_con = id_bd_con)
}


#' Calculate the balancing problem discrepancies
#'
#' @description
#' Calculate the discrepancies of a solution given a set of linear constraints
#' (`l <= Ax <= u`)
#'
#' @param Ax (vector) evaluated constraint body values.
#' @param l,u (vector) constraint lower and upper bounds.
#'
#' @return a list of three elements:
#'   - `discr_vec`: vector of constraint discrepancies (length `n`)
#'   - `max_discr`: maximum constraint discrepancy (`max(discr_vec)`)
#'   - `tot_discr`: sum of all constraint discrepancies (`sum(discr_vec)`)
#'
#' @details
#' Compare the evaluated constraints body (`Ax`) against the constraints bounds
#' (`l` and `u`). Discrepancies are positive values representing constraint bounds 
#' *violations*. A discrepancy of 0 means that the corresponding constraint bounds 
#' are fully respected.
#'   - Equality (`==`) constraints have `l = u`
#'   - Inequality (`<=` or `>=`) constraints either have `l = -Inf` or `u = Inf`
#' A given constraint discrepancy corresponds to `max(l - Ax, Ax - u, 0)`
#'
#' @noRd
calc_discr <- function(Ax, l, u) {
  
  discr_vec <- pmax.int(l - Ax, Ax - u, 0)
  list(discr_vec = discr_vec,
       max_discr = max(discr_vec),
       tot_discr = sum(discr_vec))
}


#' Return the Initial Solution
#' 
#' @description
#' Do not solve the balancing problem. Simply return the initial solution info.
#'
#' @param x (double vector) initial problem values (initial solution).
#' @param Ax (double vector) initial evaluated constraint body values.
#' @param discr (list) initial constraint discrepancies list.
#' @param validation_tol (double) solution validation tolerance (max. allowed constraint discr.).
#' @param ... receptacle for other (ignored) `solve_one_osqp()` arguments.
#'
#' @return a list with the following elements elements:
#'   - `x`: (double vector) initial solution.
#'   - `status`: (string) solution status description.
#'   - `status_val`: (integer) solution status value.
#'   - `Ax`: (double vector) evaluated constraint body values with the initial solution.
#'   - `discr`: (list) initial constraint discrepancies.
#'   - `type`: (string) solution type (`"initial"`).
#'   - `osqp_attempts`: (integer) number of OSQP attempts (`0`).
#'   - `osqp_seqno`: (integer) sequence (iteration) number of the OSQP solution (`NA`).
#'   - `osqp_settings`: (list) OSQP solution settings (`NULL`).
#'   - `osqp_info`: (list) OSQP solution info (`NULL`).
#'   - `seq_time`: solving sequence execution time in seconds.
#'
#' @details
#' This function is used when main argument `validation_only = TRUE`
#'
#' @noRd
return_initial_sol <- function(x,
                               Ax,
                               discr,
                               validation_tol,
                               ...) {

  start_time <- Sys.time()
  
  if (discr$max_discr <= validation_tol) {
    status <- "valid initial solution"
    status_val <- 1
  } else {
    status <- "invalid initial solution"
    status_val <- -1
  }
  
  list(x = x,
       status = status,
       status_val = status_val,
       Ax = Ax,
       discr = discr,
       type = "initial",
       osqp_attempts = 0,
       osqp_seqno = NA_integer_,
       osqp_settings = NULL,
       osqp_info = NULL,
       seq_time = Sys.time() - start_time)
}


#' Solve the balancing problem
#' 
#' @description
#' Solve a single balancing problem with OSQP.
#'
#' @param x (double vector) initial problem values (initial solution).
#' @param Ax (double vector) evaluated constraint body values with the initial solution.
#' @param discr (list) initial constraint discrepancies list.
#' @param validation_tol (double) solution validation tolerance (max. allowed constraint discr.).
#' @param P,q,A,l,u (double matrix, vector, matrix, vector and vector) OSQP problem elements.
#' @param settings_df (data.frame) OSQP settings sequence data frame.
#' @param full_sequence (logical) perform all steps (rows) of `settings_df`.
#' @param solve_msg_func (function) message display function.
#' @param osqpBeg_display_func (function) OSQP model and iterations display set up function.
#' @param osqpEnd_display_func (function) OSQP model and iterations display wrap up function.
#' @param osqp_output_file (string) temporary file (name and path) for the OSQP model and iterations contents.
#' @param trunc_to_zero_tol (double) tolerance for changing nonzero solution data points to 0.
#' 
#' @return a list with the following elements elements:
#'   - `x`: (double vector) final solution.
#'   - `status`: (string) solution status description.
#'   - `status_val`: (integer) solution status value.
#'   - `Ax`: (double vector) evaluated constraint body values with the final solution.
#'   - `discr`: (list) final constraint discrepancies list.
#'   - `type`: (string) solution type (`"initial"` or `"osqp"`).
#'   - `osqp_attempts`: (integer) number of OSQP attempts.
#'   - `osqp_seqno`: (integer) sequence (iteration) number of the OSQP solution (`NA` if `type = "initial"`).
#'   - `osqp_settings`: (list) OSQP solution settings (`NULL` if `type = "initial"`).
#'   - `osqp_info`: (list) OSQP solution info (`NULL` if `type = "initial"`).
#'   - `seq_time`: solving sequence execution time in seconds.
#'
#' @details
#' This function is used when main argument `validation_only = FALSE`
#'
#' The settings data frame (argument `settings_df`) dictates the total number of attempts made
#' at solving the problem. Each attempt is made with the list of OSQP settings for the corresponding
#' `settings_df` observation (row). Unless a full sequence is specified (argument `full_sequence = TRUE`), 
#' the solving sequence stops as soon as a valid solution is obtained (a solution for which all constraint 
#' discrepancies are smaller or equal to the tolerance specified with argument `validation_tol`) unless setting 
#' `require_polished = TRUE` is specified in the settings data frame for that step, in which case a polished 
#' solution from OSQP (`status_polish = 1`) would also be required to stop the sequence. 
#' 
#' Note that the constraint discrepancies are calculated on the *zero truncated* solution values, when applicable 
#' (i.e., when argument `trunc_to_zero_tol > 0`), therefore ensuring accurate validation of the final solution.
#'
#' @noRd
solve_one_osqp <- function(x,
                           Ax,
                           discr,
                           validation_tol,
                           P, q, A, l, u,
                           settings_df,
                           full_sequence,
                           solve_msg_func,
                           osqpBeg_display_func,
                           osqpEnd_display_func,
                           osqp_output_file,
                           trunc_to_zero_tol) {


  start_time <- Sys.time()
  
  # Print the initial solution discrepancies
  solve_msg_func("  Initial solution:\n    - Maximum discrepancy = ", format(discr$max_discr, big.mark = ","),
                 "\n    - Total discrepancy   = ", format(discr$tot_discr, big.mark = ","), "\n")

  
  # No initial discrepancies: return the initial solution
  if (discr$tot_discr <= gs.min_tolerance) {
    
    solve_msg_func("  No discrepancies. Returning the intial values.\n")
    
    status <- "valid initial solution"
    status_val <- 1
    type <- "initial"
    osqp_attempts <- 0
    osqp_seqno <- NA_integer_
    osqp_settings <- NULL
    osqp_info <- NULL

    
  # Fixed problem (all values are either 0 or binding): return the initial solution
  } else if (any(is.na(q))) {
    
    if (discr$max_discr <= validation_tol) {
      solve_msg_func("  Fixed balancing problem (all values are either 0 or binding) with a valid initial solution (maximum discrepancy <= ", 
                     format(validation_tol), " = `validation_tol`).\n")
      status <- "valid initial solution"
      status_val <- 1
    } else {
      warning("Unsolvable fixed balancing problem (all values are either 0 or binding) with an invalid initial solution (maximum discrepancy > ", 
              format(validation_tol), " = `validation_tol`).", call. = FALSE, immediate. = TRUE)
      message("\n")
      status <- "unsolvable fixed problem"
      status_val <- -4
    }
    type <- "initial"
    osqp_attempts <- 0
    osqp_seqno <- NA_integer_
    osqp_settings <- NULL
    osqp_info <- NULL


  # At least one free (nonbinding) value: try to solve with OSQP
  } else {
    
    # Note regarding flags `valid` and `satisfactory` used for the OSQP solving loop:
    #  - `valid`       : is TRUE if a valid solution (`max_discr <= validation_tol`) hasbeen found (initial or OSQP).
    #  - `satisfactory`: is TRUE if a satisfactory OSQP solution (`valid & polished` when required, `valid` otherwise) 
    #                    has been found.
    #
    # `satisfactory == TRUE` stops the solving loop (unless 'full_sequence == TRUE`); it is therefore initialized to 
    # `FALSE` here (as we hope to improve the initial solution with OSQP).
    satisfactory <- FALSE
    if (discr$max_discr <= validation_tol) {
      valid <- TRUE
      solve_msg_func("  Valid solution (maximum discrepancy <= ", format(validation_tol), " = `validation_tol`).\n")
      status <- "valid initial solution"
      status_val <- 1
    } else {
      valid <- FALSE
      status <- "invalid initial solution"
      status_val <- -1
    }
    solve_msg_func("  Try to find a better solution with OSQP.\n")

    # Calculate the scaling factor
    mean_abs_x <- mean(abs(x))

    
    # Solving loop
    best_osqp <- NA_integer_
    for (ii in seq_along(settings_df$verbose)) {

      # Set the scaling factor
      if (settings_df$prior_scaling[ii]) {
        factor_ii <- mean_abs_x
      } else {
        factor_ii <- 1
      }
      
      osqpBeg_display_func(osqp_output_file)

      # Build the model (ii-th row of `settings_df`)
      model_ii <- osqp::osqp(P * factor_ii, q, A, l / factor_ii, u / factor_ii, as.list(settings_df[ii, , drop = FALSE]))

      # Solve
      sol_ii <- model_ii@Solve() 
      sol_ii$x <- sol_ii$x * factor_ii  # original scale

      osqpEnd_display_func(osqp_output_file)

      # Trunc to zero
      sol_ii$x[abs(sol_ii$x) <= trunc_to_zero_tol] <- 0

      
      # Evaluate:
      #  a) update the "current solution" when relevant (smallest `tot_discr` among valid/invalid solutions so far)
      #  b) stop the solving loop if a satisfactory OSQP solution was found and not doing a full OSQP sequence
      Ax_ii <- A %*% sol_ii$x
      discr_ii <- calc_discr(Ax_ii, l, u)
      polished_ii <- sol_ii$info$status_polish == 1
      solve_msg_func("  OSQP iteration ", ii, ":\n    - Maximum discrepancy = ", format(discr_ii$max_discr, big.mark = ","),
                     "\n    - Total discrepancy   = ", format(discr_ii$tot_discr, big.mark = ","), "\n")
      
      # Valid solution (`max_discr <= validation_tol`)
      if (discr_ii$max_discr <= validation_tol) {
        solve_msg_func("  Valid solution (maximum discrepancy <= ", format(validation_tol), " = `validation_tol`).\n")

        # Polished solution required
        if (settings_df$require_polished[ii]) {
          if (polished_ii) {
            satisfactory <- TRUE
            solve_msg_func("  Required polished solution achieved.\n")
            status_ii <- "valid polished osqp solution"
            status_val_ii <- 2
          } else {
            solve_msg_func("  Required polished solution NOT achieved.\n")
            status_ii <- "valid unpolished osqp solution"
            status_val_ii <- 3
          }
          
        # Polished solution NOT required
        } else {
          satisfactory <- TRUE
          if (polished_ii) {
            status_ii <- "valid polished osqp solution"
            status_val_ii <- 2
          } else {
            status_ii <- "valid unpolished osqp solution"
            status_val_ii <- 3
          }
        }
        
        # Update the "current (valid) solution"
        if (!valid || discr_ii$tot_discr < discr$tot_discr) {
          valid <- TRUE
          status <- status_ii
          status_val <- status_val_ii
          Ax <- Ax_ii
          discr <- discr_ii
          best_osqp <- ii
          model <- model_ii
          sol <- sol_ii
          scaling_factor <- factor_ii
        }
        
        # Stop the solving loop
        if (satisfactory && !full_sequence) {
          break
        }
        
      # Invalid solution (`max_discr > validation_tol`)
      } else {
        solve_msg_func("  Invalid solution (maximum discrepancy > ", format(validation_tol), " = `validation_tol`).\n")
        
        # Update the "current (invalid) solution"
        if (!valid && discr_ii$tot_discr < discr$tot_discr) {
          if (polished_ii) {
            status <- "invalid polished osqp solution"
            status_val <- -2
          } else {
            status <- "invalid unpolished osqp solution"
            status_val <- -3
          }
          Ax <- Ax_ii
          discr <- discr_ii
          best_osqp <- ii
          model <- model_ii
          sol <- sol_ii
          scaling_factor <- factor_ii
        }
      }
    }

    
    # Wrap up the final solution    
    osqp_attempts <- ii
    
    # Initial solution
    if (is.na(best_osqp)) {
      solve_msg_func("  The initial solution could not be improved with OSQP. Returning the initial solution.\n")
      type <- "initial"
      osqp_seqno <- NA_integer_
      osqp_settings <- NULL
      osqp_info <- NULL
      
    # OSQP solution
    } else {
      if (!satisfactory) {
        if (valid) {
          word <-  "valid"
        } else {
          word <-  "all"
        }
        solve_msg_func("  A satisfactory OSQP solution could not be obtained. Returning the solution with the smallest ",
                       "total discrepancy among ", word, " solutions (OSQP iteration ", best_osqp, ").\n")
      }
      x <- sol$x
      type <- "osqp"
      osqp_seqno <- best_osqp
      osqp_settings <- c(
        model@GetParams(), 
        list(
          prior_scaling = settings_df$prior_scaling[best_osqp], 
          require_polished = settings_df$require_polished[best_osqp]
        )
      )
      # Remove old `polish` setting (replaced with `polishing` since OSQP v1.0 while both `polish` and 
      # `polishing` are returned by the `GetParams()` method...)
      if (!is.null(osqp_settings$polishing)) {
        osqp_settings$polish <- NULL
      }
      osqp_info <- c(sol$info, list(prior_scaling_factor = scaling_factor))
    }
  }

    
  # Return the final solution
  list(x = x,
       status = status,
       status_val = status_val,
       Ax = Ax,
       discr = discr,
       type = type,
       osqp_attempts = osqp_attempts,
       osqp_seqno = osqp_seqno,
       osqp_settings = osqp_settings,
       osqp_info = osqp_info,
       seq_time = Sys.time() - start_time)
}

#' Restore cross-sectional (contemporaneous) aggregation constraints
#'
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/tsraking.html}})}
#' 
#' _Replication of the G-Series 2.0 SAS\eqn{^\circledR}{®} TSRAKING procedure (PROC TSRAKING). See the
#' G-Series 2.0 documentation for details (Statistics Canada 2016)._
#'
#' This function will restore cross-sectional aggregation constraints in a system of time series. The
#' aggregation constraints may come from a 1 or 2-dimensional table. Optionally, temporal constraints
#' can also be preserved.
#'
#' [tsraking()] is usually called in practice through [tsraking_driver()] in order to reconcile
#' all periods of the time series system in a single function call.
#'
#'
#' @usage
#' tsraking(
#'   data_df,
#'   metadata_df,
#'   alterability_df = NULL,
#'   alterSeries = 1,
#'   alterTotal1 = 0,
#'   alterTotal2 = 0,
#'   alterAnnual = 0,
#'   tolV = 0.001,
#'   tolP = NA,
#'   warnNegResult = TRUE,
#'   tolN = -0.001,
#'   id = NULL,
#'   verbose = FALSE,
#'
#'   # New in G-Series 3.0
#'   Vmat_option = 1,
#'   warnNegInput = TRUE,
#'   quiet = FALSE
#' )
#'
#'
#' @param data_df (mandatory)
#'
#' Data frame (object of class "data.frame") that contains the time series data to be reconciled. It must minimally 
#' contain variables corresponding to the component series and cross-sectional control totals specified in the 
#' metadata data frame (argument `metadata_df`). If more than one observation (period) is provided, the sum of 
#' the provided component series values will also be preserved as part of implicit temporal constraints.
#'
#' @param metadata_df (mandatory)
#'
#' Data frame (object of class "data.frame") that describes the cross-sectional aggregation constraints 
#' (additivity rules) for the raking problem. Two character variables must be included in the metadata data frame: 
#' `series` and `total1`. Two variables are optional: `total2` (character) and `alterAnnual` (numeric). The values 
#' of variable `series` represent the variable names of the component series in the input time series data frame 
#' (argument `data_df`). Similarly, the values of variables `total1` and `total2` represent the variable names of 
#' the 1<sup>st</sup> and 2<sup>nd</sup> dimension cross-sectional control totals in the input time series data 
#' frame. Variable `alterAnnual` contains the alterability coefficient for the temporal constraint associated to 
#' each component series. When specified, the latter will override the default alterability coefficient specified 
#' with argument `alterAnnual`.
#'
#' @param alterability_df (optional)
#'
#' Data frame (object of class "data.frame"), or `NULL`, that contains the alterability coefficients variables. 
#' They must correspond to a component series or a cross-sectional control total, that is, a variable with the same 
#' name must exist in the input time series data frame (argument `data_df`). The values of these alterability 
#' coefficients will override the default alterability coefficients specified with arguments `alterSeries`, 
#' `alterTotal1` and `alterTotal2`. When the input time series data frame contains several observations and the 
#' alterability coefficients data frame contains only one, the alterability coefficients are used (repeated) for 
#' all observations of the input time series data frame. Alternatively, the alterability coefficients data frame 
#' may contain as many observations as the input time series data frame.
#'
#' **Default value** is `alterability_df = NULL` (default alterability coefficients).
#'
#' @param alterSeries (optional)
#'
#' Nonnegative real number specifying the default alterability coefficient for the component series values. It 
#' will apply to component series for which alterability coefficients have not already been specified in the 
#' alterability coefficients data frame (argument `alterability_df`).
#'
#' **Default value** is `alterSeries = 1.0` (nonbinding component series values).
#'
#' @param alterTotal1 (optional)
#'
#' Nonnegative real number specifying the default alterability coefficient for the 1<sup>st</sup> dimension
#' cross-sectional control totals. It will apply to cross-sectional control totals for which alterability 
#' coefficients have not already been specified in the alterability coefficients data frame (argument 
#' `alterability_df`).
#'
#' **Default value** is `alterTotal1 = 0.0` (binding 1<sup>st</sup> dimension cross-sectional control totals)
#'
#' @param alterTotal2 (optional)
#'
#' Nonnegative real number specifying the default alterability coefficient for the 2<sup>nd</sup> dimension
#' cross-sectional control totals. It will apply to cross-sectional control totals for which alterability 
#' coefficients have not already been specified in the alterability coefficients data frame (argument 
#' `alterability_df`).
#'
#' **Default value** is `alterTotal2 = 0.0` (binding 2<sup>nd</sup> dimension cross-sectional control totals).
#'
#' @param alterAnnual (optional)
#'
#' Nonnegative real number specifying the default alterability coefficient for the component series temporal
#' constraints (e.g., annual totals). It will apply to component series for which alterability coefficients
#' have not already been specified in the metadata data frame (argument `metadata_df`).
#'
#' **Default value** is `alterAnnual = 0.0` (binding temporal control totals).
#'
#' @param tolV,tolP (optional)
#'
#' Nonnegative real number, or `NA`, specifying the tolerance, in absolute value or percentage, to be used 
#' when performing the ultimate test in the case of binding totals (alterability coefficient of \eqn{0.0} 
#' for temporal or cross-sectional control totals). The test compares the input binding control totals with 
#' the ones calculated from the reconciled (output) component series. Arguments `tolV` and `tolP` cannot be both 
#' specified together (one must be specified while the other must be `NA`).
#'
#' **Example:** to set a tolerance of 10 *units*, specify `tolV = 10, tolP = NA`; to set a tolerance of 1%, 
#' specify `tolV = NA, tolP = 0.01`.
#'
#' **Default values** are `tolV = 0.001` and `tolP = NA`.
#'
#' @param warnNegResult (optional)
#'
#' Logical argument specifying whether a warning message is generated when a negative value created by the
#' function in the reconciled (output) series is smaller than the threshold specified by argument `tolN`.
#'
#' **Default value** is `warnNegResult = TRUE`.
#'
#' @param tolN (optional)
#'
#' Negative real number specifying the threshold for the identification of negative values. A value is
#' considered negative when it is smaller than this threshold.
#'
#' **Default value** is `tolN = -0.001`.
#'
#' @param id (optional)
#'
#' String vector (minimum length of 1), or `NULL`, specifying the name of additional variables to be transferred
#' from the input time series data frame (argument `data_df`) to the output time series data frame, the
#' object returned by the function (see section **Value**). By default, the output series data frame only contains
#' the variables listed in the metadata data frame (argument `metadata_df`).
#'
#' **Default value** is `id = NULL`.
#'
#' @param verbose (optional)
#'
#' Logical argument specifying whether information on intermediate steps with execution time (real time,
#' not CPU time) should be displayed. Note that specifying argument `quiet = TRUE` would *nullify* argument 
#' `verbose`.
#'
#' **Default value** is `verbose = FALSE`.
#'
#' @param Vmat_option (optional)
#'
#' Specification of the option for the variance matrices (\eqn{V_e} and \eqn{V_\epsilon}; see section **Details**):
#' | **Value** | **Description** |
#' | :-------: | :-------------- |
#' | `1` | Use vectors \eqn{x} and \eqn{g} in the variance matrices. |
#' | `2` | Use vectors \eqn{|x|} and \eqn{|g|} in the variance matrices. |
#'
#' See Ferland (2016) and subsection **Arguments `Vmat_option` and `warnNegInput`** in section **Details** for 
#' more information.
#'
#' **Default value** is `Vmat_option = 1`.
#'
#' @param warnNegInput (optional)
#'
#' Logical argument specifying whether a warning message is generated when a negative value smaller than
#' the threshold specified by argument `tolN` is found in the input time series data frame (argument `data_df`).
#'
#' **Default value** is `warnNegInput = TRUE`.
#'
#' @param quiet (optional)
#'
#' Logical argument specifying whether or not to display only essential information such as warnings and errors.
#' Specifying `quiet = TRUE` would also *nullify* argument `verbose` and is equivalent to _wrapping_ your 
#' [tsraking()] call with [suppressMessages()].
#'
#' **Default value** is `quiet = FALSE`.
#'
#'
#' @details
#' This function returns the generalized least squared solution of a specific, simple variant of the general 
#' regression-based raking model proposed by Dagum and Cholette (Dagum and Cholette 2006). The model, in matrix form, is:
#' \deqn{\displaystyle
#' \begin{bmatrix} x \\ g \end{bmatrix} = 
#' \begin{bmatrix} I \\ G \end{bmatrix} \theta + 
#' \begin{bmatrix} e \\ \varepsilon \end{bmatrix}
#' }{[x; g] = [I; G] theta + [e; epsilion]}
#' where
#' * \eqn{x} is the vector of the initial component series values.
#' * \eqn{\theta} is the vector of the final (reconciled) component series values.
#' * \eqn{e \sim \left( 0, V_e \right)}{e ~ (0, V_e)} is the vector of the measurement errors of \eqn{x} with covariance
#' matrix \eqn{V_e = \mathrm{diag} \left( c_x x \right)}{V_e = diag(c_x x)}, or \eqn{V_e = \mathrm{diag} \left( \left| 
#' c_x x \right| \right)}{V_e = diag(|c_x x|)} when argument `Vmat_option = 2`, where \eqn{c_x} is the vector of the 
#' alterability coefficients of \eqn{x}.
#' * \eqn{g} is the vector of the initial control totals, including the component series temporal totals (when 
#' applicable).
#' * \eqn{\varepsilon \sim (0, V_\varepsilon)}{epsilon ~ (0, V_epsilon)} is the vector of the measurement errors of 
#' \eqn{g} with covariance matrix \eqn{V_\varepsilon = \mathrm{diag} \left( c_g g \right)}{V_epsilion = diag(c_g g)}, or 
#' \eqn{V_\varepsilon = \mathrm{diag} \left( \left| c_g g \right| \right)}{V_epsilon = diag(|c_g g|)} when argument 
#' `Vmat_option = 2`, where \eqn{c_g} is the vector of the alterability coefficients of \eqn{g}.
#' * \eqn{G} is the matrix of aggregation constraints, including the implicit temporal constraints (when applicable). 
#' 
#' The generalized least squared solution is:
#' \deqn{\displaystyle 
#' \hat{\theta} = x + V_e G^{\mathrm{T}} \left( G V_e G^{\mathrm{T}} + V_\varepsilon \right)^+ \left( g - G x \right)
#' }{theta^hat = x + V_e G' (G V_e G' + V_epsilion)^{+} (g - G x)}
#' where \eqn{A^{+}} designates the Moore-Penrose inverse of matrix \eqn{A}.
#' 
#' [tsraking()] solves a single raking problem, i.e., either a single period of the time series system, or a single 
#' temporal group (e.g., all periods of a given year) when temporal total preservation is required. Several call to 
#' [tsraking()] are therefore necessary in order to reconcile all the periods of the time series system. 
#' [tsraking_driver()] can achieve this in a single call: it conveniently determines the required set of raking 
#' problems to be solved and internally generates the individual calls to [tsraking()].
#'
######
# This subsection differs slightly between `tsraking()` and `tsbalancing` and is therefore maintained for both functions 
# (in both sets of roxygen2 comments) as opposed to being shared with `roxygen2 tag `@inheritSection`.
# => the "Temporal total preservation* paragraph is the SAME, however: keep them "in sync"!.
######
#' ## Alterability Coefficients
#' Alterability coefficients \eqn{c_x} and \eqn{c_g} conceptually represent the measurement errors associated with the 
#' input component series values \eqn{x} and control totals \eqn{g} respectively. They are nonnegative real numbers which,  
#' in practice, specify the extent to which an initial value can be modified in relation to other values. Alterability 
#' coefficients of \eqn{0.0} define fixed (binding) values while alterability coefficients greater than \eqn{0.0} define 
#' free (nonbinding) values. Increasing the alterability coefficient of an intial value results in more changes for that 
#' value in the reconciled (output) data and, conversely, less changes when decreasing the alterability coefficient. The 
#' default alterability coefficients are \eqn{1.0} for the component series values and \eqn{0.0} for the cross-sectional 
#' control totals and, when applicable, the component series temporal totals. These default alterability coefficients 
#' result in a proportional allocation of the discrepancies to the component series. Setting the component series 
#' alterability coefficients to the inverse of the component series initial values would result in a uniform allocation 
#' of the discrepancies instead. *Almost binding* totals can be obtained in practice by specifying very small 
#' (almost \eqn{0.0}) alterability coefficients relative to those of the (nonbinding) component series.
#' 
#' **Temporal total preservation** refers to the fact that temporal totals, when applicable, are usually kept 
#' “as close as possible” to their initial value. *Pure preservation* is achieved by default with binding temporal 
#' totals while the change is minimized with nonbinding temporal totals (in accordance with the set of alterability 
#' coefficients).
#'
#' ## Arguments `Vmat_option` and `warnNegInput`
#' These arguments allow for an alternative handling of negative values in the input data, similar to that of [tsbalancing()]. 
#' Their default values correspond to the G-Series 2.0 behaviour (SAS\eqn{^\circledR}{®} PROC TSRAKING) for which equivalent 
#' options are not defined. The latter was developed with "nonnegative input data only" in mind, similar to SAS\eqn{^\circledR}{®}
#' PROC BENCHMARKING in G-Series 2.0 that did not allow negative values either with proportional benchmarking, which explains
#' the "suspicious use of proportional raking" warning in presence of negative values with PROC TSRAKING in G-Series 2.0 and 
#' when `warnNegInput = TRUE` (default). However, (proportional) raking in the presence of negative values generally works well 
#' with `Vmat_option = 2` and produces reasonable, intuitive solutions. E.g., while the default `Vmat_option = 1` fails at 
#' solving constraint `A + B = C` with input data `A = 2`, `B = -2`, `C = 1` and the default alterability coefficients, 
#' `Vmat_option = 2` returns the (intuitive) solution `A = 2.5`, `B = -1.5`, `C = 1` (25% increase for `A` and `B`). See 
#' Ferland (2016) for more details.
#'
#' ## Treatment of Missing (`NA`) Values
#' Missing values in the input time series data frame (argument `data_df`) or alterability coefficients data frame
#' (argument `alterability_df`) for any of the raking problem data (variables listed in the metadata data frame
#' with argument `metadata_df`) will generate an error message and stop the function execution.
#' 
#' 
#' @inheritSection tsbalancing Comparing [tsraking()] and [tsbalancing()]
#' 
#' 
#' @returns
#' The function returns a data frame containing the reconciled component series, reconciled cross-sectional control 
#' totals and variables specified with  argument `id`. Note that the "data.frame" object can be explicitly coerced to
#' another type of object with the appropriate `as*()` function (e.g., `tibble::as_tibble()` would coerce it to a tibble).
#'
#'
#' @references Bérubé, J. and S. Fortier (2009). "PROC TSRAKING: An in-house SAS\eqn{^\circledR}{®} procedure for balancing 
#' time series". In **JSM Proceedings, Business and Economic Statistics Section**. Alexandria, VA: American Statistical 
#' Association.
#'
#' @references Dagum, E. B. and P. Cholette (2006). **Benchmarking, Temporal Distribution and Reconciliation Methods
#' of Time Series**. Springer-Verlag, New York, Lecture Notes in Statistics, Vol. 186. \doi{10.1007/0-387-35439-5}
#' 
#' @references Ferland, M. (2016). "Negative Values with PROC TSRAKING". **Internal document**. Statistics Canada, Ottawa, 
#' Canada.
#' 
#' @references Fortier, S. and B. Quenneville (2009). "Reconciliation and Balancing of Accounts and Time Series". 
#' In **JSM Proceedings, Business and Economic Statistics Section**. Alexandria, VA: American Statistical Association.
#'
#' @references Quenneville, B. and S. Fortier (2012). "Restoring Accounting Constraints in Time Series – Methods and
#' Software for a Statistical Agency". **Economic Time Series: Modeling and Seasonality**. Chapman & Hall, New York.
#'
#' @references Statistics Canada (2016). "The TSRAKING Procedure". **G-Series 2.0 User Guide**. Statistics Canada,
#' Ottawa, Canada.
#'
#' @references Statistics Canada (2018). **Theory and Application of Reconciliation (Course code 0437)**.
#' Statistics Canada, Ottawa, Canada.
#'
#'
#' @seealso [tsraking_driver()] [tsbalancing()] [rkMeta_to_blSpecs()] [gs.gInv_MP()] [build_raking_problem()] [aliases]
#'
#'
#' @example misc/function_examples/tsraking-ex.R
#'
#'
#' @export
tsraking <- function(data_df,
                     metadata_df,
                     alterability_df = NULL,
                     alterSeries = 1,
                     alterTotal1 = 0,
                     alterTotal2 = 0,
                     alterAnnual = 0,
                     tolV = 0.001,
                     tolP = NA,
                     warnNegResult = TRUE,
                     tolN = -0.001,
                     id = NULL,
                     verbose = FALSE,

                     # New in G-Series 3.0
                     Vmat_option = 1,
                     warnNegInput = TRUE,
                     quiet = FALSE) {




  ### Internal functions ###


  # Binding totals validation functions according to (main function) arguments 'tolV' and 'tolP':
  #   - binding_tot_diff_valid   : tolV is specified (check differences against tolV)
  #   - binding_tot_relDiff_valid: tolP is specified (check relative differences against tolP)
  binding_tot_diff_valid <- function(g_in, g_out, c_g, tol, ...) {
    max_discr <- max(abs(g_out - g_in) * (c_g == 0))
    if (max_discr > tol) {
      warning("Binding totals are not met (maximum difference: ", format(max_discr), "). Inconsistencies ",
              "in the data are suspected.\n", call. = FALSE, immediate. = TRUE)
    }
    invisible(NULL)
  }
  binding_tot_relDiff_valid <- function(g_in, g_out, c_g, tol, zero_tol = gs.tolerance) {
    g_in[abs(g_in) <= zero_tol] <- NA
    max_discr <- max(abs(g_out / g_in - 1) * 100 * (c_g == 0))
    tol <- tol * 100
    if (max_discr > tol) {
      warning("Binding totals are not met (maximum relative difference: ", format(max_discr), "%). Inconsistencies ",
              "in the data are suspected.\n", call. = FALSE, immediate. = TRUE)
    }
    invisible(NULL)
  }




  ### Main function ###

  start_time <- Sys.time()
  start_time0 <- start_time

  # Enforce the default R "error" option (`options(error = NULL)`). E.g. this Turns off traceback
  # generated by calls to the stop() function inside internal functions in R Studio.
  ini_error_opt <- getOption("error")
  on.exit(options(error = ini_error_opt))
  options(error = NULL)
  
  # Validate argument `quiet` and implement the quiet setting
  quiet <- gs.validate_arg_logi(quiet)
  if (quiet) {
    quiet_msg_func <- gs.NULL_func
    quiet_lab <- ""  # won't be displayed anyway
  } else {
    quiet_msg_func <- message
    quiet_lab <- "    (*)quiet        = FALSE (default)"
  }

  # Display the function header
  quiet_msg_func("\n\n", gs.header, "\n\ntsraking() function:\n")


  # Initial argument validation

  # Mandatory arguments (without default values)
  data_df_name <- deparse1(substitute(data_df))
  tmp <- nchar(data_df_name)
  if (tmp == 0) {
    stop("Argument 'data_df' is mandatory (it must be specified).\n\n", call. = FALSE)
  } else if (tmp >= 60) {
    data_df_name <- paste0(substr(data_df_name, 1, 55), "<...>")
  }
  if (grepl("structure(", data_df_name, fixed = TRUE)) {
    data_df_name <- "<argument 'data_df'>"
  }
  data_df <- data_df
  if (!is.data.frame(data_df)) {
    stop("Argument 'data_df' is not a 'data.frame' object.\n\n", call. = FALSE)
  }
  data_df <- as.data.frame(data_df)
  row.names(data_df) <- NULL
  metadata_df_name <- deparse1(substitute(metadata_df))
  tmp <- nchar(metadata_df_name)
  if (tmp == 0) {
    stop("Argument 'metadata_df' is mandatory (it must be specified).\n\n", call. = FALSE)
  } else if (tmp >= 60) {
    metadata_df_name <- paste0(substr(metadata_df_name, 1, 55), "<...>")
  }
  if (grepl("structure(", metadata_df_name, fixed = TRUE)) {
    metadata_df_name <- "<argument 'metadata_df'>"
  }
  metadata_df <- metadata_df
  if (!is.data.frame(metadata_df)) {
    stop("Argument 'metadata_df' is not a 'data.frame' object.\n\n", call. = FALSE)
  }
  metadata_df <- as.data.frame(metadata_df)
  row.names(metadata_df) <- NULL

  # Column names in 'data_df' are case sensitive (e.g. 'metadata_df' values must reflect the correct "casing")
  # Column names in 'metadata_df' are NOT case sensitive (e.g. "Series", "series" and "SERIES" are all valid)
  dataDF_cols <- names(data_df)
  metaDF_cols <- toupper(names(metadata_df))
  names(metadata_df) <- metaDF_cols
  P <- nrow(data_df)  # number of periods
    
  # Alterability coefficients data frame (optional argument `alterability_df`)
  alterDF_lab <- "    alterability_df ="
  dup_alter_msg <- ""
  if (!is.null(alterability_df) && P > 0) {
    alter_df_name <- deparse1(substitute(alterability_df))
    if (alter_df_name >= 60) {
      alter_df_name <- paste0(substr(alter_df_name, 1, 55), "<...>")
    }
    if (grepl("structure(", alter_df_name, fixed = TRUE)) {
      alter_df_name <- "<argument 'alterability_df'>"
    }
    alterability_df <- alterability_df
    if (!is.data.frame(alterability_df)) {
      # Accept `alterability_df = NA` as `alterability_df = NULL`
      tmp <- (unlist(alterability_df))[1]
      if (!identical(alterability_df, tmp) || !is.na(tmp)) {
        warning("Argument 'alterability_df' is not a 'data.frame' object. It will be ignored.\n",
                call. = FALSE, immediate. = TRUE)
      }
      alterability_df <- NULL
      alterDF_lab <- paste0(alterDF_lab, " NULL (default)")
      
    } else {
      alterability_df <- as.data.frame(alterability_df)
      n_perAlter <- nrow(alterability_df)
      if (n_perAlter == 0) {
        alterability_df <- NULL
        alterDF_lab <- paste0(alterDF_lab, " NULL (default)")
        
      } else if (n_perAlter != 1 && n_perAlter != P) {
        warning("The alterability data frame (argument 'alterability_df') must contain either a single row ",
                "or as many rows as the input series data frame. It will be ignored.\n",
                call. = FALSE, immediate. = TRUE)
        alterability_df <- NULL
        alterDF_lab <- paste0(alterDF_lab, " NULL (default)")
        
      } else {
        
        # Duplicate the alterability coefficients
        if (n_perAlter == 1 && P > 1) {
          alterability_df <- alterability_df[rep(1, P), , drop = FALSE]
          dup_alter_msg <- paste0("\nThe alterability data frame contains only one observation (row). The ",
                                  "coefficients will be reused for every row of the input series data frame.\n")
        } else {
          
        }
        row.names(alterability_df) <- NULL
        alterDF_lab <- paste(alterDF_lab, alter_df_name)
      }
    }
  } else {
    alterDF_lab <- paste0(alterDF_lab, " NULL (default)")
  }

  # Other optional arguments
  tmp <- (unlist(alterSeries))[1]
  if (!identical(alterSeries, tmp) || is.null(tmp) || !is.finite(tmp) || tmp < 0) {
    stop("Argument 'alterSeries' must be a nonnegative real number.\n\n", call. = FALSE)
  }
  tmp <- (unlist(alterTotal1))[1]
  if (!identical(alterTotal1, tmp) || is.null(tmp) || !is.finite(tmp) || tmp < 0) {
    stop("Argument 'alterTotal1' must be a nonnegative real number.\n\n", call. = FALSE)
  }
  tmp <- (unlist(alterTotal2))[1]
  if (!identical(alterTotal2, tmp) || is.null(tmp) || !is.finite(tmp) || tmp < 0) {
    stop("Argument 'alterTotal2' must be a nonnegative real number.\n\n", call. = FALSE)
  }
  tmp <- (unlist(alterAnnual))[1]
  if (!identical(alterAnnual, tmp) || is.null(tmp) || !is.finite(tmp) || tmp < 0) {
    stop("Argument 'alterAnnual' must be a nonnegative real number.\n\n", call. = FALSE)
  }
  tmp <- (unlist(tolV))[1]
  if (is.null(tmp)) {
    tolV <- NA_real_
  } else if (!identical(tolV, tmp) || !is.finite(tmp) && !is.na(tmp) || is.finite(tmp) && tmp < 0) {
    stop("Argument 'tolV' must be a nonnegative real number or NA.\n\n", call. = FALSE)
  }
  tmp <- (unlist(tolP))[1]
  if (is.null(tmp)) {
    tolP <- NA_real_
  } else if (!identical(tolP, tmp) || !is.finite(tmp) && !is.na(tmp) || is.finite(tmp) && tmp < 0) {
    stop("Argument 'tolP' must be a nonnegative real number or NA.\n\n", call. = FALSE)
  }
  warnNegResult <- gs.validate_arg_logi(warnNegResult)
  tmp <- (unlist(tolN))[1]
  if (!identical(tolN, tmp) || is.null(tmp) || !is.finite(tmp) || tmp >= 0) {
    stop("Argument 'tolN' must be a negative real number.\n\n", call. = FALSE)
  }
  id <- gs.cleanup_col_list(id)
  if (length(id) == 0) {
    id_lab <- "    id              = NULL (default)"
    id <- NULL
  } else {
    gs.validate_cols(id, dataDF_cols, data_df_name, source_str = "argument 'id'")
    id_lab <- paste0("    id              = ", paste(id, collapse = " "))
  }
  verbose <- gs.validate_arg_logi(verbose)
  tmp <- (unlist(Vmat_option))[1]
  if (!identical(Vmat_option, tmp) || is.null(tmp) || !(tmp %in% 1:2)) {
    stop("Argument 'Vmat_option' must take value 1 or 2.\n\n", call. = FALSE)
  }
  Vmat_option <- as.integer(Vmat_option)
  warnNegInput <- gs.validate_arg_logi(warnNegInput)
  # Argument `quiet` was already validated
  
  # Implement the verbose setting
  if (verbose && !quiet) {
    verbose_func <- gs.display_difftime
    verbose_lab <- "    verbose         = TRUE"
  } else {
    verbose_func <- gs.NULL_func
    verbose_lab <- "    verbose         = FALSE (default)"
  }
  
  # Negative results validation function and validate the binding total tolerances
  if (warnNegResult) {
    neg_res_func <- gs.check_neg
    warnNegResult_lab <- "    warnNegResult   = TRUE (default)"
  } else {
    neg_res_func <- gs.FALSE_func
    warnNegResult_lab <- "    warnNegResult   = FALSE"
  }
  if (!is.na(tolV)) {
    if (!is.na(tolP)) {
      stop("Arguments 'tolV' and 'tolP' cannot be both specified (one must be NA).\n\n", call. = FALSE)
    }
    binding_tot_valid_func <- binding_tot_diff_valid
    tol_parm <- tolV
    tol_lab <- paste0("    tolV            = ", format(tolV))
    if (abs(tolV - 0.001) < gs.tolerance) {
      tol_lab <- paste0(tol_lab, " (default)")
    }
  } else {
    if (is.na(tolP)) {
      stop("Arguments 'tolV' and 'tolP' cannot be both NA (one must be specified).\n\n", call. = FALSE)
    }
    binding_tot_valid_func <- binding_tot_relDiff_valid
    tol_parm <- tolP
    tol_lab <- paste0("    tolP            = ", format(tolP))
  }
  
  
  # Display the function call (argument values)
  quiet_msg_func("    data_df         = ", data_df_name)
  quiet_msg_func("    metadata_df     = ", metadata_df_name)
  quiet_msg_func(alterDF_lab)
  lab <- paste0("    alterSeries     = ", format(alterSeries))
  if (alterSeries == 1) {
    lab <- paste0(lab, " (default)")
  }
  quiet_msg_func(lab)
  lab <- paste0("    alterTotal1     = ", format(alterTotal1))
  if (alterTotal1 == 0) {
    lab <- paste0(lab, " (default)")
  }
  quiet_msg_func(lab)
  lab <- paste0("    alterTotal2     = ", format(alterTotal2))
  if (alterTotal2 == 0) {
    lab <- paste0(lab, " (default)")
  }
  quiet_msg_func(lab)
  lab <- paste0("    alterAnnual     = ", format(alterAnnual))
  if (alterAnnual == 0) {
    lab <- paste0(lab, " (default)")
  }
  quiet_msg_func(lab)
  quiet_msg_func(tol_lab)
  quiet_msg_func(warnNegResult_lab)
  lab <- paste0("    tolN            = ", format(tolN))
  if (tolN == -0.001) {
    lab <- paste0(lab, " (default)")
  }
  quiet_msg_func(lab)
  quiet_msg_func(id_lab)
  quiet_msg_func(verbose_lab, "\n")
  lab <- paste0("    (*)Vmat_option  = ", format(Vmat_option))
  if (Vmat_option == 1) {
    lab <- paste0(lab, " (default)")
  }
  quiet_msg_func(lab)
  lab <- paste0("    (*)warnNegInput = ", format(warnNegInput))
  if (warnNegInput) {
    lab <- paste0(lab, " (default)")
  }
  quiet_msg_func(lab)
  quiet_msg_func(quiet_lab, "\n")
  quiet_msg_func("    (*) indicates new arguments in G-Series 3.0\n")
  quiet_msg_func(dup_alter_msg)
  
  verbose_func(start_time, Sys.time(), "Set up phase")
  start_time <- Sys.time()
  
  # Return the (empty) input series data frame
  if (P == 0) {
    out_df <- data_df 
    
  # Attempt to reconcile the input series data frame
  } else {

    # Build the elements of the raking problem (while validating the specified info), 
    # excluding the component series temporal totals info (will be added later):
    #   - x        : vector of component series initial values
    #   - c_x      : vector of component series alterability coefficients
    #   - comp_cols: vector of component series (column) names 
    #   - g        : vector of marginal total series initial values
    #   - c_g      : vector of marginal total series alterability coefficients
    #   - tot_cols : vector of marginal total series (column) names 
    #   - G        : marginal total series aggregation matrix (`g = G %*% x` for coherent/reconciled data)
    #
    # Notes: 
    #
    #   - The returned raking problem elements do not include the implicit component series temporal totals 
    #     when applicable (i.e., elements `g` and `G` only contain the marginal totals info).
    #
    #   - When the input data contains multiple periods (temporal total preservation scenario), raking problem 
    #     elements `x`, `c_x`, `g`, `c_g` and `G` are constructed in "column-major order", corresponding to the 
    #     default behaviour of R for converting matrices into vectors.
    #
    rk <- build_raking_problem(data_df          = data_df,
                               metadata_df      = metadata_df,
                               data_df_name     = data_df_name,
                               metadata_df_name = metadata_df_name,
                               alterability_df  = alterability_df,
                               alterSeries      = alterSeries, 
                               alterTotal1      = alterTotal1,
                               alterTotal2      = alterTotal2)
    
    n_comp <- length(rk$comp_cols)  # number of component series
    K <- length(rk$x)               # number of component series data points
    n_tot <- length(rk$tot_cols)    # number of marginal total series
    L <- length(rk$g)               # number of marginal total series data points
    
    
    # Validate the problem data
    if (any(!is.finite(rk$x)) || any(!is.finite(rk$g))) {
      stop("The input series data frame (argument 'data_df') contains invalid or NA values.\n\n",
           call. = FALSE)
    }
    if (gs.check_alter(rk$c_x) || gs.check_alter(rk$c_g)) {
      stop("The alterability data frame (argument 'alterability_df') contains invalid or NA values.\n\n",
           call. = FALSE)
    }
    if (warnNegInput) {
      if (gs.check_neg(rk$x, tol = 0) || gs.check_neg(rk$g, tol = 0)) {
        warning("The input series data frame (argument 'data_df') contains negative values. ",
                "Suspicious use of proportional raking.\n", call. = FALSE, immediate. = TRUE)
      }
    }
    
    
    # Process component series temporal totals info
    # (update aggregation matrix `G`, total data vector `g` and total alter coefs vector `c_g`
    # to include temporal totals)
    #
    # Note: count `L` is not updated here; it remains `n_tot * P`, i.e. the number of elements
    #       in vectors `g` and `c_g` associated to marginal (contemporaneous) totals only (excluding
    #       temporal totals). In other words, `L < length(g)` when `P > 1`
    if (P > 1) {
      Y <- matrix(0, nrow = n_comp, ncol = K)
      for (ii in 1:n_comp) {
        Y[ii, ((ii - 1) * P + 1):(ii * P)] <- 1
      }
      rk$G <- rbind(rk$G, Y)
      rk$g <- c(rk$g, Y %*% rk$x)
      
      if ("ALTERANNUAL" %in% metaDF_cols) {
        # Note: missing (NA) values are allowed for user-defined temporal total alter coefs (in the metadata file),
        #       but not for the component series and marginal (contemporaneous) totals (in the alterability file)
        #       in SAS G-Series PROC TSRAKING. I don't have a good  explanation for this (!)
        tmp <- metadata_df$ALTERANNUAL
        if (gs.check_alter(tmp[!is.na(tmp)])) {
          stop("Column 'alterAnnual' in the input metadata data frame (argument 'metadata_df') contains invalid values.\n\n",
               call. = FALSE)
        }
        tmp[is.na(tmp)] <- alterAnnual
        rk$c_g <- c(rk$c_g, tmp)
      } else {
        rk$c_g <- c(rk$c_g, rep(alterAnnual, n_comp))
      }
    }
    
    verbose_func(start_time, Sys.time(), "Metadata processing")
    start_time <- Sys.time()
    start_time1 <- start_time
    
    
    # Solve the raking problem (unless there's no discrepancies)
    discr_ini <- rk$g - rk$G %*% rk$x
    if (any(abs(discr_ini) > gs.min_tolerance)) {
      
      # Diagonal elements of the variance matrices (V_e and V_eps)
      if (Vmat_option == 1) {
        V_e <- rk$c_x * rk$x
        V_eps <- rk$c_g * rk$g
        mean_Ve <- mean(V_e)
        V_factor <- mean(abs(V_e))
      } else {
        V_e <- rk$c_x * abs(rk$x)
        V_eps <- rk$c_g * abs(rk$g)
        mean_Ve <- mean(V_e)
        V_factor <- mean_Ve
      }
      
      
      # Solve the raking problem
      if (mean_Ve != 0) {
        
        # Rescaled variance matrices (rescaling helps with the inverse calculation)
        # Note: `V_factor != 0` if `mean_Ve != 0`, regardless of argument `Vmat_option`
        V_e <- diag(V_e / V_factor, nrow = K)
        V_eps <- diag(V_eps / V_factor, nrow = length(rk$g))
        
        verbose_func(start_time, Sys.time(), "Variance matrices construction (V_e and V_eps)")
        start_time <- Sys.time()
        
        V_c <- V_e %*% t(rk$G)
        verbose_func(start_time, Sys.time(), "V_c = V_e * t(G)")
        start_time <- Sys.time()
        
        # Report totals for which all component series are binding (i.e. deterministic totals)
        # => may help users identify the cause of binding totals that can't be met (when applicable)
        all_binding_id <- which(colSums(V_c[, 1:L, drop = FALSE] == 0) == K)
        if (length(all_binding_id) > 0) {
          if (P == 1) {
            quiet_msg_func("All component series are binding (non alterable) for the following totals:",
                           paste0("\n  ", rk$tot_cols[all_binding_id], collapse = ""), "\n")
          } else {
            tot_id <- (all_binding_id - 1) %/% P + 1
            per_id <- (all_binding_id - 1) %% P + 1
            # `sapply()` is safe: it will always returns a character vector of length minimum 1
            quiet_msg_func("All component series are binding (non alterable) for the following totals (applicable periods in brackets):",
                           paste0("\n  ", 
                                  sapply(unique(tot_id),
                                         function(x) {
                                           paste0(rk$tot_cols[x],
                                                  "[",
                                                  paste0(per_id[which(tot_id == x)], collapse = ", "),
                                                  "]")  
                                         }), 
                                  collapse = ""),
                           "\n")
          }
        }
        verbose_func(start_time, Sys.time(), "Deterministic totals check")
        start_time <- Sys.time()
        
        V_d <- rk$G %*% V_c + V_eps
        verbose_func(start_time, Sys.time(), "V_d = G * V_c + V_eps")
        start_time <- Sys.time()
        
        V_d_inv <- gs.gInv_MP(V_d)
        verbose_func(start_time, Sys.time(), "Inverse V_d")
        start_time <- Sys.time()
        
        x_out <- rk$x + V_c %*% V_d_inv %*% discr_ini
        verbose_func(start_time, Sys.time(), "Solution = x + V_c * Inv(V_d) * (g - G * x)")
        verbose_func(start_time1, Sys.time(), "Total problem solving")
        start_time <- Sys.time()
        
        # Just for safety: `x_out` could contain `NaN` values in very extreme cases (matrix inversion problems)...
        if (any(!is.finite(x_out))) {
          warning("Unable to solve the raking problem. Returning the input component series data (and resulting totals).\n",
                  call. = FALSE, immediate. = TRUE)
          x_out <- rk$x
        }
        
        
      # Unsolvable (fixed) raking problem (sum of the elements of the component series variance matrix `V_e` is zero)
      } else {
        
        if (V_factor == 0) {
          if (P == 1) {
            quiet_msg_func("All component series are binding (non alterable). Nothing to solve. ",
                           "Returning the input component series data (and resulting totals).\n")
          } else {
            quiet_msg_func("All component series are binding (non alterable) for all periods. Nothing to solve. ",
                           "Returning the input component series data (and resulting totals).\n")
          }
          
          # May happen in presence of negative values in the input data with `Vmat_option = 1`
          # (impossible with `Vmat_option = 2`)
        } else {
          warning("Unsolvable raking problem. Returning the input component series data (and resulting totals).\n", call. = FALSE,
                  immediate. = TRUE)
        }
        x_out <- rk$x
        
        verbose_func(start_time, Sys.time(), "Problem feasibility checking")
        start_time <- Sys.time()
      }
      
      
      # (Re)calculate the totals
      g_out <- rk$G %*% x_out
      
      
    # Return the initial values (no initial discrepancies)
    } else {
      x_out <- rk$x
      g_out <- rk$g
      verbose_func(start_time, Sys.time(), "Solution = initial values (no discrepancy)")
      start_time <- Sys.time()
    }
    
    
    # Results validation
    if (neg_res_func(c(x_out, g_out), tol = -tolN)) {
      warning("The reconciled data contains negative values (threshold = ", format(tolN), ").\n", call. = FALSE,
              immediate. = TRUE)
    }
    binding_tot_valid_func(rk$g, g_out, rk$c_g, tol_parm, zero_tol = 0)
    
    verbose_func(start_time, Sys.time(), "Results validation")
    start_time <- Sys.time()
    
    
    # Create the output (series) data frame
    
    # Initialize with the input series data
    out_cols <- intersect(dataDF_cols, c(rk$comp_cols, rk$tot_cols, id))
    out_df <- data_df[out_cols]
    
    # Update the component series
    out_df[rk$comp_cols] <- as.data.frame(matrix(x_out, ncol = n_comp))
    
    # Update the marginal total(s)
    #   => exclude the temporal totals at the end of `g_out` (when applicable)
    out_df[rk$tot_cols] <- as.data.frame(matrix(g_out[1:L], ncol = n_tot))
    
    verbose_func(start_time, Sys.time(), "Wrap up phase")
  }
  
  verbose_func(start_time0, Sys.time(), "Total execution time")
  out_df
}


#' Build the elements of raking problems.
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/build_raking_problem.html}})}
#' 
#' This function is used internally by [tsraking()] to build the elements of the raking problem. It can also 
#' be useful to derive the cross-sectional (marginal) totals of the raking problem manually (outside of the [tsraking()] 
#' context). 
#' 
#' 
#' @inheritParams tsraking
#'
#' @param data_df_name (optional) 
#' 
#' String containing the value of argument `data_df`.
#'
#' **Default value** is `data_df_name = deparse1(substitute(data_df))`.
#' 
#' @param metadata_df_name (optional)
#' 
#' String containing the value of argument `metadata_df`.
#' 
#' **Default value** is `metadata_df_name = deparse1(substitute(metadata_df))`.
#'
#'
#' @returns
#' A list with the elements of the raking problem (excluding the implicit temporal totals):
#' - `x`        : vector of component series initial values
#' - `c_x`      : vector of component series alterability coefficients
#' - `comp_cols`: vector of component series (column) names 
#' - `g`        : vector of cross-sectional total initial values
#' - `c_g`      : vector of cross-sectional total alterability coefficients
#' - `tot_cols` : vector of cross-sectional total (column) names 
#' - `G`        : cross-sectional total aggregation matrix (`g = G %*% x`)
#'
#'
#' @details
#' See [tsraking()] for a detailed description of _time series raking_ problems.
#' 
#' The returned raking problem elements do not include the implicit component series temporal totals 
#' when applicable (i.e., elements `g` and `G` only contain the cross-sectional totals info).
#'       
#' When the input data contains multiple periods (temporal total preservation scenario), raking problem 
#' elements `x`, `c_x`, `g`, `c_g` and `G` are constructed _column by column_ (in "column-major order"), 
#' corresponding to the default behaviour of R for converting objects of class "matrix" into vectors.
#' 
#' Note: argument validation is not performed here; it is (bluntly) assumed that the function is called 
#' by [tsraking()] where a thorough validation of the arguments is done.
#'
#'
#' @seealso [tsraking()] [build_balancing_problem()]
#' 
#' @example misc/function_examples/build_raking_problem-ex.R
#' 
#' @export
build_raking_problem <- function(data_df,
                                 metadata_df,
                                 data_df_name = deparse1(substitute(data_df)),
                                 metadata_df_name = deparse1(substitute(metadata_df)),
                                 alterability_df = NULL,
                                 alterSeries = 1, 
                                 alterTotal1 = 0,
                                 alterTotal2 = 0) {
  
  
  # Alterability data frame creation function, taking into account both the specified
  # default (arguments `alter...`) and user-defined coefs (argument `alterability_df`)
  #
  # Main (parent) function objects used by the function (expected to exist):
  #  - P              : number of periods in the raking problem
  #  - alterability_df: user-defined alterability copefficients data frame
  #  - alterDF_cols   : list of column names of data frame `alterability_df` 
  #                     (`alterDF_cols == NULL` when `alterability_df == NULL`)
  create_alterDF <- function(cols, default_coef) {
    
    # Start with the default alter coefs
    default_cols <- setdiff(cols, alterDF_cols)
    alter_df <- as.data.frame(matrix(rep(default_coef, length(default_cols) * P),
                                     nrow = P,
                                     dimnames = list(NULL, default_cols)))
    
    # Add the user-defined alter coefs
    user_cols <- intersect(cols, alterDF_cols)
    if (length(user_cols) > 0) {
      alter_df <- cbind(alter_df, alterability_df[user_cols])
    }
    
    alter_df[cols]
  }
  
  
  ### Main function ###

  # Create the vectors of column names and other elementary objects
  dataDF_cols <- names(data_df)
  metaDF_cols <- toupper(names(metadata_df))
  names(metadata_df) <- metaDF_cols
  alterDF_cols <- names(alterability_df)
  P <- nrow(data_df)           # number of periods
  n_comp <- nrow(metadata_df)  # number of component series
  

  # Validate the metadata and build the raking problem matrices and vectors
  
  if (n_comp == 0) {
    stop("The input metadata data frame (argument 'metadata_df') must contain at least one observation (row).\n\n", 
         call. = FALSE)
  }
  gs.validate_cols(tolower(setdiff(c("SERIES", "TOTAL1"), metaDF_cols)), NULL, metadata_df_name)
  
  
  # Process component series info
  
  comp_cols <- gs.cleanup_col_list(metadata_df$SERIES)
  if (n_comp != length(comp_cols)) {
    stop("Column 'series' in the input metadata data frame (argument 'metadata_df') contains missing (NA) or blank (\"\") values.\n\n",
         call. = FALSE)
  }
  if (n_comp > length(unique(comp_cols))) {
    stop("Column 'series' in the input metadata data frame (argument 'metadata_df') contains duplicate values.\n\n",
         call. = FALSE)
  }
  gs.validate_cols(comp_cols, dataDF_cols, data_df_name, source_str = "metadata column 'series'")
  
  # Component series data vector `x` (column-major version of the input series data frame)
  x <- unlist(data_df[comp_cols], use.names = FALSE)
  K <- length(x)  # number of component series data points
  
  # Component series alter coef vector c_x (column-major version of the alter coef data frame data)
  c_x <- unlist(create_alterDF(comp_cols, alterSeries), use.names = FALSE)
  
  
  # Process marginal (contemporaneous) totals info
  
  temp <- gs.cleanup_col_list(metadata_df$TOTAL1)
  if (n_comp != length(temp)) {
    stop("Column 'total1' in the input metadata data frame (argument 'metadata_df') contains missing (NA) ", 
         "or blank (\"\") values.\n\n", call. = FALSE)
  }
  tot1_cols <- unique(temp)
  gs.validate_cols(tot1_cols, dataDF_cols, data_df_name, source_str = "metadata column 'total1'")
  N <- length(tot1_cols)
  tot_cols <- tot1_cols

  if ("TOTAL2" %in% metaDF_cols) {
    temp <- gs.cleanup_col_list(metadata_df$TOTAL2)
    if (length(temp) > 0) {
      if (n_comp != length(temp)) {
        stop("Column 'total2' in the input metadata data frame (argument 'metadata_df') contans missing (NA) ", 
             "or blank (\"\") values.\n\n", call. = FALSE)
      }
      tot2_cols <- unique(temp)
      M <- length(tot2_cols)
      if (M == 1) {
        stop("Column 'total2' in the input metadata data frame (argument 'metadata_df') defines a single dim. 2 total. ",
             "It should either be removed from the data frame for a 1-dimensional raking problem or be used to define ",
             "multiple dim. 2 totals for a 2-dimensional problem.\n\n", call. = FALSE)
      }
      gs.validate_cols(tot2_cols, dataDF_cols, data_df_name, source_str = "metadata column 'total2'")
      tot_cols <- c(tot_cols, tot2_cols)
    } else {
      tot2_cols <- NULL
      M <- 1
    }
  } else {
    tot2_cols <- NULL
    M <- 1
  }
  
  
  # Initialize aggregation matrix `G` and data vector `g` for the marginal totals
  
  # 1-dim raking (single marginal total): order of the component series doesn't matter
  if (M == 1) {
    if (length(tot_cols) != 1) {
      stop("Column 'total1' in the input metadata data frame (argument 'metadata_df') defines multiple totals ",
           "while a single total is expected in a 1-dimensional raking problem.\n\n", call. = FALSE)
    }
    n_tot <- 1
    L <- P
    
    # For each component series corresponds P rows and columns in matrix `G`
    G <- matrix(rep(diag(1, nrow = P), n_comp), nrow = P)
    
    # Totals alter coef vector c_g (column-major version of the alter coef data frame data)
    c_g <- unlist(create_alterDF(tot_cols, alterTotal1), use.names = FALSE)
    
  # 2-dim raking (multiple marginal totals): order of the component series is important (!)
  } else {
    if (n_comp < M * N) {
      stop("Incomplete specification of a 2-dimensional raking problem in the input metadata data frame ",
           "(argument 'metadata_df'). Some dim. 1 and dim. 2 total combinations (columns 'total1' and 'total2') ",
           "are missing from the data frame (have not been assigned a component series).\n\n", call. = FALSE)
    }
    if (any(duplicated(metadata_df[c("TOTAL1", "TOTAL2")], MARGIN = 1))) {
      stop("Invalid specification of a 2-dimensional raking problem in the input metadata data frame ",
           "(argument 'metadata_df'). Some dim. 1 and dim. 2 total combinations (columns 'total1' and 'total2') ",
           "appear more than once in the data frame (have been assigned more than one component series).\n\n",
           call. = FALSE)
    }
    n_tot <- M + N
    L <- n_tot * P
    
    # For each component series corresponds P rows and columns in matrix `G`
    # Initialize `G` with 0's
    G <- matrix(0, nrow = L, ncol = K)
    
    # 1st dim. totals aggregation vectors (update `G` with 1's in the right place)
    for (ii in 1:N) {
      # kk, the row number in the metadata data frame, represents column number(s) in matrix `G`.
      kk <- which(metadata_df$TOTAL1 == tot_cols[ii])
      for (jj in 1:P) {
        G[(ii - 1) * P + jj, (kk - 1) * P + jj] <- 1
      }
    }
    
    # 2nd dim. totals aggregation vectors (update `G` with 1's in the right place)
    for (ii in 1:M + N) {
      # same as for TOTAL1
      kk <- which(metadata_df$TOTAL2 == tot_cols[ii])
      for (jj in 1:P) {
        G[(ii - 1) * P + jj, (kk - 1) * P + jj] <- 1
      }
    }
    
    # Totals alter coef vector c_g (column-major version of the alter coef data frame data)
    c_g <- c(unlist(create_alterDF(tot1_cols, alterTotal1), use.names = FALSE),
             unlist(create_alterDF(tot2_cols, alterTotal2), use.names = FALSE))
  }
  
  # Totals data vector `g` (column-major version of the input series data frame data)
  g <- unlist(data_df[tot_cols], use.names = FALSE)
  
  
  # Return the raking problem elements
  list(x = x,
       c_x = c_x,
       comp_cols = comp_cols,
       g = g,
       c_g = c_g,
       tot_cols = tot_cols,
       G = G)
}

#' Generate benchmarking graphics in a PDF file
#'
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/plot_graphTable.html}})}
#' 
#' Create a PDF file (US Letter paper size format in landscape view) containing benchmarking graphics for the set of series 
#' contained in the specified benchmarking function ([benchmarking()] or [stock_benchmarking()]) output `graphTable` 
#' data frame. Four types of benchmarking graphics can be generated for each series:
#' - **Original Scale Plot** (argument `ori_plot_flag`) - overlay graph of:
#'   - Indicator series
#'   - Average indicator series
#'   - Bias corrected indicator series (when \eqn{\rho < 1})
#'   - Benchmarked series
#'   - Average benchmark
#' - **Adjustment Scale Plot** (argument `adj_plot_flag`) - overlay graph of:
#'   - Benchmarking adjustments
#'   - Average benchmarking adjustments
#'   - Bias line (when \eqn{\rho < 1})
#' - **Growth Rates Plot** (argument `GR_plot_flag`) - bar chart of the indicator and benchmarked series growth rates.
#' - **Growth Rates Table** (argument `GR_table_flag`) - table of the indicator and benchmarked series growth rates.
#'
#' These graphics can be useful to assess the quality of the benchmarking results. Any of the four types of
#' benchmarking graphics can be enabled or disabled with the corresponding flag. The first three types of graphics 
#' (the plots) are generated by default while the fourth (growth rates table) is not.
#'
#'
#' @param graphTable (mandatory)
#'
#' Data frame (object of class "data.frame") corresponding to the benchmarking function output`graphTable` data frame.
#'
#' @param pdf_file (mandatory)
#'
#' Name (and path) of the PDF file that will contain the benchmarking graphics. The name should include the ".pdf" 
#' file extension. The PDF file is created in the R session working directory (as returned by `getwd()`) if a path 
#' is not specified. Specifying `NULL` would cancel the creation of a PDF file.
#'
#' @param ori_plot_flag,adj_plot_flag,GR_plot_flag,GR_table_flag (optional)
#'
#' Logical arguments indicating whether or not the corresponding type of benchmarking graphic should be generated. 
#' All three plots are generated by default but not the growth rates tables.
#'
#' **Default values** are `ori_plot_flag = TRUE`, `adj_plot_flag = TRUE`, `GR_plot_flag = TRUE` and 
#' `GR_table_flag = FALSE`.
#'
#' @param add_bookmarks
#'
#' Logical argument indicating whether or not bookmarks should be added to the PDF file. See **Bookmarks** in section 
#' **Details** for more information.
#'
#' **Default value** is `add_bookmarks = TRUE`.
#'
#'
#' @details
#' List of the `graphTable` data frame variables corresponding to each element of the four types of benchmarking 
#' graphics:
#' - Original Scale Plot (argument `ori_plot_flag`)
#'   - `subAnnual` for the *Indicator Series* line
#'   - `avgSubAnnual` for the *Avg. Indicator Series* segments
#'   - `subAnnualCorrected` for the *Bias Corr. Indicator Series* line (when \eqn{\rho < 1})
#'   - `benchmarked` for the *Benchmarked Series* line
#'   - `avgBenchmark` for the *Average Benchmark* segments
#' - Adjustment Scale Plot (argument `adj_plot_flag`)
#'   - `benchmarkedSubAnnualRatio` for the *BI Ratios (Benchmarked Series / Indicator Series)* line \eqn{^{(*)}}{(*)}
#'   - `avgBenchmarkSubAnnualRatio` for the *Average BI Ratios* segments \eqn{^{(*)}}{(*)}
#'   - `bias` for the *Bias* line (when \eqn{\rho < 1})
#' - Growth Rates Plot (argument `GR_plot_flag`)
#'   - `growthRateSubAnnual` for the *Growth R. in Indicator Series* bars \eqn{^{(*)}}{(*)}
#'   - `growthRateBenchmarked` for the *Growth R. in Benchmarked Series* bars \eqn{^{(*)}}{(*)}
#' - Growth Rates Table (argument `GR_table_flag`)
#'   - `year` for the *Year* column
#'   - `period` for the *Period* column
#'   - `subAnnual` for the *Indicator Series* column
#'   - `benchmarked` for the *Benchmarked Series* column
#'   - `growthRateSubAnnual` for the *Growth Rate in Indicator Series* column \eqn{^{(*)}}{(*)}
#'   - `growthRateBenchmarked` for the *Growth Rate in Benchmarked Series* column \eqn{^{(*)}}{(*)}
#' 
#' \eqn{^{(*)}}{(*)} _BI ratios_ and _growth rates_ actually correspond to _differences_ when \eqn{\lambda = 0} (additive
#' benchmarking).
#' 
#' The function uses the extra columns of the `graphTable` data frame (columns not listed in the **Value** section of 
#' [benchmarking()] and [stock_benchmarking()]), if any, to build BY-groups. See section **Benchmarking Multiple Series**  
#' of [benchmarking()] for more details.
#' 
#' ## Performance
#' The two types of growth rates graphics, i.e., the bar chart (`GR_plot_flag`) and table (`GR_table_flag`), often requires 
#' the generation of several pages in the PDF file, especially for long monthly series with several years of data. This creation of 
#' extra pages slows down the execution of [plot_graphTable()]. This is why only the bar chart is generated by default 
#' (`GR_plot_flag = TRUE` and `GR_table_flag = FALSE`). Deactivating both types of growth rates graphics (`GR_plot_flag = FALSE` 
#' and `GR_table_flag = FALSE`) or reducing the size of the input `graphTable` data frame for very long series (e.g., keeping 
#' only recent years) could therefore improve execution time. Also note that the impact of benchmarking on the growth rates can be 
#' deduced from the adjustment scale plot (`adj_plot_flag`) by examining the extent of vertical movement (downward or upward) of the 
#' benchmarking adjustments between adjacent periods: the greater the vertical movement, the greater the impact on corresponding growth 
#' rate. Execution time of [plot_graphTable()] could therefore be reduced, if needed, by only generating the first two types of graphics 
#' while focusing on the adjustment scale plot to assess period-to-period movement preservation, i.e., the impact of benchmarking on 
#' the initial growth rates.
#' 
#' ## ggplot2 themes
#' The plots are generated with the ggplot2 package which comes with a convenient set of [complete 
#' themes](https://ggplot2.tidyverse.org/reference/ggtheme.html) for the general look and feel of the plots (with `theme_grey()` 
#' as the default theme). Use function `theme_set()` to change the theme applied to the plots generated by [plot_graphTable()] 
#' (see the **Examples**).
#' 
#' ## Bookmarks
#' Bookmarks are added to the PDF file with `xmpdf::set_bookmarks()` when argument `add_bookmarks = TRUE` (default), which 
#' requires a command-line tool such as [Ghostscript](https://www.ghostscript.com/)  or 
#' [PDFtk](https://www.pdflabs.com/tools/pdftk-the-pdf-toolkit/). See section **Installation** in 
#' `vignette("xmpdf", package = "xmpdf")` for details.
#' 
#' **Important**: bookmarks will be successfully added to the PDF file **if and only if** \ifelse{latex}{\code{xmpdf::supports 
#' _set_bookmarks()}}{\code{xmpdf::supports_set_bookmarks()}} returns `TRUE` **and** the execution of `xmpdf::set_bookmarks()` is 
#' successful. If Ghostscript is installed on your machine but `xmpdf::supports_set_bookmarks()` still returns `FALSE`, try 
#' specifying the path of the Ghostscript executable in environment variable `R_GSCMD` (e.g., 
#' `Sys.setenv(R_GSCMD = "C:/Program Files/.../bin/gswin64c.exe")` on Windows). On the other hand, if 
#' `xmpdf::supports_set_bookmarks()}` returns `TRUE` but you are experiencing (irresolvable) issues with `xmpdf::set_bookmarks()` 
#' (e.g., error related to the Ghostscript executable), bookmarks creation can be disabled by specifying `add_bookmarks = FALSE`.
#'
#'
#' @returns
#' In addition to creating a PDF file containing the benchmarking graphics (except when `pdf_file = NULL`), this function 
#' also invisibly returns a list with the following elements: 
#' - `graphTable`: Character string (character vector of length one) that contains the complete name and path of the PDF 
#' file if it was successfully created and `invisible(NA_character_)` otherwise or if `pdf_file = NULL` was specified.
#' - `graph_list`: List of the generated benchmarking graphics (one per series) with the following elements:
#'   - `name`: Character string describing the series (matches the bookmark name in the PDF file).
#'   - `page`: Integer representing the sequence number of the first graphic for the series in the entire sequence of 
#'   graphics for all series (matches the page number in the PDF file).
#'   - `ggplot_list`: List of ggplot objects (one per graphic or page in the PDF file) corresponding to the generated 
#'   benchmarking graphics for the series. See section **Value** in [bench_graphs] for details.
#' 
#' Note that the returned ggplot objects can be displayed _manually_ with [print()], in which case some updates to the 
#' ggplot2 theme defaults are recommended in order to produce graphics with a similar look and feel as those generated in 
#' the PDF file (see section **Value** in [bench_graphs] for details). Also keep in mind that these graphics are optimized 
#' for the US Letter paper size format in landscape view (as displayed in the PDF file), i.e., 11in wide (27.9cm, 1056px 
#' with 96 DPI) and 8.5in tall (21.6cm, 816px with 96 DPI).
#'
#'
#' @seealso [bench_graphs] [plot_benchAdj()] [benchmarking()] [stock_benchmarking()]
#'
#'
#' @example misc/function_examples/plot_graphTable-ex.R
#' 
#' 
#' @export
plot_graphTable <- function(graphTable,
                            pdf_file,
                            ori_plot_flag = TRUE,
                            adj_plot_flag = TRUE,
                            GR_plot_flag = TRUE,
                            GR_table_flag = FALSE,
                            add_bookmarks = TRUE) {
  
  
  # Initialize the object to be returned by the function via `on.exit()`
  out_list <- list(pdf_name = NA_character_,
                   graph_list = NULL)
  on.exit(return(invisible(out_list)))
  
  # Turn off debugging/traceback generated by calls to the stop() function inside internal functions
  ini_error_opt <- getOption("error")
  on.exit(options(error = ini_error_opt), add = TRUE)
  options(error = NULL)
  
  # Display warnings immediately (equivalent of `warning(..., immediate. = TRUE)` for warning messages
  # not written/generated by you directly)
  ini_warn_opt <- getOption("warn")
  on.exit(options(warn = ini_warn_opt), add = TRUE)
  options(warn = 1)
  
  
  # Validation
  df_name <- deparse1(substitute(graphTable))
  tmp <- nchar(df_name)
  if (tmp == 0) {
    stop("Argument 'graphTable' is mandatory (it must be specified).\n\n", call. = FALSE)
  } else if (tmp >= 60) {
    df_name <- paste0(substr(df_name, 1, 55), "<...>")
  }
  if (grepl("structure(", df_name, fixed = TRUE)) {
    df_name <- "<argument 'graphTable'>"
  }
  graphTable <- graphTable
  if (!is.data.frame(graphTable)) {
    stop("Argument 'graphTable' is not a 'data.frame' object.\n\n", call. = FALSE)
  }
  graphTable <- as.data.frame(graphTable)
  ori_plot_flag <- gs.validate_arg_logi(ori_plot_flag)
  adj_plot_flag <- gs.validate_arg_logi(adj_plot_flag)
  GR_plot_flag <- gs.validate_arg_logi(GR_plot_flag)
  GR_table_flag <- gs.validate_arg_logi(GR_table_flag)
  add_bookmarks <- gs.validate_arg_logi(add_bookmarks)
  
  
  # Set mandatory argument `pdf_file` to NULL if it's not specified
  # (do not create a PDF file)
  if (nchar(deparse1(substitute(pdf_file))) == 0) {
    pdf_file <- NULL
  }
  
  
  
  # Validate the graph table contents
  df_cols <- names(graphTable)
  mandatory_cols <- c("varSeries",
                      "m",
                      "year",
                      "period",
                      "constant",
                      "rho",
                      "lambda",
                      "bias",
                      "periodicity",
                      "subAnnual",
                      "benchmarked",
                      "avgBenchmark",
                      "avgSubAnnual",
                      "subAnnualCorrected",
                      "benchmarkedSubAnnualRatio",
                      "avgBenchmarkSubAnnualRatio",
                      "growthRateSubAnnual",
                      "growthRateBenchmarked")
  missing_cols <- setdiff(mandatory_cols, df_cols)
  if (length(missing_cols) > 0) {
    stop("The following columns are missing from the 'graphTable' data frame (", df_name, "): ",
         paste0("\n  ", missing_cols, collapse = ""), "\n\n", call. = FALSE)
  }
  
  # Get the by-group columns ("extras" columns)
  by_cols <- setdiff(df_cols, 
                     c("varSeries",
                       "varBenchmarks",
                       "altSeries",
                       "altSeriesValue",
                       "altbenchmarks",
                       "altBenchmarksValue",
                       "t",
                       "m",
                       "year",
                       "period",
                       "constant",
                       "rho",
                       "lambda",
                       "bias",
                       "periodicity",
                       "date",
                       "subAnnual",
                       "benchmarked",
                       "avgBenchmark",
                       "avgSubAnnual",
                       "subAnnualCorrected",
                       "benchmarkedSubAnnualRatio",
                       "avgBenchmarkSubAnnualRatio",
                       "growthRateSubAnnual",
                       "growthRateBenchmarked"))
  
  # Determine the set of series to plot (combination of by-group column values plus column "varSeries" values)
  ser_list_cols <- c(by_cols, "varSeries")
  ser_list_df <- unique(graphTable[ser_list_cols])
  n_ser <- nrow(ser_list_df)
  
  
  if (n_ser == 0) {
    warning("Nothing to plot!\n", call. = FALSE, immediate. = TRUE)
  } else {
    
    if (is.null(pdf_file)) {
      display_flag <- FALSE
    } else {
      
      # Create the temporary (local) PDF file (landscape letter size format with 1 inch borders)
      tmp_pdf <- tempfile(fileext = ".pdf")
      grDevices::pdf(tmp_pdf, onefile = TRUE, width = 11, height = 8.5, encoding = "CP1250")
      # NOTE: encoding CP1250 is used to prevent a warning with the default encoding on Unix-like environments  
      #       (e.g., StatCan GitLab pipeline's Linux Unbuntu) regarding the replacement of characters "..." 
      #       (e.g., long `graphTable` data frame name) with the 'Horizontal Ellipsis' unicode character ("\u2026").
      if (grDevices::dev.cur() == 1) {
        warning("PDF file not created (`grDevices::pdf()` failed)!\n", call. = FALSE, immediate. = TRUE)
        display_flag <- FALSE
      } else {
        display_flag <- TRUE
      }
    }
    
    
    # Display a "starting" message
    message("\nGenerating the benchmarking graphics. Please be patient...\n")
    
    # ggplot2 initialization
    pt_sz <- 2
    ini_theme <- ggplot2::theme_get()
    intialize_theme()
    on.exit(ggplot2::theme_set(ini_theme), add = TRUE)
    
    # Initialize the PDF bookmarks data frame
    bookmarks_df <- data.frame(title = rep.int("", n_ser),
                               page = rep.int(NA_integer_, n_ser))
    
    # Assign the plotting function names according to each individual flag
    # (single "if" instead of repeated "ifs" for each series to plot)
    if (ori_plot_flag) {
      ori_plot_func <- ori_plot
    } else {
      ori_plot_func <- gs.NULL_func
    }
    if (adj_plot_flag) {
      adj_plot_func <- adj_plot
    } else {
      adj_plot_func <- gs.NULL_func
    }
    if (GR_plot_flag) {
      GR_plot_func <- GR_plot
    } else {
      GR_plot_func <- gs.NULL_func
    }
    if (GR_table_flag) {
      GR_table_func <- GR_table
    } else {
      GR_table_func <- gs.NULL_func
    }
    
    
    # Build the plot subtitles and PDF bookmark labels
    plot_subtitle <- rep.int(paste0("Graphics Table <strong>", mask_underscore(df_name), "</strong>"), n_ser)
    
    # Remove unnecessary columns (those with a single value for all rows of the graphTable data frame, if any)
    ser_list_cols <- ser_list_cols[lapply(lapply(ser_list_df, unique), length) > 1]
    if (length(ser_list_cols) == 0) {
      bookmarks_df$title <- rep.int("<Single Series>", n_ser)
      by_cols <- character(0L)
    } else {
      
      # "varSeries" column info
      if ("varSeries" %in% ser_list_cols) {
        bookmarks_df$title <- ser_list_df[["varSeries"]]
        plot_subtitle <- paste0(plot_subtitle, " - Variable <strong>", mask_underscore(bookmarks_df$title), "</strong>")
        label_sep <- " - "
      } else {
        label_sep <- ""
      }
      
      # By-group columns info
      by_cols <- setdiff(ser_list_cols, "varSeries")
      if (length(by_cols) > 0) {
        if (length(by_cols) == 1) {
          byGrp_label <- as.character(ser_list_df[[by_cols]])
        } else {
          # `sapply()` is safe: it always return a character vector (`by_cols` is a vector of minimum length 2)
          byGrp_label <- apply(sapply(by_cols, function(x) paste0(x, "=", ser_list_df[[x]])),
                               1, paste0, collapse = " & ")
        }
        bookmarks_df$title <- paste0(bookmarks_df$title, label_sep, byGrp_label)
        plot_subtitle <- paste0(plot_subtitle, " - BY-Group <strong>", mask_underscore(byGrp_label), "</strong>")
      }
    }
    
    
    # Reduce the data frame to essential columns 
    graphTable <- graphTable[c(by_cols, mandatory_cols)]
    
    
    # Process each series
    
    pdf_page <- 1
    for (ii in 1:n_ser) {
      
      # Set the PDF bookmark page
      bookmarks_df$page[ii] <- pdf_page
      
      # Initialize `graph_list` objects 
      out_list$graph_list <- c(out_list$graph_list, 
                               list(list(name = bookmarks_df$title[ii],
                                         page = bookmarks_df$page[ii],
                                         ggplot_list = NULL)))
      
      # Extract the relevant graphTable rows
      plot_df <- merge(ser_list_df[ii, ser_list_cols, drop = FALSE], 
                       graphTable, 
                       by = ser_list_cols)
      plot_df <- plot_df[order(plot_df$year, plot_df$period), ]
      
      # Extract the benchmarking parms
      parms <- extract_parms(plot_df)
      
      
      # Generate the plots
      
      subtitle_str <- paste0(plot_subtitle[ii], "<br>(", parms$parms_str, ")")
      
      tmp_list <- ori_plot_func(graphTable = plot_df,
                                subtitle_str = subtitle_str,
                                mth_gap = parms$mth_gap,
                                points_set = parms$ori_plot_set,
                                pt_sz = pt_sz,
                                display_ggplot = display_flag,
                                .setup = FALSE)
      out_list$graph_list[[ii]]$ggplot_list <- c(out_list$graph_list[[ii]]$ggplot_list, tmp_list)
      pdf_page <- pdf_page + length(tmp_list)
      
      tmp_list <- adj_plot_func(graphTable = plot_df,
                                subtitle_str = subtitle_str,
                                mth_gap = parms$mth_gap,
                                full_set = parms$adj_plot_set,
                                pt_sz = pt_sz,
                                display_ggplot = display_flag,
                                .setup = FALSE)
      out_list$graph_list[[ii]]$ggplot_list <- c(out_list$graph_list[[ii]]$ggplot_list, tmp_list)
      pdf_page <- pdf_page + length(tmp_list)
      
      tmp_list <- GR_plot_func(graphTable = plot_df,
                               subtitle_str = subtitle_str,
                               factor = parms$GR_factor,
                               type_chars = parms$GR_type_chars,
                               periodicity = parms$periodicity,
                               display_ggplot = display_flag,
                               .setup = FALSE)
      out_list$graph_list[[ii]]$ggplot_list <- c(out_list$graph_list[[ii]]$ggplot_list, tmp_list)
      pdf_page <- pdf_page + length(tmp_list)
      
      tmp_list <- GR_table_func(graphTable = plot_df,
                                subtitle_str = subtitle_str,
                                factor = parms$GR_factor,
                                type_chars = parms$GR_type_chars,
                                display_ggplot = display_flag,
                                .setup = FALSE)
      out_list$graph_list[[ii]]$ggplot_list <- c(out_list$graph_list[[ii]]$ggplot_list, tmp_list)
      pdf_page <- pdf_page + length(tmp_list)
      
    }
    
    
    if (display_flag) {
      
      # Close the temporary PDF file and add bookmarks
      grDevices::dev.off()
      if (add_bookmarks) {
        if (xmpdf::supports_set_bookmarks()) {
          tmp_pdf2 <- try(xmpdf::set_bookmarks(bookmarks_df, tmp_pdf, tempfile(fileext = ".pdf")))
          if (exists("tmp_pdf2") && file.exists(tmp_pdf2)) {
            unlink(tmp_pdf)
            tmp_pdf <- tmp_pdf2
            bookmark_msg_flag <- FALSE
          } else {
            bookmark_msg_flag <- TRUE
          }
        } else {
          bookmark_msg_flag <- TRUE
        }
      } else {
        bookmark_msg_flag <- FALSE
      }
      
      # Create the final (remote) PDF file
      out_pdf <- as.character((unlist(pdf_file))[1])
      if (length(out_pdf) > 0 && !is.na(out_pdf)) {
        len <- nchar(out_pdf)
        if (toupper(substr(out_pdf, len - 3, len)) != ".PDF") {
          out_pdf <- paste0(out_pdf, ".pdf")
        }
        # NOTE: moving a file with `file.copy()` + `unlink()` is safer than with `file.rename()` 
        #       (see section "Warning" in `help(file.rename)`)
        if (file.copy(from = tmp_pdf, to = out_pdf, overwrite = TRUE)) {
          unlink(tmp_pdf)
        } else {
          # `file.copy()` failed, most likely due to a "file locking issue"
          message("")
          out_pdf <- tmp_pdf
        }
      } else {
        warning("Invalid PDF file specification (argument 'pdf_file'). A temporary PDF file will be used instead.\n", 
                call. = FALSE, immediate. = TRUE)
        out_pdf <- tmp_pdf
      }
      
      # Completion message
      out_list$pdf_name <- utils::fileSnapshot(out_pdf)$path
      message("Benchmarking graphics generated for ", format(n_ser), " series in the following PDF file:\n  ",
              normalizePath(out_list$pdf_name), "\n")
      if (bookmark_msg_flag) {
        message("Bookmarks could not be added to the PDF file (see section \"Bookmarks\" in `help(plot_graphTable)`).\n")
      }
      
    } else {
      message("Benchmarking graphics generated for ", format(n_ser), " series.\n")
    }
  }
  
  # Output object returned via function `on.exit()`
}




#' Generate a benchmarking graphic
#' 
#' @name bench_graphs
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/bench_graphs.html}})}
#' 
#' Functions used internally by [plot_graphTable()] to generate the benchmarking graphics in a PDF file:
#' - [ori_plot()]: Original Scale Plot ([plot_graphTable()] argument `ori_plot_flag = TRUE`)
#' - [adj_plot()]: Adjustment Scale Plot ([plot_graphTable()] argument `adj_plot_flag = TRUE`)
#' - [GR_plot()]: Growth Rates Plot ([plot_graphTable()] argument `GR_plot_flag = TRUE`)
#' - [GR_table()]: Growth Rates Table ([plot_graphTable()] argument `GR_table_flag = TRUE`)
#' 
#' When these functions are called directly, the `graphTable` data frame should only  contain a **single series** 
#' and the graphic is generated in the current (active) graphics device.
#' 
#' 
#' @inheritParams plot_graphTable
#' 
#' @param title_str,subtitle_str (optional)
#' 
#' Graphic title and subtitle strings (character constants). `subtitle_str` is automatically built from the 
#' `graphTable` data frame contents when `NULL` and contains the `graphTable` data frame name on the 2<sup>nd</sup> line 
#' and the benchmarking parameters on the 3<sup>rd</sup> line. Specifying empty strings (`""`) would remove the titles. 
#' Simple Markdown and HTML syntax is allowed (e.g., for bold, italic or colored fonts) through package [ggtext][ggtext::ggtext] 
#' (see `help(package = "ggtext")`).
#' 
#' **Default values** are `subtitle_str = NULL` and a function specific string for `title_str` (see **Usage**).
#' 
#' @param mth_gap (optional)
#' 
#' Number of months between consecutive periods (e.g. 1 for monthly data, 3 for quarterly data, etc.). Based  
#' on the `graphTable` data frame contents  when `NULL` (calculated as `12 / graphTable$periodicity[1]`).
#' 
#' **Default value** is `mth_gap = NULL`.
#' 
#' @param points_set,full_set (optional)
#' 
#' Character vector of the elements (variables of the `graphTable` data frame) to include in the plot.
#' Automatically built when `NULL`. See [plot_graphTable()] for the (default) list of variables used for each 
#' type of graphic.
#' 
#' **Default values** are `points_set = NULL` and `full_set = NULL`.
#' 
#' @param pt_sz (optional)
#' 
#' Size of the data points shape (symbol) for ggplot2.
#' 
#' **Default value** is `pt_sz = 2`.
#' 
#' @param factor,type_chars (optional)
#' 
#' Growth rates factor (1 or 100) and value label suffix ("" or "(%)") according to the adjustment model parameter 
#' \eqn{\lambda}. Based on the `graphTable` data frame contents when `NULL` (based on `graphTable$lambda[1]`).
#' 
#' **Default values** are `factor = NULL` and `type_chars = NULL`.
#' 
#' @param periodicity (optional)
#' 
#' Number of periods in a year. Based on the `graphTable` data frame contents when `NULL` (defined as 
#' `graphTable$periodicity[1]`).
#' 
#' **Default value** is `periodicity = NULL`.
#'
#' @param display_ggplot (optional)
#' 
#' Logical arguments indicating whether or not the ggplot object(s) should be displayed in the current (active) 
#' graphics device.
#'
#' **Default value** is `display_ggplot = TRUE`.
#'
#' @param .setup (optional)
#' 
#' Logical argument indicating whether the setup steps must be executed or not. Must be `TRUE` when the function 
#' is called directly (i.e., outside of the [plot_graphTable()] context). 
#' 
#' **Default value** is `.setup = TRUE`.
#'
#'
#' @details
#' See [plot_graphTable()] for a detailed description of the four benchmarking graphics associated to these 
#' individual functions. These graphics are optimized for the US Letter paper size format in landscape view, i.e., 
#' 11in wide (27.9cm, 1056px with 96 DPI) and 8.5in tall (21.6cm, 816px with 96 DPI). Keep this in mind when 
#' viewing or saving graphics generated by calls to these individual functions (i.e., outside of the [plot_graphTable()] 
#' context). Also note that [GR_plot()] and [GR_table()] will often generate more than one graphic (more than one 
#' *page*), unless the number of periods included in the input `graphTable` data frame is reduced (e.g., subsetting 
#' the data frame by ranges of calendar years). 
#'
#'
#' @returns
#' In addition to displaying the corresponding graphic(s) in the current (active) graphics device (except when 
#' `display_ggplot = FALSE`), each function also invisibly returns a list containing the generated ggplot object(s). 
#' Notes:
#' - [ori_plot()] and [adj_plot()] generate a single ggplot object (single graphic) while [GR_plot()] and [GR_table()] 
#' will often generate several ggplot objects (several graphics).
#' - The returned ggplot object(s) can be displayed _manually_ with [print()], in which case the following ggplot2 theme 
#' updates (used internally when `display_ggplot = TRUE`) are suggested:
#'   ```R
#'   ggplot2::theme_update(
#'     plot.title = ggtext::element_markdown(hjust = 0.5),
#'     plot.subtitle = ggtext::element_markdown(hjust = 0.5),
#'     legend.position = "bottom",
#'     plot.margin = ggplot2::margin(t = 1.5, r = 1.5, b = 1.5, l = 1.5, unit = "cm")
#'   )
#'   ```
#'
#'
#' @seealso [plot_graphTable()] [plot_benchAdj()] [benchmarking()] [stock_benchmarking()]
#'
#'
#' @example misc/function_examples/bench_graphs-ex.R
#' 
#' 
#' @importFrom rlang .data


#' @rdname bench_graphs
#' @export
ori_plot <- function(graphTable, 
                     title_str = "Original Scale",
                     subtitle_str = NULL, 
                     mth_gap = NULL, 
                     points_set = NULL,
                     pt_sz = 2,
                     display_ggplot = TRUE,
                     .setup = TRUE) {
  
  
  # Original Scale graph function (called when `ori_plot_flag = TRUE`)
  # Approach:
  #   - create 3 tall (stacked) data frames for plotting with ggplot:
  #       - data points (geom_point) for series values and repeated average values
  #       - lines (geom_line) for series lines only (not average lines)
  #       - segments (geom_segment) for average lines
  #   - generate "by name (series)" geoms for each type of element to plot
  #     (geom_line, geom_segment and geom_point)
  
  
  # Initialize the object to be returned by the function via `on.exit()`
  ggplot_list <- NULL
  on.exit(return(invisible(ggplot_list)))
  
  # Enforce the default R "error" option (`options(error = NULL)`). E.g. this Turns off traceback
  # generated by calls to the stop() function inside internal functions in R Studio.
  ini_error_opt <- getOption("error")
  on.exit(options(error = ini_error_opt), add = TRUE)
  options(error = NULL)
  
  # Set the display function
  if (gs.validate_arg_logi(display_ggplot)) {
    display_func <- print
  } else {
    display_func <- gs.NULL_func
  }
  
  # Setup part (manual call outside on `plot_graphTable()`)
  if (gs.validate_arg_logi(.setup)) {

    # Set PDF encoding to CP1250 to avoid warnings on Unix-like environments 
    # (see `plot_graphTable()` comment above)
    ini_pdf_encoding <- grDevices::pdf.options()$encoding
    on.exit(grDevices::pdf.options(encoding = ini_pdf_encoding), add = TRUE)
    grDevices::pdf.options(encoding = "CP1250")
    
    # ggplot2 initialization
    ini_theme <- ggplot2::theme_get()
    intialize_theme()
    on.exit(ggplot2::theme_set(ini_theme), add = TRUE)
    
    # Data frame name string
    df_name <- deparse1(substitute(graphTable))
    if (nchar(df_name) >= 60) {
      df_name <- paste0(substr(df_name, 1, 55), "<...>")
    }
    if (grepl("structure(", df_name, fixed = TRUE)) {
      df_name <- "<argument 'graphTable'>"
    }
    graphTable <- graphTable
    if (!is.data.frame(graphTable)) {
      stop("Argument 'graphTable' is not a 'data.frame' object.\n\n", call. = FALSE)
    }
    graphTable <- as.data.frame(graphTable)
    
    # Set the benchmarking parms
    parms <- extract_parms(graphTable)
    if (is.null(subtitle_str)) {
      subtitle_str <- paste0("Graphics Table <strong>", mask_underscore(df_name), "</strong><br>(", parms$parms_str, ")")
    }
    if (is.null(mth_gap)) {
      mth_gap <- parms$mth_gap
    }
    if (is.null(points_set)) {
      points_set <- parms$ori_plot_set
      segments_set <- c("avgSubAnnual", "avgBenchmark")
    } else {
      segments_set <- intersect(c("avgSubAnnual", "avgBenchmark"), points_set)
    }
    
    graphTable <- graphTable[order(graphTable$year, graphTable$period), ]
  } else {
    ini_theme <- NULL
    segments_set <- c("avgSubAnnual", "avgBenchmark")
  }
  lines_set <- setdiff(points_set, segments_set)
  
  # Create the points data frame for geom_point
  points_df <- stack_tsDF(graphTable[!is.na(graphTable[points_set[1]]), c("year", "period", points_set), drop = FALSE])
  points_df <- points_df[!duplicated(points_df), ]
  points_df$series <- factor(points_df$series, levels = points_set, ordered = TRUE)
  points_df$rDate <- as.Date(paste(points_df$year, 1 + (points_df$period - 1) * mth_gap, "1", sep = "-"), "%Y-%m-%d")
  
  # Extract rows corresponding to lines for geom_line
  lines_df <- points_df[points_df$series %in% lines_set, , drop = FALSE]
  
  # Create the segments data frame for geom_segment
  if (length(segments_set) > 0) {
    df_temp <- graphTable[!is.na(graphTable[segments_set[1]]), c("m", "year", "period", segments_set), drop = FALSE]
    m_vec <- unique(df_temp$m)
    M <- length(m_vec)
    df_beg <- data.frame(matrix(NA_real_, nrow = M, ncol = 2 + length(segments_set)))
    df_end <- df_beg
    names(df_beg) <- c("year", "period", segments_set)
    names(df_end) <- c("year_end", "period_end", segments_set)
    for (kk in m_vec) {
      df_beg[kk, ] <- utils::head(df_temp[df_temp$m == m_vec[kk], c("year", "period", segments_set)], n = 1)
      df_end[kk, ] <- utils::tail(df_temp[df_temp$m == m_vec[kk], c("year", "period", segments_set)], n = 1)
    }
    segments_df <- cbind(stack_tsDF(df_beg), stack_tsDF(df_end,
                                                        yr_cName = "year_end",
                                                        per_cName = "period_end",
                                                        val_cName = "value_end")[, -1])
    segments_df$series <- factor(segments_df$series, levels = segments_set, ordered = TRUE)
    segments_df$rDate <- as.Date(paste(segments_df$year, 1 + (segments_df$period - 1) * mth_gap, "1",
                                       sep = "-"), "%Y-%m-%d")
    segments_df$rDate_end <- as.Date(paste(segments_df$year_end, 1 + (segments_df$period_end - 1) * mth_gap, "1",
                                           sep = "-"), "%Y-%m-%d")
  } else {
    segments_df <- data.frame(rDate = numeric(0L),
                              value = numeric(0L),
                              rDate_end = numeric(0L),
                              value_end = numeric(0L),
                              series = character(0L))
  }
  
  labels <-    c("subAnnual"          = "Indicator Series",
                 "avgSubAnnual"       = "Avg. Indicator Series",
                 "subAnnualCorrected" = "Bias Corr. Indicator Series",
                 "benchmarked"        = "Benchmarked Series",
                 "avgBenchmark"       = "Average Benchmark")
  colors <-    c("subAnnual"          = "blue",
                 "avgSubAnnual"       = "blue",
                 "subAnnualCorrected" = "orange",
                 "benchmarked"        = "red",
                 "avgBenchmark"       = "red")
  pt_shapes <- c("subAnnual"          = "plus",
                 "avgSubAnnual"       = "diamond open",
                 "subAnnualCorrected" = "circle open",
                 "benchmarked"        = "asterisk",
                 "avgBenchmark"       = "triangle open")
  pt_sizes <-  c("subAnnual"          = pt_sz,
                 "avgSubAnnual"       = pt_sz,
                 "subAnnualCorrected" = pt_sz + 0.5,
                 "benchmarked"        = pt_sz,
                 "avgBenchmark"       = pt_sz)
  
  original_scale <- ggplot2::ggplot() +
    
    # Series lines
    ggplot2::geom_line(data = lines_df, ggplot2::aes(x = .data$rDate, y = .data$value,
                                                     color = .data$series)) +
    
    # Average lines (segments)
    ggplot2::geom_segment(data = segments_df, ggplot2::aes(x    = .data$rDate,     y    = .data$value,
                                                           xend = .data$rDate_end, yend = .data$value_end,
                                                           color = .data$series)) +
    
    # Data points for series and repeated average values
    ggplot2::geom_point(data = points_df, ggplot2::aes(x = .data$rDate, y = .data$value,
                                                       color = .data$series,
                                                       shape = .data$series,
                                                       size = .data$series)) +
    
    ggplot2::scale_color_manual(name = NULL, values = colors, limits = points_set, labels = labels) +
    ggplot2::scale_shape_manual(name = NULL, values = pt_shapes, limits = points_set, labels = labels) +
    ggplot2::scale_size_manual(name = NULL, values = pt_sizes, limits = points_set, labels = labels) +
    ggplot2::scale_x_date(name = NULL, date_labels = "%Y", breaks = "1 year", minor_breaks = NULL) +
    ggplot2::labs(title = title_str, subtitle = subtitle_str, y = "Original Scale")
  
  ggplot_list <- list(original_scale)
  display_func(original_scale)

  # Output object returned via function `on.exit()`
}


#' @rdname bench_graphs
#' @export
adj_plot <- function(graphTable, 
                     title_str = "Adjustment Scale",
                     subtitle_str = NULL, 
                     mth_gap = NULL, 
                     full_set = NULL,
                     pt_sz = 2,
                     display_ggplot = TRUE,
                     .setup = TRUE) {
  
  
  # Adjustment Scale graph function (called when `adj_plot_flag = TRUE`)
  # Approach:
  #   - create 3 tall (stacked) data frames for plotting with ggplot:
  #       - data points (geom_point) for adjustment values and repeated average adj. values
  #         (no bias values)
  #       - lines (geom_line) for adjustment and bias lines (not average adj. lines)
  #       - segments (geom_segment) for average adj. lines
  #   - generate "by name (series)" geoms for each type of element to plot
  #     (geom_line, geom_segment and geom_point)
  
  
  # Initialize the object to be returned by the function via `on.exit()`
  ggplot_list <- NULL
  on.exit(return(invisible(ggplot_list)))
  
  # Enforce the default R "error" option (`options(error = NULL)`). E.g. this Turns off traceback
  # generated by calls to the stop() function inside internal functions in R Studio.
  ini_error_opt <- getOption("error")
  on.exit(options(error = ini_error_opt), add = TRUE)
  options(error = NULL)
  
  # Set the display function
  if (gs.validate_arg_logi(display_ggplot)) {
    display_func <- print
  } else {
    display_func <- gs.NULL_func
  }
  
  # Setup part (manual call outside on `plot_graphTable()`)
  if (gs.validate_arg_logi(.setup)) {

    # Set PDF encoding to CP1250 to avoid warnings on Unix-like environments 
    # (see `plot_graphTable()` comment above)
    ini_pdf_encoding <- grDevices::pdf.options()$encoding
    on.exit(grDevices::pdf.options(encoding = ini_pdf_encoding), add = TRUE)
    grDevices::pdf.options(encoding = "CP1250")
    
    # ggplot2 initialization
    ini_theme <- ggplot2::theme_get()
    intialize_theme()
    on.exit(ggplot2::theme_set(ini_theme), add = TRUE)
    
    # Data frame name string
    df_name <- deparse1(substitute(graphTable))
    if (nchar(df_name) >= 60) {
      df_name <- paste0(substr(df_name, 1, 55), "<...>")
    }
    if (grepl("structure(", df_name, fixed = TRUE)) {
      df_name <- "<argument 'graphTable'>"
    }
    graphTable <- graphTable
    if (!is.data.frame(graphTable)) {
      stop("Argument 'graphTable' is not a 'data.frame' object.\n\n", call. = FALSE)
    }
    graphTable <- as.data.frame(graphTable)
    
    # Set the benchmarking parms
    parms <- extract_parms(graphTable)
    if (is.null(subtitle_str)) {
      subtitle_str <- paste0("Graphics Table <strong>", mask_underscore(df_name), "</strong><br>(", parms$parms_str, ")")
    }
    if (is.null(mth_gap)) {
      mth_gap <- parms$mth_gap
    }
    if (is.null(full_set)) {
      full_set <- parms$adj_plot_set
      segments_set <- "avgBenchmarkSubAnnualRatio"
    } else {
      segments_set <- intersect("avgBenchmarkSubAnnualRatio", full_set)
    }
    
    graphTable <- graphTable[order(graphTable$year, graphTable$period), ]
  } else {
    ini_theme <- NULL
    segments_set <- "avgBenchmarkSubAnnualRatio"
  }
  lines_set <- setdiff(full_set, segments_set)
  points_set <- setdiff(full_set, "bias")
  
  # Create the values data frame for lines (geom_lines) and points (geom_point)
  temp_set <- intersect(full_set, union(lines_set, points_set))
  values_df <- stack_tsDF(graphTable[!is.na(graphTable[lines_set[1]]), c("year", "period", temp_set), drop = FALSE])
  values_df <- values_df[!duplicated(values_df), ]
  values_df$series <- factor(values_df$series, levels = temp_set, ordered = TRUE)
  values_df$rDate <- as.Date(paste(values_df$year, 1 + (values_df$period - 1) * mth_gap, "1", sep = "-"), "%Y-%m-%d")
  
  # Extract rows corresponding to lines for geom_line
  lines_df <- values_df[values_df$series %in% lines_set, , drop = FALSE]
  
  # Extract rows corresponding to points for geom_point
  points_df <- values_df[values_df$series %in% points_set, , drop = FALSE]
  
  # Create the segments data frame for geom_segment
  if (length(segments_set) > 0) {
    df_temp <- graphTable[!is.na(graphTable[segments_set[1]]), c("m", "year", "period", segments_set), drop = FALSE]
    m_vec <- unique(df_temp$m)
    M <- length(m_vec)
    df_beg <- data.frame(matrix(NA_real_, nrow = M, ncol = 2 + length(segments_set)))
    df_end <- df_beg
    names(df_beg) <- c("year", "period", segments_set)
    names(df_end) <- c("year_end", "period_end", segments_set)
    for (kk in m_vec) {
      df_beg[kk, ] <- utils::head(df_temp[df_temp$m == m_vec[kk], c("year", "period", segments_set)], n = 1)
      df_end[kk, ] <- utils::tail(df_temp[df_temp$m == m_vec[kk], c("year", "period", segments_set)], n = 1)
    }
    segments_df <- cbind(stack_tsDF(df_beg), stack_tsDF(df_end,
                                                        yr_cName = "year_end",
                                                        per_cName = "period_end",
                                                        val_cName = "value_end")[, -1])
    segments_df$series <- factor(segments_df$series, levels = segments_set, ordered = TRUE)
    segments_df$rDate <- as.Date(paste(segments_df$year, 1 + (segments_df$period - 1) * mth_gap, "1",
                                       sep = "-"), "%Y-%m-%d")
    segments_df$rDate_end <- as.Date(paste(segments_df$year_end, 1 + (segments_df$period_end - 1) * mth_gap, "1",
                                           sep = "-"), "%Y-%m-%d")
  } else {
    segments_df <- data.frame(rDate = numeric(0L),
                              value = numeric(0L),
                              rDate_end = numeric(0L),
                              value_end = numeric(0L),
                              series = character(0L))
  }
  
  labels <-    c("benchmarkedSubAnnualRatio"  = "BI Ratios (Benchmarked Series / Indicator Series)",
                 "avgBenchmarkSubAnnualRatio" = "Avgerage BI Ratios",
                 "bias"                       = "Bias")
  colors <-    c("benchmarkedSubAnnualRatio"  = "blue",
                 "avgBenchmarkSubAnnualRatio" = "black",
                 "bias"                       = "green4")
  ln_types <-  c("benchmarkedSubAnnualRatio"  = "solid",
                 "avgBenchmarkSubAnnualRatio" = "solid",
                 "bias"                       = "longdash")
  pt_shapes <- c("benchmarkedSubAnnualRatio"  = "plus",
                 "avgBenchmarkSubAnnualRatio" = "triangle open",
                 "bias"                       = "")
  pt_sizes  <- c("benchmarkedSubAnnualRatio"  = pt_sz,
                 "avgBenchmarkSubAnnualRatio" = pt_sz,
                 "bias"                       = 0)
  
  adjustment_scale <- ggplot2::ggplot() +
    
    # Adjustment and bias lines
    ggplot2::geom_line(data = lines_df, ggplot2::aes(x = .data$rDate, y = .data$value,
                                                     color = .data$series,
                                                     linetype = .data$series)) +
    
    # Average adjustments (segments)
    ggplot2::geom_segment(data = segments_df, ggplot2::aes(x    = .data$rDate,     y    = .data$value,
                                                           xend = .data$rDate_end, yend = .data$value_end,
                                                           color = .data$series,
                                                           linetype = .data$series)) +
    
    # Data points for adjustment values and repeated average adj. values
    ggplot2::geom_point(data = points_df, ggplot2::aes(x = .data$rDate, y = .data$value,
                                                       color = .data$series,
                                                       shape = .data$series,
                                                       size = .data$series)) +
    
    ggplot2::scale_color_manual(name = NULL, values = colors, limits = full_set, labels = labels) +
    ggplot2::scale_linetype_manual(name = NULL, values = ln_types, limits = full_set, labels = labels) +
    ggplot2::scale_shape_manual(name = NULL, values = pt_shapes, limits = full_set, labels = labels) +
    ggplot2::scale_size_manual(name = NULL, values = pt_sizes, limits = full_set, labels = labels) +
    ggplot2::scale_x_date(name = NULL, date_labels = "%Y", breaks = "1 year", minor_breaks = NULL) +
    ggplot2::labs(title = title_str, subtitle = subtitle_str, y = "Adjustment Scale")
  
  ggplot_list <- list(adjustment_scale)
  display_func(adjustment_scale)

  # Output object returned via function `on.exit()`
}


#' @rdname bench_graphs
#' @export
GR_plot <- function(graphTable, 
                    title_str = "Growth Rates",
                    subtitle_str = NULL, 
                    factor = NULL, 
                    type_chars = NULL, 
                    periodicity = NULL,
                    display_ggplot = TRUE,
                    .setup = TRUE) {
  
  
  # Growth Rates graph function (called when `GR_plot_flag = TRUE`)
  # Approach:
  #   - create a tall (stacked) data frame for plotting with ggplot
  #   - generate a "by name (series)" geom_col
  #   - use geom_col position = "dodge" to stick the 2 GR bars (before/after) together
  #   - use facet_wrap to group the GR bars by year
  #   - create separate graphs (PDF pages) depending on the number of years and the periodicity
  #     (4 years per page for quarterly data, 2 years per page for monthly data)
  #   - use the same X and Y scales (1 to # of periods and GR min/max values) to ensure
  #     proper visual comparability across all pages
  #   - complete the last page with "empty years" to ensure the same look for all years
  #     (same yearly graph width across all pages)

 
  # Initialize the object to be returned by the function via `on.exit()`
  ggplot_list <- NULL
  on.exit(return(invisible(ggplot_list)))
  
  # Enforce the default R "error" option (`options(error = NULL)`). E.g. this Turns off traceback
  # generated by calls to the stop() function inside internal functions in R Studio.
  ini_error_opt <- getOption("error")
  on.exit(options(error = ini_error_opt), add = TRUE)
  options(error = NULL)
  
  # Set the display function
  if (gs.validate_arg_logi(display_ggplot)) {
    display_func <- print
  } else {
    display_func <- gs.NULL_func
  }
  
  # Setup part (manual call outside on `plot_graphTable()`)
  if (gs.validate_arg_logi(.setup)) {

    # Set PDF encoding to CP1250 to avoid warnings on Unix-like environments 
    # (see `plot_graphTable()` comment above)
    ini_pdf_encoding <- grDevices::pdf.options()$encoding
    on.exit(grDevices::pdf.options(encoding = ini_pdf_encoding), add = TRUE)
    grDevices::pdf.options(encoding = "CP1250")
    
    # ggplot2 initialization
    ini_theme <- ggplot2::theme_get()
    intialize_theme()
    on.exit(ggplot2::theme_set(ini_theme), add = TRUE)
    
    # Data frame name string
    df_name <- deparse1(substitute(graphTable))
    if (nchar(df_name) >= 60) {
      df_name <- paste0(substr(df_name, 1, 55), "<...>")
    }
    if (grepl("structure(", df_name, fixed = TRUE)) {
      df_name <- "<argument 'graphTable'>"
    }
    graphTable <- graphTable
    if (!is.data.frame(graphTable)) {
      stop("Argument 'graphTable' is not a 'data.frame' object.\n\n", call. = FALSE)
    }
    graphTable <- as.data.frame(graphTable)
    
    # Set the benchmarking parms
    parms <- extract_parms(graphTable)
    if (is.null(subtitle_str)) {
      subtitle_str <- paste0("Graphics Table <strong>", mask_underscore(df_name), "</strong><br>(", parms$parms_str, ")")
    }
    if (is.null(factor)) {
      factor <- parms$GR_factor
    }
    if (is.null(type_chars)) {
      type_chars <- parms$GR_type_chars
    }
    if (is.null(periodicity)) {
      periodicity <- parms$periodicity
    }
    
    graphTable <- graphTable[order(graphTable$year, graphTable$period), ]
  } else {
    ini_theme <- NULL
  }
  bars_set <- c("growthRateSubAnnual", "growthRateBenchmarked")
  temp_df <- stack_tsDF(graphTable[!duplicated(graphTable[c("year", "period")]), c("year", "period", bars_set)])
  temp_df$series <- factor(temp_df$series, levels = bars_set, ordered = TRUE)
  temp_df$value <- temp_df$value * factor
  y_lab <- paste("Growth Rates", type_chars)
  
  labels <- c("growthRateSubAnnual"   = "Growth R. in Indicator Series",
              "growthRateBenchmarked" = "Growth R. in Benchmarked Series")
  colors <- c("growthRateSubAnnual"   = "steelblue",
              "growthRateBenchmarked" = "indianred3")
  y_limits <- c(min(temp_df$value, na.rm = TRUE), max(temp_df$value, na.rm = TRUE))
  
  years <- graphTable[!is.na(graphTable[bars_set[1]]), "year"]
  beg_yr <- min(years)
  end_yr <- max(years)
  step <- ceiling(16 / periodicity)
  
  # Template plot object
  p <- ggplot2::ggplot() +
    ggplot2::scale_fill_manual(name = NULL, values = colors, labels = labels) +
    ggplot2::scale_x_discrete(name = NULL, limits = as.character(1:periodicity)) +
    ggplot2::scale_y_continuous(name = y_lab, limits = y_limits) +
    ggplot2::labs(title = title_str, subtitle = subtitle_str)
  
  seq_beg_yr <- seq(beg_yr, end_yr, step)
  for (kk in seq_beg_yr) {
    end <- kk + step - 1
    bars_df <- temp_df[temp_df$year >= kk & temp_df$year <= end, , drop = FALSE]
    
    # Add missing panels (years) at the end when necessary (more than 1 row/page of GR plots
    # and the last row/page is not full/complete)
    if (kk > beg_yr && end_yr < end) {
      n_xtra_yrs <- end - end_yr
      bars_df <- rbind(bars_df, data.frame(year = rep((end_yr + 1):end, each = 2),
                                           period = rep(1, n_xtra_yrs * 2),
                                           series = rep(bars_set, n_xtra_yrs),
                                           value = rep(NA_real_, n_xtra_yrs * 2)))
    }
    
    # Create the plot object from the template
    gRates_plot <- p +
      ggplot2::geom_col(data = bars_df, ggplot2::aes(x = .data$period, y = .data$value,
                                                     fill = .data$series),
                        position = "dodge", na.rm = TRUE) +
      ggplot2::facet_wrap(~year, ncol = step)
    
    ggplot_list <- c(ggplot_list, list(gRates_plot))
    display_func(gRates_plot)
  }
  
  # Output object returned via function `on.exit()`
}


#' @rdname bench_graphs
#' @export
GR_table <- function(graphTable, 
                     title_str = "Growth Rates Table",
                     subtitle_str = NULL, 
                     factor = NULL, 
                     type_chars = NULL,
                     display_ggplot = TRUE,
                     .setup = TRUE) {
  
  
  # Growth Rates table function (called when `GR_table_flag = TRUE`)
  # Approach:
  #   - use ggplot to generate page titles that look the same as the previous pages (plots)
  #   - generate a dummy geom (geom_blank) to define an invisible grid to which is attached the
  #     GR table with a tableGrob object (gridExtra package) in an annotation_custom geom
  #   - use element_blank() in the ggplot theme in order to draw nothing else but the GR table
  #     (invisible grid)
  #   - generate as many plots (PDF pages) as necessary in order to display the entire GR table
  #     (21 periods per page, i.e. 23 rows in total including 2 rows for the table header)
  
  
  # Initialize the object to be returned by the function via `on.exit()`
  ggplot_list <- NULL
  on.exit(return(invisible(ggplot_list)))
  
  # Enforce the default R "error" option (`options(error = NULL)`). E.g. this Turns off traceback
  # generated by calls to the stop() function inside internal functions in R Studio.
  ini_error_opt <- getOption("error")
  on.exit(options(error = ini_error_opt), add = TRUE)
  options(error = NULL)
  
  # Set the display function
  if (gs.validate_arg_logi(display_ggplot)) {
    display_func <- print
  } else {
    display_func <- gs.NULL_func
  }
  
  # Setup part (manual call outside on `plot_graphTable()`)
  if (gs.validate_arg_logi(.setup)) {

    # Set PDF encoding to CP1250 to avoid warnings on Unix-like environments 
    # (see `plot_graphTable()` comment above)
    ini_pdf_encoding <- grDevices::pdf.options()$encoding
    on.exit(grDevices::pdf.options(encoding = ini_pdf_encoding), add = TRUE)
    grDevices::pdf.options(encoding = "CP1250")
    
    # ggplot2 initialization
    ini_theme <- ggplot2::theme_get()
    intialize_theme()
    on.exit(ggplot2::theme_set(ini_theme), add = TRUE)
    
    # Data frame name string
    df_name <- deparse1(substitute(graphTable))
    if (nchar(df_name) >= 60) {
      df_name <- paste0(substr(df_name, 1, 55), "<...>")
    }
    if (grepl("structure(", df_name, fixed = TRUE)) {
      df_name <- "<argument 'graphTable'>"
    }
    graphTable <- graphTable
    if (!is.data.frame(graphTable)) {
      stop("Argument 'graphTable' is not a 'data.frame' object.\n\n", call. = FALSE)
    }
    graphTable <- as.data.frame(graphTable)
    
    # Set the benchmarking parms
    parms <- extract_parms(graphTable)
    if (is.null(subtitle_str)) {
      subtitle_str <- paste0("Graphics Table <strong>", mask_underscore(df_name), "</strong><br>(", parms$parms_str, ")")
    }
    if (is.null(factor)) {
      factor <- parms$GR_factor
    }
    if (is.null(type_chars)) {
      type_chars <- parms$GR_type_chars
    }
    
    graphTable <- graphTable[order(graphTable$year, graphTable$period), ]
  } else {
    ini_theme <- NULL
  }
  
  # Extract the GR table info
  GR_df <- graphTable[!duplicated(graphTable[c("year", "period")]),
                      c("year", "period", "subAnnual", "benchmarked", "growthRateSubAnnual", "growthRateBenchmarked")]
  row.names(GR_df) <- NULL
  GR_df$growthRateSubAnnual <- GR_df$growthRateSubAnnual * factor
  GR_df$growthRateBenchmarked <- GR_df$growthRateBenchmarked * factor
  names(GR_df) <- c("Year", "Period", "Indicator\nSeries", "Benchmarked\nSeries",
                    paste("Growth Rate in\nIndicator Series", type_chars),
                    paste("Growth Rate in \nBenchmarked Series", type_chars))
  # Apply default formatting with commas (thousands separator) to all columns except "Year" and "Period"
  GR_df <- cbind(GR_df[1:2], format(GR_df[-(1:2)], big.mark = ","))
  
  # 22 or even 23 rows (periods) would actually fit on a page, but 21 rows looks better: same
  # location for the top of the table on all pages (full set of 21 periods or fewer periods)
  max_rows <- 21
  max_y <- max_rows + 2
  grid_df <- data.frame(x = 1, y = 1:max_y)
  n_rows <- nrow(GR_df)
  n_pages <- ceiling(n_rows / max_rows)
  
  # Template plot object
  p <- ggplot2::ggplot(grid_df, ggplot2::aes(x = .data$x, y = .data$y)) +
    ggplot2::geom_blank() +
    ggplot2::labs(title = title_str, subtitle = subtitle_str) +
    ggplot2::theme(legend.position = "none",
                   axis.title = ggplot2::element_blank(),
                   axis.text = ggplot2::element_blank(),
                   axis.ticks = ggplot2::element_blank(),
                   axis.line = ggplot2::element_blank(),
                   panel.background = ggplot2::element_blank(),
                   panel.border = ggplot2::element_blank(),
                   panel.grid = ggplot2::element_blank())
  
  for (kk in 1:n_pages) {
    id_vec <- (1 + (kk - 1) * max_rows):min(n_rows, kk * max_rows)
    
    # Create the plot object from the template
    gRates_table <- p +
      ggplot2::annotation_custom(gridExtra::tableGrob(GR_df[id_vec, , drop = FALSE], rows = NULL), 
                                 ymin = max_rows - length(id_vec) + 1, ymax = max_y) 
    
    ggplot_list <- c(ggplot_list, list(gRates_table))
    display_func(gRates_table)
  }
  
  # Output object returned via function `on.exit()`
}




#' Initialize the ggplot2 theme for the benchmarking graphics
#' @noRd
intialize_theme <- function() {
  
  # IMPORTANT: keep roxygen2 @return tag up to date with modifications made here!
  
  ggplot2::theme_update(
    plot.title = ggtext::element_markdown(hjust = 0.5),
    plot.subtitle = ggtext::element_markdown(hjust = 0.5),
    legend.position = "bottom",
    plot.margin = ggplot2::margin(t = 1.5, r = 1.5, b = 1.5, l = 1.5, unit = "cm"))
  
  invisible(NULL)
}




#' Extract benchmarking parms for the plots 
#' @noRd
extract_parms <- function(graphTable) {
  
  periodicity <- graphTable$periodicity[1]
  lambda <- graphTable$lambda[1]
  rho <- graphTable$rho[1]
  bias <- graphTable$bias[1]
  constant <- graphTable$constant[1]
  
  # Define the subtitle and plot elements
  # (bias-related info not displayed for Denton benchmarking (`rho = 1`))
  if (abs(rho - 1) > gs.tolerance) {
    parms_str <- paste0("Lambda = ", format(lambda), ", Rho = ", format(rho), ", Bias = ", format(bias))
    ori_plot_set <- c("subAnnual", "avgSubAnnual", "subAnnualCorrected", "benchmarked", "avgBenchmark")
    adj_plot_set <- c("benchmarkedSubAnnualRatio", "avgBenchmarkSubAnnualRatio", "bias")
  } else {
    parms_str <- paste0("Lambda = ", format(lambda), ", Rho = ", format(rho))
    ori_plot_set <- c("subAnnual", "avgSubAnnual", "benchmarked", "avgBenchmark")
    adj_plot_set <- c("benchmarkedSubAnnualRatio", "avgBenchmarkSubAnnualRatio")
  }
  if (abs(constant) > gs.tolerance) {
    parms_str <- paste0(parms_str, ", Constant = ", format(constant))
  }
  
  # Growth rates scale factor (100 or 1) and type (percent or not) for the plot labels
  # and table headers according to lambda
  if (lambda == 0) {
    GR_factor <- 1
    GR_type_chars <- ""
  } else {
    GR_factor <- 100
    GR_type_chars <- "(%)"
  }
  
  # Return the list of parms
  list(
    periodicity = periodicity,
    mth_gap = 12 / periodicity,
    ori_plot_set = ori_plot_set,
    adj_plot_set = adj_plot_set,
    parms_str = parms_str,
    GR_factor = GR_factor,
    GR_type_chars = GR_type_chars)
}




#' Mask undersores (_) for Markdown markup (prevent them from triggering italic formatting)
#' @noRd
mask_underscore <- function(str) {
  gsub("_", "\\\\_", str)
}

######
# Objects and functions in this "common utility script" are prefixed with "gs."
# to make them easy to spot in other (regular) scripts.



######
# Constants

# Package info (DESCRIPTION file fields)
gs.pkg_info <- utils::packageDescription("gseries")
attr(gs.pkg_info, "file") <- NULL  # drop the DESCRIPTION file name and path 

# Extra package info fields (CRAN does not allow extra files in DESCRIPTION file)
gs.pkg_created <- "August 31, 2026, at 2:52:43 PM EDT"
gs.pkg_email <- "g-series@statcan.gc.ca"  # comma separated emails (when applicable)

# Core function header string
gs.header <- paste0("--- Package gseries ", gs.pkg_info$Version)
if (length(gs.pkg_info$Title) > 0) {
  gs.header <- paste0(gs.header, " - ", gs.pkg_info$Title, " ---")
} else {
  gs.header <- paste0(gs.header, " ---")
}
if (length(gs.pkg_created) > 0) {
  gs.header <- paste0(gs.header, "\nCreated on ", gs.pkg_created)
}
if (length(gs.pkg_info$URL) > 0) {
  # Multiple URLs on separate lines
  gs.header <- paste0(gs.header, "\nURL: ", 
                      gsub("[[:blank:]]*,[[:blank:]]*[\n]?[[:blank:]]*", "\n     ", gs.pkg_info$URL))
} 
if (length(gs.pkg_email) > 0) {
  # Multiple Emails on separate lines
  gs.header <- paste0(gs.header, "\nEmail: ", 
                      gsub("[[:blank:]]*,[[:blank:]]*[\n]?[[:blank:]]*", "\n       ", gs.pkg_email))
} 

# Tolerances for numerical comparisons and number of significant digits for rounding
gs.min_tolerance <- .Machine$double.eps  # smallest positive floating-point number x such that `1 + x != 1`
gs.tolerance <- sqrt(gs.min_tolerance)  # default tolerance used in function `all.equal()`
gs.max_signif_digits <- ceiling(log10(1 / gs.min_tolerance))
gs.signif_digits <- ceiling(log10(1 / gs.tolerance))



######
# Functions

# Generic functions that do no processing
gs.FALSE_func <- function(...) {
  FALSE
}
gs.NULL_func <- function(...) {
  invisible(NULL)
}
gs.passThrough_func <- function(x) {
  x
}


# Generic calculation functions (e.g., for columns of output data frames)
gs.calc_ratio <- function(x, y, tol = gs.tolerance) {
  y[abs(y) <= tol] <- NA
  x / y
}
gs.calc_diff <- function(x, y, ...) {
  x - y
}
gs.calc_relDiff <- function(x, y, tol = gs.tolerance) {
  y[abs(y) <= tol] <- NA
  (x - y) / y
}
gs.calc_firstDiff <- function(x, ...) {
  x - c(NA, x[seq_len(length(x) - 1)])
}
gs.calc_relFirstDiff <- function(x, tol = gs.tolerance) {
  y <- c(NA, x[seq_len(length(x) - 1)])
  y[abs(y) <= tol] <- NA
  (x - y) / abs(y)
}


# Data validation functions
gs.check_alter <- function(x) {
  any(!is.finite(x)) || any(x < 0)
}
gs.check_neg <- function(x, tol = gs.tolerance) {
  any(x < -tol)
}


# Returns a character vector of trimmed strings (leading and trailing blanks removed) where missing (NA) 
# and empty ("") strings have been removed (i.e., not included in the returned vector).
#
# Note: lists corresponding to non atomic vectors (i.e., created with `list()` instead of `c()`) are 
#       transformed into atomic vectors (with `unlist()`), resulting in the same character vector 
#       being returned whether argument `str_vec` was created with `list()` or `c()`.
gs.cleanup_col_list <- function(str_vec) {
  x <- unlist(str_vec)
  logi_vec <- is.character(x) & !is.na(x)
  y <- trimws(x[logi_vec])
  y[y != ""]
}


# Utility function to implement argument `verbose = TRUE`, i.e. display of the
# processing steps with elapsed time
gs.display_difftime <- function(time1, time2, label) {
  time_d <- time2 - time1
  message("--- ", label, ": ", paste(format(time_d[[1]]), attr(time_d, "units")), "\n")
  invisible(NULL)
}


# Display/messaging functions

# Utility function to display an object in the console (development mode debugging)
gs.display_obj <- function(x, title = NA, head = FALSE, str = TRUE) {

  if (is.na(title))
    cat("\n", deparse1(substitute(x), collapse = "\n  "), "\n")
  else
    cat("\n", title, "\n")

  if (str)
    str(x)
  
  if (head)
    print(head(x))
  else
    print(x)

  invisible(NULL)
}

# Muted error message (`try(stop())` message)
gs.try_stop <- function(msg) {
  try(stop(msg, call. = FALSE))
}


# Returns a list of string vectors (same length as the input string vector) corresponding to the
# substrings (stripped of leading/trailing blanks) of the input string vector split according to the
# specified pattern (regular expression). The length of the returned list is 1 plus the maximum
# number of occurrences of the pattern in the input string vector. If the pattern is not found in
# the input string vector, then the returned list contains a single element corresponding to the
# input string vector.
#
# Example 1: single-string vector, 1 match
#
#   > gs.split_str("\\/", "abc / def")
#
#   returns a list of 2 strings (2 single-string vectors):
#     [[1]]  character [1]  "abc"
#     [[2]]  character [1]  "def"
#
# Example 2: 4-string vector, maximum of 2 matches
#
#   > gs.split_str("\\/", c("abc / def", "gh ijk", "lmno / p / qr", ""))
#
#   returns a list of 3 (three) 4-string vectors:
#     [[1]]  character [4]  "abc" "gh ijk" "lmno" ""
#     [[2]]  character [4]  "def" ""       "p"    ""
#     [[3]]  character [4]  ""    ""       "qr"   ""
#
# Example 3: 2-string vector, no match
#
#   > gs.split_str("\\/", c("abc def", "g"))
#
#   returns the input vector in a (single-element) list:
#     [[1]]  character [2]  "abc def" "g"
#
#
# Note: similar to base function `strsplit` but where the returned object is "transposed"
#       and the returned strings are stripped of leading and trailing blanks
gs.split_str <- function(pattern, str) {
  out_list <- list()
  while (length(grep(pattern, str)) > 0) {
    pos <- regexpr(pattern, str)
    pos <- pmax(pos, -pos * (nchar(str) + 1))
    out_list <- append(out_list, list(trimws(substr(str, rep(1, length(str)), pos - 1))))
    str <- trimws(substr(str, pos + 1, nchar(str)))
  }
  out_list <- append(out_list, list(str))
  out_list
}


# Validate a logical argument. The function Generates an error message if the argument is invalid 
# and invisibly returns the logical value of the argument otherwise.
gs.validate_arg_logi <- function(arg) {
  out_arg <- as.logical(unlist(arg))
  if (!identical(out_arg, TRUE) && !identical(out_arg, FALSE)) {
    stop("Argument '", deparse1(substitute(arg)), "' must either be TRUE or FALSE.\n\n", call. = FALSE)
  }
  invisible(out_arg)
}


# Validate string vector `cols` against string vector `against`: report (with an error message) 
# strings in `cols` that are not in `against`
gs.validate_cols <- function(cols, against, df_name, source_str = NA) {
  missing_cols <- setdiff(unique(cols), against)
  if (length(missing_cols) > 0) {
    if (is.na(source_str)) {
      stop("The following columns are missing from data frame \"", df_name, "\": ",
           paste0("\n  ", missing_cols, collapse = ""), "\n\n", call. = FALSE)
    } else {
      stop("The following columns listed with ", source_str, " are missing from data frame \"",
           df_name, "\": ", paste0("\n  ", missing_cols, collapse = ""), "\n\n", call. = FALSE)
    }
  } else {
    invisible(NULL)
  }
}




######
# Documented functions


#' Moore-Penrose inverse
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/gs.gInv_MP.html}})}
#' 
#' This function calculates the Moore-Penrose (pseudo) inverse of a square or rectangular matrix 
#' using Singular Value Decomposition (SVD). It is used internally by [tsraking()] and [benchmarking()].
#' 
#' @param X (mandatory) 
#' 
#' Matrix to invert.
#' 
#' @param tol (optional) 
#' 
#' Real number that specifies the tolerance for identifying zero singular values. When `tol = NA` (default), 
#' the tolerance is calculated as the product of the size (dimension) of the matrix, the norm of the matrix 
#' (largest singular value) and the _machine epsilon_ (`.Machine$double.eps`). 
#' 
#' **Default value** is `tol = NA`.
#' 
#' @returns
#' The Moore-Penrose (pseudo) inverse of matrix `X`.
#' 
#' @details
#' The default tolerance (argument `tol = NA`) is coherent with the tolerance used by the MATLAB and GNU Octave 
#' software in their general inverse functions. In our testing, this default tolerance also produced solutions 
#' (results) comparable to G-Series 2.0 in SAS\eqn{^\circledR}{®}.
#' 
#' @seealso [tsraking()] [benchmarking()]
#' 
#' @examples
#' # Invertible matrix
#' X1 <- matrix(c(3, 2, 8, 
#'                6, 3, 2,
#'                5, 2, 4), nrow = 3, byrow = TRUE)
#' Y1 <- gs.gInv_MP(X1)
#' all.equal(Y1, solve(X1))
#' X1 %*% Y1
#' 
#' # Rectangular matrix
#' X2 <- X1[-1, ]
#' try(solve(X2))
#' X2 %*% gs.gInv_MP(X2)
#' 
#' # Non-invertible square matrix
#' X3 <- matrix(c(3, 0, 0, 
#'                0, 0, 0, 
#'                0, 0, 4), nrow = 3, byrow = TRUE)
#' try(solve(X3))
#' X3 %*% gs.gInv_MP(X3)
#' 
#' @export
gs.gInv_MP <- function(X, tol = NA) {
  
  # Related contents on Wikipedia:
  # - "Moore-Penrose inverse" (<https://en.wikipedia.org/wiki/Moore%E2%80%93Penrose_inverse>)
  # - "Singular Value Decomposition" (<https://en.wikipedia.org/wiki/Singular_value_decomposition>)
  
  # Notes on function `svd(X)` where X is a matrix of dimension M x N:
  #   - Returned vector `d` is of dimension MIN(M, N) and contains the (nonnegative) singular
  #     values of `X` sorted decreasingly:
  #       - no need to take the absolute value of `d` in the inversion process;
  #       - the largest singular value is the 1st element of `d`.
  #   - Returned matrix `u` is of dimension M x N if M > N and M x M otherwise.
  #   - Returned matrix `v` is of dimension M x N if M < N and N x N otherwise.
  
  m <- nrow(X)
  n <- ncol(X)
  s <- svd(X)
  if (is.na(tol)) {
    tol <- max(m, n) * s$d[1] * .Machine$double.eps
  }
  d_inv <- ifelse(s$d <= tol, 0, 1 / s$d)
  s$v %*% diag(d_inv, min(m, n)) %*% t(s$u)
}


#' Time values conversion functions
#' 
#' @name time_values_conv
#' 
#' @description
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/time_values_conv.html}})}
#' 
#' Time values conversion functions used internally by other gseries functions.
#' 
#' @param ts (mandatory) 
#' 
#' Time series (object of class "ts" or "mts").
#' 
#' @param sep (optional) 
#' 
#' String (character constant) specifying the separator to use between the year 
#' and period values.
#' 
#' **Default value** is `sep = "-"`.
#' 
#' @returns 
#' [gs.time2year()] returns an integer vector of the "nearest" year (time unit) values. This function is the 
#' equivalent of [stats::cycle()] for time unit values. 
#' 
#' [gs.time2per()] returns an integer vector of the period (cycle) values (see [stats::cycle()]).
#' 
#' [gs.time2str()] returns a character vector corresponding to `gs.time2year(ts)` if `stats::frequency(ts) == 1` 
#' or `gs.time2year(ts)` and `gs.time2per(ts)` separated with `sep` otherwise. 
#' 
#' @seealso [ts_to_tsDF()] [ts_to_bmkDF()] [gs.build_proc_grps()] 
#'  
#' @examples
#' # Dummy monthly time series 
#' mth_ts <- ts(rep(NA, 15), start = c(2019, 1), frequency = 12)
#' mth_ts
#' gs.time2year(mth_ts)
#' gs.time2per(mth_ts)
#' gs.time2str(mth_ts)
#' gs.time2str(mth_ts, sep = "m")
#' 
#' # Dummy quarterly time series 
#' qtr_ts <- ts(rep(NA, 5), start = c(2019, 1), frequency = 4)
#' qtr_ts
#' gs.time2year(qtr_ts)
#' gs.time2per(qtr_ts)
#' gs.time2str(qtr_ts)
#' gs.time2str(qtr_ts, sep = "q")

#' @rdname time_values_conv
#' @export
gs.time2year <- function(ts) {
  # Truncated time values shifted forward by "half a cycle"
  as.integer(signif(stats::time(ts) + 0.5 * stats::deltat(ts), gs.signif_digits))
}

#' @rdname time_values_conv
#' @export
gs.time2per <- function(ts) {
  as.integer(stats::cycle(ts))
}

#' @rdname time_values_conv
#' @export
gs.time2str <- function(ts, sep = "-") {
  if (stats::frequency(ts) == 1) {
    as.character(gs.time2year(ts))
  } else {
    paste(gs.time2year(ts), gs.time2per(ts), sep = sep)
  }
}


#' Build reconciliation processing groups
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/gs.build_proc_grps.html}})}
#' 
#' This function builds the processing groups data frame for reconciliation problems. It is used internally by 
#' [tsraking_driver()] and [tsbalancing()].
#' 
#' @param ts_yr_vec (mandatory) 
#' 
#' Vector of the time series year (time unit) values (see [gs.time2year()]).
#' 
#' @param ts_per_vec (mandatory) 
#' 
#' Vector of the time series period (cycle) values (see [gs.time2per()]).
#' 
#' @param n_per (mandatory) 
#' 
#' Time series length (number of periods).
#' 
#' @param ts_freq (mandatory) 
#' 
#' Time series frequency (see [stats::frequency()]).
#' 
#' @param temporal_grp_periodicity (mandatory) 
#' 
#' Number of periods in temporal groups.
#' 
#' @param temporal_grp_start (mandatory) 
#' 
#' First period of temporal groups.
#' 
#' @inheritSection tsbalancing Processing groups
#' 
#' @returns
#' A data frame with the following variables (columns):
#' - `grp`         : integer vector identifying the processing group (`1:<number-of-groups>`).
#' - `beg_per`     : integer vector identifying the first period of the processing group.
#' - `end_per`     : integer vector identifying the last period of the processing group.
#' - `complete_grp`: logical vector indicating if the processing group corresponds to a complete temporal group.
#' 
#' @seealso [tsraking_driver()] [tsbalancing()] [time_values_conv] 
#' 
#' @example misc/function_examples/gs.build_proc_grps-ex.R
#' 
#' @export
gs.build_proc_grps <- function(ts_yr_vec,
                               ts_per_vec,
                               n_per,
                               ts_freq,
                               temporal_grp_periodicity,
                               temporal_grp_start) {


  # Define processing groups for period-by-period processing (no temporal total preservation)
  if (temporal_grp_periodicity == 1) {

    # Create the processing groups data frame (each period is a processing group)
    per_vec <- 1:n_per
    grp_df <- data.frame(grp = per_vec,
                         beg_per = per_vec,
                         end_per = per_vec,
                         complete_grp = rep(FALSE, n_per))


  # Define processing groups for temporal groups processing (temporal total preservation)
  #
  # => The idea is to create a "ts" object of the temporal group frequency (frequency = `temporal_grp_periodicity`) 
  #    where the "unit values" (integer portion of the "ts" object time values) represent the temporal group and the
  #    "cycle values" represent the periods within the temporal group.
  #
  # => Then...
  #      - the set of periods of each complete temporal group (# of periods = `temporal_grp_periodicity`) will 
  #        constitute individual processing groups (with temporal total preservation)
  #      - each period of each incomplete temporal group (# of periods < `temporal_grp_periodicity`) will 
  #        constitute individual processing groups (without temporal total preservation)
  #
  } else {

    # Define the temporal groups vector (`tmpGrp_vec = 1..[# of temporal groups]` or length `n_per`) identifying 
    # the temporal group associated to each period

    
    # Temporal groups = years (annual total preservation: keep the original start time (and frequency))
    if (temporal_grp_periodicity == ts_freq) {
      start_time <- c(ts_yr_vec[1], ts_per_vec[1])
      start_per <- temporal_grp_start

      
    # Other (non-year) temporal groups (temporal groups where the "time series unit" no longer represents a "year")
    } else {

      # Temporal groups shorter than a year
      if (temporal_grp_periodicity < ts_freq) {

        # Map the input "ts" start to a temporal group period
        start_time <- c(ts_yr_vec[1], (ts_per_vec[1] - 1) %% temporal_grp_periodicity + 1)
        start_per <- temporal_grp_start

      # Temporal groups longer than a year
      } else {

        # Map the temporal group start period to an input "ts" period
        start_per <- (temporal_grp_start - 1) %% ts_freq + 1
        start_id_vec <- which(ts_per_vec == start_per)

        # Calculate extra years for the temporal group start (as necessary)
        if (length(start_id_vec) > 0) {

          # Number of years in the temporal group (capped upwards, i.e., 1.5 -> 2)
          tmpGrp_yrs <- ceiling(temporal_grp_periodicity / ts_freq)

          # Keep `xtra_yrs` in the [0..(`tmpGrp_yrs` - 1)] interval
          xtra_yrs <- (
            # Groups should start on a "whole year" (a multiple of `tmpGrp_yrs`) by default (e.g., even years of biennial groups)
            tmpGrp_yrs - ts_yr_vec[start_id_vec[1]] %% tmpGrp_yrs
            # Shift for groups that shouldn't start on a "whole year" (e.g., odd year start for biennial groups)
            + (temporal_grp_start - 1) %/% ts_freq
          ) %% tmpGrp_yrs

        } else {
          xtra_yrs <- 0
        }

        # At least 1 occurrence of the temporal group start: start time corresponds to the start of a "ts" object
        # (with the new frequency) that ends with the start of the 1st temporal group
        if (length(start_id_vec) > xtra_yrs) {
          start_time <- stats::time(stats::ts(rep(NA, start_id_vec[1 + xtra_yrs]),
                                              end = c(ts_yr_vec[start_id_vec[1]] + xtra_yrs, start_per),
                                              frequency = temporal_grp_periodicity))[1]

        # No occurrence of the temporal group start (no complete temporal group): keep the "original" start time
        } else {
          start_time <- c(ts_yr_vec[1], ts_per_vec[1])
        }
      }
    }
    
    # Create the (new) "ts" object (with the new frequency and start)
    temp_ts <- stats::ts(data = rep.int(NA, n_per),
                         start = start_time,
                         frequency = temporal_grp_periodicity)
    
    # This is where it all happens: the temporal group ids are simply the (new) "ts" object time units 
    # (integer portion of the time values), shifted to take into account the temporal group start, 
    # and then finally "adjusted/realigned/translated" so that they start at 1.
    #
    # => Proper temporal groups determination (assignment) is ensured by the fact that the time values of `temp_ts` 
    #    are always aligned at the start of an interval ("pure ts object" time values), even for cases where the 
    #    original `in_ts` time values are not ("non-pure ts object" time values). I.e., no need to use `gs.time2year()` 
    #    to get the "unit values" (simply "truncating" the time values works fine)
    #
    temp_ts_unit <- as.integer(signif(stats::time(temp_ts) - (start_per - 1) * stats::deltat(temp_ts),
                                      gs.signif_digits))
    tmpGrp_vec <- temp_ts_unit - temp_ts_unit[1] + 1L


    # Create the processing groups data frame
    #  - Periods of a complete temporal group are processed together (in a single processing group)
    #  - Periods of an incomplete temporal group are processed separately (in individual processing groups of 1 period)
    #  - Only the 1st and last temporal groups can be incomplete (middle temporal groups are always complete)

    # 1st temporal group (can be "incomplete")
    grp_n_per <- sum(tmpGrp_vec == 1)

    # Complete
    if (grp_n_per == temporal_grp_periodicity) {
      grp_df <- data.frame(grp = 1,
                           beg_per = 1,
                           end_per = grp_n_per,
                           complete_grp = TRUE)
      grp_padding <- 0

    # Incomplete
    } else {
      per_vec <- 1:grp_n_per
      grp_df <- data.frame(grp = per_vec,
                           beg_per = per_vec,
                           end_per = per_vec,
                           complete_grp = rep(FALSE, grp_n_per))
      grp_padding <- grp_n_per - 1
    }

    # Middle temporal groups (always "complete")
    mid_tmpGrp_vec <- seq_len(max(0, tmpGrp_vec[n_per] - 2)) + 1
    n_mid_grp <- length(mid_tmpGrp_vec)
    beg_per_vec <- (mid_tmpGrp_vec - 2) * temporal_grp_periodicity + 1 + grp_n_per
    end_per_vec <- beg_per_vec + temporal_grp_periodicity - 1
    grp_df <- rbind(grp_df,
                    data.frame(grp = mid_tmpGrp_vec + grp_padding,
                               beg_per = beg_per_vec,
                               end_per = end_per_vec,
                               complete_grp = rep(TRUE, n_mid_grp)))

    # Last temporal group (can be "incomplete")
    if (grp_df$end_per[length(grp_df$end_per)] < n_per) {
      grp_n_per <- n_per - grp_df$end_per[length(grp_df$end_per)]

      # Complete
      if (grp_n_per == temporal_grp_periodicity) {
        beg_per <- grp_df$end_per[length(grp_df$end_per)] + 1
        end_per <- beg_per + temporal_grp_periodicity - 1
        grp_df <- rbind(grp_df,
                        data.frame(grp = tmpGrp_vec[n_per] + grp_padding,
                                   beg_per = beg_per,
                                   end_per = end_per,
                                   complete_grp = TRUE)
        )

      # Incomplete
      } else {
        per_vec <- (grp_df$end_per[length(grp_df$end_per)] + 1):n_per
        grp_df <- rbind(grp_df,
                        data.frame(grp = 1:grp_n_per + grp_df$grp[length(grp_df$grp)],
                                   beg_per = per_vec,
                                   end_per = per_vec,
                                   complete_grp = rep(FALSE, grp_n_per)))
      }
    }
  }

  # Return the processing groups data frame, converting the group and period id's to integers
  # `sapply()` is safe: it always returns a compatible object for the assignment operator 
  # (a vector of length 3 if `grp_df` has a single row or a matrix of matching dimension otherwise)
  grp_df[, 1:3] <- sapply(grp_df[, 1:3], as.integer)
  grp_df
}

#' Function aliases
#' @name aliases
#' 
#' @description
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/aliases.html}})}
#' 
#' [proc_benchmarking()] is an alias for [benchmarking()]
#' 
#' [proc_tsraking()] is an alias for [tsraking()]
#' 
#' [macro_gseriestsbalancing()] is an alias for [tsbalancing()]
#' 
#' @param ... Corresponding function arguments.
#' 
#' @details 
#' The name of these aliases is reminiscent of the corresponding function's SAS\eqn{^\circledR}{®} origin 
#' (PROC BENCHMARKING, PROC TSRAKING and macro ***GSeriesTSBalancing*** in G-Series 2.0). These aliases also 
#' ensure backward compatibility with early development versions of the R package.
#'
#' See the corresponding function for more details, including complete examples, detailed description of the 
#' arguments and the returned value.
#'
#' @returns
#' See the corresponding function:
#' - [benchmarking()] for alias [proc_benchmarking()]
#' - [tsraking()] for alias [proc_tsraking()]
#' - [tsbalancing()] for alias [macro_gseriestsbalancing()]

#' @rdname aliases
#' @export
#' @keywords internal
proc_benchmarking <- function(...) benchmarking(...)

#' @rdname aliases
#' @export
#' @keywords internal
proc_tsraking <- function(...) tsraking(...)

#' @rdname aliases
#' @export
#' @keywords internal
macro_gseriestsbalancing <- function(...) tsbalancing(...)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, include = FALSE---------------------------------------------------
library(gseries)

## ----eval = FALSE-------------------------------------------------------------
# # Load G-Series in R (package gseries)
# library(gseries)
# 
# # Set the working directory (for the PDF graph files)
# iniwd <- setwd("C:/Temp")
# 
# 
# 
# 
# #### Example 1 (single quarterly series) ####
# # Simple case with a single quarterly series to benchmark to annual values
# 
# # Quarterly indicator series
# my_ts_qtr <- ts(
#   c(1.9, 2.4, 3.1, 2.2, 2.0, 2.6, 3.4, 2.4, 2.3),
#   start = c(2015, 1),
#   frequency = 4
# )
# my_series1 <- ts_to_tsDF(my_ts_qtr)
# 
# # Annual benchmarks for quarterly data
# my_ts_ann <- ts(
#   c(10.3, 10.2),
#   start = 2015,
#   frequency = 1
# )
# my_benchmarks1 <- ts_to_bmkDF(my_ts_ann, ind_frequency = 4)
# 
# # Benchmarking using...
# #   - recommended `rho` value for quarterly series (`rho = 0.729`)
# #   - proportional model (`lambda = 1`)
# #   - bias-corrected indicator series with the estimated bias (`biasOption = 3`)
# out_bench1 <- benchmarking(
#   my_series1,
#   my_benchmarks1,
#   rho = 0.729,
#   lambda = 1,
#   biasOption = 3
# )
# 
# # Outputs
# View(out_bench1$series)
# View(out_bench1$benchmarks)
# View(out_bench1$graphTable)
# 
# # Graphics (default set of plots)
# plot_graphTable(
#   out_bench1$graphTable,
#   "Ex1_graphs.pdf"
# )
# # With the G. R. Table
# plot_graphTable(
#   out_bench1$graphTable,
#   "Ex1_graphs_with_GRTable.pdf",
#   GR_table_flag = TRUE
# )
# 
# 
# # Compare with package tempdisagg that allows Denton benchmarking
# # (modified version by Cholette)
# 
# #install.packages("tempdisagg")
# library(tempdisagg)
# #install.packages("dplyr")
# library(dplyr)
# 
# 
# # Proportional Denton
# mult_denton <- benchmarking(
#   my_series1,
#   my_benchmarks1,
#   rho = 1,
#   lambda = 1
# )
# 
# td_mult_denton <- td(
#   formula = my_ts_ann ~ 0 + my_ts_qtr,
#   method = "denton-cholette",
#   criterion = "proportional"
# )
# 
# compare_mult <- ts.union(
#   gseries = tsDF_to_ts(mult_denton$series, frequency = 4),
#   tempdisagg = predict(td_mult_denton),
#   dframe = TRUE
# ) %>%
#   cbind(diff = .[[1]] - .[[2]])
# 
# # Same results
# View(compare_mult)
# 
# 
# # Additive Denton
# add_denton <- benchmarking(
#   my_series1,
#   my_benchmarks1,
#   rho = 1,
#   lambda = 0
# )
# 
# td_add_denton <- td(
#   formula = my_ts_ann ~ 0 + my_ts_qtr,
#   method = "denton-cholette",
#   criterion = "additive"
# )
# 
# compare_add <- ts.union(
#   gseries = tsDF_to_ts(add_denton$series, frequency = 4),
#   tempdisagg = predict(td_add_denton),
#   dframe = TRUE
# ) %>%
#   cbind(diff = .[[1]] - .[[2]])
# 
# # Same results
# View(compare_add)
# 
# 
# 
# 
# #### Example 2 (multiple quarterly series) ####
# # Two quarterly series to benchmark to annual values,
# # with BY-groups and user-defined alterability coefficients
# 
# # Sales data (same sales for groups A and B; only alter coefs for van sales differ)
# qtr_sales <- ts(
#   matrix(
#     c(
#       # Car sales
#       1851, 2436, 3115, 2205, 1987, 2635, 3435, 2361, 2183, 2822,
#       3664, 2550, 2342, 3001, 3779, 2538, 2363, 3090, 3807, 2631,
#       2601, 3063, 3961, 2774, 2476, 3083, 3864, 2773, 2489, 3082,
#       # Van sales
#       1900, 2200, 3000, 2000, 1900, 2500, 3800, 2500, 2100, 3100,
#       3650, 2950, 3300, 4000, 3290, 2600, 2010, 3600, 3500, 2100,
#       2050, 3500, 4290, 2800, 2770, 3080, 3100, 2800, 3100, 2860
#     ),
#     ncol = 2
#   ),
#   start = c(2011, 1),
#   frequency = 4,
#   names = c("car_sales", "van_sales")
# )
# 
# class(qtr_sales)
# 
# ann_sales <- ts(
#   matrix(
#     c(
#       # Car sales
#       10324, 10200, 10582, 11097, 11582, 11092,
#       # Van sales
#       12000, 10400, 11550, 11400, 14500, 16000
#     ),
#     ncol = 2
#   ),
#   start = 2011,
#   frequency = 1,
#   names = c("car_sales", "van_sales")
# )
# 
# # Quarterly indicator series (with default alter coefs for now)
# my_series2 <- rbind(
#   cbind(
#     data.frame(
#       group = rep("A", nrow(qtr_sales)),
#       alt_van = rep(1, nrow(qtr_sales))
#     ),
#     ts_to_tsDF(qtr_sales)
#   ),
#   cbind(
#     data.frame(
#       group = rep("B", nrow(qtr_sales)),
#       alt_van = rep(1, nrow(qtr_sales))
#     ),
#     ts_to_tsDF(qtr_sales)
#   )
# )
# 
# # Set binding van sales (alter coef = 0) for 2012 Q1 and Q2 in group A (rows 5 and 6)
# my_series2$alt_van[c(5, 6)] <- 0
# 
# # Annual benchmarks for quarterly data (without alter coefs)
# my_benchmarks2 <- rbind(
#   cbind(
#     data.frame(group = rep("A", nrow(ann_sales))),
#     ts_to_bmkDF(ann_sales, ind_frequency = 4)
#   ),
#   cbind(
#     data.frame(group = rep("B", nrow(ann_sales))),
#     ts_to_bmkDF(ann_sales, ind_frequency = 4)
#   )
# )
# 
# # Benchmarking using...
# #   - recommended RHO value for quarterly series (rho = 0.729)
# #   - proportional model (lambda = 1)
# #   - without bias correction (biasOption = 1 and bias not specified)
# out_bench2 <- benchmarking(
#   my_series2,
#   my_benchmarks2,
#   rho = 0.729,
#   lambda = 1,
#   biasOption = 1,
#   var = c("car_sales", "van_sales / alt_van"),
#   with = c("car_sales", "van_sales"),
#   by = "group"
# )
# 
# # Outputs
# View(out_bench2$series)
# View(out_bench2$benchmarks)
# View(out_bench2$graphTable)
# 
# # Graphics
# plot_graphTable(
#   out_bench2$graphTable,
#   "Ex2_graphs.pdf"
# )
# 
# 
# 
# 
# #### Example 3 (multiple quarterly series) ####
# # Same as example 2, but benchmarking all 4 series as BY-groups
# # (4 BY-groups of 1 series instead of 2 BY-groups of 2 series)
# 
# my_series3 <- stack_tsDF(ts_to_tsDF(ts.union(A = qtr_sales, B = qtr_sales)))
# my_series3$alter <- 1
# my_series3$alter[
#   my_series3$series == "A.van_sales" &
#     my_series3$year == 2012 &
#     my_series3$period <= 2
# ] <- 0
# my_benchmarks3 <- stack_bmkDF(
#   ts_to_bmkDF(ts.union(A = ann_sales, B = ann_sales), ind_frequency = 4)
# )
# out_bench3 <- benchmarking(
#   my_series3,
#   my_benchmarks3,
#   rho = 0.729,
#   lambda = 1,
#   biasOption = 1,
#   var = "value / alter",
#   with = "value",
#   by = "series"
# )
# 
# # Outputs
# View(out_bench3$series)
# View(out_bench3$benchmarks)
# View(out_bench3$graphTable)
# 
# # Graphics
# plot_graphTable(
#   out_bench3$graphTable,
#   "Ex3_graphs.pdf"
# )
# 
# # Convert the benchmarked series as a "mts" object
# my_out_series3 <- tsDF_to_ts(unstack_tsDF(out_bench3$series), frequency = 4)
# class(my_out_series3)
# plot(my_out_series3)
# 
# 
# 
# 
# #### Monthly data (Box & Jenkins airline series) ####
# 
# data(AirPassengers)
# my_AP_ind <- ts_to_tsDF(AirPassengers)
# 
# # Create annual benchmarks by changing the level (5 times larger), adding some random
# # noise and dropping the last 2 benchmarks
# set.seed(as.Date("2003-03-25"))  # for results reproducibility (select any date you want)
# #set.seed(NULL)
# ann_AP <- round(
#   jitter(
#     aggregate.ts(AirPassengers, nfrequency = 1, FUN = sum) * 5, amount = 2500
#   )
# )
# my_AP_bmk <- (ts_to_bmkDF(ann_AP, ind_frequency = 12))[1:10, ]
# 
# 
# # With bias correction (estimated bias with `biasOption = 3`)
# #   => everything looks good
# out_bench_AP <- benchmarking(
#   my_AP_ind,
#   my_AP_bmk,
#   rho = 0.9,
#   lambda = 1,
#   biasOption = 3
# )
# View(out_bench_AP$series)
# View(out_bench_AP$benchmarks)
# View(out_bench_AP$graphTable)
# 
# plot_graphTable(
#   out_bench_AP$graphTable,
#   "AP_graphs.pdf"
# )
# 
# 
# # Without bias correction (`biasOption = 1`)
# #   => issues with the projected adjustments at the end of the series for periods
# #      not covered by a benchmark
# out_bench_AP_noBias <- benchmarking(
#   my_AP_ind,
#   my_AP_bmk,
#   rho = 0.9,
#   lambda = 1,
#   biasOption = 1
# )
# View(out_bench_AP_noBias$series)
# View(out_bench_AP_noBias$benchmarks)
# View(out_bench_AP_noBias$graphTable)
# 
# plot_graphTable(
#   out_bench_AP_noBias$graphTable,
#   "AP_graphs_noBias.pdf"
# )
# 
# 
# # Denton benchmarking (`rho = 1`, bias correction is irrelevant)
# #   => last adjustment repeated (forever) at the end of the series
# #      (strong assumption, but not necessarily problematic)
# out_bench_AP_denton <- benchmarking(
#   my_AP_ind,
#   my_AP_bmk,
#   rho = 1,
#   lambda = 1
# )
# View(out_bench_AP_denton$series)
# View(out_bench_AP_denton$benchmarks)
# View(out_bench_AP_denton$graphTable)
# 
# plot_graphTable(
#   out_bench_AP_denton$graphTable,
#   "AP_graphs_Denton.pdf"
# )
# 
# 
# # Denton benchmarking approximation (`rho = 0.999`, `biasOption = 3`)
# #   => regression benchmarking model approximation of the "pure" Denton method
# #   => last adjustment repeated at the end of the series (mild convergence to the bias
# #      compared to the "pure" Denton method)
# out_bench_AP_dentonApprox <- benchmarking(
#   my_AP_ind,
#   my_AP_bmk,
#   rho = 0.999,
#   lambda = 1,
#   biasOption = 3
# )
# View(out_bench_AP_dentonApprox$series)
# View(out_bench_AP_dentonApprox$benchmarks)
# View(out_bench_AP_dentonApprox$graphTable)
# 
# plot_graphTable(
#   out_bench_AP_dentonApprox$graphTable,
#   "AP_graphs_DentonApprox.pdf"
# )
# 
# 
# # Pro-rating (`lambda = 0.5` and `rho = 0`)
# #   => no movement preservation: all adjustments are lumped into January every year
# #      (with a disastrous impact on some of the initial December to January movements)
# #   => immediate convergence to the bias (estimated with `biasOption = 3`) for the
# #      projected adjustments at the end of the series
# out_bench_AP_proRate <- benchmarking(
#   my_AP_ind,
#   my_AP_bmk,
#   rho = 0,
#   lambda = 0.5,
#   biasOption = 3
# )
# View(out_bench_AP_proRate$series)
# View(out_bench_AP_proRate$benchmarks)
# View(out_bench_AP_proRate$graphTable)
# 
# plot_graphTable(
#   out_bench_AP_proRate$graphTable,
#   "AP_graphs_proRating.pdf"
# )
# 
# 
# 
# 
# #### End of year stocks ("kinks" in the adjustments with `benchmarking()`) ####
# 
# # Quarterly indicator stock series (same pattern repeated every year)
# my_stock_ind <- ts_to_tsDF(
#   ts(
#     rep(c(85, 95, 125, 95), 7),
#     start = c(2013, 1),
#     frequency = 4
#   )
# )
# 
# # Annual benchmarks (end-of-year stocks)
# my_stock_bmk <- ts_to_bmkDF(
#   ts(
#     c(135, 125, 155, 145, 165),
#     start = 2013,
#     frequency = 1
#   ),
#   discrete_flag = TRUE,
#   alignment = "e",
#   ind_frequency = 4
# )
# 
# # With `benchmarking()` ("Proc Benchmarking" approach)
# out_stock_PB <- benchmarking(
#   my_stock_ind,
#   my_stock_bmk,
#   rho = 0.729,
#   lambda = 1,
#   biasOption = 3
# )
# 
# # With `stock_benchmarking()` ("Stock Benchmarking" approach)
# out_stock_SB <- stock_benchmarking(
#   my_stock_ind,
#   my_stock_bmk,
#   rho = 0.729,
#   lambda = 1,
#   biasOption = 3
# )
# 
# # Benchmarking adjustments of both approaches
# plot_benchAdj(
#   PB_graphTable = out_stock_PB$graphTable,
#   SB_graphTable = out_stock_SB$graphTable
# )
# 
# # Have you noticed how smoother the `stock_benchmarking()` adjustments are compared
# # to the `benchmarking()` ones?
# 
# # The gain in the quality of the resulting benchmarked stocks might not necessarily
# # be obvious in this example.
# plot(
#   out_stock_SB$graphTable$t, out_stock_SB$graphTable$benchmarked,
#   type = "b", col = "red", xlab = "t", ylab = "Benchmarked Stock"
# )
# lines(
#   out_stock_PB$graphTable$t, out_stock_PB$graphTable$benchmarked,
#   type = "b", col = "blue"
# )
# legend(
#   x = "topleft", bty = "n", inset = 0.05, lty = 1, pch = 1,
#   col = c("red", "blue"), legend = c("out_stock_SB", "out_stock_PB")
# )
# title("Benchmarked Stock")
# 
# # PDF graphics
# plot_graphTable(out_stock_PB$graphTable, "Stock_graphs_PB.pdf")
# plot_graphTable(out_stock_SB$graphTable, "Stock_graphs_SB.pdf")
# 
# # What about cases where a flat indicator is used, which may happen in practice
# # in absence of a good indicator of the quarterly (sub-annual) movement?
# my_flat_ind <- my_stock_ind
# my_flat_ind$value <- 1
# out_stock_PB2 <- benchmarking(
#   my_flat_ind,
#   my_stock_bmk,
#   rho = 0.729,
#   lambda = 1,
#   biasOption = 3
# )
# out_stock_SB2 <- stock_benchmarking(
#   my_flat_ind,
#   my_stock_bmk,
#   rho = 0.729,
#   lambda = 1,
#   biasOption = 3
# )
# 
# plot(
#   out_stock_SB2$graphTable$t, out_stock_SB2$graphTable$benchmarked,
#   type = "b", col = "red", xlab = "t", ylab = "Benchmarked Stock"
# )
# lines(
#   out_stock_PB2$graphTable$t, out_stock_PB2$graphTable$benchmarked,
#   type = "b", col = "blue"
# )
# legend(
#   x = "bottomright", bty = "n", inset = 0.05, lty = 1, pch = 1,
#   col = c("red", "blue"), legend = c("out_stock_SB2", "out_stock_PB2")
# )
# title("Benchmarked Stock - Flat Indicator")
# 
# # The awkwardness of the benchmarked stocks produced by `benchmarking()` suddenly
# # becomes obvious. That's because the benchmarked series corresponds to the
# # benchmarking adjustments when using a flat indicator (e.g., a series of 1's
# # with proportional benchmarking):
# plot_benchAdj(
#   PB_graphTable = out_stock_PB2$graphTable,
#   SB_graphTable = out_stock_SB2$graphTable
# )
# 
# # The shortcomings of the "Proc Benchmarking" approach (function `benchmarking()`)
# # with stocks is also quite noticeable in this case when looking at the resulting
# # quarterly growth rates, which are conveniently produced by `plot_graphTable()`.
# # Pay particular attention to the transition in the growth rates from Q4 to Q1
# # every year in the generated PDF graphs.
# plot_graphTable(out_stock_PB2$graphTable, "Stock_graphs_PB_flat_indicator.pdf")
# plot_graphTable(out_stock_SB2$graphTable, "Stock_graphs_SB_flat_indicator.pdf")
# 
# 
# # Reset the working directory to its initial location
# setwd(iniwd)

#' OSQP settings sequence data frame
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/osqp_settings_sequence.html}})}
#' 
#' Data frame containing a sequence of OSQP settings for [tsbalancing()] specified with argument 
#' `osqp_settings_df`. The package includes two predefined OSQP settings sequence data frames:
#' * [`default_osqp_sequence`][osqp_settings_sequence]: fast and effective (default `osqp_settings_df` argument value);
#' * [`alternate_osqp_sequence`][osqp_settings_sequence]: geared towards precision at the expense of execution time.
#' 
#' See `vignette("osqp-settings-sequence-dataframe")` for the actual contents of these data frames.
#'
#' @format A data frame with at least one row and at least one column, the *most common* columns being:
#' \describe{
#'   \item{max_iter}{Maximum number of iterations (integer)}
#'   \item{sigma}{Alternating direction method of multipliers (ADMM) sigma step (double)}
#'   \item{eps_abs}{Absolute tolerance (double)}
#'   \item{eps_rel}{Relative tolerance (double)}
#'   \item{eps_prim_inf}{Primal infeasibility tolerance (double)}
#'   \item{eps_dual_inf}{Dual infeasibility tolerance (double)}
#'   \item{polishing}{Perform solution polishing (logical)}
#'   \item{scaling}{ Number of scaling iterations (integer)}
#'   \item{prior_scaling}{Scale problem data prior to solving with OSQP (logical)}
#'   \item{require_polished}{Require a polished solution to stop the sequence (logical)}
#'   \item{\[*any-other-OSQP-setting*\]}{Value of the corresponding OSQP setting}
#' }
#'
#' @details
#' With the exception of `prior_scaling` and `require_polished`, all columns of the data frame must 
#' correspond to a OSQP setting. Default OSQP values are used for any setting not specified in this 
#' data frame. Visit <https://osqp.org/docs/interfaces/solver_settings.html> for all available OSQP 
#' settings. Note that the OSQP `verbose` setting is actually controlled through [tsbalancing()] arguments 
#' `quiet` and `display_level` (i.e., column `verbose` in a *OSQP settings sequence data frame* would be ignored).
#' 
#' Each row of a *OSQP settings sequence data frame* represents one attempt at solving a balancing problem 
#' with the corresponding OSQP settings. The solving sequence stops as soon as a valid solution is obtained 
#' (a solution for which all constraint discrepancies are smaller or equal to the tolerance 
#' specified with [tsbalancing()] argument `validation_tol`) unless column `require_polished = TRUE`, 
#' in which case a polished solution from OSQP (`status_polish = 1`) would also be required 
#' to stop the sequence. Constraint discrepancies correspond to \eqn{\mathrm{max}(0, l - Ax, Ax - u)} 
#' with constraints defined as \eqn{l \le Ax \le u}{l <= Ax <= u}. In the event where a satisfactory solution 
#' cannot be obtained after having gone through the entire sequence, [tsbalancing()] returns the solution that 
#' generated the smallest total constraint discrepancies among valid solutions, if any, or among all solutions, 
#' otherwise. Note that running the entire solving sequence can be *enforced* by specifying [tsbalancing()] argument 
#' `full_sequence = TRUE`. Rows with column `prior_scaling = TRUE` have the problem data scaled prior to 
#' solving with OSQP, using the average of the free (nonbinding) problem values as the scaling factor.
#' 
#' In addition to specifying a custom-made *OSQP settings sequence data frame* with argument `osqp_settings_df`, 
#' one can also specify `osqp_settings_df = NULL` which would result in a single solving attempt with default OSQP 
#' values for all settings along with `prior_scaling = FALSE` and `require_polished = FALSE`. Note that it is 
#' recommended, however, to first try data frames `default_osqp_sequence` and `alternate_osqp_sequence`, along with 
#' `full_sequence = TRUE` if necessary, before considering other alternatives.
#' 
#' Vignette *OSQP Settings Sequence Data Frame* (`vignette("osqp-settings-sequence-dataframe")`) contains additional 
#' information.
#' 
#' @name osqp_settings_sequence
NULL

# NOTES (usage examples): 
#   a) bringing them all under subject `osqp_settings_sequence` above would generate 2 residual/extra lines 
#      ("default_osqp_sequence" and "alternate_osqp_sequence") in the Rd file (!)
#   b) the `tsbalancing()` call examples MUST be commented out, otherwise:
#      - the reference page is indexed under subject `tsbalancing()` instead of subjects 
#        `osqp_settings_sequence`, `default_osqp_sequence` and `alternate_osqp_sequence`.
#      - `R CMD check` fails: it complains about the `...` argument not being defined with `tsbalancing()` 
#        (removing the dot arguments from the calls would work but not solve the preceding issue).

#' @rdname osqp_settings_sequence
#' @usage 
#' # Default sequence:
#' # tsbalancing(..., osqp_settings_df = default_osqp_sequence)
"default_osqp_sequence"

#' @rdname osqp_settings_sequence
#' @usage 
#' 
#' # Alternative (slower) sequence:
#' # tsbalancing(..., osqp_settings_df = alternate_osqp_sequence)
#' 
#' # Custom-made sequence (use with caution!):
#' # tsbalancing(..., osqp_settings_df = <my-osqp-sequence-dataframe>)
#' 
#' # Single solving attempt with the default OSQP settings (not recommended!):
#' # tsbalancing(..., osqp_settings_df = NULL)
"alternate_osqp_sequence"

#' Reciprocal function of [stack_tsDF()]
#'
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/unstack_tsDF.html}})}
#' 
#' Convert a stacked (tall) multivariate time series data frame ([benchmarking()] and [stock_benchmarking()]
#' data format) into a non-stacked (wide) multivariate time series data frame.
#'
#' This function, combined with [tsDF_to_ts()], is useful to convert the benchmarked data frame returned by a call to 
#' [benchmarking()] or [stock_benchmarking()] back into a "mts" object, where multiple series were benchmarked in 
#' *BY-group* processing mode.
#'
#'
#' @param ts_df (mandatory)
#'
#' Data frame (object of class "data.frame") that contains the multivariate time series data to be *unstacked*.
#'
#' @param ser_cName (optional)
#'
#' String specifying the name of the character variable (column) in the input time series data frame that 
#' contains the series identifier (the time series variable names in the output data frame).
#'
#' **Default value** is `ser_cName = "series"`.
#'
#' @param yr_cName,per_cName (optional)
#'
#' Strings specifying the name of the numeric variables (columns) in the input time series data frame that
#' contain the data point year and period identifiers. These variables are *transferred* to the output data 
#' frame with the same names.
#'
#' **Default values** are `yr_cName = "year"` and `per_cName   = "period"`.
#'
#' @param val_cName (optional)
#'
#' String specifying the name of the numeric variable (column) in the input time series data frame that
#' contains the data point values.
#'
#' **Default value** is `val_cName = "value"`.
#'
#'
#' @returns
#' The function returns a data frame with three or more variables:
#' * Data point year, type numeric (see argument `yr_cName`)
#' * Data point period, type numeric (see argument `per_cName`)
#' * One time series data variable for each distinct value of the input data frame variable specified with
#' argument `ser_cName`, type numeric (see arguments `ser_cName` and `val_cName`)
#'
#' Note: the function returns a "data.frame" object than can be explicitly coerced to another type of object 
#' with the appropriate `as*()` function (e.g., `tibble::as_tibble()` would coerce it to a tibble).
#'
#'
#' @seealso [stack_tsDF()] [tsDF_to_ts()] [benchmarking()] [stock_benchmarking()]
#'
#'
#' @example misc/function_examples/unstack_tsDF-ex.R
#'
#'
#' @export
unstack_tsDF <- function(ts_df,
                         ser_cName = "series",
                         yr_cName = "year",
                         per_cName = "period",
                         val_cName = "value") {

  # Validate object
  if (!is.data.frame(ts_df)) {
    stop("Argument 'ts_df' is not a 'data.frame' object.\n\n", call. = FALSE)
  }
  ts_df <- as.data.frame(ts_df)
  df_cols <- names(ts_df)
  cols <- c(ser_cName, yr_cName, per_cName, val_cName)
  args <- c("ser_cName", "yr_cName", "per_cName", "val_cName")
  for (ii in seq_along(cols)) {
    if (!(cols[ii] %in% df_cols)) {
      stop("The input data frame does not contain column \"", cols[ii], "\" (argument '", args[ii],
           "').\n\n", call. = FALSE)
    }
  }

  # Unstack the data frame
  ser_list <- unique(ts_df[[ser_cName]])
  date_cols <- c(yr_cName, per_cName)
  out_df <- unique(ts_df[date_cols])
  for (ser in ser_list) {
    tmp_df <- ts_df[ts_df[ser_cName] == ser, c(date_cols, val_cName)]
    names(tmp_df)[3] <- ser
    out_df <- merge(out_df, tmp_df, by = date_cols, all = TRUE)
  }

  # Sort the data frame by date and reset the row names (numbers)
  out_df <- out_df[order(out_df[[yr_cName]], out_df[[per_cName]]), ]
  row.names(out_df) <- NULL

  out_df
}

#' Restore temporal constraints
#'
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/benchmarking.html}})}
#' 
#' _Replication of the G-Series 2.0 SAS\eqn{^\circledR}{®} BENCHMARKING procedure (PROC BENCHMARKING).
#' See the G-Series 2.0 documentation for details (Statistics Canada 2016)._
#'
#' This function ensures coherence between time series data of the same target variable measured at different 
#' frequencies (e.g., sub-annually and annually). Benchmarking consists of imposing the level of the benchmark 
#' series (e.g., annual data) while minimizing the revisions of the observed movement in the indicator series 
#' (e.g.,  sub-annual data) as much as possible. The function also allows nonbinding benchmarking where the 
#' benchmark series can also be revised.
#'
#' The function may also be used for benchmarking-related topics such as *temporal distribution*
#' (the reciprocal action of benchmarking: disaggregation of the benchmark series into more frequent observations), 
#' *calendarization* (a special case of temporal distribution) and *linking* (the connection of different time series 
#' segments into a single consistent time series).
#'
#' Several series can be benchmarked in a single function call.
#'
#'
#' @usage
#' benchmarking(
#'   series_df,
#'   benchmarks_df,
#'   rho,
#'   lambda,
#'   biasOption,
#'   bias = NA,
#'   tolV = 0.001,
#'   tolP = NA,
#'   warnNegResult = TRUE,
#'   tolN = -0.001,
#'   var = "value",
#'   with = NULL,
#'   by = NULL,
#'   verbose = FALSE,
#'
#'   # New in G-Series 3.0
#'   constant = 0,
#'   negInput_option = 0,
#'   allCols = FALSE,
#'   quiet = FALSE
#' )
#'
#'
#' @param series_df (mandatory)
#'
#' Data frame (object of class "data.frame") that contains the indicator time series data to be 
#' benchmarked. In addition to the series data variable(s), specified with argument `var`, the data 
#' frame must also contain two numeric variables, `year` and `period`, identifying the periods of 
#' the indicator time series.
#'
#' @param benchmarks_df (mandatory)
#'
#' Data frame (object of class "data.frame") that contains the benchmarks. In addition to the benchmarks 
#' data variable(s), specified with argument `with`, the data frame must also contain four numeric 
#' variables, `startYear`, `startPeriod`, `endYear` and `endPeriod`, identifying the indicator time 
#' series periods covered by each benchmark.
#'
#' @param rho (mandatory)
#'
#' Real number in the \eqn{[0,1]} interval that specifies the value of the autoregressive parameter
#' \eqn{\rho}. See section **Details** for more information on the effect of parameter \eqn{\rho}.
#'
#' @param lambda (mandatory)
#'
#' Real number, with suggested values in the \eqn{[-3,3]} interval, that specifies the value of the
#' adjustment model parameter \eqn{\lambda}. Typical values are `lambda = 0.0` for an additive model
#' and `lambda = 1.0` for a proportional model.
#'
#' @param biasOption (mandatory)
#'
#' Specification of the bias estimation option:
#' 
#' * `1`: Do not estimate the bias. The bias used to correct the indicator series will be the value 
#' specified with argument `bias`.
#' * `2`: Estimate the bias, display the result, but do not use it. The bias used to correct the indicator 
#' series will be the value specified with argument `bias`.
#' * `3`: Estimate the bias, display the result and use the estimated bias to correct the indicator series. 
#' Any value specified with argument `bias` will be ignored.
#'
#' Argument `biasOption` is ignored when `rho = 1.0`. See section **Details** for more information on the bias.
#'
#' @param bias (optional)
#'
#' Real number, or `NA`, specifying the value of the user-defined bias to be used for the correction of the
#' indicator series prior to benchmarking. The bias is added to the indicator series with an additive model
#' (argument  `lambda = 0.0`) while it is multiplied otherwise (argument `lambda != 0.0`). No bias
#' correction is applied when `bias = NA`, which is equivalent to specifying `bias = 0.0` when `lambda = 0.0`
#' and `bias = 1.0` otherwise. Argument `bias` is ignored when `biasOption = 3` or `rho = 1.0`. See section 
#' **Details** for more information on the bias.
#'
#' **Default value** is `bias = NA` (no user-defined bias).
#'
#' @param tolV,tolP (optional)
#'
#' Nonnegative real number, or `NA`, specifying the tolerance, in absolute value or percentage, to be used
#' for the validation of the output binding benchmarks (alterability coefficient of \eqn{0.0}). This validation 
#' compares the input binding benchmark values with the equivalent values calculated from the benchmarked series 
#' (output) data. Arguments `tolV` and `tolP` cannot be both specified (one must be specified while the other 
#' must be `NA`).
#'
#' **Example:** to set a tolerance of 10 *units*, specify `tolV = 10, tolP = NA`; to set a tolerance of 1%, 
#' specify `tolV = NA, tolP = 0.01`.
#'
#' **Default values** are `tolV = 0.001` and `tolP = NA`.
#'
#' @param warnNegResult (optional)
#'
#' Logical argument specifying whether a warning message is generated when a negative value created by the
#' function in the benchmarked (output) series is smaller than the threshold specified by argument `tolN`.
#'
#' **Default value** is `warnNegResult = TRUE`.
#'
#' @param tolN (optional)
#'
#' Negative real number specifying the threshold for the identification of negative values. A value is
#' considered negative when it is smaller than this threshold.
#'
#' **Default value** is `tolN = -0.001`.
#'
#' @param var (optional)
#'
#' String vector (minimum length of 1) specifying the variable name(s) in the indicator series data frame 
#' (argument `series_df`) containing the values and (optionally) the user-defined alterability coefficients 
#' of the series to be benchmarked. These variables must be numeric.
#'
#' The syntax is `var = c("series1 </ alt_ser1>", "series2 </ alt_ser2>", ...)`. Default alterability
#' coefficients of \eqn{1.0} are used when a user-defined alterability coefficients variable is not specified
#' alongside an indicator series variable. See section **Details** for more information on alterability coefficients.
#'
#' **Example:** `var = "value / alter"` would benchmark indicator series data frame variable `value` with the
#' alterability coefficients contained in variable `alter` while `var = c("value / alter", "value2")` would
#' additionally benchmark variable `value2` with default alterability coefficients of \eqn{1.0}.
#'
#' **Default value** is `var = "value"` (benchmark variable `value` using default alterability coefficients of
#' \eqn{1.0}).
#'
#' @param with (optional)
#'
#' String vector (same length as argument `var`), or `NULL`, specifying the variable name(s) in the benchmarks data 
#' frame (argument `benchmarks_df`) containing the values and (optionally) the user-defined alterability coefficients 
#' of the benchmarks. These variables must be numeric. Specifying `with = NULL` results in using benchmark variable(s) 
#' with the same names(s) as those specified with argument `var` without user-defined benchmark alterability coefficients 
#' (i.e., default alterability coefficients of \eqn{0.0} corresponding to binding benchmarks).
#'
#' The syntax is `with = NULL` or `with = c("bmk1 </ alt_bmk1>", "bmk2 </ alt_bmk2>", ...)`. Default alterability
#' coefficients of \eqn{0.0} (binding benchmarks) are used when a user-defined alterability coefficients variable
#' is not specified alongside a benchmark variable. See section **Details** for more information on alterability
#' coefficients.
#'
#' **Example:** `with = "val_bmk"` would use benchmarks data frame variable `val_bmk` with default benchmark
#' alterability coefficients of \eqn{0.0} to benchmark the indicator series while
#' `with = c("val_bmk", "val_bmk2 / alt_bmk2")` would additionally benchmark a second indicator series using
#' benchmark variable `val_bmk2` with the benchmark alterability coefficients contained in variable `alt_bmk2`.
#'
#' **Default value** is `with = NULL` (same benchmark variable(s) as argument `var` using default benchmark
#' alterability coefficients of \eqn{0.0}).
#'
#' @param by (optional)
#'
#' String vector (minimum length of 1), or `NULL`, specifying the variable name(s) in the input data frames 
#' (arguments `series_df` and `benchmarks_df`) to be used to form groups (for _BY-group_ processing) and allow 
#' the benchmarking of multiple series in a single function call. BY-group variables can be numeric or 
#' character (factors or not), must be present in both input data frames and will appear in all three output 
#' data frames (see section **Value**). BY-group processing is not  implemented when `by = NULL`. See 
#' "Benchmarking Multiple Series" in section **Details** for more information.
#'
#' **Default value** is `by = NULL` (no BY-group processing).
#'
#' @param verbose (optional)
#'
#' Logical argument specifying whether information on intermediate steps with execution time (real time,
#' not CPU time) should be displayed. Note that specifying argument `quiet = TRUE` would *nullify* argument 
#' `verbose`.
#'
#' **Default value** is `verbose = FALSE`.
#'
#' @param constant (optional)
#'
#' Real number that specifies a value to be temporarily added to both the indicator series and the
#' benchmarks before solving proportional benchmarking problems (`lambda != 0.0`). The temporary
#' constant is removed from the final output benchmarked series. E.g., specifying a (small) constant
#' would allow proportional benchmarking with `rho = 1` (e.g., proportional Denton benchmarking) on
#' indicator series that include values of 0. Otherwise, proportional benchmarking with values of 0 in
#' the indicator series is only possible when `rho < 1`. Specifying a constant with additive benchmarking
#' (`lambda = 0.0`) has no impact on the resulting benchmarked data. The data variables in the `graphTable`
#' output data frame include the constant, corresponding to the benchmarking problem that was actually solved.
#'
#' **Default value** is `constant = 0` (no temporary additive constant).
#'
#' @param negInput_option (optional)
#'
#' Handling of negative values in the input data for proportional benchmarking (`lambda != 0.0`):
#' 
#' * `0`: Do not allow negative values with proportional benchmarking. An error message is displayed in the 
#' presence of negative values in the input indicator series or benchmarks and missing (`NA`) values are 
#' returned for the benchmarked series. This corresponds to the G-Series 2.0 behaviour.
#' * `1`: Allow negative values with proportional benchmarking but display a warning message.
#' * `2`: Allow negative values with proportional benchmarking without displaying any message.
#' 
#' **Default value** is `negInput_option = 0` (do not allow negative values with proportional benchmarking).
#'
#' @param allCols (optional)
#'
#' Logical argument specifying whether all variables in the indicator series data frame (argument `series_df`), 
#' other than `year` and `period`, determine the set of series to benchmark. Values specified with arguments `var` 
#' and `with` are ignored when `allCols = TRUE`, which automatically implies default alterability coefficients, 
#' and variables with the same names as the indicator series must exist in the benchmarks data frame (argument 
#' `benchmarks_df`).
#'
#' **Default value** is `allCols = FALSE`.
#'
#' @param quiet (optional)
#'
#' Logical argument specifying whether or not to display only essential information such as warning messages,
#' error messages and variable (series) or BY-group information when multiple series are benchmarked in a single call 
#' to the function. We advise against *wrapping* your [benchmarking()] call with [suppressMessages()] to further suppress 
#' the display of variable (series) or BY-group information when processing multiple series as this would make 
#' troubleshooting difficult in case of issues with individual series. Note that specifying `quiet = TRUE` would 
#' also *nullify* argument `verbose`.
#'
#' **Default value** is `quiet = FALSE`.
#'
#'
#' @details
#' When \eqn{\rho < 1}, this function returns the generalized least squared solution of a special case of the general 
#' regression-based benchmarking model proposed by Dagum and Cholette (2006). The model, in matrix form, is:
#' \deqn{\displaystyle
#' \begin{bmatrix} s^\dagger \\ a \end{bmatrix} = 
#' \begin{bmatrix} I \\ J \end{bmatrix} \theta + 
#' \begin{bmatrix} e \\ \varepsilon \end{bmatrix}
#' }{[s^dag; a] = [I; J] theta + [e; epsilion]}
#' where
#' * \eqn{a} is the vector of length \eqn{M} of the benchmarks.
#' * \eqn{s^\dagger = \left\{
#'     \begin{array}{cl}
#'       s + b & \text{if } \lambda = 0 \\
#'       s \cdot b  & \text{otherwise}
#'     \end{array} \right.
#'   }{s^dag = s + b if lambda = 0, s^dag = s * b otherwise} is the vector of length \eqn{T} of the bias corrected 
#'   indicator series values, with \eqn{s} denoting the initial (input) indicator series.
#' * \eqn{b} is the bias, which is specified with argument `bias` when argument `bias_option != 3` or, when `bias_option = 3`, 
#' is estimated as \eqn{\hat{b} = \left\{
#'     \begin{array}{cl}
#'       \frac{{1_M}^\mathrm{T} (a - Js)}{{1_M}^\mathrm{T} J 1_T} & \text{if } \lambda = 0 \\
#'       \frac{{1_M}^\mathrm{T} a}{{1_M}^\mathrm{T} Js} & \text{otherwise}
#'     \end{array} \right.
#'   }{b^hat = ({1_M}'(a - J s)) / ({1_M}' J 1_T) if \lambda = 0, b^hat = ({1_M}' a) / ({1_M}' J s) otherwise}, where 
#'   \eqn{1_X = (1, ..., 1)^\mathrm{T}}{1_X = (1, ..., 1)'} is a vector of \eqn{1} of length \eqn{X}.
#' * \eqn{J} is the \eqn{M \times T}{M x T} matrix of temporal aggregation constraints with elements \if{latex}{\cr} 
#'   \eqn{j_{m, t} = \left\{
#'     \begin{array}{cl}
#'       1 & \text{if benchmark } m \text{ covers period } t \\
#'       0 & \text{otherwise}
#'     \end{array} \right.
#'   }{j_{m,t} = 1 if benchmark m covers period t, j_{m,t} = 0 otherwise}.
#' * \eqn{\theta} is the vector of the final (benchmarked) series values.
#' * \eqn{e \sim \left( 0, V_e \right)}{e ~ (0, V_e)} is the vector of the measurement errors of \eqn{s^\dagger}{s^dag} 
#' with covariance matrix \eqn{V_e = C \Omega_e C}.
#' * \eqn{C = \mathrm{diag} \left( \sqrt{c_{s^\dagger}} \left| s^\dagger \right|^\lambda \right)}{C = diag(\sqrt{c_{s^dag}} 
#' |s^dag|^lambda)} where \eqn{c_{s^\dagger}}{c_{s^dag}} is the vector of the alterability coefficients of \eqn{s^\dagger}{s^dag}, 
#' assuming \eqn{0^0 = 1}.
#' * \eqn{\Omega_e} is a \eqn{T \times T}{T x T} matrix with elements \eqn{\omega_{e_{i,j}} = \rho^{|i-j|}} representing the 
#' autocorrelation of an AR(1) process, again assuming \eqn{0^0 = 1}.
#' * \eqn{\varepsilon \sim (0, V_\varepsilon)}{epsilon ~ (0, V_epsilon)} is the vector of the measurement errors of the benchmarks 
#' \eqn{a} with covariance matrix \eqn{V_\varepsilon = \mathrm{diag} \left( c_a a \right)}{V_epsilion = diag(c_a a)} where \eqn{c_a} 
#' is the vector of the alterability coefficients of the benchmarks \eqn{a}.
#' 
#' The generalized least squared solution is:
#' \deqn{\displaystyle 
#' \hat{\theta} = s^\dagger + V_e J^{\mathrm{T}} \left( J V_e J^{\mathrm{T}} + V_\varepsilon \right)^+ \left( a - J s^\dagger \right)
#' }{theta^hat = s^dag + V_e J' (J V_e J' + V_epsilion)^{+} (a - J s^dag)}
#' where \eqn{A^{+}} designates the Moore-Penrose inverse of matrix \eqn{A}.
#' 
#' When \eqn{\rho = 1}, the function returns the solution of the (modified) Denton method:
#' \deqn{\displaystyle 
#' \hat{\theta} = s + W \left( a - J s \right)
#' }{theta^hat = s + W (a - J s)}
#' where
#' * \eqn{W} is the upper-right corner matrix from the following matrix product
#'   \deqn{
#'     \left[\begin{array}{cc}
#'       D^{+} \Delta^{\mathrm{T}} \Delta D^{+} & J^{\mathrm{T}} \\
#'       J & 0
#'     \end{array} \right]^{+}
#'     \left[\begin{array}{cc}
#'       D^{+} \Delta^{\mathrm{T}} \Delta D^{+} & 0 \\
#'       J & I_M
#'     \end{array} \right] = 
#'     \left[\begin{array}{cc}
#'       I_T & W \\
#'       0 & W_\nu
#'     \end{array} \right]
#'   }{[D^{+} Delta' Delta D^{+}, J'; J, 0]^{+} [D^{+} Delta' Delta D^{+}, 0; J, I_M] = [I_T, W; 0, W_nu]}
#' * \eqn{D = \mathrm{diag} \left( \left| s \right|^\lambda \right)}{D = diag(|s|^\lambda)}, assuming \eqn{0^0 = 1}. Note that 
#' \eqn{D} corresponds to \eqn{C} with \eqn{c_{s^\dagger} = 1.0}{c_{s^dag} = 1.0} and without bias correction (arguments 
#' `bias_option = 1` and `bias = NA`).
#' * \eqn{\Delta}{Delta} is a \eqn{T-1 \times T}{T-1 x T} matrix with elements \eqn{\delta_{i,j} = \left\{
#'     \begin{array}{cl}
#'       -1 & \text{if } i=j \\
#'       1 & \text{if } j=i+1 \\
#'       0 & \text{otherwise}
#'     \end{array} \right.
#'   }{delta_{i,j} = 1 if i = j, delta_{i,j} = 1 if j = i + 1, delta_{i,j} = 0 otherwise}.
#' * \eqn{W_\nu} is a \eqn{M \times M}{M x M} matrix associated with the Lagrange multipliers of the corresponding 
#' minimization problem, expressed as:
#' \deqn{\displaystyle 
#' \begin{aligned}
#' & \underset{\theta}{\text{minimize}} 
#' & & \sum_{t \ge 2} \left[ \frac{\left( s_t - \theta_t \right)}{\left| s_t\right|^\lambda}
#'       - \frac{\left( s_{t-1} - \theta_{t-1} \right)}{\left| s_{t-1}\right|^\lambda} \right]^2 \\
#' & \text{subject to} 
#' & & a = J \theta
#' \end{aligned}
#' }{min(theta) sum_{t>1}{[(s_t - theta_t) / |s_t|^lambda - (s_{t-1} - theta_{t-1}) / |s_{t-1}|^lambda]^2}, subject to a = J theta}
#' 
#' See Quenneville et al. (2006) and Dagum and Cholette (2006) for details.
#' 
#' ## Autoregressive Parameter \eqn{\rho} and *bias*
#' Parameter \eqn{\rho} (argument `rho`) is associated to the change between the (input) indicator and the (output)
#' benchmarked series for two consecutive periods and is often called the *movement preservation parameter*. The larger
#' the value of \eqn{\rho}, the more the indicator series period to period movements are preserved in the
#' benchmarked series. With \eqn{\rho = 0}, period to period movement preservation is not enforced and the
#' resulting benchmarking adjustments are not smooth, as in the case of prorating (\eqn{\rho = 0} and
#' \eqn{\lambda = 0.5}) where the adjustments take the shape of a *step function*. At the other end of the
#' spectrum is \eqn{\rho = 1}, referred to as *Denton benchmarking*, where period to period movement preservation
#' is maximized, which results in the smoothest possible set of benchmarking adjustments available with the function.
#'
#' The *bias* represents the expected discrepancies between the benchmarks and the indicator series. It can be
#' used to pre-adjust the indicator series in order to reduce, on average, the discrepancies between the two
#' sources of data. Bias correction, which is specified with arguments `biasOption` and `bias`, can be particularly 
#' useful for periods not covered by benchmarks when \eqn{\rho < 1}. In this context, parameter \eqn{\rho} dictates the 
#' speed at which the projected benchmarking adjustments converge to the bias (or converge to *no adjustment* without bias 
#' correction) for periods not covered by a benchmark. The smaller the value of \eqn{\rho}, the faster the convergence to 
#' the bias, with immediate convergence when \eqn{\rho = 0} and no convergence at all (the adjustment of the last period 
#' covered by a benchmark is repeated) when \eqn{\rho = 1} (Denton benchmarking). Arguments `biasOption` and `bias` are 
#' actually ignored when \eqn{\rho = 1} since correcting for the bias has no impact on Denton benchmarking solutions.
#' The suggested value for \eqn{\rho} is \eqn{0.9} for monthly indicators and \eqn{0.9^3 = 0.729} for quarterly indicators, 
#' representing a reasonable compromise between maximizing movement preservation and reducing revisions as new benchmarks 
#' become available in the future (benchmarking *timeliness issue*). In practice, note that Denton benchmarking could be 
#' *approximated* with the regression-based model by using a \eqn{\rho} value that is smaller than, but very close to, 
#' \eqn{1.0} (e.g., \eqn{\rho = 0.999}). See Dagum and Cholette (2006) for a complete discussion on this topic.
#'
#' ## Alterability Coefficients
#' Alterability coefficients \eqn{c_{s^\dagger}}{c_{s^dag}} and \eqn{c_a} conceptually represent the measurement errors 
#' associated with the (bias corrected) indicator time series values \eqn{s^\dagger}{s^dag} and benchmarks \eqn{a} 
#' respectively. They are nonnegative real numbers which, in practice, specify the extent to which an initial value can be 
#' modified in relation to other values. Alterability coefficients of \eqn{0.0} define fixed (binding) values while 
#' alterability coefficients greater than \eqn{0.0} define free (nonbinding) values. Increasing the alterability coefficient 
#' of an intial value results in more changes for that value in the benchmarking solution and, conversely, less changes when 
#' decreasing the alterability coefficient. The default alterability coefficients are \eqn{0.0} for the benchmarks (binding 
#' benchmarks) and \eqn{1.0} for the indicator series values (nonbinding indicator series). Important notes:
#' * With a value of \eqn{\rho = 1} (argument `rho = 1`, associated to Denton Benchmarking), only the default 
#' alterability coefficients (\eqn{0.0} for a benchmark and \eqn{1.0} for a an indicator series value) are valid. The 
#' specification of user-defined alterability coefficients variables is therefore not allowed. If such variables are 
#' specified (see arguments `var` and `with`), the function ignores them and displays a warning message in the console.
#' * Alterability coefficients \eqn{c_{s^\dagger}}{c_{s^dag}} come into play after the indicator series has been corrected for 
#' the bias, when applicable (\eqn{c_{s^\dagger}}{c_{s^dag}} is associated to \eqn{s^\dagger}{s^dag}, not \eqn{s}). This means 
#' that specifying an alterability coefficient of \eqn{0.0} for a given indicator series value **will not** result in an unchanged 
#' value after benchmarking **with bias correction** (see arguments `biasOption` and `bias`).
#'
#' Nonbinding benchmarks, when applicable, can be recovered (calculated) from the benchmarked series (see output data frame
#' `series` in section **Value**). The output `benchmarks` data frame always contains the original benchmarks provided 
#' in the input benchmarks data frame (argument `benchmarks_df`).
#'
#' ## Benchmarking Multiple Series
#' Multiple series can be benchmarked in a single [benchmarking()] call, by specifying `allCols = TRUE`, by 
#' (manually) specifying multiple variables with argument `var` (and argument `with`) or with BY-group processing 
#' (argument `by != NULL`). An important distinction is that all indicator series specified with `allCols = TRUE` 
#' or with argument `var` (and benchmarks with argument `with`) are expected to be of the same length, i.e., same 
#' set of periods and same set (number) of benchmarks. Benchmarking series of different lengths (different sets of 
#' periods) or with different sets (number) of benchmarks must be done with BY-group processing on stacked indicator 
#' series and benchmarks input data frames (see utility functions [stack_tsDF()] and [stack_bmkDF()]). Arguments 
#' `by` and `var` can be combined in order to implement BY-group processing for multiple series as illustrated by 
#' *Example 2* in the **Examples** section. While multiple variables with argument `var` (or `allCols = TRUE`) 
#' without BY-group processing (argument `by = NULL`) is slightly more efficient (faster), a BY-group approach with 
#' a single series variable is usually recommended as it is more general (works in all contexts). The latter is 
#' illustrated by *Example 3* in the **Examples** section. The BY variables specified with argument `by` appear in 
#' all three output data frames.
#'
#' ## Arguments `constant` and `negInput_option`
#' These arguments extend the usage of proportional benchmarking to a larger set of problems. Their default 
#' values correspond to the G-Series 2.0 behaviour (SAS\eqn{^\circledR}{®} PROC BENCHMARKING) for which equivalent 
#' options are not defined. Although proportional benchmarking may not necessarily be the most appropriate approach 
#' (additive benchmarking may be more appropriate) when the values of the indicator series approach 0 (unstable 
#' period-to-period ratios) or "cross the 0 line" and can therefore go from positive to negative and vice versa 
#' (confusing, difficult to interpret period-to-period ratios), these cases are not invalid mathematically 
#' speaking (i.e., the associated proportional benchmarking problem can be solved). It is strongly recommended, 
#' however, to carefully analyze and validate the resulting benchmarked data in these situations and make sure they 
#' correspond to reasonable, interpretable solutions.
#'
#' ## Treatment of Missing (`NA`) Values
#' * If a missing value appears in one of the variables of the benchmarks input data frame (other than the
#' BY variables), the observations with the missing values are dropped, a warning message is displayed and the
#' function executes.
#' * If a missing value appears in the `year` and/or `period` variables of the indicator series input data frame
#' and BY variables are specified, the corresponding BY-group is skipped, a warning message is displayed and
#' the function moves on to the next BY-group. If no BY variables are specified, a warning message is displayed
#' and no processing is done.
#' * If a missing value appears in one of the indicator series variables in the indicator series input data
#' frame and BY variables are specified, the corresponding BY-group is skipped, a warning message is displayed
#' and the function moves on to the next BY-group. If no BY variables are specified, the affected indicator
#' series is not processed, a warning message is displayed and the function moves on to the next indicator
#' series (when applicable).
#'
#'
#' @returns
#' The function returns is a list of three data frames:
#' 
#' * `series`: data frame containing the benchmarked data (primary function output). BY variables specified 
#' with argument `by` would be included in the data frame but not alterability coefficient variables 
#' specified with argument `var`.
#' 
#' * `benchmarks`: copy of the input benchmarks data frame (excluding invalid benchmarks when applicable).
#' BY variables specified with argument `by` would be included in the data frame but not alterability
#' coefficient variables specified with argument `with`.
#' 
#' * `graphTable`: data frame containing supplementary data useful for producing analytical tables and graphs
#' (see function [plot_graphTable()]). It contains the following variables in addition to the BY variables
#' specified with argument `by`:
#'   * `varSeries`: Name of the indicator series variable
#'   * `varBenchmarks`: Name of the benchmark variable
#'   * `altSeries`: Name of the user-defined indicator series alterability coefficients variable
#'   * `altSeriesValue`: Indicator series alterability coefficients
#'   * `altbenchmarks`: Name of the user-defined benchmark alterability coefficients variable
#'   * `altBenchmarksValue`: Benchmark alterability coefficients
#'   * `t`: Indicator series period identifier (1 to \eqn{T})
#'   * `m`: Benchmark coverage periods identifier (1 to \eqn{M})
#'   * `year`: Data point calendar year
#'   * `period`: Data point period (cycle) value (1 to `periodicity`)
#'   * `constant`: Temporary additive constant (argument `constant`)
#'   * `rho`: Autoregressive parameter \eqn{\rho} (argument `rho`)
#'   * `lambda`: Adjustment model parameter \eqn{\lambda} (argument `lambda`)
#'   * `bias`: Bias adjustment (default, user-defined or estimated bias according to arguments `biasOption` and `bias`)
#'   * `periodicity`: The maximum number of periods in a year (e.g. 4 for a quarterly indicator series)
#'   * `date`: Character string combining the values of variables `year` and `period`
#'   * `subAnnual`: Indicator series values
#'   * `benchmarked`: Benchmarked series values
#'   * `avgBenchmark`: Benchmark values divided by the number of coverage periods
#'   * `avgSubAnnual`: Indicator series values (variable `subAnnual`) averaged over the benchmark coverage period
#'   * `subAnnualCorrected`: Bias corrected indicator series values
#'   * `benchmarkedSubAnnualRatio`: Difference (\eqn{\lambda = 0}) or ratio (\eqn{\lambda \ne 0}{lambda != 0}) of the values 
#'   of variables `benchmarked` and `subAnnual`
#'   * `avgBenchmarkSubAnnualRatio`: Difference (\eqn{\lambda = 0}) or ratio (\eqn{\lambda \ne 0}{lambda != 0}) of the values 
#'   of variables `avgBenchmark` and `avgSubAnnual`
#'   * `growthRateSubAnnual`: Period to period difference (\eqn{\lambda = 0}) or relative difference (\eqn{\lambda \ne 0}{
#'   lambda != 0}) of the indicator series values (variable `subAnnual`)
#'   * `growthRateBenchmarked`: Period to period difference (\eqn{\lambda = 0}) or relative difference (\eqn{\lambda \ne 0}{
#'   lambda != 0}) of the benchmarked series values (variable `benchmarked`)
#'
#' Notes:
#' * The output `benchmarks` data frame always contains the original benchmarks provided in the input
#' benchmarks data frame. Modified nonbinding benchmarks, when applicable, can be recovered (calculated)
#' from the output `series` data frame.
#' * The function returns a `NULL` object if an error occurs before data processing could start. Otherwise,
#' if execution gets far enough so that data processing could start, then an incomplete object would be
#' returned in case of errors (e.g., output `series` data frame with `NA` values for the benchmarked data).
#' * The function returns "data.frame" objects that can be explicitly coerced to other types of objects with 
#' the appropriate `as*()` function (e.g., `tibble::as_tibble()` would coerce any of them to a tibble).
#'
#'
#' @references Dagum, E. B. and P. Cholette (2006). **Benchmarking, Temporal Distribution and Reconciliation Methods 
#' of Time Series**. Springer-Verlag, New York, Lecture Notes in Statistics, Vol. 186. \doi{10.1007/0-387-35439-5}
#'
#' @references Fortier, S. and B. Quenneville (2007). "Theory and Application of Benchmarking in Business Surveys".
#' **Proceedings of the Third International Conference on Establishment Surveys (ICES-III)**. Montréal, June 2007.
#'
#' @references Latendresse, E., M. Djona and S. Fortier (2007). "Benchmarking Sub-Annual Series to Annual Totals –
#' From Concepts to SAS\eqn{^\circledR}{®} Procedure and Enterprise Guide\eqn{^\circledR}{®} Custom Task". **Proceedings 
#' of the SAS\eqn{^\circledR}{®} Global Forum 2007 Conference**. Cary, NC: SAS Institute Inc.
#'
#' @references Quenneville, B., S. Fortier, Z.-G. Chen and E. Latendresse (2006). "Recent Developments in Benchmarking to 
#' Annual Totals in X-12-ARIMA and at Statistics Canada". **Proceedings of the Eurostat Conference on Seasonality, 
#' Seasonal Adjustment and Their Implications for Short-Term Analysis and Forecasting**. Luxembourg, May 2006.
#'
#' @references Quenneville, B., P. Cholette, S. Fortier and J. Bérubé (2010). "Benchmarking Sub-Annual Indicator
#' Series to Annual Control Totals (Forillon v1.04.001)". **Internal document**. Statistics Canada, Ottawa, Canada.
#'
#' @references Quenneville, B. and S. Fortier (2012). "Restoring Accounting Constraints in Time Series – Methods and
#' Software for a Statistical Agency". **Economic Time Series: Modeling and Seasonality**. Chapman & Hall, New York.
#'
#' @references Statistics Canada (2012). **Theory and Application of Benchmarking (Course code 0436)**.
#' Statistics Canada, Ottawa, Canada.
#'
#' @references Statistics Canada (2016). "The BENCHMARKING Procedure". **G-Series 2.0 User Guide**.
#' Statistics Canada, Ottawa, Canada.
#'
#'
#' @seealso [stock_benchmarking()] [plot_graphTable()] [bench_graphs] [plot_benchAdj()] [gs.gInv_MP()] [aliases]
#'
#'
#' @example misc/function_examples/benchmarking-ex.R
#'
#'
#' @export
benchmarking <- function(series_df,
                         benchmarks_df,
                         rho,
                         lambda,
                         biasOption,
                         bias = NA,
                         tolV = 0.001,
                         tolP = NA,
                         warnNegResult = TRUE,
                         tolN = -0.001,
                         var = "value",
                         with = NULL,
                         by = NULL,
                         verbose = FALSE,

                         # New in G-Series 3.0
                         constant = 0,
                         negInput_option = 0,
                         allCols = FALSE,
                         quiet = FALSE) {




  ### Internal functions ###


  # NOTE: some internal functions shared with `stock_benchmarking()` are defined in script `aaa_bench_utils.R`
  #       where environment `the` is defined and contains objects that are not (cannot be) passed as arguments
  #       to the shared internal benchmarking functions.


  # Zero values input data validation functions (proportional benchmarking)
  # => return `TRUE` if the situation corresponds to an "error" and return `FALSE` otherwise
  check_any_zero_ind <- function(s, tol = gs.tolerance, ...) {
    if (any(abs(s) <= tol)) {
      
      # Zero values are NOT allowed and some were found (this is an "error")
      TRUE
      
    } else {
      
      # Zero values not found (this is NOT an "error")
      FALSE
    }
  }
  check_nonZero_bindingBmk <- function(s, J, a, c_a, tol = gs.tolerance) {
    abs_s_lowFreq <- as.vector(J %*% abs(s))

    # Elementwise comparisons are necessary here (cannot use &&)
    if (any(abs_s_lowFreq <= tol & (abs(a) * (c_a == 0)) > tol)) {
      warning("The indicator series is zero for all periods of a nonzero binding benchmark. ",
              "This benchmark cannot be met with proportional benchmarking.\n", call. = FALSE, immediate. = TRUE)
      bk.e$warning_flag <- TRUE
    }
    
    # Zero values are allowed (whether zero values are found or not, this is NOT an "error")
    FALSE
  }


  # Benchmarking functions according to (main function) argument 'rho':
  #   - Denton benchmarking    : rho = 1
  #   - Non-Denton benchmarking: otherwise (rho < 1)
  # => No arguments: they refer to objects that exist in the parent (main function) calling environment and 
  #                  use environment pointer `main.e` (when necessary) to update objects in that environment 
  #                  (instead of using operator `<<-`)
  Denton_bench <- function() {

    # Data rescaling (standardization) of matrix C
    tmp_C <- abs(s_b)^lambda
    mean_C <- mean(tmp_C)
    if (mean_C != 0) {
      # `mean_C` should never be 0 in theory, but it could become zero in practice with 
      # "extreme negative" values for `lambda` (e.g., `lambda = -9999` )
      tmp_C <- tmp_C / mean_C
    }
    verbose_func(start_time, Sys.time(), "Data rescaling for matrix C")
    main.e$start_time <- Sys.time()

    # Build the matrices and solve
    C <- diag(tmp_C, nrow = nT)
    delta <- stats::toeplitz(c(-1, 1, rep.int(0, nT - 2)))
    delta[lower.tri(delta)] <- 0
    delta <- delta[1:(nT - 1), , drop = FALSE]
    verbose_func(start_time, Sys.time(), "C and Delta matrices construction")
    main.e$start_time <- Sys.time()

    C_inv <- gs.gInv_MP(C)
    verbose_func(start_time, Sys.time(), "Inverse C calculation")
    main.e$start_time <- Sys.time()

    M0 <- C_inv %*% t(delta) %*% delta %*% C_inv
    verbose_func(start_time, Sys.time(), "Inv(C) x Delta' x Delta x Inv(C)")
    main.e$start_time <- Sys.time()

    M1_inv <- gs.gInv_MP(rbind(cbind(M0, t(J)), cbind(J, matrix(0, nrow = M, ncol = M))))
    verbose_func(start_time, Sys.time(), "Inverse BigMat1 calculation")
    main.e$start_time <- Sys.time()

    bigMat <- M1_inv %*% rbind(cbind(M0, matrix(0, nrow = nT, ncol = M)), cbind(J, diag(1, nrow = M)))
    verbose_func(start_time, Sys.time(), "BigMat3 = Inv(BigMat1) x BigMat2")
    main.e$start_time <- Sys.time()

    theta <- as.vector(s_b + bigMat[1:nT, (nT + 1):(nT + M), drop = FALSE] %*% (a - J %*% s_b))
    verbose_func(start_time, Sys.time(), "Theta = s* + W x (a - J x s*)")
    main.e$start_time <- Sys.time()

    theta
  }
  nonDenton_bench <- function() {

    # Data rescaling (standardization) of matrices C and V_eps
    tmp_C <- sqrt(c_s) * abs(s_b)^lambda
    tmp_V <- c_a * a
    mean_C <- mean(tmp_C)
    if (mean_C != 0) {
      tmp_C <- tmp_C / mean_C
      tmp_V <- tmp_V / mean_C^2
    }
    verbose_func(start_time, Sys.time(), "Data rescaling for matrices C and V_eps")
    main.e$start_time <- Sys.time()

    # Build the matrices and solve
    C <- diag(tmp_C, nrow = nT)
    V_eps <- diag(tmp_V, nrow = M)
    # `sapply()` is safe: Omega_e will always be a "matrix" object, even when `nT = 1`
    Omega_e <- stats::toeplitz(sapply(1:nT, function(x) rho^(x - 1)))
    verbose_func(start_time, Sys.time(), "C, V_eps and Omega matrices construction")
    main.e$start_time <- Sys.time()

    V_e <- C %*% Omega_e %*% C
    verbose_func(start_time, Sys.time(), "V_e = C x Omega x C")
    main.e$start_time <- Sys.time()

    V_d <- J %*% V_e %*% t(J) + V_eps
    verbose_func(start_time, Sys.time(), "V_d = J x V_e x J' + V_eps")
    main.e$start_time <- Sys.time()

    V_d_inv <- gs.gInv_MP(V_d)
    verbose_func(start_time, Sys.time(), "Inverse V_d calculation")
    main.e$start_time <- Sys.time()

    theta <- as.vector(s_b + V_e %*% t(J) %*% V_d_inv %*% (a - J %*% s_b))
    verbose_func(start_time, Sys.time(), "Theta = s* + V_e x J' x Inv(V_d) x (a - J x s*)")
    main.e$start_time <- Sys.time()

    theta
  }




  ### Main function ###
  
  # Define a pointer to the main function environment and create a NULL object
  main.e <- environment()
  main.e$dummy <- NULL
  
  # Time stamps (when argument `verbose = TRUE`)
  start_time <- Sys.time()
  start_time0 <- start_time

  # Initialize the object to be returned by the function via `on.exit()`
  out_list <- NULL
  on.exit(return(out_list))
  try_error <- FALSE
  try_error_msg <- ""
  bk.e$warning_flag <- FALSE
  
  # Enforce the default R "error" option (`options(error = NULL)`). E.g. this Turns off traceback
  # generated by calls to the stop() function inside internal functions in R Studio.
  ini_error_opt <- getOption("error")
  on.exit(options(error = ini_error_opt), add = TRUE)
  options(error = NULL)
  
  # Validate argument `quiet` and implement the quiet setting
  quiet <- gs.validate_arg_logi(quiet)
  if (quiet) {
    quiet_msg_func <- gs.NULL_func
    quiet_lab <- ""  # won't be displayed anyway
  } else {
    quiet_msg_func <- message
    quiet_lab <- "    (*)quiet           = FALSE (default)"
  }

  # Display the function header
  quiet_msg_func("\n\n", gs.header, "\n\nbenchmarking() function:\n")


  # Initial argument validation

  # Mandatory arguments (without default values)
  ser_df_name <- deparse1(substitute(series_df))
  tmp <- nchar(ser_df_name)
  if (tmp == 0) {
    stop("Argument 'series_df' is mandatory (it must be specified).\n\n", call. = FALSE)
  } else if (tmp >= 60) {
    ser_df_name <- paste0(substr(ser_df_name, 1, 55), "<...>")
  }
  if (grepl("structure(", ser_df_name, fixed = TRUE)) {
    ser_df_name <- "<argument 'series_df'>"
  }
  series_df <- series_df
  if (!is.data.frame(series_df)) {
    stop("Argument 'series_df' is not a 'data.frame' object.\n\n", call. = FALSE)
  }
  series_df <- as.data.frame(series_df)
  row.names(series_df) <- NULL
  bmk_df_name <- deparse1(substitute(benchmarks_df))
  tmp <- nchar(bmk_df_name)
  if (tmp == 0) {
    stop("Argument 'benchmarks_df' is mandatory (it must be specified).\n\n", call. = FALSE)
  } else if (tmp >= 60) {
    bmk_df_name <- paste0(substr(bmk_df_name, 1, 55), "<...>")
  }
  if (grepl("structure(", bmk_df_name, fixed = TRUE)) {
    bmk_df_name <- "<argument 'benchmarks_df'>"
  }
  benchmarks_df <- benchmarks_df
  if (!is.data.frame(benchmarks_df)) {
    stop("Argument 'benchmarks_df' is not a 'data.frame' object.\n\n", call. = FALSE)
  }
  benchmarks_df <- as.data.frame(benchmarks_df)
  row.names(benchmarks_df) <- NULL
  if (nchar(deparse1(substitute(lambda))) == 0) {
    stop("Argument 'lambda' is mandatory (it must be specified).\n\n", call. = FALSE)
  }
  tmp <- (unlist(lambda))[1]
  if (!identical(lambda, tmp) || is.null(tmp) || !is.finite(tmp)) {
    stop("Argument 'lambda' must be a real number.\n\n", call. = FALSE)
  }
  if (nchar(deparse1(substitute(rho))) == 0) {
    stop("Argument 'rho' is mandatory (it must be specified).\n\n", call. = FALSE)
  }
  tmp <- (unlist(rho))[1]
  if (!identical(rho, tmp) || is.null(tmp) || !is.finite(tmp) || tmp < 0 || tmp > 1) {
    stop("Argument 'rho' must be a real number in the [0, 1] interval.\n\n", call. = FALSE)
  }
  if (nchar(deparse1(substitute(biasOption))) == 0) {
    if (rho == 1) {
      # bias is irrelevant for Denton benchmarking (`rho = 1`)
      biasOption <- 1L
    } else {
      stop("Argument 'biasOption' is mandatory (it must be specified).\n\n", call. = FALSE)
    }
  }
  tmp <- (unlist(biasOption))[1]
  if (!identical(biasOption, tmp) || is.null(tmp) || !(tmp %in% 1:3)) {
    stop("Argument 'biasOption' must take value 1, 2 or 3.\n\n", call. = FALSE)
  }
  biasOption <- as.integer(biasOption)

  # Optional arguments (with default values)
  tmp <- (unlist(bias))[1]
  if (is.null(tmp)) {
    bias <- NA_real_
  } else if (!identical(bias, tmp) || !is.finite(tmp) && !is.na(tmp)) {
    stop("Argument 'bias' must either be a real number or NA.\n\n", call. = FALSE)
  }
  tmp <- (unlist(tolV))[1]
  if (is.null(tmp)) {
    tolV <- NA_real_
  } else if (!identical(tolV, tmp) || !is.finite(tmp) && !is.na(tmp) || is.finite(tmp) && tmp < 0) {
    stop("Argument 'tolV' must be a nonnegative real number or NA.\n\n", call. = FALSE)
  }
  tmp <- (unlist(tolP))[1]
  if (is.null(tmp)) {
    tolP <- NA_real_
  } else if (!identical(tolP, tmp) || !is.finite(tmp) && !is.na(tmp) || is.finite(tmp) && tmp < 0) {
    stop("Argument 'tolP' must be a nonnegative real number or NA.\n\n", call. = FALSE)
  }
  warnNegResult <- gs.validate_arg_logi(warnNegResult)
  tmp <- (unlist(tolN))[1]
  if (!identical(tolN, tmp) || is.null(tmp) || !is.finite(tmp) || tmp >= 0) {
    stop("Argument 'tolN' must be a negative real number.\n\n", call. = FALSE)
  }
  verbose <- gs.validate_arg_logi(verbose)
  tmp <- (unlist(constant))[1]
  if (!identical(constant, tmp) || is.null(tmp) || !is.finite(tmp)) {
    stop("Argument 'constant' must be a real number.\n\n", call. = FALSE)
  }
  tmp <- (unlist(negInput_option))[1]
  if (!identical(negInput_option, tmp) || is.null(tmp) || !(tmp %in% 0:2)) {
    stop("Argument 'negInput_option' must take value 0, 1 or 2.\n\n", call. = FALSE)
  }
  negInput_option <- as.integer(negInput_option)
  allCols <- gs.validate_arg_logi(allCols)
  # Argument `quiet` was already validated

  # Implement the verbose setting
  if (verbose && !quiet) {
    verbose_func <- gs.display_difftime
    verbose_lab <- "    verbose            = TRUE"
  } else {
    verbose_func <- gs.NULL_func
    verbose_lab <- "    verbose            = FALSE (default)"
  }
  
  # Initialize the final warning/error message flag for multiple series processing
  final_msg_flag <- FALSE
  try_stop_func <- gs.NULL_func
  

  # Initialize the list of "non-data columns" (date and by-group columns)
  info_cols_serDF <- c("year", "period")
  info_cols_bmkDF <- c("startYear", "startPeriod", "endYear", "endPeriod")

  all_cols_serDF <- names(series_df)
  gs.validate_cols(info_cols_serDF, all_cols_serDF, ser_df_name)
  if (any(!is.finite(as.matrix(series_df[info_cols_serDF])))) {
    stop("Indicator series date columns (\"year\" and \"period\") contain invalid or NA values.\n\n", call. = FALSE)
  }

  all_cols_bmkDF <- names(benchmarks_df)
  gs.validate_cols(info_cols_bmkDF, all_cols_bmkDF, bmk_df_name)
  if (any(!is.finite(as.matrix(benchmarks_df[info_cols_bmkDF])))) {
    stop("Benchmarks date columns (\"startYear\", \"startPeriod\", \"endYear\" and \"endPeriod\") contain invalid or NA values.\n\n",
         call. = FALSE)
  }

  # Initialize the list of "series data columns" from both input files
  data_cols_serDF <- setdiff(all_cols_serDF, info_cols_serDF)
  data_cols_bmkDF <- setdiff(all_cols_bmkDF, info_cols_bmkDF)

  # Validation of argument 'by' ("non-data columns" as well)
  by <- unique(gs.cleanup_col_list(by))
  n_byVars <- length(by)
  if (n_byVars > 0) {
    by_lab <- paste0("    by                 = ", paste(by, collapse = " "))
    gs.validate_cols(by, data_cols_serDF, ser_df_name, source_str = "argument 'by'")
    gs.validate_cols(by, data_cols_bmkDF, bmk_df_name, source_str = "argument 'by'")
    by_grps <- unique(series_df[by])
    n_byGrps <- nrow(by_grps)

    # Build the by-group expression (for message display)
    if (n_byGrps > 0) {
      by_grps$PB._expr_ <- paste0(by[1], "=", by_grps[, by[1]])
      for (ii in seq_along(by)[-1]) {
        by_grps$PB._expr_ <- paste(by_grps$PB._expr_, paste0(by[ii], "=", by_grps[, by[ii]]), sep = " & ")
      }
    }
    
    # Add the by variables to the list of "non-data columns"
    info_cols_serDF <- c(by, info_cols_serDF)
    info_cols_bmkDF <- c(by, info_cols_bmkDF)
    
    # Remove by variables from the list of "series data columns"
    data_cols_serDF <- setdiff(data_cols_serDF, by)
    data_cols_bmkDF <- setdiff(data_cols_bmkDF, by)
    
    # Set message display info
    if (n_byGrps > 1) {
      byGrp_ini_func <- bk.byGrp_ini
      byGrp_msg_func <- message
      final_msg_flag <- TRUE
      try_stop_func <- gs.try_stop
    } else {
      byGrp_ini_func <- bk.noByGrp_ini
      byGrp_msg_func <- gs.NULL_func
    }

  # No by-groups
  } else {
    by <- NULL
    by_lab <- "    by                 = NULL (default)"
    by_grps <- data.frame(PB._expr_ = "")
    n_byGrps <- 1
    byGrp_ini_func <- bk.noByGrp_ini
    byGrp_msg_func <- gs.NULL_func
  }

  # Ignore arguments 'var' and 'with' and process all data columns of the input indicator series
  # data frame, expecting corresponding columns (same names) in the input benchmarks data frame
  if (allCols) {
    var_lab <- "    var                (ignored)"
    with_lab <- "    with               (ignored)"
    allCols_lab <- "    (*)allCols         = TRUE"
    var <- data_cols_serDF
    with <- var
    n_vars <- length(var)
    alter_ser <- rep.int("", n_vars)
    default_alter_ser_id <- 1:n_vars
    user_alter_ser_id <- integer(0)  # value returned by which() on an "all FALSE" logical vector
    alter_bmk <- alter_ser
    default_alter_bmk_id <- 1:n_vars
    user_alter_bmk_id <- integer(0)
    gs.validate_cols(with, data_cols_bmkDF, bmk_df_name)

    # Process arguments 'var' and 'with'
  } else {
    allCols_lab <- "    (*)allCols         = FALSE (default)"

    # Validation of argument 'var'
    var <- gs.cleanup_col_list(var)
    var_lab <- paste0("    var                = ", paste(var, collapse = " "))
    if (var_lab == "    var                = value") {
      var_lab <- "    var                = value (default)"
    }
    tmp <- gs.split_str("\\/", var)
    var <- tmp[[1]]
    n_vars <- length(var)
    len <- length(tmp)
    if (len == 2) {
      alter_ser <- tmp[[2]]
      default_alter_ser_id <- which(alter_ser == "")
      user_alter_ser_id <- setdiff(1:n_vars, default_alter_ser_id)
    } else {
      if (len > 2) {
        stop("Invalid specification of alterability coefficients in argument 'var'.\n\n", call. = FALSE)
      }
      alter_ser <- rep.int("", n_vars)
      default_alter_ser_id <- 1:n_vars
      user_alter_ser_id <- integer(0)
    }
    prob_cols <- intersect(var, by)
    if (length(prob_cols) > 0) {
      stop("The following columns are listed with arguments 'var' and 'by' (these arguments are mutually exclusive):",
           paste0("\n  ", prob_cols, collapse = ""), "\n\n", call. = FALSE)
    }
    gs.validate_cols(c(var, alter_ser[user_alter_ser_id]), data_cols_serDF, ser_df_name, source_str = "argument 'var'")

    # Validation of argument 'with'
    with <- gs.cleanup_col_list(with)
    if (length(with) == 0) {
      with_lab <- "    with               = NULL (default)"
      with <- var
      alter_bmk <- rep.int("", n_vars)
      default_alter_bmk_id <- 1:n_vars
      user_alter_bmk_id <- integer(0)
    } else {
      with_lab <- paste0("    with               = ", paste(with, collapse = " "))
      tmp <- gs.split_str("\\/", with)
      with <- tmp[[1]]
      len <- length(tmp)
      if (len == 2) {
        alter_bmk <- tmp[[2]]
        default_alter_bmk_id <- which(alter_bmk == "")
        user_alter_bmk_id <- setdiff(1:n_vars, default_alter_bmk_id)
      } else {
        if (len > 2) {
          stop("Invalid specification of alterability coefficients in argument 'with'.\n\n", call. = FALSE)
        }
        alter_bmk <- rep.int("", n_vars)
        default_alter_bmk_id <- 1:n_vars
        user_alter_bmk_id <- integer(0)
      }
    }
    prob_cols <- intersect(with, by)
    if (length(prob_cols) > 0) {
      stop("The following columns are listed with arguments 'with' and 'by' (these arguments are mutually exclusive):",
           paste0("\n  ", prob_cols, collapse = ""), "\n\n", call. = FALSE)
    }
    gs.validate_cols(c(with, alter_bmk[user_alter_bmk_id]), data_cols_bmkDF, bmk_df_name, source_str = "argument 'with'")
    if (n_vars != length(with)) {
      stop("Arguments 'var' and 'with' do not contain the same number of elements.\n\n", call. = FALSE)
    }
  }
  if (n_vars == 0) {
    stop("Argument 'var' results in 0 series to be benchmarked.\n\n", call. = FALSE)
  } else if (n_vars == 1) {
    var_msg_func <- gs.NULL_func
  } else {
    var_msg_func <- message
    final_msg_flag <- TRUE
    try_stop_func <- gs.try_stop
  }
  
  # Shrink the input data frames with only the necessary columns
  series_df <- series_df[unique(c(info_cols_serDF, var, alter_ser[user_alter_ser_id]))]
  benchmarks_df <- benchmarks_df[unique(c(info_cols_bmkDF, with, alter_bmk[user_alter_bmk_id]))]

  # Add the default alter coefs to the input data frames and specify them when relevant
  series_df$PB._default_alter_ <- rep.int(1, nrow(series_df))
  benchmarks_df$PB._default_alter_ <- rep.int(0, nrow(benchmarks_df))
  actual_alter_ser <- alter_ser
  actual_alter_bmk <- alter_bmk
  alter_ser[default_alter_ser_id] <- "PB._default_alter_"
  alter_bmk[default_alter_bmk_id] <- "PB._default_alter_"
  alter_ser_check_func_str <- rep.int("gs.FALSE_func", n_vars)
  alter_bmk_check_func_str <- rep.int("gs.FALSE_func", n_vars)
  alter_ser_check_func_str[user_alter_ser_id] <- "gs.check_alter"
  alter_bmk_check_func_str[user_alter_bmk_id] <- "gs.check_alter"

  # Initialize the output data frames (as NULL objects for now)
  out_series_df <- NULL
  out_benchmarks_df <- NULL
  out_graphTable_df <- NULL

  # Set parameters and function names according to the benchmarking model (additive benchmarking
  # when 'lambda == 0' and proportional benchmarking otherwise):
  #   - negative values verification functions
  #   - bias calculation
  #   - ratio and growth rate calculation functions
  if (lambda == 0) {
    zeros_verif_func <- gs.FALSE_func
    neg_verif_func <- gs.FALSE_func
    default_bias <- 0
    bias_calc_func <- bk.calc_add_bias
    bias_apply_func <- bk.apply_add_bias
    ratio_func <- gs.calc_diff
    growthRate_func <- gs.calc_firstDiff
  } else {
    if (rho == 1 || lambda < 0) {
      zeros_verif_func <- check_any_zero_ind
    } else {
      zeros_verif_func <- check_nonZero_bindingBmk
    }
    if (negInput_option == 0) {
      neg_verif_func <- bk.check_neg_err
    } else if (negInput_option == 1) {
      neg_verif_func <- bk.check_neg_warn
    } else {
      neg_verif_func <- gs.FALSE_func
    }
    default_bias <- 1
    bias_calc_func <- bk.calc_mult_bias
    bias_apply_func <- bk.apply_mult_bias
    ratio_func <- gs.calc_ratio
    growthRate_func <- gs.calc_relFirstDiff
    if (!is.na(bias) && bias < 0) {
      stop("Argument 'bias' must be positive with proportional benchmarking.\n\n", call. = FALSE)
    }
  }

  # Set the bias option function according to argument 'biasOption'
  if (biasOption == 3) {
    biasOption_func <- bk.apply_biasOption3
    biasOption_lab <- "    biasOption         = 3 (Calculate bias, use calculated bias)"
    bias_parm <- NA_real_
    bias_str <- NA_character_
    bias_lab <- "    bias               (ignored)"
  } else {
    if (biasOption == 1) {
      biasOption_func <- bk.apply_biasOption1
      biasOption_lab <- "    biasOption         = 1 (Use user-defined or default bias)"
    } else if (biasOption == 2) {
      biasOption_func <- bk.apply_biasOption2
      biasOption_lab <- "    biasOption         = 2 (Calculate bias, but use user-defined or default bias)"
    }
    if (is.na(bias)) {
      bias_parm <- default_bias
      bias_str <- "default"
      bias_lab <- "    bias               = NA (default)"
    } else {
      bias_parm <- bias
      bias_str <- "user-defined"
      bias_lab <- paste0("    bias               = ", format(bias))
    }
  }
  bk.e$actual_bias <- NULL

  # Set parameters and functions according to argument 'rho' (Denton or Non-Denton benchmarking):
  #   - minimum required number of periods in the indicator series
  #   - benchmarking function
  if (rho == 1) {
    min_nT <- 2
    bench_func <- Denton_bench
    
    # Impose no bias correction (same as `biasOption = 1` and `bias = NA`)
    biasOption_func <- bk.apply_biasOption1
    biasOption_lab <- "    biasOption         (ignored)"
    bias_parm <- default_bias
    bias_str <- "default"
    bias_lab <- "    bias               (ignored)"
    
    # Check for user (non-default) alter coefs (not allowed when rho=1)
    if (length(user_alter_ser_id) > 0 || length(user_alter_bmk_id) > 0) {
      warning("Alterability coefficients are not available when rho=1.0. ",
              "The default values will be used.\n", call. = FALSE, immediate. = TRUE)
      bk.e$warning_flag <- TRUE
      if (length(user_alter_ser_id) > 0) {
        var_lab <- paste0("    var                = ", paste(var, collapse = " "))
        if (var_lab == "    var                = value") {
          var_lab <- "    var                = value (default)"
        }
      }
      if (length(user_alter_bmk_id) > 0) {
        with_lab <- paste0("    with               = ", paste(with, collapse = " "))
      }
      alter_ser[user_alter_ser_id] <- "PB._default_alter_"
      alter_bmk[user_alter_bmk_id] <- "PB._default_alter_"
      actual_alter_ser[user_alter_ser_id] <- ""
      actual_alter_bmk[user_alter_bmk_id] <- ""
      alter_ser_check_func_str[user_alter_ser_id] <- "gs.FALSE_func"
      alter_bmk_check_func_str[user_alter_bmk_id] <- "gs.FALSE_func"
      user_alter_ser_id <- integer(0)
      user_alter_bmk_id <- integer(0)
      default_alter_ser_id <- 1:n_vars
      default_alter_bmk_id <- 1:n_vars
    }
  } else {
    min_nT <- 1
    bench_func <- nonDenton_bench
  }

  # Results validation function and binding benchmarks tolerance
  if (warnNegResult) {
    neg_res_func <- gs.check_neg
    warnNegResult_lab <- "    warnNegResult      = TRUE (default)"
  } else {
    neg_res_func <- gs.FALSE_func
    warnNegResult_lab <- "    warnNegResult      = FALSE"
  }
  if (!is.na(tolV)) {
    if (!is.na(tolP)) {
      stop("Arguments 'tolV' and 'tolP' cannot be both specified (one must be NA).\n\n", call. = FALSE)
    }
    binding_bmk_valid_func <- bk.binding_bmk_diff_valid
    tol_parm <- tolV
    tol_lab <- paste0("    tolV               = ", format(tolV))
    if (abs(tolV - 0.001) < gs.tolerance) {
      tol_lab <- paste0(tol_lab, " (default)")
    }
  } else {
    if (is.na(tolP)) {
      stop("Arguments 'tolV' and 'tolP' cannot be both NA (one must be specified).\n\n", call. = FALSE)
    }
    binding_bmk_valid_func <- bk.binding_bmk_relDiff_valid
    tol_parm <- tolP
    tol_lab <- paste0("    tolP               = ", format(tolP))
  }


  # Display the function call (argument values)
  quiet_msg_func("    series_df          = ", ser_df_name)
  quiet_msg_func("    benchmarks_df      = ", bmk_df_name)
  quiet_msg_func("    rho                = ", format(rho))
  quiet_msg_func("    lambda             = ", format(lambda))
  quiet_msg_func(biasOption_lab)
  quiet_msg_func(bias_lab)
  quiet_msg_func(tol_lab)
  quiet_msg_func(warnNegResult_lab)
  lab <- paste0("    tolN               = ", format(tolN))
  if (abs(tolN - (-0.001)) < gs.tolerance) {
    lab <- paste0(lab, " (default)")
  }
  quiet_msg_func(lab)
  quiet_msg_func(var_lab)
  quiet_msg_func(with_lab)
  quiet_msg_func(by_lab)
  quiet_msg_func(verbose_lab, "\n")
  lab <- paste0("    (*)constant        = ", format(constant))
  if (constant == 0) {
    lab <- paste0(lab, " (default)")
  }
  quiet_msg_func(lab)
  lab <- paste0("    (*)negInput_option = ", format(negInput_option))
  if (negInput_option == 0) {
    lab <- paste0(lab, " (default)")
  }
  quiet_msg_func(lab)
  quiet_msg_func(allCols_lab)
  quiet_msg_func(quiet_lab, "\n")
  quiet_msg_func("    (*) indicates new arguments in G-Series 3.0\n")

  verbose_func(start_time, Sys.time(), "Set up phase")
  start_time <- Sys.time()

  
  # Return the input series data frames
  if (nrow(series_df) == 0 || nrow(benchmarks_df) == 0) {
    out_series_df <- series_df[intersect(all_cols_serDF, names(series_df))]
    out_benchmarks_df <- benchmarks_df[intersect(all_cols_bmkDF, names(benchmarks_df))]

  # Attempt to benchmark the data
  } else {
    
    # Enforce integer period id values
    cols_vec <- setdiff(info_cols_serDF, by)
    series_df[cols_vec] <- as.integer(as.matrix(series_df[cols_vec]))
    cols_vec <- setdiff(info_cols_bmkDF, by)
    benchmarks_df[cols_vec] <- as.integer(as.matrix(benchmarks_df[cols_vec]))
    
    
    # Process each by-group
    
    on.exit(rlang::env_unbind(bk.e, c("ser_df_byGrp", "bmk_df_byGrp")), add = TRUE)
    for (ii in 1:n_byGrps) {
      
      # By-group initialization/setup
      bk.e$ser_df_byGrp <- NULL
      bk.e$bmk_df_byGrp <- NULL
      byGrp_ini_func(series_df, benchmarks_df, by_grps, ii, by)
      
      msg_str <- paste0("\nBenchmarking by-group ", ii, " (", by_grps$PB._expr_[ii], ")")
      byGrp_msg_func(msg_str)
      byGrp_msg_func(strrep("=", nchar(msg_str) - 1), "\n")
      
      # Reject rows with invalid benchmark data (for any of the benchmarks)
      M_ini <- nrow(bk.e$bmk_df_byGrp)
      if (nrow(bk.e$bmk_df_byGrp) > 1) {
        prob_id <- which(apply(!apply(bk.e$bmk_df_byGrp[with], 2, is.finite), 1, any))
      } else {
        prob_id <- which(any(!apply(bk.e$bmk_df_byGrp[with], 2, is.finite)))
      }
      if (length(prob_id) > 0) {
        warning("Rows from the benchmarks data frame were dropped because of invalid or NA values.\n",
                call. = FALSE, immediate. = TRUE)
        bk.e$warning_flag <- TRUE
        bk.e$bmk_df_byGrp <- bk.e$bmk_df_byGrp[-prob_id, , drop = FALSE]
        M <- nrow(bk.e$bmk_df_byGrp)
      } else {
        M <- M_ini
      }
      
      # Initialize the by-group output data frames
      out_bmk_df_byGrp <- bk.e$bmk_df_byGrp[c(info_cols_bmkDF, with)]
      out_ser_df_byGrp <- bk.e$ser_df_byGrp[c(info_cols_serDF, var)]
      out_ser_df_byGrp[var] <- NA_real_
      
      nT <- nrow(bk.e$ser_df_byGrp)
      if (nT < min_nT) {
        try_error_msg <- paste0("The minimum number of periods (", min_nT, ") for the indicator series ", 
                                "is not met.\n\n")
        try_stop_func(try_error_msg)
        try_error <- TRUE
        
      } else if (M == 0) {
        try_error_msg <- "A minimum of 1 benchmark is required.\n\n"
        try_stop_func(try_error_msg)
        try_error <- TRUE
        
      } else {
        
        # Vectors for periods and benchmarks coverage validation
        periodicity <- max(bk.e$ser_df_byGrp$period)
        periods <- paste(bk.e$ser_df_byGrp$year, bk.e$ser_df_byGrp$period, sep = "-")
        bmk_start <- paste(bk.e$bmk_df_byGrp$startYear, bk.e$bmk_df_byGrp$startPeriod, sep = "-")
        bmk_end <- paste(bk.e$bmk_df_byGrp$endYear, bk.e$bmk_df_byGrp$endPeriod, sep = "-")
        
        # Validate indicator series periods (period-to-period gap != 1 / periodicity)
        time_val <- bk.e$ser_df_byGrp$year + (bk.e$ser_df_byGrp$period - 1) / periodicity
        if (nT > 1) {
          # `sapply()` is safe: `which()` always returns a "vector" object, even when `nT = 2`
          prob_id <- which(c(FALSE,
                             sapply(2:nT,
                                    function(x) {
                                      abs(time_val[x] - time_val[x - 1] - 1 / periodicity) > gs.tolerance
                                    })))
        } else {
          prob_id <- integer(0)
        }
        if (length(prob_id) > 0) {
          try_error_msg <- paste0("Non-contiguous periods found in the indicator series: ", 
                                  paste0("\n  ", periods[prob_id - 1], " - ", periods[prob_id], 
                                         collapse = ""), 
                                  "\n\n")
          try_stop_func(try_error_msg)
          try_error <- TRUE
          
        } else {
          
          # Build benchmarks to periods mapping matrix of dimension 2 X M:
          #   - Row 1: benchmark starting period ids (positions in vector `periods`)
          #   - Row 2: benchmark ending period ids (positions in vector `periods`)
          mapping <- rbind(match(bmk_start, periods),
                           match(bmk_end, periods))
          
          # Validate the benchmark coverage (i.e. not inside the set of indicator periods)
          prob_id <- which(apply(is.na(mapping), 2, any))
          if (length(prob_id) > 0) {
            try_error_msg <- paste0("Benchmark coverage not fully inside the indicator series span: ",
                                    paste0("\n  [", bmk_start[prob_id], ", ", bmk_end[prob_id], "]", 
                                           collapse = ""), 
                                    "\n\n")
            try_stop_func(try_error_msg)
            try_error <- TRUE
            
          } else {
            
            # Display a message about the number of observations in the benchmarks and series data frames
            # Note: all indicator series observations are valid at this point, making the SERIES
            #       message not quite relevant... but that's what SAS G-Series PROC BENCHMARKING displays
            #       (there might be cases where the 2 numbers differ, but I don't see it...)
            
            # Widths without commas
            width_nT <- floor(log10(nT)) + 1
            width_M_ini <- floor(log10(M_ini)) + 1
            width_M <- floor(log10(M)) + 1
            
            # Adjust the widths of the benchmarks info considering the commas (for `format(..., big.mark = ",")`)
            # => add the difference in the number of commas between the series obs. number (`nT`) and each of the 
            #    two (initial and final) benchmarks obs. numbers (`M_ini` and `M`)
            # => it's (reasonably) assumed here that there will always be more observations in the series data 
            #    frame than in the benchmarks data frame (i.e., `nT` >= `M_ini` >= `M`)
            if (width_M_ini < width_nT) {
              width_M_ini <- width_nT + floor((width_nT - 1) / 3) - floor((width_M_ini - 1) / 3)
            }
            if (width_M < width_nT) {
              width_M <- width_nT + floor((width_nT - 1) / 3) - floor((width_M - 1) / 3)
            }
            quiet_msg_func("Number of observations in the BENCHMARKS data frame .............: ",
                           format(M_ini, scientific = FALSE, big.mark = ",", width = width_M_ini, justify = "right"))
            quiet_msg_func("Number of valid observations in the BENCHMARKS data frame .......: ",
                           format(M, scientific = FALSE, big.mark = ",", width = width_M, justify = "right"), "\n")
            quiet_msg_func("Number of observations in the SERIES data frame .................: ",
                           format(nT, scientific = FALSE, big.mark = ",", width = width_nT, justify = "right"))
            quiet_msg_func("Number of valid observations in the SERIES data frame ...........: ",
                           format(nT, scientific = FALSE, big.mark = ",", width = width_nT, justify = "right"), "\n")
            
            verbose_func(start_time, Sys.time(), "Data validation 1")
            start_time <- Sys.time()
            start_time1 <- start_time
            
            # Process each series
            for (jj in 1:n_vars) {
              
              msg_str <- paste0("\nBenchmarking indicator series [", var[jj], "] with benchmarks [", with[jj], "]")
              var_msg_func(msg_str)
              var_msg_func(strrep("-", nchar(msg_str) - 1), "\n")
              
              # Series/benchmarks data and alterablity coefficients validation
              if (any(!is.finite(bk.e$ser_df_byGrp[[var[jj]]]))) {
                warning("The indicator series contains invalid or NA values. It will not be processed.\n", call. = FALSE,
                        immediate. = TRUE)
                bk.e$warning_flag <- TRUE
                
              } else if (do.call(alter_ser_check_func_str[jj], list(bk.e$ser_df_byGrp[[alter_ser[jj]]]))) {
                try_error_msg <- "Invalid indicator series alterability coefficients (must be nonnegative numbers).\n\n"
                try_stop_func(try_error_msg)
                try_error <- TRUE
                
              } else if (do.call(alter_bmk_check_func_str[jj], list(bk.e$bmk_df_byGrp[[alter_bmk[jj]]]))) {
                try_error_msg <- "Invalid benchmarks alterability coefficients (must be nonnegative numbers).\n\n"
                try_stop_func(try_error_msg)
                try_error <- TRUE
                
                # Process (benchmark) the series
              } else {
                
                verbose_func(start_time, Sys.time(), "Data validation 2")
                start_time <- Sys.time()
                
                # Build the J matrix (temporal sum operator)
                J <- matrix(0, nrow = M, ncol = nT)
                for (kk in 1:M) {
                  J[kk, mapping[1, kk]:mapping[2, kk]] <- 1
                }
                
                # Build the elementary vectors and matrices
                #   => the temporary constant is added here
                s <- bk.e$ser_df_byGrp[[var[jj]]] + constant
                c_s <- bk.e$ser_df_byGrp[[alter_ser[jj]]]
                s_lowFreq <- as.vector(J %*% s)
                nT_lowFreq <- as.vector(J %*% rep.int(1, nT))
                nT_bmk <- sum(nT_lowFreq)
                a <- bk.e$bmk_df_byGrp[[with[jj]]] + constant * nT_lowFreq
                c_a <- bk.e$bmk_df_byGrp[[alter_bmk[jj]]]
                
                verbose_func(start_time, Sys.time(), "J matrix construction")
                start_time <- Sys.time()
                
                # Additional series/benchmarks data validation for proportional benchmarking (lambda != 0)
                #   - `rho = 1` or `lambda < 0`: zero indicator series values (infeasible proportional benchmarking problem)
                #     `rho < 1`: nonzero binding benchmarks associated to "all zero" indicator series values
                #                (infeasible proportional benchmarking problem)
                #   - negative benchmarks or indicator series values (according to argument `negInput_option`)
                if (zeros_verif_func(s, J, a, c_a, tol = gs.tolerance)) {
                  try_error_msg <- paste0("The indicator series has zero values. This is not permitted for proportional ",
                                          "benchmarking when `rho = 1` or `lambda < 0`.\n\n")
                  try_stop_func(try_error_msg)
                  try_error <- TRUE
                  
                } else if (neg_verif_func(a, tol = 0, data_str = "benchmarks")) {
                  try_error_msg <- paste0("Negative values found in the benchmarks. This is not permitted for ", 
                                          "proportional benchmarking.\n\n")
                  try_stop_func(try_error_msg)
                  try_error <- TRUE
                  
                } else if (neg_verif_func(s, tol = 0, data_str = "indicator series")) {
                  try_error_msg <- paste0("Negative values found in the indicator series. This is not permitted for ", 
                                          "proportional benchmarking.\n\n")
                  try_stop_func(try_error_msg)
                  try_error <- TRUE
                  
                } else {
                  
                  # Bias correction
                  s_b <- biasOption_func(quiet_msg_func,                                      # message function
                                         bias_calc_func, s_lowFreq, a, nT_bmk, gs.tolerance,  # bias calculation arguments
                                         bias_apply_func, s, bias_parm,                       # bias application arguments
                                         bias_str)
                  
                  verbose_func(start_time, Sys.time(), "Bias adjustment (s*)")
                  start_time <- Sys.time()
                  
                  # Solve the benchmarking problem (unless there's no discrepancies)
                  if (any(abs(as.vector(J %*% s_b) - a) > gs.min_tolerance)) {
                    theta <- bench_func()
                  } else {
                    theta <- s_b
                  }
                  
                  # Just for safety: solution could contain `NaN` values in very extreme cases (matrix inversion problems)
                  if (any(!is.finite(theta))) {
                    warning("Unable to solve the benchmarking problem.\n", call. = FALSE, immediate. = TRUE)
                    bk.e$warning_flag <- TRUE
                    out_ser_df_byGrp[[var[jj]]] <- NA_real_
                    
                  } else {
                    
                    
                    # Cumulate the graphTable output data frame info
                    #   => the graphTable includes more periods than the indicator series (`n_obs_GT > nT`)
                    #      in the case of overlapping benchmarks

                    # Map benchmark level info (vectors of length `M`) to period level info (vectors of length
                    # `n >= nT`, with `n = nT` for non-overlapping benchmarks and `n > nT` for overlapping benchmarks)
                    # according to matrix `mapping` of dimension 2 X `M`:
                    #   - Row 1: benchmark starting period ids
                    #   - Row 2: benchmark ending period ids
                    #
                    # The resulting `bmk_info` data frame has `n >= nT` observations and 5 variables:
                    #   - `t`: period id
                    #   - `m`: benchmark id (`NA` for periods not associated to any benchmark)
                    #   - `avg_s`: average indicator series value  (`NA` for periods not associated to any benchmark)
                    #   - `avg_a`: average benchmark value (`NA` for periods not associated to any benchmark)
                    #   - `c_a`: benchmark alter coef (`NA` for periods not associated to any benchmark)
                    temp_df <- data.frame(t = rep.int(NA_integer_, nT_bmk),
                                          m = rep.int(NA_integer_, nT_bmk))
                    last_row <- 0
                    for (ii in 1:M) {
                      per_vec <- mapping[1, ii]:mapping[2, ii]
                      n_per <- length(per_vec)
                      temp_df[(last_row + 1):(last_row + n_per), ] <- c(per_vec,             # period ids (`t`)
                                                                        rep.int(ii, n_per))  # benchmark ids (`m`)
                      last_row <- last_row + n_per
                    }
                    bmk_info <- merge(data.frame(t = 1:nT), # period ids
                                      merge(temp_df,
                                            # Benchmark level info
                                            data.frame(m = 1:M,
                                                       avg_s = s_lowFreq / nT_lowFreq,
                                                       avg_a = a / nT_lowFreq,
                                                       c_a = c_a),
                                            by = "m"), 
                                      by = "t", 
                                      all.x = TRUE)
                    bmk_info <- bmk_info[order(bmk_info$t, bmk_info$m), ]

                    # Generate the graphTable data frame info
                    n_obs_GT <- length(bmk_info$t)
                    out_GT_df_var <- data.frame(varSeries = rep.int(var[jj], n_obs_GT),
                                                varBenchmarks = rep.int(with[jj], n_obs_GT),
                                                altSeries = rep.int(actual_alter_ser[jj], n_obs_GT),
                                                altSeriesValue = c_s[bmk_info$t],
                                                altbenchmarks = rep.int(actual_alter_bmk[jj], n_obs_GT),
                                                altBenchmarksValue = bmk_info$c_a,
                                                t = bmk_info$t,
                                                m = bmk_info$m,
                                                year = bk.e$ser_df_byGrp$year[bmk_info$t],
                                                period = bk.e$ser_df_byGrp$period[bmk_info$t],
                                                constant = rep.int(constant, n_obs_GT),
                                                rho = rep.int(rho, n_obs_GT),
                                                lambda = rep.int(lambda, n_obs_GT),
                                                bias = rep.int(bk.e$actual_bias, n_obs_GT),
                                                periodicity = rep.int(periodicity, n_obs_GT),
                                                date = paste(bk.e$ser_df_byGrp$year[bmk_info$t], 
                                                             sprintf("%06d", bk.e$ser_df_byGrp$period[bmk_info$t]), sep = "-"),
                                                subAnnual = s[bmk_info$t],
                                                benchmarked = theta[bmk_info$t],
                                                avgBenchmark = bmk_info$avg_a,
                                                avgSubAnnual = bmk_info$avg_s,
                                                subAnnualCorrected = s_b[bmk_info$t],
                                                benchmarkedSubAnnualRatio = ratio_func(theta[bmk_info$t], s[bmk_info$t]),
                                                avgBenchmarkSubAnnualRatio = ratio_func(bmk_info$avg_a, bmk_info$avg_s),
                                                growthRateSubAnnual = growthRate_func(s)[bmk_info$t],
                                                growthRateBenchmarked = growthRate_func(theta)[bmk_info$t],
                                                stringsAsFactors = FALSE)
                    
                    # Cumulate the graphTable data frame info
                    out_graphTable_df <- rbind(out_graphTable_df, cbind(bk.e$ser_df_byGrp[bmk_info$t, by, drop = FALSE],
                                                                        out_GT_df_var))
                    
                    verbose_func(start_time, Sys.time(), "Output grahTable creation")
                    start_time <- Sys.time()
                    
                    # Remove the temporary constant
                    out_ser_df_byGrp[[var[jj]]] <- theta - constant
                    
                    # Results validation
                    if (neg_res_func(out_ser_df_byGrp[[var[jj]]], tol = -tolN)) {
                      warning("The benchmarked series contains negative values (threshold = ", format(tolN), ").\n",
                              call. = FALSE, immediate. = TRUE)
                      bk.e$warning_flag <- TRUE
                    }
                    binding_bmk_valid_func(bk.e$bmk_df_byGrp[[with[jj]]], as.vector(J %*% out_ser_df_byGrp[[var[jj]]]), c_a,
                                           bmk_start, bmk_end, tol_parm, zero_tol = 0)
                    
                    verbose_func(start_time, Sys.time(), "Results validation")
                    start_time <- Sys.time()
                  }
                  
                  verbose_func(start_time1, Sys.time(), "Benchmarking step total execution time")
                  start_time1 <- Sys.time()
                }
              }
            }
          }
        }
      }
      
      # Cumulate the output series and benchmarks data frame info
      out_series_df <- rbind(out_series_df, out_ser_df_byGrp)
      out_benchmarks_df <- rbind(out_benchmarks_df, out_bmk_df_byGrp)
    }
  }

  
  # Create the output list
  row.names(out_series_df) <- NULL
  row.names(out_benchmarks_df) <- NULL
  row.names(out_graphTable_df) <- NULL
  out_list <- list(series = out_series_df, 
                   benchmarks = out_benchmarks_df, 
                   graphTable = out_graphTable_df)
  
  verbose_func(start_time, Sys.time(), "Wrap up phase")
  verbose_func(start_time0, Sys.time(), "Total execution time")


  # Display a final warning/error message for multiple series processing
  if (final_msg_flag) {
    if (bk.e$warning_flag) {
      warning("Warnings were generated during processing. See preceeding warning message(s) for details.\n",
              call. = FALSE, immediate. = TRUE)
    }
    # Non-muted error message (for proper condition catching by users of the function)
    if (try_error) {
      stop("Problems were encontered during processing. See preceeding error message(s) for details.\n\n",
           call. = FALSE)
    }
    
  # Display the error message for single-series processing
  } else if (try_error) {
    stop(try_error_msg, call. = FALSE)
  }

  # Output object returned via function `on.exit()`
}

#' Restore temporal constraints for stock series
#'
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/stock_benchmarking.html}})}
#' 
#' Function specifically aimed at benchmarking stock series where the benchmarks are anchor points covering a
#' single period of the indicator series. Benchmarks covering more than one period of the indicator series
#' cannot be used with this function. Function [benchmarking()] should be used instead to benchmark
#' non-stock series (flows).
#'
#' Several stock series can be benchmarked in a single function call.
#'
#' Note that functions [stock_benchmarking()] and [benchmarking()] mainly share the same arguments and
#' return the same type of object. Differences are listed below:
#' * Argument `verbose` is not defined for [stock_benchmarking()].
#' * Extra arguments defined for [stock_benchmarking()]:
#'   * `low_freq_periodicity`
#'   * `n_low_freq_proj`
#'   * `proj_knots_rho_bd`
#' * The list returned by [stock_benchmarking()] contains an extra (fourth) data frame:
#'   * `splineKnots`
#'
#'  See section **Details** for more information on the similarities and differences of functions
#'  [stock_benchmarking()] and [benchmarking()].
#'
#'  *A direct equivalent of [stock_benchmarking()] does not exist in SAS\eqn{^\circledR}{®} G-Series 2.0.*
#'
#'
#' @inheritParams benchmarking
#'
#' @param low_freq_periodicity (optional)
#'
#' Positive integer representing the number of periods defining the *low* (e.g., benchmarks) frequency for adding
#' the extra spline knots (before the first benchmark and after the last benchmark). For example, `low_freq_periodicity = 3`
#' with monthly indicators would define quarterly knots. Annual knots are added when `low_freq_periodicity = NA`.
#'
#' **Default value** is `low_freq_periodicity = NA` (annual knots).
#'
#' @param n_low_freq_proj (optional)
#'
#' Nonnegative integer representing the number of low frequency knots (as defined with argument
#' `low_freq_periodicity`) to add at both ends (before the first benchmark and after the last benchmark)
#' before starting to add *high* (indicator series) frequency knots.
#'
#' **Default value** is `n_low_freq_proj = 1`.
#'
#' @param proj_knots_rho_bd (optional)
#'
#' Bound that applies to the value specified with argument `rho` and determines the type of extra knots to be added at 
#' both ends (before the first benchmark and after the last benchmark). When `rho > proj_knots_rho_bd`, *high* (indicator 
#' series) frequency knots are used right away. Otherwise, when `rho <= proj_knots_rho_bd`, *low* frequency knots (see 
#' arguments `low_freq_periodicity` and `n_low_freq_proj`) are first projected on either side. Note that for quarterly 
#' stocks, the cube of the specified `proj_knots_rho_bd` value is actually used. Therefore, the value for argument 
#' `proj_knots_rho_bd` should correspond to monthly stock indicators; it is internally adjusted for quarterly stocks. 
#' This argument aims at reaching a compromise for the set periods outside (before or after) the provided benchmarks 
#' (anchor points), i.e., Denton-type (straight line) adjustments as `rho` approaches 1 (when `rho > proj_knots_rho_bd`) 
#' and a natural looking (not overly contorted) spline otherwise (when `rho <= proj_knots_rho_bd`). Section **Details** 
#' contains more information on this subject and some illustrative cases are provided in section **Examples**.
#'
#' **Default value** is `proj_knots_rho_bd = 0.995` (\eqn{0.995^3} for quarterly stock indicators).
#'
#'
#' @details
#' ## Comparison with [benchmarking()]
#' With stock series, [benchmarking()] is known to produce breaks in the benchmarking adjustments at
#' periods corresponding to the benchmark stocks (anchor points). [stock_benchmarking()] addresses this issue
#' by working directly on the benchmarking adjustments. Smooth adjustments for stocks are ensured by
#' estimating a *slope=0* cubic spline (a spline that is flat at the end knots) going through knots
#' corresponding to the difference (when argument `lambda = 0.0`) or ratio (otherwise) between the benchmarks
#' (anchor points) and the corresponding indicator series values. These knots are sometimes referred to as
#' *BI* (***B***enchmark-to-***I***ndicator) *differences* or *BI ratios*. Interpolations from the estimated
#' cubic spline then provide the adjustments for the periods between benchmarks.
#'
#' Arguments `rho`, `lambda`, `biasOption`and `bias` play a similar role as in [benchmarking()]. However,
#' note that for [stock_benchmarking()], argument `rho` only affects the results for periods outside of, or
#' around the, first and last benchmarks and `lambda` only takes two values in practice: `lambda = 0.0` for
#' additive adjustments (spline interpolations where the knots are *BI differences*) or `lambda = 1.0` for
#' multiplicative adjustments (spline interpolations where the knots are *BI ratios*). Any nonzero value for
#' `lambda` would return the same result as `lambda = 1.0`. Alterability coefficients also play a similar role
#' as in [benchmarking()] and have the same default values, i.e., \eqn{1.0} for the indicator series
#' (nonbinding values) and \eqn{0.0} for the benchmarks (binding benchmarks). However, similar to argument
#' `lambda`, alterability coefficients in this function only take two values in practice: \eqn{0.0} for binding
#' values or \eqn{1.0} for nonbinding values. Any nonzero alterability coefficient would return the same result
#' as a coefficient of \eqn{1.0}. Another difference with [benchmarking()] is that user-defined
#' alterability coefficients are allowed even when `rho = 1` with [stock_benchmarking()]. Finally, specifying a
#' nonbinding benchmark with [stock_benchmarking()]  is equivalent to ignoring the benchmark entirely, as if the
#' benchmark was not included in the input benchmarks file. Compared to [benchmarking()], this generally
#' translates into nonbinding benchmarks having a larger impact on the resulting benchmarked stocks.
#'
#' ## Solution around the first and last benchmarks (benchmarking *timeliness issue*)
#' A *slope=0* spline is chosen because it conceptually corresponds to the (popular) *Denton benchmarking*
#' approach (`rho = 1`). In order to provide a solution before the first benchmark and after the last benchmark
#' that is similar to [benchmarking()] when `rho < 1`, i.e., adjustments converging to the bias at a speed
#' dictated by argument `rho`, extra knots are added at both ends before estimating the spline. By default, one
#' extra low frequency (as defined with argument `low_freq_periodicity`) knot is added on each side (beginning
#' and end), i.e. one extra knot is added before the first benchmark and after the last benchmark. Then, high
#' (indicator series) frequency knots are added to cover the indicator series span to which is added an extra
#' year worth of high frequency knots. The value of all those extra knots is based on arguments `rho`, `biasOption` 
#' and `bias`. This produces natural looking, smooth adjustments for periods outside of or around the first and last
#' benchmarks that gradually converge to the bias, similarly to [benchmarking()]. The number of extra low
#' frequency knots to be added can be modified with argument `n_low_freq_proj`. Using high frequency knots right
#' away (`n_low_freq_proj = 0`) would produce the same projected adjustments as [benchmarking()]. However,
#' note that this tends to produce an unnatural looking (overly contorted) spline around the first and last 
#' benchmarks that could be substantially revised once the next benchmark is available. Using the default 
#' `n_low_freq_proj = 1` generally works better. However, when `rho` is *close to 1* (see argument `proj_knots_rho_bd`), 
#' high frequency knots are immediately added on each side in order to ensure Denton-type (straight line) projected 
#' adjustments for periods outside of the first and last benchmarks. Finally, a *slope=0* cubic spline is fitted through 
#' the (original and extra) knots. Note that in practice, the *slope=0* spline is actually approximated by 
#' replicating the value of the end knots 100 times within the following period (at a frequency corresponding 
#' to 100 times the indicator series frequency).
#'
#' A *natural spline* at the original end knots (first and last benchmarks) can be approximated by specifying
#' a large value for argument `low_freq_periodicity`. The larger the value of `low_freq_periodicity`, the more
#' the cubic spline at the end knots will behave like a *natural spline* (2<sup>nd</sup> derivative equal to 0
#' at the end knots, i.e., a spline that keeps a constant slope at the end knots as opposed to being flat like
#' a *slope=0* spline).
#'
#' In summary, the projected adjustments are controlled with arguments `rho`, `bias` (and `biasOption`),
#' `n_low_freq_proj`, `proj_knots_rho_bd` and `low_freq_periodicity`:
#' * Default values for these arguments produce [benchmarking()] function-like projected adjustments
#' (reasonably slow convergence to the bias).
#' * Smaller values of `rho` would generate faster convergence to the bias.
#' * Specifying a user-defined bias with argument `bias` when `rho < 1` is another way to influence the shape
#' of the projected adjustments.
#' * Specifying `rho = 1` produce Denton-like projected adjustments (repeated first/last adjustments without
#' convergence to the bias).
#' * Specifying a large value for `low_freq_periodicity` generates projected adjustments that behave more
#' like a natural spline, i.e., adjustments that continue in the same direction at the first/last benchmark.
#' The larger the value of `low_freq_periodicity`, the more the projected adjustments keep on going in the
#' same direction before *turning around*.
#'
#' ## Note on revisions to the benchmarking adjustments
#' [benchmarking()] adjustments would not be revised if all future benchmarks were to fall exactly on the
#' projected ones (based on the bias and value of `rho`) and the bias was fixed. The same could be achieved
#' with [stock_benchmarking()] if *enough* low (e.g., benchmarks) frequency knots were projected. The problem
#' with this approach, however, is that the projected adjustments may not look natural as the spline may
#' oscillate more than desired around the projected knots. This is clearly noticeable as `rho` approaches 1
#' and the spline oscillates around the horizontally aligned projected knots instead of being aligned in a
#' perfectly straight line. The default implementation of the spline around the first and last benchmarks
#' described previously aims at reaching a *best compromise* solution:
#' * a natural looking spline around the end knots avoiding oscillations and excessive contortions;
#' * small revisions to the spline if the next benchmark is close to the projected one when `rho` is *far enough*
#' from 1 (`rho <= proj_knots_rho_bd`);
#' * projected adjustments that are in a straight line (free of oscillations) as `rho` approaches 1 (`rho >
#' proj_knots_rho_bd`).
#' 
#' Subsections *Benchmarking Multiple Series*, *Arguments `constant` and `negInput_option`* and *Treatment 
#' of Missing (`NA`) Values* at the end of the [benchmarking()] **Details** section are also relevant for 
#' [stock_benchmarking()]. Consult them as necessary.
#' 
#' Finally, note that the cubic spline associated to the [stock_benchmarking()] adjustments can be conveniently 
#' plotted with [plot_benchAdj()]. The latter is used in the **Examples** to illustrate some of the topics discussed 
#' above.
#' 
#'
#' @returns
#' The function returns is a list of four data frames:
#' 
#' * `series`: data frame containing the benchmarked data (primary function output). BY-group variables
#' specified with argument `by` would be included in the data frame but not alterability coefficient variables
#' specified with argument `var`.
#' 
#' * `benchmarks`: copy of the input benchmarks data frame (excluding invalid benchmarks when applicable).
#' BY-group variables specified with argument `by` would be included in the data frame but not alterability
#' coefficient variables specified with argument `with`.
#' 
#' * `graphTable`: data frame containing supplementary data useful to produce analytical tables and graphs
#' (see function [plot_graphTable()]). It contains the following variables in addition to the BY-group variables
#' specified with argument `by`:
#'   * `varSeries`: Name of the indicator series variable
#'   * `varBenchmarks`: Name of the benchmark variable
#'   * `altSeries`: Name of the user-defined indicator series alterability coefficients variable
#'   * `altSeriesValue`: Indicator series alterability coefficients
#'   * `altbenchmarks`: Name of the user-defined benchmark alterability coefficients variable
#'   * `altBenchmarksValue`: Benchmark alterability coefficients
#'   * `t`: Indicator series period identifier (1 to \eqn{T})
#'   * `m`: Benchmark coverage periods identifier (1 to \eqn{M})
#'   * `year`: Data point calendar year
#'   * `period`: Data point period (cycle) value (1 to `periodicity`)
#'   * `rho`: Autoregressive parameter \eqn{\rho} (argument `rho`)
#'   * `lambda`: Adjustment model parameter \eqn{\lambda} (argument `lambda`)
#'   * `bias`: Bias adjustment (default, user-defined or estimated bias according to arguments `biasOption` and `bias`)
#'   * `periodicity`: The maximum number of periods in a year (e.g. 4 for a quarterly indicator series)
#'   * `date`: Character string combining the values of variables `year` and `period`
#'   * `subAnnual`: Indicator series values
#'   * `benchmarked`: Benchmarked series values
#'   * `avgBenchmark`: Benchmark values divided by the number of coverage periods
#'   * `avgSubAnnual`: Indicator series values (variable `subAnnual`) averaged over the benchmark coverage period
#'   * `subAnnualCorrected`: Bias corrected indicator series values
#'   * `benchmarkedSubAnnualRatio`: Difference (\eqn{\lambda = 0}) or ratio (\eqn{\lambda \ne 0}{lambda != 0}) of the values 
#'   of variables `benchmarked` and `subAnnual`
#'   * `avgBenchmarkSubAnnualRatio`: Difference (\eqn{\lambda = 0}) or ratio (\eqn{\lambda \ne 0}{lambda != 0}) of the values 
#'   of variables `avgBenchmark` and `avgSubAnnual`
#'   * `growthRateSubAnnual`: Period to period difference (\eqn{\lambda = 0}) or relative difference (\eqn{\lambda \ne 0}{
#'   lambda != 0}) of the indicator series values (variable `subAnnual`)
#'   * `growthRateBenchmarked`: Period to period difference (\eqn{\lambda = 0}) or relative difference (\eqn{\lambda \ne 0}{
#'   lambda != 0}) of the benchmarked series values (variable `benchmarked`)
#'   
#' * `splineKnots`: set of `x` and `y` coordinates (knots) used to estimate the natural cubic spline with
#' function `stats::spline()`. In addition to the original set of knots corresponding to binding benchmarks
#' (anchor points), extra knots are also added at the beginning and end in order to deal with the *benchmarking
#' timeliness issue* and approximate a *slope=0* spline at both ends (see section **Details**). It contains the following 
#' variables in addition to the BY-group variables specified with argument `by`:
#'   * `varSeries`: Name of the indicator series variable
#'   * `varBenchmarks`: Name of the benchmark variable
#'   * `x`: Cubic spline `x` coordinate
#'   * `y`: Cubic spline `y` coordinate
#'   * `extraKnot`: Logical value identifying the extra knots added at the beginning and end
#'   
#'   Rows for which `extraKnot == FALSE` correspond to rows in the `graphTable` output data frame for which 
#'   `m` is not missing (not `NA`), with `x = t` and `y = benchmarkedSubAnnualRatio`.
#'
#' Notes:
#' * The output `benchmarks` data frame always contains the original benchmarks provided in the input
#' benchmarks data frame. Modified nonbinding benchmarks, when applicable, can be recovered (calculated)
#' from the output `series` data frame.
#' * The function returns a `NULL` object if an error occurs before data processing could start. Otherwise,
#' if execution gets far enough so that data processing could start, then an incomplete object would be
#' returned in case of errors (e.g., output `series` data frame with `NA` values for the benchmarked data).
#' * The function returns "data.frame" objects that can be explicitly coerced to other types of objects with 
#' the appropriate `as*()` function (e.g., `tibble::as_tibble()` would coerce any of them to a tibble).
#'
#'
#' @references Statistics Canada (2012). "Chapter 5: Benchmarking Stock". **Theory and Application of Benchmarking
#' (Course code 0436)**. Statistics Canada, Ottawa, Canada.
#'
#'
#' @seealso [benchmarking()] [plot_graphTable()] [bench_graphs] [plot_benchAdj()]
#'
#'
#' @example misc/function_examples/stock_benchmarking-ex.R
#'
#'
#' @export
stock_benchmarking <- function(series_df,
                               benchmarks_df,
                               rho,
                               lambda,
                               biasOption,
                               bias = NA,
                               low_freq_periodicity = NA,
                               n_low_freq_proj = 1,
                               proj_knots_rho_bd = 0.995,
                               tolV = 0.001,
                               tolP = NA,
                               warnNegResult = TRUE,
                               tolN = -0.001,
                               var = "value",
                               with = NULL,
                               by = NULL,
                               constant = 0,
                               negInput_option = 0,
                               allCols = FALSE,
                               quiet = FALSE) {




  ### Internal functions ###


  # NOTE: some internal functions shared with `benchmarking()` are defined in script `aaa_bench_utils.R`
  #       where environment `the` is defined and contains objects that are not (cannot be) passed as arguments
  #       to the shared internal benchmarking functions.


  # Binding benchmarks validation function (proportional benchmarking0)
  check_nonZero_bindingBmk <- function(s_lowFreq, a, c_a, tol = gs.tolerance) {

    # Elementwise comparisons are necessary here (cannot use &&)
    if (any(abs(s_lowFreq) <= tol & (abs(a) * (c_a == 0)) > tol)) {
      warning("The indicator series is zero for a nonzero binding benchmark (anchor point). ",
              "This benchmark cannot be met with proportional benchmarking.\n", call. = FALSE, immediate. = TRUE)
      bk.e$warning_flag <- TRUE
    }
    
    invisible(NULL)
  }




  ### Main function ###

  # Initialize the object to be returned by the function via `on.exit()`
  out_list <- NULL
  on.exit(return(out_list))
  try_error <- FALSE
  try_error_msg <- ""
  bk.e$warning_flag <- FALSE
  
  # Enforce the default R "error" option (`options(error = NULL)`). E.g. this Turns off traceback
  # generated by calls to the stop() function inside internal functions in R Studio.
  ini_error_opt <- getOption("error")
  on.exit(options(error = ini_error_opt), add = TRUE)
  options(error = NULL)
  
  # Validate argument `quiet` and implement the quiet setting
  quiet <- gs.validate_arg_logi(quiet)
  if (quiet) {
    quiet_msg_func <- gs.NULL_func
    quiet_lab <- ""  # won't be displayed anyway
  } else {
    quiet_msg_func <- message
    quiet_lab <- "    quiet                = FALSE (default)"
  }

  # Display the function header
  quiet_msg_func("\n\n", gs.header, "\n\nstock_benchmarking() function:\n")


  # Initial argument validation

  # Mandatory arguments (without default values)
  ser_df_name <- deparse1(substitute(series_df))
  tmp <- nchar(ser_df_name)
  if (tmp == 0) {
    stop("Argument 'series_df' is mandatory (it must be specified).\n\n", call. = FALSE)
  } else if (tmp >= 60) {
    ser_df_name <- paste0(substr(ser_df_name, 1, 55), "<...>")
  }
  if (grepl("structure(", ser_df_name, fixed = TRUE)) {
    ser_df_name <- "<argument 'series_df'>"
  }
  series_df <- series_df
  if (!is.data.frame(series_df)) {
    stop("Argument 'series_df' is not a 'data.frame' object.\n\n", call. = FALSE)
  }
  series_df <- as.data.frame(series_df)
  row.names(series_df) <- NULL
  bmk_df_name <- deparse1(substitute(benchmarks_df))
  tmp <- nchar(bmk_df_name)
  if (tmp == 0) {
    stop("Argument 'benchmarks_df' is mandatory (it must be specified).\n\n", call. = FALSE)
  } else if (tmp >= 60) {
    bmk_df_name <- paste0(substr(bmk_df_name, 1, 55), "<...>")
  }
  if (grepl("structure(", bmk_df_name, fixed = TRUE)) {
    bmk_df_name <- "<argument 'benchmarks_df'>"
  }
  benchmarks_df <- benchmarks_df
  if (!is.data.frame(benchmarks_df)) {
    stop("Argument 'benchmarks_df' is not a 'data.frame' object.\n\n", call. = FALSE)
  }
  benchmarks_df <- as.data.frame(benchmarks_df)
  row.names(benchmarks_df) <- NULL
  if (nchar(deparse1(substitute(lambda))) == 0) {
    stop("Argument 'lambda' is mandatory (it must be specified).\n\n", call. = FALSE)
  }
  tmp <- (unlist(lambda))[1]
  if (!identical(lambda, tmp) || is.null(tmp) || !is.finite(tmp)) {
    stop("Argument 'lambda' must be a real number.\n\n", call. = FALSE)
  }
  if (nchar(deparse1(substitute(rho))) == 0) {
    stop("Argument 'rho' is mandatory (it must be specified).\n\n", call. = FALSE)
  }
  tmp <- (unlist(rho))[1]
  if (!identical(rho, tmp) || is.null(tmp) || !is.finite(tmp) || tmp < 0 || tmp > 1) {
    stop("Argument 'rho' must be a real number in the [0, 1] interval.\n\n", call. = FALSE)
  }
  if (nchar(deparse1(substitute(biasOption))) == 0) {
    if (rho == 1) {
      # bias is irrelevant for Denton benchmarking (`rho = 1`)
      biasOption <- 1L
    } else {
      stop("Argument 'biasOption' is mandatory (it must be specified).\n\n", call. = FALSE)
    }
  }
  tmp <- (unlist(biasOption))[1]
  if (!identical(biasOption, tmp) || is.null(tmp) || !(tmp %in% 1:3)) {
    stop("Argument 'biasOption' must take value 1, 2 or 3.\n\n", call. = FALSE)
  }
  biasOption <- as.integer(biasOption)

  # Optional arguments (with default values)
  tmp <- (unlist(bias))[1]
  if (is.null(tmp)) {
    bias <- NA_real_
  } else if (!identical(bias, tmp) || !is.finite(tmp) && !is.na(tmp)) {
    stop("Argument 'bias' must either be a real number or NA.\n\n", call. = FALSE)
  }
  tmp <- (unlist(low_freq_periodicity))[1]
  if (is.null(tmp)) {
    low_freq_periodicity <- NA_integer_
  } else {
    if (!identical(low_freq_periodicity, tmp) || !is.finite(tmp) && !is.na(tmp) ||
        is.finite(tmp) && (tmp <= 0 || tmp != as.integer(tmp))) {
      stop("Argument 'n_low_freq_proj' must be a positive integer or NA.\n\n", call. = FALSE)
    }
    low_freq_periodicity <- as.integer(low_freq_periodicity)    
  } 
  tmp <- (unlist(n_low_freq_proj))[1]
  if (!identical(n_low_freq_proj, tmp) || is.null(tmp) || !is.finite(tmp) || tmp < 0 || tmp != as.integer(tmp)) {
    stop("Argument 'n_low_freq_proj' must be a nonnegative integer.\n\n", call. = FALSE)
  }
  n_low_freq_proj <- as.integer(n_low_freq_proj)
  tmp <- (unlist(proj_knots_rho_bd))[1]
  if (!identical(proj_knots_rho_bd, tmp) || is.null(tmp) || !is.finite(tmp) || tmp < 0 || tmp > 1) {
    stop("Argument 'proj_knots_rho_bd' must be a real number in the [0, 1] interval.\n\n", call. = FALSE)
  }
  tmp <- (unlist(tolV))[1]
  if (is.null(tmp)) {
    tolV <- NA_real_
  } else if (!identical(tolV, tmp) || !is.finite(tmp) && !is.na(tmp) || is.finite(tmp) && tmp < 0) {
    stop("Argument 'tolV' must be a nonnegative real number or NA.\n\n", call. = FALSE)
  }
  tmp <- (unlist(tolP))[1]
  if (is.null(tmp)) {
    tolP <- NA_real_
  } else if (!identical(tolP, tmp) || !is.finite(tmp) && !is.na(tmp) || is.finite(tmp) && tmp < 0) {
    stop("Argument 'tolP' must be a nonnegative real number or NA.\n\n", call. = FALSE)
  }
  warnNegResult <- gs.validate_arg_logi(warnNegResult)
  tmp <- (unlist(tolN))[1]
  if (!identical(tolN, tmp) || is.null(tmp) || !is.finite(tmp) || tmp >= 0) {
    stop("Argument 'tolN' must be a negative real number.\n\n", call. = FALSE)
  }
  tmp <- (unlist(constant))[1]
  if (!identical(constant, tmp) || is.null(tmp) || !is.finite(tmp)) {
    stop("Argument 'constant' must be a real number.\n\n", call. = FALSE)
  }
  tmp <- (unlist(negInput_option))[1]
  if (!identical(negInput_option, tmp) || is.null(tmp) || !(tmp %in% 0:2)) {
    stop("Argument 'negInput_option' must take value 0, 1 or 2.\n\n", call. = FALSE)
  }
  negInput_option <- as.integer(negInput_option)
  allCols <- gs.validate_arg_logi(allCols)
  # Argument `quiet` was already validated

  # Initialize the final warning/error message flag for multiple series processing
  final_msg_flag <- FALSE
  try_stop_func <- gs.NULL_func

  # Initialize the list of "non-data columns" (date and by-group columns)
  info_cols_serDF <- c("year", "period")
  info_cols_bmkDF <- c("startYear", "startPeriod", "endYear", "endPeriod")

  all_cols_serDF <- names(series_df)
  gs.validate_cols(info_cols_serDF, all_cols_serDF, ser_df_name)
  if (any(!is.finite(as.matrix(series_df[info_cols_serDF])))) {
    stop("Indicator series date columns (\"year\" and \"period\") contain invalid or NA values.\n\n", call. = FALSE)
  }

  all_cols_bmkDF <- names(benchmarks_df)
  gs.validate_cols(info_cols_bmkDF, all_cols_bmkDF, bmk_df_name)
  if (any(!is.finite(as.matrix(benchmarks_df[info_cols_bmkDF])))) {
    stop("Benchmarks date columns (\"startYear\", \"startPeriod\", \"endYear\" and \"endPeriod\") contain invalid or NA values.\n\n",
         call. = FALSE)
  }

  # Initialize the list of "series data columns" from both input files
  data_cols_serDF <- setdiff(all_cols_serDF, info_cols_serDF)
  data_cols_bmkDF <- setdiff(all_cols_bmkDF, info_cols_bmkDF)

  # Validation of argument 'by' ("non-data columns" as well)
  by <- unique(gs.cleanup_col_list(by))
  n_byVars <- length(by)
  if (n_byVars > 0) {
    by_lab <- paste0("    by                   = ", paste(by, collapse = " "))
    gs.validate_cols(by, data_cols_serDF, ser_df_name, source_str = "argument 'by'")
    gs.validate_cols(by, data_cols_bmkDF, bmk_df_name, source_str = "argument 'by'")
    by_grps <- unique(series_df[by])
    n_byGrps <- nrow(by_grps)

    # Build the by-group expression (for message display)
    if (n_byGrps > 0) {
      by_grps$SB._expr_ <- paste0(by[1], "=", by_grps[, by[1]])
      for (ii in seq_along(by)[-1]) {
        by_grps$SB._expr_ <- paste(by_grps$SB._expr_, paste0(by[ii], "=", by_grps[, by[ii]]), sep = " & ")
      }
    }

    # Add the by variables to the list of "non-data columns"
    info_cols_serDF <- c(by, info_cols_serDF)
    info_cols_bmkDF <- c(by, info_cols_bmkDF)

    # Remove by variables from the list of "series data columns"
    data_cols_serDF <- setdiff(data_cols_serDF, by)
    data_cols_bmkDF <- setdiff(data_cols_bmkDF, by)

    # Set message display info
    if (n_byGrps > 1) {
      byGrp_ini_func <- bk.byGrp_ini
      byGrp_msg_func <- message
      final_msg_flag <- TRUE
      try_stop_func <- gs.try_stop
    } else {
      byGrp_ini_func <- bk.noByGrp_ini
      byGrp_msg_func <- gs.NULL_func
    }

  # No by-groups
  } else {
    by <- NULL
    by_lab <- "    by                   = NULL (default)"
    by_grps <- data.frame(SB._expr_ = "")
    n_byGrps <- 1
    byGrp_ini_func <- bk.noByGrp_ini
    byGrp_msg_func <- gs.NULL_func
  }

  # Ignore arguments 'var' and 'with' and process all data columns of the input indicator series
  # data frame, expecting corresponding columns (same names) in the input benchmarks data frame
  if (allCols) {
    var_lab <- "    var                  (ignored)"
    with_lab <- "    with                 (ignored)"
    allCols_lab <- "    allCols              = TRUE"
    var <- data_cols_serDF
    with <- var
    n_vars <- length(var)
    alter_ser <- rep.int("", n_vars)
    default_alter_ser_id <- 1:n_vars
    user_alter_ser_id <- integer(0)  # value returned by which() on an "all FALSE" logical vector
    alter_bmk <- alter_ser
    default_alter_bmk_id <- 1:n_vars
    user_alter_bmk_id <- integer(0)
    gs.validate_cols(with, data_cols_bmkDF, bmk_df_name)

    # Process arguments 'var' and 'with'
  } else {
    allCols_lab <- "    allCols              = FALSE (default)"

    # Validation of argument 'var'
    var <- gs.cleanup_col_list(var)
    var_lab <- paste0("    var                  = ", paste(var, collapse = " "))
    if (var_lab == "    var                  = value") {
      var_lab <- "    var                  = value (default)"
    }
    tmp <- gs.split_str("\\/", var)
    var <- tmp[[1]]
    n_vars <- length(var)
    len <- length(tmp)
    if (len == 2) {
      alter_ser <- tmp[[2]]
      default_alter_ser_id <- which(alter_ser == "")
      user_alter_ser_id <- setdiff(1:n_vars, default_alter_ser_id)
    } else {
      if (len > 2) {
        stop("Invalid specification of alterability coefficients in argument 'var'.\n\n", call. = FALSE)
      }
      alter_ser <- rep.int("", n_vars)
      default_alter_ser_id <- 1:n_vars
      user_alter_ser_id <- integer(0)
    }
    prob_cols <- intersect(var, by)
    if (length(prob_cols) > 0) {
      stop("The following columns are listed with arguments 'var' and 'by' (these arguments are mutually exclusive):",
           paste0("\n  ", prob_cols, collapse = ""), "\n\n", call. = FALSE)
    }
    gs.validate_cols(c(var, alter_ser[user_alter_ser_id]), data_cols_serDF, ser_df_name, source_str = "argument 'var'")

    # Validation of argument 'with'
    with <- gs.cleanup_col_list(with)
    if (length(with) == 0) {
      with_lab <- "    with                 = NULL (default)"
      with <- var
      alter_bmk <- rep.int("", n_vars)
      default_alter_bmk_id <- 1:n_vars
      user_alter_bmk_id <- integer(0)
    } else {
      with_lab <- paste0("    with                 = ", paste(with, collapse = " "))
      tmp <- gs.split_str("\\/", with)
      with <- tmp[[1]]
      len <- length(tmp)
      if (len == 2) {
        alter_bmk <- tmp[[2]]
        default_alter_bmk_id <- which(alter_bmk == "")
        user_alter_bmk_id <- setdiff(1:n_vars, default_alter_bmk_id)
      } else {
        if (len > 2) {
          stop("Invalid specification of alterability coefficients in argument 'with'.", call. = FALSE)
        }
        alter_bmk <- rep.int("", n_vars)
        default_alter_bmk_id <- 1:n_vars
        user_alter_bmk_id <- integer(0)
      }
    }
    prob_cols <- intersect(with, by)
    if (length(prob_cols) > 0) {
      stop("The following columns are listed with arguments 'with' and 'by' (these arguments are mutually exclusive):",
           paste0("\n  ", prob_cols, collapse = ""), "\n\n", call. = FALSE)
    }
    gs.validate_cols(c(with, alter_bmk[user_alter_bmk_id]), data_cols_bmkDF, bmk_df_name, source_str = "argument 'with'")
    if (n_vars != length(with)) {
      stop("Arguments 'var' and 'with' do not contain the same number of elements.\n\n", call. = FALSE)
    }
  }
  if (n_vars == 0) {
    stop("Argument 'var' results in 0 series to be benchmarked.\n\n", call. = FALSE)
  } else if (n_vars == 1) {
    var_msg_func <- gs.NULL_func
  } else {
    var_msg_func <- message
    final_msg_flag <- TRUE
    try_stop_func <- gs.try_stop
  }


  # Shrink the input data frames with only the necessary columns
  series_df <- series_df[unique(c(info_cols_serDF, var, alter_ser[user_alter_ser_id]))]
  benchmarks_df <- benchmarks_df[unique(c(info_cols_bmkDF, with, alter_bmk[user_alter_bmk_id]))]

  # Add the default alter coefs to the input data frames and specify them when relevant
  series_df$SB._default_alter_ <- rep.int(1, nrow(series_df))
  benchmarks_df$SB._default_alter_ <- rep.int(0, nrow(benchmarks_df))
  actual_alter_ser <- alter_ser
  actual_alter_bmk <- alter_bmk
  alter_ser[default_alter_ser_id] <- "SB._default_alter_"
  alter_bmk[default_alter_bmk_id] <- "SB._default_alter_"
  alter_ser_check_func_str <- rep.int("gs.FALSE_func", n_vars)
  alter_bmk_check_func_str <- rep.int("gs.FALSE_func", n_vars)
  alter_ser_check_func_str[user_alter_ser_id] <- "gs.check_alter"
  alter_bmk_check_func_str[user_alter_bmk_id] <- "gs.check_alter"

  # Initialize the output data frames (as NULL objects for now)
  out_series_df <- NULL
  out_benchmarks_df <- NULL
  out_graphTable_df <- NULL
  out_splineKnots_df <- NULL

  # Set parameters and function names according to the benchmarking model (additive benchmarking
  # when 'lambda == 0' and proportional benchmarking otherwise):
  #   - negative values verification functions
  #   - bias calculation and BI (ratios or differences) application
  #   - ratio and growth rate calculation functions
  if (lambda == 0) {
    zeros_verif_func <- gs.NULL_func
    neg_verif_func <- gs.FALSE_func
    default_bias <- 0
    bias_calc_func <- bk.calc_add_bias
    bias_apply_func <- bk.apply_add_bias
    BI_apply_func <- bk.apply_add_bias
    ratio_func <- gs.calc_diff
    growthRate_func <- gs.calc_firstDiff
  } else {
    zeros_verif_func <- check_nonZero_bindingBmk
    if (negInput_option == 0) {
      neg_verif_func <- bk.check_neg_err
    } else if (negInput_option == 1) {
      neg_verif_func <- bk.check_neg_warn
    } else {
      neg_verif_func <- gs.FALSE_func
    }
    default_bias <- 1
    bias_calc_func <- bk.calc_mult_bias
    bias_apply_func <- bk.apply_mult_bias
    BI_apply_func <- bk.apply_mult_bias
    ratio_func <- gs.calc_ratio
    growthRate_func <- gs.calc_relFirstDiff
    if (!is.na(bias) && bias < 0) {
      stop("Argument 'bias' must be positive with proportional benchmarking.\n\n", call. = FALSE)
    }
  }

  # Set the bias option function according to argument 'biasOption'
  if (rho == 1) {
    # The bias has no effect and can be ignored (impose `biasOption = 1` and `bias = NA`)
    biasOption_func <- bk.apply_biasOption1
    biasOption_lab <- "    biasOption           (ignored)"
    bias_parm <- default_bias
    bias_str <- "default"
    bias_lab <- "    bias                 (ignored)"
  } else {
    if (biasOption == 3) {
      biasOption_func <- bk.apply_biasOption3
      biasOption_lab <- "    biasOption           = 3 (Calculate bias, use calculated bias)"
      bias_parm <- NA_real_
      bias_str <- NA_character_
      bias_lab <- "    bias                 (ignored)"
    } else {
      if (biasOption == 1) {
        biasOption_func <- bk.apply_biasOption1
        biasOption_lab <- "    biasOption           = 1 (Use user-defined or default bias)"
      } else if (biasOption == 2) {
        biasOption_func <- bk.apply_biasOption2
        biasOption_lab <- "    biasOption           = 2 (Calculate bias, but use user-defined or default bias)"
      }
      if (is.na(bias)) {
        bias_parm <- default_bias
        bias_str <- "default"
        bias_lab <- "    bias                 = NA (default)"
      } else {
        bias_parm <- bias
        bias_str <- "user-defined"
        bias_lab <- paste0("    bias                 = ", format(bias))
      }
    }
  }
  bk.e$actual_bias <- NULL

  # Minimum required number of periods in the indicator series
  min_nT <- 1

  # Results validation function and binding benchmarks tolerance
  if (warnNegResult) {
    neg_res_func <- gs.check_neg
    warnNegResult_lab <- "    warnNegResult        = TRUE (default)"
  } else {
    neg_res_func <- gs.FALSE_func
    warnNegResult_lab <- "    warnNegResult        = FALSE"
  }
  if (!is.na(tolV)) {
    if (!is.na(tolP)) {
      stop("Arguments 'tolV' and 'tolP' cannot be both specified (one must be NA).\n\n", call. = FALSE)
    }
    binding_bmk_valid_func <- bk.binding_bmk_diff_valid
    tol_parm <- tolV
    tol_lab <- paste0("    tolV                 = ", format(tolV))
    if (abs(tolV - 0.001) < gs.tolerance) {
      tol_lab <- paste0(tol_lab, " (default)")
    }
  } else {
    if (is.na(tolP)) {
      stop("Arguments 'tolV' and 'tolP' cannot be both NA (one must be specified).\n\n", call. = FALSE)
    }
    binding_bmk_valid_func <- bk.binding_bmk_relDiff_valid
    tol_parm <- tolP
    tol_lab <- paste0("    tolP                 = ", format(tolP))
  }


  # Display the function call (argument values)
  quiet_msg_func("    series_df            = ", ser_df_name)
  quiet_msg_func("    benchmarks_df        = ", bmk_df_name)
  quiet_msg_func("    rho                  = ", format(rho))
  quiet_msg_func("    lambda               = ", format(lambda))
  quiet_msg_func(biasOption_lab)
  quiet_msg_func(bias_lab)
  lab <- paste0("    low_freq_periodicity = ", low_freq_periodicity)
  if (is.na(low_freq_periodicity)) {
    lab <- paste0(lab, " (default)")
  }
  quiet_msg_func(lab)
  lab <- paste0("    n_low_freq_proj      = ", n_low_freq_proj)
  if (n_low_freq_proj == 1) {
    lab <- paste0(lab, " (default)")
  }
  quiet_msg_func(lab)
  lab <- paste0("    proj_knots_rho_bd    = ", format(proj_knots_rho_bd))
  if (abs(proj_knots_rho_bd - 0.995) < gs.tolerance) {
    lab <- paste0(lab, " (default)")
  }
  quiet_msg_func(lab)
  quiet_msg_func(tol_lab)
  quiet_msg_func(warnNegResult_lab)
  lab <- paste0("    tolN                 = ", format(tolN))
  if (abs(tolN - (-0.001)) < gs.tolerance) {
    lab <- paste0(lab, " (default)")
  }
  quiet_msg_func(lab)
  quiet_msg_func(var_lab)
  quiet_msg_func(with_lab)
  quiet_msg_func(by_lab)
  lab <- paste0("    constant             = ", format(constant))
  if (constant == 0) {
    lab <- paste0(lab, " (default)")
  }
  quiet_msg_func(lab)
  lab <- paste0("    negInput_option      = ", format(negInput_option))
  if (negInput_option == 0) {
    lab <- paste0(lab, " (default)")
  }
  quiet_msg_func(lab)
  quiet_msg_func(allCols_lab)
  quiet_msg_func(quiet_lab, "\n")
  
  
  # Return the input series data frames
  if (nrow(series_df) == 0 || nrow(benchmarks_df) == 0) {
    out_series_df <- series_df[intersect(all_cols_serDF, names(series_df))]
    out_benchmarks_df <- benchmarks_df[intersect(all_cols_bmkDF, names(benchmarks_df))]
    
    # Attempt to benchmark the data
  } else {
    
    # Enforce integer period id values
    cols_vec <- setdiff(info_cols_serDF, by)
    series_df[cols_vec] <- as.integer(as.matrix(series_df[cols_vec]))
    cols_vec <- setdiff(info_cols_bmkDF, by)
    benchmarks_df[cols_vec] <- as.integer(as.matrix(benchmarks_df[cols_vec]))
    
    
    # Process each by-group

    on.exit(rlang::env_unbind(bk.e, c("ser_df_byGrp", "bmk_df_byGrp")), add = TRUE)
    for (ii in 1:n_byGrps) {
  
      # By-group initialization/setup
      bk.e$ser_df_byGrp <- NULL
      bk.e$bmk_df_byGrp <- NULL
      byGrp_ini_func(series_df, benchmarks_df, by_grps, ii, by)
  
      msg_str <- paste0("\nBenchmarking by-group ", ii, " (", by_grps$SB._expr_[ii], ")")
      byGrp_msg_func(msg_str)
      byGrp_msg_func(strrep("=", nchar(msg_str) - 1), "\n")
  
      # Reject rows with invalid benchmark data (for any of the benchmarks)
      M_ini <- nrow(bk.e$bmk_df_byGrp)
      if (nrow(bk.e$bmk_df_byGrp) > 1) {
        prob_id <- which(apply(!apply(bk.e$bmk_df_byGrp[with], 2, is.finite), 1, any))
      } else {
        prob_id <- which(any(!apply(bk.e$bmk_df_byGrp[with], 2, is.finite)))
      }
      if (length(prob_id) > 0) {
        warning("Rows from the benchmarks data frame were dropped because of invalid or NA values.\n",
                call. = FALSE, immediate. = TRUE)
        bk.e$warning_flag <- TRUE
        bk.e$bmk_df_byGrp <- bk.e$bmk_df_byGrp[-prob_id, , drop = FALSE]
        M <- nrow(bk.e$bmk_df_byGrp)
      } else {
        M <- M_ini
      }
  
      # Initialize the by-group output data frames
      out_bmk_df_byGrp <- bk.e$bmk_df_byGrp[c(info_cols_bmkDF, with)]
      out_ser_df_byGrp <- bk.e$ser_df_byGrp[c(info_cols_serDF, var)]
      out_ser_df_byGrp[var] <- NA_real_
  
      nT <- nrow(bk.e$ser_df_byGrp)
      if (nT < min_nT) {
        try_error_msg <- paste0("The minimum number of periods (", min_nT, ") for the indicator series ", 
                                "is not met.\n\n")
        try_stop_func(try_error_msg)
        try_error <- TRUE
  
      } else if (M == 0) {
        try_error_msg <- "A minimum of 1 benchmark is required.\n\n"
        try_stop_func(try_error_msg)
        try_error <- TRUE
  
      } else {
  
        # Vectors for periods and benchmarks coverage validation
        periodicity <- max(bk.e$ser_df_byGrp$period)
        if (is.na(low_freq_periodicity)) {
          low_freq_periodicity <- periodicity
        }
        periods <- paste(bk.e$ser_df_byGrp$year, bk.e$ser_df_byGrp$period, sep = "-")
        bmk_start <- paste(bk.e$bmk_df_byGrp$startYear, bk.e$bmk_df_byGrp$startPeriod, sep = "-")
        bmk_end <- paste(bk.e$bmk_df_byGrp$endYear, bk.e$bmk_df_byGrp$endPeriod, sep = "-")
  
        # Validate indicator series periods (period-to-period gap != 1 / periodicity)
        time_val <- bk.e$ser_df_byGrp$year + (bk.e$ser_df_byGrp$period - 1) / periodicity
        if (nT > 1) {
          # `sapply()` is safe: `which()` always returns a "vector" object, even when `nT = 2`
          prob_id <- which(c(FALSE,
                             sapply(2:nT,
                                    function(x) {
                                      abs(time_val[x] - time_val[x - 1] - 1 / periodicity) > gs.tolerance
                                    })))
        } else {
          prob_id <- integer(0)
        }
        if (length(prob_id) > 0) {
          try_error_msg <- paste0("Non-contiguous periods found in the indicator series: ",
                                  paste0("\n  ", periods[prob_id - 1], " - ", periods[prob_id], collapse = ""), 
                                  "\n\n")
          try_stop_func(try_error_msg)
          try_error <- TRUE
  
        } else {
  
          # Validate the benchmark coverage: anchor points (single-period coverage)
          prob_id <- which(bk.e$bmk_df_byGrp$startYear != bk.e$bmk_df_byGrp$endYear |
                           bk.e$bmk_df_byGrp$startPeriod != bk.e$bmk_df_byGrp$endPeriod)
          if (length(prob_id) > 0) {
            try_error_msg <- paste0("Benchmark coverage not compatible with stock benchmarking (not an anchor point): ",
                                    paste0("\n  [", bmk_start[prob_id], ", ", bmk_end[prob_id], "]", 
                                           collapse = ""), 
                                    "\n\n")
            try_stop_func(try_error_msg)
            try_error <- TRUE
  
          } else {
  
            # Build benchmarks to periods mapping vector of dimension M
            # (period id of each "anchor point" benchmark)
            bmk_per_id <- match(bmk_start, periods)
  
            # Validate the benchmark coverage: not inside the set of indicator periods
            prob_id <- which(is.na(bmk_per_id))
            if (length(prob_id) > 0) {
              try_error_msg <- paste0("Benchmark coverage not fully inside the indicator series span: ",
                                      paste0("\n  [", bmk_start[prob_id], ", ", bmk_end[prob_id], "]", 
                                             collapse = ""), 
                                      "\n\n")
              try_stop_func(try_error_msg)
              try_error <- TRUE
  
            } else {
  
              # Display a message about the number of observations in the benchmarks and series data frames
              # Note: all indicator series observations are valid at this point, making the SERIES
              #       message not quite relevant... but that's what SAS G-Series PROC BENCHMARKING displays
              #       (there might be cases where the 2 numbers differ, but I don't see it...)
  
              # Widths without commas
              width_nT <- floor(log10(nT)) + 1
              width_M_ini <- floor(log10(M_ini)) + 1
              width_M <- floor(log10(M)) + 1
  
              # Adjust the widths of the benchmarks info considering the commas (for `format(..., big.mark = ",")`)
              # => add the difference in the number of commas between the series obs. number (`nT`) and each of the 
              #    two (initial and final) benchmarks obs. numbers (`M_ini` and `M`)
              # => it's (reasonably) assumed here that there will always be more observations in the series data 
              #    frame than in the benchmarks data frame (i.e., `nT` >= `M_ini` >= `M`)
              if (width_M_ini < width_nT) {
                width_M_ini <- width_nT + floor((width_nT - 1) / 3) - floor((width_M_ini - 1) / 3)
              }
              if (width_M < width_nT) {
                width_M <- width_nT + floor((width_nT - 1) / 3) - floor((width_M - 1) / 3)
              }
              quiet_msg_func("Number of observations in the BENCHMARKS data frame .............: ",
                             format(M_ini, scientific = FALSE, big.mark = ",", width = width_M_ini, justify = "right"))
              quiet_msg_func("Number of valid observations in the BENCHMARKS data frame .......: ",
                             format(M, scientific = FALSE, big.mark = ",", width = width_M, justify = "right"), "\n")
              quiet_msg_func("Number of observations in the SERIES data frame .................: ",
                             format(nT, scientific = FALSE, big.mark = ",", width = width_nT, justify = "right"))
              quiet_msg_func("Number of valid observations in the SERIES data frame ...........: ",
                             format(nT, scientific = FALSE, big.mark = ",", width = width_nT, justify = "right"), "\n")
  
              # Set K1 for the BY-group...
              #   K1 = the number of low (benchmark) frequency extra knots at the start and end
              if (rho > proj_knots_rho_bd^(12 / periodicity)) {
                K1 <- 0
              } else {
                K1 <- n_low_freq_proj
              }

  
              # Process each series
              for (jj in 1:n_vars) {
  
                msg_str <- paste0("\nBenchmarking indicator series [", var[jj], "] with benchmarks [", with[jj], "]")
                var_msg_func(msg_str)
                var_msg_func(strrep("-", nchar(msg_str) - 1), "\n")
  
                # Series/benchmarks data and alterablity coefficients validation
                if (any(!is.finite(bk.e$ser_df_byGrp[[var[jj]]]))) {
                  warning("The indicator series contains invalid or NA values. It will not be processed.\n", call. = FALSE,
                          immediate. = TRUE)
                  bk.e$warning_flag <- TRUE
  
                } else if (do.call(alter_ser_check_func_str[jj], list(bk.e$ser_df_byGrp[[alter_ser[jj]]]))) {
                  try_error_msg <- "Invalid indicator series alterability coefficients (must be nonnegative numbers).\n\n"
                  try_stop_func(try_error_msg)
                  try_error <- TRUE
  
                } else if (do.call(alter_bmk_check_func_str[jj], list(bk.e$bmk_df_byGrp[[alter_bmk[jj]]]))) {
                  try_error_msg <- "Invalid benchmarks alterability coefficients (must be nonnegative numbers).\n\n"
                  try_stop_func(try_error_msg)
                  try_error <- TRUE
  
                  # Process (benchmark) the series
                } else {
  
                  # Build the elementary vectors and matrices
                  #   => the temporary constant is added here
                  s <- bk.e$ser_df_byGrp[[var[jj]]] + constant
                  c_s <- bk.e$ser_df_byGrp[[alter_ser[jj]]]
                  a <- bk.e$bmk_df_byGrp[[with[jj]]] + constant
                  c_a <- bk.e$bmk_df_byGrp[[alter_bmk[jj]]]
  
                  # Get the indicator series values associated to the benchmarks (anchor points)
                  s_lowFreq <- s[bmk_per_id]
  
                  # Additional series/benchmarks data validation for proportional benchmarking (lambda != 0)
                  #   - nonzero binding benchmarks associated to zero indicator series values (undefined BI value)
                  #   - negative benchmark or indicator series values (according to argument `negInput_option`)
                  zeros_verif_func(s_lowFreq, a, c_a, tol = gs.tolerance)
                  if (neg_verif_func(a, tol = 0, data_str = "benchmarks")) {
                    try_error_msg <- paste0("Negative values found in the benchmarks. This is not permitted for ", 
                                            "proportional benchmarking.\n\n")
                    try_stop_func(try_error_msg)
                    try_error <- TRUE
  
                  } else if (neg_verif_func(s, tol = 0, data_str = "indicator series")) {
                    try_error_msg <- paste0("Negative values found in the indicator series. This is not permitted for ", 
                                            "proportional benchmarking.\n\n")
                    try_stop_func(try_error_msg)
                    try_error <- TRUE
  
                  } else {
  
                    # Bias correction
                    s_b <- biasOption_func(quiet_msg_func,                                 # message function
                                           bias_calc_func, s_lowFreq, a, M, gs.tolerance,  # bias calculation arguments
                                           bias_apply_func, s, bias_parm,                  # bias application arguments
                                           bias_str)
  
  
                    # Calculate the initial set of knots corresponding to binding benchmarks, taking the average value
                    # for duplicated binding benchmarks (spline knot x = t, spline knot y = BI differences or ratios)
                    binding_m <- which(c_a == 0)
                    t_vec <- bmk_per_id[binding_m]
                    if (anyDuplicated(t_vec)) {
                      t_vec <- unique(t_vec)
                      knots_ini <- data.frame(t = t_vec,
                                              BI = ratio_func(as.vector(tapply(a[binding_m], bmk_start[binding_m], mean)),
                                                              s[t_vec]),
                                              extra_knot = FALSE)
                    } else {
                      knots_ini <- data.frame(t  = t_vec,
                                              BI = ratio_func(a[binding_m], s[t_vec]),
                                              extra_knot = FALSE)
                    }
  
                    # Remove missing (`NA`) BI's, i.e. irrelevant spline knots (BI values) corresponding to indicator
                    # values of 0 with proportional benchmarking (0's remain 0's with proportional benchmarking)
                    knots_ini <- knots_ini[!is.na(knots_ini$BI), , drop = FALSE]
                    n_knots_ini <- nrow(knots_ini)
  
                    # Validate the initial set of knots
                    if (n_knots_ini == 0) {
                      quiet_msg_func("Benchmarking is not required for this series (no relevant benchmark remains). ",
                                     "New observations will not be added to the output spline knots data frame.\n")
                      theta <- s
  
                    } else {
  
  
                      # Add extra knots at the start and end (benchmarking timeliness issue)
  
                      # Set K2...
                      #   K2 = the number of high (indicator series) frequency extra knots to add after the
                      #        low (benchmarks) frequency extra knots (K1)
                      #      = the number of periods required to cover the first/last plus 1 year worth of
                      #        high frequency knots
                      K2.beg <- periodicity + max(0, knots_ini$t[1] - K1 * low_freq_periodicity - 1)
                      K2.end <- periodicity + max(0, nT - knots_ini$t[n_knots_ini] - K1 * low_freq_periodicity)
  
                      # Initialize the spline knots with NA for columns "t" and "BI" for the extra knots (for now)
                      # (the 100 extra knots will be used to approximate a "slope=0" spline at both ends)
                      n_extra.beg <- K1 + K2.beg + 100
                      n_extra.end <- K1 + K2.end + 100
                      knots <- rbind(data.frame(t = rep.int(NA_real_, n_extra.beg),
                                                BI = rep.int(NA_real_, n_extra.beg),
                                                extra_knot = rep.int(TRUE, n_extra.beg)),
                                     knots_ini,
                                     data.frame(t = rep.int(NA_real_, n_extra.end),
                                                BI = rep.int(NA_real_, n_extra.end),
                                                extra_knot = rep.int(TRUE, n_extra.end)))
  
                      # K1 low frequency knots
                      kk.beg <- n_extra.beg + 1
                      kk.end <- n_extra.beg + n_knots_ini
                      for (kk in seq_len(K1)) {
                        # ... at the start
                        knots$t[kk.beg - kk] <- knots$t[kk.beg] - kk * low_freq_periodicity
                        knots$BI[kk.beg - kk] <- bk.e$actual_bias + rho^(kk * low_freq_periodicity) *
                          (knots$BI[kk.beg] - bk.e$actual_bias)
                        # ... at the end
                        knots$t[kk.end + kk] <- knots$t[kk.end] + kk * low_freq_periodicity
                        knots$BI[kk.end + kk] <- bk.e$actual_bias + rho^(kk * low_freq_periodicity) *
                          (knots$BI[kk.end] - bk.e$actual_bias)
                      }
  
                      # K2 high frequency knots
                      #... at the start
                      kk.beg <- kk.beg - K1
                      for (kk in seq_len(K2.beg)) {
                        knots$t[kk.beg - kk] <- knots$t[kk.beg - kk + 1] - 1
                        knots$BI[kk.beg - kk] <- bk.e$actual_bias + rho * (knots$BI[kk.beg - kk + 1] - bk.e$actual_bias)
                      }
                      #... at the end
                      kk.end <- kk.end + K1
                      for (kk in seq_len(K2.end)) {
                        knots$t[kk.end + kk] <- knots$t[kk.end + kk - 1] + 1
                        knots$BI[kk.end + kk] <- bk.e$actual_bias + rho * (knots$BI[kk.end + kk - 1] - bk.e$actual_bias)
                      }
  
                      # 100 extra knots for "slope=0" spline approximation
                      kk.beg <- kk.beg - K2.beg
                      kk.end <- kk.end + K2.end
                      for (kk in 1:100) {
                        # ... at the start
                        knots$t[kk.beg - kk] <- knots$t[kk.beg] - kk / 100
                        knots$BI[kk.beg - kk] <- knots$BI[kk.beg]
                        # ... at the end
                        knots$t[kk.end + kk] <- knots$t[kk.end] + kk / 100
                        knots$BI[kk.end + kk] <- knots$BI[kk.end]
                      }
  
                      # Implement the spline interpolations
                      interpol <- stats::spline(knots$t, knots$BI, xout = 1:nT, method = "natural")
                      BI <- interpol$y
  
                      # Benchmark the indicator series by applying the interpolated BI ratios or differences
                      # to the indicator series, taking into account binding indicator values (BI = bias)
                      BI[which(c_s == 0)] <- bk.e$actual_bias
                      theta <- BI_apply_func(s, BI)
  
  
                      # Cumulate the splineKnots output data frame info
                      n_knots <- nrow(knots)
                      out_splineKnots_df <- rbind(out_splineKnots_df,
                                                  cbind((bk.e$ser_df_byGrp[by])[rep.int(1, n_knots), , drop = FALSE],
                                                        data.frame(varSeries = rep.int(var[jj], n_knots),
                                                                   varBenchmarks = rep.int(with[jj], n_knots),
                                                                   x = knots$t,
                                                                   y = knots$BI,
                                                                   extraKnot = knots$extra_knot)))
                    }
                    
                    
                    # Cumulate the graphTable output data frame info
                    #   => the graphTable includes more periods than the indicator series (`n_obs_GT > nT`)
                    #      in the case of duplicate benchmarks (anchor points)
                    
                    # Map benchmark level info (vectors of length `M`) to period level info (vectors of length
                    # `n >= nT`, with `n = nT` for distinct benchmarks and `n > nT` for duplicate benchmarks)
                    #
                    # The resulting `bmk_info` data frame has `n >= nT` observations and 5 variables:
                    #   - `t`: period id
                    #   - `m`: benchmark id (`NA` for periods not associated to any benchmark)
                    #   - `avg_s`: indicator series value  (`NA` for periods not associated to any benchmark)
                    #   - `avg_a`: benchmark value (`NA` for periods not associated to any benchmark)
                    #   - `c_a`: benchmark alter coef (`NA` for periods not associated to any benchmark)
                    bmk_info <- merge(data.frame(t = 1:nT), # period ids
                                      # Benchmark level info
                                      data.frame(t = bmk_per_id, # benchmark (anchor point) period ids
                                                 m = 1:M,
                                                 avg_s = s_lowFreq,
                                                 avg_a = a,
                                                 c_a = c_a),
                                      by = "t", 
                                      all.x = TRUE)

                    # Generate the graphTable data frame info
                    n_obs_GT <- length(bmk_info$t)
                    out_GT_df_var <- data.frame(varSeries = rep.int(var[jj], n_obs_GT),
                                                varBenchmarks = rep.int(with[jj], n_obs_GT),
                                                altSeries = rep.int(actual_alter_ser[jj], n_obs_GT),
                                                altSeriesValue = c_s[bmk_info$t],
                                                altbenchmarks = rep.int(actual_alter_bmk[jj], n_obs_GT),
                                                altBenchmarksValue = bmk_info$c_a,
                                                t = bmk_info$t,
                                                m = bmk_info$m,
                                                year = bk.e$ser_df_byGrp$year[bmk_info$t],
                                                period = bk.e$ser_df_byGrp$period[bmk_info$t],
                                                constant = rep.int(constant, n_obs_GT),
                                                rho = rep.int(rho, n_obs_GT),
                                                lambda = rep.int(lambda, n_obs_GT),
                                                bias = rep.int(bk.e$actual_bias, n_obs_GT),
                                                periodicity = rep.int(periodicity, n_obs_GT),
                                                date = paste(bk.e$ser_df_byGrp$year[bmk_info$t], 
                                                             sprintf("%06d", bk.e$ser_df_byGrp$period[bmk_info$t]), sep = "-"),
                                                subAnnual = s[bmk_info$t],
                                                benchmarked = theta[bmk_info$t],
                                                avgBenchmark = bmk_info$avg_a,
                                                avgSubAnnual = bmk_info$avg_s,
                                                subAnnualCorrected = s_b[bmk_info$t],
                                                benchmarkedSubAnnualRatio = ratio_func(theta[bmk_info$t], s[bmk_info$t]),
                                                avgBenchmarkSubAnnualRatio = ratio_func(bmk_info$avg_a, bmk_info$avg_s),
                                                growthRateSubAnnual = growthRate_func(s)[bmk_info$t],
                                                growthRateBenchmarked = growthRate_func(theta)[bmk_info$t],
                                                stringsAsFactors = FALSE)
  
                    out_graphTable_df <- rbind(out_graphTable_df, cbind(bk.e$ser_df_byGrp[bmk_info$t, by, drop = FALSE],
                                                                        out_GT_df_var))
  
                    # Remove the temporary constant
                    out_ser_df_byGrp[[var[jj]]] <- theta - constant
  
                    # Results validation
                    if (neg_res_func(out_ser_df_byGrp[[var[jj]]], tol = -tolN)) {
                      warning("The benchmarked series contains negative values (threshold = ", format(tolN), ").\n",
                              call. = FALSE, immediate. = TRUE)
                      bk.e$warning_flag <- TRUE
                    }
                    binding_bmk_valid_func(bk.e$bmk_df_byGrp[[with[jj]]], out_ser_df_byGrp[bmk_per_id, var[jj], drop = TRUE], c_a,
                                           bmk_start, bmk_end, tol_parm, zero_tol = 0)
                  }
                }
              }
            }
          }
        }
      }
  
      # Cumulate the output series and benchmarks data frame info
      out_series_df <- rbind(out_series_df, out_ser_df_byGrp)
      out_benchmarks_df <- rbind(out_benchmarks_df, out_bmk_df_byGrp)
    }
  }


  # Create the output list
  row.names(out_series_df) <- NULL
  row.names(out_benchmarks_df) <- NULL
  row.names(out_graphTable_df) <- NULL
  row.names(out_splineKnots_df) <- NULL
  out_list <- list(series = out_series_df, 
                   benchmarks = out_benchmarks_df, 
                   graphTable = out_graphTable_df, 
                   splineKnots = out_splineKnots_df)


  # Display a final warning/error message for multiple series processing
  if (final_msg_flag) {
    if (bk.e$warning_flag) {
      warning("Warnings were generated during processing. See preceeding warning message(s) for details.\n",
              call. = FALSE, immediate. = TRUE)
    }
    # Non-muted error message (for proper condition catching by users of the function)
    if (try_error) {
      stop("Problems were encontered during processing. See preceeding error message(s) for details.\n\n",
           call. = FALSE)
    }

  # Display the error message for single-series processing
  } else if (try_error) {
    stop(try_error_msg, call. = FALSE)
  }

  # Output object returned via function `on.exit()`
}

#' Convert reconciliation metadata
#' 
#' 
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/rkMeta_to_blSpecs.html}})}
#' 
#' Convert a [tsraking()] metadata data frame to a [tsbalancing()] problem specs data frame.
#'
#'
#' @usage
#' rkMeta_to_blSpecs(
#'   metadata_df,
#'   alterability_df = NULL,
#'   alterSeries = 1, 
#'   alterTotal1 = 0,
#'   alterTotal2 = 0,
#'   alterability_df_only = FALSE
#' )
#'
#'
#' @inheritParams tsraking
#' 
#' @param alterability_df_only (optional)
#'
#' Logical argument specifying whether or not only the set of alterability ceofficients found in the alterability 
#' file (argument `alterability_df`) should be included in the returned [tsbalancing()] problem specs data frame.
#' When `alterability_df_only = FALSE` (the default), the alterability coefficients specified with arguments 
#' `alterSeries`, `alterTotal1` and `alterTotal2` are combined with those found in `alterability_df` (the latter 
#' coefficients overwriting the former) and the returned data frame therefore contains alterability coefficients 
#' for all component and cross-sectional control total series. This argument does not affect the set of temporal 
#' total alterability coefficients (associated to [tsraking()] argument `alterAnnual`) that are included in the 
#' returned [tsbalancing()] problem specs data frame. The latter always strictly contains those specified in 
#' `metadata_df` with a non-missing (non-`NA`) value for column `alterAnnual`.
#'
#' **Default value** is `alterability_df_only = FALSE`.
#'
#'
#' @details
#' 
#' The preceding description of argument `alterability_df` comes from [tsraking()]. This function 
#' (`rkMeta_to_blSpecs()`) slightly changes the specification of alterability coefficients with argument 
#' `alterability_df` by allowing either
#' * a single observation, specifying the set of alterability coefficients to use for all periods, 
#' * or one or several observations with an additional column named `timeVal` allowing the specification of 
#' both period-specific alterability coefficients (`timeVal` is not `NA`) and generic coefficients to use for 
#' all other periods (`timeVal` is `NA`). Values for column `timeVal` correspond to the *time values* 
#' of a "ts" object as returned by [stats::time()], conceptually corresponding to \eqn{year + (period - 1) / 
#' frequency}.
#' 
#' Another difference with [tsraking()] is that missing (`NA`) values are allowed in the alterability coefficients 
#' data frame (argument `alterability_df`) and result in using the generic coefficients (observations for which 
#' `timeVal` is `NA`) or the default coefficients (arguments `alterSeries`, `alterTotal1` and `alterTotal2`).
#'
#' Note that apart from discarding alterability coefficients for series not listed in the [tsraking()] metadata 
#' data frame (argument `metadata_df`), this function does not validate the values specified in the alterability 
#' coefficients data frame (argument `alterability_df`) nor the ones specified with column `alterAnnual` in the 
#' [tsraking()] metadata data frame (argument `metadata_df`). The function transfers them *as is* in the 
#' returned [tsbalancing()] problem specs data frame.
#'
#'
#' @returns
#' A [tsbalancing()] problem specs data frame (argument `problem_specs_df`).
#' 
#' 
#' @seealso [tsraking()] [tsbalancing()]
#'
#'
#' @example misc/function_examples/rkMeta_to_blSpecs-ex.R
#'
#'
#' @export
rkMeta_to_blSpecs <- function(metadata_df,
                              alterability_df = NULL,
                              alterSeries = 1, 
                              alterTotal1 = 0,
                              alterTotal2 = 0,
                              alterability_df_only = FALSE) {
  
  # Enforce the default R "error" option (`options(error = NULL)`). E.g. this Turns off traceback
  # generated by calls to the stop() function inside internal functions in R Studio.
  ini_error_opt <- getOption("error")
  on.exit(options(error = ini_error_opt))
  options(error = NULL)
  
  # Validate `metadata_df`
  if (nchar(deparse1(substitute(metadata_df))) == 0) {
    stop("Argument 'metadata_df' is mandatory (it must be specified).\n\n", call. = FALSE)
  }
  metadata_df <- metadata_df
  if (!is.data.frame(metadata_df)) {
    stop("Argument 'metadata_df' is not a 'data.frame' object.\n\n", call. = FALSE)
  }
  metadata_df <- as.data.frame(metadata_df)
  metaDF_cols <- toupper(names(metadata_df))
  names(metadata_df) <- metaDF_cols
  row.names(metadata_df) <- NULL

  # Validate the alterability coefficients arguments
  tmp <- (unlist(alterSeries))[1]
  if (!identical(alterSeries, tmp) || is.null(tmp) || !is.finite(tmp) || tmp < 0) {
    stop("Argument 'alterSeries' must be a nonnegative real number.\n\n", call. = FALSE)
  }
  tmp <- (unlist(alterTotal1))[1]
  if (!identical(alterTotal1, tmp) || is.null(tmp) || !is.finite(tmp) || tmp < 0) {
    stop("Argument 'alterTotal1' must be a nonnegative real number.\n\n", call. = FALSE)
  }
  tmp <- (unlist(alterTotal2))[1]
  if (!identical(alterTotal2, tmp) || is.null(tmp) || !is.finite(tmp) || tmp < 0) {
    stop("Argument 'alterTotal2' must be a nonnegative real number.\n\n", call. = FALSE)
  }
  
  # Validate `alterability_df_only`
  alterability_df_only <- gs.validate_arg_logi(alterability_df_only)

  
  # Generate a fake series data (data frame)
  data_cols <- unique(gs.cleanup_col_list(metadata_df[intersect(c("SERIES", "TOTAL1", "TOTAL2"), 
                                                             metaDF_cols)]))
  n_ser <- length(data_cols)
  fake_df <- as.data.frame(matrix(rep(0, n_ser), 
                                  nrow = 1,
                                  dimnames = list(NULL, data_cols)))

  
  # Process the alterability coefficients data frame
  if (!is.null(alterability_df)) {
    alterability_df <- alterability_df
    if (is.data.frame(alterability_df)) {
      alterability_df <- as.data.frame(alterability_df)
      alterDF_cols <- toupper(names(alterability_df))
      alter_ser <- setdiff(alterDF_cols, "TIMEVAL")
      names(alterability_df) <- alterDF_cols
      row.names(alterability_df) <- NULL
      
      # Split period-specific (dated) and generic (non-dated) alterability coefficients
      if ("TIMEVAL" %in% alterDF_cols) {
        tmp <- !is.na(alterability_df$TIMEVAL)
        dated_id <- which(tmp)
        if (length(dated_id) > 0) {
          dated_alter_df <- alterability_df[dated_id, , drop = FALSE]
        } else {
          dated_alter_df <- alterability_df[NULL, , drop = FALSE]
        }
        nondated_id <- which(!tmp)
        if (length(nondated_id) > 0) {
          nondated_alter_df <- alterability_df[nondated_id[1], alter_ser, drop = FALSE]
        } else {
          nondated_alter_df <- alterability_df[NULL, alter_ser, drop = FALSE]
        }
        
      } else {
        if (nrow(alterability_df) > 1) {
          warning("The specified 'alterability_df' contains more than one observation but does not include column ", 
                  "`timeVal`. Only the set of alterability coefficients from the first observation will be used.\n", 
                  call. = FALSE, immediate. = TRUE)
          nondated_alter_df <- alterability_df[1, , drop = FALSE]
        } else {
          nondated_alter_df <- alterability_df
        }
        dated_alter_df <- nondated_alter_df
        dated_alter_df$TIMEVAL <- NA_real_
        dated_alter_df <- dated_alter_df[NULL, , drop = FALSE]
      }
      
      # Enforce matching series names (case-wise) and reject series that don't exist in `metadata_df`
      data_cols_uc <- toupper(data_cols)
      alter_col_id <- match(alter_ser, data_cols_uc, nomatch = 0)
      nondated_alter_df <- nondated_alter_df[data_cols_uc[alter_col_id]]
      alter_cols <- data_cols[alter_col_id]
      names(nondated_alter_df) <- alter_cols
      dated_alter_df <- dated_alter_df[c("TIMEVAL", data_cols_uc[alter_col_id])]
      names(dated_alter_df) <- c("TIMEVAL", alter_cols)
      
      # Reject missing (NA) generic (non-dated) alter coefs, if any
      nondated_alter_df <- nondated_alter_df[, !is.na(nondated_alter_df), drop = FALSE]
      if (nrow(nondated_alter_df) == 0 || ncol(nondated_alter_df)  == 0) {
        nondated_alter_df <- NULL
      }
      
    } else {
      # Accept `alterability_df = NA` as `alterability_df = NULL`
      tmp <- (unlist(alterability_df))[1]
      if (!identical(alterability_df, tmp) || !is.na(tmp)) {
        warning("Argument 'alterability_df' is not a 'data.frame' object. It will be ignored.\n",
                call. = FALSE, immediate. = TRUE)
        alterability_df <- NULL
      }
      nondated_alter_df <- NULL
      dated_alter_df <- as.data.frame(NULL)
    }
  } else {
    nondated_alter_df <- NULL
    dated_alter_df <- as.data.frame(NULL)
  }

  
  # Build the elements of the raking problem (while validating the specified info), 
  # excluding the component series temporal totals info:
  #   - x        : vector of component series initial values
  #   - c_x      : vector of component series alterability coefficients
  #   - comp_cols: vector of component series (column) names 
  #   - g        : vector of marginal total series initial values
  #   - c_g      : vector of marginal total series alterability coefficients
  #   - tot_cols : vector of marginal total series (column) names 
  #   - G        : marginal total series aggregation matrix (`g = G %*% x` for coherent/reconciled data)
  #
  # The returned raking problem elements do not include the implicit component series temporal totals 
  # when applicable (i.e., elements `g` and `G` only contain the marginal totals info).
  #
  rk <- build_raking_problem(data_df         = fake_df,
                             metadata_df     = metadata_df,
                             alterability_df = nondated_alter_df,
                             alterSeries     = alterSeries, 
                             alterTotal1     = alterTotal1,
                             alterTotal2     = alterTotal2)
  all_cols <- c(rk$comp_cols, rk$tot_cols)
  
  
  # Start with the balancing constraints 
  n_con <- length(rk$tot_cols)
  specs_df <- data.frame(sort.1 = 1:n_con,
                         sort.2 = rep.int(0, n_con),
                         type = rep.int("EQ", n_con),
                         col = rep.int(NA_character_, n_con),
                         row = paste0("Marginal Total ", 1:n_con, " (", rk$tot_cols, ")"),
                         coef = rep.int(NA_real_, n_con),
                         timeVal = rep.int(NA_real_, n_con))
  
  for (ii in 1:n_con) {
    comp_id <- which(rk$G[ii, ] == 1)
    n_comp <- length(comp_id)
    specs_df <- rbind(specs_df,
                      data.frame(sort.1 = rep.int(ii, n_comp + 1),
                                 sort.2 = 1:(n_comp + 1),
                                 type = rep.int(NA_character_, n_comp + 1),
                                 col = c(rk$comp_cols[comp_id], rk$tot_cols[ii]),
                                 row = rep.int(paste0("Marginal Total ", ii, " (", rk$tot_cols[ii], ")"), n_comp + 1),
                                 coef = c(rep.int(1, n_comp), -1),
                                 timeVal = rep.int(NA_real_, n_comp + 1)))
  }
  
  specs_df <- specs_df[order(specs_df$sort.1, specs_df$sort.2), ]

  
  # Add the generic (non-dated) alterability coefficients (from `alterability_df' only or not)
  if (!alterability_df_only) {
    nondated_alter <- TRUE
    alter_df <- data.frame(sort = 0:n_ser,
                           type = c("alter", rep(NA_character_, n_ser)),
                           col = c(NA_character_, all_cols),
                           row = rep.int("Period Value Alterability", n_ser + 1),
                           coef = c(NA_real_, rk$c_x, rk$c_g),
                           timeVal = rep.int(NA_real_, n_ser + 1))
    
    
  } else if (!is.null(nondated_alter_df)) {  
    nondated_alter <- TRUE
    col_id <- match(names(nondated_alter_df), all_cols)
    n_alter <- length(col_id)
    alter_df <- data.frame(sort = c(0, col_id),
                           type = c("alter", rep(NA_character_, n_alter)),
                           col = c(NA_character_, all_cols[col_id]),
                           row = rep.int("Period Value Alterability", n_alter + 1),
                           coef = c(NA_real_, c(rk$c_x, rk$c_g)[col_id]),
                           timeVal = rep.int(NA_real_, n_alter + 1))

  } else {
    nondated_alter <- FALSE
    alter_df <- NULL
  }
  
  
  # Add the period-specific (dated) alterability coefficients
  if (nrow(dated_alter_df) > 0) {
    
    dated_alter <- FALSE
    for (ii in seq_along(alter_cols)) {
      logi_vec <- !is.na(dated_alter_df[[alter_cols[ii]]])
      n_alter <- sum(logi_vec)
      if (n_alter > 0) {
        dated_alter <- TRUE
        alter_df <- rbind(alter_df,
                          data.frame(sort = rep.int(alter_col_id[ii], n_alter),
                                     type = rep.int(NA_character_, n_alter),
                                     col = rep.int(alter_cols[ii], n_alter),
                                     row = rep.int("Period Value Alterability", n_alter),
                                     coef = dated_alter_df[[alter_cols[ii]]][logi_vec],
                                     timeVal = dated_alter_df[["TIMEVAL"]][logi_vec]))
      }
    }
    
    # Add the label definition record (if needed)
    if (!nondated_alter && dated_alter) {
      alter_df <- rbind(data.frame(sort = 0,
                                   type = "alter",
                                   col = NA_character_,
                                   row = "Period Value Alterability",
                                   coef = NA_real_,
                                   timeVal = NA_real_),
                        alter_df)
    }    
  }
  
    
  # Add the period value alterability coefficients
  if (!is.null(alter_df)) {
    alter_df <- alter_df[order(alter_df$sort, alter_df$timeVal, na.last = FALSE), ]
    specs_df <- rbind(specs_df[, -c(1, 2)], alter_df[, -1])
  } else  {
    specs_df <- specs_df[, -c(1, 2)]
  }

  
  # Add the temporal total alterability coefficients
  if ("ALTERANNUAL" %in% metaDF_cols) {
    logi_vec <- !is.na(metadata_df$ALTERANNUAL)
    n_comp <- sum(logi_vec)
    if (n_comp > 0) {
      specs_df <- rbind(specs_df,
                        data.frame(type = c("alterTmp", rep(NA_character_, n_comp)),
                                   col = c(NA_character_, metadata_df$SERIES[logi_vec]),
                                   row = rep.int("Temporal Total Alterability", n_comp + 1),
                                   coef = c(NA_real_, metadata_df$ALTERANNUAL[logi_vec]),
                                   timeVal = rep(NA_real_, n_comp + 1)))
    }
  }
  
  
  # Reset the row labels
  row.names(specs_df) <- NULL
  specs_df
}

#' Stack benchmarks data
#'
#'
#' @description
#'
#' \if{html,text}{(\emph{version française: 
#' \url{https://StatCan.github.io/gensol-gseries/fr/reference/stack_bmkDF.html}})}
#' 
#' Convert a multivariate benchmarks data frame (see [ts_to_bmkDF()]) for the benchmarking functions
#' ([benchmarking()] and [stock_benchmarking()]) into a stacked (tall) data frame with six variables (columns):
#' * one (1) for the benchmark name (e.g., series name)
#' * four (4) for the benchmark coverage
#' * one (1) for the benchmark value
#'
#' Missing (`NA`) benchmark values are not included in the output stacked data frame by default. Specify argument 
#' `keep_NA = TRUE` in order to keep them.
#'
#' This function is useful when intending to use the `by` argument (*BY-group* processing mode) of the benchmarking
#' functions in order to benchmark multiple series in a single function call.
#'
#'
#' @param bmk_df (mandatory)
#'
#' Data frame (object of class "data.frame") that contains the multivariate benchmarks to be stacked.
#'
#' @param ser_cName (optional)
#'
#' String specifying the name of the character variable (column) in the output stacked data frame that will
#' contain the benchmark names (name of the benchmark variables in the input multivariate benchmarks data
#' frame). This variable can then be used as the BY-group variable (argument `by`) with the benchmarking functions.
#'
#' **Default value** is `ser_cName = "series"`.
#'
#' @param startYr_cName,startPer_cName,endYr_cName,endPer_cName (optional)
#'
#' Strings specifying the name of the numeric variables (columns) in the input multivariate benchmarks data frame
#' that define the benchmark coverage, i.e., the starting and ending year and period (cycle) identifiers. These variables
#' are *transferred* to the output stacked data frame with the same variable names.
#'
#' **Default values** are `startYr_cName = "startYear"`, `startPer_cName = "startPeriod"`
#' `endYr_cName = "endYear"` and `endPer_cName   = "endPeriod"`.
#'
#' @param val_cName (optional)
#'
#' String specifying the name of the numeric variable (column) in the output stacked data frame that will
#' contain the benchmark values.
#'
#' **Default value** is `val_cName = "value"`.
#'
#' @param keep_NA (optional)
#'
#' Logical argument specifying whether missing (`NA`) benchmark values in the input multivariate benchmarks data frame
#' should be kept in the output stacked data frame.
#'
#' **Default value** is `keep_NA = FALSE`.
#'
#'
#' @returns
#' The function returns a data frame with six variables:
#' * Benchmark (series) name, type character (see argument `ser_cName`)
#' * Benchmark coverage starting year, type numeric (see argument `startYr_cName`)
#' * Benchmark coverage starting period, type numeric (see argument `startPer_cName`)
#' * Benchmark coverage ending year, type numeric (see argument `endtYr_cName`)
#' * Benchmark coverage ending period, type numeric (see argument `endPer_cName`)
#' * Benchmark value, type numeric (see argument `val_cName`)
#'
#' Note: the function returns a "data.frame" object than can be explicitly coerced to another type of object 
#' with the appropriate `as*()` function (e.g., `tibble::as_tibble()` would coerce it to a tibble).
#'
#'
#' @seealso [stack_tsDF()] [ts_to_bmkDF()] [benchmarking()] [stock_benchmarking()]
#'
#'
#' @example misc/function_examples/stack_bmkDF-ex.R
#'
#'
#' @export
stack_bmkDF <- function(bmk_df,
                        ser_cName = "series",
                        startYr_cName = "startYear",
                        startPer_cName = "startPeriod",
                        endYr_cName = "endYear",
                        endPer_cName = "endPeriod",
                        val_cName = "value",
                        keep_NA = FALSE) {
  
  # Enforce the default R "error" option (`options(error = NULL)`). E.g. this Turns off traceback
  # generated by calls to the stop() function inside internal functions in R Studio.
  ini_error_opt <- getOption("error")
  on.exit(options(error = ini_error_opt))
  options(error = NULL)
  
  # validate object
  if (!is.data.frame(bmk_df)) {
    stop("Argument 'bmk_df' is not a 'data.frame' object.\n\n", call. = FALSE)
  }
  bmk_df <- as.data.frame(bmk_df)
  df_cols <- names(bmk_df)
  date_cols <- c(startYr_cName, startPer_cName, endYr_cName, endPer_cName)
  date_args <- c("startYr_cName", "startPer_cName", "endYr_cName", "endPer_cName")
  for (ii in seq_along(date_cols)) {
    if (!(date_cols[ii] %in% df_cols)) {
      stop("The input data frame does not contain column \"", date_cols[ii], "\" (argument '",
           date_args[ii], "').\n\n", call. = FALSE)
    }
  }
  
  ser_list <- setdiff(df_cols, date_cols)
  out_df <- data.frame(col1 = character(),
                       col2 = integer(),
                       col3 = integer(),
                       col4 = integer(),
                       col5 = integer(),
                       col6 = double(),
                       stringsAsFactors = FALSE)
  for (ser in ser_list) {
    if (gs.validate_arg_logi(keep_NA)) {
      tmp_df <- bmk_df[c(date_cols, ser)]
    } else {
      tmp_df <- bmk_df[!is.na(bmk_df[ser]), c(date_cols, ser)]
    }
    out_df <- rbind(out_df, data.frame(col1 = ser,
                                       col2 = as.integer(tmp_df[[startYr_cName]]),
                                       col3 = as.integer(tmp_df[[startPer_cName]]),
                                       col4 = as.integer(tmp_df[[endYr_cName]]),
                                       col5 = as.integer(tmp_df[[endPer_cName]]),
                                       col6 = as.numeric(tmp_df[[ser]]),
                                       stringsAsFactors = FALSE))
  }
  
  # Set the column names and reset the now names (numbers)
  names(out_df) <- c(ser_cName, date_cols, val_cName)
  row.names(out_df) <- NULL
  out_df  
}

---
title: "Benchmarking Cookbook"
output:
  rmarkdown::html_vignette: null
  pdf_document: null
  html_document:
    df_print: paged
urlcolor: blue
linkcolor: blue
vignette: >
  %\VignetteIndexEntry{Benchmarking Cookbook}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, include = FALSE}
library(gseries)
```

<!-- Display the link to the French web page only when rendering an HTML document
     (e.g., not when rendering the PDF version)
     
     => the Pandoc "fenced_div" below (::: {.pkgdown-devel} <...> :::) is used to avoid 
        having the link generated in the pkgdown website vignette page
     => the link would only show in the "development" version of the pkgdown website
        (`development: mode: devel` in `_pkdown.yml` or `development: mode: auto` with a 4-level 
        version number in the DESCRIPTION file), which we do not use for gseries (we set 
        `development: mode: release` in `_pkdown.yml`, resulting in a single "release" website 
        regardless of the version number -->
::: {.pkgdown-devel}
```{asis, echo=knitr::is_html_output()}
(*version française : <https://StatCan.github.io/gensol-gseries/fr/articles/benchmarking-cookbook.html>*)
```
:::


Description of the steps of a typical benchmarking project with G-Series (R package gseries).

Other useful resources related to the time series benchmarking problem discussed here and solved using the G-Series benchmarking functions (`benchmarking()` and `stock_benchmarking()`), in ascending order of technical complexity:

- Fortier and Quenneville (2007), in ***References*** for function `benchmarking()`, for an overview of the time series benchmarking methodology implemented in G-Series including detailed examples. This document is available in GCdocs for Statistics Canada employees (search for "ICES2007_Fortier.pdf" in GCDocs).

- Course 0436, “Theory and Application of Benchmarking”, in ***References*** for function `benchmarking()` and `stock_benchmarking()`. Visit the [Course 0436](https://www.statcan.gc.ca/en/training/statistical/0436) web page (Statistics Canada general public website) for more details.

- Dagum and Cholette (2006), in ***References*** for function `benchmarking()`, for a complete technical discussion and presentation of time series benchmarking problems and their solution.

<br>

***

<br>

### 1. Prepare the input data

The first step usually involves converting "ts" or "mts" objects (stats package) into the proper format for the G-Series benchmarking functions with the following two utility functions;

- `ts_to_tsDF()` for the indicator series

- `ts_to_bmkDF()` for the benchmarks

It is possible to benchmark multiple series in a single call to the benchmarking functions. This can be done by specifying the appropriate list of data frame variables with arguments `var` and `with`, which can be cumbersome and make your code look _cluttered_. A more convenient option to achieve the same result would be to use the `allCols` argument. However, these two alternatives have important limitations as they both require that all the indicator series are of the same length (same number of periods) and have the same set (number) of benchmarks. The values of the benchmarks can obviously differ for each indicator series, but their coverage must be the same.

A more flexible approach, which doesn't suffer from the aforementioned limitations, is to use the _BY-group_ mode (argument `by`) of the benchmarking functions after having converted the input data frames into _stacked_ (tall) versions with the following utility functions:

- `stack_tsDF()` for the indicator series

- `stack_bmkDF()` for the benchmarks

Stacked versions of the data frames use only two variables to specify information about the different indicator series or benchmarks: one variable for the identifiers and another for the values. A stacked data frame therefore contains more rows but fewer variables (columns) than a non-stacked data frame as the time series are _stacked_ on top of each other instead of being _layed out_ side by side. _BY-group_ processing with stacked data frames is the recommended approach to benchmark several series in a single benchmarking function call, unless the number of series to benchmark is extremely large and processing time is a really important issue (processing multiple time series with arguments `var` or `allCols` should be slightly faster than the BY-group approach with argument `by`).

<br>

### 2. Run benchmarking

The number of calls to the benchmarking functions depends on the values of arguments `rho`, `lambda`, `biasOption` and `bias` (plus `low_freq_periodicity`, `n_low_freq_proj` and `proj_knots_rho_bd` for function `stock_benchmarking()`). One call is necessary for each distinct combination of these argument values. In practice, however, only argument `lambda` will usually require distinct values to process the entire set of series: `lambda = 1` for proportional benchmarking and `lambda = 0` for additive benchmarking. Two calls to the benchmarking functions are therefore often enough.

When more than one call is required, the input indicator series and benchmarks data frames need to be split into distinct data frames: one for each call with the relevant set of indicator series and benchmarks. Alternatively, one can add one or several columns to the stacked versions of the input data frames in order to identify (and extract) the indicator series and benchmarks of each call.

**Note on proportional and additive benchmarking**\
Proportional benchmarking ($\lambda \ne 0$) is normally used when the main focus is the preservation of period-to-period ratios (relative differences) and additive benchmarking ($\lambda = 0$) for the preservation of period-to-period differences. Focusing on period-to-period ratios is usually preferable for time series for which the amplitude (e.g., seasonal and irregular components) changes with the level of the series. On the other hand, if the amplitude of the series stays relatively constant regardless of the level of the series, then looking at period-to-period differences is appropriate.

<center> ![**Series Amplitude**](amplitude.png){width=80%} </center>
<p line-height: 0.5> </p>

The most common _pitfall_ would likely be using an additive benchmarking approach when changes in the amplitude of the indicator series are important and follow the level (the amplitude increases/decreases with the level). The main advantage of additive benchmarking is that it _works_ (returns _a solution_) in all contexts while proportional benchmarking will fail in some particular cases (e.g., nonzero benchmark with zero indicator series values for all periods covered by the benchmark). Proportional benchmarking with $\rho < 1$ (regression-based benchmarking) generally works well in practice (provides _reasonable_ solutions) with problems involving values of zero for the indicator series and/or the benchmarks. Some people may actually appreciate (find appealing) the fact that values of zero in the initial indicator series always remain zero in proportionally benchmarked series, which is not the case for additively benchmarked series. As for proportional benchmarking with $\rho = 1$ (Denton benchmarking), the indicator series must be strictly positive. However, one could try using argument `constant` in order to add a (relatively small) temporary constant to the input data and solve cases that involve values of zero in the indicator series. In practice, note that proportional Denton benchmarking could also be _approximated_ with the regression-based approach by using a $\rho$ value that is smaller than, but very close to, $1.0$ (e.g., $\rho = 0.999$). Finally, although proportional benchmarking ($\lambda \ne 0, \forall \rho$) is not allowed by default in presence of negative values, this behaviour can be modified with argument `negInput_option`. In any case, one should closely monitor proportionally benchmarked series involving negative values or values of zero (or almost zero) where ratios may be undefined, unstable or difficult to interpret. The resulting proportionally benchmarked data should be carefully analyzed and validated in such cases to make sure they correspond to reasonable, interpretable solutions. 

<br>

### 3. Validate the results

Any warning and error message generated by the benchmarking functions should be investigated in order to fix the corresponding issue(s). Once clean executions of the benchmarking functions (free of warning and error messages) are obtained, one should validate the resulting benchmarked series data. Utility function `plot_graphTable()` generates useful graphics to perform such a task.

Examples of things to look for in the benchmarking results:

- **Inadequate projected benchmarking adjustments**. The benchmarking solution for periods not covered by a benchmark at the end of the series is driven by parameter $\rho$ (argument `rho`) and the specified bias adjustment (arguments `biasOption`and `bias`). Bias adjustment is generally recommended when the levels of the benchmarks and the indicator series are systematically different (e.g., benchmarks always, or almost always, larger than the indicator series or vice versa). Not correcting for an (important) bias  may result in poor results for the benchmarked series for periods not covered by a benchmark (poor projected benchmarking adjustments), which may then lead to large revisions as new benchmarks become available in the future. Correcting for the bias should help in such cases. An exception is _Denton benchmarking_ ($\rho = 1$) where bias adjustment has no impact on the benchmarking solution. Changing the value of parameter $\rho$, which dictates the speed at which the projected adjustments converge to the bias for periods not covered by a benchmark, may also improve the situation. The smaller the value of $\rho$ the faster the convergence, with immediate convergence when $\rho = 0$ and no convergence at all (the adjustment of the last period covered by a benchmark is repeated) when $\rho = 1$ (Denton benchmarking). A general recommendation that works reasonably well in most cases is to adjust with the estimated average bias (`biasOption = 3` and `bias = NA`) and use $\rho = 0.9$ for monthly indicators and $\rho = 0.9^3 = 0.729$ for quarterly indicators. Specifying a user-defined bias (argument `bias`) may be relevant if the discrepancies between the two sources of data have changed over time (e.g., specifying a value more representative of the _recent bias_). An alternative would be to use _explicit_ forecasts for the benchmarks at the end of the series instead of relying on the _implicit_ projected benchmarks associated to the bias adjustment and parameter $\rho$. For example, available auxiliary information could be used to generate explicit benchmarks and (if the projections are good) reduce revisions once the true benchmark values are known. The first two benchmarking graphs (_Original Scale Plot_ and _Adjustment Scale Plot_) of function `plot_graphTable()` should help identify potential issues with the projections.

- **Inadequate autoregressive parameter $\rho$ (argument `rho`)**. The goal of benchmarking usually is to preserve the period to period movements of the indicator series, to which correspond values of $\rho$ relatively close to 1 and smooth benchmarking adjustments. This being said, some particular cases may warrant small values of $\rho$ corresponding to _less smooth_ adjustments and reduced movement preservation. The 2^nd^ benchmarking graph (_Adjustment Scale Plot_) of function `plot_graphTable()` shows the benchmarking adjustments and can therefore be used to assess the _smoothness_ of the adjustments and modify, if necessary, the value of parameter $\rho$. The corresponding degree of movement preservation is illustrated by the 3^rd^ and 4^th^ benchmarking graphs (_Growth Rates Plot_ and _Table_) of function `plot_graphTable()`. Note that the benchmarking adjustments can also be plotted using utility function `plot_benchAdj()`.

- **Inadequate adjustment model parameter $\lambda$ (argument `lambda`)**. Additive benchmarking is implemented when $\lambda = 0$ and proportional benchmarking otherwise (when $\lambda \ne 0$). Choosing the _ideal_ adjustment model may not necessarily be obvious. Refer to the _Note on proportional and additive benchmarking_ above for some insights. Trying both an additive and a proportional benchmarking approach and comparing the benchmarking graphics may help choose the benchmarking adjustment model. For example, the approach that generates a more natural looking benchmarked series in the _Original Scale Plot_ (1^st^ graph), smoother benchmarking adjustments in the _Adjustment Scale Plot_ (2^nd^ graph) and better movement preservation in the _Growth Rates Plot_ and _Table_ (3^rd^ and 4^th^ graphs) should be favoured. Looking at the benchmarking graphics of function `plot_graphTable()` should also help identify problematic solutions that may require a change of adjustment model. For example, problematic cases of proportional benchmarking problems with negative values (see argument `negInput_option`) or values of zero or almost zero for the benchmarks or the indicator series would most likely generate _odd looking_ benchmarked series in the _Original Scale Plot_, extreme or non-smooth adjustments in the _Adjustment Scale Plot_ or poor movement preservation in the _Growth Rates Plot_  and _Table_. An additive benchmarking approach may be a better alternative in such cases.

Systematically looking at all the graphics may not necessarily be feasible in practice for large benchmarking projects involving many series. A (crude) classification analysis of the benchmarking functions' `graphTable` output data frame (the input of function `plot_graphTable()`) may help identify cases requiring further investigation and a _closer look_ at the benchmarking graphics.

**Note on stock series**\
Benchmarking stock series with the `benchmarking()` function generates non smooth adjustments (_kinks_) around each benchmark, regardless of the values of $\rho$ and $\lambda$. This is due to the nature of the benchmarks, i.e., discrete values covering a single period (anchor points). Function `stock_benchmarking()`, specifically aimed at benchmarking stock series, usually provides better results (i.e., improved movement preservation and smoother adjustments). Function `plot_benchAdj()` is especially useful to compare (overlay) the adjustments of stock series generated by functions `benchmarking()` and `stock_benchmarking()`. See ***Details*** for the `stock_benchmarking()` function for more information on the benchmarking of stock series. 

<br>

### 4. Process the benchmarked data

The final step usually involves converting the output benchmarked series data (benchmarking functions' `series` output data frame) into "ts" (or "mts") objects with utility function `tsDF_to_ts()`. When _BY-group_ processing (`by` argument) is used, one would first need to _unstack_ the benchmarked series data using utility function `unstack_tsDF()` before calling the `tsDF_to_ts()` function.

**Nonbinding benchmarks**\
Although benchmarking problems involving nonbinding benchmarks (benchmarking alterability coefficients greater than 0) are relatively rare in practice, it's important to remember that the benchmarking functions' output `benchmarks` data frame always contains the original (unmodified) benchmarks provided as input. In such cases, the modified nonbinding benchmarks would be recovered (calculated) from the output `series` data frame instead. For example, flows resulting from a `benchmarking()` call can be aggregated using function `stats::aggregate.ts()` after having converted the `series` output data frame into a "ts" object with utility function `tsDF_to_ts()`.

---
title: "A Beginner's Benchmarking Demo Script"
output:
  rmarkdown::html_vignette: null
  pdf_document: null
  html_document:
    df_print: paged
urlcolor: blue
linkcolor: blue
vignette: >
  %\VignetteIndexEntry{A Beginner's Benchmarking Demo Script}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, include = FALSE}
library(gseries)
```

<!-- Display the link to the French web page only when rendering an HTML document
     (e.g., not when rendering the PDF version)
     
     => the Pandoc "fenced_div" below (::: {.pkgdown-devel} <...> :::) is used to avoid 
        having the link generated in the pkgdown website vignette page
     => the link would only show in the "development" version of the pkgdown website
        (`development: mode: devel` in `_pkdown.yml` or `development: mode: auto` with a 4-level 
        version number in the DESCRIPTION file), which we do not use for gseries (we set 
        `development: mode: release` in `_pkdown.yml`, resulting in a single "release" website 
        regardless of the version number -->
::: {.pkgdown-devel}
```{asis, echo=knitr::is_html_output()}
(*version française : <https://StatCan.github.io/gensol-gseries/fr/articles/benchmarking-demo-script.html>*)
```
:::


```{r, eval = FALSE}
# Load G-Series in R (package gseries)
library(gseries)

# Set the working directory (for the PDF graph files)
iniwd <- getwd() 
setwd("C:/Temp")




#### Example 1 (single quarterly series) ####
# Simple case with a single quarterly series to benchmark to annual values

# Quarterly indicator series
my_ts_qtr <- ts(c(1.9, 2.4, 3.1, 2.2, 2.0, 2.6, 3.4, 2.4, 2.3), 
                start = c(2015, 1), 
                frequency = 4)
my_series1 <- ts_to_tsDF(my_ts_qtr)

# Annual benchmarks for quarterly data
my_ts_ann <- ts(c(10.3, 10.2), 
                start = 2015, 
                frequency = 1)
my_benchmarks1 <- ts_to_bmkDF(my_ts_ann, ind_frequency = 4)

# Benchmarking using...
#   - recommended `rho` value for quarterly series (`rho = 0.729`)
#   - proportional model (`lambda = 1`)
#   - bias-corrected indicator series with the estimated bias (`biasOption = 3`)
out_bench1 <- benchmarking(my_series1,
                           my_benchmarks1,
                           rho = 0.729,
                           lambda = 1,
                           biasOption = 3)

# Outputs
View(out_bench1$series)
View(out_bench1$benchmarks)
View(out_bench1$graphTable)

# Graphics (default set of plots)
plot_graphTable(out_bench1$graphTable, 
                "Ex1_graphs.pdf")
# With the G. R. Table
plot_graphTable(out_bench1$graphTable, 
                "Ex1_graphs_with_GRTable.pdf",
                GR_table_flag = TRUE)


# Compare with package tempdisagg that allows Denton benchmarking 
# (modified version by Cholette)

#install.packages("tempdisagg")
library(tempdisagg)
#install.packages("dplyr")
library(dplyr)


# Proportional Denton
mult_denton <- benchmarking(my_series1,
                            my_benchmarks1,
                            rho = 1,
                            lambda = 1)

td_mult_denton <- td(formula = my_ts_ann ~ 0 + my_ts_qtr,
                     method = "denton-cholette",
                     criterion = "proportional")

compare_mult <- ts.union(gseries = tsDF_to_ts(mult_denton$series, frequency = 4),
                         tempdisagg = predict(td_mult_denton),
                         dframe = TRUE) %>%
  cbind(., diff = .[, 1] - .[, 2])

# Same results
View(compare_mult)


# Additive Denton
add_denton <- benchmarking(my_series1,
                           my_benchmarks1,
                           rho = 1,
                           lambda = 0)

td_add_denton <- td(formula = my_ts_ann ~ 0 + my_ts_qtr,
                    method = "denton-cholette",
                    criterion = "additive")

compare_add <- ts.union(gseries = tsDF_to_ts(add_denton$series, frequency = 4),
                        tempdisagg = predict(td_add_denton),
                        dframe = TRUE) %>%
  cbind(., diff = .[, 1] - .[, 2])

# Same results
View(compare_add)




#### Example 2 (multiple quarterly series) ####
# Two quarterly series to benchmark to annual values, 
# with BY-groups and user-defined alterability coefficients

# Sales data (same sales for groups A and B; only alter coefs for van sales differ)
qtr_sales <- ts(matrix(c(# Car sales
  1851, 2436, 3115, 2205, 1987, 2635, 3435, 2361, 2183, 2822, 
  3664, 2550, 2342, 3001, 3779, 2538, 2363, 3090, 3807, 2631, 
  2601, 3063, 3961, 2774, 2476, 3083, 3864, 2773, 2489, 3082,
  # Van sales
  1900, 2200, 3000, 2000, 1900, 2500, 3800, 2500, 2100, 3100, 
  3650, 2950, 3300, 4000, 3290, 2600, 2010, 3600, 3500, 2100, 
  2050, 3500, 4290, 2800, 2770, 3080, 3100, 2800, 3100, 2860),
  ncol = 2),
  start = c(2011, 1), 
  frequency = 4,
  names = c("car_sales", "van_sales"))

class(qtr_sales)

ann_sales <- ts(matrix(c(# Car sales
  10324, 10200, 10582, 11097, 11582, 11092,
  # Van sales
  12000, 10400, 11550, 11400, 14500, 16000),
  ncol = 2),
  start = 2011, 
  frequency = 1,
  names = c("car_sales", "van_sales"))

# Quarterly indicator series (with default alter coefs for now)
my_series2 <- rbind(cbind(data.frame(group = rep("A", nrow(qtr_sales)),
                                     alt_van = rep(1, nrow(qtr_sales))), 
                          ts_to_tsDF(qtr_sales)),
                    cbind(data.frame(group = rep("B", nrow(qtr_sales)),
                                     alt_van = rep(1, nrow(qtr_sales))), 
                          ts_to_tsDF(qtr_sales)))

# Set binding van sales (alter coef = 0) for 2012 Q1 and Q2 in group A (rows 5 and 6)
my_series2$alt_van[c(5, 6)] <- 0

# Annual benchmarks for quarterly data (without alter coefs)
my_benchmarks2 <- rbind(cbind(data.frame(group = rep("A", nrow(ann_sales))), 
                              ts_to_bmkDF(ann_sales, ind_frequency = 4)),
                        cbind(data.frame(group = rep("B", nrow(ann_sales))), 
                              ts_to_bmkDF(ann_sales, ind_frequency = 4)))

# Benchmarking using...
#   - recommended RHO value for quarterly series (rho = 0.729)
#   - proportional model (lambda = 1)
#   - without bias correction (biasOption = 1 and bias not specified)
out_bench2 <- benchmarking(my_series2,
                           my_benchmarks2,
                           rho = 0.729,
                           lambda = 1,
                           biasOption = 1,
                           var = c("car_sales", "van_sales / alt_van"),
                           with = c("car_sales", "van_sales"),
                           by = "group")

# Outputs
View(out_bench2$series)
View(out_bench2$benchmarks)
View(out_bench2$graphTable)

# Graphics
plot_graphTable(out_bench2$graphTable, 
                "Ex2_graphs.pdf")




#### Example 3 (multiple quarterly series) ####
# Same as example 2, but benchmarking all 4 series as BY-groups
# (4 BY-groups of 1 series instead of 2 BY-groups of 2 series)

my_series3 <- stack_tsDF(ts_to_tsDF(ts.union(A = qtr_sales, B = qtr_sales)))
my_series3$alter <- 1
my_series3$alter[my_series3$series == "A.van_sales" & 
                   my_series3$year == 2012 & my_series3$period <= 2] <- 0
my_benchmarks3 <- stack_bmkDF(ts_to_bmkDF(ts.union(A = ann_sales, B = ann_sales),
                                          ind_frequency = 4))
out_bench3 <- benchmarking(my_series3,
                           my_benchmarks3,
                           rho = 0.729,
                           lambda = 1,
                           biasOption = 1,
                           var = "value / alter",
                           with = "value",
                           by = "series")

# Outputs
View(out_bench3$series)
View(out_bench3$benchmarks)
View(out_bench3$graphTable)

# Graphics
plot_graphTable(out_bench3$graphTable, 
                "Ex3_graphs.pdf")

# Convert the benchmarked series as a "mts" object
my_out_series3 <- tsDF_to_ts(unstack_tsDF(out_bench3$series), frequency = 4)
class(my_out_series3)
plot(my_out_series3)




#### Monthly data (Box & Jenkins airline series) ####

data(AirPassengers)
my_AP_ind <- ts_to_tsDF(AirPassengers)

# Create annual benchmarks by changing the level (5 times larger), adding some random 
# noise and dropping the last 2 benchmarks
set.seed(as.Date("2003-03-25"))  # for results reproducibility (select any date you want)
#set.seed(NULL)
ann_AP <- round(jitter(aggregate.ts(AirPassengers, nfrequency = 1, FUN = sum) * 5, 
                       amount = 2500))
my_AP_bmk <- (ts_to_bmkDF(ann_AP, ind_frequency = 12))[1:10, ]


# With bias correction (estimated bias with `biasOption = 3`)
#   => everything looks good
out_bench_AP <- benchmarking(my_AP_ind,
                             my_AP_bmk,
                             rho = 0.9,
                             lambda = 1,
                             biasOption = 3)
View(out_bench_AP$series)
View(out_bench_AP$benchmarks)
View(out_bench_AP$graphTable)

plot_graphTable(out_bench_AP$graphTable, 
                "AP_graphs.pdf")


# Without bias correction (`biasOption = 1`)
#   => issues with the projected adjustments at the end of the series for periods 
#      not covered by a benchmark
out_bench_AP_noBias <- benchmarking(my_AP_ind,
                                    my_AP_bmk,
                                    rho = 0.9,
                                    lambda = 1,
                                    biasOption = 1)
View(out_bench_AP_noBias$series)
View(out_bench_AP_noBias$benchmarks)
View(out_bench_AP_noBias$graphTable)

plot_graphTable(out_bench_AP_noBias$graphTable, 
                "AP_graphs_noBias.pdf")


# Denton benchmarking (`rho = 1`, bias correction is irrelevant)
#   => last adjustment repeated (forever) at the end of the series 
#      (strong assumption, but not necessarily problematic)
out_bench_AP_denton <- benchmarking(my_AP_ind,
                                    my_AP_bmk,
                                    rho = 1,
                                    lambda = 1)
View(out_bench_AP_denton$series)
View(out_bench_AP_denton$benchmarks)
View(out_bench_AP_denton$graphTable)

plot_graphTable(out_bench_AP_denton$graphTable, 
                "AP_graphs_Denton.pdf")


# Denton benchmarking approximation (`rho = 0.999`, `biasOption = 3`)
#   => regression benchmarking model approximation of the "pure" Denton method
#   => last adjustment repeated at the end of the series (mild convergence to the bias 
#      compared to the "pure" Denton method)
out_bench_AP_dentonApprox <- benchmarking(my_AP_ind,
                                          my_AP_bmk,
                                          rho = 0.999,
                                          lambda = 1,
                                          biasOption = 3)
View(out_bench_AP_dentonApprox$series)
View(out_bench_AP_dentonApprox$benchmarks)
View(out_bench_AP_dentonApprox$graphTable)

plot_graphTable(out_bench_AP_dentonApprox$graphTable, 
                "AP_graphs_DentonApprox.pdf")


# Pro-rating (`lambda = 0.5` and `rho = 0`)
#   => no movement preservation: all adjustments are lumped into January every year 
#      (with a disastrous impact on some of the initial December to January movements)
#   => immediate convergence to the bias (estimated with `biasOption = 3`) for the 
#      projected adjustments at the end of the series
out_bench_AP_proRate <- benchmarking(my_AP_ind,
                                     my_AP_bmk,
                                     rho = 0,
                                     lambda = 0.5,
                                     biasOption = 3)
View(out_bench_AP_proRate$series)
View(out_bench_AP_proRate$benchmarks)
View(out_bench_AP_proRate$graphTable)

plot_graphTable(out_bench_AP_proRate$graphTable, 
                "AP_graphs_proRating.pdf")




#### End of year stocks ("kinks" in the adjustments with `benchmarking()`) ####

# Quarterly indicator stock series (same pattern repeated every year)
my_stock_ind <- ts_to_tsDF(ts(rep(c(85, 95, 125, 95), 7), 
                              start = c(2013, 1), 
                              frequency = 4))

# Annual benchmarks (end-of-year stocks)
my_stock_bmk <- ts_to_bmkDF(ts(c(135, 125, 155, 145, 165), 
                               start = 2013, 
                               frequency = 1),
                            discrete_flag = TRUE,
                            alignment = "e",
                            ind_frequency = 4)

# With `benchmarking()` ("Proc Benchmarking" approach)
out_stock_PB <- benchmarking(my_stock_ind,
                             my_stock_bmk,
                             rho = 0.729,
                             lambda = 1,
                             biasOption = 3)

# With `stock_benchmarking()` ("Stock Benchmarking" approach)
out_stock_SB <- stock_benchmarking(my_stock_ind,
                                   my_stock_bmk,
                                   rho = 0.729,
                                   lambda = 1,
                                   biasOption = 3)

# Benchmarking adjustments of both approaches
plot_benchAdj(PB_graphTable = out_stock_PB$graphTable,
              SB_graphTable = out_stock_SB$graphTable)

# Have you noticed how smoother the `stock_benchmarking()` adjustments are compared 
# to the `benchmarking()` ones?

# The gain in the quality of the resulting benchmarked stocks might not necessarily
# be obvious in this example.
plot(out_stock_SB$graphTable$t, out_stock_SB$graphTable$benchmarked, 
     type = "b", col = "red", xlab = "t", ylab = "Benchmarked Stock")
lines(out_stock_PB$graphTable$t, out_stock_PB$graphTable$benchmarked, 
      type = "b", col = "blue")
legend(x = "topleft", bty = "n", inset = 0.05, lty = 1, pch = 1, 
       col = c("red", "blue"), legend = c("out_stock_SB", "out_stock_PB"))
title("Benchmarked Stock")

# PDF graphics
plot_graphTable(out_stock_PB$graphTable, "Stock_graphs_PB.pdf")
plot_graphTable(out_stock_SB$graphTable, "Stock_graphs_SB.pdf")

# What about cases where a flat indicator is used, which may happen in practice
# in absence of a good indicator of the quarterly (sub-annual) movement?
my_flat_ind <- my_stock_ind
my_flat_ind$value <- 1
out_stock_PB2 <- benchmarking(my_flat_ind,
                              my_stock_bmk,
                              rho = 0.729,
                              lambda = 1,
                              biasOption = 3)
out_stock_SB2 <- stock_benchmarking(my_flat_ind,
                                    my_stock_bmk,
                                    rho = 0.729,
                                    lambda = 1,
                                    biasOption = 3)

plot(out_stock_SB2$graphTable$t, out_stock_SB2$graphTable$benchmarked, 
     type = "b", col = "red", xlab = "t", ylab = "Benchmarked Stock")
lines(out_stock_PB2$graphTable$t, out_stock_PB2$graphTable$benchmarked, 
      type = "b", col = "blue")
legend(x = "bottomright", bty = "n", inset = 0.05, lty = 1, pch = 1, 
       col = c("red", "blue"), legend = c("out_stock_SB2", "out_stock_PB2"))
title("Benchmarked Stock - Flat Indicator")

# The awkwardness of the benchmarked stocks produced by `benchmarking()` suddenly
# becomes obvious. That's because the benchmarked series corresponds to the
# benchmarking adjustments when using a flat indicator (e.g., a series of 1's
# with proportional benchmarking):
plot_benchAdj(PB_graphTable = out_stock_PB2$graphTable,
              SB_graphTable = out_stock_SB2$graphTable)

# The shortcomings of the "Proc Benchmarking" approach (function `benchmarking()`)
# with stocks is also quite noticeable in this case when looking at the resulting
# quarterly growth rates, which are conveniently produced by `plot_graphTable()`.
# Pay particular attention to the transition in the growth rates from Q4 to Q1
# every year in the generated PDF graphs.
plot_graphTable(out_stock_PB2$graphTable, "Stock_graphs_PB_flat_indicator.pdf")
plot_graphTable(out_stock_SB2$graphTable, "Stock_graphs_SB_flat_indicator.pdf")


# Reset the working directory to its initial location
setwd(iniwd)
```

---
title: "G-Series"
output:
  rmarkdown::html_vignette: null
  pdf_document: null
  html_document:
    df_print: paged
urlcolor: blue
linkcolor: blue
vignette: >
  %\VignetteIndexEntry{G-Series}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, include = FALSE}
library(gseries)
```

<!-- Display the link to the French web page only when rendering an HTML document
     (e.g., not when rendering the PDF version)
     
     => the Pandoc "fenced_div" below (::: {.pkgdown-devel} <...> :::) is used to avoid 
        having the link generated in the pkgdown website vignette page
     => the link would only show in the "development" version of the pkgdown website
        (`development: mode: devel` in `_pkdown.yml` or `development: mode: auto` with a 4-level 
        version number in the DESCRIPTION file), which we do not use for gseries (we set 
        `development: mode: release` in `_pkdown.yml`, resulting in a single "release" website 
        regardless of the version number -->
::: {.pkgdown-devel}
```{asis, echo=knitr::is_html_output()}
(*version française : <https://StatCan.github.io/gensol-gseries/fr/articles/gseries.html>*)
```
:::

## Description

G<span>&#8209;</span>Series is Statistics Canada's (StatCan) generalized system devoted to the benchmarking and reconciliation of time series data. The methods used in G<span>&#8209;</span>Series essentially come from

* Dagum, E. B., and P. Cholette (2006). *Benchmarking, Temporal Distribution and Reconciliation Methods of Time Series*. Springer-Verlag, New York, Lecture Notes in Statistics, #186.

### Time Series Benchmarking
Goal: restore coherence between time series data of the same target variable measured at different frequencies (e.g., sub-annually and annually). 

The family of topics included under the *benchmarking umbrella* in G<span>&#8209;</span>Series includes, among others, *temporal distribution* (the reciprocal action of benchmarking: disaggregation of the benchmark series into more frequent observations), *calendarization* (a special case of temporal distribution) and *linking* (the connection of different time series segments into a single consistent time series). 

### Time Series Reconciliation
Goal: restore cross-sectional (contemporaneous) constraints in a system of time series measured at the same frequency (e.g., provincial and national series) with the optional preservation of temporal constraints. 

The reconciliation of aggregation tables (data cubes) involving only additivity constraints is called *raking* in G<span>&#8209;</span>Series while *balancing* refers to a more general class of reconciliation problems involving any type of linear constraints (including inequality constraints). 


## Software Availability

While early versions of G<span>&#8209;</span>Series (v1.04 and v2.0) were developed in SAS^®^, the software became an open-source tool with the release of G<span>&#8209;</span>Series 3.0 (R package gseries 3.0.0). This project is devoted to the open-source version of G<span>&#8209;</span>Series (R package gseries). Email us at [g-series@statcan.gc.ca](mailto:g-series@statcan.gc.ca) for information about the SAS^®^ versions. StatCan employees can also visit the G<span>&#8209;</span>Series Confluence page on the agency's intranet (search for "G-Series | G-Séries" in Confluence).


## Training

StatCan offers training on these topics. Visit the following pages on the agency's website for more information:

* [Theory and Application of Benchmarking (Course code 0436)](https://www.statcan.gc.ca/en/training/statistical/0436)
* [Theory and Application of Reconciliation (Course code 0437)](https://www.statcan.gc.ca/en/training/surveys/0437)


## Contact - Support

G<span>&#8209;</span>Series support is provided by the Time Series Research and Analysis Centre (TSRAC) in the Economic Statistics Methods Division (ESMD) and the Digital Processing Solutions Division (DPSD). Email us at [g-series@statcan.gc.ca](mailto:g-series@statcan.gc.ca) for information or help using G<span>&#8209;</span>Series. GitHub account holders can also request information, ask questions or report problems through the G<span>&#8209;</span>Series GitHub project [Issues](https://github.com/StatCan/gensol-gseries/issues) page. StatCan employees can do the same through the **Issues** page of the G<span>&#8209;</span>Series GitLab development project hosted on the agency's intranet (search for "G-Series in R - G-Séries en R" in GitLab).

---
title: "OSQP Settings Sequence Data Frame"
output:
  rmarkdown::html_vignette: null
  pdf_document: null
  html_document:
    df_print: paged
urlcolor: blue
linkcolor: blue
vignette: >
  %\VignetteIndexEntry{OSQP Settings Sequence Data Frame}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, include = FALSE}
library(gseries)
```

<!-- Display the link to the French web page only when rendering an HTML document
     (e.g., not when rendering the PDF version)
     
     => the Pandoc "fenced_div" below (::: {.pkgdown-devel} <...> :::) is used to avoid 
        having the link generated in the pkgdown website vignette page
     => the link would only show in the "development" version of the pkgdown website
        (`development: mode: devel` in `_pkdown.yml` or `development: mode: auto` with a 4-level 
        version number in the DESCRIPTION file), which we do not use for gseries (we set 
        `development: mode: release` in `_pkdown.yml`, resulting in a single "release" website 
        regardless of the version number -->
::: {.pkgdown-devel}
```{asis, echo=knitr::is_html_output()}
(*version française : <https://StatCan.github.io/gensol-gseries/fr/articles/osqp-settings-sequence-dataframe.html>*)
```
:::


Data frame containing a sequence of OSQP settings for `tsbalancing()` specified with argument `osqp_settings_df`. The package includes two predefined OSQP settings sequence data frames that are presented below.


## **Data frame `default_osqp_sequence`**

Fast and effective sequence of OSQP settings that should be suitable for accurately solving most time series balancing problems. It is the default value of `tsbalancing()` argument `osqp_settings_df`.

```{r, echo = FALSE}
load(file = "../data/default_osqp_sequence.rda")
format(as.data.frame(default_osqp_sequence), big.mark = ",", digits = 4)
```


## **Data frame `alternate_osqp_sequence`**

Alternative slower sequence of OSQP settings that could help achieve more precision, when needed, especially when combined with argument `full_sequence = TRUE`.

```{r, echo = FALSE}
load(file = "../data/alternate_osqp_sequence.rda")
format(as.data.frame(alternate_osqp_sequence), big.mark = ",", digits = 4)
```


## Details

With the exception of `prior_scaling` and `require_polished`, all columns of the data frame must correspond to a OSQP setting. Default OSQP values are used for any setting not specified in this data frame. Visit  <https://osqp.org/docs/interfaces/solver_settings.html> for all available OSQP settings. Note that the OSQP `verbose` setting is actually controlled through `tsbalancing()` arguments `quiet` and `display_level` (i.e., column `verbose` in a *OSQP settings sequence data frame* would be ignored). 

Each row of a *OSQP settings sequence data frame* represents one attempt at solving a balancing problem with the corresponding OSQP settings. The solving sequence stops as soon as a valid solution is obtained (a solution for which all constraint discrepancies are smaller or equal to the tolerance specified with `tsbalancing()` argument `validation_tol`) unless column `require_polished = TRUE`, in which case a polished solution from OSQP (`status_polish = 1`) would also be required to stop the sequence. Constraint discrepancies correspond to $\mathrm{max}(0, l - Ax, Ax - u)$ with constraints defined as $l \le Ax \le u$. In the event where a satisfactory solution cannot be obtained after having gone through the entire sequence, `tsbalancing()` returns the solution that generated the smallest total constraint discrepancies among valid solutions, if any, or among all solutions, otherwise. Note that running the entire solving sequence can be *enforced* by specifying `tsbalancing()` argument `full_sequence = TRUE`. Rows with column `prior_scaling = TRUE` have the problem data scaled prior to solving with OSQP, using the average of the free (nonbinding) problem values as the scaling factor.

In addition to specifying a custom-made *OSQP settings sequence data frame* with argument `osqp_settings_df`, one can also specify `osqp_settings_df = NULL` which would result in a single solving attempt with default OSQP values for all settings along with `prior_scaling = FALSE` and `require_polished = FALSE`. Note that it is recommended, however, to first try data frames `default_osqp_sequence` and `alternate_osqp_sequence`, along with `full_sequence = TRUE` if necessary, before considering other alternatives.


## Recommended Approach

Start with the default `tsbalancing()` solving sequence (`osqp_settings_df = default_osqp_sequence` and `full_sequence = FALSE`). Then, if more precision is needed, try with:

1. `full_sequence = TRUE`
2. `osqp_settings_df = alternate_osqp_sequence`
3. `osqp_settings_df = alternate_osqp_sequence` and `full_sequence = TRUE`

In practice, specifying `full_sequence = TRUE` should be enough when more precision is needed (at the expense of execution time, obviously). Only in rare occasions should you need to use the `alternate_osqp_sequence` data frame, which will often be even more costly in terms of execution time especially when combined with `full_sequence = TRUE`.


## Guiding Principles

The following is a summary of the lessons learned while developing `tsbalancing()` and experimenting with the OSQP solver. They are the guiding principles that lead to both *OSQP settings sequence data frames* presented earlier. Note that these observations apply to **_time series balancing problems_** as solved by `tsbalancing()` and may not directly apply to other types of *quadratic problems*.

+ Data preconditioning options available in OSQP (with the `scaling` setting) are not sufficient for some (badly scaled) problems. External (prior) data scaling (`prior_scaling = TRUE`) is sometimes necessary for OSQP to converge at a decent pace and generate *precise enough* solutions in a reasonable number of iterations.
+ Prior data scaling often reduces execution time (the required number of iterations to achieve the specified precision) and greatly increases the likelihood of polished solutions.
+ Polished solutions are always very precise, even when prior data scaling is performed (i.e., the solution in the original scale will usually still be *precise enough*).
+ While polished solutions from prior-scaled data are usually more precise than unpolished solutions from non prior-scaled data, the most precise solutions correspond to polished solutions from non prior-scaled data.
+ Smaller `sigma` and tolerance (`eps_*`) settings result in more precise solutions but take longer to run (require more iterations).
+ Enough precision is usually obtained after 10,000 iterations with small values for the `sigma` and tolerance (`eps_*`) settings.
+ The default OSQP values for `alpha` and the various settings associated to $\rho$ (`*rho*`) are sufficient (they work well). Reducing the `sigma` and tolerance (`eps_*`) settings and performing prior data scaling is sufficient to obtain precise solutions in a reasonable number of iterations. 
+ Keeping the same scale between the `sigma` and tolerance (`eps_*`) settings as the default OSQP values works well and is used  in both *OSQP settings sequence data frames*, i.e.:
  + `eps_abs = eps_rel = 1,000 * sigma`
  + `eps_prim_inf = eps_dual_inf = 100 * sigma`
  + (and consequently) `eps_abs = eps_rel = 10 * eps_prim_inf = 10 * eps_dual_inf`
+ The *machine epsilon* (`.Machine$double.eps`) for the `sigma` and tolerance (`eps_*`) settings, which basically forces the maximum number of iterations, is used as a *last resort* in both *OSQP settings sequence data frames*.

**Summary - default sequence** (data frame `default_osqp_sequence`)

+ Geared towards achieving both fast and precise solutions.
+ First try to get (fast) polished solutions on prior-scaled data before attempting solving the original (non-scaled) problem data.
+ Make a final attempt on prior-scaled data with the *machine epsilon* for the `sigma` and tolerance (`eps_*`) settings.

**Summary - alternative sequence** (data frame `alternate_osqp_sequence`)

+ Geared towards achieving precise solutions at the expense of execution time.
+ Somewhat similar to a *brute force* or *try all* approach (especially when combined with `full_sequence = TRUE`).
+ Small `sigma` and tolerance (`eps_*`) settings that gradually increase, with the *machine epsilon* as a last attempt.
+ Polished solutions are required for every step of the sequence (the best unpolished solution is returned if no polished solution could be obtained at the end of the entire sequence).
+ Maximum of 10,000 iterations for every step of the sequence
+ First try to get polished solutions from non prior-scaled data (the most precise solutions), then try with prior-scaled data.

---
title: "A Beginner's Benchmarking Demo Script"
output:
  rmarkdown::html_vignette: null
  pdf_document: null
  html_document:
    df_print: paged
urlcolor: blue
linkcolor: blue
vignette: >
  %\VignetteIndexEntry{A Beginner's Benchmarking Demo Script}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, include = FALSE}
library(gseries)
```

<!-- Display the link to the French web page only when rendering an HTML document
     (e.g., not when rendering the PDF version)
     
     => the Pandoc "fenced_div" below (::: {.pkgdown-devel} <...> :::) is used to avoid 
        having the link generated in the pkgdown website vignette page
     => the link would only show in the "development" version of the pkgdown website
        (`development: mode: devel` in `_pkdown.yml` or `development: mode: auto` with a 4-level 
        version number in the DESCRIPTION file), which we do not use for gseries (we set 
        `development: mode: release` in `_pkdown.yml`, resulting in a single "release" website 
        regardless of the version number -->
::: {.pkgdown-devel}
```{asis, echo=knitr::is_html_output()}
(*version française : <https://StatCan.github.io/gensol-gseries/fr/articles/benchmarking-demo-script.html>*)
```
:::


```{r, eval = FALSE}
# Load G-Series in R (package gseries)
library(gseries)

# Set the working directory (for the PDF graph files)
iniwd <- setwd("C:/Temp")




#### Example 1 (single quarterly series) ####
# Simple case with a single quarterly series to benchmark to annual values

# Quarterly indicator series
my_ts_qtr <- ts(
  c(1.9, 2.4, 3.1, 2.2, 2.0, 2.6, 3.4, 2.4, 2.3), 
  start = c(2015, 1), 
  frequency = 4
)
my_series1 <- ts_to_tsDF(my_ts_qtr)

# Annual benchmarks for quarterly data
my_ts_ann <- ts(
  c(10.3, 10.2), 
  start = 2015, 
  frequency = 1
)
my_benchmarks1 <- ts_to_bmkDF(my_ts_ann, ind_frequency = 4)

# Benchmarking using...
#   - recommended `rho` value for quarterly series (`rho = 0.729`)
#   - proportional model (`lambda = 1`)
#   - bias-corrected indicator series with the estimated bias (`biasOption = 3`)
out_bench1 <- benchmarking(
  my_series1,
  my_benchmarks1,
  rho = 0.729,
  lambda = 1,
  biasOption = 3
)

# Outputs
View(out_bench1$series)
View(out_bench1$benchmarks)
View(out_bench1$graphTable)

# Graphics (default set of plots)
plot_graphTable(
  out_bench1$graphTable, 
  "Ex1_graphs.pdf"
)
# With the G. R. Table
plot_graphTable(
  out_bench1$graphTable, 
  "Ex1_graphs_with_GRTable.pdf",
  GR_table_flag = TRUE
)


# Compare with package tempdisagg that allows Denton benchmarking 
# (modified version by Cholette)

#install.packages("tempdisagg")
library(tempdisagg)
#install.packages("dplyr")
library(dplyr)


# Proportional Denton
mult_denton <- benchmarking(
  my_series1,
  my_benchmarks1,
  rho = 1,
  lambda = 1
)

td_mult_denton <- td(
  formula = my_ts_ann ~ 0 + my_ts_qtr,
  method = "denton-cholette",
  criterion = "proportional"
)

compare_mult <- ts.union(
  gseries = tsDF_to_ts(mult_denton$series, frequency = 4),
  tempdisagg = predict(td_mult_denton),
  dframe = TRUE
) %>% 
  cbind(diff = .[[1]] - .[[2]])

# Same results
View(compare_mult)


# Additive Denton
add_denton <- benchmarking(
  my_series1,
  my_benchmarks1,
  rho = 1,
  lambda = 0
)

td_add_denton <- td(
  formula = my_ts_ann ~ 0 + my_ts_qtr,
  method = "denton-cholette",
  criterion = "additive"
)

compare_add <- ts.union(
  gseries = tsDF_to_ts(add_denton$series, frequency = 4),
  tempdisagg = predict(td_add_denton),
  dframe = TRUE
) %>% 
  cbind(diff = .[[1]] - .[[2]])

# Same results
View(compare_add)




#### Example 2 (multiple quarterly series) ####
# Two quarterly series to benchmark to annual values, 
# with BY-groups and user-defined alterability coefficients

# Sales data (same sales for groups A and B; only alter coefs for van sales differ)
qtr_sales <- ts(
  matrix(
    c(
      # Car sales
      1851, 2436, 3115, 2205, 1987, 2635, 3435, 2361, 2183, 2822, 
      3664, 2550, 2342, 3001, 3779, 2538, 2363, 3090, 3807, 2631, 
      2601, 3063, 3961, 2774, 2476, 3083, 3864, 2773, 2489, 3082,
      # Van sales
      1900, 2200, 3000, 2000, 1900, 2500, 3800, 2500, 2100, 3100, 
      3650, 2950, 3300, 4000, 3290, 2600, 2010, 3600, 3500, 2100, 
      2050, 3500, 4290, 2800, 2770, 3080, 3100, 2800, 3100, 2860
    ),
    ncol = 2
  ),
  start = c(2011, 1), 
  frequency = 4,
  names = c("car_sales", "van_sales")
)

class(qtr_sales)

ann_sales <- ts(
  matrix(
    c(
      # Car sales
      10324, 10200, 10582, 11097, 11582, 11092,
      # Van sales
      12000, 10400, 11550, 11400, 14500, 16000
    ),
    ncol = 2
  ),
  start = 2011, 
  frequency = 1,
  names = c("car_sales", "van_sales")
)

# Quarterly indicator series (with default alter coefs for now)
my_series2 <- rbind(
  cbind(
    data.frame(
      group = rep("A", nrow(qtr_sales)),
      alt_van = rep(1, nrow(qtr_sales))
    ), 
    ts_to_tsDF(qtr_sales)
  ),
  cbind(
    data.frame(
      group = rep("B", nrow(qtr_sales)),
      alt_van = rep(1, nrow(qtr_sales))
    ), 
    ts_to_tsDF(qtr_sales)
  )
)

# Set binding van sales (alter coef = 0) for 2012 Q1 and Q2 in group A (rows 5 and 6)
my_series2$alt_van[c(5, 6)] <- 0

# Annual benchmarks for quarterly data (without alter coefs)
my_benchmarks2 <- rbind(
  cbind(
    data.frame(group = rep("A", nrow(ann_sales))), 
    ts_to_bmkDF(ann_sales, ind_frequency = 4)
  ),
  cbind(
    data.frame(group = rep("B", nrow(ann_sales))),
    ts_to_bmkDF(ann_sales, ind_frequency = 4)
  )
)

# Benchmarking using...
#   - recommended RHO value for quarterly series (rho = 0.729)
#   - proportional model (lambda = 1)
#   - without bias correction (biasOption = 1 and bias not specified)
out_bench2 <- benchmarking(
  my_series2,
  my_benchmarks2,
  rho = 0.729,
  lambda = 1,
  biasOption = 1,
  var = c("car_sales", "van_sales / alt_van"),
  with = c("car_sales", "van_sales"),
  by = "group"
)

# Outputs
View(out_bench2$series)
View(out_bench2$benchmarks)
View(out_bench2$graphTable)

# Graphics
plot_graphTable(
  out_bench2$graphTable, 
  "Ex2_graphs.pdf"
)




#### Example 3 (multiple quarterly series) ####
# Same as example 2, but benchmarking all 4 series as BY-groups
# (4 BY-groups of 1 series instead of 2 BY-groups of 2 series)

my_series3 <- stack_tsDF(ts_to_tsDF(ts.union(A = qtr_sales, B = qtr_sales)))
my_series3$alter <- 1
my_series3$alter[
  my_series3$series == "A.van_sales" & 
    my_series3$year == 2012 & 
    my_series3$period <= 2
] <- 0
my_benchmarks3 <- stack_bmkDF(
  ts_to_bmkDF(ts.union(A = ann_sales, B = ann_sales), ind_frequency = 4)
)
out_bench3 <- benchmarking(
  my_series3,
  my_benchmarks3,
  rho = 0.729,
  lambda = 1,
  biasOption = 1,
  var = "value / alter",
  with = "value",
  by = "series"
)

# Outputs
View(out_bench3$series)
View(out_bench3$benchmarks)
View(out_bench3$graphTable)

# Graphics
plot_graphTable(
  out_bench3$graphTable, 
  "Ex3_graphs.pdf"
)

# Convert the benchmarked series as a "mts" object
my_out_series3 <- tsDF_to_ts(unstack_tsDF(out_bench3$series), frequency = 4)
class(my_out_series3)
plot(my_out_series3)




#### Monthly data (Box & Jenkins airline series) ####

data(AirPassengers)
my_AP_ind <- ts_to_tsDF(AirPassengers)

# Create annual benchmarks by changing the level (5 times larger), adding some random 
# noise and dropping the last 2 benchmarks
set.seed(as.Date("2003-03-25"))  # for results reproducibility (select any date you want)
#set.seed(NULL)
ann_AP <- round(
  jitter(
    aggregate.ts(AirPassengers, nfrequency = 1, FUN = sum) * 5, amount = 2500
  )
)
my_AP_bmk <- (ts_to_bmkDF(ann_AP, ind_frequency = 12))[1:10, ]


# With bias correction (estimated bias with `biasOption = 3`)
#   => everything looks good
out_bench_AP <- benchmarking(
  my_AP_ind,
  my_AP_bmk,
  rho = 0.9,
  lambda = 1,
  biasOption = 3
)
View(out_bench_AP$series)
View(out_bench_AP$benchmarks)
View(out_bench_AP$graphTable)

plot_graphTable(
  out_bench_AP$graphTable, 
  "AP_graphs.pdf"
)


# Without bias correction (`biasOption = 1`)
#   => issues with the projected adjustments at the end of the series for periods 
#      not covered by a benchmark
out_bench_AP_noBias <- benchmarking(
  my_AP_ind,
  my_AP_bmk,
  rho = 0.9,
  lambda = 1,
  biasOption = 1
)
View(out_bench_AP_noBias$series)
View(out_bench_AP_noBias$benchmarks)
View(out_bench_AP_noBias$graphTable)

plot_graphTable(
  out_bench_AP_noBias$graphTable, 
  "AP_graphs_noBias.pdf"
)


# Denton benchmarking (`rho = 1`, bias correction is irrelevant)
#   => last adjustment repeated (forever) at the end of the series 
#      (strong assumption, but not necessarily problematic)
out_bench_AP_denton <- benchmarking(
  my_AP_ind,
  my_AP_bmk,
  rho = 1,
  lambda = 1
)
View(out_bench_AP_denton$series)
View(out_bench_AP_denton$benchmarks)
View(out_bench_AP_denton$graphTable)

plot_graphTable(
  out_bench_AP_denton$graphTable, 
  "AP_graphs_Denton.pdf"
)


# Denton benchmarking approximation (`rho = 0.999`, `biasOption = 3`)
#   => regression benchmarking model approximation of the "pure" Denton method
#   => last adjustment repeated at the end of the series (mild convergence to the bias 
#      compared to the "pure" Denton method)
out_bench_AP_dentonApprox <- benchmarking(
  my_AP_ind,
  my_AP_bmk,
  rho = 0.999,
  lambda = 1,
  biasOption = 3
)
View(out_bench_AP_dentonApprox$series)
View(out_bench_AP_dentonApprox$benchmarks)
View(out_bench_AP_dentonApprox$graphTable)

plot_graphTable(
  out_bench_AP_dentonApprox$graphTable, 
  "AP_graphs_DentonApprox.pdf"
)


# Pro-rating (`lambda = 0.5` and `rho = 0`)
#   => no movement preservation: all adjustments are lumped into January every year 
#      (with a disastrous impact on some of the initial December to January movements)
#   => immediate convergence to the bias (estimated with `biasOption = 3`) for the 
#      projected adjustments at the end of the series
out_bench_AP_proRate <- benchmarking(
  my_AP_ind,
  my_AP_bmk,
  rho = 0,
  lambda = 0.5,
  biasOption = 3
)
View(out_bench_AP_proRate$series)
View(out_bench_AP_proRate$benchmarks)
View(out_bench_AP_proRate$graphTable)

plot_graphTable(
  out_bench_AP_proRate$graphTable, 
  "AP_graphs_proRating.pdf"
)




#### End of year stocks ("kinks" in the adjustments with `benchmarking()`) ####

# Quarterly indicator stock series (same pattern repeated every year)
my_stock_ind <- ts_to_tsDF(
  ts(
    rep(c(85, 95, 125, 95), 7), 
    start = c(2013, 1), 
    frequency = 4
  )
)

# Annual benchmarks (end-of-year stocks)
my_stock_bmk <- ts_to_bmkDF(
  ts(
    c(135, 125, 155, 145, 165), 
    start = 2013, 
    frequency = 1
  ),
  discrete_flag = TRUE,
  alignment = "e",
  ind_frequency = 4
)

# With `benchmarking()` ("Proc Benchmarking" approach)
out_stock_PB <- benchmarking(
  my_stock_ind,
  my_stock_bmk,
  rho = 0.729,
  lambda = 1,
  biasOption = 3
)

# With `stock_benchmarking()` ("Stock Benchmarking" approach)
out_stock_SB <- stock_benchmarking(
  my_stock_ind,
  my_stock_bmk,
  rho = 0.729,
  lambda = 1,
  biasOption = 3
)

# Benchmarking adjustments of both approaches
plot_benchAdj(
  PB_graphTable = out_stock_PB$graphTable,
  SB_graphTable = out_stock_SB$graphTable
)

# Have you noticed how smoother the `stock_benchmarking()` adjustments are compared 
# to the `benchmarking()` ones?

# The gain in the quality of the resulting benchmarked stocks might not necessarily
# be obvious in this example.
plot(
  out_stock_SB$graphTable$t, out_stock_SB$graphTable$benchmarked, 
  type = "b", col = "red", xlab = "t", ylab = "Benchmarked Stock"
)
lines(
  out_stock_PB$graphTable$t, out_stock_PB$graphTable$benchmarked, 
  type = "b", col = "blue"
)
legend(
  x = "topleft", bty = "n", inset = 0.05, lty = 1, pch = 1, 
  col = c("red", "blue"), legend = c("out_stock_SB", "out_stock_PB")
)
title("Benchmarked Stock")

# PDF graphics
plot_graphTable(out_stock_PB$graphTable, "Stock_graphs_PB.pdf")
plot_graphTable(out_stock_SB$graphTable, "Stock_graphs_SB.pdf")

# What about cases where a flat indicator is used, which may happen in practice
# in absence of a good indicator of the quarterly (sub-annual) movement?
my_flat_ind <- my_stock_ind
my_flat_ind$value <- 1
out_stock_PB2 <- benchmarking(
  my_flat_ind,
  my_stock_bmk,
  rho = 0.729,
  lambda = 1,
  biasOption = 3
)
out_stock_SB2 <- stock_benchmarking(
  my_flat_ind,
  my_stock_bmk,
  rho = 0.729,
  lambda = 1,
  biasOption = 3
)

plot(
  out_stock_SB2$graphTable$t, out_stock_SB2$graphTable$benchmarked, 
  type = "b", col = "red", xlab = "t", ylab = "Benchmarked Stock"
)
lines(
  out_stock_PB2$graphTable$t, out_stock_PB2$graphTable$benchmarked, 
  type = "b", col = "blue"
)
legend(
  x = "bottomright", bty = "n", inset = 0.05, lty = 1, pch = 1, 
  col = c("red", "blue"), legend = c("out_stock_SB2", "out_stock_PB2")
)
title("Benchmarked Stock - Flat Indicator")

# The awkwardness of the benchmarked stocks produced by `benchmarking()` suddenly
# becomes obvious. That's because the benchmarked series corresponds to the
# benchmarking adjustments when using a flat indicator (e.g., a series of 1's
# with proportional benchmarking):
plot_benchAdj(
  PB_graphTable = out_stock_PB2$graphTable,
  SB_graphTable = out_stock_SB2$graphTable
)

# The shortcomings of the "Proc Benchmarking" approach (function `benchmarking()`)
# with stocks is also quite noticeable in this case when looking at the resulting
# quarterly growth rates, which are conveniently produced by `plot_graphTable()`.
# Pay particular attention to the transition in the growth rates from Q4 to Q1
# every year in the generated PDF graphs.
plot_graphTable(out_stock_PB2$graphTable, "Stock_graphs_PB_flat_indicator.pdf")
plot_graphTable(out_stock_SB2$graphTable, "Stock_graphs_SB_flat_indicator.pdf")


# Reset the working directory to its initial location
setwd(iniwd)
```

---
title: "G-Series"
output:
  rmarkdown::html_vignette: null
  pdf_document: null
  html_document:
    df_print: paged
urlcolor: blue
linkcolor: blue
vignette: >
  %\VignetteIndexEntry{G-Series}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, include = FALSE}
library(gseries)
```

<!-- Display the link to the French web page only when rendering an HTML document
     (e.g., not when rendering the PDF version)
     
     => the Pandoc "fenced_div" below (::: {.pkgdown-devel} <...> :::) is used to avoid 
        having the link generated in the pkgdown website vignette page
     => the link would only show in the "development" version of the pkgdown website
        (`development: mode: devel` in `_pkdown.yml` or `development: mode: auto` with a 4-level 
        version number in the DESCRIPTION file), which we do not use for gseries (we set 
        `development: mode: release` in `_pkdown.yml`, resulting in a single "release" website 
        regardless of the version number -->
::: {.pkgdown-devel}
```{asis, echo=knitr::is_html_output()}
(*version française : <https://StatCan.github.io/gensol-gseries/fr/articles/gseries.html>*)
```
:::

## Description

G<span>&#8209;</span>Series is Statistics Canada's (StatCan) generalized system devoted to the benchmarking and reconciliation of time series data. The methods used in G<span>&#8209;</span>Series essentially come from

* Dagum, E. B., and P. Cholette (2006). *Benchmarking, Temporal Distribution and Reconciliation Methods of Time Series*. Springer-Verlag, New York, Lecture Notes in Statistics, Vol. 186. [doi:10.1007/0-387-35439-5](https://doi.org/10.1007/0-387-35439-5)

### Time Series Benchmarking
Goal: restore coherence between time series data of the same target variable measured at different frequencies (e.g., sub-annually and annually). 

The family of topics included under the *benchmarking umbrella* in G<span>&#8209;</span>Series includes, among others, *temporal distribution* (the reciprocal action of benchmarking: disaggregation of the benchmark series into more frequent observations), *calendarization* (a special case of temporal distribution) and *linking* (the connection of different time series segments into a single consistent time series). 

### Time Series Reconciliation
Goal: restore cross-sectional (contemporaneous) constraints in a system of time series measured at the same frequency (e.g., provincial and national series) with the optional preservation of temporal constraints. 

The reconciliation of aggregation tables (data cubes) involving only additivity constraints is called *raking* in G<span>&#8209;</span>Series while *balancing* refers to a more general class of reconciliation problems involving any type of linear constraints (including inequality constraints). 


## Software Availability

While early versions of G<span>&#8209;</span>Series (v1.04 and v2.0) were developed in SAS^®^, the software became an open-source tool with the release of G<span>&#8209;</span>Series 3.0 (R package gseries 3.0.0). This project is devoted to the open-source version of G<span>&#8209;</span>Series (R package gseries). Email us at [g-series@statcan.gc.ca](mailto:g-series@statcan.gc.ca) for information about the SAS^®^ versions. StatCan employees can also visit the G<span>&#8209;</span>Series Confluence page on the agency's intranet (search for "G-Series | G-Séries" in Confluence).


## Training

StatCan offers training on these topics. Visit the following pages on the agency's website for more information:

* [Theory and Application of Benchmarking (Course code 0436)](https://www.statcan.gc.ca/en/training/statistical/0436)
* [Theory and Application of Reconciliation (Course code 0437)](https://www.statcan.gc.ca/en/training/statistical/0437)


## Contact - Support

G<span>&#8209;</span>Series support is provided by the Time Series Research and Analysis Centre (TSRAC) in the Economic Statistics Methods Division (ESMD) and the Digital Processing Solutions Division (DPSD). Email us at [g-series@statcan.gc.ca](mailto:g-series@statcan.gc.ca) for information or help using G<span>&#8209;</span>Series. GitHub account holders can also request information, ask questions or report problems through the G<span>&#8209;</span>Series GitHub project [Issues](https://github.com/StatCan/gensol-gseries/issues) page. StatCan employees can do the same through the **Issues** page of the G<span>&#8209;</span>Series GitLab development project hosted on the agency's intranet (search for "G-Series in R - G-Séries en R" in GitLab).

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_graphTable.R
\name{plot_graphTable}
\alias{plot_graphTable}
\title{Generate benchmarking graphics in a PDF file}
\usage{
plot_graphTable(
  graphTable,
  pdf_file,
  ori_plot_flag = TRUE,
  adj_plot_flag = TRUE,
  GR_plot_flag = TRUE,
  GR_table_flag = FALSE,
  add_bookmarks = TRUE
)
}
\arguments{
\item{graphTable}{(mandatory)

Data frame (object of class "data.frame") corresponding to the benchmarking function output\code{graphTable} data frame.}

\item{pdf_file}{(mandatory)

Name (and path) of the PDF file that will contain the benchmarking graphics. The name should include the ".pdf"
file extension. The PDF file is created in the R session working directory (as returned by \code{getwd()}) if a path
is not specified. Specifying \code{NULL} would cancel the creation of a PDF file.}

\item{ori_plot_flag, adj_plot_flag, GR_plot_flag, GR_table_flag}{(optional)

Logical arguments indicating whether or not the corresponding type of benchmarking graphic should be generated.
All three plots are generated by default but not the growth rates tables.

\strong{Default values} are \code{ori_plot_flag = TRUE}, \code{adj_plot_flag = TRUE}, \code{GR_plot_flag = TRUE} and
\code{GR_table_flag = FALSE}.}

\item{add_bookmarks}{Logical argument indicating whether or not bookmarks should be added to the PDF file. See \strong{Bookmarks} in section
\strong{Details} for more information.

\strong{Default value} is \code{add_bookmarks = TRUE}.}
}
\value{
In addition to creating a PDF file containing the benchmarking graphics (except when \code{pdf_file = NULL}), this function
also invisibly returns a list with the following elements:
\itemize{
\item \code{graphTable}: Character string (character vector of length one) that contains the complete name and path of the PDF
file if it was successfully created and \code{invisible(NA_character_)} otherwise or if \code{pdf_file = NULL} was specified.
\item \code{graph_list}: List of the generated benchmarking graphics (one per series) with the following elements:
\itemize{
\item \code{name}: Character string describing the series (matches the bookmark name in the PDF file).
\item \code{page}: Integer representing the sequence number of the first graphic for the series in the entire sequence of
graphics for all series (matches the page number in the PDF file).
\item \code{ggplot_list}: List of ggplot objects (one per graphic or page in the PDF file) corresponding to the generated
benchmarking graphics for the series. See section \strong{Value} in \link{bench_graphs} for details.
}
}

Note that the returned ggplot objects can be displayed \emph{manually} with \code{\link[=print]{print()}}, in which case some updates to the
ggplot2 theme defaults are recommended in order to produce graphics with a similar look and feel as those generated in
the PDF file (see section \strong{Value} in \link{bench_graphs} for details). Also keep in mind that these graphics are optimized
for the US Letter paper size format in landscape view (as displayed in the PDF file), i.e., 11in wide (27.9cm, 1056px
with 96 DPI) and 8.5in tall (21.6cm, 816px with 96 DPI).
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/plot_graphTable.html}})}

Create a PDF file (US Letter paper size format in landscape view) containing benchmarking graphics for the set of series
contained in the specified benchmarking function (\code{\link[=benchmarking]{benchmarking()}} or \code{\link[=stock_benchmarking]{stock_benchmarking()}}) output \code{graphTable}
data frame. Four types of benchmarking graphics can be generated for each series:
\itemize{
\item \strong{Original Scale Plot} (argument \code{ori_plot_flag}) - overlay graph of:
\itemize{
\item Indicator series
\item Average indicator series
\item Bias corrected indicator series (when \eqn{\rho < 1})
\item Benchmarked series
\item Average benchmark
}
\item \strong{Adjustment Scale Plot} (argument \code{adj_plot_flag}) - overlay graph of:
\itemize{
\item Benchmarking adjustments
\item Average benchmarking adjustments
\item Bias line (when \eqn{\rho < 1})
}
\item \strong{Growth Rates Plot} (argument \code{GR_plot_flag}) - bar chart of the indicator and benchmarked series growth rates.
\item \strong{Growth Rates Table} (argument \code{GR_table_flag}) - table of the indicator and benchmarked series growth rates.
}

These graphics can be useful to assess the quality of the benchmarking results. Any of the four types of
benchmarking graphics can be enabled or disabled with the corresponding flag. The first three types of graphics
(the plots) are generated by default while the fourth (growth rates table) is not.
}
\details{
List of the \code{graphTable} data frame variables corresponding to each element of the four types of benchmarking
graphics:
\itemize{
\item Original Scale Plot (argument \code{ori_plot_flag})
\itemize{
\item \code{subAnnual} for the \emph{Indicator Series} line
\item \code{avgSubAnnual} for the \emph{Avg. Indicator Series} segments
\item \code{subAnnualCorrected} for the \emph{Bias Corr. Indicator Series} line (when \eqn{\rho < 1})
\item \code{benchmarked} for the \emph{Benchmarked Series} line
\item \code{avgBenchmark} for the \emph{Average Benchmark} segments
}
\item Adjustment Scale Plot (argument \code{adj_plot_flag})
\itemize{
\item \code{benchmarkedSubAnnualRatio} for the \emph{BI Ratios (Benchmarked Series / Indicator Series)} line \eqn{^{(*)}}{(*)}
\item \code{avgBenchmarkSubAnnualRatio} for the \emph{Average BI Ratios} segments \eqn{^{(*)}}{(*)}
\item \code{bias} for the \emph{Bias} line (when \eqn{\rho < 1})
}
\item Growth Rates Plot (argument \code{GR_plot_flag})
\itemize{
\item \code{growthRateSubAnnual} for the \emph{Growth R. in Indicator Series} bars \eqn{^{(*)}}{(*)}
\item \code{growthRateBenchmarked} for the \emph{Growth R. in Benchmarked Series} bars \eqn{^{(*)}}{(*)}
}
\item Growth Rates Table (argument \code{GR_table_flag})
\itemize{
\item \code{year} for the \emph{Year} column
\item \code{period} for the \emph{Period} column
\item \code{subAnnual} for the \emph{Indicator Series} column
\item \code{benchmarked} for the \emph{Benchmarked Series} column
\item \code{growthRateSubAnnual} for the \emph{Growth Rate in Indicator Series} column \eqn{^{(*)}}{(*)}
\item \code{growthRateBenchmarked} for the \emph{Growth Rate in Benchmarked Series} column \eqn{^{(*)}}{(*)}
}
}

\eqn{^{(*)}}{(*)} \emph{BI ratios} and \emph{growth rates} actually correspond to \emph{differences} when \eqn{\lambda = 0} (additive
benchmarking).

The function uses the extra columns of the \code{graphTable} data frame (columns not listed in the \strong{Value} section of
\code{\link[=benchmarking]{benchmarking()}} and \code{\link[=stock_benchmarking]{stock_benchmarking()}}), if any, to build BY-groups. See section \strong{Benchmarking Multiple Series}
of \code{\link[=benchmarking]{benchmarking()}} for more details.
\subsection{Performance}{

The two types of growth rates graphics, i.e., the bar chart (\code{GR_plot_flag}) and table (\code{GR_table_flag}), often requires
the generation of several pages in the PDF file, especially for long monthly series with several years of data. This creation of
extra pages slows down the execution of \code{\link[=plot_graphTable]{plot_graphTable()}}. This is why only the bar chart is generated by default
(\code{GR_plot_flag = TRUE} and \code{GR_table_flag = FALSE}). Deactivating both types of growth rates graphics (\code{GR_plot_flag = FALSE}
and \code{GR_table_flag = FALSE}) or reducing the size of the input \code{graphTable} data frame for very long series (e.g., keeping
only recent years) could therefore improve execution time. Also note that the impact of benchmarking on the growth rates can be
deduced from the adjustment scale plot (\code{adj_plot_flag}) by examining the extent of vertical movement (downward or upward) of the
benchmarking adjustments between adjacent periods: the greater the vertical movement, the greater the impact on corresponding growth
rate. Execution time of \code{\link[=plot_graphTable]{plot_graphTable()}} could therefore be reduced, if needed, by only generating the first two types of graphics
while focusing on the adjustment scale plot to assess period-to-period movement preservation, i.e., the impact of benchmarking on
the initial growth rates.
}

\subsection{ggplot2 themes}{

The plots are generated with the ggplot2 package which comes with a convenient set of \href{https://ggplot2.tidyverse.org/reference/ggtheme.html}{complete themes} for the general look and feel of the plots (with \code{theme_grey()}
as the default theme). Use function \code{theme_set()} to change the theme applied to the plots generated by \code{\link[=plot_graphTable]{plot_graphTable()}}
(see the \strong{Examples}).
}

\subsection{Bookmarks}{

Bookmarks are added to the PDF file with \code{xmpdf::set_bookmarks()} when argument \code{add_bookmarks = TRUE} (default), which
requires a command-line tool such as \href{https://www.ghostscript.com/}{Ghostscript}  or
\href{https://www.pdflabs.com/tools/pdftk-the-pdf-toolkit/}{PDFtk}. See section \strong{Installation} in
\code{vignette("xmpdf", package = "xmpdf")} for details.

\strong{Important}: bookmarks will be successfully added to the PDF file \strong{if and only if} \ifelse{latex}{\code{xmpdf::supports 
_set_bookmarks()}}{\code{xmpdf::supports_set_bookmarks()}} returns \code{TRUE} \strong{and} the execution of \code{xmpdf::set_bookmarks()} is
successful. If Ghostscript is installed on your machine but \code{xmpdf::supports_set_bookmarks()} still returns \code{FALSE}, try
specifying the path of the Ghostscript executable in environment variable \code{R_GSCMD} (e.g.,
\code{Sys.setenv(R_GSCMD = "C:/Program Files/.../bin/gswin64c.exe")} on Windows). On the other hand, if
\verb{xmpdf::supports_set_bookmarks()\}} returns \code{TRUE} but you are experiencing (irresolvable) issues with \code{xmpdf::set_bookmarks()}
(e.g., error related to the Ghostscript executable), bookmarks creation can be disabled by specifying \code{add_bookmarks = FALSE}.
}
}
\examples{
# Set the working directory (for the PDF files)
iniwd <- setwd(tempdir())

# Quarterly car and van sales (indicator series)
qtr_ind <- ts_to_tsDF(
  ts(
    matrix(
      c(
        # Car sales
        1851, 2436, 3115, 2205, 1987, 2635, 3435, 2361, 2183, 2822,
        3664, 2550, 2342, 3001, 3779, 2538, 2363, 3090, 3807, 2631,
        2601, 3063, 3961, 2774, 2476, 3083, 3864, 2773, 2489, 3082,
        # Van sales
        1900, 2200, 3000, 2000, 1900, 2500, 3800, 2500, 2100, 3100,
        3650, 2950, 3300, 4000, 3290, 2600, 2010, 3600, 3500, 2100,
        2050, 3500, 4290, 2800, 2770, 3080, 3100, 2800, 3100, 2860
      ),
      ncol = 2
    ),
    start = c(2011, 1),
    frequency = 4,
    names = c("car_sales", "van_sales")
  )
)

# Annual car and van sales (benchmarks)
ann_bmk <- ts_to_bmkDF(
  ts(matrix(c(# Car sales
              10324, 10200, 10582, 11097, 11582, 11092,
              # Van sales
              12000, 10400, 11550, 11400, 14500, 16000),
            ncol = 2),
     start = 2011,
     frequency = 1,
     names = c("car_sales", "van_sales")), 
  ind_frequency = 4)

# Proportional benchmarking without bias correction
out_bench <- benchmarking(qtr_ind, ann_bmk, 
                          rho = 0.729, lambda = 1, biasOption = 1,
                          allCols = TRUE,
                          quiet = TRUE)


# Default set of graphics (the first 3 types of plots)
\donttest{plot_graphTable(out_bench$graphTable, "bench_graphs.pdf")}

# Temporarily use ggplot2 `theme_bw()` for the plots
library(ggplot2)
ini_theme <- theme_get()
theme_set(theme_bw())
\donttest{plot_graphTable(out_bench$graphTable, "bench_graphs_bw.pdf")}
theme_set(ini_theme)

# Generate all 4 types of graphics (including the growth rates table)
\donttest{plot_graphTable(out_bench$graphTable, "bench_graphs_with_GRTable.pdf",
                GR_table_flag = TRUE)}

# Reduce execution time by disabling both types of growth rates graphics
\donttest{plot_graphTable(out_bench$graphTable, "bench_graphs_no_GR.pdf",
                GR_plot_flag = FALSE)}


# Reset the working directory to its initial location
setwd(iniwd)
}
\seealso{
\link{bench_graphs} \code{\link[=plot_benchAdj]{plot_benchAdj()}} \code{\link[=benchmarking]{benchmarking()}} \code{\link[=stock_benchmarking]{stock_benchmarking()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aliases.R
\name{aliases}
\alias{aliases}
\alias{proc_benchmarking}
\alias{proc_tsraking}
\alias{macro_gseriestsbalancing}
\title{Function aliases}
\usage{
proc_benchmarking(...)

proc_tsraking(...)

macro_gseriestsbalancing(...)
}
\arguments{
\item{...}{Corresponding function arguments.}
}
\value{
See the corresponding function:
\itemize{
\item \code{\link[=benchmarking]{benchmarking()}} for alias \code{\link[=proc_benchmarking]{proc_benchmarking()}}
\item \code{\link[=tsraking]{tsraking()}} for alias \code{\link[=proc_tsraking]{proc_tsraking()}}
\item \code{\link[=tsbalancing]{tsbalancing()}} for alias \code{\link[=macro_gseriestsbalancing]{macro_gseriestsbalancing()}}
}
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/aliases.html}})}

\code{\link[=proc_benchmarking]{proc_benchmarking()}} is an alias for \code{\link[=benchmarking]{benchmarking()}}

\code{\link[=proc_tsraking]{proc_tsraking()}} is an alias for \code{\link[=tsraking]{tsraking()}}

\code{\link[=macro_gseriestsbalancing]{macro_gseriestsbalancing()}} is an alias for \code{\link[=tsbalancing]{tsbalancing()}}
}
\details{
The name of these aliases is reminiscent of the corresponding function's SAS\eqn{^\circledR}{®} origin
(PROC BENCHMARKING, PROC TSRAKING and macro \emph{\strong{GSeriesTSBalancing}} in G-Series 2.0). These aliases also
ensure backward compatibility with early development versions of the R package.

See the corresponding function for more details, including complete examples, detailed description of the
arguments and the returned value.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rkMeta_to_blSpecs.R
\name{rkMeta_to_blSpecs}
\alias{rkMeta_to_blSpecs}
\title{Convert reconciliation metadata}
\usage{
rkMeta_to_blSpecs(
  metadata_df,
  alterability_df = NULL,
  alterSeries = 1, 
  alterTotal1 = 0,
  alterTotal2 = 0,
  alterability_df_only = FALSE
)
}
\arguments{
\item{metadata_df}{(mandatory)

Data frame (object of class "data.frame") that describes the cross-sectional aggregation constraints
(additivity rules) for the raking problem. Two character variables must be included in the metadata data frame:
\code{series} and \code{total1}. Two variables are optional: \code{total2} (character) and \code{alterAnnual} (numeric). The values
of variable \code{series} represent the variable names of the component series in the input time series data frame
(argument \code{data_df}). Similarly, the values of variables \code{total1} and \code{total2} represent the variable names of
the 1\if{html}{\out{<sup>}}st\if{html}{\out{</sup>}} and 2\if{html}{\out{<sup>}}nd\if{html}{\out{</sup>}} dimension cross-sectional control totals in the input time series data
frame. Variable \code{alterAnnual} contains the alterability coefficient for the temporal constraint associated to
each component series. When specified, the latter will override the default alterability coefficient specified
with argument \code{alterAnnual}.}

\item{alterability_df}{(optional)

Data frame (object of class "data.frame"), or \code{NULL}, that contains the alterability coefficients variables.
They must correspond to a component series or a cross-sectional control total, that is, a variable with the same
name must exist in the input time series data frame (argument \code{data_df}). The values of these alterability
coefficients will override the default alterability coefficients specified with arguments \code{alterSeries},
\code{alterTotal1} and \code{alterTotal2}. When the input time series data frame contains several observations and the
alterability coefficients data frame contains only one, the alterability coefficients are used (repeated) for
all observations of the input time series data frame. Alternatively, the alterability coefficients data frame
may contain as many observations as the input time series data frame.

\strong{Default value} is \code{alterability_df = NULL} (default alterability coefficients).}

\item{alterSeries}{(optional)

Nonnegative real number specifying the default alterability coefficient for the component series values. It
will apply to component series for which alterability coefficients have not already been specified in the
alterability coefficients data frame (argument \code{alterability_df}).

\strong{Default value} is \code{alterSeries = 1.0} (nonbinding component series values).}

\item{alterTotal1}{(optional)

Nonnegative real number specifying the default alterability coefficient for the 1\if{html}{\out{<sup>}}st\if{html}{\out{</sup>}} dimension
cross-sectional control totals. It will apply to cross-sectional control totals for which alterability
coefficients have not already been specified in the alterability coefficients data frame (argument
\code{alterability_df}).

\strong{Default value} is \code{alterTotal1 = 0.0} (binding 1\if{html}{\out{<sup>}}st\if{html}{\out{</sup>}} dimension cross-sectional control totals)}

\item{alterTotal2}{(optional)

Nonnegative real number specifying the default alterability coefficient for the 2\if{html}{\out{<sup>}}nd\if{html}{\out{</sup>}} dimension
cross-sectional control totals. It will apply to cross-sectional control totals for which alterability
coefficients have not already been specified in the alterability coefficients data frame (argument
\code{alterability_df}).

\strong{Default value} is \code{alterTotal2 = 0.0} (binding 2\if{html}{\out{<sup>}}nd\if{html}{\out{</sup>}} dimension cross-sectional control totals).}

\item{alterability_df_only}{(optional)

Logical argument specifying whether or not only the set of alterability ceofficients found in the alterability
file (argument \code{alterability_df}) should be included in the returned \code{\link[=tsbalancing]{tsbalancing()}} problem specs data frame.
When \code{alterability_df_only = FALSE} (the default), the alterability coefficients specified with arguments
\code{alterSeries}, \code{alterTotal1} and \code{alterTotal2} are combined with those found in \code{alterability_df} (the latter
coefficients overwriting the former) and the returned data frame therefore contains alterability coefficients
for all component and cross-sectional control total series. This argument does not affect the set of temporal
total alterability coefficients (associated to \code{\link[=tsraking]{tsraking()}} argument \code{alterAnnual}) that are included in the
returned \code{\link[=tsbalancing]{tsbalancing()}} problem specs data frame. The latter always strictly contains those specified in
\code{metadata_df} with a non-missing (non-\code{NA}) value for column \code{alterAnnual}.

\strong{Default value} is \code{alterability_df_only = FALSE}.}
}
\value{
A \code{\link[=tsbalancing]{tsbalancing()}} problem specs data frame (argument \code{problem_specs_df}).
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/rkMeta_to_blSpecs.html}})}

Convert a \code{\link[=tsraking]{tsraking()}} metadata data frame to a \code{\link[=tsbalancing]{tsbalancing()}} problem specs data frame.
}
\details{
The preceding description of argument \code{alterability_df} comes from \code{\link[=tsraking]{tsraking()}}. This function
(\code{rkMeta_to_blSpecs()}) slightly changes the specification of alterability coefficients with argument
\code{alterability_df} by allowing either
\itemize{
\item a single observation, specifying the set of alterability coefficients to use for all periods,
\item or one or several observations with an additional column named \code{timeVal} allowing the specification of
both period-specific alterability coefficients (\code{timeVal} is not \code{NA}) and generic coefficients to use for
all other periods (\code{timeVal} is \code{NA}). Values for column \code{timeVal} correspond to the \emph{time values}
of a "ts" object as returned by \code{\link[stats:time]{stats::time()}}, conceptually corresponding to \eqn{year + (period - 1) / 
frequency}.
}

Another difference with \code{\link[=tsraking]{tsraking()}} is that missing (\code{NA}) values are allowed in the alterability coefficients
data frame (argument \code{alterability_df}) and result in using the generic coefficients (observations for which
\code{timeVal} is \code{NA}) or the default coefficients (arguments \code{alterSeries}, \code{alterTotal1} and \code{alterTotal2}).

Note that apart from discarding alterability coefficients for series not listed in the \code{\link[=tsraking]{tsraking()}} metadata
data frame (argument \code{metadata_df}), this function does not validate the values specified in the alterability
coefficients data frame (argument \code{alterability_df}) nor the ones specified with column \code{alterAnnual} in the
\code{\link[=tsraking]{tsraking()}} metadata data frame (argument \code{metadata_df}). The function transfers them \emph{as is} in the
returned \code{\link[=tsbalancing]{tsbalancing()}} problem specs data frame.
}
\examples{
# `tsraking()` metadata for a 2-dimensional raking problem (2 x 2 table)
my_metadata <- data.frame(
  series = c("A1", "A2", "B1", "B2"),
  total1 = c("totA", "totA", "totB", "totB"),
  total2 = c("tot1", "tot2", "tot1", "tot2")
)
my_metadata


# Convert to `tsbalancing()` specifications

# Include the default `tsraking()` alterability coefficients
rkMeta_to_blSpecs(my_metadata)

# Almost binding 1st marginal totals (small alter. coef for columns `totA` and `totB`)
tail(rkMeta_to_blSpecs(my_metadata, alterTotal1 = 1e-6))

# Do not include alterability coefficients (aggregation constraints only)
rkMeta_to_blSpecs(my_metadata, alterability_df_only = TRUE)

# With an alterability coefficients file (argument `alterability_df`)
my_alter = data.frame(B2 = 0.5)
tail(rkMeta_to_blSpecs(my_metadata, alterability_df = my_alter))

# Only include the alterability coefficients from `alterability_df` 
# (i.e., for column `B2` only)
tail(
  rkMeta_to_blSpecs(
    my_metadata, alterability_df = my_alter, 
    alterability_df_only = TRUE
  )
)
}
\seealso{
\code{\link[=tsraking]{tsraking()}} \code{\link[=tsbalancing]{tsbalancing()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_graphTable.R
\name{bench_graphs}
\alias{bench_graphs}
\alias{ori_plot}
\alias{adj_plot}
\alias{GR_plot}
\alias{GR_table}
\title{Generate a benchmarking graphic}
\usage{
ori_plot(
  graphTable,
  title_str = "Original Scale",
  subtitle_str = NULL,
  mth_gap = NULL,
  points_set = NULL,
  pt_sz = 2,
  display_ggplot = TRUE,
  .setup = TRUE
)

adj_plot(
  graphTable,
  title_str = "Adjustment Scale",
  subtitle_str = NULL,
  mth_gap = NULL,
  full_set = NULL,
  pt_sz = 2,
  display_ggplot = TRUE,
  .setup = TRUE
)

GR_plot(
  graphTable,
  title_str = "Growth Rates",
  subtitle_str = NULL,
  factor = NULL,
  type_chars = NULL,
  periodicity = NULL,
  display_ggplot = TRUE,
  .setup = TRUE
)

GR_table(
  graphTable,
  title_str = "Growth Rates Table",
  subtitle_str = NULL,
  factor = NULL,
  type_chars = NULL,
  display_ggplot = TRUE,
  .setup = TRUE
)
}
\arguments{
\item{graphTable}{(mandatory)

Data frame (object of class "data.frame") corresponding to the benchmarking function output\code{graphTable} data frame.}

\item{title_str, subtitle_str}{(optional)

Graphic title and subtitle strings (character constants). \code{subtitle_str} is automatically built from the
\code{graphTable} data frame contents when \code{NULL} and contains the \code{graphTable} data frame name on the 2\if{html}{\out{<sup>}}nd\if{html}{\out{</sup>}} line
and the benchmarking parameters on the 3\if{html}{\out{<sup>}}rd\if{html}{\out{</sup>}} line. Specifying empty strings (\code{""}) would remove the titles.
Simple Markdown and HTML syntax is allowed (e.g., for bold, italic or colored fonts) through package \link[ggtext:ggtext]{ggtext}
(see \code{help(package = "ggtext")}).

\strong{Default values} are \code{subtitle_str = NULL} and a function specific string for \code{title_str} (see \strong{Usage}).}

\item{mth_gap}{(optional)

Number of months between consecutive periods (e.g. 1 for monthly data, 3 for quarterly data, etc.). Based
on the \code{graphTable} data frame contents  when \code{NULL} (calculated as \code{12 / graphTable$periodicity[1]}).

\strong{Default value} is \code{mth_gap = NULL}.}

\item{points_set, full_set}{(optional)

Character vector of the elements (variables of the \code{graphTable} data frame) to include in the plot.
Automatically built when \code{NULL}. See \code{\link[=plot_graphTable]{plot_graphTable()}} for the (default) list of variables used for each
type of graphic.

\strong{Default values} are \code{points_set = NULL} and \code{full_set = NULL}.}

\item{pt_sz}{(optional)

Size of the data points shape (symbol) for ggplot2.

\strong{Default value} is \code{pt_sz = 2}.}

\item{display_ggplot}{(optional)

Logical arguments indicating whether or not the ggplot object(s) should be displayed in the current (active)
graphics device.

\strong{Default value} is \code{display_ggplot = TRUE}.}

\item{.setup}{(optional)

Logical argument indicating whether the setup steps must be executed or not. Must be \code{TRUE} when the function
is called directly (i.e., outside of the \code{\link[=plot_graphTable]{plot_graphTable()}} context).

\strong{Default value} is \code{.setup = TRUE}.}

\item{factor, type_chars}{(optional)

Growth rates factor (1 or 100) and value label suffix ("" or "(\%)") according to the adjustment model parameter
\eqn{\lambda}. Based on the \code{graphTable} data frame contents when \code{NULL} (based on \code{graphTable$lambda[1]}).

\strong{Default values} are \code{factor = NULL} and \code{type_chars = NULL}.}

\item{periodicity}{(optional)

Number of periods in a year. Based on the \code{graphTable} data frame contents when \code{NULL} (defined as
\code{graphTable$periodicity[1]}).

\strong{Default value} is \code{periodicity = NULL}.}
}
\value{
In addition to displaying the corresponding graphic(s) in the current (active) graphics device (except when
\code{display_ggplot = FALSE}), each function also invisibly returns a list containing the generated ggplot object(s).
Notes:
\itemize{
\item \code{\link[=ori_plot]{ori_plot()}} and \code{\link[=adj_plot]{adj_plot()}} generate a single ggplot object (single graphic) while \code{\link[=GR_plot]{GR_plot()}} and \code{\link[=GR_table]{GR_table()}}
will often generate several ggplot objects (several graphics).
\item The returned ggplot object(s) can be displayed \emph{manually} with \code{\link[=print]{print()}}, in which case the following ggplot2 theme
updates (used internally when \code{display_ggplot = TRUE}) are suggested:

\if{html}{\out{<div class="sourceCode R">}}\preformatted{ggplot2::theme_update(
  plot.title = ggtext::element_markdown(hjust = 0.5),
  plot.subtitle = ggtext::element_markdown(hjust = 0.5),
  legend.position = "bottom",
  plot.margin = ggplot2::margin(t = 1.5, r = 1.5, b = 1.5, l = 1.5, unit = "cm")
)
}\if{html}{\out{</div>}}
}
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/bench_graphs.html}})}

Functions used internally by \code{\link[=plot_graphTable]{plot_graphTable()}} to generate the benchmarking graphics in a PDF file:
\itemize{
\item \code{\link[=ori_plot]{ori_plot()}}: Original Scale Plot (\code{\link[=plot_graphTable]{plot_graphTable()}} argument \code{ori_plot_flag = TRUE})
\item \code{\link[=adj_plot]{adj_plot()}}: Adjustment Scale Plot (\code{\link[=plot_graphTable]{plot_graphTable()}} argument \code{adj_plot_flag = TRUE})
\item \code{\link[=GR_plot]{GR_plot()}}: Growth Rates Plot (\code{\link[=plot_graphTable]{plot_graphTable()}} argument \code{GR_plot_flag = TRUE})
\item \code{\link[=GR_table]{GR_table()}}: Growth Rates Table (\code{\link[=plot_graphTable]{plot_graphTable()}} argument \code{GR_table_flag = TRUE})
}

When these functions are called directly, the \code{graphTable} data frame should only  contain a \strong{single series}
and the graphic is generated in the current (active) graphics device.
}
\details{
See \code{\link[=plot_graphTable]{plot_graphTable()}} for a detailed description of the four benchmarking graphics associated to these
individual functions. These graphics are optimized for the US Letter paper size format in landscape view, i.e.,
11in wide (27.9cm, 1056px with 96 DPI) and 8.5in tall (21.6cm, 816px with 96 DPI). Keep this in mind when
viewing or saving graphics generated by calls to these individual functions (i.e., outside of the \code{\link[=plot_graphTable]{plot_graphTable()}}
context). Also note that \code{\link[=GR_plot]{GR_plot()}} and \code{\link[=GR_table]{GR_table()}} will often generate more than one graphic (more than one
\emph{page}), unless the number of periods included in the input \code{graphTable} data frame is reduced (e.g., subsetting
the data frame by ranges of calendar years).
}
\examples{
# Deactivate the graphics device creation for the pkgdown website HTML reference page
# (irrelevant in that context)
new_grDev <- !(identical(Sys.getenv("IN_PKGDOWN"), "true"))


# Initial quarterly time series (indicator series to be benchmarked)
qtr_ts <- ts(
  c(1.9, 2.4, 3.1, 2.2, 2.0, 2.6, 3.4, 2.4, 2.3),
  start = c(2015, 1), frequency = 4
)

# Annual time series (benchmarks)
ann_ts <- ts(c(10.3, 10.2), start = 2015, frequency = 1)

# Proportional benchmarking
out_bench <- benchmarking(
  ts_to_tsDF(qtr_ts), 
  ts_to_bmkDF(ann_ts, ind_frequency = 4),
  rho = 0.729, lambda = 1, biasOption = 3,
  quiet = TRUE
)


# Open a new graphics device that is 11in wide and 8.5in tall 
# (US Letter paper size format in landscape view)
if (new_grDev) {
  dev.new(width = 11, height = 8.5, unit = "in", noRStudioGD = TRUE)
}

# Generate the benchmarking graphics
ori_plot(out_bench$graphTable)
adj_plot(out_bench$graphTable)
GR_plot(out_bench$graphTable)
GR_table(out_bench$graphTable)


# Simulate multiple series benchmarking (3 series)

qtr_mts <- ts.union(ser1 = qtr_ts, ser2 = qtr_ts * 100, ser3 = qtr_ts * 10)
ann_mts <- ts.union(ser1 = ann_ts, ser2 = ann_ts * 100, ser3 = ann_ts * 10)

# Using argument `allCols = TRUE` (identify series with column `varSeries`)
out_bench2 <- benchmarking(
  ts_to_tsDF(qtr_mts),
  ts_to_bmkDF(ann_mts, ind_frequency = 4),
  rho = 0.729, lambda = 1, biasOption = 3,
  allCols = TRUE,
  quiet = TRUE
)

# Original and adjustment scale plots for the 2nd series (ser2)
ser2_res <- out_bench2$graphTable[out_bench2$graphTable$varSeries == "ser2", ]
ori_plot(ser2_res)
adj_plot(ser2_res)

# Using argument `by = "series"` (identify series with column `series`)
out_bench3 <- benchmarking(
  stack_tsDF(ts_to_tsDF(qtr_mts)),
  stack_bmkDF(ts_to_bmkDF(ann_mts, ind_frequency = 4)),
  rho = 0.729, lambda = 1, biasOption = 3,
  by = "series",
  quiet = TRUE
)
						   
# Growth rates plot for the 3rd series (ser3)
ser3_res <- out_bench3$graphTable[out_bench3$graphTable$series == "ser3", ]
GR_plot(ser3_res)


# Close the graphics device 
if (new_grDev) {
  dev.off()
}
}
\seealso{
\code{\link[=plot_graphTable]{plot_graphTable()}} \code{\link[=plot_benchAdj]{plot_benchAdj()}} \code{\link[=benchmarking]{benchmarking()}} \code{\link[=stock_benchmarking]{stock_benchmarking()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsraking.R
\name{tsraking}
\alias{tsraking}
\title{Restore cross-sectional (contemporaneous) aggregation constraints}
\usage{
tsraking(
  data_df,
  metadata_df,
  alterability_df = NULL,
  alterSeries = 1,
  alterTotal1 = 0,
  alterTotal2 = 0,
  alterAnnual = 0,
  tolV = 0.001,
  tolP = NA,
  warnNegResult = TRUE,
  tolN = -0.001,
  id = NULL,
  verbose = FALSE,

  # New in G-Series 3.0
  Vmat_option = 1,
  warnNegInput = TRUE,
  quiet = FALSE
)
}
\arguments{
\item{data_df}{(mandatory)

Data frame (object of class "data.frame") that contains the time series data to be reconciled. It must minimally
contain variables corresponding to the component series and cross-sectional control totals specified in the
metadata data frame (argument \code{metadata_df}). If more than one observation (period) is provided, the sum of
the provided component series values will also be preserved as part of implicit temporal constraints.}

\item{metadata_df}{(mandatory)

Data frame (object of class "data.frame") that describes the cross-sectional aggregation constraints
(additivity rules) for the raking problem. Two character variables must be included in the metadata data frame:
\code{series} and \code{total1}. Two variables are optional: \code{total2} (character) and \code{alterAnnual} (numeric). The values
of variable \code{series} represent the variable names of the component series in the input time series data frame
(argument \code{data_df}). Similarly, the values of variables \code{total1} and \code{total2} represent the variable names of
the 1\if{html}{\out{<sup>}}st\if{html}{\out{</sup>}} and 2\if{html}{\out{<sup>}}nd\if{html}{\out{</sup>}} dimension cross-sectional control totals in the input time series data
frame. Variable \code{alterAnnual} contains the alterability coefficient for the temporal constraint associated to
each component series. When specified, the latter will override the default alterability coefficient specified
with argument \code{alterAnnual}.}

\item{alterability_df}{(optional)

Data frame (object of class "data.frame"), or \code{NULL}, that contains the alterability coefficients variables.
They must correspond to a component series or a cross-sectional control total, that is, a variable with the same
name must exist in the input time series data frame (argument \code{data_df}). The values of these alterability
coefficients will override the default alterability coefficients specified with arguments \code{alterSeries},
\code{alterTotal1} and \code{alterTotal2}. When the input time series data frame contains several observations and the
alterability coefficients data frame contains only one, the alterability coefficients are used (repeated) for
all observations of the input time series data frame. Alternatively, the alterability coefficients data frame
may contain as many observations as the input time series data frame.

\strong{Default value} is \code{alterability_df = NULL} (default alterability coefficients).}

\item{alterSeries}{(optional)

Nonnegative real number specifying the default alterability coefficient for the component series values. It
will apply to component series for which alterability coefficients have not already been specified in the
alterability coefficients data frame (argument \code{alterability_df}).

\strong{Default value} is \code{alterSeries = 1.0} (nonbinding component series values).}

\item{alterTotal1}{(optional)

Nonnegative real number specifying the default alterability coefficient for the 1\if{html}{\out{<sup>}}st\if{html}{\out{</sup>}} dimension
cross-sectional control totals. It will apply to cross-sectional control totals for which alterability
coefficients have not already been specified in the alterability coefficients data frame (argument
\code{alterability_df}).

\strong{Default value} is \code{alterTotal1 = 0.0} (binding 1\if{html}{\out{<sup>}}st\if{html}{\out{</sup>}} dimension cross-sectional control totals)}

\item{alterTotal2}{(optional)

Nonnegative real number specifying the default alterability coefficient for the 2\if{html}{\out{<sup>}}nd\if{html}{\out{</sup>}} dimension
cross-sectional control totals. It will apply to cross-sectional control totals for which alterability
coefficients have not already been specified in the alterability coefficients data frame (argument
\code{alterability_df}).

\strong{Default value} is \code{alterTotal2 = 0.0} (binding 2\if{html}{\out{<sup>}}nd\if{html}{\out{</sup>}} dimension cross-sectional control totals).}

\item{alterAnnual}{(optional)

Nonnegative real number specifying the default alterability coefficient for the component series temporal
constraints (e.g., annual totals). It will apply to component series for which alterability coefficients
have not already been specified in the metadata data frame (argument \code{metadata_df}).

\strong{Default value} is \code{alterAnnual = 0.0} (binding temporal control totals).}

\item{tolV, tolP}{(optional)

Nonnegative real number, or \code{NA}, specifying the tolerance, in absolute value or percentage, to be used
when performing the ultimate test in the case of binding totals (alterability coefficient of \eqn{0.0}
for temporal or cross-sectional control totals). The test compares the input binding control totals with
the ones calculated from the reconciled (output) component series. Arguments \code{tolV} and \code{tolP} cannot be both
specified together (one must be specified while the other must be \code{NA}).

\strong{Example:} to set a tolerance of 10 \emph{units}, specify \verb{tolV = 10, tolP = NA}; to set a tolerance of 1\%,
specify \verb{tolV = NA, tolP = 0.01}.

\strong{Default values} are \code{tolV = 0.001} and \code{tolP = NA}.}

\item{warnNegResult}{(optional)

Logical argument specifying whether a warning message is generated when a negative value created by the
function in the reconciled (output) series is smaller than the threshold specified by argument \code{tolN}.

\strong{Default value} is \code{warnNegResult = TRUE}.}

\item{tolN}{(optional)

Negative real number specifying the threshold for the identification of negative values. A value is
considered negative when it is smaller than this threshold.

\strong{Default value} is \code{tolN = -0.001}.}

\item{id}{(optional)

String vector (minimum length of 1), or \code{NULL}, specifying the name of additional variables to be transferred
from the input time series data frame (argument \code{data_df}) to the output time series data frame, the
object returned by the function (see section \strong{Value}). By default, the output series data frame only contains
the variables listed in the metadata data frame (argument \code{metadata_df}).

\strong{Default value} is \code{id = NULL}.}

\item{verbose}{(optional)

Logical argument specifying whether information on intermediate steps with execution time (real time,
not CPU time) should be displayed. Note that specifying argument \code{quiet = TRUE} would \emph{nullify} argument
\code{verbose}.

\strong{Default value} is \code{verbose = FALSE}.}

\item{Vmat_option}{(optional)

Specification of the option for the variance matrices (\eqn{V_e} and \eqn{V_\epsilon}; see section \strong{Details}):\tabular{cl}{
   \strong{Value} \tab \strong{Description} \cr
   \code{1} \tab Use vectors \eqn{x} and \eqn{g} in the variance matrices. \cr
   \code{2} \tab Use vectors \eqn{|x|} and \eqn{|g|} in the variance matrices. \cr
}


See Ferland (2016) and subsection \strong{Arguments \code{Vmat_option} and \code{warnNegInput}} in section \strong{Details} for
more information.

\strong{Default value} is \code{Vmat_option = 1}.}

\item{warnNegInput}{(optional)

Logical argument specifying whether a warning message is generated when a negative value smaller than
the threshold specified by argument \code{tolN} is found in the input time series data frame (argument \code{data_df}).

\strong{Default value} is \code{warnNegInput = TRUE}.}

\item{quiet}{(optional)

Logical argument specifying whether or not to display only essential information such as warnings and errors.
Specifying \code{quiet = TRUE} would also \emph{nullify} argument \code{verbose} and is equivalent to \emph{wrapping} your
\code{\link[=tsraking]{tsraking()}} call with \code{\link[=suppressMessages]{suppressMessages()}}.

\strong{Default value} is \code{quiet = FALSE}.}
}
\value{
The function returns a data frame containing the reconciled component series, reconciled cross-sectional control
totals and variables specified with  argument \code{id}. Note that the "data.frame" object can be explicitly coerced to
another type of object with the appropriate \verb{as*()} function (e.g., \code{tibble::as_tibble()} would coerce it to a tibble).
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/tsraking.html}})}

\emph{Replication of the G-Series 2.0 SAS\eqn{^\circledR}{®} TSRAKING procedure (PROC TSRAKING). See the
G-Series 2.0 documentation for details (Statistics Canada 2016).}

This function will restore cross-sectional aggregation constraints in a system of time series. The
aggregation constraints may come from a 1 or 2-dimensional table. Optionally, temporal constraints
can also be preserved.

\code{\link[=tsraking]{tsraking()}} is usually called in practice through \code{\link[=tsraking_driver]{tsraking_driver()}} in order to reconcile
all periods of the time series system in a single function call.
}
\details{
This function returns the generalized least squared solution of a specific, simple variant of the general
regression-based raking model proposed by Dagum and Cholette (Dagum and Cholette 2006). The model, in matrix form, is:
\deqn{\displaystyle
\begin{bmatrix} x \\ g \end{bmatrix} = 
\begin{bmatrix} I \\ G \end{bmatrix} \theta + 
\begin{bmatrix} e \\ \varepsilon \end{bmatrix}
}{[x; g] = [I; G] theta + [e; epsilion]}
where
\itemize{
\item \eqn{x} is the vector of the initial component series values.
\item \eqn{\theta} is the vector of the final (reconciled) component series values.
\item \eqn{e \sim \left( 0, V_e \right)}{e ~ (0, V_e)} is the vector of the measurement errors of \eqn{x} with covariance
matrix \eqn{V_e = \mathrm{diag} \left( c_x x \right)}{V_e = diag(c_x x)}, or \eqn{V_e = \mathrm{diag} \left( \left| 
c_x x \right| \right)}{V_e = diag(|c_x x|)} when argument \code{Vmat_option = 2}, where \eqn{c_x} is the vector of the
alterability coefficients of \eqn{x}.
\item \eqn{g} is the vector of the initial control totals, including the component series temporal totals (when
applicable).
\item \eqn{\varepsilon \sim (0, V_\varepsilon)}{epsilon ~ (0, V_epsilon)} is the vector of the measurement errors of
\eqn{g} with covariance matrix \eqn{V_\varepsilon = \mathrm{diag} \left( c_g g \right)}{V_epsilion = diag(c_g g)}, or
\eqn{V_\varepsilon = \mathrm{diag} \left( \left| c_g g \right| \right)}{V_epsilon = diag(|c_g g|)} when argument
\code{Vmat_option = 2}, where \eqn{c_g} is the vector of the alterability coefficients of \eqn{g}.
\item \eqn{G} is the matrix of aggregation constraints, including the implicit temporal constraints (when applicable).
}

The generalized least squared solution is:
\deqn{\displaystyle 
\hat{\theta} = x + V_e G^{\mathrm{T}} \left( G V_e G^{\mathrm{T}} + V_\varepsilon \right)^+ \left( g - G x \right)
}{theta^hat = x + V_e G' (G V_e G' + V_epsilion)^{+} (g - G x)}
where \eqn{A^{+}} designates the Moore-Penrose inverse of matrix \eqn{A}.

\code{\link[=tsraking]{tsraking()}} solves a single raking problem, i.e., either a single period of the time series system, or a single
temporal group (e.g., all periods of a given year) when temporal total preservation is required. Several call to
\code{\link[=tsraking]{tsraking()}} are therefore necessary in order to reconcile all the periods of the time series system.
\code{\link[=tsraking_driver]{tsraking_driver()}} can achieve this in a single call: it conveniently determines the required set of raking
problems to be solved and internally generates the individual calls to \code{\link[=tsraking]{tsraking()}}.
\subsection{Alterability Coefficients}{

Alterability coefficients \eqn{c_x} and \eqn{c_g} conceptually represent the measurement errors associated with the
input component series values \eqn{x} and control totals \eqn{g} respectively. They are nonnegative real numbers which,
in practice, specify the extent to which an initial value can be modified in relation to other values. Alterability
coefficients of \eqn{0.0} define fixed (binding) values while alterability coefficients greater than \eqn{0.0} define
free (nonbinding) values. Increasing the alterability coefficient of an intial value results in more changes for that
value in the reconciled (output) data and, conversely, less changes when decreasing the alterability coefficient. The
default alterability coefficients are \eqn{1.0} for the component series values and \eqn{0.0} for the cross-sectional
control totals and, when applicable, the component series temporal totals. These default alterability coefficients
result in a proportional allocation of the discrepancies to the component series. Setting the component series
alterability coefficients to the inverse of the component series initial values would result in a uniform allocation
of the discrepancies instead. \emph{Almost binding} totals can be obtained in practice by specifying very small
(almost \eqn{0.0}) alterability coefficients relative to those of the (nonbinding) component series.

\strong{Temporal total preservation} refers to the fact that temporal totals, when applicable, are usually kept
“as close as possible” to their initial value. \emph{Pure preservation} is achieved by default with binding temporal
totals while the change is minimized with nonbinding temporal totals (in accordance with the set of alterability
coefficients).
}

\subsection{Arguments \code{Vmat_option} and \code{warnNegInput}}{

These arguments allow for an alternative handling of negative values in the input data, similar to that of \code{\link[=tsbalancing]{tsbalancing()}}.
Their default values correspond to the G-Series 2.0 behaviour (SAS\eqn{^\circledR}{®} PROC TSRAKING) for which equivalent
options are not defined. The latter was developed with "nonnegative input data only" in mind, similar to SAS\eqn{^\circledR}{®}
PROC BENCHMARKING in G-Series 2.0 that did not allow negative values either with proportional benchmarking, which explains
the "suspicious use of proportional raking" warning in presence of negative values with PROC TSRAKING in G-Series 2.0 and
when \code{warnNegInput = TRUE} (default). However, (proportional) raking in the presence of negative values generally works well
with \code{Vmat_option = 2} and produces reasonable, intuitive solutions. E.g., while the default \code{Vmat_option = 1} fails at
solving constraint \code{A + B = C} with input data \code{A = 2}, \code{B = -2}, \code{C = 1} and the default alterability coefficients,
\code{Vmat_option = 2} returns the (intuitive) solution \code{A = 2.5}, \code{B = -1.5}, \code{C = 1} (25\% increase for \code{A} and \code{B}). See
Ferland (2016) for more details.
}

\subsection{Treatment of Missing (\code{NA}) Values}{

Missing values in the input time series data frame (argument \code{data_df}) or alterability coefficients data frame
(argument \code{alterability_df}) for any of the raking problem data (variables listed in the metadata data frame
with argument \code{metadata_df}) will generate an error message and stop the function execution.
}
}
\section{Comparing \code{\link[=tsraking]{tsraking()}} and \code{\link[=tsbalancing]{tsbalancing()}}}{
\itemize{
\item \code{\link[=tsraking]{tsraking()}} is limited to one- and two-dimensional aggregation table raking problems (with temporal
total preservation if required) while \code{\link[=tsbalancing]{tsbalancing()}} handles more general balancing problems (e.g.,
higher dimensional raking problems, nonnegative solutions, general linear equality and inequality
constraints as opposed to aggregation rules only, etc.).
\item \code{\link[=tsraking]{tsraking()}} returns the generalized least squared solution of the Dagum and Cholette regression-based
raking model (Dagum and Cholette 2006) while \code{\link[=tsbalancing]{tsbalancing()}} solves the corresponding quadratic
minimization problem using a numerical solver. In most cases, \emph{convergence to the minimum} is achieved
and the \code{\link[=tsbalancing]{tsbalancing()}} solution matches the (exact) \code{\link[=tsraking]{tsraking()}} least square solution. It may not be
the case, however, if convergence could not be achieved after a reasonable number of iterations.
Having said that, only in very rare occasions will the \code{\link[=tsbalancing]{tsbalancing()}} solution \emph{significantly} differ
from the \code{\link[=tsraking]{tsraking()}} solution.
\item \code{\link[=tsbalancing]{tsbalancing()}} is usually faster than \code{\link[=tsraking]{tsraking()}}, especially for large raking problems, but is
generally more sensitive to the presence of (small) inconsistencies in the input data associated to
the redundant constraints of fully specified (over-specified) raking problems. \code{\link[=tsraking]{tsraking()}} handles
these inconsistencies by using the Moore-Penrose inverse (uniform distribution among all binding totals).
\item \code{\link[=tsbalancing]{tsbalancing()}} accommodates the specification of sparse problems in their reduced form. This is not
true in the case of \code{\link[=tsraking]{tsraking()}} where aggregation rules must always be fully specified since a
\emph{complete data cube} without missing data is expected as input (every single \emph{inner-cube} component
series must contribute to all dimensions of the cube, i.e., to every single \emph{outer-cube} marginal
total series).
\item Both tools handle negative values in the input data differently by default. While the solutions of
raking problems obtained from \code{\link[=tsbalancing]{tsbalancing()}} and \code{\link[=tsraking]{tsraking()}} are identical when all input data
points are positive, they will differ if some data points are negative (unless argument
\code{Vmat_option = 2} is specified with \code{\link[=tsraking]{tsraking()}}).
\item While both \code{\link[=tsbalancing]{tsbalancing()}} and \code{\link[=tsraking]{tsraking()}} allow the preservation of temporal totals, time
management is not incorporated in \code{\link[=tsraking]{tsraking()}}. For example, the construction of the processing groups
(sets of periods of each raking problem) is left to the user with \code{\link[=tsraking]{tsraking()}} and separate calls must
be submitted for each processing group (each raking problem). That's where helper function
\code{\link[=tsraking_driver]{tsraking_driver()}} comes in handy with \code{\link[=tsraking]{tsraking()}}.
\item The marginal totals (\emph{outer-cube} marginal total series) returned by \code{\link[=tsraking]{tsraking()}} are always
recalculated as the sum of the relevant component series (\emph{inner-cube} component series). The output
values of \strong{binding} marginal totals may therefore slightly differ from the input values. The
magnitude of the differences depends on the initial inconsistencies (that end up being uniformly
distributed by the Moore-Penrose inverse). This is not the case for \code{\link[=tsbalancing]{tsbalancing()}} where the values
of \strong{binding} marginal totals always remain perfectly unchanged, potentially resulting in (small)
discrepancies between the sum of the returned component series and the corresponding marginal totals.
\item \code{\link[=tsbalancing]{tsbalancing()}} returns the same set of series as the input time series object while \code{\link[=tsraking]{tsraking()}}
returns the set of series involved in the raking problem plus those specified with argument \code{id}
(which could correspond to a subset of the input series).
}
}

\examples{
###########
# Example 1: Simple 1-dimensional raking problem where the values of `cars` and `vans`
#            must sum up to the value of `total`.

# Problem metadata
my_metadata1 <- data.frame(
  series = c("cars", "vans"),
  total1 = c("total", "total")
)
my_metadata1

# Problem data
my_series1 <- data.frame(cars = 25, vans = 5, total = 40)

# Reconcile the data
out_raked1 <- tsraking(my_series1, my_metadata1)

# Initial data
my_series1

# Reconciled data
out_raked1

# Check the output cross-sectional constraint
all.equal(rowSums(out_raked1[c("cars", "vans")]), out_raked1$total)

# Check the control total (fixed)
all.equal(my_series1$total, out_raked1$total)


###########
# Example 2: 2-dimensional raking problem similar to the 1st example but adding
#            regional sales for the 3 prairie provinces (Alb., Sask. and Man.)
#            and where the sales of vans in Sask. are non-alterable
#            (alterability coefficient = 0), with `quiet = TRUE` to avoid
#            displaying the function header.

# Problem metadata
my_metadata2 <- data.frame(
  series = c(
    "cars_alb", "cars_sask", "cars_man",
    "vans_alb", "vans_sask", "vans_man"
  ),
  total1 = c(
    rep("cars_total", 3),
    rep("vans_total", 3)
  ),
  total2 = rep(c("alb_total", "sask_total", "man_total"), 2)
)
my_metadata2

# Problem data
my_series2 <- data.frame(
  cars_alb = 12, cars_sask = 14, cars_man = 13,
  vans_alb = 20, vans_sask = 20, vans_man = 24,
  alb_total = 30, sask_total = 31, man_total = 32,
  cars_total = 40, vans_total = 53
)

# Reconciled data
out_raked2 <- tsraking(
  my_series2, my_metadata2,
  alterability_df = data.frame(vans_sask = 0),
  quiet = TRUE
)

# Initial data
my_series2

# Reconciled data
out_raked2

# Check the output cross-sectional constraints
all.equal(
  rowSums(out_raked2[c("cars_alb", "cars_sask", "cars_man")]), 
  out_raked2$cars_total
)
all.equal(
  rowSums(out_raked2[c("vans_alb", "vans_sask", "vans_man")]), 
  out_raked2$vans_total
)
all.equal(
  rowSums(out_raked2[c("cars_alb", "vans_alb")]), 
  out_raked2$alb_total
)
all.equal(
  rowSums(out_raked2[c("cars_sask", "vans_sask")]), 
  out_raked2$sask_total
)
all.equal(
  rowSums(out_raked2[c("cars_man", "vans_man")]), 
  out_raked2$man_total
)

# Check the control totals (fixed)
tot_cols <- union(unique(my_metadata2$total1), unique(my_metadata2$total2))
all.equal(my_series2[tot_cols], out_raked2[tot_cols])

# Check the value of vans in Saskatchewan (fixed at 20)
all.equal(my_series2$vans_sask, out_raked2$vans_sask)
}
\references{
Bérubé, J. and S. Fortier (2009). "PROC TSRAKING: An in-house SAS\eqn{^\circledR}{®} procedure for balancing
time series". In \strong{JSM Proceedings, Business and Economic Statistics Section}. Alexandria, VA: American Statistical
Association.

Dagum, E. B. and P. Cholette (2006). \strong{Benchmarking, Temporal Distribution and Reconciliation Methods
of Time Series}. Springer-Verlag, New York, Lecture Notes in Statistics, Vol. 186. \doi{10.1007/0-387-35439-5}

Ferland, M. (2016). "Negative Values with PROC TSRAKING". \strong{Internal document}. Statistics Canada, Ottawa,
Canada.

Fortier, S. and B. Quenneville (2009). "Reconciliation and Balancing of Accounts and Time Series".
In \strong{JSM Proceedings, Business and Economic Statistics Section}. Alexandria, VA: American Statistical Association.

Quenneville, B. and S. Fortier (2012). "Restoring Accounting Constraints in Time Series – Methods and
Software for a Statistical Agency". \strong{Economic Time Series: Modeling and Seasonality}. Chapman & Hall, New York.

Statistics Canada (2016). "The TSRAKING Procedure". \strong{G-Series 2.0 User Guide}. Statistics Canada,
Ottawa, Canada.

Statistics Canada (2018). \strong{Theory and Application of Reconciliation (Course code 0437)}.
Statistics Canada, Ottawa, Canada.
}
\seealso{
\code{\link[=tsraking_driver]{tsraking_driver()}} \code{\link[=tsbalancing]{tsbalancing()}} \code{\link[=rkMeta_to_blSpecs]{rkMeta_to_blSpecs()}} \code{\link[=gs.gInv_MP]{gs.gInv_MP()}} \code{\link[=build_raking_problem]{build_raking_problem()}} \link{aliases}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stack_tsDF.R
\name{stack_tsDF}
\alias{stack_tsDF}
\title{Stack time series data}
\usage{
stack_tsDF(
  ts_df,
  ser_cName = "series",
  yr_cName = "year",
  per_cName = "period",
  val_cName = "value",
  keep_NA = FALSE
)
}
\arguments{
\item{ts_df}{(mandatory)

Data frame (object of class "data.frame") that contains the multivariate time series data to be stacked.}

\item{ser_cName}{(optional)

String specifying the name of the character variable (column) in the output stacked data frame that will
contain the series names (name of the time series variables in the input multivariate time series data
frame). This variable can then be used as the BY-group variable (argument \code{by}) with the benchmarking functions.

\strong{Default value} is \code{ser_cName = "series"}.}

\item{yr_cName, per_cName}{(optional)

Strings specifying the name of the numeric variables (columns) in the input multivariate time series data frame
that contain the data point year and period (cycle) identifiers. These variables are \emph{transferred} to the output
stacked data frame with the same variable names.

\strong{Default values} are \code{yr_cName = "year"} and \code{per_cName   = "period"}.}

\item{val_cName}{(optional)

String specifying the name of the numeric variable (column) in the output stacked data frame that will
contain the data point values.

\strong{Default value} is \code{val_cName = "value"}.}

\item{keep_NA}{(optional)

Logical argument specifying whether \code{NA} time series values in the input multivariate time series data frame
should be kept in the output stacked data frame.

\strong{Default value} is \code{keep_NA = FALSE}.}
}
\value{
The function returns a data frame with four variables:
\itemize{
\item Series name, type character (see argument \code{ser_cName})
\item Data point year, type numeric (see argument \code{yr_cName})
\item Data point period, type numeric (see argument \code{per_cName})
\item Data point value, type numeric (see argument \code{val_cName})
}

Note: the function returns a "data.frame" object than can be explicitly coerced to another type of object
with the appropriate \verb{as*()} function (e.g., \code{tibble::as_tibble()} would coerce it to a tibble).
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/stack_tsDF.html}})}

Convert a multivariate time series data frame (see \code{\link[=ts_to_tsDF]{ts_to_tsDF()}}) for the benchmarking functions
(\code{\link[=benchmarking]{benchmarking()}} and \code{\link[=stock_benchmarking]{stock_benchmarking()}}) into a stacked (tall) data frame with four variables (columns):
\itemize{
\item one (1) for the series name
\item two (2) for the data point identification (year and period)
\item one (1) for the data point value
}

Missing (\code{NA}) series values are not included in the output stacked data frame by default. Specify argument
\code{keep_NA = TRUE} in order to keep them.

This function is useful when intending to use the \code{by} argument (\emph{BY-group} processing mode) of the benchmarking
functions in order to benchmark multiple series in a single function call.
}
\examples{
# Create a data frame with 2 quarterly indicators series
# (with missing values for the last 2 quarters)
my_indicators <- ts_to_tsDF(
  ts(
    data.frame(
      ser1 = c(1:5 *  10, NA, NA),
      ser2 = c(1:5 * 100, NA, NA)
    ), 
    start = c(2019, 1), frequency = 4
  )
)
my_indicators


# Stack the indicator series ...

# discarding `NA` values in the output stacked data frame (default behavior)
stack_tsDF(my_indicators)

# keeping `NA` values in the output stacked data frame
stack_tsDF(my_indicators, keep_NA = TRUE)

# using custom variable (column) names
stack_tsDF(my_indicators, ser_cName = "ser_name", val_cName = "ser_val")
}
\seealso{
\code{\link[=unstack_tsDF]{unstack_tsDF()}} \code{\link[=stack_bmkDF]{stack_bmkDF()}} \code{\link[=ts_to_tsDF]{ts_to_tsDF()}} \code{\link[=benchmarking]{benchmarking()}} \code{\link[=stock_benchmarking]{stock_benchmarking()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts_to_tsDF.R
\name{ts_to_tsDF}
\alias{ts_to_tsDF}
\title{Convert a "ts" object to a time series data frame}
\usage{
ts_to_tsDF(
  in_ts,
  yr_cName = "year",
  per_cName = "period",
  val_cName = "value"
)
}
\arguments{
\item{in_ts}{(mandatory)

Time series (object of class "ts" or "mts") to be converted.}

\item{yr_cName, per_cName}{(optional)

Strings specifying the name of the numeric variables (columns) in the output data frame that will contain
the data point year and period identifiers.

\strong{Default values} are \code{yr_cName = "year"} and \code{per_cName   = "period"}.}

\item{val_cName}{(optional)

String specifying the name of the numeric variable (column) in the output data frame that will contain the
data point value. This argument has no effect for "mts" objects (time series data variable names are
automatically inherited from the "mts" object).

\strong{Default value} is \code{val_cName = "value"}.}
}
\value{
The function returns a data frame with three or more variables:
\itemize{
\item Data point year, type numeric (see argument \code{startYr_cName})
\item Data point period, type numeric (see argument \code{startPer_cName})
\item One ("ts" object) or many ("mts" object) time series data variable(s), type numeric (see argument \code{val_cName})
}

Note: the function returns a "data.frame" object than can be explicitly coerced to another type of object
with the appropriate \verb{as*()} function (e.g., \code{tibble::as_tibble()} would coerce it to a tibble).
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/ts_to_tsDF.html}})}

Convert a "ts" (or "mts") object into a time series data frame for the benchmarking functions with three or more
variables (columns):
\itemize{
\item two (2) for the data point identification (year and period)
\item one (1) for each time series
}
}
\examples{
# Quarterly time series
my_ts <- ts(1:10 * 100, start = 2019, frequency = 4)
my_ts


# With the default variable (column) names
ts_to_tsDF(my_ts)

# Using a custom name for the time series data variable (column)
ts_to_tsDF(my_ts, val_cName = "ser_val")


# Multiple time series: argument `val_cName` ignored
# (the "mts" object column names are always used)
ts_to_tsDF(
  ts.union(ser1 = my_ts, ser2 = my_ts / 10),
  val_cName = "useless_column_name"
)
}
\seealso{
\code{\link[=tsDF_to_ts]{tsDF_to_ts()}} \code{\link[=ts_to_bmkDF]{ts_to_bmkDF()}} \code{\link[=stack_tsDF]{stack_tsDF()}} \code{\link[=benchmarking]{benchmarking()}} \code{\link[=stock_benchmarking]{stock_benchmarking()}} \link{time_values_conv}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/benchmarking.R
\name{benchmarking}
\alias{benchmarking}
\title{Restore temporal constraints}
\usage{
benchmarking(
  series_df,
  benchmarks_df,
  rho,
  lambda,
  biasOption,
  bias = NA,
  tolV = 0.001,
  tolP = NA,
  warnNegResult = TRUE,
  tolN = -0.001,
  var = "value",
  with = NULL,
  by = NULL,
  verbose = FALSE,

  # New in G-Series 3.0
  constant = 0,
  negInput_option = 0,
  allCols = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{series_df}{(mandatory)

Data frame (object of class "data.frame") that contains the indicator time series data to be
benchmarked. In addition to the series data variable(s), specified with argument \code{var}, the data
frame must also contain two numeric variables, \code{year} and \code{period}, identifying the periods of
the indicator time series.}

\item{benchmarks_df}{(mandatory)

Data frame (object of class "data.frame") that contains the benchmarks. In addition to the benchmarks
data variable(s), specified with argument \code{with}, the data frame must also contain four numeric
variables, \code{startYear}, \code{startPeriod}, \code{endYear} and \code{endPeriod}, identifying the indicator time
series periods covered by each benchmark.}

\item{rho}{(mandatory)

Real number in the \eqn{[0,1]} interval that specifies the value of the autoregressive parameter
\eqn{\rho}. See section \strong{Details} for more information on the effect of parameter \eqn{\rho}.}

\item{lambda}{(mandatory)

Real number, with suggested values in the \eqn{[-3,3]} interval, that specifies the value of the
adjustment model parameter \eqn{\lambda}. Typical values are \code{lambda = 0.0} for an additive model
and \code{lambda = 1.0} for a proportional model.}

\item{biasOption}{(mandatory)

Specification of the bias estimation option:
\itemize{
\item \code{1}: Do not estimate the bias. The bias used to correct the indicator series will be the value
specified with argument \code{bias}.
\item \code{2}: Estimate the bias, display the result, but do not use it. The bias used to correct the indicator
series will be the value specified with argument \code{bias}.
\item \code{3}: Estimate the bias, display the result and use the estimated bias to correct the indicator series.
Any value specified with argument \code{bias} will be ignored.
}

Argument \code{biasOption} is ignored when \code{rho = 1.0}. See section \strong{Details} for more information on the bias.}

\item{bias}{(optional)

Real number, or \code{NA}, specifying the value of the user-defined bias to be used for the correction of the
indicator series prior to benchmarking. The bias is added to the indicator series with an additive model
(argument  \code{lambda = 0.0}) while it is multiplied otherwise (argument \code{lambda != 0.0}). No bias
correction is applied when \code{bias = NA}, which is equivalent to specifying \code{bias = 0.0} when \code{lambda = 0.0}
and \code{bias = 1.0} otherwise. Argument \code{bias} is ignored when \code{biasOption = 3} or \code{rho = 1.0}. See section
\strong{Details} for more information on the bias.

\strong{Default value} is \code{bias = NA} (no user-defined bias).}

\item{tolV, tolP}{(optional)

Nonnegative real number, or \code{NA}, specifying the tolerance, in absolute value or percentage, to be used
for the validation of the output binding benchmarks (alterability coefficient of \eqn{0.0}). This validation
compares the input binding benchmark values with the equivalent values calculated from the benchmarked series
(output) data. Arguments \code{tolV} and \code{tolP} cannot be both specified (one must be specified while the other
must be \code{NA}).

\strong{Example:} to set a tolerance of 10 \emph{units}, specify \verb{tolV = 10, tolP = NA}; to set a tolerance of 1\%,
specify \verb{tolV = NA, tolP = 0.01}.

\strong{Default values} are \code{tolV = 0.001} and \code{tolP = NA}.}

\item{warnNegResult}{(optional)

Logical argument specifying whether a warning message is generated when a negative value created by the
function in the benchmarked (output) series is smaller than the threshold specified by argument \code{tolN}.

\strong{Default value} is \code{warnNegResult = TRUE}.}

\item{tolN}{(optional)

Negative real number specifying the threshold for the identification of negative values. A value is
considered negative when it is smaller than this threshold.

\strong{Default value} is \code{tolN = -0.001}.}

\item{var}{(optional)

String vector (minimum length of 1) specifying the variable name(s) in the indicator series data frame
(argument \code{series_df}) containing the values and (optionally) the user-defined alterability coefficients
of the series to be benchmarked. These variables must be numeric.

The syntax is \code{var = c("series1 </ alt_ser1>", "series2 </ alt_ser2>", ...)}. Default alterability
coefficients of \eqn{1.0} are used when a user-defined alterability coefficients variable is not specified
alongside an indicator series variable. See section \strong{Details} for more information on alterability coefficients.

\strong{Example:} \code{var = "value / alter"} would benchmark indicator series data frame variable \code{value} with the
alterability coefficients contained in variable \code{alter} while \code{var = c("value / alter", "value2")} would
additionally benchmark variable \code{value2} with default alterability coefficients of \eqn{1.0}.

\strong{Default value} is \code{var = "value"} (benchmark variable \code{value} using default alterability coefficients of
\eqn{1.0}).}

\item{with}{(optional)

String vector (same length as argument \code{var}), or \code{NULL}, specifying the variable name(s) in the benchmarks data
frame (argument \code{benchmarks_df}) containing the values and (optionally) the user-defined alterability coefficients
of the benchmarks. These variables must be numeric. Specifying \code{with = NULL} results in using benchmark variable(s)
with the same names(s) as those specified with argument \code{var} without user-defined benchmark alterability coefficients
(i.e., default alterability coefficients of \eqn{0.0} corresponding to binding benchmarks).

The syntax is \code{with = NULL} or \code{with = c("bmk1 </ alt_bmk1>", "bmk2 </ alt_bmk2>", ...)}. Default alterability
coefficients of \eqn{0.0} (binding benchmarks) are used when a user-defined alterability coefficients variable
is not specified alongside a benchmark variable. See section \strong{Details} for more information on alterability
coefficients.

\strong{Example:} \code{with = "val_bmk"} would use benchmarks data frame variable \code{val_bmk} with default benchmark
alterability coefficients of \eqn{0.0} to benchmark the indicator series while
\code{with = c("val_bmk", "val_bmk2 / alt_bmk2")} would additionally benchmark a second indicator series using
benchmark variable \code{val_bmk2} with the benchmark alterability coefficients contained in variable \code{alt_bmk2}.

\strong{Default value} is \code{with = NULL} (same benchmark variable(s) as argument \code{var} using default benchmark
alterability coefficients of \eqn{0.0}).}

\item{by}{(optional)

String vector (minimum length of 1), or \code{NULL}, specifying the variable name(s) in the input data frames
(arguments \code{series_df} and \code{benchmarks_df}) to be used to form groups (for \emph{BY-group} processing) and allow
the benchmarking of multiple series in a single function call. BY-group variables can be numeric or
character (factors or not), must be present in both input data frames and will appear in all three output
data frames (see section \strong{Value}). BY-group processing is not  implemented when \code{by = NULL}. See
"Benchmarking Multiple Series" in section \strong{Details} for more information.

\strong{Default value} is \code{by = NULL} (no BY-group processing).}

\item{verbose}{(optional)

Logical argument specifying whether information on intermediate steps with execution time (real time,
not CPU time) should be displayed. Note that specifying argument \code{quiet = TRUE} would \emph{nullify} argument
\code{verbose}.

\strong{Default value} is \code{verbose = FALSE}.}

\item{constant}{(optional)

Real number that specifies a value to be temporarily added to both the indicator series and the
benchmarks before solving proportional benchmarking problems (\code{lambda != 0.0}). The temporary
constant is removed from the final output benchmarked series. E.g., specifying a (small) constant
would allow proportional benchmarking with \code{rho = 1} (e.g., proportional Denton benchmarking) on
indicator series that include values of 0. Otherwise, proportional benchmarking with values of 0 in
the indicator series is only possible when \code{rho < 1}. Specifying a constant with additive benchmarking
(\code{lambda = 0.0}) has no impact on the resulting benchmarked data. The data variables in the \code{graphTable}
output data frame include the constant, corresponding to the benchmarking problem that was actually solved.

\strong{Default value} is \code{constant = 0} (no temporary additive constant).}

\item{negInput_option}{(optional)

Handling of negative values in the input data for proportional benchmarking (\code{lambda != 0.0}):
\itemize{
\item \code{0}: Do not allow negative values with proportional benchmarking. An error message is displayed in the
presence of negative values in the input indicator series or benchmarks and missing (\code{NA}) values are
returned for the benchmarked series. This corresponds to the G-Series 2.0 behaviour.
\item \code{1}: Allow negative values with proportional benchmarking but display a warning message.
\item \code{2}: Allow negative values with proportional benchmarking without displaying any message.
}

\strong{Default value} is \code{negInput_option = 0} (do not allow negative values with proportional benchmarking).}

\item{allCols}{(optional)

Logical argument specifying whether all variables in the indicator series data frame (argument \code{series_df}),
other than \code{year} and \code{period}, determine the set of series to benchmark. Values specified with arguments \code{var}
and \code{with} are ignored when \code{allCols = TRUE}, which automatically implies default alterability coefficients,
and variables with the same names as the indicator series must exist in the benchmarks data frame (argument
\code{benchmarks_df}).

\strong{Default value} is \code{allCols = FALSE}.}

\item{quiet}{(optional)

Logical argument specifying whether or not to display only essential information such as warning messages,
error messages and variable (series) or BY-group information when multiple series are benchmarked in a single call
to the function. We advise against \emph{wrapping} your \code{\link[=benchmarking]{benchmarking()}} call with \code{\link[=suppressMessages]{suppressMessages()}} to further suppress
the display of variable (series) or BY-group information when processing multiple series as this would make
troubleshooting difficult in case of issues with individual series. Note that specifying \code{quiet = TRUE} would
also \emph{nullify} argument \code{verbose}.

\strong{Default value} is \code{quiet = FALSE}.}
}
\value{
The function returns is a list of three data frames:
\itemize{
\item \code{series}: data frame containing the benchmarked data (primary function output). BY variables specified
with argument \code{by} would be included in the data frame but not alterability coefficient variables
specified with argument \code{var}.
\item \code{benchmarks}: copy of the input benchmarks data frame (excluding invalid benchmarks when applicable).
BY variables specified with argument \code{by} would be included in the data frame but not alterability
coefficient variables specified with argument \code{with}.
\item \code{graphTable}: data frame containing supplementary data useful for producing analytical tables and graphs
(see function \code{\link[=plot_graphTable]{plot_graphTable()}}). It contains the following variables in addition to the BY variables
specified with argument \code{by}:
\itemize{
\item \code{varSeries}: Name of the indicator series variable
\item \code{varBenchmarks}: Name of the benchmark variable
\item \code{altSeries}: Name of the user-defined indicator series alterability coefficients variable
\item \code{altSeriesValue}: Indicator series alterability coefficients
\item \code{altbenchmarks}: Name of the user-defined benchmark alterability coefficients variable
\item \code{altBenchmarksValue}: Benchmark alterability coefficients
\item \code{t}: Indicator series period identifier (1 to \eqn{T})
\item \code{m}: Benchmark coverage periods identifier (1 to \eqn{M})
\item \code{year}: Data point calendar year
\item \code{period}: Data point period (cycle) value (1 to \code{periodicity})
\item \code{constant}: Temporary additive constant (argument \code{constant})
\item \code{rho}: Autoregressive parameter \eqn{\rho} (argument \code{rho})
\item \code{lambda}: Adjustment model parameter \eqn{\lambda} (argument \code{lambda})
\item \code{bias}: Bias adjustment (default, user-defined or estimated bias according to arguments \code{biasOption} and \code{bias})
\item \code{periodicity}: The maximum number of periods in a year (e.g. 4 for a quarterly indicator series)
\item \code{date}: Character string combining the values of variables \code{year} and \code{period}
\item \code{subAnnual}: Indicator series values
\item \code{benchmarked}: Benchmarked series values
\item \code{avgBenchmark}: Benchmark values divided by the number of coverage periods
\item \code{avgSubAnnual}: Indicator series values (variable \code{subAnnual}) averaged over the benchmark coverage period
\item \code{subAnnualCorrected}: Bias corrected indicator series values
\item \code{benchmarkedSubAnnualRatio}: Difference (\eqn{\lambda = 0}) or ratio (\eqn{\lambda \ne 0}{lambda != 0}) of the values
of variables \code{benchmarked} and \code{subAnnual}
\item \code{avgBenchmarkSubAnnualRatio}: Difference (\eqn{\lambda = 0}) or ratio (\eqn{\lambda \ne 0}{lambda != 0}) of the values
of variables \code{avgBenchmark} and \code{avgSubAnnual}
\item \code{growthRateSubAnnual}: Period to period difference (\eqn{\lambda = 0}) or relative difference (\eqn{\lambda \ne 0}{
  lambda != 0}) of the indicator series values (variable \code{subAnnual})
\item \code{growthRateBenchmarked}: Period to period difference (\eqn{\lambda = 0}) or relative difference (\eqn{\lambda \ne 0}{
  lambda != 0}) of the benchmarked series values (variable \code{benchmarked})
}
}

Notes:
\itemize{
\item The output \code{benchmarks} data frame always contains the original benchmarks provided in the input
benchmarks data frame. Modified nonbinding benchmarks, when applicable, can be recovered (calculated)
from the output \code{series} data frame.
\item The function returns a \code{NULL} object if an error occurs before data processing could start. Otherwise,
if execution gets far enough so that data processing could start, then an incomplete object would be
returned in case of errors (e.g., output \code{series} data frame with \code{NA} values for the benchmarked data).
\item The function returns "data.frame" objects that can be explicitly coerced to other types of objects with
the appropriate \verb{as*()} function (e.g., \code{tibble::as_tibble()} would coerce any of them to a tibble).
}
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/benchmarking.html}})}

\emph{Replication of the G-Series 2.0 SAS\eqn{^\circledR}{®} BENCHMARKING procedure (PROC BENCHMARKING).
See the G-Series 2.0 documentation for details (Statistics Canada 2016).}

This function ensures coherence between time series data of the same target variable measured at different
frequencies (e.g., sub-annually and annually). Benchmarking consists of imposing the level of the benchmark
series (e.g., annual data) while minimizing the revisions of the observed movement in the indicator series
(e.g.,  sub-annual data) as much as possible. The function also allows nonbinding benchmarking where the
benchmark series can also be revised.

The function may also be used for benchmarking-related topics such as \emph{temporal distribution}
(the reciprocal action of benchmarking: disaggregation of the benchmark series into more frequent observations),
\emph{calendarization} (a special case of temporal distribution) and \emph{linking} (the connection of different time series
segments into a single consistent time series).

Several series can be benchmarked in a single function call.
}
\details{
When \eqn{\rho < 1}, this function returns the generalized least squared solution of a special case of the general
regression-based benchmarking model proposed by Dagum and Cholette (2006). The model, in matrix form, is:
\deqn{\displaystyle
\begin{bmatrix} s^\dagger \\ a \end{bmatrix} = 
\begin{bmatrix} I \\ J \end{bmatrix} \theta + 
\begin{bmatrix} e \\ \varepsilon \end{bmatrix}
}{[s^dag; a] = [I; J] theta + [e; epsilion]}
where
\itemize{
\item \eqn{a} is the vector of length \eqn{M} of the benchmarks.
\item \eqn{s^\dagger = \left\{
    \begin{array}{cl}
      s + b & \text{if } \lambda = 0 \\
      s \cdot b  & \text{otherwise}
    \end{array} \right.
  }{s^dag = s + b if lambda = 0, s^dag = s * b otherwise} is the vector of length \eqn{T} of the bias corrected
indicator series values, with \eqn{s} denoting the initial (input) indicator series.
\item \eqn{b} is the bias, which is specified with argument \code{bias} when argument \code{bias_option != 3} or, when \code{bias_option = 3},
is estimated as \eqn{\hat{b} = \left\{
    \begin{array}{cl}
      \frac{{1_M}^\mathrm{T} (a - Js)}{{1_M}^\mathrm{T} J 1_T} & \text{if } \lambda = 0 \\
      \frac{{1_M}^\mathrm{T} a}{{1_M}^\mathrm{T} Js} & \text{otherwise}
    \end{array} \right.
  }{b^hat = ({1_M}'(a - J s)) / ({1_M}' J 1_T) if \lambda = 0, b^hat = ({1_M}' a) / ({1_M}' J s) otherwise}, where
\eqn{1_X = (1, ..., 1)^\mathrm{T}}{1_X = (1, ..., 1)'} is a vector of \eqn{1} of length \eqn{X}.
\item \eqn{J} is the \eqn{M \times T}{M x T} matrix of temporal aggregation constraints with elements \if{latex}{\cr}
\eqn{j_{m, t} = \left\{
    \begin{array}{cl}
      1 & \text{if benchmark } m \text{ covers period } t \\
      0 & \text{otherwise}
    \end{array} \right.
  }{j_{m,t} = 1 if benchmark m covers period t, j_{m,t} = 0 otherwise}.
\item \eqn{\theta} is the vector of the final (benchmarked) series values.
\item \eqn{e \sim \left( 0, V_e \right)}{e ~ (0, V_e)} is the vector of the measurement errors of \eqn{s^\dagger}{s^dag}
with covariance matrix \eqn{V_e = C \Omega_e C}.
\item \eqn{C = \mathrm{diag} \left( \sqrt{c_{s^\dagger}} \left| s^\dagger \right|^\lambda \right)}{C = diag(\sqrt{c_{s^dag}} 
|s^dag|^lambda)} where \eqn{c_{s^\dagger}}{c_{s^dag}} is the vector of the alterability coefficients of \eqn{s^\dagger}{s^dag},
assuming \eqn{0^0 = 1}.
\item \eqn{\Omega_e} is a \eqn{T \times T}{T x T} matrix with elements \eqn{\omega_{e_{i,j}} = \rho^{|i-j|}} representing the
autocorrelation of an AR(1) process, again assuming \eqn{0^0 = 1}.
\item \eqn{\varepsilon \sim (0, V_\varepsilon)}{epsilon ~ (0, V_epsilon)} is the vector of the measurement errors of the benchmarks
\eqn{a} with covariance matrix \eqn{V_\varepsilon = \mathrm{diag} \left( c_a a \right)}{V_epsilion = diag(c_a a)} where \eqn{c_a}
is the vector of the alterability coefficients of the benchmarks \eqn{a}.
}

The generalized least squared solution is:
\deqn{\displaystyle 
\hat{\theta} = s^\dagger + V_e J^{\mathrm{T}} \left( J V_e J^{\mathrm{T}} + V_\varepsilon \right)^+ \left( a - J s^\dagger \right)
}{theta^hat = s^dag + V_e J' (J V_e J' + V_epsilion)^{+} (a - J s^dag)}
where \eqn{A^{+}} designates the Moore-Penrose inverse of matrix \eqn{A}.

When \eqn{\rho = 1}, the function returns the solution of the (modified) Denton method:
\deqn{\displaystyle 
\hat{\theta} = s + W \left( a - J s \right)
}{theta^hat = s + W (a - J s)}
where
\itemize{
\item \eqn{W} is the upper-right corner matrix from the following matrix product
\deqn{
    \left[\begin{array}{cc}
      D^{+} \Delta^{\mathrm{T}} \Delta D^{+} & J^{\mathrm{T}} \\
      J & 0
    \end{array} \right]^{+}
    \left[\begin{array}{cc}
      D^{+} \Delta^{\mathrm{T}} \Delta D^{+} & 0 \\
      J & I_M
    \end{array} \right] = 
    \left[\begin{array}{cc}
      I_T & W \\
      0 & W_\nu
    \end{array} \right]
  }{[D^{+} Delta' Delta D^{+}, J'; J, 0]^{+} [D^{+} Delta' Delta D^{+}, 0; J, I_M] = [I_T, W; 0, W_nu]}
\item \eqn{D = \mathrm{diag} \left( \left| s \right|^\lambda \right)}{D = diag(|s|^\lambda)}, assuming \eqn{0^0 = 1}. Note that
\eqn{D} corresponds to \eqn{C} with \eqn{c_{s^\dagger} = 1.0}{c_{s^dag} = 1.0} and without bias correction (arguments
\code{bias_option = 1} and \code{bias = NA}).
\item \eqn{\Delta}{Delta} is a \eqn{T-1 \times T}{T-1 x T} matrix with elements \eqn{\delta_{i,j} = \left\{
    \begin{array}{cl}
      -1 & \text{if } i=j \\
      1 & \text{if } j=i+1 \\
      0 & \text{otherwise}
    \end{array} \right.
  }{delta_{i,j} = 1 if i = j, delta_{i,j} = 1 if j = i + 1, delta_{i,j} = 0 otherwise}.
\item \eqn{W_\nu} is a \eqn{M \times M}{M x M} matrix associated with the Lagrange multipliers of the corresponding
minimization problem, expressed as:
\deqn{\displaystyle 
\begin{aligned}
& \underset{\theta}{\text{minimize}} 
& & \sum_{t \ge 2} \left[ \frac{\left( s_t - \theta_t \right)}{\left| s_t\right|^\lambda}
      - \frac{\left( s_{t-1} - \theta_{t-1} \right)}{\left| s_{t-1}\right|^\lambda} \right]^2 \\
& \text{subject to} 
& & a = J \theta
\end{aligned}
}{min(theta) sum_{t>1}{[(s_t - theta_t) / |s_t|^lambda - (s_{t-1} - theta_{t-1}) / |s_{t-1}|^lambda]^2}, subject to a = J theta}
}

See Quenneville et al. (2006) and Dagum and Cholette (2006) for details.
\subsection{Autoregressive Parameter \eqn{\rho} and \emph{bias}}{

Parameter \eqn{\rho} (argument \code{rho}) is associated to the change between the (input) indicator and the (output)
benchmarked series for two consecutive periods and is often called the \emph{movement preservation parameter}. The larger
the value of \eqn{\rho}, the more the indicator series period to period movements are preserved in the
benchmarked series. With \eqn{\rho = 0}, period to period movement preservation is not enforced and the
resulting benchmarking adjustments are not smooth, as in the case of prorating (\eqn{\rho = 0} and
\eqn{\lambda = 0.5}) where the adjustments take the shape of a \emph{step function}. At the other end of the
spectrum is \eqn{\rho = 1}, referred to as \emph{Denton benchmarking}, where period to period movement preservation
is maximized, which results in the smoothest possible set of benchmarking adjustments available with the function.

The \emph{bias} represents the expected discrepancies between the benchmarks and the indicator series. It can be
used to pre-adjust the indicator series in order to reduce, on average, the discrepancies between the two
sources of data. Bias correction, which is specified with arguments \code{biasOption} and \code{bias}, can be particularly
useful for periods not covered by benchmarks when \eqn{\rho < 1}. In this context, parameter \eqn{\rho} dictates the
speed at which the projected benchmarking adjustments converge to the bias (or converge to \emph{no adjustment} without bias
correction) for periods not covered by a benchmark. The smaller the value of \eqn{\rho}, the faster the convergence to
the bias, with immediate convergence when \eqn{\rho = 0} and no convergence at all (the adjustment of the last period
covered by a benchmark is repeated) when \eqn{\rho = 1} (Denton benchmarking). Arguments \code{biasOption} and \code{bias} are
actually ignored when \eqn{\rho = 1} since correcting for the bias has no impact on Denton benchmarking solutions.
The suggested value for \eqn{\rho} is \eqn{0.9} for monthly indicators and \eqn{0.9^3 = 0.729} for quarterly indicators,
representing a reasonable compromise between maximizing movement preservation and reducing revisions as new benchmarks
become available in the future (benchmarking \emph{timeliness issue}). In practice, note that Denton benchmarking could be
\emph{approximated} with the regression-based model by using a \eqn{\rho} value that is smaller than, but very close to,
\eqn{1.0} (e.g., \eqn{\rho = 0.999}). See Dagum and Cholette (2006) for a complete discussion on this topic.
}

\subsection{Alterability Coefficients}{

Alterability coefficients \eqn{c_{s^\dagger}}{c_{s^dag}} and \eqn{c_a} conceptually represent the measurement errors
associated with the (bias corrected) indicator time series values \eqn{s^\dagger}{s^dag} and benchmarks \eqn{a}
respectively. They are nonnegative real numbers which, in practice, specify the extent to which an initial value can be
modified in relation to other values. Alterability coefficients of \eqn{0.0} define fixed (binding) values while
alterability coefficients greater than \eqn{0.0} define free (nonbinding) values. Increasing the alterability coefficient
of an intial value results in more changes for that value in the benchmarking solution and, conversely, less changes when
decreasing the alterability coefficient. The default alterability coefficients are \eqn{0.0} for the benchmarks (binding
benchmarks) and \eqn{1.0} for the indicator series values (nonbinding indicator series). Important notes:
\itemize{
\item With a value of \eqn{\rho = 1} (argument \code{rho = 1}, associated to Denton Benchmarking), only the default
alterability coefficients (\eqn{0.0} for a benchmark and \eqn{1.0} for a an indicator series value) are valid. The
specification of user-defined alterability coefficients variables is therefore not allowed. If such variables are
specified (see arguments \code{var} and \code{with}), the function ignores them and displays a warning message in the console.
\item Alterability coefficients \eqn{c_{s^\dagger}}{c_{s^dag}} come into play after the indicator series has been corrected for
the bias, when applicable (\eqn{c_{s^\dagger}}{c_{s^dag}} is associated to \eqn{s^\dagger}{s^dag}, not \eqn{s}). This means
that specifying an alterability coefficient of \eqn{0.0} for a given indicator series value \strong{will not} result in an unchanged
value after benchmarking \strong{with bias correction} (see arguments \code{biasOption} and \code{bias}).
}

Nonbinding benchmarks, when applicable, can be recovered (calculated) from the benchmarked series (see output data frame
\code{series} in section \strong{Value}). The output \code{benchmarks} data frame always contains the original benchmarks provided
in the input benchmarks data frame (argument \code{benchmarks_df}).
}

\subsection{Benchmarking Multiple Series}{

Multiple series can be benchmarked in a single \code{\link[=benchmarking]{benchmarking()}} call, by specifying \code{allCols = TRUE}, by
(manually) specifying multiple variables with argument \code{var} (and argument \code{with}) or with BY-group processing
(argument \code{by != NULL}). An important distinction is that all indicator series specified with \code{allCols = TRUE}
or with argument \code{var} (and benchmarks with argument \code{with}) are expected to be of the same length, i.e., same
set of periods and same set (number) of benchmarks. Benchmarking series of different lengths (different sets of
periods) or with different sets (number) of benchmarks must be done with BY-group processing on stacked indicator
series and benchmarks input data frames (see utility functions \code{\link[=stack_tsDF]{stack_tsDF()}} and \code{\link[=stack_bmkDF]{stack_bmkDF()}}). Arguments
\code{by} and \code{var} can be combined in order to implement BY-group processing for multiple series as illustrated by
\emph{Example 2} in the \strong{Examples} section. While multiple variables with argument \code{var} (or \code{allCols = TRUE})
without BY-group processing (argument \code{by = NULL}) is slightly more efficient (faster), a BY-group approach with
a single series variable is usually recommended as it is more general (works in all contexts). The latter is
illustrated by \emph{Example 3} in the \strong{Examples} section. The BY variables specified with argument \code{by} appear in
all three output data frames.
}

\subsection{Arguments \code{constant} and \code{negInput_option}}{

These arguments extend the usage of proportional benchmarking to a larger set of problems. Their default
values correspond to the G-Series 2.0 behaviour (SAS\eqn{^\circledR}{®} PROC BENCHMARKING) for which equivalent
options are not defined. Although proportional benchmarking may not necessarily be the most appropriate approach
(additive benchmarking may be more appropriate) when the values of the indicator series approach 0 (unstable
period-to-period ratios) or "cross the 0 line" and can therefore go from positive to negative and vice versa
(confusing, difficult to interpret period-to-period ratios), these cases are not invalid mathematically
speaking (i.e., the associated proportional benchmarking problem can be solved). It is strongly recommended,
however, to carefully analyze and validate the resulting benchmarked data in these situations and make sure they
correspond to reasonable, interpretable solutions.
}

\subsection{Treatment of Missing (\code{NA}) Values}{
\itemize{
\item If a missing value appears in one of the variables of the benchmarks input data frame (other than the
BY variables), the observations with the missing values are dropped, a warning message is displayed and the
function executes.
\item If a missing value appears in the \code{year} and/or \code{period} variables of the indicator series input data frame
and BY variables are specified, the corresponding BY-group is skipped, a warning message is displayed and
the function moves on to the next BY-group. If no BY variables are specified, a warning message is displayed
and no processing is done.
\item If a missing value appears in one of the indicator series variables in the indicator series input data
frame and BY variables are specified, the corresponding BY-group is skipped, a warning message is displayed
and the function moves on to the next BY-group. If no BY variables are specified, the affected indicator
series is not processed, a warning message is displayed and the function moves on to the next indicator
series (when applicable).
}
}
}
\examples{
# Set the working directory (for the PDF files)
iniwd <- setwd(tempdir())


###########
# Example 1: Simple case with a single quarterly series to benchmark to annual values

# Quarterly indicator series
my_series1 <- ts_to_tsDF(
  ts(
    c(1.9, 2.4, 3.1, 2.2, 2.0, 2.6, 3.4, 2.4, 2.3),
    start = c(2015, 1),
    frequency = 4
  )
)
my_series1

# Annual benchmarks for quarterly data
my_benchmarks1 <- ts_to_bmkDF(
  ts(
    c(10.3, 10.2),
    start = 2015,
    frequency = 1
  ),
  ind_frequency = 4
)
my_benchmarks1

# Benchmarking using...
#   - recommended `rho` value for quarterly series (`rho = 0.729`)
#   - proportional model (`lambda = 1`)
#   - bias-corrected indicator series with the estimated bias (`biasOption = 3`)
out_bench1 <- benchmarking(
  my_series1,
  my_benchmarks1,
  rho = 0.729,
  lambda = 0,
  biasOption = 3
)

# Generate the benchmarking graphs
\donttest{plot_graphTable(out_bench1$graphTable, "Ex1_graphs.pdf")}


###########
# Example 2: Two quarterly series to benchmark to annual values,
#            with BY-groups and user-defined alterability coefficients

# Sales data (same sales for groups A and B; only alter coefs for van sales differ)
qtr_sales <- ts(
  matrix(
    c(
      # Car sales
      1851, 2436, 3115, 2205, 1987, 2635, 3435, 2361, 2183, 2822,
      3664, 2550, 2342, 3001, 3779, 2538, 2363, 3090, 3807, 2631,
      2601, 3063, 3961, 2774, 2476, 3083, 3864, 2773, 2489, 3082,
      # Van sales
      1900, 2200, 3000, 2000, 1900, 2500, 3800, 2500, 2100, 3100,
      3650, 2950, 3300, 4000, 3290, 2600, 2010, 3600, 3500, 2100,
      2050, 3500, 4290, 2800, 2770, 3080, 3100, 2800, 3100, 2860
    ),
    ncol = 2
  ),
  start = c(2011, 1),
  frequency = 4,
  names = c("car_sales", "van_sales")
)

ann_sales <- ts(
  matrix(
    c(
      # Car sales
       10324, 10200, 10582, 11097, 11582, 11092,
       # Van sales
       12000, 10400, 11550, 11400, 14500, 16000),
     ncol = 2
    ),
  start = 2011,
  frequency = 1,
  names = c("car_sales", "van_sales")
)

# Quarterly indicator series (with default alter coefs for now)
my_series2 <- rbind(
  cbind(
    data.frame(
      group = rep("A", nrow(qtr_sales)),
      alt_van = rep(1, nrow(qtr_sales))
    ),
    ts_to_tsDF(qtr_sales)
  ),
  cbind(
    data.frame(
      group = rep("B", nrow(qtr_sales)),
      alt_van = rep(1, nrow(qtr_sales))
    ),
    ts_to_tsDF(qtr_sales)
  )
)

# Set binding van sales (alter coef = 0) for 2012 Q1 and Q2 in group A (rows 5 and 6)
my_series2$alt_van[c(5,6)] <- 0
head(my_series2, n = 10)
tail(my_series2)


# Annual benchmarks for quarterly data (without alter coefs)
my_benchmarks2 <- rbind(
  cbind(
    data.frame(group = rep("A", nrow(ann_sales))),
    ts_to_bmkDF(ann_sales, ind_frequency = 4)
  ),
  cbind(
    data.frame(group = rep("B", nrow(ann_sales))),
    ts_to_bmkDF(ann_sales, ind_frequency = 4))
)
my_benchmarks2

# Benchmarking using...
#   - recommended `rho` value for quarterly series (`rho = 0.729`)
#   - proportional model (`lambda = 1`)
#   - without bias correction (`biasOption = 1` and `bias` not specified)
#   - `quiet = TRUE` to avoid generating the function header
out_bench2 <- benchmarking(
  my_series2,
  my_benchmarks2,
  rho = 0.729,
  lambda = 1,
  biasOption = 1,
  var = c("car_sales", "van_sales / alt_van"),
  with = c("car_sales", "van_sales"),
  by = "group",
  quiet = TRUE
)

# Generate the benchmarking graphs
\donttest{plot_graphTable(out_bench2$graphTable, "Ex2_graphs.pdf")}

# Check the value of van sales for 2012 Q1 and Q2 in group A (fixed values)
all.equal(my_series2$van_sales[c(5,6)], out_bench2$series$van_sales[c(5,6)])


###########
# Example 3: same as example 2, but benchmarking all 4 series as BY-groups
#            (4 BY-groups of 1 series instead of 2 BY-groups of 2 series)

qtr_sales2 <- ts.union(A = qtr_sales, B = qtr_sales)
my_series3 <- stack_tsDF(ts_to_tsDF(qtr_sales2))
my_series3$alter <- 1
my_series3$alter[
  my_series3$series == "A.van_sales" & 
    my_series3$year == 2012 & 
    my_series3$period <= 2
] <- 0
head(my_series3)
tail(my_series3)


ann_sales2 <- ts.union(A = ann_sales, B = ann_sales)
my_benchmarks3 <- stack_bmkDF(ts_to_bmkDF(ann_sales2, ind_frequency = 4))
head(my_benchmarks3)
tail(my_benchmarks3)

out_bench3 <- benchmarking(
  my_series3,
  my_benchmarks3,
  rho = 0.729,
  lambda = 1,
  biasOption = 1,
  var = "value / alter",
  with = "value",
  by = "series",
  quiet = TRUE
)

# Generate the benchmarking graphs
\donttest{plot_graphTable(out_bench3$graphTable, "Ex3_graphs.pdf")}

# Convert data frame `out_bench3$series` to a "mts" object
qtr_sales2_bmked <- tsDF_to_ts(unstack_tsDF(out_bench3$series), frequency = 4)

# Print the first 10 observations
ts(qtr_sales2_bmked[1:10, ], start = start(qtr_sales2), deltat = deltat(qtr_sales2))

# Check the value of van sales for 2012 Q1 and Q2 in group A (fixed values)
all.equal(
  window(qtr_sales2[, "A.van_sales"], start = c(2012, 1), end = c(2012, 2)),
  window(qtr_sales2_bmked[, "A.van_sales"], start = c(2012, 1), end = c(2012, 2))
)


# Reset the working directory to its initial location
setwd(iniwd)
}
\references{
Dagum, E. B. and P. Cholette (2006). \strong{Benchmarking, Temporal Distribution and Reconciliation Methods
of Time Series}. Springer-Verlag, New York, Lecture Notes in Statistics, Vol. 186. \doi{10.1007/0-387-35439-5}

Fortier, S. and B. Quenneville (2007). "Theory and Application of Benchmarking in Business Surveys".
\strong{Proceedings of the Third International Conference on Establishment Surveys (ICES-III)}. Montréal, June 2007.

Latendresse, E., M. Djona and S. Fortier (2007). "Benchmarking Sub-Annual Series to Annual Totals –
From Concepts to SAS\eqn{^\circledR}{®} Procedure and Enterprise Guide\eqn{^\circledR}{®} Custom Task". \strong{Proceedings
of the SAS\eqn{^\circledR}{®} Global Forum 2007 Conference}. Cary, NC: SAS Institute Inc.

Quenneville, B., S. Fortier, Z.-G. Chen and E. Latendresse (2006). "Recent Developments in Benchmarking to
Annual Totals in X-12-ARIMA and at Statistics Canada". \strong{Proceedings of the Eurostat Conference on Seasonality,
Seasonal Adjustment and Their Implications for Short-Term Analysis and Forecasting}. Luxembourg, May 2006.

Quenneville, B., P. Cholette, S. Fortier and J. Bérubé (2010). "Benchmarking Sub-Annual Indicator
Series to Annual Control Totals (Forillon v1.04.001)". \strong{Internal document}. Statistics Canada, Ottawa, Canada.

Quenneville, B. and S. Fortier (2012). "Restoring Accounting Constraints in Time Series – Methods and
Software for a Statistical Agency". \strong{Economic Time Series: Modeling and Seasonality}. Chapman & Hall, New York.

Statistics Canada (2012). \strong{Theory and Application of Benchmarking (Course code 0436)}.
Statistics Canada, Ottawa, Canada.

Statistics Canada (2016). "The BENCHMARKING Procedure". \strong{G-Series 2.0 User Guide}.
Statistics Canada, Ottawa, Canada.
}
\seealso{
\code{\link[=stock_benchmarking]{stock_benchmarking()}} \code{\link[=plot_graphTable]{plot_graphTable()}} \link{bench_graphs} \code{\link[=plot_benchAdj]{plot_benchAdj()}} \code{\link[=gs.gInv_MP]{gs.gInv_MP()}} \link{aliases}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ts_to_bmkDF.R
\name{ts_to_bmkDF}
\alias{ts_to_bmkDF}
\title{Convert a "ts" object to a benchmarks data frame}
\usage{
ts_to_bmkDF(
  in_ts,
  ind_frequency,
  discrete_flag = FALSE,
  alignment = "b",
  bmk_interval_start = 1,
  startYr_cName = "startYear",
  startPer_cName = "startPeriod",
  endYr_cName = "endYear",
  endPer_cName = "endPeriod",
  val_cName = "value"
)
}
\arguments{
\item{in_ts}{(mandatory)

Time series (object of class "ts" or "mts") to be converted.}

\item{ind_frequency}{(mandatory)

Integer specifying the frequency of the indicator (high frequency) series for which the benchmarks (low
frequency series) are related to. The frequency of a time series corresponds to the maximum number of periods
in a year (e.g., 12 for a monthly data, 4 for a quarterly data, 1 for annual data).}

\item{discrete_flag}{(optional)

Logical argument specifying whether the benchmarks correspond to discrete values (anchor points covering a single
period of the indicator series, e.g., end of year stocks) or not. \code{discrete_flag = FALSE} defines non-discrete
benchmarks, i.e., benchmarks that cover several periods of the indicator series (e.g. annual benchmarks cover
4 quarters or 12 months, quarterly benchmarks cover 3 months, etc.).

\strong{Default value} is \code{discrete_flag = FALSE}.}

\item{alignment}{(optional)

Character identifying the alignment of discrete benchmarks (argument \code{discrete_flag = TRUE}) in the benchmark
(low frequency series) interval coverage window:
\itemize{
\item \code{alignment = "b"}: beginning of the benchmark interval window (first period)
\item \code{alignment = "e"}: end of the benchmark interval window (last period)
\item \code{alignment = "m"}: middle of the benchmark interval window (middle period)
}

This argument has no effect for non-discrete benchmarks (\code{discrete_flag = FALSE}).

\strong{Default value} is \code{alignment = "b"}.}

\item{bmk_interval_start}{(optional)

Integer in the [1 .. \code{ind_frequency}] interval specifying the period (cycle) of the indicator (high frequency)
series at which the benchmark (low frequency series) interval window starts. E.g., annual benchmarks corresponding to
fiscal years defined from April to March of the following year would be specified with \code{bmk_interval_start = 4}
for a monthly indicator series (\code{ind_frequency = 12}) and \code{bmk_interval_start = 2} for a quarterly indicator
series (\code{ind_frequency = 4}).

\strong{Default value} is \code{bmk_interval_start = 1}.}

\item{startYr_cName, startPer_cName, endYr_cName, endPer_cName}{(optional)

Strings specifying the name of the numeric variables (columns) in the output data frame that will define the
benchmarks coverage, i.e., the starting and ending year and period (cycle) identifiers.

\strong{Default values} are \code{startYr_cName = "startYear"}, \code{startPer_cName = "startPeriod"}
\code{endYr_cName = "endYear"} and \code{endPer_cName   = "endPeriod"}.}

\item{val_cName}{(optional)

String specifying the name of the numeric variable (column) in the output data frame that will contain the
benchmark values. This argument has no effect for "mts" objects (benchmark variable names are automatically
inherited from the "mts" object).

\strong{Default value} is \code{val_cName = "value"}.}
}
\value{
The function returns a data frame with five or more variables:
\itemize{
\item Benchmark coverage starting year, type numeric (see argument \code{startYr_cName})
\item Benchmark coverage starting period (cycle), type numeric (see argument \code{startPer_cName})
\item Benchmark coverage ending year, type numeric (see argument \code{endtYr_cName})
\item Benchmark coverage ending period (cycle), type numeric (see argument \code{endPer_cName})
\item One ("ts" object) or many ("mts" object) benchmark data variable(s), type numeric (see argument \code{val_cName})
}

Note: the function returns a "data.frame" object than can be explicitly coerced to another type of object
with the appropriate \verb{as*()} function (e.g., \code{tibble::as_tibble()} would coerce it to a tibble).
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/ts_to_bmkDF.html}})}

Convert a "ts" (or "mts") object into a benchmarks data frame for the benchmarking functions with five or more variables
(columns):
\itemize{
\item four (4) for the benchmark coverage
\item one (1) for each benchmark time series
}

For discrete benchmarks (anchor points covering a single period of the indicator series, e.g., end of year
stocks), specify \code{discrete_flag = TRUE} and \code{alignment = "b"}, \code{"e"} or \code{"m"}.
}
\examples{
# Annual and quarterly time series
my_ann_ts <- ts(1:5 * 100, start = 2019, frequency = 1)
my_ann_ts
my_qtr_ts <- ts(my_ann_ts, frequency = 4)
my_qtr_ts


# Annual benchmarks for a monthly indicator series
ts_to_bmkDF(my_ann_ts, ind_frequency = 12)

# Annual benchmarks for a quarterly indicator series
ts_to_bmkDF(my_ann_ts, ind_frequency = 4)

# Quarterly benchmarks for a monthly indicator series
ts_to_bmkDF(my_qtr_ts, ind_frequency = 12)

# Start of year stocks for a quarterly indicator series
ts_to_bmkDF(
  my_ann_ts, ind_frequency = 4, 
  discrete_flag = TRUE
)

# End of quarter stocks for a monthly indicator series
ts_to_bmkDF(
  my_qtr_ts, ind_frequency = 12, 
  discrete_flag = TRUE, alignment = "e"
)

# April to March annual benchmarks for a ...
# ... monthly indicator series
ts_to_bmkDF(
  my_ann_ts, ind_frequency = 12, 
  bmk_interval_start = 4
)
# ... quarterly indicator series
ts_to_bmkDF(
  my_ann_ts, ind_frequency = 4, 
  bmk_interval_start = 2
)

# End-of-year (April to March) stocks for a ...
# ... monthly indicator series
ts_to_bmkDF(
  my_ann_ts, ind_frequency = 12, 
  discrete_flag = TRUE, alignment = "e", bmk_interval_start = 4
)
# ... quarterly indicator series
ts_to_bmkDF(
  my_ann_ts, ind_frequency = 4,
  discrete_flag = TRUE, alignment = "e", bmk_interval_start = 2
)

# Custom name for the benchmark data variable (column)
ts_to_bmkDF(
  my_ann_ts, ind_frequency = 12,
  val_cName = "bmk_val"
)

# Multiple time series: argument `val_cName` ignored
# (the "mts" object column names are always used)
ts_to_bmkDF(
  ts.union(ser1 = my_ann_ts, ser2 = my_ann_ts / 10), ind_frequency = 12,
  val_cName = "useless_column_name"
)
}
\seealso{
\code{\link[=ts_to_tsDF]{ts_to_tsDF()}} \code{\link[=stack_bmkDF]{stack_bmkDF()}} \code{\link[=benchmarking]{benchmarking()}} \code{\link[=stock_benchmarking]{stock_benchmarking()}} \link{time_values_conv}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsraking_driver.R
\name{tsraking_driver}
\alias{tsraking_driver}
\title{Helper function for \code{\link[=tsraking]{tsraking()}}}
\usage{
tsraking_driver(
  in_ts,
  ...,  # `tsraking()` arguments excluding `data_df`
  temporal_grp_periodicity = 1,
  temporal_grp_start = 1
)
}
\arguments{
\item{in_ts}{(mandatory)

Time series (object of class "ts" or "mts") that contains the time series data to be reconciled.
They are the raking problems' input data (initial solutions).}

\item{...}{
  Arguments passed on to \code{\link{tsraking}}
  \describe{
    \item{\code{metadata_df}}{(mandatory)

Data frame (object of class "data.frame") that describes the cross-sectional aggregation constraints
(additivity rules) for the raking problem. Two character variables must be included in the metadata data frame:
\code{series} and \code{total1}. Two variables are optional: \code{total2} (character) and \code{alterAnnual} (numeric). The values
of variable \code{series} represent the variable names of the component series in the input time series data frame
(argument \code{data_df}). Similarly, the values of variables \code{total1} and \code{total2} represent the variable names of
the 1\if{html}{\out{<sup>}}st\if{html}{\out{</sup>}} and 2\if{html}{\out{<sup>}}nd\if{html}{\out{</sup>}} dimension cross-sectional control totals in the input time series data
frame. Variable \code{alterAnnual} contains the alterability coefficient for the temporal constraint associated to
each component series. When specified, the latter will override the default alterability coefficient specified
with argument \code{alterAnnual}.}
    \item{\code{alterability_df}}{(optional)

Data frame (object of class "data.frame"), or \code{NULL}, that contains the alterability coefficients variables.
They must correspond to a component series or a cross-sectional control total, that is, a variable with the same
name must exist in the input time series data frame (argument \code{data_df}). The values of these alterability
coefficients will override the default alterability coefficients specified with arguments \code{alterSeries},
\code{alterTotal1} and \code{alterTotal2}. When the input time series data frame contains several observations and the
alterability coefficients data frame contains only one, the alterability coefficients are used (repeated) for
all observations of the input time series data frame. Alternatively, the alterability coefficients data frame
may contain as many observations as the input time series data frame.

\strong{Default value} is \code{alterability_df = NULL} (default alterability coefficients).}
    \item{\code{alterSeries}}{(optional)

Nonnegative real number specifying the default alterability coefficient for the component series values. It
will apply to component series for which alterability coefficients have not already been specified in the
alterability coefficients data frame (argument \code{alterability_df}).

\strong{Default value} is \code{alterSeries = 1.0} (nonbinding component series values).}
    \item{\code{alterTotal1}}{(optional)

Nonnegative real number specifying the default alterability coefficient for the 1\if{html}{\out{<sup>}}st\if{html}{\out{</sup>}} dimension
cross-sectional control totals. It will apply to cross-sectional control totals for which alterability
coefficients have not already been specified in the alterability coefficients data frame (argument
\code{alterability_df}).

\strong{Default value} is \code{alterTotal1 = 0.0} (binding 1\if{html}{\out{<sup>}}st\if{html}{\out{</sup>}} dimension cross-sectional control totals)}
    \item{\code{alterTotal2}}{(optional)

Nonnegative real number specifying the default alterability coefficient for the 2\if{html}{\out{<sup>}}nd\if{html}{\out{</sup>}} dimension
cross-sectional control totals. It will apply to cross-sectional control totals for which alterability
coefficients have not already been specified in the alterability coefficients data frame (argument
\code{alterability_df}).

\strong{Default value} is \code{alterTotal2 = 0.0} (binding 2\if{html}{\out{<sup>}}nd\if{html}{\out{</sup>}} dimension cross-sectional control totals).}
    \item{\code{alterAnnual}}{(optional)

Nonnegative real number specifying the default alterability coefficient for the component series temporal
constraints (e.g., annual totals). It will apply to component series for which alterability coefficients
have not already been specified in the metadata data frame (argument \code{metadata_df}).

\strong{Default value} is \code{alterAnnual = 0.0} (binding temporal control totals).}
    \item{\code{tolV,tolP}}{(optional)

Nonnegative real number, or \code{NA}, specifying the tolerance, in absolute value or percentage, to be used
when performing the ultimate test in the case of binding totals (alterability coefficient of \eqn{0.0}
for temporal or cross-sectional control totals). The test compares the input binding control totals with
the ones calculated from the reconciled (output) component series. Arguments \code{tolV} and \code{tolP} cannot be both
specified together (one must be specified while the other must be \code{NA}).

\strong{Example:} to set a tolerance of 10 \emph{units}, specify \verb{tolV = 10, tolP = NA}; to set a tolerance of 1\%,
specify \verb{tolV = NA, tolP = 0.01}.

\strong{Default values} are \code{tolV = 0.001} and \code{tolP = NA}.}
    \item{\code{warnNegResult}}{(optional)

Logical argument specifying whether a warning message is generated when a negative value created by the
function in the reconciled (output) series is smaller than the threshold specified by argument \code{tolN}.

\strong{Default value} is \code{warnNegResult = TRUE}.}
    \item{\code{tolN}}{(optional)

Negative real number specifying the threshold for the identification of negative values. A value is
considered negative when it is smaller than this threshold.

\strong{Default value} is \code{tolN = -0.001}.}
    \item{\code{id}}{(optional)

String vector (minimum length of 1), or \code{NULL}, specifying the name of additional variables to be transferred
from the input time series data frame (argument \code{data_df}) to the output time series data frame, the
object returned by the function (see section \strong{Value}). By default, the output series data frame only contains
the variables listed in the metadata data frame (argument \code{metadata_df}).

\strong{Default value} is \code{id = NULL}.}
    \item{\code{verbose}}{(optional)

Logical argument specifying whether information on intermediate steps with execution time (real time,
not CPU time) should be displayed. Note that specifying argument \code{quiet = TRUE} would \emph{nullify} argument
\code{verbose}.

\strong{Default value} is \code{verbose = FALSE}.}
    \item{\code{Vmat_option}}{(optional)

Specification of the option for the variance matrices (\eqn{V_e} and \eqn{V_\epsilon}; see section \strong{Details}):\tabular{cl}{
   \strong{Value} \tab \strong{Description} \cr
   \code{1} \tab Use vectors \eqn{x} and \eqn{g} in the variance matrices. \cr
   \code{2} \tab Use vectors \eqn{|x|} and \eqn{|g|} in the variance matrices. \cr
}


See Ferland (2016) and subsection \strong{Arguments \code{Vmat_option} and \code{warnNegInput}} in section \strong{Details} for
more information.

\strong{Default value} is \code{Vmat_option = 1}.}
    \item{\code{warnNegInput}}{(optional)

Logical argument specifying whether a warning message is generated when a negative value smaller than
the threshold specified by argument \code{tolN} is found in the input time series data frame (argument \code{data_df}).

\strong{Default value} is \code{warnNegInput = TRUE}.}
    \item{\code{quiet}}{(optional)

Logical argument specifying whether or not to display only essential information such as warnings and errors.
Specifying \code{quiet = TRUE} would also \emph{nullify} argument \code{verbose} and is equivalent to \emph{wrapping} your
\code{\link[=tsraking]{tsraking()}} call with \code{\link[=suppressMessages]{suppressMessages()}}.

\strong{Default value} is \code{quiet = FALSE}.}
  }}

\item{temporal_grp_periodicity}{(optional)

Positive integer defining the number of periods in temporal groups for which the totals should be preserved.
E.g., specify \code{temporal_grp_periodicity = 3} with a monthly time series for quarterly total preservation and
\code{temporal_grp_periodicity = 12} (or \code{temporal_grp_periodicity = frequency(in_ts)}) for annual total preservation.
Specifying \code{temporal_grp_periodicity = 1} (\emph{default}) corresponds to period-by-period processing without
temporal total preservation.

\strong{Default value} is \code{temporal_grp_periodicity = 1} (period-by-period processing without temporal total preservation).}

\item{temporal_grp_start}{(optional)

Integer in the [1 .. \code{temporal_grp_periodicity}] interval specifying the starting period (cycle) for temporal
total preservation. E.g., annual totals corresponding to fiscal years defined from April to March of the
following year would be specified with \code{temporal_grp_start = 4} for a monthly time series (\code{frequency(in_ts) = 12})
and \code{temporal_grp_start = 2} for a quarterly time series (\code{frequency(in_ts) = 4}). This argument has no effect
for period-by-period processing without temporal total preservation (\code{temporal_grp_periodicity = 1}).

\strong{Default value} is \code{temporal_grp_start = 1}.}
}
\value{
The function returns a time series object (class "ts" or "mts") containing the reconciled component series,
reconciled cross-sectional control totals and other series specified with \code{\link[=tsraking]{tsraking()}} argument \code{id}. It can be
explicitly coerced to another type of object with the appropriate \verb{as*()} function (e.g., \code{tsibble::as_tsibble()}
would coerce it to a tsibble).

Note that a \code{NULL} object is returned if an error occurs before data processing could start. Otherwise, if execution
gets far enough so that data processing could start, then an incomplete object (with \code{NA} values) would be returned
in case of errors.
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/tsraking_driver.html}})}

Helper function for the \code{\link[=tsraking]{tsraking()}} function that conveniently determines the required set of raking
problems to be solved and internally generates the individual calls to \code{\link[=tsraking]{tsraking()}}. It is especially
useful in the context of temporal total (e.g., annual total) preservation where each individual raking
problem either involves a single period for incomplete temporal groups (e.g., incomplete years) or several
periods for complete temporal groups (e.g., the set of periods of a complete year).
}
\details{
This function solves one raking problem with \code{\link[=tsraking]{tsraking()}} per processing group (see section \strong{Processing groups} for
details). The mathematical expression of these raking problem can be found in the \strong{Details} section of the \code{\link[=tsraking]{tsraking()}}
documentation.

The alterability coefficients data frame (argument \code{alterability_df}) specified with \code{\link[=tsraking_driver]{tsraking_driver()}} can either
contain:
\itemize{
\item A single observation: the specified coefficients will be used for all periods of input time series object (argument
\code{in_ts}).
\item A number of observations equal to \code{frequency(in_ts)}: the specified coefficients will be used for the
corresponding \emph{cycle} of the input time series object (argument \code{in_ts}) periods. Monthly data example: 1\if{html}{\out{<sup>}}st\if{html}{\out{</sup>}}
observation for January, 2\if{html}{\out{<sup>}}nd\if{html}{\out{</sup>}} observation for February, etc.).
\item A number of observations equal to \code{nrow(in_ts)}: the specified coefficients will be used for the corresponding
periods of the input time series object (argument \code{in_ts}), i.e., 1\if{html}{\out{<sup>}}st\if{html}{\out{</sup>}} observation for the 1\if{html}{\out{<sup>}}st\if{html}{\out{</sup>}}
period, 2\if{html}{\out{<sup>}}nd\if{html}{\out{</sup>}} observation for the 2\if{html}{\out{<sup>}}nd\if{html}{\out{</sup>}} period, etc.).
}

Specifying \code{quiet = TRUE} will suppress the \code{\link[=tsraking]{tsraking()}} messages (e.g., function header) and only
display essential information such as warnings, errors and the period (or set of periods) being reconciled.
We advise against \emph{wrapping} your \code{\link[=tsraking_driver]{tsraking_driver()}} function call with \code{\link[=suppressMessages]{suppressMessages()}} to further
suppress the display of the \emph{raking period(s)} information as this would make troubleshooting difficult
in case of issues with individual raking problems.

Although \code{\link[=tsraking]{tsraking()}} could be called with \verb{*apply()} to successively reconcile all the periods of the input time
series (\code{in_ts}), using \code{\link[=tsraking_driver]{tsraking_driver()}} has a few advantages, namely:
\itemize{
\item temporal total preservation (only period-by-period processing, without temporal total preservation,
would be possible with \verb{*apply()});
\item more flexibility in the specification of user-defined alterability coefficients (e.g., period-specific values);
\item display of the period being processed (reconciled) in the console, which is useful for troubleshooting individual
raking problems;
\item improved error handling, i.e., better management of warnings or errors if they were to occur only for some
raking problems (periods);
\item readily returns a "ts" ("mts") object.
}
}
\section{Processing groups}{
The set of periods of a given reconciliation (raking or balancing) problem is called a \emph{processing group}
and either corresponds to:
\itemize{
\item a \strong{single period} with period-by-period processing or, when preserving temporal totals, for the individual
periods of an incomplete temporal group (e.g., an incomplete year)
\item or the \strong{set of periods of a complete temporal group} (e.g., a complete year) when preserving temporal
totals.
}

The total number of processing groups (total number of reconciliation problems) depends on the set of
periods in the input time series object (argument \code{in_ts}) and on the value of arguments
\code{temporal_grp_periodicity} and \code{temporal_grp_start}.

Common scenarios include \code{temporal_grp_periodicity = 1} (default) for period-by period processing without
temporal total preservation and \code{temporal_grp_periodicity = frequency(in_ts)} for the preservation of annual
totals (calendar years by default). Argument \code{temporal_grp_start} allows the specification of other types of
(\emph{non-calendar}) years. E.g., fiscal years starting on April correspond to \code{temporal_grp_start = 4} with monthly
data and \code{temporal_grp_start = 2} with quarterly data. Preserving quarterly totals with monthly data would
correspond to \ifelse{latex}{\code{temporal_grp _periodicity = 3}}{\code{temporal_grp_periodicity = 3}}.

By default, temporal groups covering more than a year (i.e., corresponding to \ifelse{latex}{\code{
temporal_grp _periodicity > frequency(in_ts)}}{\code{temporal_grp_periodicity > frequency(in_ts)}} start on a
year that is a multiple of \ifelse{latex}{\code{ceiling(temporal_grp _periodicity / frequency(in_ts))}}{\code{
ceiling(temporal_grp_periodicity / frequency(in_ts))}}. E.g., biennial groups corresponding to \ifelse{latex}{
\code{temporal_grp_per iodicity = 2 * frequency(in_ts)}}{\code{temporal_grp_periodicity = 2 * frequency(in_ts)}}
start on an \emph{even year} by default. This behaviour can be changed with argument \code{temporal_grp_start}. E.g., the
preservation of biennial totals starting on an \emph{odd year} instead of an \emph{even year} (default) corresponds to
\code{temporal_grp_start = frequency(in_ts) + 1} (along with \code{temporal_grp_periodicity = 2 * frequency(in_ts)}).

See the \code{\link[=gs.build_proc_grps]{gs.build_proc_grps()}} \strong{Examples} for common processing group scenarios.
}

\examples{
# 1-dimensional raking problem where the quarterly sales of cars in the 3 prairie
# provinces (Alb., Sask. and Man.) for 8 quarters, from 2019 Q2 to 2021 Q1, must
# sum up to the total (`cars_tot`).

# Problem metadata
my_metadata <- data.frame(
  series = c("cars_alb", "cars_sask", "cars_man"),
  total1 = rep("cars_tot", 3)
)
my_metadata

# Problem data
my_series <- ts(
  matrix(
    c(
      14, 18, 14, 58,
      17, 14, 16, 44,
      14, 19, 18, 58,
      20, 18, 12, 53,
      16, 16, 19, 44,
      14, 15, 16, 50,
      19, 20, 14, 52,
      16, 15, 19, 51
    ),
    ncol = 4,
    byrow = TRUE
  ),
  start = c(2019, 2),
  frequency = 4,
  names = c("cars_alb", "cars_sask", "cars_man", "cars_tot")
)


###########
# Example 1: Period-by-period processing without temporal total preservation.

# Reconcile the data
out_raked1 <- tsraking_driver(my_series, my_metadata)

# Initial data
my_series

# Reconciled data
out_raked1

# Check the output cross-sectional constraint
all.equal(
  rowSums(out_raked1[, my_metadata$series]), 
  as.vector(out_raked1[, "cars_tot"])
)

# Check the control total (fixed)
all.equal(my_series[, "cars_tot"], out_raked1[, "cars_tot"])


###########
# Example 2: Annual total preservation for year 2020 (period-by-period processing
#            for incomplete years 2019 and 2021), with `quiet = TRUE` to avoid
#            displaying the function header for all processing groups.

# First, check that the 2020 annual total for the total series (`cars_tot`) and the
# sum of the component series (`cars_alb`, `cars_sask` and `cars_man`) matches.
# Otherwise, this "grand total" discrepancy would first have to be resolved before
# calling `tsraking_driver()`.
tot2020 <- aggregate.ts(window(my_series, start = c(2020, 1), end = c(2020, 4)))
all.equal(
  as.numeric(tot2020[, "cars_tot"]), 
  sum(tot2020[, my_metadata$series])
)

# Reconcile the data
out_raked2 <- tsraking_driver(
  in_ts = my_series,
  metadata_df = my_metadata,
  quiet = TRUE,
  temporal_grp_periodicity = frequency(my_series)
)

# Initial data
my_series

# Reconciled data
out_raked2

# Check the output cross-sectional constraint
all.equal(
  rowSums(out_raked2[, my_metadata$series]), 
  as.vector(out_raked2[, "cars_tot"])
)

# Check the output temporal constraints (2020 annual totals for each series)
all.equal(
  tot2020,
  aggregate.ts(window(out_raked2, start = c(2020, 1), end = c(2020, 4)))
)

# Check the control total (fixed)
all.equal(my_series[, "cars_tot"], out_raked2[, "cars_tot"])


###########
# Example 3: Annual total preservation for fiscal years defined from April to March
#            (2019Q2-2020Q1 and 2020Q2-2021Q1).

# Calculate the fiscal year totals (as an annual "ts" object)
fiscalYr_tot <- ts(
  rbind(
    aggregate.ts(window(my_series, start = c(2019, 2), end = c(2020, 1))),
    aggregate.ts(window(my_series, start = c(2020, 2), end = c(2021, 1)))
  ),
  start = 2019,
  frequency = 1
)

# Discrepancies in both fiscal year totals (total series vs. sum of the component series)
as.numeric(fiscalYr_tot[, "cars_tot"]) - rowSums(fiscalYr_tot[, my_metadata$series])


# 3a) Reconcile the fiscal year totals (rake the fiscal year totals of the component series
#     to those of the total series).
new_fiscalYr_tot <- tsraking_driver(
  in_ts = fiscalYr_tot,
  metadata_df = my_metadata,
  quiet = TRUE
)

# Confirm that the previous discrepancies are now "gone" (are both zero)
as.numeric(new_fiscalYr_tot[, "cars_tot"]) - 
  rowSums(new_fiscalYr_tot[, my_metadata$series])

# 3b) Benchmark the quarterly component series to these new (coherent) fiscal year totals.
out_bench <- benchmarking(
  series_df = ts_to_tsDF(my_series[, my_metadata$series]),
  benchmarks_df = ts_to_bmkDF(
    new_fiscalYr_tot[, my_metadata$series],
    ind_frequency = frequency(my_series),
    
    # Fiscal years starting on Q2 (April)
    bmk_interval_start = 2
  ),
  rho = 0.729,
  lambda = 1,
  biasOption = 2,
  allCols = TRUE,
  quiet = TRUE
)
my_new_ser <- tsDF_to_ts(
  cbind(out_bench$series, cars_tot = my_series[, "cars_tot"]),
  frequency = frequency(my_series)
)

# 3c) Reconcile the quarterly data with preservation of fiscal year totals.
out_raked3 <- tsraking_driver(
  in_ts = my_new_ser,
  metadata_df = my_metadata,
  temporal_grp_periodicity = frequency(my_series),
  
  # Fiscal years starting on Q2 (April)
  temporal_grp_start = 2,
  
  quiet = TRUE
)

# Initial data
my_series

# With coherent fiscal year totals
my_new_ser

# Reconciled data
out_raked3

# Check the output cross-sectional constraint
all.equal(
  rowSums(out_raked3[, my_metadata$series]), 
  as.vector(out_raked3[, "cars_tot"])
)

# Check the output temporal constraints (both fiscal year totals for all series)
all.equal(
  rbind(
    aggregate.ts(window(my_new_ser, start = c(2019, 2), end = c(2020, 1))),
    aggregate.ts(window(my_new_ser, start = c(2020, 2), end = c(2021, 1)))
  ),
  rbind(
    aggregate.ts(window(out_raked3, start = c(2019, 2), end = c(2020, 1))),
    aggregate.ts(window(out_raked3, start = c(2020, 2), end = c(2021, 1)))
  )
)

# Check the control total (fixed)
all.equal(my_series[, "cars_tot"], out_raked3[, "cars_tot"])
}
\references{
Statistics Canada (2018). "Chapter 6: Advanced topics", \strong{Theory and Application of Reconciliation
(Course code 0437)}, Statistics Canada, Ottawa, Canada.
}
\seealso{
\code{\link[=tsraking]{tsraking()}} \code{\link[=tsbalancing]{tsbalancing()}} \code{\link[=rkMeta_to_blSpecs]{rkMeta_to_blSpecs()}} \code{\link[=gs.build_proc_grps]{gs.build_proc_grps()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsbalancing.R
\name{build_balancing_problem}
\alias{build_balancing_problem}
\title{Build the elements of balancing problems.}
\usage{
build_balancing_problem(
  in_ts,
  problem_specs_df,
  in_ts_name = deparse1(substitute(in_ts)),
  ts_freq = stats::frequency(in_ts),
  periods = gs.time2str(in_ts),
  n_per = nrow(as.matrix(in_ts)),
  specs_df_name = deparse1(substitute(problem_specs_df)),
  temporal_grp_periodicity = 1,
  alter_pos = 1,
  alter_neg = 1,
  alter_mix = 1,
  lower_bound = -Inf,
  upper_bound = Inf,
  validation_only = FALSE
)
}
\arguments{
\item{in_ts}{(mandatory)

Time series (object of class "ts" or "mts") that contains the time series data to be reconciled.
They are the balancing problems' input data (initial solutions).}

\item{problem_specs_df}{(mandatory)

Balancing problem specifications data frame (object of class "data.frame"). Using a sparse format inspired from the
SAS/OR\eqn{^\circledR}{®} LP procedure’s \emph{sparse data input format} (SAS Institute 2015), it contains only the relevant
information such as the nonzero coefficients of the balancing constraints as well as the non-default alterability
coefficients and lower/upper bounds (i.e., values that would take precedence over those defined with arguments \code{alter_pos},
\code{alter_neg}, \code{alter_mix}, \code{alter_temporal}, \code{lower_bound} and \code{upper_bound}).

The information is provided using four mandatory variables (\code{type}, \code{col}, \code{row} and \code{coef}) and one optional variable
(\code{timeVal}). An observation (a row) in the problem specs data frame either defines a label for one of the seven types of the
balancing problem elements with columns \code{type} and \code{row} (see \emph{Label definition records} below) or specifies coefficients
(numerical values) for those balancing problem elements with variables \code{col}, \code{row}, \code{coef} and \code{timeVal} (see \emph{Information
specification records} below).
\itemize{
\item \strong{Label definition records} (\code{type} is not missing (is not \code{NA}))
\itemize{
\item \code{type} (chr): reserved keyword identifying the type of problem element being defined:
\itemize{
\item \code{EQ}: equality (\eqn{=}) balancing constraint
\item \code{LE}: lower or equal (\eqn{\le}{<=}) balancing constraint
\item \code{GE}: greater or equal (\eqn{\ge}{>=}) balancing constraint
\item \code{lowerBd}: period value lower bound
\item \code{upperBd}: period value upper bound
\item \code{alter}: period values alterability coefficient
\item \code{alterTmp}: temporal total alterability coefficient
}
\item \code{row} (chr): label to be associated to the problem element (\emph{\code{type} keyword})
\item \emph{all other variables are irrelevant and should contain missing data (\code{NA} values)} \cr \cr
}
\item \strong{Information specification records} (\code{type} is missing (is \code{NA}))
\itemize{
\item \code{type} (chr): not applicable (\code{NA})
\item \code{col} (chr): series name or reserved word \verb{_rhs_} to specify a balancing constraint right-hand side (RHS) value.
\item \code{row} (chr): problem element label.
\item \code{coef} (num): problem element value:
\itemize{
\item balancing constraint series coefficient or RHS value
\item series period value lower or upper bound
\item series period value or temporal total alterability coefficient
}
\item \code{timeVal} (num): optional time value to restrict the application of series bounds or alterability coefficients
to a specific time period (or temporal group). It corresponds to the time value, as returned by \code{stats::time()}, of a given
input time series (argument \code{in_ts}) period (observation) and is conceptually equivalent to \eqn{year + (period - 1) / 
  frequency}.
}
}

Note that empty strings (\code{""} or \code{''}) for character variables are interpreted as missing (\code{NA}) by the function. Variable
\code{row} identifies the elements of the balancing problem and is the key variable that makes the link between both types of
records. The same label (\code{row}) cannot be associated with more than one type of problem element (\code{type}) and multiple labels
(\code{row}) cannot be defined for the same given type of problem element (\code{type}), except for balancing constraints (values
\code{"EQ"}, \code{"LE"} and \code{"GE"} of column \code{type}). User-friendly features of the problem specs data frame include:
\itemize{
\item The order of the observations (rows) is not important.
\item Character values (variables \code{type}, \code{row} and \code{col}) are not case sensitive (e.g., strings \code{"Constraint 1"} and
\code{"CONSTRAINT 1"} for \code{row} would be considered as the same problem element label), except when \code{col} is used to specify a
series name (a column of the input time series object) where \strong{case sensitivity is enforced}.
\item The variable names of the problem specs data frame are also not case sensitive (e.g., \code{type}, \code{Type} or \code{TYPE} are all
valid) and \code{time_val} is an accepted variable name (instead of \code{timeVal}).
}

Finally, the following table lists valid aliases for the \emph{\code{type} keywords} (type of problem element):\tabular{cl}{
   \strong{Keyword} \tab \strong{Aliases} \cr
   \code{EQ} \tab \code{==}, \code{=} \cr
   \code{LE} \tab \code{<=}, \code{<} \cr
   \code{GE} \tab \code{>=}, \code{>} \cr
   \code{lowerBd} \tab \code{lowerBound}, \code{lowerBnd}, + \emph{same terms with '_', '.' or ' ' between words} \cr
   \code{upperBd} \tab \code{upperBound}, \code{upperBnd}, + \emph{same terms with '_', '.' or ' ' between words} \cr
   \code{alterTmp} \tab \code{alterTemporal}, \code{alterTemp}, + \emph{same terms with '_', '.' or ' ' between words} \cr
}


Reviewing the \strong{Examples} should help conceptualize the balancing problem specifications data frame.}

\item{in_ts_name}{(optional)

String containing the value of argument \code{in_ts}.

\strong{Default value} is \code{in_ts_name = deparse1(substitute(in_ts))}.}

\item{ts_freq}{(optional)

Frequency of the time series object (argument \code{in_ts}).

\strong{Default value} is \code{ts_freq = stats::frequency(in_ts)}.}

\item{periods}{(optional)

Character vector describing the time series object (argument \code{in_ts}) periods.

\strong{Default value} is \code{periods = gs.time2str(in_ts)}.}

\item{n_per}{(optional)

Number of periods of the time series object (argument \code{in_ts}).

\strong{Default value} is \code{n_per = nrow(as.matrix(in_ts))}.}

\item{specs_df_name}{(optional)

String containing the value of argument \code{problem_specs_df}.

\strong{Default value} is \code{specs_df_name = deparse1(substitute(problem_specs_df))}.}

\item{temporal_grp_periodicity}{(optional)

Positive integer defining the number of periods in temporal groups for which the totals should be preserved.
E.g., specify \code{temporal_grp_periodicity = 3} with a monthly time series for quarterly total preservation and
\code{temporal_grp_periodicity = 12} (or \code{temporal_grp_periodicity = frequency(in_ts)}) for annual total preservation.
Specifying \code{temporal_grp_periodicity = 1} (\emph{default}) corresponds to period-by-period processing without
temporal total preservation.

\strong{Default value} is \code{temporal_grp_periodicity = 1} (period-by-period processing without temporal total preservation).}

\item{alter_pos}{(optional)

Nonnegative real number specifying the default alterability coefficient associated to the values of time series with \strong{positive}
coefficients in all balancing constraints in which they are involved (e.g., component series in aggregation table raking problems).
Alterability coefficients provided in the problem specification data frame (argument \code{problem_specs_df}) override this value.

\strong{Default value} is \code{alter_pos = 1.0} (nonbinding values).}

\item{alter_neg}{(optional)

Nonnegative real number specifying the default alterability coefficient associated to the values of time series with \strong{negative}
coefficients in all balancing constraints in which they are involved (e.g., marginal totals in aggregation table raking problems).
Alterability coefficients provided in the problem specification data frame (argument \code{problem_specs_df}) override this value.

\strong{Default value} is \code{alter_neg = 1.0} (nonbinding values).}

\item{alter_mix}{(optional)

Nonnegative real number specifying the default alterability coefficient associated to the values of time series with a mix of
\strong{positive and negative} coefficients in the balancing constraints in which they are involved. Alterability coefficients provided
in the problem specification data frame (argument \code{problem_specs_df}) override this value.

\strong{Default value} is \code{alter_mix = 1.0} (nonbinding values).}

\item{lower_bound}{(optional)

Real number specifying the default lower bound for the time series values. Lower bounds provided in the problem specification
data frame (argument \ifelse{latex}{\code{problem _specs_df}}{\code{problem_specs_df}}) override this value.

\strong{Default value} is \code{lower_bound = -Inf} (unbounded).}

\item{upper_bound}{(optional)

Real number specifying the default upper bound for the time series values. Upper bounds provided in the problem specification
data frame (argument \ifelse{latex}{\code{problem _specs_df}}{\code{problem_specs_df}}) override this value.

\strong{Default value} is \code{upper_bound = Inf} (unbounded).}

\item{validation_only}{(optional)

Logical argument specifying whether the function should only perform input data validation or not. When
\code{validation_only = TRUE}, the specified \emph{balancing constraints} and \emph{period value (lower and upper) bounds} constraints
are validated against the input time series data, allowing for discrepancies up to the value specified with argument
\code{validation_tol}. Otherwise, when \code{validation_only = FALSE} (default), the input data are first reconciled and the
resulting (output) data are then validated.

\strong{Default value} is \code{validation_only = FALSE}.}
}
\value{
A list with the elements of the balancing problems (excluding the temporal totals info):
\itemize{
\item \code{labels_df}: cleaned-up version of the \emph{label definition records} from \code{problem_specs_df}
(\code{type} is not missing (is not \code{NA})); extra columns:
\itemize{
\item \code{type.lc} : \code{tolower(type)}
\item \code{row.lc}  : \code{tolower(row)}
\item \code{con.flag}: \code{type.lc \%in\% c("eq", "le", "ge")}
}
\item \code{coefs_df} : cleaned-up version of the information specification records from \code{problem_specs_df}
(\code{type} is missing (is \code{NA}); extra columns:
\itemize{
\item \code{row.lc}  : \code{tolower(row)}
\item \code{con.flag}: \code{labels_df$con.flag} allocated through \code{row.lc}
}
\item \code{values_ts}: reduced version of \code{in_ts} with only the relevant series (see vector \code{ser_names})
\item \code{lb}       : lower bound info (\code{type.lc  = "lowerbd"}) for the relevant series; list object with the
following elements:
\itemize{
\item \code{coefs_ts}       : lower bound values for series and period
\item \code{nondated_coefs} : vector of nondated lower bounds from \code{problem_specs_df} (\code{timeVal} is \code{NA})
\item \code{nondated_id_vec}: vector of \code{ser_names} id's associated to vector \code{nondated_coefs}
\item \code{dated_id_vec}   : vector of \code{ser_names} id's associated to dated lower bounds from
\code{problem_specs_df} (\code{timeVal} is not \code{NA})
}
\item \code{ub}       : \code{lb} equivalent for upper bounds (\code{type.lc = "upperbd"})
\item \code{alter}    : \code{lb} equivalent for period value alterability coefficients (\code{type.lc = "alter"})
\item \code{altertmp} : \code{lb} equivalent for temporal total alterability coefficients (\code{type.lc = "altertmp"})
\item \code{ser_names}: vector of the relevant series names (set of series involved in the balancing constraints)
\item \code{pos_ser}  : vector of series names that have only positive nonzero coefficients across all balancing constraints
\item \code{neg_ser}  : vector of series names that have only negative nonzero coefficients across all balancing constraints
\item \code{mix_ser}  : vector of series names that have both positive and negative nonzero coefficients across all balancing
constraints
\item \code{A1},\code{op1},\code{b1}: balancing constraint elements for problems involving a single period (e.g., each period of an
incomplete temporal group)
\item \code{A2},\code{op2},\code{b2}: balancing constraint elements for problems involving \code{temporal_grp_periodicity} periods (e.g., the set
of periods of a complete temporal group)
}
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/build_balancing_problem.html}})}

This function is used internally by \code{\link[=tsbalancing]{tsbalancing()}} to build the elements of the balancing problems.
It can also be useful to derive the indirect series associated to equality balancing constraints manually
(outside of the \code{\link[=tsbalancing]{tsbalancing()}} context).
}
\details{
See \code{\link[=tsbalancing]{tsbalancing()}} for a detailed description of \emph{time series balancing} problems.

Any missing (\code{NA}) value found in the input time series object (argument \code{in_ts}) would be replaced with 0 in \code{values_ts}
and trigger a warning message.

The returned elements of the balancing problems do not include the implicit temporal totals (i.e., elements \code{A2}, \code{op2}
and \code{b2} only contain the balancing constraints).

Multi-period balancing problem elements \code{A2}, \code{op2} and \code{b2} (when \code{temporal_grp_periodicity > 1}) are constructed
\emph{column by column} (in "column-major order"), corresponding to the default behaviour of R for converting objects of class
"matrix" into vectors. I.e., the balancing constraints conceptually correspond to:
\itemize{
\item \verb{A1 \%*\% values_ts[t, ] op1 b1} for problems involving a single period (\code{t})
\item \verb{A2 \%*\% as.vector(values_ts[t1:t2, ]) op2 b2} for problems involving \code{temporal_grp_periodicity} periods (\code{t1:t2}).
}

Notes:
\itemize{
\item Argument \code{alter_temporal} has not been applied yet at this point and \code{altertmp$coefs_ts} only contains the
coefficients specified in the problem specs data frame (argument \code{problem_specs_df}). I.e., \code{altertmp$coefs_ts} contains
missing (\code{NA}) values except for the temporal total alterability coefficients included in (specified with) \code{problem_specs_df}.
This is done in order to simplify the identification of the first non missing (non \code{NA}) temporal total alterability
coefficient of each complete temporal group (to occur later, when applicable, inside \code{\link[=tsbalancing]{tsbalancing()}}).
\item Argument validation is not performed here; it is (bluntly) assumed that the function is called by \code{\link[=tsbalancing]{tsbalancing()}}
where a thorough validation of the arguments is done.
}
}
\examples{
######################################################################################
#        Indirect series derivation framework with `tsbalancing()` metadata
######################################################################################
#
# Is is assumed (agreed) that...
#
# a) All balancing constraints are equality constraints (`type = EQ`).
# b) All constraints have only one nonbinding (free) series: the series to be derived
#    (i.e., all series have an alter. coef of 0 except the series to be derived).
# c) Each constraint derives a different (new) series.
# d) Constraints are the same for all periods (i.e., no "dated" alter. coefs 
#    specified with column `timeVal`).
######################################################################################


# Derive the 5 marginal totals of a 2 x 3 two-dimensional data cube using `tsbalancing()` 
# metadata (data cube aggregation constraints respect the above assumptions).


# Build the balancing problem specs through the (simpler) raking metadata.
my_specs <- rkMeta_to_blSpecs(
  data.frame(
    series = c(
      "A1", "A2", "A3",
      "B1", "B2", "B3"
    ),
    total1 = c(
      rep("totA", 3),
      rep("totB", 3)
    ),
    total2 = rep(c("tot1", "tot2", "tot3"), 2)
  ),
  alterSeries = 0,  # binding (fixed) component series
  alterTotal1 = 1,  # nonbinding (free) marginal totals (to be derived)
  alterTotal2 = 1)  # nonbinding (free) marginal totals (to be derived)
my_specs

# 6 periods (quarters) of data with marginal totals set to zero (0): they MUST exist
# in the input data AND contain valid (non missing) data.
my_ts <- ts(
  data.frame(
    A1 = c(12, 10, 12,  9, 15,  7),
    B1 = c(20, 21, 15, 17, 19, 18),
    A2 = c(14,  9,  8,  9, 11, 10),
    B2 = c(20, 29, 20, 24, 21, 17),
    A3 = c(13, 15, 17, 14, 16, 12),
    B3 = c(24, 20, 30, 23, 21, 19),
    tot1 = rep(0, 6),
    tot2 = rep(0, 6),
    tot3 = rep(0, 6),
    totA = rep(0, 6),
    totB = rep(0, 6)
  ),
  start = 2019, frequency = 4
)

# Get the balancing problem elements.
n_per <- nrow(my_ts)
p <- build_balancing_problem(
  my_ts, my_specs, 
  temporal_grp_periodicity = n_per
)

# `A2`, `op2` and `b2` define 30 constraints (5 marginal totals X 6 periods) 
# involving a total of 66 time series data points (11 series X 6 periods) of which 
# 36 belong to the 6 component series and 30 belong to the 5 marginal totals.
dim(p$A2)

# Get the names of the marginal totals (series with a nonzero alter. coef), in the order 
# in which the corresponding constraints appear in the specs (constraints specification 
# order).
tmp <- p$coefs_df$col[p$coefs_df$con.flag]
tot_names <- tmp[
  tmp \%in\% p$ser_names[p$alter$nondated_id_vec[p$alter$nondated_coefs != 0]]
]

# Define logical flags identifying the marginal total columns:
# - `tot_col_logi1`: for single-period elements (of length 11 = number of series)
# - `tot_col_logi2`: for multi-period elements (of length 66 = number of data points), 
#                    in "column-major order" (the `A2` matrix element construction order)
tot_col_logi1 <- p$ser_names \%in\% tot_names
tot_col_logi2 <- rep(tot_col_logi1, each = n_per)

# Order of the marginal totals to be derived based on
# ... the input data columns ("mts" object `my_ts`)
p$ser_names[tot_col_logi1]
# ... the constraints specification (data frame `my_specs`)
tot_names


# Calculate the 5 marginal totals for all 6 periods
# Note: the following calculation allows for general linear equality constraints, i.e.,
#       a) nonzero right-hand side (RHS) constraint values (`b2`) and 
#       b) nonzero constraint coefs other than 1 for the component series and -1 for 
#          the derived series.
my_ts[, tot_names] <- {
  (
    # Constraints RHS.
    p$b2 - 

    # Sums of the components ("weighted" by the constraint coefficients).
    p$A2[, !tot_col_logi2, drop = FALSE] \%*\% as.vector(p$values_ts[, !tot_col_logi1])
  ) /

  # Derived series constraint coefficients: `t()` allows for a "row-major order" search 
  # in matrix `A2` (i.e., according to the constraints specification order).
  # Note: `diag(p$A2[, tot_col_logi2])` would work if `p$ser_names[tot_col_logi1]` and 
  #       `tot_names` were identical (same totals order); however, the following search 
  #       in "row-major order" will always work (and is necessary in the current case).
  t(p$A2[, tot_col_logi2])[t(p$A2[, tot_col_logi2]) != 0]
}
my_ts
}
\seealso{
\code{\link[=tsbalancing]{tsbalancing()}} \code{\link[=build_raking_problem]{build_raking_problem()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-common.R
\name{time_values_conv}
\alias{time_values_conv}
\alias{gs.time2year}
\alias{gs.time2per}
\alias{gs.time2str}
\title{Time values conversion functions}
\usage{
gs.time2year(ts)

gs.time2per(ts)

gs.time2str(ts, sep = "-")
}
\arguments{
\item{ts}{(mandatory)

Time series (object of class "ts" or "mts").}

\item{sep}{(optional)

String (character constant) specifying the separator to use between the year
and period values.

\strong{Default value} is \code{sep = "-"}.}
}
\value{
\code{\link[=gs.time2year]{gs.time2year()}} returns an integer vector of the "nearest" year (time unit) values. This function is the
equivalent of \code{\link[stats:cycle]{stats::cycle()}} for time unit values.

\code{\link[=gs.time2per]{gs.time2per()}} returns an integer vector of the period (cycle) values (see \code{\link[stats:cycle]{stats::cycle()}}).

\code{\link[=gs.time2str]{gs.time2str()}} returns a character vector corresponding to \code{gs.time2year(ts)} if \code{stats::frequency(ts) == 1}
or \code{gs.time2year(ts)} and \code{gs.time2per(ts)} separated with \code{sep} otherwise.
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/time_values_conv.html}})}

Time values conversion functions used internally by other gseries functions.
}
\examples{
# Dummy monthly time series 
mth_ts <- ts(rep(NA, 15), start = c(2019, 1), frequency = 12)
mth_ts
gs.time2year(mth_ts)
gs.time2per(mth_ts)
gs.time2str(mth_ts)
gs.time2str(mth_ts, sep = "m")

# Dummy quarterly time series 
qtr_ts <- ts(rep(NA, 5), start = c(2019, 1), frequency = 4)
qtr_ts
gs.time2year(qtr_ts)
gs.time2per(qtr_ts)
gs.time2str(qtr_ts)
gs.time2str(qtr_ts, sep = "q")
}
\seealso{
\code{\link[=ts_to_tsDF]{ts_to_tsDF()}} \code{\link[=ts_to_bmkDF]{ts_to_bmkDF()}} \code{\link[=gs.build_proc_grps]{gs.build_proc_grps()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gseries-package.R
\docType{package}
\name{gseries-package}
\alias{gseries}
\alias{gseries-package}
\title{gseries: Improve the Coherence of Your Time Series Data}
\description{
'R' version of 'G-Series', Statistics Canada's generalized system devoted to the benchmarking and reconciliation of time series data. The methods used in 'G-Series' essentially come from Dagum, E. B., and P. Cholette (2006) \doi{10.1007/0-387-35439-5}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://StatCan.github.io/gensol-gseries/en/}
  \item \url{https://StatCan.github.io/gensol-gseries/fr/}
  \item Report bugs at \url{https://github.com/StatCan/gensol-gseries/issues/}
}

}
\author{
\strong{Maintainer}: Michel Ferland \email{michel.ferland@statcan.gc.ca}

Authors:
\itemize{
  \item Michel Ferland \email{michel.ferland@statcan.gc.ca}
}

Other contributors:
\itemize{
  \item Statistics Canada [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-common.R
\name{gs.build_proc_grps}
\alias{gs.build_proc_grps}
\title{Build reconciliation processing groups}
\usage{
gs.build_proc_grps(
  ts_yr_vec,
  ts_per_vec,
  n_per,
  ts_freq,
  temporal_grp_periodicity,
  temporal_grp_start
)
}
\arguments{
\item{ts_yr_vec}{(mandatory)

Vector of the time series year (time unit) values (see \code{\link[=gs.time2year]{gs.time2year()}}).}

\item{ts_per_vec}{(mandatory)

Vector of the time series period (cycle) values (see \code{\link[=gs.time2per]{gs.time2per()}}).}

\item{n_per}{(mandatory)

Time series length (number of periods).}

\item{ts_freq}{(mandatory)

Time series frequency (see \code{\link[stats:frequency]{stats::frequency()}}).}

\item{temporal_grp_periodicity}{(mandatory)

Number of periods in temporal groups.}

\item{temporal_grp_start}{(mandatory)

First period of temporal groups.}
}
\value{
A data frame with the following variables (columns):
\itemize{
\item \code{grp}         : integer vector identifying the processing group (\verb{1:<number-of-groups>}).
\item \code{beg_per}     : integer vector identifying the first period of the processing group.
\item \code{end_per}     : integer vector identifying the last period of the processing group.
\item \code{complete_grp}: logical vector indicating if the processing group corresponds to a complete temporal group.
}
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/gs.build_proc_grps.html}})}

This function builds the processing groups data frame for reconciliation problems. It is used internally by
\code{\link[=tsraking_driver]{tsraking_driver()}} and \code{\link[=tsbalancing]{tsbalancing()}}.
}
\section{Processing groups}{
The set of periods of a given reconciliation (raking or balancing) problem is called a \emph{processing group}
and either corresponds to:
\itemize{
\item a \strong{single period} with period-by-period processing or, when preserving temporal totals, for the individual
periods of an incomplete temporal group (e.g., an incomplete year)
\item or the \strong{set of periods of a complete temporal group} (e.g., a complete year) when preserving temporal
totals.
}

The total number of processing groups (total number of reconciliation problems) depends on the set of
periods in the input time series object (argument \code{in_ts}) and on the value of arguments
\code{temporal_grp_periodicity} and \code{temporal_grp_start}.

Common scenarios include \code{temporal_grp_periodicity = 1} (default) for period-by period processing without
temporal total preservation and \code{temporal_grp_periodicity = frequency(in_ts)} for the preservation of annual
totals (calendar years by default). Argument \code{temporal_grp_start} allows the specification of other types of
(\emph{non-calendar}) years. E.g., fiscal years starting on April correspond to \code{temporal_grp_start = 4} with monthly
data and \code{temporal_grp_start = 2} with quarterly data. Preserving quarterly totals with monthly data would
correspond to \ifelse{latex}{\code{temporal_grp _periodicity = 3}}{\code{temporal_grp_periodicity = 3}}.

By default, temporal groups covering more than a year (i.e., corresponding to \ifelse{latex}{\code{
temporal_grp _periodicity > frequency(in_ts)}}{\code{temporal_grp_periodicity > frequency(in_ts)}} start on a
year that is a multiple of \ifelse{latex}{\code{ceiling(temporal_grp _periodicity / frequency(in_ts))}}{\code{
ceiling(temporal_grp_periodicity / frequency(in_ts))}}. E.g., biennial groups corresponding to \ifelse{latex}{
\code{temporal_grp_per iodicity = 2 * frequency(in_ts)}}{\code{temporal_grp_periodicity = 2 * frequency(in_ts)}}
start on an \emph{even year} by default. This behaviour can be changed with argument \code{temporal_grp_start}. E.g., the
preservation of biennial totals starting on an \emph{odd year} instead of an \emph{even year} (default) corresponds to
\code{temporal_grp_start = frequency(in_ts) + 1} (along with \code{temporal_grp_periodicity = 2 * frequency(in_ts)}).

See the \code{\link[=gs.build_proc_grps]{gs.build_proc_grps()}} \strong{Examples} for common processing group scenarios.
}

\examples{
#######
# Preliminary setup

# Dummy monthly and quarterly time series (2.5 years long)
mth_ts <- ts(rep(NA, 30), start = c(2019, 1), frequency = 12)
mth_ts
qtr_ts <- ts(rep(NA, 10), start = c(2019, 1), frequency = 4)
qtr_ts

# Summarized time series info
ts_info <- function(ts, sep = "-") {
  list(
    y = gs.time2year(ts),     # years
    p = gs.time2per(ts),      # periods
    n = length(ts),           # length
    f = frequency(ts),        # frequency
    l = gs.time2str(ts, sep)  # labels
  )
}
mth_info <- ts_info(mth_ts)
qtr_info <- ts_info(qtr_ts, sep = "q")

# Function to add a description label for the processing group
add_desc <- function(grp_df, lab_vec, word) {
  grp_df$description <- ifelse(
    grp_df$complete_grp,
    paste0(
      "--- ", grp_df$end_per - grp_df$beg_per + 1, " ", word, "s: ",
      lab_vec[grp_df$beg_per], " to ", lab_vec[grp_df$end_per], " --- "
    ),
    paste0("--- 1 ", word, ": ", lab_vec[grp_df$beg_per], " ---")
  )
  
  grp_df
}




#######
# Common processing group scenarios for monthly data


# 0- Month-by-month processing (every single month is a processing group)
mth_grps0 <- gs.build_proc_grps(
  mth_info$y, mth_info$p, mth_info$n, mth_info$f,
  temporal_grp_periodicity = 1,
  temporal_grp_start = 1
)
tmp <- add_desc(mth_grps0, mth_info$l, "month")
head(tmp)
tail(tmp)


# Temporal groups corresponding to ...

# 1- calendar years
mth_grps1 <- gs.build_proc_grps(
  mth_info$y, mth_info$p, mth_info$n, mth_info$f,
  temporal_grp_periodicity = 12,
  temporal_grp_start = 1
)
add_desc(mth_grps1, mth_info$l, "month")

# 2- fiscal years starting on April
mth_grps2 <- gs.build_proc_grps(
  mth_info$y, mth_info$p, mth_info$n, mth_info$f,
  temporal_grp_periodicity = 12,
  temporal_grp_start = 4
)
add_desc(mth_grps2, mth_info$l, "month")

# 3- regular quarters (starting on Jan, Apr, Jul and Oct)
mth_grps3 <- gs.build_proc_grps(
  mth_info$y, mth_info$p, mth_info$n, mth_info$f,
  temporal_grp_periodicity = 3,
  temporal_grp_start = 1
)
add_desc(mth_grps3, mth_info$l, "month")

# 4- quarters shifted by one month (starting on Feb, May, Aug and Nov)
mth_grps4 <- gs.build_proc_grps(
  mth_info$y, mth_info$p, mth_info$n, mth_info$f,
  temporal_grp_periodicity = 3,
  temporal_grp_start = 2
)
add_desc(mth_grps4, mth_info$l, "month")




#######
# Common processing group scenarios for quarterly data


# 0- Quarter-by-quarter processing (every single quarter is a processing group)
qtr_grps0 <- gs.build_proc_grps(
  qtr_info$y, qtr_info$p, qtr_info$n, qtr_info$f,
  temporal_grp_periodicity = 1,
  temporal_grp_start = 1
)
add_desc(qtr_grps0, qtr_info$l, "quarter")


# Temporal groups corresponding to ...

# 1- calendar years
qtr_grps1 <- gs.build_proc_grps(
  qtr_info$y, qtr_info$p, qtr_info$n, qtr_info$f,
  temporal_grp_periodicity = 4,
  temporal_grp_start = 1
)
add_desc(qtr_grps1, qtr_info$l, "quarter")

# 2- fiscal years starting on April (2nd quarter)
qtr_grps2 <- gs.build_proc_grps(
  qtr_info$y, qtr_info$p, qtr_info$n, qtr_info$f,
  temporal_grp_periodicity = 4,
  temporal_grp_start = 2
)
add_desc(qtr_grps2, qtr_info$l, "quarter")
}
\seealso{
\code{\link[=tsraking_driver]{tsraking_driver()}} \code{\link[=tsbalancing]{tsbalancing()}} \link{time_values_conv}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsraking.R
\name{build_raking_problem}
\alias{build_raking_problem}
\title{Build the elements of raking problems.}
\usage{
build_raking_problem(
  data_df,
  metadata_df,
  data_df_name = deparse1(substitute(data_df)),
  metadata_df_name = deparse1(substitute(metadata_df)),
  alterability_df = NULL,
  alterSeries = 1,
  alterTotal1 = 0,
  alterTotal2 = 0
)
}
\arguments{
\item{data_df}{(mandatory)

Data frame (object of class "data.frame") that contains the time series data to be reconciled. It must minimally
contain variables corresponding to the component series and cross-sectional control totals specified in the
metadata data frame (argument \code{metadata_df}). If more than one observation (period) is provided, the sum of
the provided component series values will also be preserved as part of implicit temporal constraints.}

\item{metadata_df}{(mandatory)

Data frame (object of class "data.frame") that describes the cross-sectional aggregation constraints
(additivity rules) for the raking problem. Two character variables must be included in the metadata data frame:
\code{series} and \code{total1}. Two variables are optional: \code{total2} (character) and \code{alterAnnual} (numeric). The values
of variable \code{series} represent the variable names of the component series in the input time series data frame
(argument \code{data_df}). Similarly, the values of variables \code{total1} and \code{total2} represent the variable names of
the 1\if{html}{\out{<sup>}}st\if{html}{\out{</sup>}} and 2\if{html}{\out{<sup>}}nd\if{html}{\out{</sup>}} dimension cross-sectional control totals in the input time series data
frame. Variable \code{alterAnnual} contains the alterability coefficient for the temporal constraint associated to
each component series. When specified, the latter will override the default alterability coefficient specified
with argument \code{alterAnnual}.}

\item{data_df_name}{(optional)

String containing the value of argument \code{data_df}.

\strong{Default value} is \code{data_df_name = deparse1(substitute(data_df))}.}

\item{metadata_df_name}{(optional)

String containing the value of argument \code{metadata_df}.

\strong{Default value} is \code{metadata_df_name = deparse1(substitute(metadata_df))}.}

\item{alterability_df}{(optional)

Data frame (object of class "data.frame"), or \code{NULL}, that contains the alterability coefficients variables.
They must correspond to a component series or a cross-sectional control total, that is, a variable with the same
name must exist in the input time series data frame (argument \code{data_df}). The values of these alterability
coefficients will override the default alterability coefficients specified with arguments \code{alterSeries},
\code{alterTotal1} and \code{alterTotal2}. When the input time series data frame contains several observations and the
alterability coefficients data frame contains only one, the alterability coefficients are used (repeated) for
all observations of the input time series data frame. Alternatively, the alterability coefficients data frame
may contain as many observations as the input time series data frame.

\strong{Default value} is \code{alterability_df = NULL} (default alterability coefficients).}

\item{alterSeries}{(optional)

Nonnegative real number specifying the default alterability coefficient for the component series values. It
will apply to component series for which alterability coefficients have not already been specified in the
alterability coefficients data frame (argument \code{alterability_df}).

\strong{Default value} is \code{alterSeries = 1.0} (nonbinding component series values).}

\item{alterTotal1}{(optional)

Nonnegative real number specifying the default alterability coefficient for the 1\if{html}{\out{<sup>}}st\if{html}{\out{</sup>}} dimension
cross-sectional control totals. It will apply to cross-sectional control totals for which alterability
coefficients have not already been specified in the alterability coefficients data frame (argument
\code{alterability_df}).

\strong{Default value} is \code{alterTotal1 = 0.0} (binding 1\if{html}{\out{<sup>}}st\if{html}{\out{</sup>}} dimension cross-sectional control totals)}

\item{alterTotal2}{(optional)

Nonnegative real number specifying the default alterability coefficient for the 2\if{html}{\out{<sup>}}nd\if{html}{\out{</sup>}} dimension
cross-sectional control totals. It will apply to cross-sectional control totals for which alterability
coefficients have not already been specified in the alterability coefficients data frame (argument
\code{alterability_df}).

\strong{Default value} is \code{alterTotal2 = 0.0} (binding 2\if{html}{\out{<sup>}}nd\if{html}{\out{</sup>}} dimension cross-sectional control totals).}
}
\value{
A list with the elements of the raking problem (excluding the implicit temporal totals):
\itemize{
\item \code{x}        : vector of component series initial values
\item \code{c_x}      : vector of component series alterability coefficients
\item \code{comp_cols}: vector of component series (column) names
\item \code{g}        : vector of cross-sectional total initial values
\item \code{c_g}      : vector of cross-sectional total alterability coefficients
\item \code{tot_cols} : vector of cross-sectional total (column) names
\item \code{G}        : cross-sectional total aggregation matrix (\code{g = G \%*\% x})
}
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/build_raking_problem.html}})}

This function is used internally by \code{\link[=tsraking]{tsraking()}} to build the elements of the raking problem. It can also
be useful to derive the cross-sectional (marginal) totals of the raking problem manually (outside of the \code{\link[=tsraking]{tsraking()}}
context).
}
\details{
See \code{\link[=tsraking]{tsraking()}} for a detailed description of \emph{time series raking} problems.

The returned raking problem elements do not include the implicit component series temporal totals
when applicable (i.e., elements \code{g} and \code{G} only contain the cross-sectional totals info).

When the input data contains multiple periods (temporal total preservation scenario), raking problem
elements \code{x}, \code{c_x}, \code{g}, \code{c_g} and \code{G} are constructed \emph{column by column} (in "column-major order"),
corresponding to the default behaviour of R for converting objects of class "matrix" into vectors.

Note: argument validation is not performed here; it is (bluntly) assumed that the function is called
by \code{\link[=tsraking]{tsraking()}} where a thorough validation of the arguments is done.
}
\examples{
# Derive the 5 marginal totals of a 2 x 3 two-dimensional data cube using `tsraking()` 
# metadata.

my_metadata <- data.frame(
  series = c(
    "A1", "A2", "A3",
    "B1", "B2", "B3"
  ),
  total1 = c(
    rep("totA", 3),
    rep("totB", 3)
  ),
  total2 = rep(c("tot1", "tot2", "tot3"), 2)
)
my_metadata

# 6 periods of data with marginal totals set to `NA` (they MUST exist in the input data 
# but can be `NA`).
my_data <- data.frame(
  A1 = c(12, 10, 12,  9, 15,  7),
  B1 = c(20, 21, 15, 17, 19, 18),
  A2 = c(14,  9,  8,  9, 11, 10),
  B2 = c(20, 29, 20, 24, 21, 17),
  A3 = c(13, 15, 17, 14, 16, 12),
  B3 = c(24, 20, 30, 23, 21, 19),
  tot1 = rep(NA, 6),
  tot2 = rep(NA, 6),
  tot3 = rep(NA, 6),
  totA = rep(NA, 6),
  totB = rep(NA, 6)
)

# Get the raking problem elements.
p <- build_raking_problem(my_data, my_metadata)
str(p)

# Calculate the 5 marginal totals for all 6 periods.
my_data[p$tot_cols] <- p$G \%*\% p$x
my_data
}
\seealso{
\code{\link[=tsraking]{tsraking()}} \code{\link[=build_balancing_problem]{build_balancing_problem()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsDF_to_ts.R
\name{tsDF_to_ts}
\alias{tsDF_to_ts}
\title{Reciprocal function of \code{\link[=ts_to_tsDF]{ts_to_tsDF()}}}
\usage{
tsDF_to_ts(
  ts_df,
  frequency,
  yr_cName = "year",
  per_cName = "period"
)
}
\arguments{
\item{ts_df}{(mandatory)

Data frame (object of class "data.frame") to be converted.}

\item{frequency}{(mandatory)

Integer specifying the frequency of the time series to be converted. The frequency of a time series corresponds
to the maximum number of periods in a year (12 for a monthly data, 4 for a quarterly data, 1 for annual data).}

\item{yr_cName, per_cName}{(optional)

Strings specifying the name of the numeric variables (columns) in the input data frame that contain the data
point year and period identifiers.

\strong{Default values} are \code{yr_cName = "year"} and \code{per_cName   = "period"}.}
}
\value{
The function returns a time series object (class "ts" or "mts"), which can be explicitly coerced to another type
of object with the appropriate \verb{as*()} function (e.g., \code{tsibble::as_tsibble()} would coerce it to a tsibble).
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/tsDF_to_ts.html}})}

Convert a (non-stacked) time series data frame (\code{\link[=benchmarking]{benchmarking()}} and \code{\link[=stock_benchmarking]{stock_benchmarking()}} data format)
into a "ts" (or "mts") object.

This function is useful to convert the benchmarked data frame returned by a call to \code{\link[=benchmarking]{benchmarking()}} or
\code{\link[=stock_benchmarking]{stock_benchmarking()}} into a "ts" object, where one or several series were benchmarked in \emph{non BY-group}
processing mode. Stacked time series data frames associated to executions in \emph{BY-group} mode must first be
\emph{unstacked} with \code{\link[=unstack_tsDF]{unstack_tsDF()}}.
}
\examples{
# Initial quarterly time series (indicator series to be benchmarked)
qtr_ts <- ts(
  c(1.9, 2.4, 3.1, 2.2, 2.0, 2.6, 3.4, 2.4, 2.3),
  start = c(2015, 1), frequency = 4
)

# Annual time series (benchmarks)
ann_ts <- ts(c(10.3, 10.2), start = 2015, frequency = 1)


# Proportional benchmarking
out_bench <- benchmarking(
  ts_to_tsDF(qtr_ts),
  ts_to_bmkDF(ann_ts, ind_frequency = 4),
  rho = 0.729, lambda = 1, biasOption = 3,
  quiet = TRUE
)

# Initial and final (benchmarked) quarterly time series ("ts" objects)
qtr_ts
tsDF_to_ts(out_bench$series, frequency = 4)


# Proportional end-of-year stock benchmarking - multiple (3) series processed 
# with argument `by` (in BY-group mode)
qtr_mts <- ts.union(ser1 = qtr_ts,     ser2 = qtr_ts * 100, ser3 = qtr_ts * 10)
ann_mts <- ts.union(ser1 = ann_ts / 4, ser2 = ann_ts * 25,  ser3 = ann_ts * 2.5)
out_bench2 <- stock_benchmarking(
  stack_tsDF(ts_to_tsDF(qtr_mts)),
  stack_bmkDF(ts_to_bmkDF(
    ann_mts, ind_frequency = 4,
    discrete_flag = TRUE, alignment = "e")
  ),
  rho = 0.729, lambda = 1, biasOption = 3,
  by = "series",
  quiet = TRUE
)

# Initial and final (benchmarked) quarterly time series ("mts" objects)
qtr_mts
tsDF_to_ts(unstack_tsDF(out_bench2$series), frequency = 4)
}
\seealso{
\code{\link[=ts_to_tsDF]{ts_to_tsDF()}} \code{\link[=unstack_tsDF]{unstack_tsDF()}} \code{\link[=benchmarking]{benchmarking()}} \code{\link[=stock_benchmarking]{stock_benchmarking()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stock_benchmarking.R
\name{stock_benchmarking}
\alias{stock_benchmarking}
\title{Restore temporal constraints for stock series}
\usage{
stock_benchmarking(
  series_df,
  benchmarks_df,
  rho,
  lambda,
  biasOption,
  bias = NA,
  low_freq_periodicity = NA,
  n_low_freq_proj = 1,
  proj_knots_rho_bd = 0.995,
  tolV = 0.001,
  tolP = NA,
  warnNegResult = TRUE,
  tolN = -0.001,
  var = "value",
  with = NULL,
  by = NULL,
  constant = 0,
  negInput_option = 0,
  allCols = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{series_df}{(mandatory)

Data frame (object of class "data.frame") that contains the indicator time series data to be
benchmarked. In addition to the series data variable(s), specified with argument \code{var}, the data
frame must also contain two numeric variables, \code{year} and \code{period}, identifying the periods of
the indicator time series.}

\item{benchmarks_df}{(mandatory)

Data frame (object of class "data.frame") that contains the benchmarks. In addition to the benchmarks
data variable(s), specified with argument \code{with}, the data frame must also contain four numeric
variables, \code{startYear}, \code{startPeriod}, \code{endYear} and \code{endPeriod}, identifying the indicator time
series periods covered by each benchmark.}

\item{rho}{(mandatory)

Real number in the \eqn{[0,1]} interval that specifies the value of the autoregressive parameter
\eqn{\rho}. See section \strong{Details} for more information on the effect of parameter \eqn{\rho}.}

\item{lambda}{(mandatory)

Real number, with suggested values in the \eqn{[-3,3]} interval, that specifies the value of the
adjustment model parameter \eqn{\lambda}. Typical values are \code{lambda = 0.0} for an additive model
and \code{lambda = 1.0} for a proportional model.}

\item{biasOption}{(mandatory)

Specification of the bias estimation option:
\itemize{
\item \code{1}: Do not estimate the bias. The bias used to correct the indicator series will be the value
specified with argument \code{bias}.
\item \code{2}: Estimate the bias, display the result, but do not use it. The bias used to correct the indicator
series will be the value specified with argument \code{bias}.
\item \code{3}: Estimate the bias, display the result and use the estimated bias to correct the indicator series.
Any value specified with argument \code{bias} will be ignored.
}

Argument \code{biasOption} is ignored when \code{rho = 1.0}. See section \strong{Details} for more information on the bias.}

\item{bias}{(optional)

Real number, or \code{NA}, specifying the value of the user-defined bias to be used for the correction of the
indicator series prior to benchmarking. The bias is added to the indicator series with an additive model
(argument  \code{lambda = 0.0}) while it is multiplied otherwise (argument \code{lambda != 0.0}). No bias
correction is applied when \code{bias = NA}, which is equivalent to specifying \code{bias = 0.0} when \code{lambda = 0.0}
and \code{bias = 1.0} otherwise. Argument \code{bias} is ignored when \code{biasOption = 3} or \code{rho = 1.0}. See section
\strong{Details} for more information on the bias.

\strong{Default value} is \code{bias = NA} (no user-defined bias).}

\item{low_freq_periodicity}{(optional)

Positive integer representing the number of periods defining the \emph{low} (e.g., benchmarks) frequency for adding
the extra spline knots (before the first benchmark and after the last benchmark). For example, \code{low_freq_periodicity = 3}
with monthly indicators would define quarterly knots. Annual knots are added when \code{low_freq_periodicity = NA}.

\strong{Default value} is \code{low_freq_periodicity = NA} (annual knots).}

\item{n_low_freq_proj}{(optional)

Nonnegative integer representing the number of low frequency knots (as defined with argument
\code{low_freq_periodicity}) to add at both ends (before the first benchmark and after the last benchmark)
before starting to add \emph{high} (indicator series) frequency knots.

\strong{Default value} is \code{n_low_freq_proj = 1}.}

\item{proj_knots_rho_bd}{(optional)

Bound that applies to the value specified with argument \code{rho} and determines the type of extra knots to be added at
both ends (before the first benchmark and after the last benchmark). When \code{rho > proj_knots_rho_bd}, \emph{high} (indicator
series) frequency knots are used right away. Otherwise, when \code{rho <= proj_knots_rho_bd}, \emph{low} frequency knots (see
arguments \code{low_freq_periodicity} and \code{n_low_freq_proj}) are first projected on either side. Note that for quarterly
stocks, the cube of the specified \code{proj_knots_rho_bd} value is actually used. Therefore, the value for argument
\code{proj_knots_rho_bd} should correspond to monthly stock indicators; it is internally adjusted for quarterly stocks.
This argument aims at reaching a compromise for the set periods outside (before or after) the provided benchmarks
(anchor points), i.e., Denton-type (straight line) adjustments as \code{rho} approaches 1 (when \code{rho > proj_knots_rho_bd})
and a natural looking (not overly contorted) spline otherwise (when \code{rho <= proj_knots_rho_bd}). Section \strong{Details}
contains more information on this subject and some illustrative cases are provided in section \strong{Examples}.

\strong{Default value} is \code{proj_knots_rho_bd = 0.995} (\eqn{0.995^3} for quarterly stock indicators).}

\item{tolV, tolP}{(optional)

Nonnegative real number, or \code{NA}, specifying the tolerance, in absolute value or percentage, to be used
for the validation of the output binding benchmarks (alterability coefficient of \eqn{0.0}). This validation
compares the input binding benchmark values with the equivalent values calculated from the benchmarked series
(output) data. Arguments \code{tolV} and \code{tolP} cannot be both specified (one must be specified while the other
must be \code{NA}).

\strong{Example:} to set a tolerance of 10 \emph{units}, specify \verb{tolV = 10, tolP = NA}; to set a tolerance of 1\%,
specify \verb{tolV = NA, tolP = 0.01}.

\strong{Default values} are \code{tolV = 0.001} and \code{tolP = NA}.}

\item{warnNegResult}{(optional)

Logical argument specifying whether a warning message is generated when a negative value created by the
function in the benchmarked (output) series is smaller than the threshold specified by argument \code{tolN}.

\strong{Default value} is \code{warnNegResult = TRUE}.}

\item{tolN}{(optional)

Negative real number specifying the threshold for the identification of negative values. A value is
considered negative when it is smaller than this threshold.

\strong{Default value} is \code{tolN = -0.001}.}

\item{var}{(optional)

String vector (minimum length of 1) specifying the variable name(s) in the indicator series data frame
(argument \code{series_df}) containing the values and (optionally) the user-defined alterability coefficients
of the series to be benchmarked. These variables must be numeric.

The syntax is \code{var = c("series1 </ alt_ser1>", "series2 </ alt_ser2>", ...)}. Default alterability
coefficients of \eqn{1.0} are used when a user-defined alterability coefficients variable is not specified
alongside an indicator series variable. See section \strong{Details} for more information on alterability coefficients.

\strong{Example:} \code{var = "value / alter"} would benchmark indicator series data frame variable \code{value} with the
alterability coefficients contained in variable \code{alter} while \code{var = c("value / alter", "value2")} would
additionally benchmark variable \code{value2} with default alterability coefficients of \eqn{1.0}.

\strong{Default value} is \code{var = "value"} (benchmark variable \code{value} using default alterability coefficients of
\eqn{1.0}).}

\item{with}{(optional)

String vector (same length as argument \code{var}), or \code{NULL}, specifying the variable name(s) in the benchmarks data
frame (argument \code{benchmarks_df}) containing the values and (optionally) the user-defined alterability coefficients
of the benchmarks. These variables must be numeric. Specifying \code{with = NULL} results in using benchmark variable(s)
with the same names(s) as those specified with argument \code{var} without user-defined benchmark alterability coefficients
(i.e., default alterability coefficients of \eqn{0.0} corresponding to binding benchmarks).

The syntax is \code{with = NULL} or \code{with = c("bmk1 </ alt_bmk1>", "bmk2 </ alt_bmk2>", ...)}. Default alterability
coefficients of \eqn{0.0} (binding benchmarks) are used when a user-defined alterability coefficients variable
is not specified alongside a benchmark variable. See section \strong{Details} for more information on alterability
coefficients.

\strong{Example:} \code{with = "val_bmk"} would use benchmarks data frame variable \code{val_bmk} with default benchmark
alterability coefficients of \eqn{0.0} to benchmark the indicator series while
\code{with = c("val_bmk", "val_bmk2 / alt_bmk2")} would additionally benchmark a second indicator series using
benchmark variable \code{val_bmk2} with the benchmark alterability coefficients contained in variable \code{alt_bmk2}.

\strong{Default value} is \code{with = NULL} (same benchmark variable(s) as argument \code{var} using default benchmark
alterability coefficients of \eqn{0.0}).}

\item{by}{(optional)

String vector (minimum length of 1), or \code{NULL}, specifying the variable name(s) in the input data frames
(arguments \code{series_df} and \code{benchmarks_df}) to be used to form groups (for \emph{BY-group} processing) and allow
the benchmarking of multiple series in a single function call. BY-group variables can be numeric or
character (factors or not), must be present in both input data frames and will appear in all three output
data frames (see section \strong{Value}). BY-group processing is not  implemented when \code{by = NULL}. See
"Benchmarking Multiple Series" in section \strong{Details} for more information.

\strong{Default value} is \code{by = NULL} (no BY-group processing).}

\item{constant}{(optional)

Real number that specifies a value to be temporarily added to both the indicator series and the
benchmarks before solving proportional benchmarking problems (\code{lambda != 0.0}). The temporary
constant is removed from the final output benchmarked series. E.g., specifying a (small) constant
would allow proportional benchmarking with \code{rho = 1} (e.g., proportional Denton benchmarking) on
indicator series that include values of 0. Otherwise, proportional benchmarking with values of 0 in
the indicator series is only possible when \code{rho < 1}. Specifying a constant with additive benchmarking
(\code{lambda = 0.0}) has no impact on the resulting benchmarked data. The data variables in the \code{graphTable}
output data frame include the constant, corresponding to the benchmarking problem that was actually solved.

\strong{Default value} is \code{constant = 0} (no temporary additive constant).}

\item{negInput_option}{(optional)

Handling of negative values in the input data for proportional benchmarking (\code{lambda != 0.0}):
\itemize{
\item \code{0}: Do not allow negative values with proportional benchmarking. An error message is displayed in the
presence of negative values in the input indicator series or benchmarks and missing (\code{NA}) values are
returned for the benchmarked series. This corresponds to the G-Series 2.0 behaviour.
\item \code{1}: Allow negative values with proportional benchmarking but display a warning message.
\item \code{2}: Allow negative values with proportional benchmarking without displaying any message.
}

\strong{Default value} is \code{negInput_option = 0} (do not allow negative values with proportional benchmarking).}

\item{allCols}{(optional)

Logical argument specifying whether all variables in the indicator series data frame (argument \code{series_df}),
other than \code{year} and \code{period}, determine the set of series to benchmark. Values specified with arguments \code{var}
and \code{with} are ignored when \code{allCols = TRUE}, which automatically implies default alterability coefficients,
and variables with the same names as the indicator series must exist in the benchmarks data frame (argument
\code{benchmarks_df}).

\strong{Default value} is \code{allCols = FALSE}.}

\item{quiet}{(optional)

Logical argument specifying whether or not to display only essential information such as warning messages,
error messages and variable (series) or BY-group information when multiple series are benchmarked in a single call
to the function. We advise against \emph{wrapping} your \code{\link[=benchmarking]{benchmarking()}} call with \code{\link[=suppressMessages]{suppressMessages()}} to further suppress
the display of variable (series) or BY-group information when processing multiple series as this would make
troubleshooting difficult in case of issues with individual series. Note that specifying \code{quiet = TRUE} would
also \emph{nullify} argument \code{verbose}.

\strong{Default value} is \code{quiet = FALSE}.}
}
\value{
The function returns is a list of four data frames:
\itemize{
\item \code{series}: data frame containing the benchmarked data (primary function output). BY-group variables
specified with argument \code{by} would be included in the data frame but not alterability coefficient variables
specified with argument \code{var}.
\item \code{benchmarks}: copy of the input benchmarks data frame (excluding invalid benchmarks when applicable).
BY-group variables specified with argument \code{by} would be included in the data frame but not alterability
coefficient variables specified with argument \code{with}.
\item \code{graphTable}: data frame containing supplementary data useful to produce analytical tables and graphs
(see function \code{\link[=plot_graphTable]{plot_graphTable()}}). It contains the following variables in addition to the BY-group variables
specified with argument \code{by}:
\itemize{
\item \code{varSeries}: Name of the indicator series variable
\item \code{varBenchmarks}: Name of the benchmark variable
\item \code{altSeries}: Name of the user-defined indicator series alterability coefficients variable
\item \code{altSeriesValue}: Indicator series alterability coefficients
\item \code{altbenchmarks}: Name of the user-defined benchmark alterability coefficients variable
\item \code{altBenchmarksValue}: Benchmark alterability coefficients
\item \code{t}: Indicator series period identifier (1 to \eqn{T})
\item \code{m}: Benchmark coverage periods identifier (1 to \eqn{M})
\item \code{year}: Data point calendar year
\item \code{period}: Data point period (cycle) value (1 to \code{periodicity})
\item \code{rho}: Autoregressive parameter \eqn{\rho} (argument \code{rho})
\item \code{lambda}: Adjustment model parameter \eqn{\lambda} (argument \code{lambda})
\item \code{bias}: Bias adjustment (default, user-defined or estimated bias according to arguments \code{biasOption} and \code{bias})
\item \code{periodicity}: The maximum number of periods in a year (e.g. 4 for a quarterly indicator series)
\item \code{date}: Character string combining the values of variables \code{year} and \code{period}
\item \code{subAnnual}: Indicator series values
\item \code{benchmarked}: Benchmarked series values
\item \code{avgBenchmark}: Benchmark values divided by the number of coverage periods
\item \code{avgSubAnnual}: Indicator series values (variable \code{subAnnual}) averaged over the benchmark coverage period
\item \code{subAnnualCorrected}: Bias corrected indicator series values
\item \code{benchmarkedSubAnnualRatio}: Difference (\eqn{\lambda = 0}) or ratio (\eqn{\lambda \ne 0}{lambda != 0}) of the values
of variables \code{benchmarked} and \code{subAnnual}
\item \code{avgBenchmarkSubAnnualRatio}: Difference (\eqn{\lambda = 0}) or ratio (\eqn{\lambda \ne 0}{lambda != 0}) of the values
of variables \code{avgBenchmark} and \code{avgSubAnnual}
\item \code{growthRateSubAnnual}: Period to period difference (\eqn{\lambda = 0}) or relative difference (\eqn{\lambda \ne 0}{
  lambda != 0}) of the indicator series values (variable \code{subAnnual})
\item \code{growthRateBenchmarked}: Period to period difference (\eqn{\lambda = 0}) or relative difference (\eqn{\lambda \ne 0}{
  lambda != 0}) of the benchmarked series values (variable \code{benchmarked})
}
\item \code{splineKnots}: set of \code{x} and \code{y} coordinates (knots) used to estimate the natural cubic spline with
function \code{stats::spline()}. In addition to the original set of knots corresponding to binding benchmarks
(anchor points), extra knots are also added at the beginning and end in order to deal with the \emph{benchmarking
timeliness issue} and approximate a \emph{slope=0} spline at both ends (see section \strong{Details}). It contains the following
variables in addition to the BY-group variables specified with argument \code{by}:
\itemize{
\item \code{varSeries}: Name of the indicator series variable
\item \code{varBenchmarks}: Name of the benchmark variable
\item \code{x}: Cubic spline \code{x} coordinate
\item \code{y}: Cubic spline \code{y} coordinate
\item \code{extraKnot}: Logical value identifying the extra knots added at the beginning and end
}

Rows for which \code{extraKnot == FALSE} correspond to rows in the \code{graphTable} output data frame for which
\code{m} is not missing (not \code{NA}), with \code{x = t} and \code{y = benchmarkedSubAnnualRatio}.
}

Notes:
\itemize{
\item The output \code{benchmarks} data frame always contains the original benchmarks provided in the input
benchmarks data frame. Modified nonbinding benchmarks, when applicable, can be recovered (calculated)
from the output \code{series} data frame.
\item The function returns a \code{NULL} object if an error occurs before data processing could start. Otherwise,
if execution gets far enough so that data processing could start, then an incomplete object would be
returned in case of errors (e.g., output \code{series} data frame with \code{NA} values for the benchmarked data).
\item The function returns "data.frame" objects that can be explicitly coerced to other types of objects with
the appropriate \verb{as*()} function (e.g., \code{tibble::as_tibble()} would coerce any of them to a tibble).
}
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/stock_benchmarking.html}})}

Function specifically aimed at benchmarking stock series where the benchmarks are anchor points covering a
single period of the indicator series. Benchmarks covering more than one period of the indicator series
cannot be used with this function. Function \code{\link[=benchmarking]{benchmarking()}} should be used instead to benchmark
non-stock series (flows).

Several stock series can be benchmarked in a single function call.

Note that functions \code{\link[=stock_benchmarking]{stock_benchmarking()}} and \code{\link[=benchmarking]{benchmarking()}} mainly share the same arguments and
return the same type of object. Differences are listed below:
\itemize{
\item Argument \code{verbose} is not defined for \code{\link[=stock_benchmarking]{stock_benchmarking()}}.
\item Extra arguments defined for \code{\link[=stock_benchmarking]{stock_benchmarking()}}:
\itemize{
\item \code{low_freq_periodicity}
\item \code{n_low_freq_proj}
\item \code{proj_knots_rho_bd}
}
\item The list returned by \code{\link[=stock_benchmarking]{stock_benchmarking()}} contains an extra (fourth) data frame:
\itemize{
\item \code{splineKnots}
}
}

See section \strong{Details} for more information on the similarities and differences of functions
\code{\link[=stock_benchmarking]{stock_benchmarking()}} and \code{\link[=benchmarking]{benchmarking()}}.

\emph{A direct equivalent of \code{\link[=stock_benchmarking]{stock_benchmarking()}} does not exist in SAS\eqn{^\circledR}{®} G-Series 2.0.}
}
\details{
\subsection{Comparison with \code{\link[=benchmarking]{benchmarking()}}}{

With stock series, \code{\link[=benchmarking]{benchmarking()}} is known to produce breaks in the benchmarking adjustments at
periods corresponding to the benchmark stocks (anchor points). \code{\link[=stock_benchmarking]{stock_benchmarking()}} addresses this issue
by working directly on the benchmarking adjustments. Smooth adjustments for stocks are ensured by
estimating a \emph{slope=0} cubic spline (a spline that is flat at the end knots) going through knots
corresponding to the difference (when argument \code{lambda = 0.0}) or ratio (otherwise) between the benchmarks
(anchor points) and the corresponding indicator series values. These knots are sometimes referred to as
\emph{BI} (\emph{\strong{B}}enchmark-to-\emph{\strong{I}}ndicator) \emph{differences} or \emph{BI ratios}. Interpolations from the estimated
cubic spline then provide the adjustments for the periods between benchmarks.

Arguments \code{rho}, \code{lambda}, \code{biasOption}and \code{bias} play a similar role as in \code{\link[=benchmarking]{benchmarking()}}. However,
note that for \code{\link[=stock_benchmarking]{stock_benchmarking()}}, argument \code{rho} only affects the results for periods outside of, or
around the, first and last benchmarks and \code{lambda} only takes two values in practice: \code{lambda = 0.0} for
additive adjustments (spline interpolations where the knots are \emph{BI differences}) or \code{lambda = 1.0} for
multiplicative adjustments (spline interpolations where the knots are \emph{BI ratios}). Any nonzero value for
\code{lambda} would return the same result as \code{lambda = 1.0}. Alterability coefficients also play a similar role
as in \code{\link[=benchmarking]{benchmarking()}} and have the same default values, i.e., \eqn{1.0} for the indicator series
(nonbinding values) and \eqn{0.0} for the benchmarks (binding benchmarks). However, similar to argument
\code{lambda}, alterability coefficients in this function only take two values in practice: \eqn{0.0} for binding
values or \eqn{1.0} for nonbinding values. Any nonzero alterability coefficient would return the same result
as a coefficient of \eqn{1.0}. Another difference with \code{\link[=benchmarking]{benchmarking()}} is that user-defined
alterability coefficients are allowed even when \code{rho = 1} with \code{\link[=stock_benchmarking]{stock_benchmarking()}}. Finally, specifying a
nonbinding benchmark with \code{\link[=stock_benchmarking]{stock_benchmarking()}}  is equivalent to ignoring the benchmark entirely, as if the
benchmark was not included in the input benchmarks file. Compared to \code{\link[=benchmarking]{benchmarking()}}, this generally
translates into nonbinding benchmarks having a larger impact on the resulting benchmarked stocks.
}

\subsection{Solution around the first and last benchmarks (benchmarking \emph{timeliness issue})}{

A \emph{slope=0} spline is chosen because it conceptually corresponds to the (popular) \emph{Denton benchmarking}
approach (\code{rho = 1}). In order to provide a solution before the first benchmark and after the last benchmark
that is similar to \code{\link[=benchmarking]{benchmarking()}} when \code{rho < 1}, i.e., adjustments converging to the bias at a speed
dictated by argument \code{rho}, extra knots are added at both ends before estimating the spline. By default, one
extra low frequency (as defined with argument \code{low_freq_periodicity}) knot is added on each side (beginning
and end), i.e. one extra knot is added before the first benchmark and after the last benchmark. Then, high
(indicator series) frequency knots are added to cover the indicator series span to which is added an extra
year worth of high frequency knots. The value of all those extra knots is based on arguments \code{rho}, \code{biasOption}
and \code{bias}. This produces natural looking, smooth adjustments for periods outside of or around the first and last
benchmarks that gradually converge to the bias, similarly to \code{\link[=benchmarking]{benchmarking()}}. The number of extra low
frequency knots to be added can be modified with argument \code{n_low_freq_proj}. Using high frequency knots right
away (\code{n_low_freq_proj = 0}) would produce the same projected adjustments as \code{\link[=benchmarking]{benchmarking()}}. However,
note that this tends to produce an unnatural looking (overly contorted) spline around the first and last
benchmarks that could be substantially revised once the next benchmark is available. Using the default
\code{n_low_freq_proj = 1} generally works better. However, when \code{rho} is \emph{close to 1} (see argument \code{proj_knots_rho_bd}),
high frequency knots are immediately added on each side in order to ensure Denton-type (straight line) projected
adjustments for periods outside of the first and last benchmarks. Finally, a \emph{slope=0} cubic spline is fitted through
the (original and extra) knots. Note that in practice, the \emph{slope=0} spline is actually approximated by
replicating the value of the end knots 100 times within the following period (at a frequency corresponding
to 100 times the indicator series frequency).

A \emph{natural spline} at the original end knots (first and last benchmarks) can be approximated by specifying
a large value for argument \code{low_freq_periodicity}. The larger the value of \code{low_freq_periodicity}, the more
the cubic spline at the end knots will behave like a \emph{natural spline} (2\if{html}{\out{<sup>}}nd\if{html}{\out{</sup>}} derivative equal to 0
at the end knots, i.e., a spline that keeps a constant slope at the end knots as opposed to being flat like
a \emph{slope=0} spline).

In summary, the projected adjustments are controlled with arguments \code{rho}, \code{bias} (and \code{biasOption}),
\code{n_low_freq_proj}, \code{proj_knots_rho_bd} and \code{low_freq_periodicity}:
\itemize{
\item Default values for these arguments produce \code{\link[=benchmarking]{benchmarking()}} function-like projected adjustments
(reasonably slow convergence to the bias).
\item Smaller values of \code{rho} would generate faster convergence to the bias.
\item Specifying a user-defined bias with argument \code{bias} when \code{rho < 1} is another way to influence the shape
of the projected adjustments.
\item Specifying \code{rho = 1} produce Denton-like projected adjustments (repeated first/last adjustments without
convergence to the bias).
\item Specifying a large value for \code{low_freq_periodicity} generates projected adjustments that behave more
like a natural spline, i.e., adjustments that continue in the same direction at the first/last benchmark.
The larger the value of \code{low_freq_periodicity}, the more the projected adjustments keep on going in the
same direction before \emph{turning around}.
}
}

\subsection{Note on revisions to the benchmarking adjustments}{

\code{\link[=benchmarking]{benchmarking()}} adjustments would not be revised if all future benchmarks were to fall exactly on the
projected ones (based on the bias and value of \code{rho}) and the bias was fixed. The same could be achieved
with \code{\link[=stock_benchmarking]{stock_benchmarking()}} if \emph{enough} low (e.g., benchmarks) frequency knots were projected. The problem
with this approach, however, is that the projected adjustments may not look natural as the spline may
oscillate more than desired around the projected knots. This is clearly noticeable as \code{rho} approaches 1
and the spline oscillates around the horizontally aligned projected knots instead of being aligned in a
perfectly straight line. The default implementation of the spline around the first and last benchmarks
described previously aims at reaching a \emph{best compromise} solution:
\itemize{
\item a natural looking spline around the end knots avoiding oscillations and excessive contortions;
\item small revisions to the spline if the next benchmark is close to the projected one when \code{rho} is \emph{far enough}
from 1 (\code{rho <= proj_knots_rho_bd});
\item projected adjustments that are in a straight line (free of oscillations) as \code{rho} approaches 1 (\code{rho > proj_knots_rho_bd}).
}

Subsections \emph{Benchmarking Multiple Series}, \emph{Arguments \code{constant} and \code{negInput_option}} and \emph{Treatment
of Missing (\code{NA}) Values} at the end of the \code{\link[=benchmarking]{benchmarking()}} \strong{Details} section are also relevant for
\code{\link[=stock_benchmarking]{stock_benchmarking()}}. Consult them as necessary.

Finally, note that the cubic spline associated to the \code{\link[=stock_benchmarking]{stock_benchmarking()}} adjustments can be conveniently
plotted with \code{\link[=plot_benchAdj]{plot_benchAdj()}}. The latter is used in the \strong{Examples} to illustrate some of the topics discussed
above.
}
}
\examples{
# Quarterly stock series (same pattern repeated every year)
my_series <- ts_to_tsDF(
  ts(
    rep(c(85, 95, 125, 95), 7),
    start = c(2013, 1),
    frequency = 4
  )
)
head(my_series)

# Annual benchmarks (end-of-year stocks)
my_benchmarks <- ts_to_bmkDF(
  ts(
    c(135, 125, 155, 145, 165),
    start = 2013,
    frequency = 1
  ), 
  discrete_flag = TRUE,
  alignment = "e",
  ind_frequency = 4
)
my_benchmarks

# Benchmark using...
#   - recommended `rho` value for quarterly series (`rho = 0.729`)
#   - proportional model (`lambda = 1`)
#   - bias-corrected indicator series with the estimated bias (`biasOption = 3`)

# ... with `benchmarking()` ("Proc Benchmarking" approach)
out_PB <- benchmarking(
  my_series,
  my_benchmarks,
  rho = 0.729,
  lambda = 1,
  biasOption = 3
)

# ... with `stock_benchmarking()` ("Stock Benchmarking" approach)
out_SB <- stock_benchmarking(
  my_series,
  my_benchmarks,
  rho = 0.729,
  lambda = 1,
  biasOption = 3
)

# Compare the benchmarking adjustments of both approaches
plot_benchAdj(
  PB_graphTable = out_PB$graphTable,
  SB_graphTable = out_SB$graphTable
)


# Have you noticed how smoother the `stock_benchmarking()` adjustments are compared 
# to the `benchmarking()` ones?

# The gain in the quality of the resulting benchmarked stocks might not necessarily
# be obvious in this example
plot(
  out_SB$graphTable$t, out_SB$graphTable$benchmarked,
  type = "b", col = "red", xlab = "t", ylab = "Benchmarked Stock"
)
lines(
  out_PB$graphTable$t, out_PB$graphTable$benchmarked,
  type = "b", col = "blue"
)
legend(
  x = "topleft", bty = "n", inset = 0.05, lty = 1, pch = 1,
  col = c("red", "blue"), legend = c("out_SB", "out_PB")
)
title("Benchmarked Stock")


# What about cases where a flat indicator is used, which may happen in practice
# in absence of a good indicator of the quarterly (sub-annual) movement?
my_series2 <- my_series
my_series2$value <- 1  # flat indicator
head(my_series2)
out_PB2 <- benchmarking(
  my_series2,
  my_benchmarks,
  rho = 0.729,
  lambda = 1,
  biasOption = 3,
  quiet = TRUE  # don't show the function header
)

out_SB2 <- stock_benchmarking(
  my_series2,
  my_benchmarks,
  rho = 0.729,
  lambda = 1,
  biasOption = 3,
  quiet = TRUE  # don't show the function header
)

plot(
  out_SB2$graphTable$t, out_SB2$graphTable$benchmarked,
  type = "b", col = "red", xlab = "t", ylab = "Benchmarked Stock"
)
lines(
  out_PB2$graphTable$t, out_PB2$graphTable$benchmarked,
  type = "b", col = "blue"
)
legend(
  x = "bottomright", bty = "n", inset = 0.05, lty = 1, pch = 1,
  col = c("red", "blue"), legend = c("out_SB2", "out_PB2")
)
title("Benchmarked Stock - Flat Indicator")


# The awkwardness of the benchmarked stocks produced by `benchmarking()` suddenly
# becomes obvious. That's because the benchmarked series corresponds to the
# benchmarking adjustments when using a flat indicator (e.g., a series on 1's
# with proportional benchmarking):
plot_benchAdj(
  PB_graphTable = out_PB2$graphTable,
  SB_graphTable = out_SB2$graphTable
)


# The shortcomings of the "Proc Benchmarking" approach (function `benchmarking()`)
# with stocks is also quite noticeable in this case when looking at the resulting
# quarterly growth rates, which are conveniently produced by `plot_graphTable()`.
# Pay particular attention to the transition in the growth rates from Q4 to Q1
# every year in the generated PDF graphs.
\donttest{plot_graphTable(out_PB2$graphTable, file.path(tempdir(), "PB_stock_flat_ind.pdf"))}
\donttest{plot_graphTable(out_SB2$graphTable, file.path(tempdir(), "SB_stock_flat_ind.pdf"))}


# Illustrate approximating a natural cubic spline at the original end knots (first and 
# last benchmarks) by specifying a large `low_freq_periodicity` value.
out_SB3 <- stock_benchmarking(
  my_series,
  my_benchmarks,
  rho = 0.729,
  lambda = 1,
  biasOption = 3,
  
  # Large value to approximate a natural cubic spline
  low_freq_periodicity = 100,
  
  quiet = TRUE
)

plot_benchAdj(
  SB_graphTable = out_SB3$graphTable,
  SB_splineKnots = out_SB3$splineKnots,
  legendPos = "topleft"
)


# Illustrate "oscillations" of the cubic spline beyond the original end knots with
# Denton-type benchmarking (`rho ~ 1`) caused by using low frequency (annual) extra knots.
out_SB4 <- stock_benchmarking(
  my_series,
  my_benchmarks,
  rho = 0.999,
  lambda = 1,
  biasOption = 3,
  
  # Use 3 annual extra knots first
  n_low_freq_proj = 3,
  proj_knots_rho_bd = 1,
  
  quiet = TRUE
)

plot_benchAdj(
  SB_graphTable = out_SB4$graphTable,
  SB_splineKnots = out_SB4$splineKnots
)

# No "oscillations" with the default `proj_knots_rho_bd` value because high frequency 
# (quarterly) extra knots are used right away (`n_low_freq_proj` is ignored) since 
# `rho = 0.999` exceeds the default `proj_knots_rho_bd` value (0.995^3 for quarterly data). 
# These projected adjustments are more in line with Denton-type adjustments (straight line).
out_SB4b <- stock_benchmarking(
  my_series,
  my_benchmarks,
  rho = 0.999,
  lambda = 1,
  biasOption = 3,
  quiet = TRUE
)

plot_benchAdj(
  SB_graphTable = out_SB4b$graphTable,
  SB_splineKnots = out_SB4b$splineKnots
)


# Illustrate "contortions" of the cubic spline around the original end knots caused
# by using high frequency extra knots right away (`n_low_freq_proj = 0`), i.e., using 
# the same projected adjustments as those that would be obtained with `benchmarking()`.
#
# To exacerbate the phenomenon, we'll use monthly data (11 periods between each annual 
# benchmark compared to only 3 for quarterly data, i.e., a less constrained spline) 
# and a rather small value for `rho` (0.5 < 0.9 = recommended value for monthly data) 
# for a faster convergence to the bias of the projected adjustments.

yr_vec <- unique(my_series$year)
my_series3 <- data.frame(
  year = rep(yr_vec, each = 12),
  period = rep(1:12, length(yr_vec)),
  value = rep(1, 12 * length(yr_vec))  # flat indicator
)
my_benchmarks2 <- my_benchmarks
my_benchmarks2[c("startPeriod", "endPeriod")] <- 12

out_SB5 <- stock_benchmarking(
  my_series3,
  my_benchmarks2,
  rho = 0.5,
  lambda = 1,
  biasOption = 3,
  
  # Use monthly extra knots right away
  n_low_freq_proj = 0,
  
  quiet = TRUE
)

plot_benchAdj(
  SB_graphTable = out_SB5$graphTable,
  SB_splineKnots = out_SB5$splineKnots
)

# No excessive "contortions" around the original end knots with the default  
# `n_low_freq_proj = 1`, i.e., use 1 low frequency (annual) extra knot first.
out_SB5b <- stock_benchmarking(
  my_series3,
  my_benchmarks2,
  rho = 0.5,
  lambda = 1,
  biasOption = 3,
  quiet = TRUE
)

plot_benchAdj(
  SB_graphTable = out_SB5b$graphTable,
  SB_splineKnots = out_SB5b$splineKnots
)

# To even better highlight the potential excessive "contortions" of the cubic spline 
# when enforcing the `benchmarking()` projected adjustment (i.e., low frequency extra 
# knots right away with `n_low_freq_proj = 0`), let's plot the previous two sets of 
# adjustments on the same plot (the blue line corresponds to the `n_low_freq_proj = 0` 
# case, i.e., the `benchmarking()` projected adjustments while the red line corresponds 
# to the default `stock_benchmarking()` adjustments, i.e., `n_low_freq_proj = 1`).
plot_benchAdj(
  PB_graphTable = out_SB5$graphTable,
  SB_graphTable = out_SB5b$graphTable,
  legend = NULL
)
}
\references{
Statistics Canada (2012). "Chapter 5: Benchmarking Stock". \strong{Theory and Application of Benchmarking
(Course code 0436)}. Statistics Canada, Ottawa, Canada.
}
\seealso{
\code{\link[=benchmarking]{benchmarking()}} \code{\link[=plot_graphTable]{plot_graphTable()}} \link{bench_graphs} \code{\link[=plot_benchAdj]{plot_benchAdj()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unstack_tsDF.R
\name{unstack_tsDF}
\alias{unstack_tsDF}
\title{Reciprocal function of \code{\link[=stack_tsDF]{stack_tsDF()}}}
\usage{
unstack_tsDF(
  ts_df,
  ser_cName = "series",
  yr_cName = "year",
  per_cName = "period",
  val_cName = "value"
)
}
\arguments{
\item{ts_df}{(mandatory)

Data frame (object of class "data.frame") that contains the multivariate time series data to be \emph{unstacked}.}

\item{ser_cName}{(optional)

String specifying the name of the character variable (column) in the input time series data frame that
contains the series identifier (the time series variable names in the output data frame).

\strong{Default value} is \code{ser_cName = "series"}.}

\item{yr_cName, per_cName}{(optional)

Strings specifying the name of the numeric variables (columns) in the input time series data frame that
contain the data point year and period identifiers. These variables are \emph{transferred} to the output data
frame with the same names.

\strong{Default values} are \code{yr_cName = "year"} and \code{per_cName   = "period"}.}

\item{val_cName}{(optional)

String specifying the name of the numeric variable (column) in the input time series data frame that
contains the data point values.

\strong{Default value} is \code{val_cName = "value"}.}
}
\value{
The function returns a data frame with three or more variables:
\itemize{
\item Data point year, type numeric (see argument \code{yr_cName})
\item Data point period, type numeric (see argument \code{per_cName})
\item One time series data variable for each distinct value of the input data frame variable specified with
argument \code{ser_cName}, type numeric (see arguments \code{ser_cName} and \code{val_cName})
}

Note: the function returns a "data.frame" object than can be explicitly coerced to another type of object
with the appropriate \verb{as*()} function (e.g., \code{tibble::as_tibble()} would coerce it to a tibble).
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/unstack_tsDF.html}})}

Convert a stacked (tall) multivariate time series data frame (\code{\link[=benchmarking]{benchmarking()}} and \code{\link[=stock_benchmarking]{stock_benchmarking()}}
data format) into a non-stacked (wide) multivariate time series data frame.

This function, combined with \code{\link[=tsDF_to_ts]{tsDF_to_ts()}}, is useful to convert the benchmarked data frame returned by a call to
\code{\link[=benchmarking]{benchmarking()}} or \code{\link[=stock_benchmarking]{stock_benchmarking()}} back into a "mts" object, where multiple series were benchmarked in
\emph{BY-group} processing mode.
}
\examples{
# Proportional benchmarking for multiple (3) quarterly series processed with 
# argument `by` (in BY-group mode)

ind_vec <- c(1.9, 2.4, 3.1, 2.2, 2.0, 2.6, 3.4, 2.4, 2.3)
ind_df <- ts_to_tsDF(
  ts(
    data.frame(
      ser1 = ind_vec,
      ser2 = ind_vec * 100,
      ser3 = ind_vec * 10
    ),
    start = c(2015, 1), frequency = 4
  )
)

bmk_vec <- c(10.3, 10.2)
bmk_df <- ts_to_bmkDF(
  ts(
    data.frame(
      ser1 = bmk_vec,
      ser2 = bmk_vec * 100,
      ser3 = bmk_vec * 10
    ), 
    start = 2015, frequency = 1
  ),
  ind_frequency = 4
)

out_bench <- benchmarking(
  stack_tsDF(ind_df),
  stack_bmkDF(bmk_df),
  rho = 0.729, lambda = 1, biasOption = 3,
  by = "series",
  quiet = TRUE
)

# Initial and final (benchmarked) quarterly time series data frames
ind_df
unstack_tsDF(out_bench$series)
}
\seealso{
\code{\link[=stack_tsDF]{stack_tsDF()}} \code{\link[=tsDF_to_ts]{tsDF_to_ts()}} \code{\link[=benchmarking]{benchmarking()}} \code{\link[=stock_benchmarking]{stock_benchmarking()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-common.R
\name{gs.gInv_MP}
\alias{gs.gInv_MP}
\title{Moore-Penrose inverse}
\usage{
gs.gInv_MP(X, tol = NA)
}
\arguments{
\item{X}{(mandatory)

Matrix to invert.}

\item{tol}{(optional)

Real number that specifies the tolerance for identifying zero singular values. When \code{tol = NA} (default),
the tolerance is calculated as the product of the size (dimension) of the matrix, the norm of the matrix
(largest singular value) and the \emph{machine epsilon} (\code{.Machine$double.eps}).

\strong{Default value} is \code{tol = NA}.}
}
\value{
The Moore-Penrose (pseudo) inverse of matrix \code{X}.
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/gs.gInv_MP.html}})}

This function calculates the Moore-Penrose (pseudo) inverse of a square or rectangular matrix
using Singular Value Decomposition (SVD). It is used internally by \code{\link[=tsraking]{tsraking()}} and \code{\link[=benchmarking]{benchmarking()}}.
}
\details{
The default tolerance (argument \code{tol = NA}) is coherent with the tolerance used by the MATLAB and GNU Octave
software in their general inverse functions. In our testing, this default tolerance also produced solutions
(results) comparable to G-Series 2.0 in SAS\eqn{^\circledR}{®}.
}
\examples{
# Invertible matrix
X1 <- matrix(c(3, 2, 8, 
               6, 3, 2,
               5, 2, 4), nrow = 3, byrow = TRUE)
Y1 <- gs.gInv_MP(X1)
all.equal(Y1, solve(X1))
X1 \%*\% Y1

# Rectangular matrix
X2 <- X1[-1, ]
try(solve(X2))
X2 \%*\% gs.gInv_MP(X2)

# Non-invertible square matrix
X3 <- matrix(c(3, 0, 0, 
               0, 0, 0, 
               0, 0, 4), nrow = 3, byrow = TRUE)
try(solve(X3))
X3 \%*\% gs.gInv_MP(X3)

}
\seealso{
\code{\link[=tsraking]{tsraking()}} \code{\link[=benchmarking]{benchmarking()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsbalancing.R
\name{tsbalancing}
\alias{tsbalancing}
\title{Restore cross-sectional (contemporaneous) linear constraints}
\usage{
tsbalancing(
  in_ts,
  problem_specs_df,
  temporal_grp_periodicity = 1,
  temporal_grp_start = 1,
  osqp_settings_df = default_osqp_sequence,
  display_level = 1,
  alter_pos = 1,
  alter_neg = 1,
  alter_mix = 1,
  alter_temporal = 0,
  lower_bound = -Inf,
  upper_bound = Inf,
  tolV = 0,
  tolV_temporal = 0,
  tolP_temporal = NA,

  # New in G-Series 3.0
  validation_tol = 0.001,
  trunc_to_zero_tol = validation_tol,
  full_sequence = FALSE,
  validation_only = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{in_ts}{(mandatory)

Time series (object of class "ts" or "mts") that contains the time series data to be reconciled.
They are the balancing problems' input data (initial solutions).}

\item{problem_specs_df}{(mandatory)

Balancing problem specifications data frame (object of class "data.frame"). Using a sparse format inspired from the
SAS/OR\eqn{^\circledR}{®} LP procedure’s \emph{sparse data input format} (SAS Institute 2015), it contains only the relevant
information such as the nonzero coefficients of the balancing constraints as well as the non-default alterability
coefficients and lower/upper bounds (i.e., values that would take precedence over those defined with arguments \code{alter_pos},
\code{alter_neg}, \code{alter_mix}, \code{alter_temporal}, \code{lower_bound} and \code{upper_bound}).

The information is provided using four mandatory variables (\code{type}, \code{col}, \code{row} and \code{coef}) and one optional variable
(\code{timeVal}). An observation (a row) in the problem specs data frame either defines a label for one of the seven types of the
balancing problem elements with columns \code{type} and \code{row} (see \emph{Label definition records} below) or specifies coefficients
(numerical values) for those balancing problem elements with variables \code{col}, \code{row}, \code{coef} and \code{timeVal} (see \emph{Information
specification records} below).
\itemize{
\item \strong{Label definition records} (\code{type} is not missing (is not \code{NA}))
\itemize{
\item \code{type} (chr): reserved keyword identifying the type of problem element being defined:
\itemize{
\item \code{EQ}: equality (\eqn{=}) balancing constraint
\item \code{LE}: lower or equal (\eqn{\le}{<=}) balancing constraint
\item \code{GE}: greater or equal (\eqn{\ge}{>=}) balancing constraint
\item \code{lowerBd}: period value lower bound
\item \code{upperBd}: period value upper bound
\item \code{alter}: period values alterability coefficient
\item \code{alterTmp}: temporal total alterability coefficient
}
\item \code{row} (chr): label to be associated to the problem element (\emph{\code{type} keyword})
\item \emph{all other variables are irrelevant and should contain missing data (\code{NA} values)} \cr \cr
}
\item \strong{Information specification records} (\code{type} is missing (is \code{NA}))
\itemize{
\item \code{type} (chr): not applicable (\code{NA})
\item \code{col} (chr): series name or reserved word \verb{_rhs_} to specify a balancing constraint right-hand side (RHS) value.
\item \code{row} (chr): problem element label.
\item \code{coef} (num): problem element value:
\itemize{
\item balancing constraint series coefficient or RHS value
\item series period value lower or upper bound
\item series period value or temporal total alterability coefficient
}
\item \code{timeVal} (num): optional time value to restrict the application of series bounds or alterability coefficients
to a specific time period (or temporal group). It corresponds to the time value, as returned by \code{stats::time()}, of a given
input time series (argument \code{in_ts}) period (observation) and is conceptually equivalent to \eqn{year + (period - 1) / 
  frequency}.
}
}

Note that empty strings (\code{""} or \code{''}) for character variables are interpreted as missing (\code{NA}) by the function. Variable
\code{row} identifies the elements of the balancing problem and is the key variable that makes the link between both types of
records. The same label (\code{row}) cannot be associated with more than one type of problem element (\code{type}) and multiple labels
(\code{row}) cannot be defined for the same given type of problem element (\code{type}), except for balancing constraints (values
\code{"EQ"}, \code{"LE"} and \code{"GE"} of column \code{type}). User-friendly features of the problem specs data frame include:
\itemize{
\item The order of the observations (rows) is not important.
\item Character values (variables \code{type}, \code{row} and \code{col}) are not case sensitive (e.g., strings \code{"Constraint 1"} and
\code{"CONSTRAINT 1"} for \code{row} would be considered as the same problem element label), except when \code{col} is used to specify a
series name (a column of the input time series object) where \strong{case sensitivity is enforced}.
\item The variable names of the problem specs data frame are also not case sensitive (e.g., \code{type}, \code{Type} or \code{TYPE} are all
valid) and \code{time_val} is an accepted variable name (instead of \code{timeVal}).
}

Finally, the following table lists valid aliases for the \emph{\code{type} keywords} (type of problem element):\tabular{cl}{
   \strong{Keyword} \tab \strong{Aliases} \cr
   \code{EQ} \tab \code{==}, \code{=} \cr
   \code{LE} \tab \code{<=}, \code{<} \cr
   \code{GE} \tab \code{>=}, \code{>} \cr
   \code{lowerBd} \tab \code{lowerBound}, \code{lowerBnd}, + \emph{same terms with '_', '.' or ' ' between words} \cr
   \code{upperBd} \tab \code{upperBound}, \code{upperBnd}, + \emph{same terms with '_', '.' or ' ' between words} \cr
   \code{alterTmp} \tab \code{alterTemporal}, \code{alterTemp}, + \emph{same terms with '_', '.' or ' ' between words} \cr
}


Reviewing the \strong{Examples} should help conceptualize the balancing problem specifications data frame.}

\item{temporal_grp_periodicity}{(optional)

Positive integer defining the number of periods in temporal groups for which the totals should be preserved.
E.g., specify \code{temporal_grp_periodicity = 3} with a monthly time series for quarterly total preservation and
\code{temporal_grp_periodicity = 12} (or \code{temporal_grp_periodicity = frequency(in_ts)}) for annual total preservation.
Specifying \code{temporal_grp_periodicity = 1} (\emph{default}) corresponds to period-by-period processing without
temporal total preservation.

\strong{Default value} is \code{temporal_grp_periodicity = 1} (period-by-period processing without temporal total preservation).}

\item{temporal_grp_start}{(optional)

Integer in the [1 .. \code{temporal_grp_periodicity}] interval specifying the starting period (cycle) for temporal
total preservation. E.g., annual totals corresponding to fiscal years defined from April to March of the
following year would be specified with \code{temporal_grp_start = 4} for a monthly time series (\code{frequency(in_ts) = 12})
and \code{temporal_grp_start = 2} for a quarterly time series (\code{frequency(in_ts) = 4}). This argument has no effect
for period-by-period processing without temporal total preservation (\code{temporal_grp_periodicity = 1}).

\strong{Default value} is \code{temporal_grp_start = 1}.}

\item{osqp_settings_df}{(optional)

Data frame (object of class "data.frame") containing a sequence of OSQP settings for solving the balancing problems.
The package includes two predefined OSQP settings sequence data frames:
\itemize{
\item \link{default_osqp_sequence}: fast and effective (default);
\item \link{alternate_osqp_sequence}: geared towards precision at the expense of execution time.
}

See \code{vignette("osqp-settings-sequence-dataframe")} for more details on this topic and to see the actual contents
of these two data frames. Note that the concept of a \emph{solving sequence} with different sets of solver settings is
new in G-Series 3.0 (a single solving attempt was made in G-Series 2.0).

\strong{Default value} is \code{osqp_settings_df = default_osqp_sequence}.}

\item{display_level}{(optional)

Integer in the [0 .. 3] interval specifying the level of information to display in the console (\code{stdout()}).
Note that specifying argument \code{quiet = TRUE} would \emph{nullify} argument \code{display_level} (none of the following information
would be displayed).\tabular{lcccc}{
   \strong{Displayed information} \tab \strong{\code{0}} \tab \strong{\code{1}} \tab \strong{\code{2}} \tab \strong{\code{3}} \cr
   Function header \tab \eqn{\checkmark}{Y} \tab \eqn{\checkmark}{Y} \tab \eqn{\checkmark}{Y} \tab \eqn{\checkmark}{Y} \cr
   Balancing problem elements \tab  \tab \eqn{\checkmark}{Y} \tab \eqn{\checkmark}{Y} \tab \eqn{\checkmark}{Y} \cr
   Individual problem solving details \tab  \tab  \tab \eqn{\checkmark}{Y} \tab \eqn{\checkmark}{Y} \cr
   Individual problem results (values and constraints) \tab  \tab  \tab  \tab \eqn{\checkmark}{Y} \cr
}


\strong{Default value} is \code{display_level = 1}.}

\item{alter_pos}{(optional)

Nonnegative real number specifying the default alterability coefficient associated to the values of time series with \strong{positive}
coefficients in all balancing constraints in which they are involved (e.g., component series in aggregation table raking problems).
Alterability coefficients provided in the problem specification data frame (argument \code{problem_specs_df}) override this value.

\strong{Default value} is \code{alter_pos = 1.0} (nonbinding values).}

\item{alter_neg}{(optional)

Nonnegative real number specifying the default alterability coefficient associated to the values of time series with \strong{negative}
coefficients in all balancing constraints in which they are involved (e.g., marginal totals in aggregation table raking problems).
Alterability coefficients provided in the problem specification data frame (argument \code{problem_specs_df}) override this value.

\strong{Default value} is \code{alter_neg = 1.0} (nonbinding values).}

\item{alter_mix}{(optional)

Nonnegative real number specifying the default alterability coefficient associated to the values of time series with a mix of
\strong{positive and negative} coefficients in the balancing constraints in which they are involved. Alterability coefficients provided
in the problem specification data frame (argument \code{problem_specs_df}) override this value.

\strong{Default value} is \code{alter_mix = 1.0} (nonbinding values).}

\item{alter_temporal}{(optional)

Nonnegative real number specifying the default alterability coefficient associated to the time series temporal totals.
Alterability coefficients provided in the problem specification data frame (argument \code{problem_specs_df}) override this value.

\strong{Default value} is \code{alter_temporal = 0.0} (binding values).}

\item{lower_bound}{(optional)

Real number specifying the default lower bound for the time series values. Lower bounds provided in the problem specification
data frame (argument \ifelse{latex}{\code{problem _specs_df}}{\code{problem_specs_df}}) override this value.

\strong{Default value} is \code{lower_bound = -Inf} (unbounded).}

\item{upper_bound}{(optional)

Real number specifying the default upper bound for the time series values. Upper bounds provided in the problem specification
data frame (argument \ifelse{latex}{\code{problem _specs_df}}{\code{problem_specs_df}}) override this value.

\strong{Default value} is \code{upper_bound = Inf} (unbounded).}

\item{tolV}{(optional)

Nonnegative real number specifying the tolerance, in absolute value, for the balancing constraints right-hand side (RHS)
values:
\itemize{
\item \code{EQ} constraints: \eqn{\quad A\mathbf{x} = \mathbf{b} \quad}{Ax = b} become \eqn{\quad \mathbf{b} - \epsilon \le A\mathbf{x} 
\le \mathbf{b} + \epsilon}{b - eps <= Ax <= b + eps}
\item \code{LE} constraints: \eqn{\quad A\mathbf{x} \le \mathbf{b} \quad}{Ax <= b} become \eqn{\quad A\mathbf{x} \le \mathbf{b} + 
\epsilon}{Ax <= b + eps}
\item \code{GE} constraints: \eqn{\quad A\mathbf{x} \ge \mathbf{b} \quad}{Ax >= b} become \eqn{\quad A\mathbf{x} \ge \mathbf{b} - 
\epsilon}{Ax >= b - eps}
}

where \eqn{\epsilon}{eps} is the tolerance specified with \code{tolV}. This argument does not apply to the \emph{period value (lower
and upper) bounds} specified with arguments \code{lower_bound} and \code{upper_bound} or in the problem specs data frame (argument
\code{prob_specs_df}). I.e., \code{tolV} does not affect the time series values lower and upper bounds, unless they are specified as
\emph{balancing constraints} instead (with \code{GE} and \code{LE} constraints in the problem specs data frame).

\strong{Default value} is \code{tolV = 0.0} (no tolerance).}

\item{tolV_temporal, tolP_temporal}{(optional)

Nonnegative real number, or \code{NA}, specifying the tolerance, in percentage (\code{tolP_temporal}) or absolute value
(\code{tolV_temporal}), for the implicit temporal aggregation constraints associated to \strong{binding temporal totals}
\eqn{\left( \sum_t{x_{i,t}} = \sum_t{y_{i,t}} \right)}{(sum_t{x_{i,t}} = sum_t{y_{i,t}})}, which become:
\deqn{\sum_t{y_{i,t}} - \epsilon_\text{abs} \le \sum_t{x_{i,t}} \le \sum_t{y_{i,t}} + \epsilon_\text{abs}}{sum_t{y_{i,t}} - 
eps_abs <= sum_t{x_{i,t}} <= sum_t{y_{i,t}} + eps_abs}
or
\deqn{\sum_t{y_{i,t}} \left( 1 - \epsilon_\text{rel} \right) \le \sum_t{x_{i,t}} \le \sum_t{y_{i,t}} \left( 1 + 
\epsilon_\text{rel} \right)}{sum_t{y_{i,t}} (1 - eps_rel) <= sum_t{x_{i,t}} <= sum_t{y_{i,t}} (1 + eps_rel)}

where \eqn{\epsilon_\text{abs}}{eps_abs} and \eqn{\epsilon_\text{rel}}{eps_rel} are the absolute and percentage tolerances
specified respectively with \code{tolV_temporal} and  \code{tolP_temporal}. Both arguments cannot be specified together (one must be
specified while the other must be \code{NA}).

\strong{Example:} to set a tolerance of 10 \emph{units}, specify \ifelse{latex}{\code{tolV_temporal = 10, tolP_temporal} \code{= NA}}{
\code{tolV_temporal = 10, tolP_temporal = NA}}; to set a tolerance of 1\%, specify\verb{tolV_temporal = NA, tolP_temporal = 0.01}.

\strong{Default values} are \code{tolV_temporal = 0.0} and \code{tolP_temporal = NA} (no tolerance).}

\item{validation_tol}{(optional)

Nonnegative real number specifying the tolerance for the validation of the balancing results. The function verifies if
the final (reconciled) time series values meet the constraints, allowing for discrepancies up to the value specified with
this argument. A warning is issued as soon as one constraint is not met (discrepancy greater than \code{validation_tol}).

With constraints defined as \eqn{\mathbf{l} \le A\mathbf{x} \le \mathbf{u}}{l <= Ax <= u}, where \eqn{\mathbf{l = u}}{l = u}
for \code{EQ} constraints, \eqn{\mathbf{l} = -\infty}{l = -Inf} for \code{LE} constraints and \eqn{\mathbf{u} = \infty}{u = Inf} for
\code{GE} constraints, \strong{constraint discrepancies} correspond to \eqn{\max \left( 0, \mathbf{l} - A\mathbf{x}, A\mathbf{x} - 
\mathbf{u} \right)}{max(0, l - Ax, Ax - u)}, where constraint bounds \eqn{\mathbf{l}}{l} and \eqn{\mathbf{u}}{u} include
the tolerances, when applicable, specified with arguments \code{tolV}, \code{tolV_temporal} and \code{tolP_temporal}.

\strong{Default value} is \code{validation_tol = 0.001}.}

\item{trunc_to_zero_tol}{(optional)

Nonnegative real number specifying the tolerance, in absolute value, for replacing by zero (small) values in the output
(reconciled) time series data (output object \code{out_ts}). Specify \code{trunc_to_zero_tol = 0} to disable this \emph{truncation to zero}
process on the reconciled data. Otherwise, specify \code{trunc_to_zero_tol > 0} to replace with \eqn{0.0} any value in the
\eqn{\left[ -\epsilon, \epsilon \right]}{[-eps, eps]} interval, where \eqn{\epsilon}{eps} is the tolerance specified with
\code{trunc_to_zero_tol}.

Note that the final constraint discrepancies (see argument \code{validation_tol}) are calculated on the \emph{zero truncated}
reconciled time series values, therefore ensuring accurate validation of the actual reconciled data returned by the
function.

\strong{Default value} is \code{trunc_to_zero_tol = validation_tol}.}

\item{full_sequence}{(optional)

Logical argument specifying whether all the steps of the \emph{OSQP settings sequence data frame} should be performed or not.
See argument \code{osqp_settings_df} and \code{vignette("osqp-settings-sequence-dataframe")} for more details on this topic.

\strong{Default value} is \code{full_sequence = FALSE}.}

\item{validation_only}{(optional)

Logical argument specifying whether the function should only perform input data validation or not. When
\code{validation_only = TRUE}, the specified \emph{balancing constraints} and \emph{period value (lower and upper) bounds} constraints
are validated against the input time series data, allowing for discrepancies up to the value specified with argument
\code{validation_tol}. Otherwise, when \code{validation_only = FALSE} (default), the input data are first reconciled and the
resulting (output) data are then validated.

\strong{Default value} is \code{validation_only = FALSE}.}

\item{quiet}{(optional)

Logical argument specifying whether or not to display only essential information such as warnings, errors and the period
(or set of periods) being reconciled. You could further suppress, if desired, the display of the \emph{balancing period(s)}
information by \emph{wrapping} your \code{\link[=tsbalancing]{tsbalancing()}} call with \code{\link[=suppressMessages]{suppressMessages()}}. In that case, the \code{proc_grp_df} output
data frame can be used to identify (unsuccessful) balancing problems associated with warning messages (if any). Note that
specifying \code{quiet = TRUE} would also \emph{nullify} argument \code{display_level}.

\strong{Default value} is \code{quiet = FALSE}.}
}
\value{
The function returns is a list of seven objects:
\itemize{
\item \code{out_ts}: modified version of the input time series object (class "ts" or "mts"; see argument \code{in_ts}) with the resulting
reconciled time series values (primary function output). It can be explicitly coerced to another type of object with the
appropriate \verb{as*()} function (e.g., \code{tsibble::as_tsibble()} would coerce it to a tsibble).
\item \code{proc_grp_df}: processing group summary data frame, useful to identify problems that have succeeded or failed.
It contains one observation (row) for each balancing problem with the following columns:
\itemize{
\item \code{proc_grp} (num): processing group id.
\item \code{proc_grp_type} (chr): processing group type. Possible values are:
\itemize{
\item \code{"period"};
\item \code{"temporal group"}.
}
\item \code{proc_grp_label} (chr): string describing the processing group in the following format:
\itemize{
\item \code{"<year>-<period>"} (single periods)
\item \code{"<start year>-<start period> - <end year>-<end period>"} (temporal groups)
}
\item \code{sol_status_val}, \code{sol_status} (num, chr): solution status numerical (integer) value and description string:
\itemize{
\item \code{ 1}: \code{"valid initial solution"};
\item \code{-1}: \code{"invalid initial solution"};
\item \code{ 2}: \code{"valid polished osqp solution"};
\item \code{-2}: \code{"invalid polished osqp solution"};
\item \code{ 3}: \code{"valid unpolished osqp solution"};
\item \code{-3}: \code{"invalid unpolished osqp solution"};
\item \code{-4}: \code{"unsolvable fixed problem"} (invalid initial solution).
}
\item \code{n_unmet_con} (num): number of unmet constraints (\code{sum(prob_conf_df$unmet_flag)}).
\item \code{max_discr} (num): maximum constraint discrepancy (\code{max(prob_conf_df$discr_out)}).
\item \code{validation_tol} (num): specified tolerance for validation purposes (argument \code{validation_tol}).
\item \code{sol_type} (chr): returned solution type. Possible values are:
\itemize{
\item \code{"initial"} (initial solution, i.e., input data values);
\item \code{"osqp"} (OSQP solution).
}
\item \code{osqp_attempts} (num): number of attempts made with OSQP (depth achieved in the solving sequence).
\item \code{osqp_seqno} (num): step # of the solving sequence corresponding to the returned solution. \code{NA} when
\code{sol_type = "initial"}.
\item \code{osqp_status} (chr): OSQP status description string (\code{osqp_sol_info_df$status}). \code{NA} when \code{sol_type = "initial"}.
\item \code{osqp_polished} (logi): \code{TRUE} if the returned OSQP solution is polished (\ifelse{latex}{\code{osqp_sol_info_df 
  $status_polish = 1}}{\code{osqp_sol_info_df$status_polish = 1}}), \code{FALSE} otherwise. \code{NA} when \code{sol_type = "initial"}.
\item \code{total_solve_time} (num): total time, in seconds, of the solving sequence.
}

Column \code{proc_grp} constitutes a \emph{unique key} (distinct rows) for the data frame. Successful balancing problems
(problems with a valid solution) correspond to rows with \code{sol_status_val > 0} or, equivalently, to \code{n_unmet_con = 0}
or to \code{max_discr <= validation_tol}. The \emph{initial solution} (\code{sol_type = "initial"}) is returned only if \strong{a)}
there are no initial constraint discrepancies, \strong{b)} the problem is fixed (all values are binding) or \strong{c)} it beats
the OSQP solution (smaller total constraint discrepancies). The OSQP solving sequence is described in \ifelse{latex}{
  \code{vignette("osqp-settings -sequence-dataframe")}}{\code{vignette("osqp-settings-sequence-dataframe")}}.
\item \code{periods_df}: time periods data frame, useful to match periods to processing groups. It contains one observation
(row) for each period of the input time series object (argument \code{in_ts}) with the following columns:
\itemize{
\item \code{proc_grp} (num): processing group id.
\item \code{t} (num): time id (\code{1:nrow(in_ts)}).
\item \code{time_val} (num): time value (\code{stats::time(in_ts)}). It conceptually corresponds to \eqn{year + (period - 1) / frequency}.
}

Columns \code{t} and \code{time_val} both constitute a \emph{unique key} (distinct rows) for the data frame.
\item \code{prob_val_df}: problem values data frame, useful to analyze change diagnostics, i.e., initial vs final (reconciled)
values. It contains one observation (row) for each value involved in each balancing problem, with the following columns:
\itemize{
\item \code{proc_grp} (num): processing group id.
\item \code{val_type} (chr): problem value type. Possible values are:
\itemize{
\item \code{"period value"};
\item \code{"temporal total"}.
}
\item \code{name} (chr): time series (variable) name.
\item \code{t} (num): time id (\code{1:nrow(in_ts)}); id of the first period of the temporal group for a \emph{temporal total}.
\item \code{time_val} (num): time value (\code{stats::time(in_ts)}); value of the first period of the temporal group for a
\emph{temporal total}. It conceptually corresponds to \eqn{year + (period - 1) / frequency}.
\item \code{lower_bd}, \code{upper_bd} (num): period value bounds; always \code{-Inf} and \code{Inf} for a \emph{temporal total}.
\item \code{alter} (num): alterability coefficient.
\item \code{value_in}, \code{value_out} (num): initial and final (reconciled) values.
\item \code{dif} (num): \code{value_out - value_in}.
\item \code{rdif} (num): \code{dif / value_in}; \code{NA} if \code{value_in = 0}.
}

Columns \code{val_type + name + t} and \code{val_type + name + time_val} both constitute a \emph{unique key} (distinct rows) for the
data frame. Binding (fixed) problem values correspond to rows with \code{alter = 0} or \code{value_in = 0}. Conversely, nonbinding
(free) problem values correspond to rows with \code{alter != 0} and \code{value_in != 0}.
\item \code{prob_con_df}: problem constraints data frame, useful for troubleshooting problems that failed (identify unmet
constraints). It contains one observation (row) for each constraint involved in each balancing problem, with the following
columns:
\itemize{
\item \code{proc_grp} (num): processing group id.
\item \code{con_type} (chr): problem constraint type. Possible values are:
\itemize{
\item \code{"balancing constraint"};
\item \code{"temporal aggregation constraint"};
\item \code{"period value bounds"}.
}

While \emph{balancing constraints} are specicied by the user, the other two types of constraints (\emph{temporal aggregation
constraints} and \emph{period value bounds}) are automatically added to the problem by the function (when applicable).
\item \code{name} (chr): constraint label or time series (variable) name.
\item \code{t} (num): time id (\code{1:nrow(in_ts)}); id of the first period of the temporal group for a \emph{temporal aggregation constraint}.
\item \code{time_val} (num): time value (\code{stats::time(in_ts)}); value of the first period of the temporal group for a \emph{temporal
aggregation constraint}. It conceptually corresponds to \eqn{year + (period - 1) / frequency}.
\item \code{l}, \code{u}, \code{Ax_in}, \code{Ax_out} (num): initial and final constraint elements \eqn{\left( \mathbf{l} \le A \mathbf{x} \le 
  \mathbf{u} \right)}{(l <= Ax <= u)}.
\item \code{discr_in}, \code{discr_out} (num): initial and final constraint discrepancies \eqn{\left( \max \left( 0, \mathbf{l} - A 
  \mathbf{x}, A \mathbf{x} - \mathbf{u} \right) \right)}{(max(0, l - Ax, Ax - u))}.
\item \code{validation_tol} (num): specified tolerance for validation purposes (argument \code{validation_tol}).
\item \code{unmet_flag} (logi): \code{TRUE} if the constraint is not met (\code{discr_out > validation_tol}), \code{FALSE} otherwise.
}

Columns \code{con_type + name + t} and \code{con_type + name + time_val} both constitute a \emph{unique key} (distinct rows) for the
data frame. Constraint bounds \eqn{\mathbf{l = u}}{l = u} for \code{EQ} constraints, \eqn{\mathbf{l} = -\infty}{l = -Inf} for
\code{LE} constraints, \eqn{\mathbf{u} = \infty}{u = Inf} for \code{GE} constraints, and include the tolerances, when applicable,
specified with arguments \code{tolV}, \code{tolV_temporal} and \code{tolP_temporal}.
\item \code{osqp_settings_df}: OSQP settings data frame. It contains one observation (row) for each problem (processing group)
solved with OSQP (\code{proc_grp_df$sol_type = "osqp"}), with the following columns:
\itemize{
\item \code{proc_grp} (num): processing group id.
\item one column corresponding to each element of the list returned by the \code{osqp::GetParams()} method applied
to a \emph{OSQP solver object} (class "OSQP_Model" object as returned by \code{\link[osqp:osqp]{osqp::osqp()}}), e.g.:
\itemize{
\item Maximum iterations (\code{max_iter});
\item Primal and dual infeasibility tolerances (\code{eps_prim_inf} and \code{eps_dual_inf});
\item Solution polishing flag (\code{polishing});
\item Number of scaling iterations (\code{scaling});
\item etc.
}
\item extra settings specific to \code{\link[=tsbalancing]{tsbalancing()}}:
\itemize{
\item \code{prior_scaling} (logi): \code{TRUE} if the problem data were scaled (using the average of the free (nonbinding) problem
values as the scaling factor) prior to solving with OSQP, \code{FALSE} otherwise.
\item \code{require_polished} (logi): \code{TRUE} if a polished solution from OSQP (\ifelse{latex}{\code{osqp_sol_info_df 
    $status_polish = 1}}{\code{osqp_sol_info_df$status_polish = 1}}) was required for this step in order to end the solving
sequence, \code{FALSE} otherwise. See \code{vignette("osqp-settings-sequence-dataframe")} for more details on the solving sequence
used by \code{\link[=tsbalancing]{tsbalancing()}}.
}
}

Column \code{proc_grp} constitutes a \emph{unique key} (distinct rows) for the data frame. Visit
\url{https://osqp.org/docs/interfaces/solver_settings.html} for all available OSQP settings. Problems (processing groups) for
which the initial solution was returned (\code{proc_grp_df$sol_type = "initial"}) are not included in this data frame.
\item \code{osqp_sol_info_df}: OSQP solution information data frame. It contains one observation (row) for each problem
(processing group) solved with OSQP (\code{proc_grp_df$sol_type = "osqp"}), with the following columns:
\itemize{
\item \code{proc_grp} (num): processing group id.
\item one column corresponding to each element of the \code{info} list of a \emph{OSQP solver object} (class "OSQP_Model"
object as returned by \code{\link[osqp:osqp]{osqp::osqp()}}) after having been solved with the \code{osqp::Solve()} method, e.g.:
\itemize{
\item Solution status (\code{status} and \code{status_val});
\item Polishing status (\code{status_polish});
\item Number of iterations (\code{iter});
\item Objective function value (\code{obj_val});
\item Primal and dual residuals (\code{pri_res} and \code{dua_res});
\item Solve time (\code{solve_time});
\item etc.
}
\item extra information specific to \code{\link[=tsbalancing]{tsbalancing()}}:
\itemize{
\item \code{prior_scaling_factor} (num): value of the scaling factor when \ifelse{latex}{\code{osqp_settings_df $prior_scaling 
    = TRUE}}{\code{osqp_settings_df$prior_scaling = TRUE}} (\code{prior_scaling_factor = 1.0} otherwise).
\item \code{obj_val_ori_prob} (num): original balancing problem's objective function value, which is the OSQP objective function
value (\code{obj_val}) on the original scale (when \code{osqp_settings_df$prior_scaling = TRUE}) plus the constant term of
the original balancing problem's objective function, i.e., \ifelse{latex}{\code{obj_val_ori_prob = obj_val * 
    prior _scaling_factor + <constant term>}}{\code{obj_val_ori_prob = obj_val * prior_scaling_factor + <constant term>}},
where \verb{<constant term>} corresponds to
\eqn{\mathbf{y}^{\mathrm{T}} W \mathbf{y}}{y' W y}. See section \strong{Details} for the definition of vector
\eqn{\mathbf{y}}{y}, matrix \eqn{W} and, more generally speaking, the complete expression of the balancing problem's
objective function.
}
}

Column \code{proc_grp} constitutes a \emph{unique key} (distinct rows) for the data frame. Visit \url{https://osqp.org} for more information on
OSQP. Problems (processing groups) for which the initial solution was returned (\code{proc_grp_df$sol_type = "initial"}) are not
included in this data frame.
}

Note that the "data.frame" objects returned by the function can be explicitly coerced to other types of objects with
the appropriate \verb{as*()} function (e.g., \code{tibble::as_tibble()} would coerce any of them to a tibble).
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/tsbalancing.html}})}

\emph{Replication of the G-Series 2.0 SAS\eqn{^\circledR}{®} \strong{GSeriesTSBalancing} macro. See the
G-Series 2.0 documentation for details (Statistics Canada 2016).}

This function balances (reconciles) a system of time series according to a set of linear constraints. The balancing
solution is obtained by solving one or several quadratic minimization problems (see section \strong{Details}) with the OSQP
solver (Stellato et al. 2020). Given the feasibility of the balancing problem(s), the resulting time series data respect
the specified constraints for every time period. Linear equality and inequality constraints are allowed. Optionally, the
preservation of temporal totals may also be specified.
}
\details{
This function solves one balancing problem per processing group (see section \strong{Processing groups} for details). Each of these balancing
problems is a quadratic minimization problem of the following form:
\deqn{\displaystyle 
\begin{aligned}
& \underset{\mathbf{x}}{\text{minimize}} 
& & \mathbf{\left( y - x \right)}^{\mathrm{T}} W \mathbf{\left( y - x \right)} \\
& \text{subject to} 
& & \mathbf{l} \le A \mathbf{x} \le \mathbf{u}
\end{aligned}
}{min(x) (y - x)' W (y - x), subject to l <= Ax <= u}
where
\itemize{
\item \eqn{\mathbf{y}}{y} is the vector of the initial problem values, i.e., the initial time series period values and, when applicable,
temporal totals;
\item \eqn{\mathbf{x}}{x} is the final (reconciled) version of vector \eqn{\mathbf{y}}{y};
\item matrix \eqn{W = \mathrm{diag} \left( \mathbf{w} \right)}{W = diag(w)} with vector \eqn{\mathbf{w}}{w} elements \eqn{w_i = \left\{
    \begin{array}{cl}
      0 & \text{if } |c_i y_i| = 0 \\
      \frac{1}{|c_i y_i|} & \text{otherwise}
    \end{array} \right.
    }{w_i = 0 if |c_i y_i| = 0, w_i = 1/|c_i y_i| otherwise},
where \eqn{c_i} is the alterability coefficient of problem value \eqn{y_i} and cases corresponding to \eqn{|c_i y_i| 
    = 0} are fixed problem values (binding period values or temporal totals);
\item matrix \eqn{A} and vectors \eqn{\mathbf{l}}{l} and \eqn{\mathbf{u}}{u} specify the \emph{balancing constraints}, the \emph{implicit
temporal total aggregation constraints} (when applicable), the \emph{period value (upper and lower) bounds} as well as \emph{\eqn{x_i 
= y_i} constraints for fixed \eqn{y_i} values} \eqn{\left( \left| c_i y_i \right| = 0 \right)}{(|c_i y_i| = 0)}.
}

In practice, the objective function of the problem solved by OSQP excludes constant term \eqn{\mathbf{y}^{\mathrm{T}} W 
\mathbf{y}}{y' W y}, therefore corresponding to \eqn{\mathbf{x}^{\mathrm{T}} W \mathbf{x} - 2 \left( \mathbf{w} \mathbf{y} 
\right)^{\mathrm{T}} \mathbf{x}}{x' W x - 2 (w y)' x}, and the fixed \eqn{y_i} values \eqn{\left( \left| c_i y_i \right| = 0 
\right)}{(|c_i y_i| = 0)} are removed from the problem, adjusting the constraints accordingly, i.e.:
\itemize{
\item rows corresponding to the \emph{\eqn{x_i = y_i} constraints for fixed \eqn{y_i} values} are removed from \eqn{A}, \eqn{
\mathbf{l}}{l} and \eqn{\mathbf{u}}{u};
\item columns corresponding to fixed \eqn{y_i} values are removed from \eqn{A} while appropriately adjusting \eqn{
\mathbf{l}}{l} and \eqn{\mathbf{u}}{u}.
}
\subsection{Alterability Coefficients}{

Alterability coefficients are nonnegative numbers that change the relative cost of modifying an initial problem value.
By changing the actual objective function to minimize, they allow the generation of a wide range of solutions. Since they
appear in the denominator of the objective function (matrix \eqn{W}), the larger the alterability coefficient the less costly
it is to modify a problem value (period value or temporal total) and, conversely, the smaller the alterability coefficient
the more costly it becomes. This results in problem values with larger alterability coefficients proportionally changing more
than the ones with smaller alterability coefficients. Alterability coefficients of \eqn{0.0} define fixed (binding) problem
values while alterability coefficients greater than \eqn{0.0} define free (nonbinding) values. The default alterability
coefficients are \eqn{0.0} for temporal totals (argument \code{alter_temporal}) and \eqn{1.0} for period values (arguments
\code{alter_pos}, \code{alter_neg}, \code{alter_mix}). In the common case of aggregation table raking problems, the period values of the
marginal totals (time series with a coefficient of \eqn{-1} in the balancing constraints) are usually binding (specified
with \code{alter_neg = 0}) while the period values of the component series (time series with a coefficient of \eqn{1} in the
balancing constraints) are usually nonbinding (specified with \code{alter_pos > 0}, e.g., \code{alter_pos = 1}). \emph{Almost binding}
problem values (e.g., marginal totals or temporal totals) can be obtained in practice by specifying very small (almost
\eqn{0.0}) alterability coefficients relative to those of the other (nonbinding) problem values.

\strong{Temporal total preservation} refers to the fact that temporal totals, when applicable, are usually kept “as close as
possible” to their initial value. \emph{Pure preservation} is achieved by default with binding temporal totals while the change
is minimized with nonbinding temporal totals (in accordance with the set of alterability coefficients).
}

\subsection{Validation and troubleshooting}{

Successful balancing problems (problems with a valid solution) have \code{sol_status_val > 0} or, equivalently,
\code{n_unmet_con = 0} or \code{max_discr <= validation_tol} in the output \code{proc_grp_df} data frame. Troubleshooting
unsuccessful balancing problems is not necessarily straightforward. Following are some suggestions:
\itemize{
\item Investigate the failed constraints (\code{unmet_flag = TRUE} or, equivalently, \code{discr_out > validation_tol} in the
output \code{prob_con_df} data frame) to make sure that they do not cause an empty solution space (infeasible problem).
\item Change the OSQP solving sequence. E.g., try:
\enumerate{
\item argument \code{full_sequence = TRUE}
\item argument \code{osqp_settings_df = alternate_osqp_sequence}
\item arguments \code{osqp_settings_df = alternate_osqp_sequence} and \code{full_sequence = TRUE}
}

See \code{vignette("osqp-settings-sequence-dataframe")} for more details on this topic.
\item Increase (review) the \code{validation_tol} value. Although this may sound like \emph{cheating}, the default \code{validation_tol}
value (\eqn{1 \times 10^{-3}}) may actually be too small for balancing problems that involve very large values (e.g.,
in billions) or, conversely, too large with very small problem values (e.g, \eqn{< 1.0}). Multiplying the average scale
of the problem data by the \emph{machine tolerance} (\code{.Machine$double.eps}) gives an approximation of the average size of the
discrepancies that \code{\link[=tsbalancing]{tsbalancing()}} should be able to handle (distinguish from \eqn{0}) and should probably constitute an
\strong{absolute lower bound} for argument \code{validation_tol}. In practice, a reasonable \code{validation_tol} value would likely be
\eqn{1 \times 10^3} to \eqn{1 \times 10^6} times larger than this \emph{lower bound}.
\item Address constraints redundancy. Multi-dimensional aggregation table raking problems are over-specified (involve
redundant constraints) when all totals of all dimensions of the \emph{data cube} are binding (fixed) and a constraint is
defined for all of them. Redundancy also occurs for the implicit temporal aggregation constraints in single- or
multi-dimensional aggregation table raking problems with binding (fixed) temporal totals. Over-specification is generally
not an issue for \code{\link[=tsbalancing]{tsbalancing()}} if the input data are not contradictory with regards to the redundant constraints, i.e.,
if there are no inconsistencies (discrepancies) associated to the redundant constraints in the input data or if they are
\emph{negligible} (reasonably small relative to the scale of the problem data). Otherwise, this may lead to unsuccessful
balancing problems with \code{\link[=tsbalancing]{tsbalancing()}}. Possible solutions would then include:
\enumerate{
\item Resolve (or reduce) the discrepancies associated to the redundant constraints in the input data.
\item Select one marginal total in every dimension, but one, of the data cube and remove the corresponding balancing
constraints from the problem. \emph{This cannot be done for the implicit temporal aggregation constraints}.
\item Select one marginal total in every dimension, but one, of the data cube and make them nonbinding (alterability
coefficient of, say, \eqn{1.0}).
\item Do the same as (3) for the temporal totals of one of the inner-cube component series (make them nonbinding).
\item Make all marginal totals of every dimension, but one, of the data cube \emph{amlost binding}, i.e., specify very small
alterability coefficients (say \eqn{1 \times 10^{-6}}) compared to those of the inner-cube component series.
\item Do the same as (5) for the temporal totals of all inner-cube component series (very small alterability
coefficients, e.g., with argument \code{alter_temporal}).
\item Use \code{\link[=tsraking]{tsraking()}} (if applicable), which handles these inconsistencies by using the Moore-Penrose inverse
(uniform distribution among all binding totals).
}

Solutions (2) to (7) above should only be considered if the discrepancies associated to the redundant constraints
in the input data are \emph{reasonably small} as they would be distributed among the omitted or nonbinding totals with
\code{\link[=tsbalancing]{tsbalancing()}} and all binding totals with \code{\link[=tsraking]{tsraking()}}. Otherwise, one should first investigate solution (1) above.
\item Relax the bounds of the problem constraints, e.g.:
\itemize{
\item argument \code{tolV} for the balancing constraints;
\item arguments \code{tolV_temporal} and \code{tolP_temporal} for the implicit temporal aggregation constraints;
\item arguments \code{lower_bound} and \code{upper_bound}.
}
}
}
}
\section{Processing groups}{
The set of periods of a given reconciliation (raking or balancing) problem is called a \emph{processing group}
and either corresponds to:
\itemize{
\item a \strong{single period} with period-by-period processing or, when preserving temporal totals, for the individual
periods of an incomplete temporal group (e.g., an incomplete year)
\item or the \strong{set of periods of a complete temporal group} (e.g., a complete year) when preserving temporal
totals.
}

The total number of processing groups (total number of reconciliation problems) depends on the set of
periods in the input time series object (argument \code{in_ts}) and on the value of arguments
\code{temporal_grp_periodicity} and \code{temporal_grp_start}.

Common scenarios include \code{temporal_grp_periodicity = 1} (default) for period-by period processing without
temporal total preservation and \code{temporal_grp_periodicity = frequency(in_ts)} for the preservation of annual
totals (calendar years by default). Argument \code{temporal_grp_start} allows the specification of other types of
(\emph{non-calendar}) years. E.g., fiscal years starting on April correspond to \code{temporal_grp_start = 4} with monthly
data and \code{temporal_grp_start = 2} with quarterly data. Preserving quarterly totals with monthly data would
correspond to \ifelse{latex}{\code{temporal_grp _periodicity = 3}}{\code{temporal_grp_periodicity = 3}}.

By default, temporal groups covering more than a year (i.e., corresponding to \ifelse{latex}{\code{
temporal_grp _periodicity > frequency(in_ts)}}{\code{temporal_grp_periodicity > frequency(in_ts)}} start on a
year that is a multiple of \ifelse{latex}{\code{ceiling(temporal_grp _periodicity / frequency(in_ts))}}{\code{
ceiling(temporal_grp_periodicity / frequency(in_ts))}}. E.g., biennial groups corresponding to \ifelse{latex}{
\code{temporal_grp_per iodicity = 2 * frequency(in_ts)}}{\code{temporal_grp_periodicity = 2 * frequency(in_ts)}}
start on an \emph{even year} by default. This behaviour can be changed with argument \code{temporal_grp_start}. E.g., the
preservation of biennial totals starting on an \emph{odd year} instead of an \emph{even year} (default) corresponds to
\code{temporal_grp_start = frequency(in_ts) + 1} (along with \code{temporal_grp_periodicity = 2 * frequency(in_ts)}).

See the \code{\link[=gs.build_proc_grps]{gs.build_proc_grps()}} \strong{Examples} for common processing group scenarios.
}

\section{Comparing \code{\link[=tsraking]{tsraking()}} and \code{\link[=tsbalancing]{tsbalancing()}}}{
\itemize{
\item \code{\link[=tsraking]{tsraking()}} is limited to one- and two-dimensional aggregation table raking problems (with temporal
total preservation if required) while \code{\link[=tsbalancing]{tsbalancing()}} handles more general balancing problems (e.g.,
higher dimensional raking problems, nonnegative solutions, general linear equality and inequality
constraints as opposed to aggregation rules only, etc.).
\item \code{\link[=tsraking]{tsraking()}} returns the generalized least squared solution of the Dagum and Cholette regression-based
raking model (Dagum and Cholette 2006) while \code{\link[=tsbalancing]{tsbalancing()}} solves the corresponding quadratic
minimization problem using a numerical solver. In most cases, \emph{convergence to the minimum} is achieved
and the \code{\link[=tsbalancing]{tsbalancing()}} solution matches the (exact) \code{\link[=tsraking]{tsraking()}} least square solution. It may not be
the case, however, if convergence could not be achieved after a reasonable number of iterations.
Having said that, only in very rare occasions will the \code{\link[=tsbalancing]{tsbalancing()}} solution \emph{significantly} differ
from the \code{\link[=tsraking]{tsraking()}} solution.
\item \code{\link[=tsbalancing]{tsbalancing()}} is usually faster than \code{\link[=tsraking]{tsraking()}}, especially for large raking problems, but is
generally more sensitive to the presence of (small) inconsistencies in the input data associated to
the redundant constraints of fully specified (over-specified) raking problems. \code{\link[=tsraking]{tsraking()}} handles
these inconsistencies by using the Moore-Penrose inverse (uniform distribution among all binding totals).
\item \code{\link[=tsbalancing]{tsbalancing()}} accommodates the specification of sparse problems in their reduced form. This is not
true in the case of \code{\link[=tsraking]{tsraking()}} where aggregation rules must always be fully specified since a
\emph{complete data cube} without missing data is expected as input (every single \emph{inner-cube} component
series must contribute to all dimensions of the cube, i.e., to every single \emph{outer-cube} marginal
total series).
\item Both tools handle negative values in the input data differently by default. While the solutions of
raking problems obtained from \code{\link[=tsbalancing]{tsbalancing()}} and \code{\link[=tsraking]{tsraking()}} are identical when all input data
points are positive, they will differ if some data points are negative (unless argument
\code{Vmat_option = 2} is specified with \code{\link[=tsraking]{tsraking()}}).
\item While both \code{\link[=tsbalancing]{tsbalancing()}} and \code{\link[=tsraking]{tsraking()}} allow the preservation of temporal totals, time
management is not incorporated in \code{\link[=tsraking]{tsraking()}}. For example, the construction of the processing groups
(sets of periods of each raking problem) is left to the user with \code{\link[=tsraking]{tsraking()}} and separate calls must
be submitted for each processing group (each raking problem). That's where helper function
\code{\link[=tsraking_driver]{tsraking_driver()}} comes in handy with \code{\link[=tsraking]{tsraking()}}.
\item The marginal totals (\emph{outer-cube} marginal total series) returned by \code{\link[=tsraking]{tsraking()}} are always
recalculated as the sum of the relevant component series (\emph{inner-cube} component series). The output
values of \strong{binding} marginal totals may therefore slightly differ from the input values. The
magnitude of the differences depends on the initial inconsistencies (that end up being uniformly
distributed by the Moore-Penrose inverse). This is not the case for \code{\link[=tsbalancing]{tsbalancing()}} where the values
of \strong{binding} marginal totals always remain perfectly unchanged, potentially resulting in (small)
discrepancies between the sum of the returned component series and the corresponding marginal totals.
\item \code{\link[=tsbalancing]{tsbalancing()}} returns the same set of series as the input time series object while \code{\link[=tsraking]{tsraking()}}
returns the set of series involved in the raking problem plus those specified with argument \code{id}
(which could correspond to a subset of the input series).
}
}

\examples{
###########
# Example 1: In this first example, the objective is to balance a following simple 
#            accounting table (`Profits = Revenues – Expenses`) for 5 quarters 
#            without modifying `Profits` where `Revenues >= 0` and `Expenses >= 0`.

# Problem specifications
my_specs1 <- data.frame(
  type = c(
    "EQ", rep(NA, 3), 
    "alter", NA, 
    "lowerBd", NA, NA
  ),
  col = c(
    NA, "Revenues", "Expenses", "Profits", 
    NA, "Profits", 
    NA, "Revenues", "Expenses"
  ),
  row = c(
    rep("Accounting Rule", 4), 
    rep("Alterability Coefficient", 2), 
    rep("Lower Bound", 3)
  ),
  coef = c(
    NA, 1, -1, -1,
    NA, 0,
    NA, 0, 0
  )
)
my_specs1

# Problem data
my_series1 <- ts(
  matrix(
    c( 15,  10,  10,
        4,   8,  -1,
      250, 250,   5,
        8,  12,   0,
        0,  45, -55
    ),
    ncol = 3,
    byrow = TRUE
  ),
  start = c(2022, 1),
  frequency = 4,
  names = c("Revenues", "Expenses", "Profits")
)

# Reconcile the data
out_balanced1 <- tsbalancing(
  in_ts = my_series1,
  problem_specs_df = my_specs1,
  display_level = 3
)

# Initial data
my_series1

# Reconciled data
out_balanced1$out_ts

# Check for invalid solutions
any(out_balanced1$proc_grp_df$sol_status_val < 0)

# Display the maximum output constraint discrepancies
out_balanced1$proc_grp_df[, c("proc_grp_label", "max_discr")]


# The solution returned by `tsbalancing()` corresponds to equal proportional changes 
# (pro-rating) and is related to the default alterability coefficients of 1. Equal 
# absolute changes could be obtained instead by specifying alterability coefficients 
# equal to the inverse of the initial values. 
#
# Let’s do this for the processing group 2022Q2 (`timeVal = 2022.25`), with the default 
# displayed level of information (`display_level = 1`). 

my_specs1b <- rbind(
  cbind(my_specs1,  data.frame(timeVal = rep(NA_real_, nrow(my_specs1)))),
  data.frame(
    type = rep(NA, 2),
    col = c("Revenues", "Expenses"),
    row = rep("Alterability Coefficient", 2),
    coef = c(0.25, 0.125),
    timeVal = rep(2022.25, 2)
  )
)
my_specs1b

out_balanced1b <- tsbalancing(
  in_ts = my_series1,
  problem_specs_df = my_specs1b
)

# Display the initial 2022Q2 values and both solutions
cbind(
  data.frame(Status = c("initial", "pro-rating", "equal change")),
  rbind(
    as.data.frame(my_series1[2, , drop = FALSE]), 
    as.data.frame(out_balanced1$out_ts[2, , drop = FALSE]),
    as.data.frame(out_balanced1b$out_ts[2, , drop = FALSE])
  ),
  data.frame(
    Accounting_discr = c(
      my_series1[2, 1] - my_series1[2, 2] - my_series1[2, 3],
      out_balanced1$out_ts[2, 1] - out_balanced1$out_ts[2, 2] - 
        out_balanced1$out_ts[2, 3],
      out_balanced1b$out_ts[2, 1] - out_balanced1b$out_ts[2, 2] - 
        out_balanced1b$out_ts[2, 3]
    ),
    RelChg_Rev = c(
      NA, 
      out_balanced1$out_ts[2, 1] / my_series1[2, 1] - 1,
      out_balanced1b$out_ts[2, 1] / my_series1[2, 1] - 1
    ),
    RelChg_Exp = c(
      NA, 
      out_balanced1$out_ts[2, 2] / my_series1[2, 2] - 1,
      out_balanced1b$out_ts[2, 2] / my_series1[2, 2] - 1
    ),
    AbsChg_Rev = c(
      NA, 
      out_balanced1$out_ts[2, 1] - my_series1[2, 1],
      out_balanced1b$out_ts[2, 1] - my_series1[2, 1]
    ),
    AbsChg_Exp = c(
      NA, 
      out_balanced1$out_ts[2, 2] - my_series1[2, 2],
      out_balanced1b$out_ts[2, 2] - my_series1[2, 2]
    )
  )
)


###########
# Example 2: In this second example, consider the simulated data on quarterly 
#            vehicle sales by region (West, Centre and East), along with a national 
#            total for the three regions, and by type of vehicles (cars, trucks and 
#            a total that may include other types of vehicles). The input data correspond 
#            to directly seasonally adjusted data that have been benchmarked to the 
#            annual totals of the corresponding unadjusted time series data as part 
#            of the seasonal adjustment process (e.g., with the FORCE spec in the 
#            X-13ARIMA-SEATS software). 
#
#            The objective is to reconcile the regional sales to the national sales 
#            without modifying the latter while ensuring that the sum of the sales of 
#            cars and trucks do not exceed 95\% of the sales for all types of vehicles 
#            in any quarter. For illustrative purposes, we assume that the sales of 
#            trucks in the Centre region for the 2nd quarter of 2022 cannot be modified.

# Problem specifications
my_specs2 <- data.frame(
  
  type = c(
    "EQ", rep(NA, 4),
    "EQ", rep(NA, 4),
    "EQ", rep(NA, 4),
    "LE", rep(NA, 3),
    "LE", rep(NA, 3),
    "LE", rep(NA, 3),
    "alter", rep(NA, 4)
  ),
  
  col = c(
    NA, "West_AllTypes", "Centre_AllTypes", "East_AllTypes", "National_AllTypes", 
    NA, "West_Cars", "Centre_Cars", "East_Cars", "National_Cars", 
    NA, "West_Trucks", "Centre_Trucks", "East_Trucks", "National_Trucks", 
    NA, "West_Cars", "West_Trucks", "West_AllTypes", 
    NA, "Centre_Cars", "Centre_Trucks", "Centre_AllTypes", 
    NA, "East_Cars", "East_Trucks", "East_AllTypes",
    NA, "National_AllTypes", "National_Cars", "National_Trucks", "Centre_Trucks"
  ),
  
  row = c(
    rep("National Total - All Types", 5),
    rep("National Total - Cars", 5),
    rep("National Total - Trucks", 5),
    rep("West Region Sum", 4),
    rep("Center Region Sum", 4),
    rep("East Region Sum", 4),
    rep("Alterability Coefficient", 5)
  ),
  
  coef = c(
    NA, 1, 1, 1, -1,
    NA, 1, 1, 1, -1,
    NA, 1, 1, 1, -1,
    NA, 1, 1, -.95,
    NA, 1, 1, -.95,
    NA, 1, 1, -.95,
    NA, 0, 0, 0, 0
  ),
  
  time_val = c(rep(NA, 31), 2022.25)
)

# Beginning and end of the specifications data frame
head(my_specs2, n = 10)
tail(my_specs2)

# Problem data
my_series2 <- ts(
  matrix(
    c(
      43, 49, 47, 136, 20, 18, 12, 53, 20, 22, 26, 61,
      40, 45, 42, 114, 16, 16, 19, 44, 21, 26, 21, 59,
      35, 47, 40, 133, 14, 15, 16, 50, 19, 25, 19, 71,
      44, 44, 45, 138, 19, 20, 14, 52, 21, 18, 27, 74,
      46, 48, 55, 135, 16, 15, 19, 51, 27, 25, 28, 54
    ),
    ncol = 12,
    byrow = TRUE
  ),
  start = c(2022, 1),
  frequency = 4,
  names = c(
    "West_AllTypes", "Centre_AllTypes", "East_AllTypes", "National_AllTypes", 
    "West_Cars", "Centre_Cars", "East_Cars", "National_Cars", 
    "West_Trucks", "Centre_Trucks", "East_Trucks", "National_Trucks"
  )
)

# Reconcile without displaying the function header and enforce nonnegative data
out_balanced2 <- tsbalancing(
  in_ts                    = my_series2,
  problem_specs_df         = my_specs2,
  temporal_grp_periodicity = frequency(my_series2),
  lower_bound              = 0,
  quiet                    = TRUE
)

# Initial data
my_series2

# Reconciled data
out_balanced2$out_ts

# Check for invalid solutions
any(out_balanced2$proc_grp_df$sol_status_val < 0)

# Display the maximum output constraint discrepancies
out_balanced2$proc_grp_df[, c("proc_grp_label", "max_discr")]


###########
# Example 3: Reproduce the `tsraking_driver()` 2nd example with `tsbalancing()` 
#            (1-dimensional raking problem with annual total preservation).

# `tsraking()` metadata
my_metadata3 <- data.frame(
  series = c("cars_alb", "cars_sask", "cars_man"),
  total1 = rep("cars_tot", 3)
)
my_metadata3

# `tsbalancing()` problem specifications
my_specs3 <- rkMeta_to_blSpecs(my_metadata3)
my_specs3

# Problem data
my_series3 <- ts(
  matrix(
    c(
      14, 18, 14, 58,
      17, 14, 16, 44,
      14, 19, 18, 58,
      20, 18, 12, 53,
      16, 16, 19, 44,
      14, 15, 16, 50,
      19, 20, 14, 52,
      16, 15, 19, 51
    ),
    ncol = 4,
    byrow = TRUE
  ),
  start = c(2019, 2),
  frequency = 4,
  names = c("cars_alb", "cars_sask", "cars_man", "cars_tot")
)

# Reconcile the data with `tsraking()` (through `tsraking_driver()`)
out_raked3 <- tsraking_driver(
  in_ts = my_series3,
  metadata_df = my_metadata3,
  temporal_grp_periodicity = frequency(my_series3),
  quiet = TRUE
)

# Reconcile the data with `tsbalancing()`
out_balanced3 <- tsbalancing(
  in_ts = my_series3,
  problem_specs_df = my_specs3,
  temporal_grp_periodicity = frequency(my_series3),
  quiet = TRUE
)

# Initial data
my_series3

# Both sets of reconciled data
out_raked3
out_balanced3$out_ts

# Check for invalid `tsbalancing()` solutions
any(out_balanced3$proc_grp_df$sol_status_val < 0)

# Display the maximum output constraint discrepancies from the `tsbalancing()` solutions
out_balanced3$proc_grp_df[, c("proc_grp_label", "max_discr")]

# Confirm that both solutions (`tsraking() and `tsbalancing()`) are the same
all.equal(out_raked3, out_balanced3$out_ts)
}
\references{
Dagum, E. B. and P. Cholette (2006). \strong{Benchmarking, Temporal Distribution and Reconciliation Methods
of Time Series}. Springer-Verlag, New York, Lecture Notes in Statistics, Vol. 186. \doi{10.1007/0-387-35439-5}

Ferland, M., S. Fortier and J. Bérubé (2016). "A Mathematical Optimization Approach to Balancing Time Series:
Statistics Canada’s GSeriesTSBalancing". In \strong{JSM Proceedings, Business and Economic Statistics Section}. Alexandria,
VA: American Statistical Association. 2292-2306.

Ferland, M. (2018). "Time Series Balancing Quadratic Problem — Hessian matrix and vector of linear objective
function coefficients". \strong{Internal document}. Statistics Canada, Ottawa, Canada.

Quenneville, B. and S. Fortier (2012). "Restoring Accounting Constraints in Time Series – Methods and
Software for a Statistical Agency". \strong{Economic Time Series: Modeling and Seasonality}. Chapman & Hall, New York.

SAS Institute Inc. (2015). "The LP Procedure Sparse Data Input Format". \strong{SAS/OR\eqn{^\circledR}{®} 14.1
User's Guide: Mathematical Programming Legacy Procedures}.
\url{https://support.sas.com/documentation/cdl/en/ormplpug/68158/HTML/default/viewer.htm#ormplpug_lp_details03.htm}

Statistics Canada (2016). "The \emph{\strong{GSeriesTSBalancing}} Macro". \strong{G-Series 2.0 User Guide}.
Statistics Canada, Ottawa, Canada.

Statistics Canada (2018). \strong{Theory and Application of Reconciliation (Course code 0437)}.
Statistics Canada, Ottawa, Canada.

Stellato, B., G. Banjac, P. Goulart et al. (2020). "OSQP: an operator splitting solver for quadratic programs".
\strong{Math. Prog. Comp. 12}, 637–672 (2020). \doi{10.1007/s12532-020-00179-2}
}
\seealso{
\code{\link[=tsraking]{tsraking()}} \code{\link[=tsraking_driver]{tsraking_driver()}} \code{\link[=rkMeta_to_blSpecs]{rkMeta_to_blSpecs()}} \code{\link[=gs.build_proc_grps]{gs.build_proc_grps()}} \code{\link[=build_balancing_problem]{build_balancing_problem()}} \link{aliases}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{osqp_settings_sequence}
\alias{osqp_settings_sequence}
\alias{default_osqp_sequence}
\alias{alternate_osqp_sequence}
\title{OSQP settings sequence data frame}
\format{
A data frame with at least one row and at least one column, the \emph{most common} columns being:
\describe{
\item{max_iter}{Maximum number of iterations (integer)}
\item{sigma}{Alternating direction method of multipliers (ADMM) sigma step (double)}
\item{eps_abs}{Absolute tolerance (double)}
\item{eps_rel}{Relative tolerance (double)}
\item{eps_prim_inf}{Primal infeasibility tolerance (double)}
\item{eps_dual_inf}{Dual infeasibility tolerance (double)}
\item{polishing}{Perform solution polishing (logical)}
\item{scaling}{ Number of scaling iterations (integer)}
\item{prior_scaling}{Scale problem data prior to solving with OSQP (logical)}
\item{require_polished}{Require a polished solution to stop the sequence (logical)}
\item{[\emph{any-other-OSQP-setting}]}{Value of the corresponding OSQP setting}
}
}
\usage{
# Default sequence:
# tsbalancing(..., osqp_settings_df = default_osqp_sequence)

# Alternative (slower) sequence:
# tsbalancing(..., osqp_settings_df = alternate_osqp_sequence)

# Custom-made sequence (use with caution!):
# tsbalancing(..., osqp_settings_df = <my-osqp-sequence-dataframe>)

# Single solving attempt with the default OSQP settings (not recommended!):
# tsbalancing(..., osqp_settings_df = NULL)
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/osqp_settings_sequence.html}})}

Data frame containing a sequence of OSQP settings for \code{\link[=tsbalancing]{tsbalancing()}} specified with argument
\code{osqp_settings_df}. The package includes two predefined OSQP settings sequence data frames:
\itemize{
\item \code{\link[=osqp_settings_sequence]{default_osqp_sequence}}: fast and effective (default \code{osqp_settings_df} argument value);
\item \code{\link[=osqp_settings_sequence]{alternate_osqp_sequence}}: geared towards precision at the expense of execution time.
}

See \code{vignette("osqp-settings-sequence-dataframe")} for the actual contents of these data frames.
}
\details{
With the exception of \code{prior_scaling} and \code{require_polished}, all columns of the data frame must
correspond to a OSQP setting. Default OSQP values are used for any setting not specified in this
data frame. Visit \url{https://osqp.org/docs/interfaces/solver_settings.html} for all available OSQP
settings. Note that the OSQP \code{verbose} setting is actually controlled through \code{\link[=tsbalancing]{tsbalancing()}} arguments
\code{quiet} and \code{display_level} (i.e., column \code{verbose} in a \emph{OSQP settings sequence data frame} would be ignored).

Each row of a \emph{OSQP settings sequence data frame} represents one attempt at solving a balancing problem
with the corresponding OSQP settings. The solving sequence stops as soon as a valid solution is obtained
(a solution for which all constraint discrepancies are smaller or equal to the tolerance
specified with \code{\link[=tsbalancing]{tsbalancing()}} argument \code{validation_tol}) unless column \code{require_polished = TRUE},
in which case a polished solution from OSQP (\code{status_polish = 1}) would also be required
to stop the sequence. Constraint discrepancies correspond to \eqn{\mathrm{max}(0, l - Ax, Ax - u)}
with constraints defined as \eqn{l \le Ax \le u}{l <= Ax <= u}. In the event where a satisfactory solution
cannot be obtained after having gone through the entire sequence, \code{\link[=tsbalancing]{tsbalancing()}} returns the solution that
generated the smallest total constraint discrepancies among valid solutions, if any, or among all solutions,
otherwise. Note that running the entire solving sequence can be \emph{enforced} by specifying \code{\link[=tsbalancing]{tsbalancing()}} argument
\code{full_sequence = TRUE}. Rows with column \code{prior_scaling = TRUE} have the problem data scaled prior to
solving with OSQP, using the average of the free (nonbinding) problem values as the scaling factor.

In addition to specifying a custom-made \emph{OSQP settings sequence data frame} with argument \code{osqp_settings_df},
one can also specify \code{osqp_settings_df = NULL} which would result in a single solving attempt with default OSQP
values for all settings along with \code{prior_scaling = FALSE} and \code{require_polished = FALSE}. Note that it is
recommended, however, to first try data frames \code{default_osqp_sequence} and \code{alternate_osqp_sequence}, along with
\code{full_sequence = TRUE} if necessary, before considering other alternatives.

Vignette \emph{OSQP Settings Sequence Data Frame} (\code{vignette("osqp-settings-sequence-dataframe")}) contains additional
information.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_benchAdj.R
\name{plot_benchAdj}
\alias{plot_benchAdj}
\title{Plot benchmarking adjustments}
\usage{
plot_benchAdj(
  PB_graphTable = NULL,
  SB_graphTable = NULL,
  SB_splineKnots = NULL,
  legendPos = "bottomright"
)
}
\arguments{
\item{PB_graphTable}{(optional)

Data frame (object of class "data.frame") corresponding to the \code{\link[=benchmarking]{benchmarking()}} (PB for "Proc Benchmarking"
approach) function output \code{graphTable} data frame. Specify \code{NULL} not to include the \code{\link[=benchmarking]{benchmarking()}} adjustments
in the plot.

\strong{Default value} is \code{PB_graphTable = NULL}.}

\item{SB_graphTable}{(optional)

Data frame (object of class "data.frame") corresponding to the \code{\link[=stock_benchmarking]{stock_benchmarking()}} (SB) function output
\code{graphTable} data frame. Specify \code{NULL} not to include the \code{\link[=stock_benchmarking]{stock_benchmarking()}} adjustments in the plot.

\strong{Default value} is \code{SB_graphTable = NULL}.}

\item{SB_splineKnots}{(optional)

Data frame (object of class "data.frame") corresponding to the \code{\link[=stock_benchmarking]{stock_benchmarking()}} (SB) function output
\code{splineKnots} data frame. Specify \code{NULL} not to include the \code{\link[=stock_benchmarking]{stock_benchmarking()}} cubic spline in the plot.

\strong{Default value} is \code{SB_splineKnots = NULL}.}

\item{legendPos}{(optional)

String (keyword) specifying the location of the legend in the plot. See the description of argument \code{x} in
the documentation of \code{graphics::legend()} for the list of valid keywords. Specify \code{NULL} not to include a
legend in the plot.

\strong{Default value} is \code{legendPos = "bottomright"}.}
}
\value{
This function returns nothing (\code{invisible(NULL)}).
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/plot_benchAdj.html}})}

Plot benchmarking adjustments for a single series in the current (active) graphics device. Up to three types of
adjustments can be overlayed in the same plot:
\itemize{
\item Adjustments generated by function \code{\link[=benchmarking]{benchmarking()}}
\item Adjustments generated by function \code{\link[=stock_benchmarking]{stock_benchmarking()}}
\item Cubic spline associated to adjustments generated by function \code{\link[=stock_benchmarking]{stock_benchmarking()}}
}

These plots can be useful to assess the quality of the benchmarking results and compare the adjustments
generated by both benchmarking functions (\code{\link[=benchmarking]{benchmarking()}} and \code{\link[=stock_benchmarking]{stock_benchmarking()}}) for stock series.
}
\details{
\code{graphTable} data frame (arguments \code{PB_graphTable} and \code{SB_graphTable}) variables used in the plot:
\itemize{
\item \code{t} for the x-axis values (\emph{t})
\item \code{benchmarkedSubAnnualRatio} for the \emph{Stock Bench. (SB)} and \emph{Proc Bench. (PB)} lines
\item \code{bias} for the \emph{Bias} line (when \eqn{\rho < 1})
}

\code{splineKnots} data frame (argument \code{SB_splineKnots}) variables used in the plot:
\itemize{
\item \code{x} for the x-axis values (\emph{t})
\item \code{y} for the \emph{Cubic spline} line and the \emph{Extra knot} and \emph{Original knot} points
\item \code{extraKnot} for the type of knot (\emph{Extra knot} vs. \emph{Original knot})
}

See section \strong{Value} of \code{\link[=benchmarking]{benchmarking()}} and \code{\link[=stock_benchmarking]{stock_benchmarking()}} for more details on these data frames.
}
\examples{
#######
# Preliminary setup

# Quarterly stocks (same annual pattern repeated for 7 years)
qtr_ts <- ts(rep(c(85, 95, 125, 95), 7), start = c(2013, 1), frequency = 4)

# End-of-year stocks
ann_ts <- ts(c(135, 125, 155, 145, 165), start = 2013, frequency = 1)

# Proportional benchmarking
# ... with `benchmarking()` ("Proc Benchmarking" approach)
out_PB <- benchmarking(
  ts_to_tsDF(qtr_ts), 
  ts_to_bmkDF(ann_ts, discrete_flag = TRUE, alignment = "e", ind_frequency = 4),
  rho = 0.729, lambda = 1, biasOption = 3,
  quiet = TRUE
)
# ... with `stock_benchmarking()`
out_SB <- stock_benchmarking(
  ts_to_tsDF(qtr_ts), 
  ts_to_bmkDF(ann_ts, discrete_flag = TRUE, alignment = "e", ind_frequency = 4),
  rho = 0.729, lambda = 1, biasOption = 3,
  quiet = TRUE
)


#######
# Plot the benchmarking adjustments

# `benchmarking()` adjustments (`out_PB`), without a legend
plot_benchAdj(
  PB_graphTable = out_PB$graphTable,
  legendPos = NULL
)

# Add the `stock_benchmarking()` (`out_SB`) adjustments, with a legend this time
plot_benchAdj(
  PB_graphTable = out_PB$graphTable,
  SB_graphTable = out_SB$graphTable
)

# Add the `stock_benchmarking()` cubic spline actually used to generate the adjustments
# (incl. the extra knots at both ends), with the legend located in the top-left corner
plot_benchAdj(
  PB_graphTable = out_PB$graphTable,
  SB_graphTable = out_SB$graphTable,
  SB_splineKnots = out_SB$splineKnots,
  legendPos = "topleft"
)


#######
# Simulate multiple series benchmarking (3 stock series)

qtr_mts <- ts.union(ser1 = qtr_ts, ser2 = qtr_ts * 100, ser3 = qtr_ts * 10)
ann_mts <- ts.union(ser1 = ann_ts, ser2 = ann_ts * 100, ser3 = ann_ts * 10)

# Using argument `allCols = TRUE` (identify stocks with column `varSeries`)
out_SB2 <- stock_benchmarking(
  ts_to_tsDF(qtr_mts),
  ts_to_bmkDF(ann_mts, discrete_flag = TRUE, alignment = "e", ind_frequency = 4),
  rho = 0.729, lambda = 1, biasOption = 3,
  allCols = TRUE,
  quiet = TRUE
)

# Adjustments for 2nd stock (ser2)
plot_benchAdj(
  SB_graphTable = out_SB2$graphTable[out_SB2$graphTable$varSeries == "ser2", ]
)

# Using argument `by = "series"` (identify stocks with column `series`)
out_SB3 <- stock_benchmarking(
  stack_tsDF(ts_to_tsDF(qtr_mts)),
  stack_bmkDF(ts_to_bmkDF(
    ann_mts, discrete_flag = TRUE, alignment = "e", ind_frequency = 4)
  ),
  rho = 0.729, lambda = 1, biasOption = 3,
  by = "series",
  quiet = TRUE
)

# Cubic spline for 3nd stock (ser3)
plot_benchAdj(
  SB_splineKnots = out_SB3$splineKnots[out_SB3$splineKnots$series == "ser3", ]
)
}
\seealso{
\code{\link[=plot_graphTable]{plot_graphTable()}} \link{bench_graphs} \code{\link[=benchmarking]{benchmarking()}} \code{\link[=stock_benchmarking]{stock_benchmarking()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stack_bmkDF.R
\name{stack_bmkDF}
\alias{stack_bmkDF}
\title{Stack benchmarks data}
\usage{
stack_bmkDF(
  bmk_df,
  ser_cName = "series",
  startYr_cName = "startYear",
  startPer_cName = "startPeriod",
  endYr_cName = "endYear",
  endPer_cName = "endPeriod",
  val_cName = "value",
  keep_NA = FALSE
)
}
\arguments{
\item{bmk_df}{(mandatory)

Data frame (object of class "data.frame") that contains the multivariate benchmarks to be stacked.}

\item{ser_cName}{(optional)

String specifying the name of the character variable (column) in the output stacked data frame that will
contain the benchmark names (name of the benchmark variables in the input multivariate benchmarks data
frame). This variable can then be used as the BY-group variable (argument \code{by}) with the benchmarking functions.

\strong{Default value} is \code{ser_cName = "series"}.}

\item{startYr_cName, startPer_cName, endYr_cName, endPer_cName}{(optional)

Strings specifying the name of the numeric variables (columns) in the input multivariate benchmarks data frame
that define the benchmark coverage, i.e., the starting and ending year and period (cycle) identifiers. These variables
are \emph{transferred} to the output stacked data frame with the same variable names.

\strong{Default values} are \code{startYr_cName = "startYear"}, \code{startPer_cName = "startPeriod"}
\code{endYr_cName = "endYear"} and \code{endPer_cName   = "endPeriod"}.}

\item{val_cName}{(optional)

String specifying the name of the numeric variable (column) in the output stacked data frame that will
contain the benchmark values.

\strong{Default value} is \code{val_cName = "value"}.}

\item{keep_NA}{(optional)

Logical argument specifying whether missing (\code{NA}) benchmark values in the input multivariate benchmarks data frame
should be kept in the output stacked data frame.

\strong{Default value} is \code{keep_NA = FALSE}.}
}
\value{
The function returns a data frame with six variables:
\itemize{
\item Benchmark (series) name, type character (see argument \code{ser_cName})
\item Benchmark coverage starting year, type numeric (see argument \code{startYr_cName})
\item Benchmark coverage starting period, type numeric (see argument \code{startPer_cName})
\item Benchmark coverage ending year, type numeric (see argument \code{endtYr_cName})
\item Benchmark coverage ending period, type numeric (see argument \code{endPer_cName})
\item Benchmark value, type numeric (see argument \code{val_cName})
}

Note: the function returns a "data.frame" object than can be explicitly coerced to another type of object
with the appropriate \verb{as*()} function (e.g., \code{tibble::as_tibble()} would coerce it to a tibble).
}
\description{
\if{html,text}{(\emph{version française: 
\url{https://StatCan.github.io/gensol-gseries/fr/reference/stack_bmkDF.html}})}

Convert a multivariate benchmarks data frame (see \code{\link[=ts_to_bmkDF]{ts_to_bmkDF()}}) for the benchmarking functions
(\code{\link[=benchmarking]{benchmarking()}} and \code{\link[=stock_benchmarking]{stock_benchmarking()}}) into a stacked (tall) data frame with six variables (columns):
\itemize{
\item one (1) for the benchmark name (e.g., series name)
\item four (4) for the benchmark coverage
\item one (1) for the benchmark value
}

Missing (\code{NA}) benchmark values are not included in the output stacked data frame by default. Specify argument
\code{keep_NA = TRUE} in order to keep them.

This function is useful when intending to use the \code{by} argument (\emph{BY-group} processing mode) of the benchmarking
functions in order to benchmark multiple series in a single function call.
}
\examples{
# Create an annual benchmarks data frame for 2 quarterly indicator series 
# (with missing benchmark values for the last 2 years)
my_benchmarks <- ts_to_bmkDF(
  ts(
    data.frame(
      ser1 = c(1:3 *  10, NA, NA), 
      ser2 = c(1:3 * 100, NA, NA)
    ), 
    start = c(2019, 1), frequency = 1
  ),
  ind_frequency = 4
)
my_benchmarks


# Stack the benchmarks ...

# discarding `NA` values in the output stacked data frame (default behavior)
stack_bmkDF(my_benchmarks)

# keep `NA` values in the output stacked data frame
stack_bmkDF(my_benchmarks, keep_NA = TRUE)

# using custom variable (column) names
stack_bmkDF(my_benchmarks, ser_cName = "bmk_name", val_cName = "bmk_val")
}
\seealso{
\code{\link[=stack_tsDF]{stack_tsDF()}} \code{\link[=ts_to_bmkDF]{ts_to_bmkDF()}} \code{\link[=benchmarking]{benchmarking()}} \code{\link[=stock_benchmarking]{stock_benchmarking()}}
}
